# -*- coding: utf-8 -*-
# mypy: disable-error-code=assignment
# pylint: disable=line-too-long,consider-using-f-string
"""Application Layer Protocol Numbers
========================================

.. module:: pcapkit.const.reg.apptype

This module contains the constant enumeration for **Application Layer Protocol Numbers**,
which is automatically generated from :class:`pcapkit.vendor.reg.apptype.AppType`.

"""
from collections import defaultdict
from typing import TYPE_CHECKING

from aenum import IntFlag, StrEnum, auto, extend_enum

from pcapkit.utilities.compat import show_flag_values

__all__ = ['AppType']

if TYPE_CHECKING:
    from typing import DefaultDict, Type


class TransportProtocol(IntFlag):
    """Transport layer protocol."""

    undefined = 0

    #: Transmission Control Protocol.
    tcp = auto()
    #: User Datagram Protocol.
    udp = auto()
    #: Stream Control Transmission Protocol.
    sctp = auto()
    #: Datagram Congestion Control Protocol.
    dccp = auto()

    @staticmethod
    def get(key: 'int | str') -> 'TransportProtocol':
        """Backport support for original codes.

        Args:
            key: Key to get enum item.

        :meta private:
        """
        if isinstance(key, int):
            return TransportProtocol(key)
        if key.lower() in TransportProtocol.__members__:
            return TransportProtocol[key.lower()]  # type: ignore[misc]
        max_val = max(TransportProtocol.__members__.values())
        return extend_enum(TransportProtocol, key.lower(), max_val * 2)


class AppType(StrEnum):
    """[AppType] Application Layer Protocol Numbers"""

    if TYPE_CHECKING:
        #: Service name.
        svc: 'str'
        #: Port number.
        port: 'int'
        #: Transport protocol.
        proto: 'TransportProtocol'

    #: Mapping of members based on transport protocol.
    __members_proto__: 'DefaultDict[TransportProtocol, dict[int, AppType]]' = defaultdict(dict)

    def __new__(cls, value: 'int', name: 'str' = '<null>',
                proto: 'TransportProtocol' = TransportProtocol.undefined) -> 'Type[AppType]':
        temp = '%s [%d - %s]' % (name, value, proto.name)

        obj = str.__new__(cls, temp)
        obj._value_ = temp

        obj.svc = name
        obj.port = value
        obj.proto = proto

        for namespace in show_flag_values(proto):
            cls.__members_proto__[TransportProtocol(namespace)][value] = obj
        if proto is TransportProtocol.undefined:
            cls.__members_proto__[proto][value] = obj

        return obj

    def __repr__(self) -> 'str':
        return "<%s.%s: %d [%s]>" % (self.__class__.__name__, self.svc, self.port, self.proto.name)

    def __str__(self) -> 'str':
        return '%s [%d - %s]' % (self.svc, self.port, self.proto.name)

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_0 = 0, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TCP Port Service Multiplexer
    #: - [UDP] TCP Port Service Multiplexer
    tcpmux = 1, 'tcpmux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Management Utility
    #: - [UDP] Management Utility
    compressnet_2 = 2, 'compressnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compression Process
    #: - [UDP] Compression Process
    compressnet_3 = 3, 'compressnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_4 = 4, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Job Entry
    #: - [UDP] Remote Job Entry
    rje = 5, 'rje', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_6 = 6, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Echo
    #: - [UDP] Echo
    echo = 7, 'echo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_8 = 8, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Discard
    #: - [UDP] Discard
    #: - [SCTP] Discard [:rfc:`9260`]
    #: - [DCCP] Discard [:rfc:`4340`]
    discard = 9, 'discard', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp') | TransportProtocol.get('dccp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_10 = 10, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Active Users
    #: - [UDP] Active Users
    systat = 11, 'systat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_12 = 12, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Daytime [:rfc:`867`]
    #: - [UDP] Daytime [:rfc:`867`]
    daytime = 13, 'daytime', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_14 = 14, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Unassigned [was netstat]
    unassigned_was_netstat = 15, 'unassigned_was_netstat', TransportProtocol.get('tcp')

    #: [UDP] Unassigned
    unassigned_15 = 15, 'unassigned', TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_16 = 16, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quote of the Day
    #: - [UDP] Quote of the Day
    qotd = 17, 'qotd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Character Generator
    #: - [UDP] Character Generator
    chargen = 19, 'chargen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] File Transfer [Default Data]
    #: - [UDP] File Transfer [Default Data]
    #: - [SCTP] FTP [:rfc:`9260`]
    ftp_data = 20, 'ftp-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] File Transfer Protocol [Control] [:rfc:`959`]
    #: - [UDP] File Transfer Protocol [Control] [:rfc:`959`]
    #: - [SCTP] FTP [:rfc:`9260`]
    ftp = 21, 'ftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Telnet [:rfc:`854`]
    #: - [UDP] Telnet [:rfc:`854`]
    telnet = 23, 'telnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] any private mail system
    #: - [UDP] any private mail system
    any_private_mail_system = 24, 'any_private_mail_system', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Mail Transfer [:rfc:`5321`]
    #: - [UDP] Simple Mail Transfer [:rfc:`5321`]
    smtp = 25, 'smtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_26 = 26, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NSW User System FE
    #: - [UDP] NSW User System FE
    nsw_fe = 27, 'nsw-fe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_28 = 28, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSG ICP
    #: - [UDP] MSG ICP
    msg_icp = 29, 'msg-icp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_30 = 30, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSG Authentication
    #: - [UDP] MSG Authentication
    msg_auth = 31, 'msg-auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_32 = 32, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Display Support Protocol
    #: - [UDP] Display Support Protocol
    dsp = 33, 'dsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_34 = 34, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] any private printer server
    #: - [UDP] any private printer server
    any_private_printer_server = 35, 'any_private_printer_server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_36 = 36, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Time
    #: - [UDP] Time
    time = 37, 'time', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Resource Location Protocol
    #: - [UDP] Resource Location Protocol
    rlp = 39, 'rlp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_40 = 40, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Graphics
    #: - [UDP] Graphics
    graphics = 41, 'graphics', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Host Name Server
    #: - [UDP] Host Name Server
    name = 42, 'name', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Host Name Server
    #: - [UDP] Host Name Server
    nameserver = 42, 'nameserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Who Is
    #: - [UDP] Who Is
    nicname = 43, 'nicname', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MPM FLAGS Protocol
    #: - [UDP] MPM FLAGS Protocol
    mpm_flags = 44, 'mpm-flags', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Message Processing Module [recv]
    #: - [UDP] Message Processing Module [recv]
    mpm = 45, 'mpm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MPM [default send]
    #: - [UDP] MPM [default send]
    mpm_snd = 46, 'mpm-snd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_47 = 47, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Digital Audit Daemon
    #: - [UDP] Digital Audit Daemon
    auditd = 48, 'auditd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Login Host Protocol (TACACS)
    #: - [UDP] Login Host Protocol (TACACS)
    tacacs = 49, 'tacacs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Mail Checking Protocol
    #: - [UDP] Remote Mail Checking Protocol
    re_mail_ck = 50, 're-mail-ck', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Reserved
    reserved_51 = 51, 'reserved', TransportProtocol.get('undefined')

    #: - [TCP] XNS Time Protocol
    #: - [UDP] XNS Time Protocol
    xns_time = 52, 'xns-time', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Domain Name Server
    #: - [UDP] Domain Name Server
    domain = 53, 'domain', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XNS Clearinghouse
    #: - [UDP] XNS Clearinghouse
    xns_ch = 54, 'xns-ch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISI Graphics Language
    #: - [UDP] ISI Graphics Language
    isi_gl = 55, 'isi-gl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XNS Authentication
    #: - [UDP] XNS Authentication
    xns_auth = 56, 'xns-auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] any private terminal access
    #: - [UDP] any private terminal access
    any_private_terminal_access = 57, 'any_private_terminal_access', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XNS Mail
    #: - [UDP] XNS Mail
    xns_mail = 58, 'xns-mail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] any private file service
    #: - [UDP] any private file service
    any_private_file_service = 59, 'any_private_file_service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_60 = 60, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_61 = 61, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACA Services
    #: - [UDP] ACA Services
    acas = 62, 'acas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] whois++ IANA assigned this well-formed service name as a replacement
    #: for "whois++".
    #: - [UDP] whois++ IANA assigned this well-formed service name as a replacement
    #: for "whois++".
    whoispp = 63, 'whoispp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] whois++
    #: - [UDP] whois++
    whois = 63, 'whois++', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Communications Integrator (CI)
    #: - [UDP] Communications Integrator (CI)
    covia = 64, 'covia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TACACS-Database Service
    #: - [UDP] TACACS-Database Service
    tacacs_ds = 65, 'tacacs-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bootstrap Protocol Server [:rfc:`951`]
    #: - [UDP] Bootstrap Protocol Server
    bootps = 67, 'bootps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bootstrap Protocol Client
    #: - [UDP] Bootstrap Protocol Client
    bootpc = 68, 'bootpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trivial File Transfer
    #: - [UDP] Trivial File Transfer
    tftp = 69, 'tftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gopher
    #: - [UDP] Gopher
    gopher = 70, 'gopher', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Job Service
    #: - [UDP] Remote Job Service
    netrjs_1 = 71, 'netrjs-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Job Service
    #: - [UDP] Remote Job Service
    netrjs_2 = 72, 'netrjs-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Job Service
    #: - [UDP] Remote Job Service
    netrjs_3 = 73, 'netrjs-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Job Service
    #: - [UDP] Remote Job Service
    netrjs_4 = 74, 'netrjs-4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] any private dial out service
    #: - [UDP] any private dial out service
    any_private_dial_out_service = 75, 'any_private_dial_out_service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Distributed External Object Store
    #: - [UDP] Distributed External Object Store
    deos = 76, 'deos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] any private RJE service
    #: - [UDP] any private RJE service
    any_private_rje_service = 77, 'any_private_rje_service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vettcp
    #: - [UDP] vettcp
    vettcp = 78, 'vettcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Finger
    #: - [UDP] Finger
    finger = 79, 'finger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] World Wide Web HTTP [:rfc:`9110`]
    #: - [UDP] World Wide Web HTTP [:rfc:`9110`]
    #: - [SCTP] HTTP [:rfc:`9260`]
    http = 80, 'http', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] World Wide Web HTTP [:rfc:`9110`]
    #: - [UDP] World Wide Web HTTP [:rfc:`9110`]
    www = 80, 'www', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] World Wide Web HTTP
    #: - [UDP] World Wide Web HTTP
    www_http = 80, 'www-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_81 = 81, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] XFER Utility
    #: - [UDP] XFER Utility
    xfer = 82, 'xfer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Common Trace Facility
    #: - [UDP] Common Trace Facility
    ctf = 84, 'ctf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MIT ML Device
    #: - [UDP] MIT ML Device
    mit_ml_dev_83 = 83, 'mit-ml-dev', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MIT ML Device
    #: - [UDP] MIT ML Device
    mit_ml_dev_85 = 85, 'mit-ml-dev', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Micro Focus Cobol
    #: - [UDP] Micro Focus Cobol
    mfcobol = 86, 'mfcobol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] any private terminal link
    #: - [UDP] any private terminal link
    any_private_terminal_link = 87, 'any_private_terminal_link', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kerberos [:rfc:`4120`]
    #: - [UDP] Kerberos [:rfc:`4120`]
    kerberos = 88, 'kerberos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SU/MIT Telnet Gateway
    #: - [UDP] SU/MIT Telnet Gateway
    su_mit_tg = 89, 'su-mit-tg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNSIX Securit Attribute Token Map
    #: - [UDP] DNSIX Securit Attribute Token Map
    dnsix = 90, 'dnsix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MIT Dover Spooler
    #: - [UDP] MIT Dover Spooler
    mit_dov = 91, 'mit-dov', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Device Control Protocol
    #: - [UDP] Device Control Protocol
    dcp = 93, 'dcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tivoli Object Dispatcher
    #: - [UDP] Tivoli Object Dispatcher
    objcall = 94, 'objcall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SUPDUP
    #: - [UDP] SUPDUP
    supdup = 95, 'supdup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DIXIE Protocol Specification
    #: - [UDP] DIXIE Protocol Specification
    dixie = 96, 'dixie', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Swift Remote Virtural File Protocol
    #: - [UDP] Swift Remote Virtural File Protocol
    swift_rvf = 97, 'swift-rvf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TAC News
    #: - [UDP] TAC News
    tacnews = 98, 'tacnews', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Metagram Relay
    #: - [UDP] Metagram Relay
    metagram = 99, 'metagram', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_100 = 100, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] NIC Host Name Server
    #: - [UDP] NIC Host Name Server
    hostname = 101, 'hostname', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISO-TSAP Class 0
    #: - [UDP] ISO-TSAP Class 0
    iso_tsap = 102, 'iso-tsap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Genesis Point-to-Point Trans Net
    #: - [UDP] Genesis Point-to-Point Trans Net
    gppitnp = 103, 'gppitnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACR-NEMA Digital Imag. & Comm. 300
    #: - [UDP] ACR-NEMA Digital Imag. & Comm. 300
    acr_nema = 104, 'acr-nema', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CCSO name server protocol
    #: - [UDP] CCSO name server protocol
    cso = 105, 'cso', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mailbox Name Nameserver
    #: - [UDP] Mailbox Name Nameserver
    csnet_ns = 105, 'csnet-ns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3COM-TSMUX
    #: - [UDP] 3COM-TSMUX
    AppType_3com_tsmux = 106, '3com-tsmux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Telnet Service
    #: - [UDP] Remote Telnet Service
    rtelnet = 107, 'rtelnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNA Gateway Access Server
    #: - [UDP] SNA Gateway Access Server
    snagas = 108, 'snagas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Post Office Protocol - Version 2
    #: - [UDP] Post Office Protocol - Version 2
    pop2 = 109, 'pop2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Post Office Protocol - Version 3
    #: - [UDP] Post Office Protocol - Version 3
    pop3 = 110, 'pop3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SUN Remote Procedure Call
    #: - [UDP] SUN Remote Procedure Call
    sunrpc = 111, 'sunrpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] McIDAS Data Transmission Protocol
    #: - [UDP] McIDAS Data Transmission Protocol
    mcidas = 112, 'mcidas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    ident = 113, 'ident', TransportProtocol.get('tcp')

    #: - [TCP] Authentication Service
    #: - [UDP] Authentication Service
    auth = 113, 'auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] unassigned
    unassigned_114 = 114, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Simple File Transfer Protocol
    #: - [UDP] Simple File Transfer Protocol
    sftp = 115, 'sftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ANSA REX Notify
    #: - [UDP] ANSA REX Notify
    ansanotify = 116, 'ansanotify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UUCP Path Service
    #: - [UDP] UUCP Path Service
    uucp_path = 117, 'uucp-path', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SQL Services
    #: - [UDP] SQL Services
    sqlserv = 118, 'sqlserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network News Transfer Protocol [:rfc:`3977`]
    #: - [UDP] Network News Transfer Protocol [:rfc:`3977`]
    nntp = 119, 'nntp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CFDPTKT
    #: - [UDP] CFDPTKT
    cfdptkt = 120, 'cfdptkt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Encore Expedited Remote Pro.Call
    #: - [UDP] Encore Expedited Remote Pro.Call
    erpc = 121, 'erpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMAKYNET
    #: - [UDP] SMAKYNET
    smakynet = 122, 'smakynet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Time Protocol [:rfc:`5905`]
    #: - [UDP] Network Time Protocol [:rfc:`5905`]
    ntp = 123, 'ntp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ANSA REX Trader
    #: - [UDP] ANSA REX Trader
    ansatrader = 124, 'ansatrader', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Locus PC-Interface Net Map Ser
    #: - [UDP] Locus PC-Interface Net Map Ser
    locus_map = 125, 'locus-map', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NXEdit
    #: - [UDP] NXEdit
    nxedit = 126, 'nxedit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Locus PC-Interface Conn Server
    #: - [UDP] Locus PC-Interface Conn Server
    locus_con = 127, 'locus-con', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GSS X License Verification
    #: - [UDP] GSS X License Verification
    gss_xlicen = 128, 'gss-xlicen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Password Generator Protocol
    #: - [UDP] Password Generator Protocol
    pwdgen = 129, 'pwdgen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco FNATIVE
    #: - [UDP] cisco FNATIVE
    cisco_fna = 130, 'cisco-fna', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco TNATIVE
    #: - [UDP] cisco TNATIVE
    cisco_tna = 131, 'cisco-tna', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco SYSMAINT
    #: - [UDP] cisco SYSMAINT
    cisco_sys = 132, 'cisco-sys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Statistics Service
    #: - [UDP] Statistics Service
    statsrv = 133, 'statsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INGRES-NET Service
    #: - [UDP] INGRES-NET Service
    ingres_net = 134, 'ingres-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DCE endpoint resolution
    #: - [UDP] DCE endpoint resolution
    epmap = 135, 'epmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PROFILE Naming System
    #: - [UDP] PROFILE Naming System
    profile = 136, 'profile', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETBIOS Name Service
    #: - [UDP] NETBIOS Name Service
    netbios_ns = 137, 'netbios-ns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETBIOS Datagram Service
    #: - [UDP] NETBIOS Datagram Service
    netbios_dgm = 138, 'netbios-dgm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETBIOS Session Service
    #: - [UDP] NETBIOS Session Service
    netbios_ssn = 139, 'netbios-ssn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMFIS Data Service
    #: - [UDP] EMFIS Data Service
    emfis_data = 140, 'emfis-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMFIS Control Service
    #: - [UDP] EMFIS Control Service
    emfis_cntl = 141, 'emfis-cntl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Britton-Lee IDM
    #: - [UDP] Britton-Lee IDM
    bl_idm = 142, 'bl-idm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Internet Message Access Protocol [:rfc:`3501`][:rfc:`9051`]
    imap = 143, 'imap', TransportProtocol.get('tcp')

    #: [UDP] Reserved [:rfc:`9051`]
    reserved_143 = 143, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] UAAC Protocol
    #: - [UDP] UAAC Protocol
    uaac = 145, 'uaac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISO-IP0
    #: - [UDP] ISO-IP0
    iso_tp0 = 146, 'iso-tp0', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISO-IP
    #: - [UDP] ISO-IP
    iso_ip = 147, 'iso-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jargon
    #: - [UDP] Jargon
    jargon = 148, 'jargon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AED 512 Emulation Service
    #: - [UDP] AED 512 Emulation Service
    aed_512 = 149, 'aed-512', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle SQL*NET IANA assigned this well-formed service name as a
    #: replacement for "sql*net".
    #: - [TCP] Oracle SQL*NET
    #: - [UDP] Oracle SQL*NET IANA assigned this well-formed service name as a
    #: replacement for "sql*net".
    #: - [UDP] Oracle SQL*NET
    sql_net_66 = 66, 'sql-net', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] SQL-NET
    #: - [UDP] SQL-NET
    sql_net_150 = 150, 'sql-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HEMS
    #: - [UDP] HEMS
    hems = 151, 'hems', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Background File Transfer Program
    #: - [UDP] Background File Transfer Program
    bftp = 152, 'bftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SGMP
    #: - [UDP] SGMP
    sgmp = 153, 'sgmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETSC
    #: - [UDP] NETSC
    netsc_prod = 154, 'netsc-prod', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETSC
    #: - [UDP] NETSC
    netsc_dev = 155, 'netsc-dev', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SQL Service
    #: - [UDP] SQL Service
    sqlsrv = 156, 'sqlsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KNET/VM Command/Message Protocol
    #: - [UDP] KNET/VM Command/Message Protocol
    knet_cmp = 157, 'knet-cmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCMail Server
    #: - [UDP] PCMail Server
    pcmail_srv = 158, 'pcmail-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NSS-Routing
    #: - [UDP] NSS-Routing
    nss_routing = 159, 'nss-routing', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SGMP-TRAPS
    #: - [UDP] SGMP-TRAPS
    sgmp_traps = 160, 'sgmp-traps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNMP
    #: - [UDP] SNMP
    snmp = 161, 'snmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNMPTRAP
    #: - [UDP] SNMPTRAP
    snmptrap = 162, 'snmptrap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CMIP/TCP Manager
    #: - [UDP] CMIP/TCP Manager
    cmip_man = 163, 'cmip-man', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CMIP/TCP Agent
    #: - [UDP] CMIP/TCP Agent
    cmip_agent = 164, 'cmip-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xerox
    #: - [UDP] Xerox
    xns_courier = 165, 'xns-courier', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sirius Systems
    #: - [UDP] Sirius Systems
    s_net = 166, 's-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NAMP
    #: - [UDP] NAMP
    namp = 167, 'namp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSVD
    #: - [UDP] RSVD
    rsvd = 168, 'rsvd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SEND
    #: - [UDP] SEND
    send = 169, 'send', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network PostScript
    #: - [UDP] Network PostScript
    print_srv = 170, 'print-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Innovations Multiplex
    #: - [UDP] Network Innovations Multiplex
    multiplex = 171, 'multiplex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Innovations CL/1 IANA assigned this well-formed service name
    #: as a replacement for "cl/1".
    #: - [TCP] Network Innovations CL/1
    #: - [UDP] Network Innovations CL/1 IANA assigned this well-formed service name
    #: as a replacement for "cl/1".
    #: - [UDP] Network Innovations CL/1
    cl_1 = 172, 'cl-1', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Xyplex
    #: - [UDP] Xyplex
    xyplex_mux = 173, 'xyplex-mux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MAILQ
    #: - [UDP] MAILQ
    mailq = 174, 'mailq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VMNET
    #: - [UDP] VMNET
    vmnet = 175, 'vmnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GENRAD-MUX
    #: - [UDP] GENRAD-MUX
    genrad_mux = 176, 'genrad-mux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] X Display Manager Control Protocol
    #: - [UDP] X Display Manager Control Protocol
    xdmcp = 177, 'xdmcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NextStep Window Server
    #: - [UDP] NextStep Window Server
    nextstep = 178, 'nextstep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Border Gateway Protocol
    #: - [UDP] Border Gateway Protocol
    #: - [SCTP] BGP [:rfc:`9260`]
    bgp = 179, 'bgp', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Intergraph
    #: - [UDP] Intergraph
    ris = 180, 'ris', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unify
    #: - [UDP] Unify
    unify = 181, 'unify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unisys Audit SITP
    #: - [UDP] Unisys Audit SITP
    audit = 182, 'audit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OCBinder
    #: - [UDP] OCBinder
    ocbinder = 183, 'ocbinder', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OCServer
    #: - [UDP] OCServer
    ocserver = 184, 'ocserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote-KIS
    #: - [UDP] Remote-KIS
    remote_kis = 185, 'remote-kis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KIS Protocol
    #: - [UDP] KIS Protocol
    kis = 186, 'kis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Application Communication Interface
    #: - [UDP] Application Communication Interface
    aci = 187, 'aci', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Plus Five's MUMPS
    #: - [UDP] Plus Five's MUMPS
    mumps = 188, 'mumps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Queued File Transport
    #: - [UDP] Queued File Transport
    qft = 189, 'qft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gateway Access Control Protocol
    #: - [UDP] Gateway Access Control Protocol
    gacp = 190, 'gacp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prospero Directory Service
    #: - [UDP] Prospero Directory Service
    prospero = 191, 'prospero', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OSU Network Monitoring System
    #: - [UDP] OSU Network Monitoring System
    osu_nms = 192, 'osu-nms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Spider Remote Monitoring Protocol
    #: - [UDP] Spider Remote Monitoring Protocol
    srmp = 193, 'srmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Internet Relay Chat Protocol
    #: - [UDP] Internet Relay Chat Protocol
    irc = 194, 'irc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNSIX Network Level Module Audit
    #: - [UDP] DNSIX Network Level Module Audit
    dn6_nlm_aud = 195, 'dn6-nlm-aud', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNSIX Session Mgt Module Audit Redir
    #: - [UDP] DNSIX Session Mgt Module Audit Redir
    dn6_smm_red = 196, 'dn6-smm-red', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Directory Location Service Monitor
    #: - [UDP] Directory Location Service Monitor
    dls_mon = 198, 'dls-mon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMUX
    #: - [UDP] SMUX
    smux = 199, 'smux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM System Resource Controller
    #: - [UDP] IBM System Resource Controller
    src = 200, 'src', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleTalk Routing Maintenance
    #: - [UDP] AppleTalk Routing Maintenance
    at_rtmp = 201, 'at-rtmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleTalk Name Binding
    #: - [UDP] AppleTalk Name Binding
    at_nbp = 202, 'at-nbp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleTalk Unused
    #: - [UDP] AppleTalk Unused
    at_3 = 203, 'at-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleTalk Echo
    #: - [UDP] AppleTalk Echo
    at_echo = 204, 'at-echo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleTalk Unused
    #: - [UDP] AppleTalk Unused
    at_5 = 205, 'at-5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleTalk Zone Information
    #: - [UDP] AppleTalk Zone Information
    at_zis = 206, 'at-zis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleTalk Unused
    #: - [UDP] AppleTalk Unused
    at_7 = 207, 'at-7', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleTalk Unused
    #: - [UDP] AppleTalk Unused
    at_8 = 208, 'at-8', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] The Quick Mail Transfer Protocol
    #: - [UDP] The Quick Mail Transfer Protocol
    qmtp = 209, 'qmtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ANSI Z39.50 IANA assigned this well-formed service name as a
    #: replacement for "z39.50".
    #: - [TCP] ANSI Z39.50
    #: - [UDP] ANSI Z39.50 IANA assigned this well-formed service name as a
    #: replacement for "z39.50".
    #: - [UDP] ANSI Z39.50
    z39_50 = 210, 'z39-50', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Texas Instruments 914C/G Terminal IANA assigned this well-formed
    #: service name as a replacement for "914c/g".
    #: - [TCP] Texas Instruments 914C/G Terminal
    #: - [UDP] Texas Instruments 914C/G Terminal IANA assigned this well-formed
    #: service name as a replacement for "914c/g".
    #: - [UDP] Texas Instruments 914C/G Terminal
    AppType_914c_g = 211, '914c-g', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] ATEXSSTR
    #: - [UDP] ATEXSSTR
    anet = 212, 'anet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPX
    #: - [UDP] IPX
    ipx = 213, 'ipx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VM PWSCS
    #: - [UDP] VM PWSCS
    vmpwscs = 214, 'vmpwscs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Insignia Solutions
    #: - [UDP] Insignia Solutions
    softpc = 215, 'softpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Computer Associates Int'l License Server
    #: - [UDP] Computer Associates Int'l License Server
    cailic = 216, 'cailic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dBASE Unix
    #: - [UDP] dBASE Unix
    dbase = 217, 'dbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netix Message Posting Protocol
    #: - [UDP] Netix Message Posting Protocol
    mpp = 218, 'mpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unisys ARPs
    #: - [UDP] Unisys ARPs
    uarps = 219, 'uarps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Interactive Mail Access Protocol v3
    #: - [UDP] Interactive Mail Access Protocol v3
    imap3 = 220, 'imap3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Berkeley rlogind with SPX auth
    #: - [UDP] Berkeley rlogind with SPX auth
    fln_spx = 221, 'fln-spx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Berkeley rshd with SPX auth
    #: - [UDP] Berkeley rshd with SPX auth
    rsh_spx = 222, 'rsh-spx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Certificate Distribution Center
    #: - [UDP] Certificate Distribution Center
    cdc = 223, 'cdc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] masqdialer
    #: - [UDP] masqdialer
    masqdialer = 224, 'masqdialer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Direct
    #: - [UDP] Direct
    direct = 242, 'direct', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Survey Measurement
    #: - [UDP] Survey Measurement
    sur_meas = 243, 'sur-meas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] inbusiness
    #: - [UDP] inbusiness
    inbusiness = 244, 'inbusiness', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LINK
    #: - [UDP] LINK
    link = 245, 'link', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Display Systems Protocol
    #: - [UDP] Display Systems Protocol
    dsp3270 = 246, 'dsp3270', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SUBNTBCST_TFTP IANA assigned this well-formed service name as a
    #: replacement for "subntbcst_tftp".
    #: - [TCP] SUBNTBCST_TFTP
    #: - [UDP] SUBNTBCST_TFTP IANA assigned this well-formed service name as a
    #: replacement for "subntbcst_tftp".
    #: - [UDP] SUBNTBCST_TFTP
    subntbcst_tftp = 247, 'subntbcst-tftp', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] bhfhs
    #: - [UDP] bhfhs
    bhfhs = 248, 'bhfhs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Route Access Protocol
    #: - [UDP] Route Access Protocol
    rap_38 = 38, 'rap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RAP
    #: - [UDP] RAP
    rap_256 = 256, 'rap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Electronic Transaction
    #: - [UDP] Secure Electronic Transaction
    set = 257, 'set', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_258 = 258, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Efficient Short Remote Operations
    #: - [UDP] Efficient Short Remote Operations
    esro_gen = 259, 'esro-gen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Openport
    #: - [UDP] Openport
    openport = 260, 'openport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IIOP Name Service over TLS/SSL
    #: - [UDP] IIOP Name Service over TLS/SSL
    nsiiops = 261, 'nsiiops', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Arcisdms
    #: - [UDP] Arcisdms
    arcisdms = 262, 'arcisdms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HDAP
    #: - [UDP] HDAP
    hdap = 263, 'hdap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BGMP
    #: - [UDP] BGMP
    bgmp = 264, 'bgmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] X-Bone CTL
    #: - [UDP] X-Bone CTL
    x_bone_ctl = 265, 'x-bone-ctl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCSI on ST
    #: - [UDP] SCSI on ST
    sst = 266, 'sst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tobit David Service Layer
    #: - [UDP] Tobit David Service Layer
    td_service = 267, 'td-service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tobit David Replica
    #: - [UDP] Tobit David Replica
    td_replica = 268, 'td-replica', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MANET Protocols [:rfc:`5498`]
    #: - [UDP] MANET Protocols [:rfc:`5498`]
    manet = 269, 'manet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_270 = 270, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Q-mode encapsulation for GIST messages [:rfc:`5971`]
    gist = 270, 'gist', TransportProtocol.get('udp')

    #: [TCP] IETF Network Endpoint Assessment (NEA) Posture Transport Protocol over
    #: TLS (PT-TLS) [:rfc:`6876`]
    pt_tls = 271, 'pt-tls', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_271 = 271, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] http-mgmt
    #: - [UDP] http-mgmt
    http_mgmt = 280, 'http-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Personal Link
    #: - [UDP] Personal Link
    personal_link = 281, 'personal-link', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cable Port A/X
    #: - [UDP] Cable Port A/X
    cableport_ax = 282, 'cableport-ax', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rescap
    #: - [UDP] rescap
    rescap = 283, 'rescap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] corerjd
    #: - [UDP] corerjd
    corerjd = 284, 'corerjd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_285 = 285, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] K-BLOCK
    #: - [UDP] K-BLOCK
    k_block = 287, 'k-block', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novastor Backup
    #: - [UDP] Novastor Backup
    novastorbakcup = 308, 'novastorbakcup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EntrustTime
    #: - [UDP] EntrustTime
    entrusttime = 309, 'entrusttime', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bhmds
    #: - [UDP] bhmds
    bhmds = 310, 'bhmds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppleShare IP WebAdmin
    #: - [UDP] AppleShare IP WebAdmin
    asip_webadmin = 311, 'asip-webadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSLMP
    #: - [UDP] VSLMP
    vslmp = 312, 'vslmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Magenta Logic
    #: - [UDP] Magenta Logic
    magenta_logic = 313, 'magenta-logic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Opalis Robot
    #: - [UDP] Opalis Robot
    opalis_robot = 314, 'opalis-robot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DPSI
    #: - [UDP] DPSI
    dpsi = 315, 'dpsi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] decAuth
    #: - [UDP] decAuth
    decauth = 316, 'decauth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zannet
    #: - [UDP] Zannet
    zannet = 317, 'zannet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PKIX TimeStamp
    #: - [UDP] PKIX TimeStamp
    pkix_timestamp = 318, 'pkix-timestamp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PTP Event
    #: - [UDP] PTP Event
    ptp_event = 319, 'ptp-event', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PTP General
    #: - [UDP] PTP General
    ptp_general = 320, 'ptp-general', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RTSPS
    #: - [UDP] RTSPS
    rtsps = 322, 'rtsps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Resource PKI to Router Protocol [:rfc:`6810`]
    rpki_rtr = 323, 'rpki-rtr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_323 = 323, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Resource PKI to Router Protocol over TLS [:rfc:`6810`]
    rpki_rtr_tls = 324, 'rpki-rtr-tls', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_324 = 324, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Texar Security Port
    #: - [UDP] Texar Security Port
    texar = 333, 'texar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prospero Data Access Protocol
    #: - [UDP] Prospero Data Access Protocol
    pdap = 344, 'pdap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Perf Analysis Workbench
    #: - [UDP] Perf Analysis Workbench
    pawserv = 345, 'pawserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zebra server
    #: - [UDP] Zebra server
    zserv = 346, 'zserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fatmen Server
    #: - [UDP] Fatmen Server
    fatserv = 347, 'fatserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cabletron Management Protocol
    #: - [UDP] Cabletron Management Protocol
    csi_sgwp = 348, 'csi-sgwp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MATIP Type A [:rfc:`2351`]
    #: - [UDP] MATIP Type A [:rfc:`2351`]
    matip_type_a = 350, 'matip-type-a', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MATIP Type B [:rfc:`2351`]
    #: - [UDP] MATIP Type B [:rfc:`2351`]
    matip_type_b = 351, 'matip-type-b', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bhoetty
    #: - [UDP] bhoetty
    bhoetty = 351, 'bhoetty', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DTAG
    #: - [UDP] DTAG
    dtag_ste_sb = 352, 'dtag-ste-sb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bhoedap4
    #: - [UDP] bhoedap4
    bhoedap4 = 352, 'bhoedap4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NDSAUTH
    #: - [UDP] NDSAUTH
    ndsauth = 353, 'ndsauth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bh611
    #: - [UDP] bh611
    bh611 = 354, 'bh611', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DATEX-ASN
    #: - [UDP] DATEX-ASN
    datex_asn = 355, 'datex-asn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cloanto Net 1
    #: - [UDP] Cloanto Net 1
    cloanto_net_1 = 356, 'cloanto-net-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bhevent
    #: - [UDP] bhevent
    bhevent = 357, 'bhevent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Shrinkwrap
    #: - [UDP] Shrinkwrap
    shrinkwrap = 358, 'shrinkwrap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Security Risk Management Protocol
    #: - [UDP] Network Security Risk Management Protocol
    nsrmp = 359, 'nsrmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] scoi2odialog
    #: - [UDP] scoi2odialog
    scoi2odialog = 360, 'scoi2odialog', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Semantix
    #: - [UDP] Semantix
    semantix = 361, 'semantix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SRS Send
    #: - [UDP] SRS Send
    srssend = 362, 'srssend', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSVP Tunnel IANA assigned this well-formed service name as a
    #: replacement for "rsvp_tunnel".
    #: - [TCP] RSVP Tunnel
    #: - [UDP] RSVP Tunnel IANA assigned this well-formed service name as a
    #: replacement for "rsvp_tunnel".
    #: - [UDP] RSVP Tunnel
    rsvp_tunnel = 363, 'rsvp-tunnel', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Aurora CMGR
    #: - [UDP] Aurora CMGR
    aurora_cmgr = 364, 'aurora-cmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DTK
    #: - [UDP] DTK
    dtk = 365, 'dtk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ODMR
    #: - [UDP] ODMR
    odmr = 366, 'odmr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MortgageWare
    #: - [UDP] MortgageWare
    mortgageware = 367, 'mortgageware', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QbikGDP
    #: - [UDP] QbikGDP
    qbikgdp = 368, 'qbikgdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rpc2portmap
    #: - [UDP] rpc2portmap
    rpc2portmap = 369, 'rpc2portmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] codaauth2
    #: - [UDP] codaauth2
    codaauth2 = 370, 'codaauth2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Clearcase
    #: - [UDP] Clearcase
    clearcase = 371, 'clearcase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ListProcessor
    #: - [UDP] ListProcessor
    ulistproc = 372, 'ulistproc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Legent Corporation
    #: - [UDP] Legent Corporation
    legent_1 = 373, 'legent-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Legent Corporation
    #: - [UDP] Legent Corporation
    legent_2 = 374, 'legent-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hassle
    #: - [UDP] Hassle
    hassle = 375, 'hassle', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Amiga Envoy Network Inquiry Protocol
    #: - [UDP] Amiga Envoy Network Inquiry Protocol
    nip = 376, 'nip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NEC Corporation
    #: - [UDP] NEC Corporation
    tnetos = 377, 'tnetos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NEC Corporation
    #: - [UDP] NEC Corporation
    dsetos = 378, 'dsetos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TIA/EIA/IS-99 modem client
    #: - [UDP] TIA/EIA/IS-99 modem client
    is99c = 379, 'is99c', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TIA/EIA/IS-99 modem server
    #: - [UDP] TIA/EIA/IS-99 modem server
    is99s = 380, 'is99s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp performance data collector
    #: - [UDP] hp performance data collector
    hp_collector = 381, 'hp-collector', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp performance data managed node
    #: - [UDP] hp performance data managed node
    hp_managed_node = 382, 'hp-managed-node', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp performance data alarm manager
    #: - [UDP] hp performance data alarm manager
    hp_alarm_mgr = 383, 'hp-alarm-mgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A Remote Network Server System
    #: - [UDP] A Remote Network Server System
    arns = 384, 'arns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Application
    #: - [UDP] IBM Application
    ibm_app = 385, 'ibm-app', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASA Message Router Object Def.
    #: - [UDP] ASA Message Router Object Def.
    asa = 386, 'asa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Appletalk Update-Based Routing Pro.
    #: - [UDP] Appletalk Update-Based Routing Pro.
    aurp = 387, 'aurp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unidata LDM
    #: - [UDP] Unidata LDM
    unidata_ldm = 388, 'unidata-ldm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lightweight Directory Access Protocol
    #: - [UDP] Lightweight Directory Access Protocol
    ldap = 389, 'ldap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UIS
    #: - [UDP] UIS
    uis = 390, 'uis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SynOptics SNMP Relay Port
    #: - [UDP] SynOptics SNMP Relay Port
    synotics_relay = 391, 'synotics-relay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SynOptics Port Broker Port
    #: - [UDP] SynOptics Port Broker Port
    synotics_broker = 392, 'synotics-broker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Meta5
    #: - [UDP] Meta5
    meta5 = 393, 'meta5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMBL Nucleic Data Transfer
    #: - [UDP] EMBL Nucleic Data Transfer
    embl_ndt = 394, 'embl-ndt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetScout Control Protocol
    #: - [UDP] NetScout Control Protocol
    netcp = 395, 'netcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novell Netware over IP
    #: - [UDP] Novell Netware over IP
    netware_ip = 396, 'netware-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multi Protocol Trans. Net.
    #: - [UDP] Multi Protocol Trans. Net.
    mptn = 397, 'mptn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kryptolan
    #: - [UDP] Kryptolan
    kryptolan = 398, 'kryptolan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISO Transport Class 2 Non-Control over TCP
    #: - [UDP] ISO Transport Class 2 Non-Control over UDP
    iso_tsap_c2 = 399, 'iso-tsap-c2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle Secure Backup
    #: - [UDP] Oracle Secure Backup
    osb_sd = 400, 'osb-sd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Uninterruptible Power Supply
    #: - [UDP] Uninterruptible Power Supply
    ups = 401, 'ups', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Genie Protocol
    #: - [UDP] Genie Protocol
    genie = 402, 'genie', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] decap
    #: - [UDP] decap
    decap = 403, 'decap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nced
    #: - [UDP] nced
    nced = 404, 'nced', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ncld
    #: - [UDP] ncld
    ncld = 405, 'ncld', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Interactive Mail Support Protocol
    #: - [UDP] Interactive Mail Support Protocol
    imsp = 406, 'imsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Timbuktu
    #: - [UDP] Timbuktu
    timbuktu = 407, 'timbuktu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prospero Resource Manager Sys. Man.
    #: - [UDP] Prospero Resource Manager Sys. Man.
    prm_sm = 408, 'prm-sm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prospero Resource Manager Node Man.
    #: - [UDP] Prospero Resource Manager Node Man.
    prm_nm = 409, 'prm-nm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DECLadebug Remote Debug Protocol
    #: - [UDP] DECLadebug Remote Debug Protocol
    decladebug = 410, 'decladebug', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote MT Protocol
    #: - [UDP] Remote MT Protocol
    rmt = 411, 'rmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trap Convention Port
    #: - [UDP] Trap Convention Port
    synoptics_trap = 412, 'synoptics-trap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Storage Management Services Protocol
    #: - [UDP] Storage Management Services Protocol
    smsp = 413, 'smsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] InfoSeek
    #: - [UDP] InfoSeek
    infoseek = 414, 'infoseek', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BNet
    #: - [UDP] BNet
    bnet = 415, 'bnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Silverplatter
    #: - [UDP] Silverplatter
    silverplatter = 416, 'silverplatter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Onmux
    #: - [UDP] Onmux
    onmux = 417, 'onmux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hyper-G
    #: - [UDP] Hyper-G
    hyper_g = 418, 'hyper-g', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ariel 1
    #: - [UDP] Ariel 1
    ariel1 = 419, 'ariel1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMPTE
    #: - [UDP] SMPTE
    smpte = 420, 'smpte', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ariel 2
    #: - [UDP] Ariel 2
    ariel2 = 421, 'ariel2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ariel 3
    #: - [UDP] Ariel 3
    ariel3 = 422, 'ariel3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Operations Planning and Control Start
    #: - [UDP] IBM Operations Planning and Control Start
    opc_job_start = 423, 'opc-job-start', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Operations Planning and Control Track
    #: - [UDP] IBM Operations Planning and Control Track
    opc_job_track = 424, 'opc-job-track', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICAD
    #: - [UDP] ICAD
    icad_el = 425, 'icad-el', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] smartsdp
    #: - [UDP] smartsdp
    smartsdp = 426, 'smartsdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Server Location
    #: - [UDP] Server Location
    svrloc = 427, 'svrloc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OCS_CMU IANA assigned this well-formed service name as a replacement
    #: for "ocs_cmu".
    #: - [TCP] OCS_CMU
    #: - [UDP] OCS_CMU IANA assigned this well-formed service name as a replacement
    #: for "ocs_cmu".
    #: - [UDP] OCS_CMU
    ocs_cmu = 428, 'ocs-cmu', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] OCS_AMU IANA assigned this well-formed service name as a replacement
    #: for "ocs_amu".
    #: - [TCP] OCS_AMU
    #: - [UDP] OCS_AMU IANA assigned this well-formed service name as a replacement
    #: for "ocs_amu".
    #: - [UDP] OCS_AMU
    ocs_amu = 429, 'ocs-amu', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] UTMPSD
    #: - [UDP] UTMPSD
    utmpsd = 430, 'utmpsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UTMPCD
    #: - [UDP] UTMPCD
    utmpcd = 431, 'utmpcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IASD
    #: - [UDP] IASD
    iasd = 432, 'iasd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NNTP for transit servers (NNSP) [:rfc:`3977`]
    #: - [UDP] NNTP for transit servers (NNSP) [:rfc:`3977`]
    nnsp = 433, 'nnsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MobileIP-Agent
    #: - [UDP] MobileIP-Agent
    mobileip_agent = 434, 'mobileip-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MobilIP-MN
    #: - [UDP] MobilIP-MN
    mobilip_mn = 435, 'mobilip-mn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNA-CML
    #: - [UDP] DNA-CML
    dna_cml = 436, 'dna-cml', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] comscm
    #: - [UDP] comscm
    comscm = 437, 'comscm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dsfgw
    #: - [UDP] dsfgw
    dsfgw = 438, 'dsfgw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dasp
    #: - [UDP] dasp
    dasp = 439, 'dasp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sgcp
    #: - [UDP] sgcp
    sgcp = 440, 'sgcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] decvms-sysmgt
    #: - [UDP] decvms-sysmgt
    decvms_sysmgt = 441, 'decvms-sysmgt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cvc_hostd IANA assigned this well-formed service name as a replacement
    #: for "cvc_hostd".
    #: - [TCP] cvc_hostd
    #: - [UDP] cvc_hostd IANA assigned this well-formed service name as a replacement
    #: for "cvc_hostd".
    #: - [UDP] cvc_hostd
    cvc_hostd = 442, 'cvc-hostd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Network Paging Protocol [:rfc:`1568`]
    #: - [UDP] Simple Network Paging Protocol [:rfc:`1568`]
    snpp = 444, 'snpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft-DS
    #: - [UDP] Microsoft-DS
    microsoft_ds = 445, 'microsoft-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DDM-Remote Relational Database Access
    #: - [UDP] DDM-Remote Relational Database Access
    ddm_rdb = 446, 'ddm-rdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DDM-Distributed File Management
    #: - [UDP] DDM-Distributed File Management
    ddm_dfm = 447, 'ddm-dfm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DDM-Remote DB Access Using Secure Sockets
    #: - [UDP] DDM-Remote DB Access Using Secure Sockets
    ddm_ssl = 448, 'ddm-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AS Server Mapper
    #: - [UDP] AS Server Mapper
    as_servermap = 449, 'as-servermap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Computer Supported Telecomunication Applications
    #: - [UDP] Computer Supported Telecomunication Applications
    tserver = 450, 'tserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cray Network Semaphore server
    #: - [UDP] Cray Network Semaphore server
    sfs_smp_net = 451, 'sfs-smp-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cray SFS config server
    #: - [UDP] Cray SFS config server
    sfs_config = 452, 'sfs-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] macon-tcp
    macon_tcp = 456, 'macon-tcp', TransportProtocol.get('tcp')

    #: [UDP] macon-udp
    macon_udp = 456, 'macon-udp', TransportProtocol.get('udp')

    #: - [TCP] scohelp
    #: - [UDP] scohelp
    scohelp = 457, 'scohelp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] apple quick time
    #: - [UDP] apple quick time
    appleqtc = 458, 'appleqtc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ampr-rcmd
    #: - [UDP] ampr-rcmd
    ampr_rcmd = 459, 'ampr-rcmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] skronk
    #: - [UDP] skronk
    skronk = 460, 'skronk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DataRampSrv
    #: - [UDP] DataRampSrv
    datasurfsrv = 461, 'datasurfsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DataRampSrvSec
    #: - [UDP] DataRampSrvSec
    datasurfsrvsec = 462, 'datasurfsrvsec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] alpes
    #: - [UDP] alpes
    alpes = 463, 'alpes', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] kpasswd
    #: - [UDP] kpasswd
    kpasswd = 464, 'kpasswd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] URL Rendezvous Directory for SSM
    urd = 465, 'urd', TransportProtocol.get('tcp')

    #: [TCP] Message Submission over TLS protocol [:rfc:`8314`]
    submissions = 465, 'submissions', TransportProtocol.get('tcp')

    #: [UDP] IGMP over UDP for SSM
    igmpv3lite = 465, 'igmpv3lite', TransportProtocol.get('udp')

    #: - [TCP] digital-vrc
    #: - [UDP] digital-vrc
    digital_vrc = 466, 'digital-vrc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mylex-mapd
    #: - [UDP] mylex-mapd
    mylex_mapd = 467, 'mylex-mapd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] proturis
    #: - [UDP] proturis
    photuris = 468, 'photuris', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Radio Control Protocol
    #: - [UDP] Radio Control Protocol
    rcp = 469, 'rcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] scx-proxy
    #: - [UDP] scx-proxy
    scx_proxy = 470, 'scx-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mondex
    #: - [UDP] Mondex
    mondex = 471, 'mondex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ljk-login
    #: - [UDP] ljk-login
    ljk_login = 472, 'ljk-login', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hybrid-pop
    #: - [UDP] hybrid-pop
    hybrid_pop = 473, 'hybrid-pop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] tn-tl-w1
    tn_tl_w1 = 474, 'tn-tl-w1', TransportProtocol.get('tcp')

    #: [UDP] tn-tl-w2
    tn_tl_w2 = 474, 'tn-tl-w2', TransportProtocol.get('udp')

    #: - [TCP] tcpnethaspsrv
    #: - [UDP] tcpnethaspsrv
    tcpnethaspsrv = 475, 'tcpnethaspsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tn-tl-fd1
    #: - [UDP] tn-tl-fd1
    tn_tl_fd1 = 476, 'tn-tl-fd1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ss7ns
    #: - [UDP] ss7ns
    ss7ns = 477, 'ss7ns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] spsc
    #: - [UDP] spsc
    spsc = 478, 'spsc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iafserver
    #: - [UDP] iafserver
    iafserver = 479, 'iafserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iafdbase
    #: - [UDP] iafdbase
    iafdbase = 480, 'iafdbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ph service
    #: - [UDP] Ph service
    ph = 481, 'ph', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bgs-nsi
    #: - [UDP] bgs-nsi
    bgs_nsi = 482, 'bgs-nsi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ulpnet
    #: - [UDP] ulpnet
    ulpnet = 483, 'ulpnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Integra Software Management Environment
    #: - [UDP] Integra Software Management Environment
    integra_sme = 484, 'integra-sme', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Air Soft Power Burst
    #: - [UDP] Air Soft Power Burst
    powerburst = 485, 'powerburst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] avian
    #: - [UDP] avian
    avian = 486, 'avian', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] saft Simple Asynchronous File Transfer
    #: - [UDP] saft Simple Asynchronous File Transfer
    saft = 487, 'saft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gss-http
    #: - [UDP] gss-http
    gss_http = 488, 'gss-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nest-protocol
    #: - [UDP] nest-protocol
    nest_protocol = 489, 'nest-protocol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] micom-pfs
    #: - [UDP] micom-pfs
    micom_pfs = 490, 'micom-pfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] go-login
    #: - [UDP] go-login
    go_login = 491, 'go-login', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Transport Independent Convergence for FNA
    #: - [UDP] Transport Independent Convergence for FNA
    ticf_1 = 492, 'ticf-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Transport Independent Convergence for FNA
    #: - [UDP] Transport Independent Convergence for FNA
    ticf_2 = 493, 'ticf-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] POV-Ray
    #: - [UDP] POV-Ray
    pov_ray = 494, 'pov-ray', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] intecourier
    #: - [UDP] intecourier
    intecourier = 495, 'intecourier', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PIM-RP-DISC
    #: - [UDP] PIM-RP-DISC
    pim_rp_disc = 496, 'pim-rp-disc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] siam
    #: - [UDP] siam
    siam = 498, 'siam', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISO ILL Protocol
    #: - [UDP] ISO ILL Protocol
    iso_ill = 499, 'iso-ill', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isakmp
    #: - [UDP] isakmp
    isakmp = 500, 'isakmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STMF
    #: - [UDP] STMF
    stmf = 501, 'stmf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Modbus Application Protocol
    #: - [UDP] Modbus Application Protocol
    mbap = 502, 'mbap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intrinsa
    #: - [UDP] Intrinsa
    intrinsa = 503, 'intrinsa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] citadel
    #: - [UDP] citadel
    citadel = 504, 'citadel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mailbox-lm
    #: - [UDP] mailbox-lm
    mailbox_lm = 505, 'mailbox-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ohimsrv
    #: - [UDP] ohimsrv
    ohimsrv = 506, 'ohimsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] crs
    #: - [UDP] crs
    crs = 507, 'crs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xvttp
    #: - [UDP] xvttp
    xvttp = 508, 'xvttp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] snare
    #: - [UDP] snare
    snare = 509, 'snare', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FirstClass Protocol
    #: - [UDP] FirstClass Protocol
    fcp = 510, 'fcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PassGo
    #: - [UDP] PassGo
    passgo = 511, 'passgo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] remote process execution; authentication performed using passwords and
    #: UNIX login names
    exec = 512, 'exec', TransportProtocol.get('tcp')

    #: [UDP]
    comsat = 512, 'comsat', TransportProtocol.get('udp')

    #: [UDP] used by mail system to notify users of new mail received; currently
    #: receives messages only from processes on the same machine
    biff = 512, 'biff', TransportProtocol.get('udp')

    #: [TCP] remote login a la telnet; automatic authentication performed based on
    #: priviledged port numbers and distributed data bases which identify
    #: "authentication domains"
    login = 513, 'login', TransportProtocol.get('tcp')

    #: [UDP] maintains data bases showing who's logged in to machines on a local
    #: net and the load average of the machine
    who = 513, 'who', TransportProtocol.get('udp')

    #: [TCP] cmd like exec, but automatic authentication is performed as for login
    #: server
    shell = 514, 'shell', TransportProtocol.get('tcp')

    #: [UDP]  [:rfc:`5426`]
    syslog = 514, 'syslog', TransportProtocol.get('udp')

    #: - [TCP] spooler
    #: - [UDP] spooler
    printer = 515, 'printer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] videotex
    #: - [UDP] videotex
    videotex = 516, 'videotex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] like tenex link, but across machine - unfortunately, doesn't use link
    #: protocol (this is actually just a rendezvous port from which a tcp
    #: connection is established)
    #: - [UDP] like tenex link, but across machine - unfortunately, doesn't use link
    #: protocol (this is actually just a rendezvous port from which a tcp
    #: connection is established)
    talk = 517, 'talk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    ntalk = 518, 'ntalk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] unixtime
    #: - [UDP] unixtime
    utime = 519, 'utime', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] extended file name server
    efs = 520, 'efs', TransportProtocol.get('tcp')

    #: [UDP] local routing process (on site); uses variant of Xerox NS routing
    #: information protocol - RIP
    router = 520, 'router', TransportProtocol.get('udp')

    #: - [TCP] ripng
    #: - [UDP] ripng
    ripng = 521, 'ripng', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ULP
    #: - [UDP] ULP
    ulp = 522, 'ulp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM-DB2
    #: - [UDP] IBM-DB2
    ibm_db2 = 523, 'ibm-db2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NCP
    #: - [UDP] NCP
    ncp = 524, 'ncp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] timeserver
    #: - [UDP] timeserver
    timed = 525, 'timed', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] newdate
    #: - [UDP] newdate
    tempo = 526, 'tempo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Stock IXChange
    #: - [UDP] Stock IXChange
    stx = 527, 'stx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Customer IXChange
    #: - [UDP] Customer IXChange
    custix = 528, 'custix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IRC-SERV
    #: - [UDP] IRC-SERV
    irc_serv = 529, 'irc-serv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rpc
    #: - [UDP] rpc
    courier = 530, 'courier', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] chat
    #: - [UDP] chat
    conference = 531, 'conference', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] readnews
    #: - [UDP] readnews
    netnews = 532, 'netnews', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] for emergency broadcasts
    #: - [UDP] for emergency broadcasts
    netwall = 533, 'netwall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] windream Admin
    #: - [UDP] windream Admin
    windream = 534, 'windream', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iiop
    #: - [UDP] iiop
    iiop = 535, 'iiop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] opalis-rdv
    #: - [UDP] opalis-rdv
    opalis_rdv = 536, 'opalis-rdv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gdomap
    #: - [UDP] gdomap
    gdomap = 538, 'gdomap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apertus Technologies Load Determination
    #: - [UDP] Apertus Technologies Load Determination
    apertus_ldp = 539, 'apertus-ldp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] uucpd
    #: - [UDP] uucpd
    uucp = 540, 'uucp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] uucp-rlogin
    #: - [UDP] uucp-rlogin
    uucp_rlogin = 541, 'uucp-rlogin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] commerce
    #: - [UDP] commerce
    commerce = 542, 'commerce', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    klogin = 543, 'klogin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] krcmd
    #: - [UDP] krcmd
    kshell = 544, 'kshell', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] appleqtcsrvr
    #: - [UDP] appleqtcsrvr
    appleqtcsrvr = 545, 'appleqtcsrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DHCPv6 Client
    #: - [UDP] DHCPv6 Client [:rfc:`8415`]
    dhcpv6_client = 546, 'dhcpv6-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DHCPv6 Server [:rfc:`5460`]
    #: - [UDP] DHCPv6 Server [:rfc:`8415`]
    dhcpv6_server = 547, 'dhcpv6-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AFP over TCP
    #: - [UDP] AFP over TCP
    afpovertcp = 548, 'afpovertcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IDFP
    #: - [UDP] IDFP
    idfp = 549, 'idfp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] new-who [n/a]
    #: - [UDP] new-who [n/a]
    new_rwho = 550, 'new-rwho', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cybercash [:rfc:`1898`]
    #: - [UDP] cybercash [:rfc:`1898`]
    cybercash = 551, 'cybercash', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DeviceShare
    #: - [UDP] DeviceShare
    devshr_nts = 552, 'devshr-nts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pirp
    #: - [UDP] pirp
    pirp = 553, 'pirp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Real Time Streaming Protocol (RTSP)
    #: - [UDP] Real Time Streaming Protocol (RTSP)
    rtsp = 554, 'rtsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    dsf = 555, 'dsf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rfs server
    #: - [UDP] rfs server
    remotefs = 556, 'remotefs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] openvms-sysipc
    #: - [UDP] openvms-sysipc
    openvms_sysipc = 557, 'openvms-sysipc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SDNSKMP
    #: - [UDP] SDNSKMP
    sdnskmp = 558, 'sdnskmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TEEDTAP
    #: - [UDP] TEEDTAP
    teedtap = 559, 'teedtap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rmonitord
    #: - [UDP] rmonitord
    rmonitor = 560, 'rmonitor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    monitor = 561, 'monitor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] chcmd
    #: - [UDP] chcmd
    chshell = 562, 'chshell', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nntp protocol over TLS/SSL (was snntp) [:rfc:`4642`]
    #: - [UDP] nntp protocol over TLS/SSL (was snntp) [:rfc:`4642`]
    nntps = 563, 'nntps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] plan 9 file service
    #: - [UDP] plan 9 file service
    AppType_9pfs = 564, '9pfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] whoami
    #: - [UDP] whoami
    whoami = 565, 'whoami', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] streettalk
    #: - [UDP] streettalk
    streettalk = 566, 'streettalk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] banyan-rpc
    #: - [UDP] banyan-rpc
    banyan_rpc = 567, 'banyan-rpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] microsoft shuttle
    #: - [UDP] microsoft shuttle
    ms_shuttle = 568, 'ms-shuttle', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] microsoft rome
    #: - [UDP] microsoft rome
    ms_rome = 569, 'ms-rome', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] demon
    #: - [UDP] demon
    meter_570 = 570, 'meter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] udemon
    #: - [UDP] udemon
    meter_571 = 571, 'meter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sonar
    #: - [UDP] sonar
    sonar = 572, 'sonar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] banyan-vip
    #: - [UDP] banyan-vip
    banyan_vip = 573, 'banyan-vip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FTP Software Agent System
    #: - [UDP] FTP Software Agent System
    ftp_agent = 574, 'ftp-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VEMMI
    #: - [UDP] VEMMI
    vemmi = 575, 'vemmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ipcd
    #: - [UDP] ipcd
    ipcd = 576, 'ipcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vnas
    #: - [UDP] vnas
    vnas = 577, 'vnas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ipdd
    #: - [UDP] ipdd
    ipdd = 578, 'ipdd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] decbsrv
    #: - [UDP] decbsrv
    decbsrv = 579, 'decbsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNTP HEARTBEAT
    #: - [UDP] SNTP HEARTBEAT
    sntp_heartbeat = 580, 'sntp-heartbeat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bundle Discovery Protocol
    #: - [UDP] Bundle Discovery Protocol
    bdp = 581, 'bdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCC Security
    #: - [UDP] SCC Security
    scc_security = 582, 'scc-security', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Philips Video-Conferencing
    #: - [UDP] Philips Video-Conferencing
    philips_vc = 583, 'philips-vc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Key Server
    #: - [UDP] Key Server
    keyserver = 584, 'keyserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Password Change
    #: - [UDP] Password Change
    password_chg = 586, 'password-chg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Message Submission [:rfc:`6409`]
    #: - [UDP] Message Submission [:rfc:`6409`]
    submission = 587, 'submission', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CAL
    #: - [UDP] CAL
    cal = 588, 'cal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EyeLink
    #: - [UDP] EyeLink
    eyelink = 589, 'eyelink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TNS CML
    #: - [UDP] TNS CML
    tns_cml = 590, 'tns-cml', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Eudora Set
    #: - [UDP] Eudora Set
    eudora_set = 592, 'eudora-set', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HTTP RPC Ep Map
    #: - [UDP] HTTP RPC Ep Map
    http_rpc_epmap = 593, 'http-rpc-epmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TPIP
    #: - [UDP] TPIP
    tpip = 594, 'tpip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CAB Protocol
    #: - [UDP] CAB Protocol
    cab_protocol = 595, 'cab-protocol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMSD
    #: - [UDP] SMSD
    smsd = 596, 'smsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PTC Name Service
    #: - [UDP] PTC Name Service
    ptcnameservice = 597, 'ptcnameservice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCO Web Server Manager 3
    #: - [UDP] SCO Web Server Manager 3
    sco_websrvrmg3 = 598, 'sco-websrvrmg3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Aeolon Core Protocol
    #: - [UDP] Aeolon Core Protocol
    acp = 599, 'acp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun IPC server
    #: - [UDP] Sun IPC server
    ipcserver = 600, 'ipcserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reliable Syslog Service [:rfc:`3195`]
    #: - [UDP] Reliable Syslog Service [:rfc:`3195`]
    syslog_conn = 601, 'syslog-conn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XML-RPC over BEEP [:rfc:`3529`]
    #: - [UDP] XML-RPC over BEEP [:rfc:`3529`]
    xmlrpc_beep = 602, 'xmlrpc-beep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IDXP [:rfc:`4767`]
    #: - [UDP] IDXP [:rfc:`4767`]
    idxp = 603, 'idxp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TUNNEL [:rfc:`3620`]
    #: - [UDP] TUNNEL [:rfc:`3620`]
    tunnel = 604, 'tunnel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SOAP over BEEP [:rfc:`4227`]
    #: - [UDP] SOAP over BEEP [:rfc:`4227`]
    soap_beep = 605, 'soap-beep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cray Unified Resource Manager
    #: - [UDP] Cray Unified Resource Manager
    urm = 606, 'urm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nqs
    #: - [UDP] nqs
    nqs = 607, 'nqs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sender-Initiated/Unsolicited File Transfer
    #: - [UDP] Sender-Initiated/Unsolicited File Transfer
    sift_uft = 608, 'sift-uft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] npmp-trap
    #: - [UDP] npmp-trap
    npmp_trap = 609, 'npmp-trap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] npmp-local
    #: - [UDP] npmp-local
    npmp_local = 610, 'npmp-local', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] npmp-gui
    #: - [UDP] npmp-gui
    npmp_gui = 611, 'npmp-gui', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HMMP Indication
    #: - [UDP] HMMP Indication
    hmmp_ind = 612, 'hmmp-ind', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HMMP Operation
    #: - [UDP] HMMP Operation
    hmmp_op = 613, 'hmmp-op', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSLshell
    #: - [UDP] SSLshell
    sshell = 614, 'sshell', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Internet Configuration Manager
    #: - [UDP] Internet Configuration Manager
    sco_inetmgr = 615, 'sco-inetmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCO System Administration Server
    #: - [UDP] SCO System Administration Server
    sco_sysmgr = 616, 'sco-sysmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCO Desktop Administration Server
    #: - [UDP] SCO Desktop Administration Server
    sco_dtmgr = 617, 'sco-dtmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DEI-ICDA
    #: - [UDP] DEI-ICDA
    dei_icda = 618, 'dei-icda', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compaq EVM
    #: - [UDP] Compaq EVM
    compaq_evm = 619, 'compaq-evm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCO WebServer Manager
    #: - [UDP] SCO WebServer Manager
    sco_websrvrmgr = 620, 'sco-websrvrmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESCP
    #: - [UDP] ESCP
    escp_ip = 621, 'escp-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Collaborator
    #: - [UDP] Collaborator
    collaborator = 622, 'collaborator', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DMTF out-of-band web services management protocol
    oob_ws_http = 623, 'oob-ws-http', TransportProtocol.get('tcp')

    #: [UDP] ASF Remote Management and Control Protocol
    asf_rmcp = 623, 'asf-rmcp', TransportProtocol.get('udp')

    #: - [TCP] Crypto Admin
    #: - [UDP] Crypto Admin
    cryptoadmin = 624, 'cryptoadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DEC DLM IANA assigned this well-formed service name as a replacement
    #: for "dec_dlm".
    #: - [TCP] DEC DLM
    #: - [UDP] DEC DLM IANA assigned this well-formed service name as a replacement
    #: for "dec_dlm".
    #: - [UDP] DEC DLM
    dec_dlm = 625, 'dec-dlm', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] ASIA
    #: - [UDP] ASIA
    asia = 626, 'asia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PassGo Tivoli
    #: - [UDP] PassGo Tivoli
    passgo_tivoli = 627, 'passgo-tivoli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QMQP
    #: - [UDP] QMQP
    qmqp = 628, 'qmqp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3Com AMP3
    #: - [UDP] 3Com AMP3
    AppType_3com_amp3 = 629, '3com-amp3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPP (Internet Printing Protocol) [:rfc:`8011`]
    #: - [UDP] IPP (Internet Printing Protocol) [:rfc:`8011`]
    ipp = 631, 'ipp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Internet Printing Protocol over HTTPS [:rfc:`8011`]
    ipps = 631, 'ipps', TransportProtocol.get('tcp')

    #: - [TCP] bmpp
    #: - [UDP] bmpp
    bmpp = 632, 'bmpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Status update (Sterling Software)
    #: - [UDP] Service Status update (Sterling Software)
    servstat = 633, 'servstat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ginad
    #: - [UDP] ginad
    ginad = 634, 'ginad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RLZ DBase
    #: - [UDP] RLZ DBase
    rlzdbase = 635, 'rlzdbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ldap protocol over TLS/SSL (was sldap)
    #: - [UDP] ldap protocol over TLS/SSL (was sldap)
    ldaps = 636, 'ldaps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] lanserver
    #: - [UDP] lanserver
    lanserver = 637, 'lanserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mcns-sec
    #: - [UDP] mcns-sec
    mcns_sec = 638, 'mcns-sec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSDP
    #: - [UDP] MSDP
    msdp = 639, 'msdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] entrust-sps
    #: - [UDP] entrust-sps
    entrust_sps = 640, 'entrust-sps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] repcmd
    #: - [UDP] repcmd
    repcmd = 641, 'repcmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESRO-EMSDP V1.3
    #: - [UDP] ESRO-EMSDP V1.3
    esro_emsdp = 642, 'esro-emsdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SANity
    #: - [UDP] SANity
    sanity = 643, 'sanity', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dwr
    #: - [UDP] dwr
    dwr = 644, 'dwr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PSSC
    #: - [UDP] PSSC
    pssc = 645, 'pssc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LDP
    #: - [UDP] LDP
    ldp = 646, 'ldp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DHCP Failover [:rfc:`8156`]
    #: - [UDP] DHCP Failover
    dhcp_failover = 647, 'dhcp-failover', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Registry Registrar Protocol (RRP)
    #: - [UDP] Registry Registrar Protocol (RRP)
    rrp = 648, 'rrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cadview-3d - streaming 3d models over the internet
    #: - [UDP] Cadview-3d - streaming 3d models over the internet
    cadview_3d = 649, 'cadview-3d', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OBEX
    #: - [UDP] OBEX
    obex = 650, 'obex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IEEE MMS
    #: - [UDP] IEEE MMS
    ieee_mms = 651, 'ieee-mms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HELLO_PORT
    #: - [UDP] HELLO_PORT
    hello_port = 652, 'hello-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RepCmd
    #: - [UDP] RepCmd
    repscmd = 653, 'repscmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AODV
    #: - [UDP] AODV
    aodv = 654, 'aodv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TINC
    #: - [UDP] TINC
    tinc = 655, 'tinc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPMP
    #: - [UDP] SPMP
    spmp = 656, 'spmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RMC
    #: - [UDP] RMC
    rmc = 657, 'rmc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TenFold
    #: - [UDP] TenFold
    tenfold = 658, 'tenfold', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MacOS Server Admin
    #: - [UDP] MacOS Server Admin
    mac_srvr_admin = 660, 'mac-srvr-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HAP
    #: - [UDP] HAP
    hap = 661, 'hap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PFTP
    #: - [UDP] PFTP
    pftp = 662, 'pftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PureNoise
    #: - [UDP] PureNoise
    purenoise = 663, 'purenoise', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DMTF out-of-band secure web services management protocol
    oob_ws_https = 664, 'oob-ws-https', TransportProtocol.get('tcp')

    #: [UDP] ASF Secure Remote Management and Control Protocol
    asf_secure_rmcp = 664, 'asf-secure-rmcp', TransportProtocol.get('udp')

    #: - [TCP] Sun DR
    #: - [UDP] Sun DR
    sun_dr = 665, 'sun-dr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    mdqs = 666, 'mdqs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] doom Id Software
    #: - [UDP] doom Id Software
    doom = 666, 'doom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] campaign contribution disclosures - SDR Technologies
    #: - [UDP] campaign contribution disclosures - SDR Technologies
    disclose = 667, 'disclose', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MeComm
    #: - [UDP] MeComm
    mecomm = 668, 'mecomm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MeRegister
    #: - [UDP] MeRegister
    meregister = 669, 'meregister', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VACDSM-SWS
    #: - [UDP] VACDSM-SWS
    vacdsm_sws = 670, 'vacdsm-sws', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VACDSM-APP
    #: - [UDP] VACDSM-APP
    vacdsm_app = 671, 'vacdsm-app', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VPPS-QUA
    #: - [UDP] VPPS-QUA
    vpps_qua = 672, 'vpps-qua', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CIMPLEX
    #: - [UDP] CIMPLEX
    cimplex = 673, 'cimplex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACAP
    #: - [UDP] ACAP
    acap = 674, 'acap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DCTP
    #: - [UDP] DCTP
    dctp = 675, 'dctp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VPPS Via
    #: - [UDP] VPPS Via
    vpps_via = 676, 'vpps-via', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Presence Protocol
    #: - [UDP] Virtual Presence Protocol
    vpp = 677, 'vpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GNU Generation Foundation NCP
    #: - [UDP] GNU Generation Foundation NCP
    ggf_ncp = 678, 'ggf-ncp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MRM
    #: - [UDP] MRM
    mrm = 679, 'mrm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] entrust-aaas
    #: - [UDP] entrust-aaas
    entrust_aaas = 680, 'entrust-aaas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] entrust-aams
    #: - [UDP] entrust-aams
    entrust_aams = 681, 'entrust-aams', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XFR
    #: - [UDP] XFR
    xfr = 682, 'xfr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CORBA IIOP
    #: - [UDP] CORBA IIOP
    corba_iiop = 683, 'corba-iiop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CORBA IIOP SSL
    #: - [UDP] CORBA IIOP SSL
    corba_iiop_ssl = 684, 'corba-iiop-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MDC Port Mapper
    #: - [UDP] MDC Port Mapper
    mdc_portmapper = 685, 'mdc-portmapper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hardware Control Protocol Wismar
    #: - [UDP] Hardware Control Protocol Wismar
    hcp_wismar = 686, 'hcp-wismar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] asipregistry
    #: - [UDP] asipregistry
    asipregistry = 687, 'asipregistry', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ApplianceWare managment protocol
    #: - [UDP] ApplianceWare managment protocol
    realm_rusd = 688, 'realm-rusd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NMAP
    #: - [UDP] NMAP
    nmap = 689, 'nmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Velneo Application Transfer Protocol
    #: - [UDP] Velneo Application Transfer Protocol
    vatp = 690, 'vatp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS Exchange Routing
    #: - [UDP] MS Exchange Routing
    msexch_routing = 691, 'msexch-routing', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hyperwave-ISP
    #: - [UDP] Hyperwave-ISP
    hyperwave_isp = 692, 'hyperwave-isp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] almanid Connection Endpoint
    #: - [UDP] almanid Connection Endpoint
    connendp = 693, 'connendp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ha-cluster
    #: - [UDP] ha-cluster
    ha_cluster = 694, 'ha-cluster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IEEE-MMS-SSL
    #: - [UDP] IEEE-MMS-SSL
    ieee_mms_ssl = 695, 'ieee-mms-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RUSHD
    #: - [UDP] RUSHD
    rushd = 696, 'rushd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UUIDGEN
    #: - [UDP] UUIDGEN
    uuidgen = 697, 'uuidgen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OLSR
    #: - [UDP] OLSR
    olsr = 698, 'olsr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Access Network
    #: - [UDP] Access Network
    accessnetwork = 699, 'accessnetwork', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Link Management Protocol (LMP) [:rfc:`4204`]
    #: - [UDP] Link Management Protocol (LMP) [:rfc:`4204`]
    lmp = 701, 'lmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IRIS over BEEP [:rfc:`3983`]
    #: - [UDP] IRIS over BEEP [:rfc:`3983`]
    iris_beep = 702, 'iris-beep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_703 = 703, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] errlog copy/server daemon
    #: - [UDP] errlog copy/server daemon
    elcsd = 704, 'elcsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AgentX
    #: - [UDP] AgentX
    agentx = 705, 'agentx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SILC
    #: - [UDP] SILC
    silc = 706, 'silc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Borland DSJ
    #: - [UDP] Borland DSJ
    borland_dsj = 707, 'borland-dsj', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_708 = 708, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Entrust Key Management Service Handler
    #: - [UDP] Entrust Key Management Service Handler
    entrust_kmsh = 709, 'entrust-kmsh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Entrust Administration Service Handler
    #: - [UDP] Entrust Administration Service Handler
    entrust_ash = 710, 'entrust-ash', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cisco TDP
    #: - [UDP] Cisco TDP
    cisco_tdp = 711, 'cisco-tdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TBRPF [:rfc:`3684`]
    #: - [UDP] TBRPF [:rfc:`3684`]
    tbrpf = 712, 'tbrpf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IRIS over XPC
    #: - [UDP] IRIS over XPC
    iris_xpc = 713, 'iris-xpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IRIS over XPCS [:rfc:`4992`]
    #: - [UDP] IRIS over XPCS [:rfc:`4992`]
    iris_xpcs = 714, 'iris-xpcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IRIS-LWZ [:rfc:`4993`]
    #: - [UDP] IRIS-LWZ [:rfc:`4993`]
    iris_lwz = 715, 'iris-lwz', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] PANA Messages [:rfc:`5191`]
    pana = 716, 'pana', TransportProtocol.get('udp')

    #: - [TCP] IBM NetView DM/6000 Server/Client
    #: - [UDP] IBM NetView DM/6000 Server/Client
    netviewdm1 = 729, 'netviewdm1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM NetView DM/6000 send/tcp
    #: - [UDP] IBM NetView DM/6000 send/tcp
    netviewdm2 = 730, 'netviewdm2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM NetView DM/6000 receive/tcp
    #: - [UDP] IBM NetView DM/6000 receive/tcp
    netviewdm3 = 731, 'netviewdm3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netGW
    #: - [UDP] netGW
    netgw = 741, 'netgw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network based Rev. Cont. Sys.
    #: - [UDP] Network based Rev. Cont. Sys.
    netrcs = 742, 'netrcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_743 = 743, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Flexible License Manager
    #: - [UDP] Flexible License Manager
    flexlm = 744, 'flexlm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Device Control
    #: - [UDP] Fujitsu Device Control
    fujitsu_dev = 747, 'fujitsu-dev', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Russell Info Sci Calendar Manager
    #: - [UDP] Russell Info Sci Calendar Manager
    ris_cm = 748, 'ris-cm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] kerberos administration
    #: - [UDP] kerberos administration
    kerberos_adm = 749, 'kerberos-adm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    rfile = 750, 'rfile', TransportProtocol.get('tcp')

    #: [UDP]
    loadav = 750, 'loadav', TransportProtocol.get('udp')

    #: [UDP] kerberos version iv
    kerberos_iv = 750, 'kerberos-iv', TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    pump = 751, 'pump', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    qrh = 752, 'qrh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    rrh = 753, 'rrh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] send
    #: - [UDP] send
    tell = 754, 'tell', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    nlogin = 758, 'nlogin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    con = 759, 'con', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    ns = 760, 'ns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    rxe = 761, 'rxe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    quotad = 762, 'quotad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    cycleserv = 763, 'cycleserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    omserv = 764, 'omserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    webster = 765, 'webster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_766 = 766, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] phone
    #: - [UDP] phone
    phonebook = 767, 'phonebook', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_768 = 768, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP]
    #: - [UDP]
    vid = 769, 'vid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    cadlock = 770, 'cadlock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    rtip = 771, 'rtip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    cycleserv2 = 772, 'cycleserv2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    submit = 773, 'submit', TransportProtocol.get('tcp')

    #: [UDP]
    notify = 773, 'notify', TransportProtocol.get('udp')

    #: [TCP]
    rpasswd = 774, 'rpasswd', TransportProtocol.get('tcp')

    #: - [UDP] IANA assigned this well-formed service name as a replacement for
    #: "acmaint_dbd".
    #: - [UDP]
    acmaint_dbd = 774, 'acmaint-dbd', TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [TCP]
    entomb = 775, 'entomb', TransportProtocol.get('tcp')

    #: - [UDP] IANA assigned this well-formed service name as a replacement for
    #: "acmaint_transd".
    #: - [UDP]
    acmaint_transd = 775, 'acmaint-transd', TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    wpages = 776, 'wpages', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multiling HTTP
    #: - [UDP] Multiling HTTP
    multiling_http = 777, 'multiling-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    wpgs = 780, 'wpgs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_786 = 786, 'unassigned', TransportProtocol.get('undefined')

    #: [N/A] Unassigned
    unassigned_787 = 787, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] IANA assigned this well-formed service name as a replacement for
    #: "mdbs_daemon".
    #: - [TCP]
    #: - [UDP] IANA assigned this well-formed service name as a replacement for
    #: "mdbs_daemon".
    #: - [UDP]
    mdbs_daemon = 800, 'mdbs-daemon', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    device = 801, 'device', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Modbus Application Protocol Secure
    #: - [UDP] Modbus Application Protocol Secure
    mbap_s = 802, 'mbap-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FCP
    #: - [UDP] FCP Datagram
    fcp_udp = 810, 'fcp-udp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] itm-mcell-s
    #: - [UDP] itm-mcell-s
    itm_mcell_s = 828, 'itm-mcell-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PKIX-3 CA/RA
    #: - [UDP] PKIX-3 CA/RA
    pkix_3_ca_ra = 829, 'pkix-3-ca-ra', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETCONF over SSH [:rfc:`6242`]
    #: - [UDP] NETCONF over SSH [:rfc:`6242`]
    netconf_ssh = 830, 'netconf-ssh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETCONF over BEEP [:rfc:`4744`]
    #: - [UDP] NETCONF over BEEP [:rfc:`4744`]
    netconf_beep = 831, 'netconf-beep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETCONF for SOAP over HTTPS [:rfc:`4743`]
    #: - [UDP] NETCONF for SOAP over HTTPS [:rfc:`4743`]
    netconfsoaphttp = 832, 'netconfsoaphttp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETCONF for SOAP over BEEP [:rfc:`4743`]
    #: - [UDP] NETCONF for SOAP over BEEP [:rfc:`4743`]
    netconfsoapbeep = 833, 'netconfsoapbeep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dhcp-failover 2
    #: - [UDP] dhcp-failover 2
    dhcp_failover2 = 847, 'dhcp-failover2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GDOI [:rfc:`3547`]
    #: - [UDP] GDOI [:rfc:`3547`]
    gdoi = 848, 'gdoi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNS query-response protocol run over TLS [:rfc:`7858`]
    #: - [UDP] DNS query-response protocol run over DTLS or QUIC
    #: [:rfc:`7858`][:rfc:`8094`][:rfc:`9250`]
    domain_s = 853, 'domain-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dynamic Link Exchange Protocol (DLEP) [:rfc:`8175`]
    #: - [UDP] Dynamic Link Exchange Protocol (DLEP) [:rfc:`8175`]
    dlep = 854, 'dlep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iSCSI [:rfc:`7143`]
    #: - [UDP] iSCSI [:rfc:`7143`]
    iscsi = 860, 'iscsi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] OWAMP-Control [:rfc:`4656`]
    owamp_control = 861, 'owamp-control', TransportProtocol.get('tcp')

    #: [UDP] OWAMP-Test [:rfc:`8545`]
    owamp_test = 861, 'owamp-test', TransportProtocol.get('udp')

    #: [TCP] TWAMP-Control [:rfc:`5357`]
    twamp_control = 862, 'twamp-control', TransportProtocol.get('tcp')

    #: [UDP] TWAMP-Test Receiver Port [:rfc:`8545`]
    twamp_test = 862, 'twamp-test', TransportProtocol.get('udp')

    #: - [TCP] rsync
    #: - [UDP] rsync
    rsync = 873, 'rsync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICL coNETion locate server
    #: - [UDP] ICL coNETion locate server
    iclcnet_locate = 886, 'iclcnet-locate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICL coNETion server info IANA assigned this well-formed service name
    #: as a replacement for "iclcnet_svinfo".
    #: - [TCP] ICL coNETion server info
    #: - [UDP] ICL coNETion server info IANA assigned this well-formed service name
    #: as a replacement for "iclcnet_svinfo".
    #: - [UDP] ICL coNETion server info
    iclcnet_svinfo = 887, 'iclcnet-svinfo', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] AccessBuilder
    #: - [UDP] AccessBuilder
    accessbuilder = 888, 'accessbuilder', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CD Database Protocol
    cddbp = 888, 'cddbp', TransportProtocol.get('tcp')

    #: - [TCP] OMG Initial Refs
    #: - [UDP] OMG Initial Refs
    omginitialrefs = 900, 'omginitialrefs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMPNAMERES
    #: - [UDP] SMPNAMERES
    smpnameres = 901, 'smpnameres', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] self documenting Telnet Door
    #: - [UDP] self documenting Door: send 0x00 for info
    ideafarm_door = 902, 'ideafarm-door', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] self documenting Telnet Panic Door
    #: - [UDP] self documenting Panic Door: send 0x00 for info
    ideafarm_panic = 903, 'ideafarm-panic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kerberized Internet Negotiation of Keys (KINK) [:rfc:`4430`]
    #: - [UDP] Kerberized Internet Negotiation of Keys (KINK) [:rfc:`4430`]
    kink = 910, 'kink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xact-backup
    #: - [UDP] xact-backup
    xact_backup = 911, 'xact-backup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APEX relay-relay service
    #: - [UDP] APEX relay-relay service
    apex_mesh = 912, 'apex-mesh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APEX endpoint-relay service [:rfc:`3340`]
    #: - [UDP] APEX endpoint-relay service [:rfc:`3340`]
    apex_edge = 913, 'apex-edge', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_914 = 914, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Routing in Fat Trees Link Information Elements (TEMPORARY - registered
    #: 2023-02-17, expires 2024-02-17) [draft-ietf-rift-rift-16]
    rift_lies = 914, 'rift-lies', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_915 = 915, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Routing in Fat Trees Topology Information Elements (TEMPORARY -
    #: registered 2023-02-17, expires 2024-02-17) [draft-ietf-rift-rift-16]
    rift_ties = 915, 'rift-ties', TransportProtocol.get('udp')

    #: [TCP] BIND9 remote name daemon controller
    rndc = 953, 'rndc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_953 = 953, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ftp protocol, data, over TLS/SSL
    #: - [UDP] ftp protocol, data, over TLS/SSL
    ftps_data = 989, 'ftps-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ftp protocol, control, over TLS/SSL
    #: - [UDP] ftp protocol, control, over TLS/SSL
    ftps = 990, 'ftps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netnews Administration System [:rfc:`4707`]
    #: - [UDP] Netnews Administration System [:rfc:`4707`]
    nas = 991, 'nas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] telnet protocol over TLS/SSL
    #: - [UDP] telnet protocol over TLS/SSL
    telnets = 992, 'telnets', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] IMAP over TLS protocol [:rfc:`3501`][:rfc:`8314`][:rfc:`9051`]
    imaps = 993, 'imaps', TransportProtocol.get('tcp')

    #: [UDP] Reserved [:rfc:`9051`]
    reserved_993 = 993, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_994 = 994, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] POP3 over TLS protocol [:rfc:`8314`]
    #: - [UDP] pop3 protocol over TLS/SSL (was spop3)
    pop3s = 995, 'pop3s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vsinet
    #: - [UDP] vsinet
    vsinet = 996, 'vsinet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    maitrd = 997, 'maitrd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    busboy = 998, 'busboy', TransportProtocol.get('tcp')

    #: [UDP]
    puparp = 998, 'puparp', TransportProtocol.get('udp')

    #: [TCP]
    garcon = 999, 'garcon', TransportProtocol.get('tcp')

    #: [UDP] Applix ac
    applix = 999, 'applix', TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    puprouter = 999, 'puprouter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    cadlock2 = 1000, 'cadlock2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] HTTP Web Push [:rfc:`8030`]
    webpush = 1001, 'webpush', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_1001 = 1001, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Possibly used by Sun Solaris????
    possibly_used_by_sun_solaris = 1008, 'possibly_used_by_sun_solaris', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_1009 = 1009, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] surf
    #: - [UDP] surf
    surf = 1010, 'surf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RFC3692-style Experiment 1 [1][:rfc:`4727`][:rfc:`6335`]
    #: - [UDP] RFC3692-style Experiment 1 [1][:rfc:`4727`][:rfc:`6335`]
    #: - [SCTP] RFC3692-style Experiment 1 [1][:rfc:`4727`][:rfc:`6335`]
    #: - [DCCP] RFC3692-style Experiment 1 [1][:rfc:`4727`][:rfc:`6335`]
    exp1 = 1021, 'exp1', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp') | TransportProtocol.get('dccp')

    #: - [TCP] RFC3692-style Experiment 2 [1][:rfc:`4727`][:rfc:`6335`]
    #: - [UDP] RFC3692-style Experiment 2 [1][:rfc:`4727`][:rfc:`6335`]
    #: - [SCTP] RFC3692-style Experiment 2 [1][:rfc:`4727`][:rfc:`6335`]
    #: - [DCCP] RFC3692-style Experiment 2 [1][:rfc:`4727`][:rfc:`6335`]
    exp2 = 1022, 'exp2', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp') | TransportProtocol.get('dccp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_1023 = 1023, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_1024 = 1024, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] network blackjack
    #: - [UDP] network blackjack
    blackjack = 1025, 'blackjack', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Calendar Access Protocol
    #: - [UDP] Calendar Access Protocol
    cap = 1026, 'cap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] IPv6 Behind NAT44 CPEs [:rfc:`6751`]
    AppType_6a44 = 1027, '6a44', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_1027 = 1027, 'reserved', TransportProtocol.get('tcp')

    #: [N/A] Deprecated
    deprecated = 1028, 'deprecated', TransportProtocol.get('undefined')

    #: - [TCP] Solid Mux Server
    #: - [UDP] Solid Mux Server
    solid_mux = 1029, 'solid-mux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Reserved
    reserved_1030 = 1030, 'reserved', TransportProtocol.get('undefined')

    #: [N/A] Reserved
    reserved_1031 = 1031, 'reserved', TransportProtocol.get('undefined')

    #: [N/A] Reserved
    reserved_1032 = 1032, 'reserved', TransportProtocol.get('undefined')

    #: - [TCP] local netinfo port
    #: - [UDP] local netinfo port
    netinfo_local = 1033, 'netinfo-local', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ActiveSync Notifications
    #: - [UDP] ActiveSync Notifications
    activesync = 1034, 'activesync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MX-XR RPC
    #: - [UDP] MX-XR RPC
    mxxrlogin = 1035, 'mxxrlogin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nebula Secure Segment Transfer Protocol
    #: - [UDP] Nebula Secure Segment Transfer Protocol
    nsstp = 1036, 'nsstp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMS
    #: - [UDP] AMS
    ams = 1037, 'ams', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Message Tracking Query Protocol [:rfc:`3887`]
    #: - [UDP] Message Tracking Query Protocol [:rfc:`3887`]
    mtqp = 1038, 'mtqp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Streamlined Blackhole
    #: - [UDP] Streamlined Blackhole
    sbl = 1039, 'sbl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netarx Netcare
    #: - [UDP] Netarx Netcare
    netarx = 1040, 'netarx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AK2 Product
    #: - [UDP] AK2 Product
    danf_ak2 = 1041, 'danf-ak2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Subnet Roaming
    #: - [UDP] Subnet Roaming
    afrog = 1042, 'afrog', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BOINC Client Control
    #: - [UDP] BOINC Client Control
    boinc_client = 1043, 'boinc-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dev Consortium Utility
    #: - [UDP] Dev Consortium Utility
    dcutility = 1044, 'dcutility', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fingerprint Image Transfer Protocol
    #: - [UDP] Fingerprint Image Transfer Protocol
    fpitp = 1045, 'fpitp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WebFilter Remote Monitor
    #: - [UDP] WebFilter Remote Monitor
    wfremotertm = 1046, 'wfremotertm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun's NEO Object Request Broker
    #: - [UDP] Sun's NEO Object Request Broker
    neod1 = 1047, 'neod1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun's NEO Object Request Broker
    #: - [UDP] Sun's NEO Object Request Broker
    neod2 = 1048, 'neod2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tobit David Postman VPMN
    #: - [UDP] Tobit David Postman VPMN
    td_postman = 1049, 'td-postman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CORBA Management Agent
    #: - [UDP] CORBA Management Agent
    cma = 1050, 'cma', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Optima VNET
    #: - [UDP] Optima VNET
    optima_vnet = 1051, 'optima-vnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dynamic DNS Tools
    #: - [UDP] Dynamic DNS Tools
    ddt = 1052, 'ddt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Assistant (RA)
    #: - [UDP] Remote Assistant (RA)
    remote_as = 1053, 'remote-as', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BRVREAD
    #: - [UDP] BRVREAD
    brvread = 1054, 'brvread', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ANSYS - License Manager
    #: - [UDP] ANSYS - License Manager
    ansyslmd = 1055, 'ansyslmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VFO
    #: - [UDP] VFO
    vfo = 1056, 'vfo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STARTRON
    #: - [UDP] STARTRON
    startron = 1057, 'startron', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nim
    #: - [UDP] nim
    nim = 1058, 'nim', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nimreg
    #: - [UDP] nimreg
    nimreg = 1059, 'nimreg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] POLESTAR
    #: - [UDP] POLESTAR
    polestar = 1060, 'polestar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KIOSK
    #: - [UDP] KIOSK
    kiosk = 1061, 'kiosk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Veracity
    #: - [UDP] Veracity
    veracity = 1062, 'veracity', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KyoceraNetDev
    #: - [UDP] KyoceraNetDev
    kyoceranetdev = 1063, 'kyoceranetdev', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JSTEL
    #: - [UDP] JSTEL
    jstel = 1064, 'jstel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SYSCOMLAN
    #: - [UDP] SYSCOMLAN
    syscomlan = 1065, 'syscomlan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FPO-FNS
    #: - [UDP] FPO-FNS
    fpo_fns = 1066, 'fpo-fns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Installation Bootstrap Proto. Serv. IANA assigned this well-formed
    #: service name as a replacement for "instl_boots".
    #: - [TCP] Installation Bootstrap Proto. Serv.
    #: - [UDP] Installation Bootstrap Proto. Serv. IANA assigned this well-formed
    #: service name as a replacement for "instl_boots".
    #: - [UDP] Installation Bootstrap Proto. Serv.
    instl_boots = 1067, 'instl-boots', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Installation Bootstrap Proto. Cli. IANA assigned this well-formed
    #: service name as a replacement for "instl_bootc".
    #: - [TCP] Installation Bootstrap Proto. Cli.
    #: - [UDP] Installation Bootstrap Proto. Cli. IANA assigned this well-formed
    #: service name as a replacement for "instl_bootc".
    #: - [UDP] Installation Bootstrap Proto. Cli.
    instl_bootc = 1068, 'instl-bootc', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] COGNEX-INSIGHT
    #: - [UDP] COGNEX-INSIGHT
    cognex_insight = 1069, 'cognex-insight', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GMRUpdateSERV
    #: - [UDP] GMRUpdateSERV
    gmrupdateserv = 1070, 'gmrupdateserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BSQUARE-VOIP
    #: - [UDP] BSQUARE-VOIP
    bsquare_voip = 1071, 'bsquare-voip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CARDAX
    #: - [UDP] CARDAX
    cardax = 1072, 'cardax', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bridge Control
    #: - [UDP] Bridge Control
    bridgecontrol = 1073, 'bridgecontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Warmspot Management Protocol
    #: - [UDP] Warmspot Management Protocol
    warmspotmgmt = 1074, 'warmspotmgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RDRMSHC
    #: - [UDP] RDRMSHC
    rdrmshc = 1075, 'rdrmshc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DAB STI-C
    #: - [UDP] DAB STI-C
    dab_sti_c = 1076, 'dab-sti-c', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IMGames
    #: - [UDP] IMGames
    imgames = 1077, 'imgames', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avocent Proxy Protocol
    #: - [UDP] Avocent Proxy Protocol
    avocent_proxy = 1078, 'avocent-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASPROVATalk
    #: - [UDP] ASPROVATalk
    asprovatalk = 1079, 'asprovatalk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Socks
    #: - [UDP] Socks
    socks = 1080, 'socks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PVUNIWIEN
    #: - [UDP] PVUNIWIEN
    pvuniwien = 1081, 'pvuniwien', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMT-ESD-PROT
    #: - [UDP] AMT-ESD-PROT
    amt_esd_prot = 1082, 'amt-esd-prot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Anasoft License Manager
    #: - [UDP] Anasoft License Manager
    ansoft_lm_1 = 1083, 'ansoft-lm-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Anasoft License Manager
    #: - [UDP] Anasoft License Manager
    ansoft_lm_2 = 1084, 'ansoft-lm-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Web Objects
    #: - [UDP] Web Objects
    webobjects = 1085, 'webobjects', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CPL Scrambler Logging
    #: - [UDP] CPL Scrambler Logging
    cplscrambler_lg = 1086, 'cplscrambler-lg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CPL Scrambler Internal
    #: - [UDP] CPL Scrambler Internal
    cplscrambler_in = 1087, 'cplscrambler-in', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CPL Scrambler Alarm Log
    #: - [UDP] CPL Scrambler Alarm Log
    cplscrambler_al = 1088, 'cplscrambler-al', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FF Annunciation
    #: - [UDP] FF Annunciation
    ff_annunc = 1089, 'ff-annunc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FF Fieldbus Message Specification
    #: - [UDP] FF Fieldbus Message Specification
    ff_fms = 1090, 'ff-fms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FF System Management
    #: - [UDP] FF System Management
    ff_sm = 1091, 'ff-sm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Open Business Reporting Protocol
    #: - [UDP] Open Business Reporting Protocol
    obrpd = 1092, 'obrpd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PROOFD
    #: - [UDP] PROOFD
    proofd = 1093, 'proofd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ROOTD
    #: - [UDP] ROOTD
    rootd = 1094, 'rootd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NICELink
    #: - [UDP] NICELink
    nicelink = 1095, 'nicelink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Common Name Resolution Protocol
    #: - [UDP] Common Name Resolution Protocol
    cnrprotocol = 1096, 'cnrprotocol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun Cluster Manager
    #: - [UDP] Sun Cluster Manager
    sunclustermgr = 1097, 'sunclustermgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RMI Activation
    #: - [UDP] RMI Activation
    rmiactivation = 1098, 'rmiactivation', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RMI Registry
    #: - [UDP] RMI Registry
    rmiregistry = 1099, 'rmiregistry', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCTP
    #: - [UDP] MCTP
    mctp = 1100, 'mctp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PT2-DISCOVER
    #: - [UDP] PT2-DISCOVER
    pt2_discover = 1101, 'pt2-discover', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ADOBE SERVER 1
    #: - [UDP] ADOBE SERVER 1
    adobeserver_1 = 1102, 'adobeserver-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ADOBE SERVER 2
    #: - [UDP] ADOBE SERVER 2
    adobeserver_2 = 1103, 'adobeserver-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XRL
    #: - [UDP] XRL
    xrl = 1104, 'xrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FTRANHC
    #: - [UDP] FTRANHC
    ftranhc = 1105, 'ftranhc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISOIPSIGPORT-1
    #: - [UDP] ISOIPSIGPORT-1
    isoipsigport_1 = 1106, 'isoipsigport-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISOIPSIGPORT-2
    #: - [UDP] ISOIPSIGPORT-2
    isoipsigport_2 = 1107, 'isoipsigport-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ratio-adp
    #: - [UDP] ratio-adp
    ratio_adp = 1108, 'ratio-adp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Reserved - IANA
    reserved_iana = 1109, 'reserved_iana', TransportProtocol.get('undefined')

    #: [TCP] Start web admin server
    webadmstart = 1110, 'webadmstart', TransportProtocol.get('tcp')

    #: [UDP] Client status info
    nfsd_keepalive = 1110, 'nfsd-keepalive', TransportProtocol.get('udp')

    #: - [TCP] LM Social Server
    #: - [UDP] LM Social Server
    lmsocialserver = 1111, 'lmsocialserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intelligent Communication Protocol
    #: - [UDP] Intelligent Communication Protocol
    icp = 1112, 'icp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Licklider Transmission Protocol [:rfc:`5326`]
    #: - [UDP] Licklider Transmission Protocol [:rfc:`5326`][:rfc:`7122`]
    #: - [DCCP] Licklider Transmission Protocol [:rfc:`7122`]
    ltp_deepspace = 1113, 'ltp-deepspace', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('dccp')

    #: - [TCP] Mini SQL
    #: - [UDP] Mini SQL
    mini_sql = 1114, 'mini-sql', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARDUS Transfer
    #: - [UDP] ARDUS Transfer
    ardus_trns = 1115, 'ardus-trns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARDUS Control
    #: - [UDP] ARDUS Control
    ardus_cntl = 1116, 'ardus-cntl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARDUS Multicast Transfer
    #: - [UDP] ARDUS Multicast Transfer
    ardus_mtrns = 1117, 'ardus-mtrns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SACRED [:rfc:`3767`]
    #: - [UDP] SACRED [:rfc:`3767`]
    sacred = 1118, 'sacred', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Battle.net Chat/Game Protocol
    #: - [UDP] Battle.net Chat/Game Protocol
    bnetgame = 1119, 'bnetgame', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Battle.net File Transfer Protocol
    #: - [UDP] Battle.net File Transfer Protocol
    bnetfile = 1120, 'bnetfile', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Datalode RMPP
    #: - [UDP] Datalode RMPP
    rmpp = 1121, 'rmpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] availant-mgr
    #: - [UDP] availant-mgr
    availant_mgr = 1122, 'availant-mgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Murray
    #: - [UDP] Murray
    murray = 1123, 'murray', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP VMM Control
    #: - [UDP] HP VMM Control
    hpvmmcontrol = 1124, 'hpvmmcontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP VMM Agent
    #: - [UDP] HP VMM Agent
    hpvmmagent = 1125, 'hpvmmagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP VMM Agent
    #: - [UDP] HP VMM Agent
    hpvmmdata = 1126, 'hpvmmdata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KWDB Remote Communication
    #: - [UDP] KWDB Remote Communication
    kwdb_commn = 1127, 'kwdb-commn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SAPHostControl over SOAP/HTTP
    #: - [UDP] SAPHostControl over SOAP/HTTP
    saphostctrl = 1128, 'saphostctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SAPHostControl over SOAP/HTTPS
    #: - [UDP] SAPHostControl over SOAP/HTTPS
    saphostctrls = 1129, 'saphostctrls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CAC App Service Protocol
    #: - [UDP] CAC App Service Protocol
    casp = 1130, 'casp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CAC App Service Protocol Encripted
    #: - [UDP] CAC App Service Protocol Encripted
    caspssl = 1131, 'caspssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KVM-via-IP Management Service
    #: - [UDP] KVM-via-IP Management Service
    kvm_via_ip = 1132, 'kvm-via-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Flow Network
    #: - [UDP] Data Flow Network
    dfn = 1133, 'dfn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MicroAPL APLX
    #: - [UDP] MicroAPL APLX
    aplx = 1134, 'aplx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OmniVision Communication Service
    #: - [UDP] OmniVision Communication Service
    omnivision = 1135, 'omnivision', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HHB Gateway Control
    #: - [UDP] HHB Gateway Control
    hhb_gateway = 1136, 'hhb-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRIM Workgroup Service
    #: - [UDP] TRIM Workgroup Service
    trim = 1137, 'trim', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] encrypted admin requests IANA assigned this well-formed service name
    #: as a replacement for "encrypted_admin".
    #: - [TCP] encrypted admin requests
    #: - [UDP] encrypted admin requests IANA assigned this well-formed service name
    #: as a replacement for "encrypted_admin".
    #: - [UDP] encrypted admin requests
    encrypted_admin = 1138, 'encrypted-admin', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Enterprise Virtual Manager
    #: - [UDP] Enterprise Virtual Manager
    evm = 1139, 'evm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AutoNOC Network Operations Protocol
    #: - [UDP] AutoNOC Network Operations Protocol
    autonoc = 1140, 'autonoc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] User Message Service
    #: - [UDP] User Message Service
    mxomss = 1141, 'mxomss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] User Discovery Service
    #: - [UDP] User Discovery Service
    edtools = 1142, 'edtools', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Infomatryx Exchange
    #: - [UDP] Infomatryx Exchange
    imyx = 1143, 'imyx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fusion Script
    #: - [UDP] Fusion Script
    fuscript = 1144, 'fuscript', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] X9 iCue Show Control
    #: - [UDP] X9 iCue Show Control
    x9_icue = 1145, 'x9-icue', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] audit transfer
    #: - [UDP] audit transfer
    audit_transfer = 1146, 'audit-transfer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CAPIoverLAN
    #: - [UDP] CAPIoverLAN
    capioverlan = 1147, 'capioverlan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Elfiq Replication Service
    #: - [UDP] Elfiq Replication Service
    elfiq_repl = 1148, 'elfiq-repl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BlueView Sonar Service
    #: - [UDP] BlueView Sonar Service
    bvtsonar = 1149, 'bvtsonar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Blaze File Server
    #: - [UDP] Blaze File Server
    blaze = 1150, 'blaze', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unizensus Login Server
    #: - [UDP] Unizensus Login Server
    unizensus = 1151, 'unizensus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Winpopup LAN Messenger
    #: - [UDP] Winpopup LAN Messenger
    winpoplanmess = 1152, 'winpoplanmess', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ANSI C12.22 Port [:rfc:`6142`]
    #: - [UDP] ANSI C12.22 Port [:rfc:`6142`]
    c1222_acse = 1153, 'c1222-acse', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Community Service
    #: - [UDP] Community Service
    resacommunity = 1154, 'resacommunity', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network File Access
    #: - [UDP] Network File Access
    nfa = 1155, 'nfa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iasControl OMS
    #: - [UDP] iasControl OMS
    iascontrol_oms = 1156, 'iascontrol-oms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle iASControl
    #: - [UDP] Oracle iASControl
    iascontrol = 1157, 'iascontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dbControl OMS
    #: - [UDP] dbControl OMS
    dbcontrol_oms = 1158, 'dbcontrol-oms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle OMS
    #: - [UDP] Oracle OMS
    oracle_oms = 1159, 'oracle-oms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DB Lite Mult-User Server
    #: - [UDP] DB Lite Mult-User Server
    olsv = 1160, 'olsv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Health Polling
    #: - [UDP] Health Polling
    health_polling = 1161, 'health-polling', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Health Trap
    #: - [UDP] Health Trap
    health_trap = 1162, 'health-trap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SmartDialer Data Protocol
    #: - [UDP] SmartDialer Data Protocol
    sddp = 1163, 'sddp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSM Proxy Service
    #: - [UDP] QSM Proxy Service
    qsm_proxy = 1164, 'qsm-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSM GUI Service
    #: - [UDP] QSM GUI Service
    qsm_gui = 1165, 'qsm-gui', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSM RemoteExec
    #: - [UDP] QSM RemoteExec
    qsm_remote = 1166, 'qsm-remote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cisco IP SLAs Control Protocol
    #: - [UDP] Cisco IP SLAs Control Protocol
    #: - [SCTP] Cisco IP SLAs Control Protocol
    cisco_ipsla = 1167, 'cisco-ipsla', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] VChat Conference Service
    #: - [UDP] VChat Conference Service
    vchat = 1168, 'vchat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRIPWIRE
    #: - [UDP] TRIPWIRE
    tripwire = 1169, 'tripwire', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AT+C License Manager
    #: - [UDP] AT+C License Manager
    atc_lm = 1170, 'atc-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AT+C FmiApplicationServer
    #: - [UDP] AT+C FmiApplicationServer
    atc_appserver = 1171, 'atc-appserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNA Protocol
    #: - [UDP] DNA Protocol
    dnap = 1172, 'dnap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D-Cinema Request-Response
    #: - [UDP] D-Cinema Request-Response
    d_cinema_rrp = 1173, 'd-cinema-rrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FlashNet Remote Admin
    #: - [UDP] FlashNet Remote Admin
    fnet_remote_ui = 1174, 'fnet-remote-ui', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dossier Server
    #: - [UDP] Dossier Server
    dossier = 1175, 'dossier', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Indigo Home Server
    #: - [UDP] Indigo Home Server
    indigo_server = 1176, 'indigo-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DKMessenger Protocol
    #: - [UDP] DKMessenger Protocol
    dkmessenger = 1177, 'dkmessenger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SGI Storage Manager
    #: - [UDP] SGI Storage Manager
    sgi_storman = 1178, 'sgi-storman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Backup To Neighbor
    #: - [UDP] Backup To Neighbor
    b2n = 1179, 'b2n', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Millicent Client Proxy
    #: - [UDP] Millicent Client Proxy
    mc_client = 1180, 'mc-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3Com Net Management
    #: - [UDP] 3Com Net Management
    AppType_3comnetman = 1181, '3comnetman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] AcceleNet Control
    accelenet = 1182, 'accelenet', TransportProtocol.get('tcp')

    #: [UDP] AcceleNet Data
    accelenet_data = 1182, 'accelenet-data', TransportProtocol.get('udp')

    #: - [TCP] LL Surfup HTTP
    #: - [UDP] LL Surfup HTTP
    llsurfup_http = 1183, 'llsurfup-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LL Surfup HTTPS
    #: - [UDP] LL Surfup HTTPS
    llsurfup_https = 1184, 'llsurfup-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Catchpole port
    #: - [UDP] Catchpole port
    catchpole = 1185, 'catchpole', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MySQL Cluster Manager
    #: - [UDP] MySQL Cluster Manager
    mysql_cluster = 1186, 'mysql-cluster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alias Service
    #: - [UDP] Alias Service
    alias = 1187, 'alias', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP Web Admin
    #: - [UDP] HP Web Admin
    hp_webadmin = 1188, 'hp-webadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unet Connection
    #: - [UDP] Unet Connection
    unet = 1189, 'unet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CommLinx GPS / AVL System
    #: - [UDP] CommLinx GPS / AVL System
    commlinx_avl = 1190, 'commlinx-avl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] General Parallel File System
    #: - [UDP] General Parallel File System
    gpfs = 1191, 'gpfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] caids sensors channel
    #: - [UDP] caids sensors channel
    caids_sensor = 1192, 'caids-sensor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Five Across Server
    #: - [UDP] Five Across Server
    fiveacross = 1193, 'fiveacross', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenVPN
    #: - [UDP] OpenVPN
    openvpn = 1194, 'openvpn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSF-1 clustering
    #: - [UDP] RSF-1 clustering
    rsf_1 = 1195, 'rsf-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Magic
    #: - [UDP] Network Magic
    netmagic = 1196, 'netmagic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Carrius Remote Access
    #: - [UDP] Carrius Remote Access
    carrius_rshell = 1197, 'carrius-rshell', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cajo reference discovery
    #: - [UDP] cajo reference discovery
    cajo_discovery = 1198, 'cajo-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DMIDI
    #: - [UDP] DMIDI
    dmidi = 1199, 'dmidi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCOL
    #: - [UDP] SCOL
    scol = 1200, 'scol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nucleus Sand Database Server
    #: - [UDP] Nucleus Sand Database Server
    nucleus_sand = 1201, 'nucleus-sand', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] caiccipc
    #: - [UDP] caiccipc
    caiccipc = 1202, 'caiccipc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] License Validation
    #: - [UDP] License Validation
    ssslic_mgr = 1203, 'ssslic-mgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Log Request Listener
    #: - [UDP] Log Request Listener
    ssslog_mgr = 1204, 'ssslog-mgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Accord-MGC
    #: - [UDP] Accord-MGC
    accord_mgc = 1205, 'accord-mgc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Anthony Data
    #: - [UDP] Anthony Data
    anthony_data = 1206, 'anthony-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MetaSage
    #: - [UDP] MetaSage
    metasage = 1207, 'metasage', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SEAGULL AIS
    #: - [UDP] SEAGULL AIS
    seagull_ais = 1208, 'seagull-ais', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPCD3
    #: - [UDP] IPCD3
    ipcd3 = 1209, 'ipcd3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EOSS
    #: - [UDP] EOSS
    eoss = 1210, 'eoss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Groove DPP
    #: - [UDP] Groove DPP
    groove_dpp = 1211, 'groove-dpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] lupa
    #: - [UDP] lupa
    lupa = 1212, 'lupa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Medtronic/Physio-Control LIFENET
    #: - [UDP] Medtronic/Physio-Control LIFENET
    mpc_lifenet = 1213, 'mpc-lifenet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KAZAA
    #: - [UDP] KAZAA
    kazaa = 1214, 'kazaa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] scanSTAT 1.0
    #: - [UDP] scanSTAT 1.0
    scanstat_1 = 1215, 'scanstat-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ETEBAC 5
    #: - [UDP] ETEBAC 5
    etebac5 = 1216, 'etebac5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HPSS NonDCE Gateway
    #: - [UDP] HPSS NonDCE Gateway
    hpss_ndapi = 1217, 'hpss-ndapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AeroFlight-ADs
    #: - [UDP] AeroFlight-ADs
    aeroflight_ads = 1218, 'aeroflight-ads', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AeroFlight-Ret
    #: - [UDP] AeroFlight-Ret
    aeroflight_ret = 1219, 'aeroflight-ret', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QT SERVER ADMIN
    #: - [UDP] QT SERVER ADMIN
    qt_serveradmin = 1220, 'qt-serveradmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SweetWARE Apps
    #: - [UDP] SweetWARE Apps
    sweetware_apps = 1221, 'sweetware-apps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNI R&D network
    #: - [UDP] SNI R&D network
    nerv = 1222, 'nerv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TrulyGlobal Protocol
    #: - [UDP] TrulyGlobal Protocol
    tgp = 1223, 'tgp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VPNz
    #: - [UDP] VPNz
    vpnz = 1224, 'vpnz', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SLINKYSEARCH
    #: - [UDP] SLINKYSEARCH
    slinkysearch = 1225, 'slinkysearch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STGXFWS
    #: - [UDP] STGXFWS
    stgxfws = 1226, 'stgxfws', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNS2Go
    #: - [UDP] DNS2Go
    dns2go = 1227, 'dns2go', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FLORENCE
    #: - [UDP] FLORENCE
    florence = 1228, 'florence', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ZENworks Tiered Electronic Distribution
    #: - [UDP] ZENworks Tiered Electronic Distribution
    zented = 1229, 'zented', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Periscope
    #: - [UDP] Periscope
    periscope = 1230, 'periscope', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] menandmice-lpm
    #: - [UDP] menandmice-lpm
    menandmice_lpm = 1231, 'menandmice-lpm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote systems monitoring
    #: - [UDP] Remote systems monitoring
    first_defense = 1232, 'first-defense', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Universal App Server
    #: - [UDP] Universal App Server
    univ_appserver = 1233, 'univ-appserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Infoseek Search Agent
    #: - [UDP] Infoseek Search Agent
    search_agent = 1234, 'search-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mosaicsyssvc1
    #: - [UDP] mosaicsyssvc1
    mosaicsyssvc1 = 1235, 'mosaicsyssvc1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bvcontrol
    #: - [UDP] bvcontrol
    bvcontrol = 1236, 'bvcontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tsdos390
    #: - [UDP] tsdos390
    tsdos390 = 1237, 'tsdos390', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hacl-qs
    #: - [UDP] hacl-qs
    hacl_qs = 1238, 'hacl-qs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NMSD
    #: - [UDP] NMSD
    nmsd = 1239, 'nmsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Instantia
    #: - [UDP] Instantia
    instantia = 1240, 'instantia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nessus
    #: - [UDP] nessus
    nessus = 1241, 'nessus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NMAS over IP
    #: - [UDP] NMAS over IP
    nmasoverip = 1242, 'nmasoverip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SerialGateway
    #: - [UDP] SerialGateway
    serialgateway = 1243, 'serialgateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isbconference1
    #: - [UDP] isbconference1
    isbconference1 = 1244, 'isbconference1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isbconference2
    #: - [UDP] isbconference2
    isbconference2 = 1245, 'isbconference2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] payrouter
    #: - [UDP] payrouter
    payrouter = 1246, 'payrouter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VisionPyramid
    #: - [UDP] VisionPyramid
    visionpyramid = 1247, 'visionpyramid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hermes
    #: - [UDP] hermes
    hermes = 1248, 'hermes', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mesa Vista Co
    #: - [UDP] Mesa Vista Co
    mesavistaco = 1249, 'mesavistaco', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] swldy-sias
    #: - [UDP] swldy-sias
    swldy_sias = 1250, 'swldy-sias', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] servergraph
    #: - [UDP] servergraph
    servergraph = 1251, 'servergraph', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bspne-pcc
    #: - [UDP] bspne-pcc
    bspne_pcc = 1252, 'bspne-pcc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] q55-pcc
    #: - [UDP] q55-pcc
    q55_pcc = 1253, 'q55-pcc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] de-noc
    #: - [UDP] de-noc
    de_noc = 1254, 'de-noc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] de-cache-query
    #: - [UDP] de-cache-query
    de_cache_query = 1255, 'de-cache-query', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] de-server
    #: - [UDP] de-server
    de_server = 1256, 'de-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Shockwave 2
    #: - [UDP] Shockwave 2
    shockwave2 = 1257, 'shockwave2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Open Network Library
    #: - [UDP] Open Network Library
    opennl = 1258, 'opennl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Open Network Library Voice
    #: - [UDP] Open Network Library Voice
    opennl_voice = 1259, 'opennl-voice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ibm-ssd
    #: - [UDP] ibm-ssd
    ibm_ssd = 1260, 'ibm-ssd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mpshrsv
    #: - [UDP] mpshrsv
    mpshrsv = 1261, 'mpshrsv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QNTS-ORB
    #: - [UDP] QNTS-ORB
    qnts_orb = 1262, 'qnts-orb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dka
    #: - [UDP] dka
    dka = 1263, 'dka', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PRAT
    #: - [UDP] PRAT
    prat = 1264, 'prat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DSSIAPI
    #: - [UDP] DSSIAPI
    dssiapi = 1265, 'dssiapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DELLPWRAPPKS
    #: - [UDP] DELLPWRAPPKS
    dellpwrappks = 1266, 'dellpwrappks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eTrust Policy Compliance
    #: - [UDP] eTrust Policy Compliance
    epc = 1267, 'epc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PROPEL-MSGSYS
    #: - [UDP] PROPEL-MSGSYS
    propel_msgsys = 1268, 'propel-msgsys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WATiLaPP
    #: - [UDP] WATiLaPP
    watilapp = 1269, 'watilapp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Operations Manager
    #: - [UDP] Microsoft Operations Manager
    opsmgr = 1270, 'opsmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eXcW
    #: - [UDP] eXcW
    excw = 1271, 'excw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSPMLockMgr
    #: - [UDP] CSPMLockMgr
    cspmlockmgr = 1272, 'cspmlockmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMC-Gateway
    #: - [UDP] EMC-Gateway
    emc_gateway = 1273, 'emc-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] t1distproc
    #: - [UDP] t1distproc
    t1distproc = 1274, 't1distproc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ivcollector
    #: - [UDP] ivcollector
    ivcollector = 1275, 'ivcollector', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_1276 = 1276, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mqs
    #: - [UDP] mqs
    miva_mqs = 1277, 'miva-mqs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dell Web Admin 1
    #: - [UDP] Dell Web Admin 1
    dellwebadmin_1 = 1278, 'dellwebadmin-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dell Web Admin 2
    #: - [UDP] Dell Web Admin 2
    dellwebadmin_2 = 1279, 'dellwebadmin-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pictrography
    #: - [UDP] Pictrography
    pictrography = 1280, 'pictrography', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] healthd
    #: - [UDP] healthd
    healthd = 1281, 'healthd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Emperion
    #: - [UDP] Emperion
    emperion = 1282, 'emperion', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Product Information
    #: - [UDP] Product Information
    productinfo = 1283, 'productinfo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IEE-QFX
    #: - [UDP] IEE-QFX
    iee_qfx = 1284, 'iee-qfx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] neoiface
    #: - [UDP] neoiface
    neoiface = 1285, 'neoiface', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netuitive
    #: - [UDP] netuitive
    netuitive = 1286, 'netuitive', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RouteMatch Com
    #: - [UDP] RouteMatch Com
    routematch = 1287, 'routematch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NavBuddy
    #: - [UDP] NavBuddy
    navbuddy = 1288, 'navbuddy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JWalkServer
    #: - [UDP] JWalkServer
    jwalkserver = 1289, 'jwalkserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WinJaServer
    #: - [UDP] WinJaServer
    winjaserver = 1290, 'winjaserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SEAGULLLMS
    #: - [UDP] SEAGULLLMS
    seagulllms = 1291, 'seagulllms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dsdn
    #: - [UDP] dsdn
    dsdn = 1292, 'dsdn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PKT-KRB-IPSec
    #: - [UDP] PKT-KRB-IPSec
    pkt_krb_ipsec = 1293, 'pkt-krb-ipsec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CMMdriver
    #: - [UDP] CMMdriver
    cmmdriver = 1294, 'cmmdriver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] End-by-Hop Transmission Protocol
    #: - [UDP] End-by-Hop Transmission Protocol
    ehtp = 1295, 'ehtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dproxy
    #: - [UDP] dproxy
    dproxy = 1296, 'dproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sdproxy
    #: - [UDP] sdproxy
    sdproxy = 1297, 'sdproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] lpcp
    #: - [UDP] lpcp
    lpcp = 1298, 'lpcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp-sci
    #: - [UDP] hp-sci
    hp_sci = 1299, 'hp-sci', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] H.323 Secure Call Control Signalling
    #: - [UDP] H.323 Secure Call Control Signalling
    h323hostcallsc = 1300, 'h323hostcallsc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_1301 = 1301, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_1302 = 1302, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sftsrv
    #: - [UDP] sftsrv
    sftsrv = 1303, 'sftsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Boomerang
    #: - [UDP] Boomerang
    boomerang = 1304, 'boomerang', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pe-mike
    #: - [UDP] pe-mike
    pe_mike = 1305, 'pe-mike', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RE-Conn-Proto
    #: - [UDP] RE-Conn-Proto
    re_conn_proto = 1306, 're-conn-proto', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pacmand
    #: - [UDP] Pacmand
    pacmand = 1307, 'pacmand', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Optical Domain Service Interconnect (ODSI)
    #: - [UDP] Optical Domain Service Interconnect (ODSI)
    odsi = 1308, 'odsi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JTAG server
    #: - [UDP] JTAG server
    jtag_server = 1309, 'jtag-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Husky
    #: - [UDP] Husky
    husky = 1310, 'husky', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RxMon
    #: - [UDP] RxMon
    rxmon = 1311, 'rxmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STI Envision
    #: - [UDP] STI Envision
    sti_envision = 1312, 'sti-envision', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC_PATROLDB IANA assigned this well-formed service name as a
    #: replacement for "bmc_patroldb".
    #: - [TCP] BMC_PATROLDB
    #: - [UDP] BMC_PATROLDB IANA assigned this well-formed service name as a
    #: replacement for "bmc_patroldb".
    #: - [UDP] BMC_PATROLDB
    bmc_patroldb = 1313, 'bmc-patroldb', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Photoscript Distributed Printing System
    #: - [UDP] Photoscript Distributed Printing System
    pdps = 1314, 'pdps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] E.L.S., Event Listener Service
    #: - [UDP] E.L.S., Event Listener Service
    els = 1315, 'els', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Exbit-ESCP
    #: - [UDP] Exbit-ESCP
    exbit_escp = 1316, 'exbit-escp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vrts-ipcserver
    #: - [UDP] vrts-ipcserver
    vrts_ipcserver = 1317, 'vrts-ipcserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] krb5gatekeeper
    #: - [UDP] krb5gatekeeper
    krb5gatekeeper = 1318, 'krb5gatekeeper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMX-ICSP
    #: - [UDP] AMX-ICSP
    amx_icsp = 1319, 'amx-icsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMX-AXBNET
    #: - [UDP] AMX-AXBNET
    amx_axbnet = 1320, 'amx-axbnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PIP
    #: - [UDP] PIP
    pip_321 = 321, 'pip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PIP
    #: - [UDP] PIP
    pip_1321 = 1321, 'pip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novation
    #: - [UDP] Novation
    novation = 1322, 'novation', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] brcd
    #: - [UDP] brcd
    brcd = 1323, 'brcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] delta-mcp
    #: - [UDP] delta-mcp
    delta_mcp = 1324, 'delta-mcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DX-Instrument
    #: - [UDP] DX-Instrument
    dx_instrument = 1325, 'dx-instrument', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WIMSIC
    #: - [UDP] WIMSIC
    wimsic = 1326, 'wimsic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ultrex
    #: - [UDP] Ultrex
    ultrex = 1327, 'ultrex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EWALL
    #: - [UDP] EWALL
    ewall = 1328, 'ewall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netdb-export
    #: - [UDP] netdb-export
    netdb_export = 1329, 'netdb-export', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] StreetPerfect
    #: - [UDP] StreetPerfect
    streetperfect = 1330, 'streetperfect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] intersan
    #: - [UDP] intersan
    intersan = 1331, 'intersan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCIA RXP-B
    #: - [UDP] PCIA RXP-B
    pcia_rxp_b = 1332, 'pcia-rxp-b', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Password Policy
    #: - [UDP] Password Policy
    passwrd_policy = 1333, 'passwrd-policy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] writesrv
    #: - [UDP] writesrv
    writesrv = 1334, 'writesrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Digital Notary Protocol
    #: - [UDP] Digital Notary Protocol
    digital_notary = 1335, 'digital-notary', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Instant Service Chat
    #: - [UDP] Instant Service Chat
    ischat = 1336, 'ischat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] menandmice DNS
    #: - [UDP] menandmice DNS
    menandmice_dns = 1337, 'menandmice-dns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WMC-log-svr
    #: - [UDP] WMC-log-svr
    wmc_log_svc = 1338, 'wmc-log-svc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] kjtsiteserver
    #: - [UDP] kjtsiteserver
    kjtsiteserver = 1339, 'kjtsiteserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NAAP
    #: - [UDP] NAAP
    naap = 1340, 'naap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QuBES
    #: - [UDP] QuBES
    qubes = 1341, 'qubes', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESBroker
    #: - [UDP] ESBroker
    esbroker = 1342, 'esbroker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] re101
    #: - [UDP] re101
    re101 = 1343, 're101', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICAP
    #: - [UDP] ICAP
    icap = 1344, 'icap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VPJP
    #: - [UDP] VPJP
    vpjp = 1345, 'vpjp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alta Analytics License Manager
    #: - [UDP] Alta Analytics License Manager
    alta_ana_lm = 1346, 'alta-ana-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] multi media conferencing
    #: - [UDP] multi media conferencing
    bbn_mmc = 1347, 'bbn-mmc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] multi media conferencing
    #: - [UDP] multi media conferencing
    bbn_mmx = 1348, 'bbn-mmx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Registration Network Protocol
    #: - [UDP] Registration Network Protocol
    sbook = 1349, 'sbook', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Registration Network Protocol
    #: - [UDP] Registration Network Protocol
    editbench = 1350, 'editbench', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Digital Tool Works (MIT)
    #: - [UDP] Digital Tool Works (MIT)
    equationbuilder = 1351, 'equationbuilder', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lotus Note
    #: - [UDP] Lotus Note
    lotusnote = 1352, 'lotusnote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Relief Consulting
    #: - [UDP] Relief Consulting
    relief = 1353, 'relief', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Five Across XSIP Network
    #: - [UDP] Five Across XSIP Network
    xsip_network = 1354, 'xsip-network', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intuitive Edge
    #: - [UDP] Intuitive Edge
    intuitive_edge = 1355, 'intuitive-edge', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CuillaMartin Company
    #: - [UDP] CuillaMartin Company
    cuillamartin = 1356, 'cuillamartin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Electronic PegBoard
    #: - [UDP] Electronic PegBoard
    pegboard = 1357, 'pegboard', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CONNLCLI
    #: - [UDP] CONNLCLI
    connlcli = 1358, 'connlcli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FTSRV
    #: - [UDP] FTSRV
    ftsrv = 1359, 'ftsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MIMER
    #: - [UDP] MIMER
    mimer = 1360, 'mimer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LinX
    #: - [UDP] LinX
    linx = 1361, 'linx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TimeFlies
    #: - [UDP] TimeFlies
    timeflies = 1362, 'timeflies', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network DataMover Requester
    #: - [UDP] Network DataMover Requester
    ndm_requester = 1363, 'ndm-requester', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network DataMover Server
    #: - [UDP] Network DataMover Server
    ndm_server = 1364, 'ndm-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Software Associates
    #: - [UDP] Network Software Associates
    adapt_sna = 1365, 'adapt-sna', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novell NetWare Comm Service Platform
    #: - [UDP] Novell NetWare Comm Service Platform
    netware_csp = 1366, 'netware-csp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DCS
    #: - [UDP] DCS
    dcs = 1367, 'dcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ScreenCast
    #: - [UDP] ScreenCast
    screencast = 1368, 'screencast', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GlobalView to Unix Shell
    #: - [UDP] GlobalView to Unix Shell
    gv_us = 1369, 'gv-us', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unix Shell to GlobalView
    #: - [UDP] Unix Shell to GlobalView
    us_gv = 1370, 'us-gv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Config Protocol
    #: - [UDP] Fujitsu Config Protocol
    fc_cli = 1371, 'fc-cli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Config Protocol
    #: - [UDP] Fujitsu Config Protocol
    fc_ser = 1372, 'fc-ser', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Chromagrafx
    #: - [UDP] Chromagrafx
    chromagrafx = 1373, 'chromagrafx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPI Software Systems
    #: - [UDP] EPI Software Systems
    molly = 1374, 'molly', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bytex
    #: - [UDP] Bytex
    bytex = 1375, 'bytex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Person to Person Software
    #: - [UDP] IBM Person to Person Software
    ibm_pps = 1376, 'ibm-pps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cichlid License Manager
    #: - [UDP] Cichlid License Manager
    cichlid = 1377, 'cichlid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Elan License Manager
    #: - [UDP] Elan License Manager
    elan = 1378, 'elan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Integrity Solutions
    #: - [UDP] Integrity Solutions
    dbreporter = 1379, 'dbreporter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Telesis Network License Manager
    #: - [UDP] Telesis Network License Manager
    telesis_licman = 1380, 'telesis-licman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apple Network License Manager
    #: - [UDP] Apple Network License Manager
    apple_licman = 1381, 'apple-licman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GW Hannaway Network License Manager
    #: - [UDP] GW Hannaway Network License Manager
    gwha = 1383, 'gwha', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Objective Solutions License Manager
    #: - [UDP] Objective Solutions License Manager
    os_licman = 1384, 'os-licman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Atex Publishing License Manager IANA assigned this well-formed service
    #: name as a replacement for "atex_elmd".
    #: - [TCP] Atex Publishing License Manager
    #: - [UDP] Atex Publishing License Manager IANA assigned this well-formed service
    #: name as a replacement for "atex_elmd".
    #: - [UDP] Atex Publishing License Manager
    atex_elmd = 1385, 'atex-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] CheckSum License Manager
    #: - [UDP] CheckSum License Manager
    checksum = 1386, 'checksum', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Computer Aided Design Software Inc LM
    #: - [UDP] Computer Aided Design Software Inc LM
    cadsi_lm = 1387, 'cadsi-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Objective Solutions DataBase Cache
    #: - [UDP] Objective Solutions DataBase Cache
    objective_dbc = 1388, 'objective-dbc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Document Manager
    #: - [UDP] Document Manager
    iclpv_dm = 1389, 'iclpv-dm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Storage Controller
    #: - [UDP] Storage Controller
    iclpv_sc = 1390, 'iclpv-sc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Storage Access Server
    #: - [UDP] Storage Access Server
    iclpv_sas = 1391, 'iclpv-sas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Print Manager
    #: - [UDP] Print Manager
    iclpv_pm = 1392, 'iclpv-pm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Log Server
    #: - [UDP] Network Log Server
    iclpv_nls = 1393, 'iclpv-nls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Log Client
    #: - [UDP] Network Log Client
    iclpv_nlc = 1394, 'iclpv-nlc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PC Workstation Manager software
    #: - [UDP] PC Workstation Manager software
    iclpv_wsm = 1395, 'iclpv-wsm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DVL Active Mail
    #: - [UDP] DVL Active Mail
    dvl_activemail = 1396, 'dvl-activemail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Audio Active Mail
    #: - [UDP] Audio Active Mail
    audio_activmail = 1397, 'audio-activmail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Video Active Mail
    #: - [UDP] Video Active Mail
    video_activmail = 1398, 'video-activmail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cadkey License Manager
    #: - [UDP] Cadkey License Manager
    cadkey_licman = 1399, 'cadkey-licman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cadkey Tablet Daemon
    #: - [UDP] Cadkey Tablet Daemon
    cadkey_tablet = 1400, 'cadkey-tablet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Goldleaf License Manager
    #: - [UDP] Goldleaf License Manager
    goldleaf_licman = 1401, 'goldleaf-licman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prospero Resource Manager
    #: - [UDP] Prospero Resource Manager
    prm_sm_np = 1402, 'prm-sm-np', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prospero Resource Manager
    #: - [UDP] Prospero Resource Manager
    prm_nm_np = 1403, 'prm-nm-np', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Infinite Graphics License Manager
    #: - [UDP] Infinite Graphics License Manager
    igi_lm = 1404, 'igi-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Remote Execution Starter
    #: - [UDP] IBM Remote Execution Starter
    ibm_res = 1405, 'ibm-res', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetLabs License Manager
    #: - [UDP] NetLabs License Manager
    netlabs_lm = 1406, 'netlabs-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] TIBET Data Server
    tibet_server = 1407, 'tibet-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_1407 = 1407, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Sophia License Manager
    #: - [UDP] Sophia License Manager
    sophia_lm = 1408, 'sophia-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Here License Manager
    #: - [UDP] Here License Manager
    here_lm = 1409, 'here-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HiQ License Manager
    #: - [UDP] HiQ License Manager
    hiq = 1410, 'hiq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AudioFile
    #: - [UDP] AudioFile
    af = 1411, 'af', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] InnoSys
    #: - [UDP] InnoSys
    innosys = 1412, 'innosys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Innosys-ACL
    #: - [UDP] Innosys-ACL
    innosys_acl = 1413, 'innosys-acl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM MQSeries
    #: - [UDP] IBM MQSeries
    ibm_mqseries = 1414, 'ibm-mqseries', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DBStar
    #: - [UDP] DBStar
    dbstar = 1415, 'dbstar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novell LU6.2 IANA assigned this well-formed service name as a
    #: replacement for "novell-lu6.2".
    #: - [TCP] Novell LU6.2
    #: - [UDP] Novell LU6.2 IANA assigned this well-formed service name as a
    #: replacement for "novell-lu6.2".
    #: - [UDP] Novell LU6.2
    novell_lu6_2 = 1416, 'novell-lu6-2', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Timbuktu Service 1 Port
    #: - [UDP] Timbuktu Service 1 Port
    timbuktu_srv1 = 1417, 'timbuktu-srv1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Timbuktu Service 2 Port
    #: - [UDP] Timbuktu Service 2 Port
    timbuktu_srv2 = 1418, 'timbuktu-srv2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Timbuktu Service 3 Port
    #: - [UDP] Timbuktu Service 3 Port
    timbuktu_srv3 = 1419, 'timbuktu-srv3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Timbuktu Service 4 Port
    #: - [UDP] Timbuktu Service 4 Port
    timbuktu_srv4 = 1420, 'timbuktu-srv4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gandalf License Manager
    #: - [UDP] Gandalf License Manager
    gandalf_lm = 1421, 'gandalf-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Autodesk License Manager
    #: - [UDP] Autodesk License Manager
    autodesk_lm = 1422, 'autodesk-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Essbase Arbor Software
    #: - [UDP] Essbase Arbor Software
    essbase = 1423, 'essbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hybrid Encryption Protocol
    #: - [UDP] Hybrid Encryption Protocol
    hybrid = 1424, 'hybrid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zion Software License Manager
    #: - [UDP] Zion Software License Manager
    zion_lm = 1425, 'zion-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Satellite-data Acquisition System 1
    #: - [UDP] Satellite-data Acquisition System 1
    sais = 1426, 'sais', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mloadd monitoring tool
    #: - [UDP] mloadd monitoring tool
    mloadd = 1427, 'mloadd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Informatik License Manager
    #: - [UDP] Informatik License Manager
    informatik_lm = 1428, 'informatik-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hypercom NMS
    #: - [UDP] Hypercom NMS
    nms = 1429, 'nms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hypercom TPDU
    #: - [UDP] Hypercom TPDU
    tpdu = 1430, 'tpdu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reverse Gossip Transport
    #: - [UDP] Reverse Gossip Transport
    rgtp = 1431, 'rgtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Blueberry Software License Manager
    #: - [UDP] Blueberry Software License Manager
    blueberry_lm = 1432, 'blueberry-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft-SQL-Server
    #: - [UDP] Microsoft-SQL-Server
    ms_sql_s = 1433, 'ms-sql-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft-SQL-Monitor
    #: - [UDP] Microsoft-SQL-Monitor
    ms_sql_m = 1434, 'ms-sql-m', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM CICS
    #: - [UDP] IBM CICS
    ibm_cics = 1435, 'ibm-cics', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Satellite-data Acquisition System 2
    #: - [UDP] Satellite-data Acquisition System 2
    saism = 1436, 'saism', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tabula
    #: - [UDP] Tabula
    tabula = 1437, 'tabula', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Eicon Security Agent/Server
    #: - [UDP] Eicon Security Agent/Server
    eicon_server = 1438, 'eicon-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Eicon X25/SNA Gateway
    #: - [UDP] Eicon X25/SNA Gateway
    eicon_x25 = 1439, 'eicon-x25', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Eicon Service Location Protocol
    #: - [UDP] Eicon Service Location Protocol
    eicon_slp = 1440, 'eicon-slp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cadis License Management
    #: - [UDP] Cadis License Management
    cadis_1 = 1441, 'cadis-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cadis License Management
    #: - [UDP] Cadis License Management
    cadis_2 = 1442, 'cadis-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Integrated Engineering Software
    #: - [UDP] Integrated Engineering Software
    ies_lm = 1443, 'ies-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Marcam  License Management
    #: - [UDP] Marcam  License Management
    marcam_lm = 1444, 'marcam-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Proxima License Manager
    #: - [UDP] Proxima License Manager
    proxima_lm = 1445, 'proxima-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Optical Research Associates License Manager
    #: - [UDP] Optical Research Associates License Manager
    ora_lm = 1446, 'ora-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Applied Parallel Research LM
    #: - [UDP] Applied Parallel Research LM
    apri_lm = 1447, 'apri-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenConnect License Manager
    #: - [UDP] OpenConnect License Manager
    oc_lm = 1448, 'oc-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PEport
    #: - [UDP] PEport
    peport = 1449, 'peport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tandem Distributed Workbench Facility
    #: - [UDP] Tandem Distributed Workbench Facility
    dwf = 1450, 'dwf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Information Management
    #: - [UDP] IBM Information Management
    infoman = 1451, 'infoman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GTE Government Systems License Man
    #: - [UDP] GTE Government Systems License Man
    gtegsc_lm = 1452, 'gtegsc-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Genie License Manager
    #: - [UDP] Genie License Manager
    genie_lm = 1453, 'genie-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] interHDL License Manager IANA assigned this well-formed service name
    #: as a replacement for "interhdl_elmd".
    #: - [TCP] interHDL License Manager
    #: - [UDP] interHDL License Manager IANA assigned this well-formed service name
    #: as a replacement for "interhdl_elmd".
    #: - [UDP] interHDL License Manager
    interhdl_elmd = 1454, 'interhdl-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] ESL License Manager
    #: - [UDP] ESL License Manager
    esl_lm = 1455, 'esl-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DCA
    #: - [UDP] DCA
    dca = 1456, 'dca', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Valisys License Manager
    #: - [UDP] Valisys License Manager
    valisys_lm = 1457, 'valisys-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nichols Research Corp.
    #: - [UDP] Nichols Research Corp.
    nrcabq_lm = 1458, 'nrcabq-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Proshare Notebook Application
    #: - [UDP] Proshare Notebook Application
    proshare1 = 1459, 'proshare1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Proshare Notebook Application
    #: - [UDP] Proshare Notebook Application
    proshare2 = 1460, 'proshare2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Wireless LAN IANA assigned this well-formed service name as a
    #: replacement for "ibm_wrless_lan".
    #: - [TCP] IBM Wireless LAN
    #: - [UDP] IBM Wireless LAN IANA assigned this well-formed service name as a
    #: replacement for "ibm_wrless_lan".
    #: - [UDP] IBM Wireless LAN
    ibm_wrless_lan = 1461, 'ibm-wrless-lan', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] World License Manager
    #: - [UDP] World License Manager
    world_lm = 1462, 'world-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nucleus
    #: - [UDP] Nucleus
    nucleus = 1463, 'nucleus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSL License Manager IANA assigned this well-formed service name as a
    #: replacement for "msl_lmd".
    #: - [TCP] MSL License Manager
    #: - [UDP] MSL License Manager IANA assigned this well-formed service name as a
    #: replacement for "msl_lmd".
    #: - [UDP] MSL License Manager
    msl_lmd = 1464, 'msl-lmd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Pipes Platform
    #: - [UDP] Pipes Platform
    pipes = 1465, 'pipes', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ocean Software License Manager
    #: - [UDP] Ocean Software License Manager
    oceansoft_lm = 1466, 'oceansoft-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Active Analysis Limited License Manager
    #: - [UDP] Active Analysis Limited License Manager
    aal_lm = 1469, 'aal-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Universal Analytics
    #: - [UDP] Universal Analytics
    uaiact = 1470, 'uaiact', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSDMBASE
    #: - [UDP] CSDMBASE
    csdmbase_1467 = 1467, 'csdmbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] csdmbase
    #: - [UDP] csdmbase
    csdmbase_1471 = 1471, 'csdmbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSDM
    #: - [UDP] CSDM
    csdm_1468 = 1468, 'csdm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] csdm
    #: - [UDP] csdm
    csdm_1472 = 1472, 'csdm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenMath
    #: - [UDP] OpenMath
    openmath = 1473, 'openmath', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Telefinder
    #: - [UDP] Telefinder
    telefinder = 1474, 'telefinder', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Taligent License Manager
    #: - [UDP] Taligent License Manager
    taligent_lm = 1475, 'taligent-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] clvm-cfg
    #: - [UDP] clvm-cfg
    clvm_cfg = 1476, 'clvm-cfg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ms-sna-server
    #: - [UDP] ms-sna-server
    ms_sna_server = 1477, 'ms-sna-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ms-sna-base
    #: - [UDP] ms-sna-base
    ms_sna_base = 1478, 'ms-sna-base', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dberegister
    #: - [UDP] dberegister
    dberegister = 1479, 'dberegister', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PacerForum
    #: - [UDP] PacerForum
    pacerforum = 1480, 'pacerforum', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AIRS
    #: - [UDP] AIRS
    airs = 1481, 'airs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Miteksys License Manager
    #: - [UDP] Miteksys License Manager
    miteksys_lm = 1482, 'miteksys-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AFS License Manager
    #: - [UDP] AFS License Manager
    afs = 1483, 'afs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Confluent License Manager
    #: - [UDP] Confluent License Manager
    confluent = 1484, 'confluent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LANSource
    #: - [UDP] LANSource
    lansource = 1485, 'lansource', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nms_topo_serv IANA assigned this well-formed service name as a
    #: replacement for "nms_topo_serv".
    #: - [TCP] nms_topo_serv
    #: - [UDP] nms_topo_serv IANA assigned this well-formed service name as a
    #: replacement for "nms_topo_serv".
    #: - [UDP] nms_topo_serv
    nms_topo_serv = 1486, 'nms-topo-serv', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] LocalInfoSrvr
    #: - [UDP] LocalInfoSrvr
    localinfosrvr = 1487, 'localinfosrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DocStor
    #: - [UDP] DocStor
    docstor = 1488, 'docstor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dmdocbroker
    #: - [UDP] dmdocbroker
    dmdocbroker = 1489, 'dmdocbroker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] insitu-conf
    #: - [UDP] insitu-conf
    insitu_conf = 1490, 'insitu-conf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_1491 = 1491, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] stone-design-1
    #: - [UDP] stone-design-1
    stone_design_1 = 1492, 'stone-design-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netmap_lm IANA assigned this well-formed service name as a replacement
    #: for "netmap_lm".
    #: - [TCP] netmap_lm
    #: - [UDP] netmap_lm IANA assigned this well-formed service name as a replacement
    #: for "netmap_lm".
    #: - [UDP] netmap_lm
    netmap_lm = 1493, 'netmap-lm', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] ica
    #: - [UDP] ica
    ica = 1494, 'ica', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cvc
    #: - [UDP] cvc
    cvc = 1495, 'cvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] liberty-lm
    #: - [UDP] liberty-lm
    liberty_lm = 1496, 'liberty-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rfx-lm
    #: - [UDP] rfx-lm
    rfx_lm = 1497, 'rfx-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sybase SQL Any
    #: - [UDP] Sybase SQL Any
    sybase_sqlany = 1498, 'sybase-sqlany', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Federico Heinz Consultora
    #: - [UDP] Federico Heinz Consultora
    fhc = 1499, 'fhc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VLSI License Manager
    #: - [UDP] VLSI License Manager
    vlsi_lm = 1500, 'vlsi-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Satellite-data Acquisition System 3
    #: - [UDP] Satellite-data Acquisition System 3
    saiscm = 1501, 'saiscm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Shiva
    #: - [UDP] Shiva
    shivadiscovery = 1502, 'shivadiscovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Databeam
    #: - [UDP] Databeam
    imtc_mcs = 1503, 'imtc-mcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EVB Software Engineering License Manager
    #: - [UDP] EVB Software Engineering License Manager
    evb_elm = 1504, 'evb-elm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Funk Software, Inc.
    #: - [UDP] Funk Software, Inc.
    funkproxy = 1505, 'funkproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Universal Time daemon (utcd)
    #: - [UDP] Universal Time daemon (utcd)
    utcd = 1506, 'utcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] symplex
    #: - [UDP] symplex
    symplex = 1507, 'symplex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] diagmond
    #: - [UDP] diagmond
    diagmond = 1508, 'diagmond', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Robcad, Ltd. License Manager
    #: - [UDP] Robcad, Ltd. License Manager
    robcad_lm = 1509, 'robcad-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Midland Valley Exploration Ltd. Lic. Man.
    #: - [UDP] Midland Valley Exploration Ltd. Lic. Man.
    mvx_lm = 1510, 'mvx-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3l-l1
    #: - [UDP] 3l-l1
    AppType_3l_l1 = 1511, '3l-l1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft's Windows Internet Name Service
    #: - [UDP] Microsoft's Windows Internet Name Service
    wins = 1512, 'wins', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Systems Business of America, Inc
    #: - [UDP] Fujitsu Systems Business of America, Inc
    fujitsu_dtc = 1513, 'fujitsu-dtc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Systems Business of America, Inc
    #: - [UDP] Fujitsu Systems Business of America, Inc
    fujitsu_dtcns = 1514, 'fujitsu-dtcns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ifor-protocol
    #: - [UDP] ifor-protocol
    ifor_protocol = 1515, 'ifor-protocol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Places Audio data
    #: - [UDP] Virtual Places Audio data
    vpad = 1516, 'vpad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Places Audio control
    #: - [UDP] Virtual Places Audio control
    vpac = 1517, 'vpac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Places Video data
    #: - [UDP] Virtual Places Video data
    vpvd = 1518, 'vpvd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Places Video control
    #: - [UDP] Virtual Places Video control
    vpvc = 1519, 'vpvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] atm zip office
    #: - [UDP] atm zip office
    atm_zip_office = 1520, 'atm-zip-office', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nCube License Manager
    #: - [UDP] nCube License Manager
    ncube_lm = 1521, 'ncube-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cichild
    #: - [UDP] cichild
    cichild_lm = 1523, 'cichild-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ingres
    #: - [UDP] ingres
    ingreslock = 1524, 'ingreslock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] oracle
    #: - [UDP] oracle
    orasrv = 1525, 'orasrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prospero Directory Service non-priv
    #: - [UDP] Prospero Directory Service non-priv
    prospero_np = 1525, 'prospero-np', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prospero Data Access Prot non-priv
    #: - [UDP] Prospero Data Access Prot non-priv
    pdap_np = 1526, 'pdap-np', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] oracle
    #: - [UDP] oracle
    tlisrv = 1527, 'tlisrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Not Only a Routeing Protocol
    #: - [UDP] Not Only a Routeing Protocol
    #: - [SCTP] Not Only a Routeing Protocol
    norp = 1528, 'norp', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] oracle
    #: - [UDP] oracle
    coauthor = 1529, 'coauthor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rap-service
    #: - [UDP] rap-service
    rap_service = 1530, 'rap-service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rap-listen
    #: - [UDP] rap-listen
    rap_listen = 1531, 'rap-listen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] miroconnect
    #: - [UDP] miroconnect
    miroconnect = 1532, 'miroconnect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Places Software
    #: - [UDP] Virtual Places Software
    virtual_places = 1533, 'virtual-places', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] micromuse-lm
    #: - [UDP] micromuse-lm
    micromuse_lm = 1534, 'micromuse-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ampr-info
    #: - [UDP] ampr-info
    ampr_info = 1535, 'ampr-info', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ampr-inter
    #: - [UDP] ampr-inter
    ampr_inter = 1536, 'ampr-inter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isi-lm
    #: - [UDP] isi-lm
    sdsc_lm = 1537, 'sdsc-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3ds-lm
    #: - [UDP] 3ds-lm
    AppType_3ds_lm = 1538, '3ds-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intellistor License Manager
    #: - [UDP] Intellistor License Manager
    intellistor_lm = 1539, 'intellistor-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rds
    #: - [UDP] rds
    rds = 1540, 'rds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rds2
    #: - [UDP] rds2
    rds2 = 1541, 'rds2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gridgen-elmd
    #: - [UDP] gridgen-elmd
    gridgen_elmd = 1542, 'gridgen-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] simba-cs
    #: - [UDP] simba-cs
    simba_cs = 1543, 'simba-cs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] aspeclmd
    #: - [UDP] aspeclmd
    aspeclmd = 1544, 'aspeclmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vistium-share
    #: - [UDP] vistium-share
    vistium_share = 1545, 'vistium-share', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] abbaccuray
    #: - [UDP] abbaccuray
    abbaccuray = 1546, 'abbaccuray', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] laplink
    #: - [UDP] laplink
    laplink = 1547, 'laplink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Axon License Manager
    #: - [UDP] Axon License Manager
    axon_lm = 1548, 'axon-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Shiva Hose
    shivahose = 1549, 'shivahose', TransportProtocol.get('tcp')

    #: [UDP] Shiva Sound
    shivasound = 1549, 'shivasound', TransportProtocol.get('udp')

    #: - [TCP] Image Storage license manager 3M Company
    #: - [UDP] Image Storage license manager 3M Company
    AppType_3m_image_lm = 1550, '3m-image-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HECMTL-DB
    #: - [UDP] HECMTL-DB
    hecmtl_db = 1551, 'hecmtl-db', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pciarray
    #: - [UDP] pciarray
    pciarray = 1552, 'pciarray', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sna-cs
    #: - [UDP] sna-cs
    sna_cs = 1553, 'sna-cs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CACI Products Company License Manager
    #: - [UDP] CACI Products Company License Manager
    caci_lm = 1554, 'caci-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] livelan
    #: - [UDP] livelan
    livelan = 1555, 'livelan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS Private Branch Exchange IANA assigned this well-formed service
    #: name as a replacement for "veritas_pbx".
    #: - [TCP] VERITAS Private Branch Exchange
    #: - [UDP] VERITAS Private Branch Exchange IANA assigned this well-formed service
    #: name as a replacement for "veritas_pbx".
    #: - [UDP] VERITAS Private Branch Exchange
    veritas_pbx = 1556, 'veritas-pbx', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] ArborText License Manager
    #: - [UDP] ArborText License Manager
    arbortext_lm = 1557, 'arbortext-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xingmpeg
    #: - [UDP] xingmpeg
    xingmpeg = 1558, 'xingmpeg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] web2host
    #: - [UDP] web2host
    web2host = 1559, 'web2host', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASCI-RemoteSHADOW
    #: - [UDP] ASCI-RemoteSHADOW
    asci_val = 1560, 'asci-val', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] facilityview
    #: - [UDP] facilityview
    facilityview = 1561, 'facilityview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pconnectmgr
    #: - [UDP] pconnectmgr
    pconnectmgr = 1562, 'pconnectmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cadabra License Manager
    #: - [UDP] Cadabra License Manager
    cadabra_lm = 1563, 'cadabra-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pay-Per-View
    #: - [UDP] Pay-Per-View
    pay_per_view = 1564, 'pay-per-view', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WinDD
    #: - [UDP] WinDD
    winddlb = 1565, 'winddlb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CORELVIDEO
    #: - [UDP] CORELVIDEO
    corelvideo = 1566, 'corelvideo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] jlicelmd
    #: - [UDP] jlicelmd
    jlicelmd = 1567, 'jlicelmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tsspmap
    #: - [UDP] tsspmap
    tsspmap = 1568, 'tsspmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ets
    #: - [UDP] ets
    ets = 1569, 'ets', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] orbixd
    #: - [UDP] orbixd
    orbixd = 1570, 'orbixd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle Remote Data Base
    #: - [UDP] Oracle Remote Data Base
    rdb_dbs_disp = 1571, 'rdb-dbs-disp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Chipcom License Manager
    #: - [UDP] Chipcom License Manager
    chip_lm = 1572, 'chip-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] itscomm-ns
    #: - [UDP] itscomm-ns
    itscomm_ns = 1573, 'itscomm-ns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mvel-lm
    #: - [UDP] mvel-lm
    mvel_lm = 1574, 'mvel-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] oraclenames
    #: - [UDP] oraclenames
    oraclenames = 1575, 'oraclenames', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Moldflow License Manager
    #: - [UDP] Moldflow License Manager
    moldflow_lm = 1576, 'moldflow-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hypercube-lm
    #: - [UDP] hypercube-lm
    hypercube_lm = 1577, 'hypercube-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jacobus License Manager
    #: - [UDP] Jacobus License Manager
    jacobus_lm = 1578, 'jacobus-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ioc-sea-lm
    #: - [UDP] ioc-sea-lm
    ioc_sea_lm = 1579, 'ioc-sea-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] tn-tl-r1
    tn_tl_r1 = 1580, 'tn-tl-r1', TransportProtocol.get('tcp')

    #: [UDP] tn-tl-r2
    tn_tl_r2 = 1580, 'tn-tl-r2', TransportProtocol.get('udp')

    #: - [TCP] MIL-2045-47001
    #: - [UDP] MIL-2045-47001
    mil_2045_47001 = 1581, 'mil-2045-47001', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSIMS
    #: - [UDP] MSIMS
    msims = 1582, 'msims', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] simbaexpress
    #: - [UDP] simbaexpress
    simbaexpress = 1583, 'simbaexpress', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tn-tl-fd2
    #: - [UDP] tn-tl-fd2
    tn_tl_fd2 = 1584, 'tn-tl-fd2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] intv
    #: - [UDP] intv
    intv = 1585, 'intv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ibm-abtact
    #: - [UDP] ibm-abtact
    ibm_abtact = 1586, 'ibm-abtact', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pra_elmd IANA assigned this well-formed service name as a replacement
    #: for "pra_elmd".
    #: - [TCP] pra_elmd
    #: - [UDP] pra_elmd IANA assigned this well-formed service name as a replacement
    #: for "pra_elmd".
    #: - [UDP] pra_elmd
    pra_elmd = 1587, 'pra-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] triquest-lm
    #: - [UDP] triquest-lm
    triquest_lm = 1588, 'triquest-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VQP
    #: - [UDP] VQP
    vqp = 1589, 'vqp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gemini-lm
    #: - [UDP] gemini-lm
    gemini_lm = 1590, 'gemini-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ncpm-pm
    #: - [UDP] ncpm-pm
    ncpm_pm = 1591, 'ncpm-pm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] commonspace
    #: - [UDP] commonspace
    commonspace = 1592, 'commonspace', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mainsoft-lm
    #: - [UDP] mainsoft-lm
    mainsoft_lm = 1593, 'mainsoft-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sixtrak
    #: - [UDP] sixtrak
    sixtrak = 1594, 'sixtrak', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] radio
    #: - [UDP] radio
    radio = 1595, 'radio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] radio-sm
    radio_sm = 1596, 'radio-sm', TransportProtocol.get('tcp')

    #: [UDP] radio-bc
    radio_bc = 1596, 'radio-bc', TransportProtocol.get('udp')

    #: - [TCP] orbplus-iiop
    #: - [UDP] orbplus-iiop
    orbplus_iiop = 1597, 'orbplus-iiop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] picknfs
    #: - [UDP] picknfs
    picknfs = 1598, 'picknfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] simbaservices
    #: - [UDP] simbaservices
    simbaservices = 1599, 'simbaservices', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] issd
    #: - [UDP] issd
    issd = 1600, 'issd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] aas
    #: - [UDP] aas
    aas = 1601, 'aas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] inspect
    #: - [UDP] inspect
    inspect = 1602, 'inspect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pickodbc
    #: - [UDP] pickodbc
    picodbc = 1603, 'picodbc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icabrowser
    #: - [UDP] icabrowser
    icabrowser = 1604, 'icabrowser', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Salutation Manager (Salutation Protocol)
    #: - [UDP] Salutation Manager (Salutation Protocol)
    slp = 1605, 'slp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Salutation Manager (SLM-API)
    #: - [UDP] Salutation Manager (SLM-API)
    slm_api = 1606, 'slm-api', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] stt
    #: - [UDP] stt
    stt = 1607, 'stt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Smart Corp. License Manager
    #: - [UDP] Smart Corp. License Manager
    smart_lm = 1608, 'smart-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isysg-lm
    #: - [UDP] isysg-lm
    isysg_lm = 1609, 'isysg-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] taurus-wh
    #: - [UDP] taurus-wh
    taurus_wh = 1610, 'taurus-wh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inter Library Loan
    #: - [UDP] Inter Library Loan
    ill = 1611, 'ill', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetBill Transaction Server
    #: - [UDP] NetBill Transaction Server
    netbill_trans = 1612, 'netbill-trans', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetBill Key Repository
    #: - [UDP] NetBill Key Repository
    netbill_keyrep = 1613, 'netbill-keyrep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetBill Credential Server
    #: - [UDP] NetBill Credential Server
    netbill_cred = 1614, 'netbill-cred', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetBill Authorization Server
    #: - [UDP] NetBill Authorization Server
    netbill_auth = 1615, 'netbill-auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetBill Product Server
    #: - [UDP] NetBill Product Server
    netbill_prod = 1616, 'netbill-prod', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nimrod Inter-Agent Communication
    #: - [UDP] Nimrod Inter-Agent Communication
    nimrod_agent = 1617, 'nimrod-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] skytelnet
    #: - [UDP] skytelnet
    skytelnet = 1618, 'skytelnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xs-openstorage
    #: - [UDP] xs-openstorage
    xs_openstorage = 1619, 'xs-openstorage', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] faxportwinport
    #: - [UDP] faxportwinport
    faxportwinport = 1620, 'faxportwinport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] softdataphone
    #: - [UDP] softdataphone
    softdataphone = 1621, 'softdataphone', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ontime
    #: - [UDP] ontime
    ontime = 1622, 'ontime', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] jaleosnd
    #: - [UDP] jaleosnd
    jaleosnd = 1623, 'jaleosnd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] udp-sr-port
    #: - [UDP] udp-sr-port
    udp_sr_port = 1624, 'udp-sr-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] svs-omagent
    #: - [UDP] svs-omagent
    svs_omagent = 1625, 'svs-omagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Shockwave
    #: - [UDP] Shockwave
    shockwave = 1626, 'shockwave', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] T.128 Gateway
    #: - [UDP] T.128 Gateway
    t128_gateway = 1627, 't128-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LonTalk normal
    #: - [UDP] LonTalk normal
    lontalk_norm = 1628, 'lontalk-norm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LonTalk urgent
    #: - [UDP] LonTalk urgent
    lontalk_urgnt = 1629, 'lontalk-urgnt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle Net8 Cman
    #: - [UDP] Oracle Net8 Cman
    oraclenet8cman = 1630, 'oraclenet8cman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Visit view
    #: - [UDP] Visit view
    visitview = 1631, 'visitview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PAMMRATC
    #: - [UDP] PAMMRATC
    pammratc = 1632, 'pammratc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PAMMRPC
    #: - [UDP] PAMMRPC
    pammrpc = 1633, 'pammrpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Log On America Probe
    #: - [UDP] Log On America Probe
    loaprobe = 1634, 'loaprobe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EDB Server 1
    #: - [UDP] EDB Server 1
    edb_server1 = 1635, 'edb-server1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISP shared public data control
    #: - [UDP] ISP shared public data control
    isdc = 1636, 'isdc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISP shared local data control
    #: - [UDP] ISP shared local data control
    islc = 1637, 'islc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISP shared management control
    #: - [UDP] ISP shared management control
    ismc = 1638, 'ismc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cert-initiator
    #: - [UDP] cert-initiator
    cert_initiator = 1639, 'cert-initiator', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cert-responder
    #: - [UDP] cert-responder
    cert_responder = 1640, 'cert-responder', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] InVision
    #: - [UDP] InVision
    invision = 1641, 'invision', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isis-am
    #: - [UDP] isis-am
    isis_am = 1642, 'isis-am', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isis-ambc
    #: - [UDP] isis-ambc
    isis_ambc = 1643, 'isis-ambc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Satellite-data Acquisition System 4
    #: - [UDP] Satellite-data Acquisition System 4
    saiseh = 1644, 'saiseh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SightLine
    #: - [UDP] SightLine
    sightline = 1645, 'sightline', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sa-msg-port
    #: - [UDP] sa-msg-port
    sa_msg_port = 1646, 'sa-msg-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rsap
    #: - [UDP] rsap
    rsap = 1647, 'rsap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] concurrent-lm
    #: - [UDP] concurrent-lm
    concurrent_lm = 1648, 'concurrent-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] kermit
    #: - [UDP] kermit
    kermit = 1649, 'kermit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nkdn
    #: - [UDP] nkd
    nkd = 1650, 'nkd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] shiva_confsrvr IANA assigned this well-formed service name as a
    #: replacement for "shiva_confsrvr".
    #: - [TCP] shiva_confsrvr
    #: - [UDP] shiva_confsrvr IANA assigned this well-formed service name as a
    #: replacement for "shiva_confsrvr".
    #: - [UDP] shiva_confsrvr
    shiva_confsrvr = 1651, 'shiva-confsrvr', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] xnmp
    #: - [UDP] xnmp
    xnmp = 1652, 'xnmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] alphatech-lm
    #: - [UDP] alphatech-lm
    alphatech_lm = 1653, 'alphatech-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] stargatealerts
    #: - [UDP] stargatealerts
    stargatealerts = 1654, 'stargatealerts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dec-mbadmin
    #: - [UDP] dec-mbadmin
    dec_mbadmin = 1655, 'dec-mbadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dec-mbadmin-h
    #: - [UDP] dec-mbadmin-h
    dec_mbadmin_h = 1656, 'dec-mbadmin-h', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fujitsu-mmpdc
    #: - [UDP] fujitsu-mmpdc
    fujitsu_mmpdc = 1657, 'fujitsu-mmpdc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sixnetudr
    #: - [UDP] sixnetudr
    sixnetudr = 1658, 'sixnetudr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Silicon Grail License Manager
    #: - [UDP] Silicon Grail License Manager
    sg_lm = 1659, 'sg-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] skip-mc-gikreq
    #: - [UDP] skip-mc-gikreq
    skip_mc_gikreq = 1660, 'skip-mc-gikreq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-1
    #: - [UDP] netview-aix-1
    netview_aix_1 = 1661, 'netview-aix-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-2
    #: - [UDP] netview-aix-2
    netview_aix_2 = 1662, 'netview-aix-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-3
    #: - [UDP] netview-aix-3
    netview_aix_3 = 1663, 'netview-aix-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-4
    #: - [UDP] netview-aix-4
    netview_aix_4 = 1664, 'netview-aix-4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-5
    #: - [UDP] netview-aix-5
    netview_aix_5 = 1665, 'netview-aix-5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-6
    #: - [UDP] netview-aix-6
    netview_aix_6 = 1666, 'netview-aix-6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-7
    #: - [UDP] netview-aix-7
    netview_aix_7 = 1667, 'netview-aix-7', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-8
    #: - [UDP] netview-aix-8
    netview_aix_8 = 1668, 'netview-aix-8', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-9
    #: - [UDP] netview-aix-9
    netview_aix_9 = 1669, 'netview-aix-9', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-10
    #: - [UDP] netview-aix-10
    netview_aix_10 = 1670, 'netview-aix-10', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-11
    #: - [UDP] netview-aix-11
    netview_aix_11 = 1671, 'netview-aix-11', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netview-aix-12
    #: - [UDP] netview-aix-12
    netview_aix_12 = 1672, 'netview-aix-12', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intel Proshare Multicast
    #: - [UDP] Intel Proshare Multicast
    proshare_mc_1 = 1673, 'proshare-mc-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intel Proshare Multicast
    #: - [UDP] Intel Proshare Multicast
    proshare_mc_2 = 1674, 'proshare-mc-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pacific Data Products
    #: - [UDP] Pacific Data Products
    pdp = 1675, 'pdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] netcomm1
    netcomm1 = 1676, 'netcomm1', TransportProtocol.get('tcp')

    #: [UDP] netcomm2
    netcomm2 = 1676, 'netcomm2', TransportProtocol.get('udp')

    #: - [TCP] groupwise
    #: - [UDP] groupwise
    groupwise = 1677, 'groupwise', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] prolink
    #: - [UDP] prolink
    prolink = 1678, 'prolink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] darcorp-lm
    #: - [UDP] darcorp-lm
    darcorp_lm = 1679, 'darcorp-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] microcom-sbp
    #: - [UDP] microcom-sbp
    microcom_sbp = 1680, 'microcom-sbp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sd-elmd
    #: - [UDP] sd-elmd
    sd_elmd = 1681, 'sd-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] lanyon-lantern
    #: - [UDP] lanyon-lantern
    lanyon_lantern = 1682, 'lanyon-lantern', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ncpm-hip
    #: - [UDP] ncpm-hip
    ncpm_hip = 1683, 'ncpm-hip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SnareSecure
    #: - [UDP] SnareSecure
    snaresecure = 1684, 'snaresecure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] n2nremote
    #: - [UDP] n2nremote
    n2nremote = 1685, 'n2nremote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cvmon
    #: - [UDP] cvmon
    cvmon = 1686, 'cvmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nsjtp-ctrl
    #: - [UDP] nsjtp-ctrl
    nsjtp_ctrl = 1687, 'nsjtp-ctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nsjtp-data
    #: - [UDP] nsjtp-data
    nsjtp_data = 1688, 'nsjtp-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] firefox
    #: - [UDP] firefox
    firefox = 1689, 'firefox', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ng-umds
    #: - [UDP] ng-umds
    ng_umds = 1690, 'ng-umds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] empire-empuma
    #: - [UDP] empire-empuma
    empire_empuma = 1691, 'empire-empuma', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sstsys-lm
    #: - [UDP] sstsys-lm
    sstsys_lm = 1692, 'sstsys-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rrirtr
    #: - [UDP] rrirtr
    rrirtr = 1693, 'rrirtr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rrimwm
    #: - [UDP] rrimwm
    rrimwm = 1694, 'rrimwm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rrilwm
    #: - [UDP] rrilwm
    rrilwm = 1695, 'rrilwm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rrifmm
    #: - [UDP] rrifmm
    rrifmm = 1696, 'rrifmm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rrisat
    #: - [UDP] rrisat
    rrisat = 1697, 'rrisat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSVP-ENCAPSULATION-1
    #: - [UDP] RSVP-ENCAPSULATION-1
    rsvp_encap_1 = 1698, 'rsvp-encap-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSVP-ENCAPSULATION-2
    #: - [UDP] RSVP-ENCAPSULATION-2
    rsvp_encap_2 = 1699, 'rsvp-encap-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mps-raft
    #: - [UDP] mps-raft
    mps_raft = 1700, 'mps-raft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] l2f
    #: - [UDP] l2f
    l2f = 1701, 'l2f', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] l2tp
    #: - [UDP] l2tp
    l2tp = 1701, 'l2tp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] deskshare
    #: - [UDP] deskshare
    deskshare = 1702, 'deskshare', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hb-engine
    #: - [UDP] hb-engine
    hb_engine = 1703, 'hb-engine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bcs-broker
    #: - [UDP] bcs-broker
    bcs_broker = 1704, 'bcs-broker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] slingshot
    #: - [UDP] slingshot
    slingshot = 1705, 'slingshot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] jetform
    #: - [UDP] jetform
    jetform = 1706, 'jetform', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vdmplay
    #: - [UDP] vdmplay
    vdmplay = 1707, 'vdmplay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gat-lmd
    #: - [UDP] gat-lmd
    gat_lmd = 1708, 'gat-lmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] centra
    #: - [UDP] centra
    centra = 1709, 'centra', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] impera
    #: - [UDP] impera
    impera = 1710, 'impera', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pptconference
    #: - [UDP] pptconference
    pptconference = 1711, 'pptconference', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] resource monitoring service
    #: - [UDP] resource monitoring service
    registrar = 1712, 'registrar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ConferenceTalk
    #: - [UDP] ConferenceTalk
    conferencetalk = 1713, 'conferencetalk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sesi-lm
    #: - [UDP] sesi-lm
    sesi_lm = 1714, 'sesi-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] houdini-lm
    #: - [UDP] houdini-lm
    houdini_lm = 1715, 'houdini-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xmsg
    #: - [UDP] xmsg
    xmsg = 1716, 'xmsg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fj-hdnet
    #: - [UDP] fj-hdnet
    fj_hdnet = 1717, 'fj-hdnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] H.323 Multicast Gatekeeper Discover
    #: - [UDP] H.323 Multicast Gatekeeper Discover
    h323gatedisc = 1718, 'h323gatedisc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] H.323 Unicast Gatekeeper Signaling
    #: - [UDP] H.323 Unicast Gatekeeper Signaling
    h323gatestat = 1719, 'h323gatestat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] H.323 Call Control Signalling
    #: - [UDP] H.323 Call Control Signalling
    #: - [SCTP] H.323 Call Control
    h323hostcall = 1720, 'h323hostcall', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] caicci
    #: - [UDP] caicci
    caicci = 1721, 'caicci', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HKS License Manager
    #: - [UDP] HKS License Manager
    hks_lm = 1722, 'hks-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pptp [:rfc:`2637`]
    #: - [UDP] pptp [:rfc:`2637`]
    pptp = 1723, 'pptp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] csbphonemaster
    #: - [UDP] csbphonemaster
    csbphonemaster = 1724, 'csbphonemaster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iden-ralp
    #: - [UDP] iden-ralp
    iden_ralp = 1725, 'iden-ralp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBERIAGAMES
    #: - [UDP] IBERIAGAMES
    iberiagames = 1726, 'iberiagames', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] winddx
    #: - [UDP] winddx
    winddx = 1727, 'winddx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TELINDUS
    #: - [UDP] TELINDUS
    telindus = 1728, 'telindus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CityNL License Management
    #: - [UDP] CityNL License Management
    citynl = 1729, 'citynl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] roketz
    #: - [UDP] roketz
    roketz = 1730, 'roketz', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSICCP
    #: - [UDP] MSICCP
    msiccp = 1731, 'msiccp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] proxim
    #: - [UDP] proxim
    proxim = 1732, 'proxim', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SIMS - SIIPAT Protocol for Alarm Transmission
    #: - [UDP] SIMS - SIIPAT Protocol for Alarm Transmission
    siipat = 1733, 'siipat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Camber Corporation License Management
    #: - [UDP] Camber Corporation License Management
    cambertx_lm = 1734, 'cambertx-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PrivateChat
    #: - [UDP] PrivateChat
    privatechat = 1735, 'privatechat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] street-stream
    #: - [UDP] street-stream
    street_stream = 1736, 'street-stream', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ultimad
    #: - [UDP] ultimad
    ultimad = 1737, 'ultimad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GameGen1
    #: - [UDP] GameGen1
    gamegen1 = 1738, 'gamegen1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] webaccess
    #: - [UDP] webaccess
    webaccess = 1739, 'webaccess', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] encore
    #: - [UDP] encore
    encore = 1740, 'encore', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco-net-mgmt
    #: - [UDP] cisco-net-mgmt
    cisco_net_mgmt = 1741, 'cisco-net-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3Com-nsd
    #: - [UDP] 3Com-nsd
    AppType_3com_nsd = 1742, '3com-nsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cinema Graphics License Manager
    #: - [UDP] Cinema Graphics License Manager
    cinegrfx_lm = 1743, 'cinegrfx-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ncpm-ft
    #: - [UDP] ncpm-ft
    ncpm_ft = 1744, 'ncpm-ft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] remote-winsock
    #: - [UDP] remote-winsock
    remote_winsock = 1745, 'remote-winsock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ftrapid-1
    #: - [UDP] ftrapid-1
    ftrapid_1 = 1746, 'ftrapid-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ftrapid-2
    #: - [UDP] ftrapid-2
    ftrapid_2 = 1747, 'ftrapid-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] oracle-em1
    #: - [UDP] oracle-em1
    oracle_em1 = 1748, 'oracle-em1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] aspen-services
    #: - [UDP] aspen-services
    aspen_services = 1749, 'aspen-services', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Socket Library's PortMaster
    #: - [UDP] Simple Socket Library's PortMaster
    sslp = 1750, 'sslp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SwiftNet
    #: - [UDP] SwiftNet
    swiftnet = 1751, 'swiftnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Leap of Faith Research License Manager
    #: - [UDP] Leap of Faith Research License Manager
    lofr_lm = 1752, 'lofr-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Predatar Comms Service
    predatar_comms = 1753, 'predatar-comms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_1753 = 1753, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] oracle-em2
    #: - [UDP] oracle-em2
    oracle_em2 = 1754, 'oracle-em2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ms-streaming
    #: - [UDP] ms-streaming
    ms_streaming = 1755, 'ms-streaming', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] capfast-lmd
    #: - [UDP] capfast-lmd
    capfast_lmd = 1756, 'capfast-lmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cnhrp
    #: - [UDP] cnhrp
    cnhrp = 1757, 'cnhrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tftp-mcast
    #: - [UDP] tftp-mcast
    tftp_mcast = 1758, 'tftp-mcast', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPSS License Manager
    #: - [UDP] SPSS License Manager
    spss_lm = 1759, 'spss-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] www-ldap-gw
    #: - [UDP] www-ldap-gw
    www_ldap_gw = 1760, 'www-ldap-gw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cft-0
    #: - [UDP] cft-0
    cft_0 = 1761, 'cft-0', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cft-1
    #: - [UDP] cft-1
    cft_1 = 1762, 'cft-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cft-2
    #: - [UDP] cft-2
    cft_2 = 1763, 'cft-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cft-3
    #: - [UDP] cft-3
    cft_3 = 1764, 'cft-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cft-4
    #: - [UDP] cft-4
    cft_4 = 1765, 'cft-4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cft-5
    #: - [UDP] cft-5
    cft_5 = 1766, 'cft-5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cft-6
    #: - [UDP] cft-6
    cft_6 = 1767, 'cft-6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cft-7
    #: - [UDP] cft-7
    cft_7 = 1768, 'cft-7', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bmc-net-adm
    #: - [UDP] bmc-net-adm
    bmc_net_adm = 1769, 'bmc-net-adm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bmc-net-svc
    #: - [UDP] bmc-net-svc
    bmc_net_svc = 1770, 'bmc-net-svc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vaultbase
    #: - [UDP] vaultbase
    vaultbase = 1771, 'vaultbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EssWeb Gateway
    #: - [UDP] EssWeb Gateway
    essweb_gw = 1772, 'essweb-gw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KMSControl
    #: - [UDP] KMSControl
    kmscontrol = 1773, 'kmscontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] global-dtserv
    #: - [UDP] global-dtserv
    global_dtserv = 1774, 'global-dtserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] data interchange between visual processing containers
    vdab = 1775, 'vdab', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_1775 = 1775, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Federal Emergency Management Information System
    #: - [UDP] Federal Emergency Management Information System
    femis = 1776, 'femis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] powerguardian
    #: - [UDP] powerguardian
    powerguardian = 1777, 'powerguardian', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] prodigy-internet
    #: - [UDP] prodigy-internet
    prodigy_intrnet = 1778, 'prodigy-intrnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pharmasoft
    #: - [UDP] pharmasoft
    pharmasoft = 1779, 'pharmasoft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dpkeyserv
    #: - [UDP] dpkeyserv
    dpkeyserv = 1780, 'dpkeyserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] answersoft-lm
    #: - [UDP] answersoft-lm
    answersoft_lm = 1781, 'answersoft-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp-hcip
    #: - [UDP] hp-hcip
    hp_hcip = 1782, 'hp-hcip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Decomissioned Port 04/14/00, ms
    decomissioned_port_04_14_00_ms = 1783, 'decomissioned_port_04_14_00_ms', TransportProtocol.get('undefined')

    #: - [TCP] Finle License Manager
    #: - [UDP] Finle License Manager
    finle_lm = 1784, 'finle-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Wind River Systems License Manager
    #: - [UDP] Wind River Systems License Manager
    windlm = 1785, 'windlm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] funk-logger
    #: - [UDP] funk-logger
    funk_logger = 1786, 'funk-logger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] funk-license
    #: - [UDP] funk-license
    funk_license = 1787, 'funk-license', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] psmond
    #: - [UDP] psmond
    psmond = 1788, 'psmond', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hello
    #: - [UDP] hello
    hello = 1789, 'hello', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Networked Media Streaming Protocol
    #: - [UDP] Networked Media Streaming Protocol
    nmsp_537 = 537, 'nmsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Narrative Media Streaming Protocol
    #: - [UDP] Narrative Media Streaming Protocol
    nmsp_1790 = 1790, 'nmsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EA1
    #: - [UDP] EA1
    ea1 = 1791, 'ea1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ibm-dt-2
    #: - [UDP] ibm-dt-2
    ibm_dt_2 = 1792, 'ibm-dt-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rsc-robot
    #: - [UDP] rsc-robot
    rsc_robot = 1793, 'rsc-robot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cera-bcm
    #: - [UDP] cera-bcm
    cera_bcm = 1794, 'cera-bcm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dpi-proxy
    #: - [UDP] dpi-proxy
    dpi_proxy = 1795, 'dpi-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vocaltec Server Administration
    #: - [UDP] Vocaltec Server Administration
    vocaltec_admin = 1796, 'vocaltec-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Universal Management Architecture
    #: - [UDP] Universal Management Architecture
    uma_144 = 144, 'uma', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UMA
    #: - [UDP] UMA
    uma_1797 = 1797, 'uma', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Event Transfer Protocol
    #: - [UDP] Event Transfer Protocol
    etp = 1798, 'etp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETRISK
    #: - [UDP] NETRISK
    netrisk = 1799, 'netrisk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ANSYS-License manager
    #: - [UDP] ANSYS-License manager
    ansys_lm = 1800, 'ansys-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Message Que
    #: - [UDP] Microsoft Message Que
    msmq = 1801, 'msmq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ConComp1
    #: - [UDP] ConComp1
    concomp1 = 1802, 'concomp1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP-HCIP-GWY
    #: - [UDP] HP-HCIP-GWY
    hp_hcip_gwy = 1803, 'hp-hcip-gwy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ENL
    #: - [UDP] ENL
    enl = 1804, 'enl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ENL-Name
    #: - [UDP] ENL-Name
    enl_name = 1805, 'enl-name', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Musiconline
    #: - [UDP] Musiconline
    musiconline = 1806, 'musiconline', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Hot Standby Protocol
    #: - [UDP] Fujitsu Hot Standby Protocol
    fhsp = 1807, 'fhsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle-VP2
    #: - [UDP] Oracle-VP2
    oracle_vp2 = 1808, 'oracle-vp2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle-VP1
    #: - [UDP] Oracle-VP1
    oracle_vp1 = 1809, 'oracle-vp1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jerand License Manager
    #: - [UDP] Jerand License Manager
    jerand_lm = 1810, 'jerand-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Scientia-SDB
    #: - [UDP] Scientia-SDB
    scientia_sdb = 1811, 'scientia-sdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RADIUS [:rfc:`2865`]
    #: - [UDP] RADIUS [:rfc:`2865`]
    radius = 1812, 'radius', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RADIUS Accounting [:rfc:`2866`]
    #: - [UDP] RADIUS Accounting [:rfc:`2866`]
    radius_acct = 1813, 'radius-acct', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TDP Suite
    #: - [UDP] TDP Suite
    tdp_suite = 1814, 'tdp-suite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MMPFT
    #: - [UDP] MMPFT
    mmpft = 1815, 'mmpft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HARP
    #: - [UDP] HARP
    harp = 1816, 'harp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RKB-OSCS
    #: - [UDP] RKB-OSCS
    rkb_oscs = 1817, 'rkb-oscs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Enhanced Trivial File Transfer Protocol
    #: - [UDP] Enhanced Trivial File Transfer Protocol
    etftp = 1818, 'etftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Plato License Manager
    #: - [UDP] Plato License Manager
    plato_lm = 1819, 'plato-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mcagent
    #: - [UDP] mcagent
    mcagent = 1820, 'mcagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] donnyworld
    #: - [UDP] donnyworld
    donnyworld = 1821, 'donnyworld', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] es-elmd
    #: - [UDP] es-elmd
    es_elmd = 1822, 'es-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unisys Natural Language License Manager
    #: - [UDP] Unisys Natural Language License Manager
    unisys_lm = 1823, 'unisys-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] metrics-pas
    #: - [UDP] metrics-pas
    metrics_pas = 1824, 'metrics-pas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DirecPC Video
    #: - [UDP] DirecPC Video
    direcpc_video = 1825, 'direcpc-video', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARDT
    #: - [UDP] ARDT
    ardt = 1826, 'ardt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASI
    #: - [UDP] ASI
    asi = 1827, 'asi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] itm-mcell-u
    #: - [UDP] itm-mcell-u
    itm_mcell_u = 1828, 'itm-mcell-u', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Optika eMedia
    #: - [UDP] Optika eMedia
    optika_emedia = 1829, 'optika-emedia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle Net8 CMan Admin
    #: - [UDP] Oracle Net8 CMan Admin
    net8_cman = 1830, 'net8-cman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Myrtle
    #: - [UDP] Myrtle
    myrtle = 1831, 'myrtle', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ThoughtTreasure
    #: - [UDP] ThoughtTreasure
    tht_treasure = 1832, 'tht-treasure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] udpradio
    #: - [UDP] udpradio
    udpradio = 1833, 'udpradio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARDUS Unicast
    #: - [UDP] ARDUS Unicast
    ardusuni = 1834, 'ardusuni', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARDUS Multicast
    #: - [UDP] ARDUS Multicast
    ardusmul = 1835, 'ardusmul', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ste-smsc
    #: - [UDP] ste-smsc
    ste_smsc = 1836, 'ste-smsc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] csoft1
    #: - [UDP] csoft1
    csoft1 = 1837, 'csoft1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TALNET
    #: - [UDP] TALNET
    talnet = 1838, 'talnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netopia-vo1
    #: - [UDP] netopia-vo1
    netopia_vo1 = 1839, 'netopia-vo1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netopia-vo2
    #: - [UDP] netopia-vo2
    netopia_vo2 = 1840, 'netopia-vo2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netopia-vo3
    #: - [UDP] netopia-vo3
    netopia_vo3 = 1841, 'netopia-vo3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netopia-vo4
    #: - [UDP] netopia-vo4
    netopia_vo4 = 1842, 'netopia-vo4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netopia-vo5
    #: - [UDP] netopia-vo5
    netopia_vo5 = 1843, 'netopia-vo5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DirecPC-DLL
    #: - [UDP] DirecPC-DLL
    direcpc_dll = 1844, 'direcpc-dll', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] altalink
    #: - [UDP] altalink
    altalink = 1845, 'altalink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tunstall PNC
    #: - [UDP] Tunstall PNC
    tunstall_pnc = 1846, 'tunstall-pnc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SLP Notification [:rfc:`3082`]
    #: - [UDP] SLP Notification [:rfc:`3082`]
    slp_notify = 1847, 'slp-notify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fjdocdist
    #: - [UDP] fjdocdist
    fjdocdist = 1848, 'fjdocdist', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ALPHA-SMS
    #: - [UDP] ALPHA-SMS
    alpha_sms = 1849, 'alpha-sms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GSI
    #: - [UDP] GSI
    gsi = 1850, 'gsi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ctcd
    #: - [UDP] ctcd
    ctcd = 1851, 'ctcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Time
    #: - [UDP] Virtual Time
    virtual_time = 1852, 'virtual-time', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VIDS-AVTP
    #: - [UDP] VIDS-AVTP
    vids_avtp = 1853, 'vids-avtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Buddy Draw
    #: - [UDP] Buddy Draw
    buddy_draw = 1854, 'buddy-draw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fiorano RtrSvc
    #: - [UDP] Fiorano RtrSvc
    fiorano_rtrsvc = 1855, 'fiorano-rtrsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fiorano MsgSvc
    #: - [UDP] Fiorano MsgSvc
    fiorano_msgsvc = 1856, 'fiorano-msgsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DataCaptor
    #: - [UDP] DataCaptor
    datacaptor = 1857, 'datacaptor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PrivateArk
    #: - [UDP] PrivateArk
    privateark = 1858, 'privateark', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gamma Fetcher Server
    #: - [UDP] Gamma Fetcher Server
    gammafetchsvr = 1859, 'gammafetchsvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SunSCALAR Services
    #: - [UDP] SunSCALAR Services
    sunscalar_svc = 1860, 'sunscalar-svc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LeCroy VICP
    #: - [UDP] LeCroy VICP
    lecroy_vicp = 1861, 'lecroy-vicp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MySQL Cluster Manager Agent
    #: - [UDP] MySQL Cluster Manager Agent
    mysql_cm_agent = 1862, 'mysql-cm-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSNP
    #: - [UDP] MSNP
    msnp = 1863, 'msnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Paradym 31 Port
    #: - [UDP] Paradym 31 Port
    paradym_31port = 1864, 'paradym-31port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ENTP
    #: - [UDP] ENTP
    entp = 1865, 'entp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] swrmi
    #: - [UDP] swrmi
    swrmi = 1866, 'swrmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UDRIVE
    #: - [UDP] UDRIVE
    udrive = 1867, 'udrive', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VizibleBrowser
    #: - [UDP] VizibleBrowser
    viziblebrowser = 1868, 'viziblebrowser', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TransAct
    #: - [UDP] TransAct
    transact = 1869, 'transact', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SunSCALAR DNS Service
    #: - [UDP] SunSCALAR DNS Service
    sunscalar_dns = 1870, 'sunscalar-dns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cano Central 0
    #: - [UDP] Cano Central 0
    canocentral0 = 1871, 'canocentral0', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cano Central 1
    #: - [UDP] Cano Central 1
    canocentral1 = 1872, 'canocentral1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fjmpjps
    #: - [UDP] Fjmpjps
    fjmpjps = 1873, 'fjmpjps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fjswapsnp
    #: - [UDP] Fjswapsnp
    fjswapsnp = 1874, 'fjswapsnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] westell stats
    #: - [UDP] westell stats
    westell_stats = 1875, 'westell-stats', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ewcappsrv
    #: - [UDP] ewcappsrv
    ewcappsrv = 1876, 'ewcappsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp-webqosdb
    #: - [UDP] hp-webqosdb
    hp_webqosdb = 1877, 'hp-webqosdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] drmsmc
    #: - [UDP] drmsmc
    drmsmc = 1878, 'drmsmc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NettGain NMS
    #: - [UDP] NettGain NMS
    nettgain_nms = 1879, 'nettgain-nms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gilat VSAT Control
    #: - [UDP] Gilat VSAT Control
    vsat_control = 1880, 'vsat-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM WebSphere MQ Everyplace
    #: - [UDP] IBM WebSphere MQ Everyplace
    ibm_mqseries2 = 1881, 'ibm-mqseries2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA eTrust Common Services
    #: - [UDP] CA eTrust Common Services
    ecsqdmn = 1882, 'ecsqdmn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Internet Distance Map Svc
    #: - [UDP] Internet Distance Map Svc
    idmaps = 1884, 'idmaps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Veritas Trap Server
    #: - [UDP] Veritas Trap Server
    vrtstrapserver = 1885, 'vrtstrapserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Leonardo over IP
    #: - [UDP] Leonardo over IP
    leoip = 1886, 'leoip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileX Listening Port
    #: - [UDP] FileX Listening Port
    filex_lport = 1887, 'filex-lport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NC Config Port
    #: - [UDP] NC Config Port
    ncconfig = 1888, 'ncconfig', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unify Web Adapter Service
    #: - [UDP] Unify Web Adapter Service
    unify_adapter = 1889, 'unify-adapter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] wilkenListener
    #: - [UDP] wilkenListener
    wilkenlistener = 1890, 'wilkenlistener', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ChildKey Notification
    #: - [UDP] ChildKey Notification
    childkey_notif = 1891, 'childkey-notif', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ChildKey Control
    #: - [UDP] ChildKey Control
    childkey_ctrl = 1892, 'childkey-ctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ELAD Protocol
    #: - [UDP] ELAD Protocol
    elad = 1893, 'elad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] O2Server Port
    #: - [UDP] O2Server Port
    o2server_port = 1894, 'o2server-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] unassigned
    #: - [UDP] unassigned
    unassigned_1895 = 1895, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] b-novative license server
    #: - [UDP] b-novative license server
    b_novative_ls = 1896, 'b-novative-ls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MetaAgent
    #: - [UDP] MetaAgent
    metaagent = 1897, 'metaagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cymtec secure management
    #: - [UDP] Cymtec secure management
    cymtec_port = 1898, 'cymtec-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MC2Studios
    #: - [UDP] MC2Studios
    mc2studios = 1899, 'mc2studios', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSDP
    #: - [UDP] SSDP
    ssdp = 1900, 'ssdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu ICL Terminal Emulator Program A
    #: - [UDP] Fujitsu ICL Terminal Emulator Program A
    fjicl_tep_a = 1901, 'fjicl-tep-a', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu ICL Terminal Emulator Program B
    #: - [UDP] Fujitsu ICL Terminal Emulator Program B
    fjicl_tep_b = 1902, 'fjicl-tep-b', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Local Link Name Resolution
    #: - [UDP] Local Link Name Resolution
    linkname = 1903, 'linkname', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu ICL Terminal Emulator Program C
    #: - [UDP] Fujitsu ICL Terminal Emulator Program C
    fjicl_tep_c = 1904, 'fjicl-tep-c', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure UP.Link Gateway Protocol
    #: - [UDP] Secure UP.Link Gateway Protocol
    sugp = 1905, 'sugp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TPortMapperReq
    #: - [UDP] TPortMapperReq
    tpmd = 1906, 'tpmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IntraSTAR
    #: - [UDP] IntraSTAR
    intrastar = 1907, 'intrastar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dawn
    #: - [UDP] Dawn
    dawn = 1908, 'dawn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Global World Link
    #: - [UDP] Global World Link
    global_wlink = 1909, 'global-wlink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UltraBac Software communications port
    #: - [UDP] UltraBac Software communications port
    ultrabac = 1910, 'ultrabac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Starlight Networks Multimedia Transport Protocol
    #: - [UDP] Starlight Networks Multimedia Transport Protocol
    mtp = 1911, 'mtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rhp-iibp
    #: - [UDP] rhp-iibp
    rhp_iibp = 1912, 'rhp-iibp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] armadp
    #: - [UDP] armadp
    armadp = 1913, 'armadp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Elm-Momentum
    #: - [UDP] Elm-Momentum
    elm_momentum = 1914, 'elm-momentum', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FACELINK
    #: - [UDP] FACELINK
    facelink = 1915, 'facelink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Persoft Persona
    #: - [UDP] Persoft Persona
    persona = 1916, 'persona', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nOAgent
    #: - [UDP] nOAgent
    noagent = 1917, 'noagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Tivole Directory Service - NDS
    #: - [UDP] IBM Tivole Directory Service - NDS
    can_nds = 1918, 'can-nds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Tivoli Directory Service - DCH
    #: - [UDP] IBM Tivoli Directory Service - DCH
    can_dch = 1919, 'can-dch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Tivoli Directory Service - FERRET
    #: - [UDP] IBM Tivoli Directory Service - FERRET
    can_ferret = 1920, 'can-ferret', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NoAdmin
    #: - [UDP] NoAdmin
    noadmin = 1921, 'noadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tapestry
    #: - [UDP] Tapestry
    tapestry = 1922, 'tapestry', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPICE
    #: - [UDP] SPICE
    spice = 1923, 'spice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XIIP
    #: - [UDP] XIIP
    xiip = 1924, 'xiip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Surrogate Discovery Port
    #: - [UDP] Surrogate Discovery Port
    discovery_port = 1925, 'discovery-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Evolution Game Server
    #: - [UDP] Evolution Game Server
    egs = 1926, 'egs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Videte CIPC Port
    #: - [UDP] Videte CIPC Port
    videte_cipc = 1927, 'videte-cipc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Expnd Maui Srvr Dscovr
    #: - [UDP] Expnd Maui Srvr Dscovr
    emsd_port = 1928, 'emsd-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bandwiz System - Server
    #: - [UDP] Bandwiz System - Server
    bandwiz_system = 1929, 'bandwiz-system', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Drive AppServer
    #: - [UDP] Drive AppServer
    driveappserver = 1930, 'driveappserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMD SCHED
    #: - [UDP] AMD SCHED
    amdsched = 1931, 'amdsched', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CTT Broker
    #: - [UDP] CTT Broker
    ctt_broker = 1932, 'ctt-broker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM LM MT Agent
    #: - [UDP] IBM LM MT Agent
    xmapi = 1933, 'xmapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM LM Appl Agent
    #: - [UDP] IBM LM Appl Agent
    xaapi = 1934, 'xaapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Macromedia Flash Communications Server MX
    #: - [UDP] Macromedia Flash Communications server MX
    macromedia_fcs = 1935, 'macromedia-fcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JetCmeServer Server Port
    #: - [UDP] JetCmeServer Server Port
    jetcmeserver = 1936, 'jetcmeserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JetVWay Server Port
    #: - [UDP] JetVWay Server Port
    jwserver = 1937, 'jwserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JetVWay Client Port
    #: - [UDP] JetVWay Client Port
    jwclient = 1938, 'jwclient', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JetVision Server Port
    #: - [UDP] JetVision Server Port
    jvserver = 1939, 'jvserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JetVision Client Port
    #: - [UDP] JetVision Client Port
    jvclient = 1940, 'jvclient', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DIC-Aida
    #: - [UDP] DIC-Aida
    dic_aida = 1941, 'dic-aida', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Real Enterprise Service
    #: - [UDP] Real Enterprise Service
    res = 1942, 'res', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Beeyond Media
    #: - [UDP] Beeyond Media
    beeyond_media = 1943, 'beeyond-media', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] close-combat
    #: - [UDP] close-combat
    close_combat = 1944, 'close-combat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dialogic-elmd
    #: - [UDP] dialogic-elmd
    dialogic_elmd = 1945, 'dialogic-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tekpls
    #: - [UDP] tekpls
    tekpls = 1946, 'tekpls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SentinelSRM
    #: - [UDP] SentinelSRM
    sentinelsrm = 1947, 'sentinelsrm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eye2eye
    #: - [UDP] eye2eye
    eye2eye = 1948, 'eye2eye', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISMA Easdaq Live
    #: - [UDP] ISMA Easdaq Live
    ismaeasdaqlive = 1949, 'ismaeasdaqlive', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISMA Easdaq Test
    #: - [UDP] ISMA Easdaq Test
    ismaeasdaqtest = 1950, 'ismaeasdaqtest', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bcs-lmserver
    #: - [UDP] bcs-lmserver
    bcs_lmserver = 1951, 'bcs-lmserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mpnjsc
    #: - [UDP] mpnjsc
    mpnjsc = 1952, 'mpnjsc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rapid Base
    #: - [UDP] Rapid Base
    rapidbase = 1953, 'rapidbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ABR-API (diskbridge)
    #: - [UDP] ABR-API (diskbridge)
    abr_api = 1954, 'abr-api', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ABR-Secure Data (diskbridge)
    #: - [UDP] ABR-Secure Data (diskbridge)
    abr_secure = 1955, 'abr-secure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vertel VMF DS
    #: - [UDP] Vertel VMF DS
    vrtl_vmf_ds = 1956, 'vrtl-vmf-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] unix-status
    #: - [UDP] unix-status
    unix_status = 1957, 'unix-status', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA Administration Daemon
    #: - [UDP] CA Administration Daemon
    dxadmind = 1958, 'dxadmind', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SIMP Channel
    #: - [UDP] SIMP Channel
    simp_all = 1959, 'simp-all', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Merit DAC NASmanager
    #: - [UDP] Merit DAC NASmanager
    nasmanager = 1960, 'nasmanager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BTS APPSERVER
    #: - [UDP] BTS APPSERVER
    bts_appserver = 1961, 'bts-appserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BIAP-MP
    #: - [UDP] BIAP-MP
    biap_mp = 1962, 'biap-mp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WebMachine
    #: - [UDP] WebMachine
    webmachine = 1963, 'webmachine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SOLID E ENGINE
    #: - [UDP] SOLID E ENGINE
    solid_e_engine = 1964, 'solid-e-engine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tivoli NPM
    #: - [UDP] Tivoli NPM
    tivoli_npm = 1965, 'tivoli-npm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Slush
    #: - [UDP] Slush
    slush = 1966, 'slush', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNS Quote
    #: - [UDP] SNS Quote
    sns_quote = 1967, 'sns-quote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LIPSinc
    #: - [UDP] LIPSinc
    lipsinc = 1968, 'lipsinc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LIPSinc 1
    #: - [UDP] LIPSinc 1
    lipsinc1 = 1969, 'lipsinc1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetOp Remote Control
    #: - [UDP] NetOp Remote Control
    netop_rc = 1970, 'netop-rc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetOp School
    #: - [UDP] NetOp School
    netop_school = 1971, 'netop-school', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cache
    #: - [UDP] Cache
    intersys_cache = 1972, 'intersys-cache', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Link Switching Remote Access Protocol
    #: - [UDP] Data Link Switching Remote Access Protocol
    dlsrap = 1973, 'dlsrap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DRP
    #: - [UDP] DRP
    drp = 1974, 'drp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TCO Flash Agent
    #: - [UDP] TCO Flash Agent
    tcoflashagent = 1975, 'tcoflashagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TCO Reg Agent
    #: - [UDP] TCO Reg Agent
    tcoregagent = 1976, 'tcoregagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TCO Address Book
    #: - [UDP] TCO Address Book
    tcoaddressbook = 1977, 'tcoaddressbook', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UniSQL
    #: - [UDP] UniSQL
    unisql = 1978, 'unisql', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UniSQL Java
    #: - [UDP] UniSQL Java
    unisql_java = 1979, 'unisql-java', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PearlDoc XACT
    #: - [UDP] PearlDoc XACT
    pearldoc_xact = 1980, 'pearldoc-xact', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] p2pQ
    #: - [UDP] p2pQ
    p2pq = 1981, 'p2pq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Evidentiary Timestamp
    #: - [UDP] Evidentiary Timestamp
    estamp = 1982, 'estamp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Loophole Test Protocol
    #: - [UDP] Loophole Test Protocol
    lhtp = 1983, 'lhtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BB
    #: - [UDP] BB
    bb = 1984, 'bb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hot Standby Router Protocol [:rfc:`2281`]
    #: - [UDP] Hot Standby Router Protocol [:rfc:`2281`]
    hsrp = 1985, 'hsrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco license management
    #: - [UDP] cisco license management
    licensedaemon = 1986, 'licensedaemon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco RSRB Priority 1 port
    #: - [UDP] cisco RSRB Priority 1 port
    tr_rsrb_p1 = 1987, 'tr-rsrb-p1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco RSRB Priority 2 port
    #: - [UDP] cisco RSRB Priority 2 port
    tr_rsrb_p2 = 1988, 'tr-rsrb-p2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco RSRB Priority 3 port
    #: - [UDP] cisco RSRB Priority 3 port
    tr_rsrb_p3 = 1989, 'tr-rsrb-p3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MHSnet system
    #: - [UDP] MHSnet system
    mshnet = 1989, 'mshnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco STUN Priority 1 port
    #: - [UDP] cisco STUN Priority 1 port
    stun_p1 = 1990, 'stun-p1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco STUN Priority 2 port
    #: - [UDP] cisco STUN Priority 2 port
    stun_p2 = 1991, 'stun-p2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco STUN Priority 3 port
    #: - [UDP] cisco STUN Priority 3 port
    stun_p3 = 1992, 'stun-p3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPsendmsg
    #: - [UDP] IPsendmsg
    ipsendmsg = 1992, 'ipsendmsg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco SNMP TCP port
    #: - [UDP] cisco SNMP TCP port
    snmp_tcp_port = 1993, 'snmp-tcp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco serial tunnel port
    #: - [UDP] cisco serial tunnel port
    stun_port = 1994, 'stun-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco perf port
    #: - [UDP] cisco perf port
    perf_port = 1995, 'perf-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco Remote SRB port
    #: - [UDP] cisco Remote SRB port
    tr_rsrb_port = 1996, 'tr-rsrb-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco Gateway Discovery Protocol
    #: - [UDP] cisco Gateway Discovery Protocol
    gdp_port = 1997, 'gdp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco X.25 service (XOT)
    #: - [UDP] cisco X.25 service (XOT)
    x25_svc_port = 1998, 'x25-svc-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cisco identification port
    #: - [UDP] cisco identification port
    tcp_id_port = 1999, 'tcp-id-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cisco SCCP
    #: - [UDP] Cisco SCCp
    cisco_sccp = 2000, 'cisco-sccp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    dc = 2001, 'dc', TransportProtocol.get('tcp')

    #: [UDP] curry
    wizard = 2001, 'wizard', TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    globe = 2002, 'globe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Brutus Server
    #: - [UDP] Brutus Server
    brutus = 2003, 'brutus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    mailbox = 2004, 'mailbox', TransportProtocol.get('tcp')

    #: [UDP] CCWS mm conf
    emce = 2004, 'emce', TransportProtocol.get('udp')

    #: [TCP]
    berknet = 2005, 'berknet', TransportProtocol.get('tcp')

    #: [UDP]
    oracle = 2005, 'oracle', TransportProtocol.get('udp')

    #: [TCP]
    invokator = 2006, 'invokator', TransportProtocol.get('tcp')

    #: [TCP]
    dectalk = 2007, 'dectalk', TransportProtocol.get('tcp')

    #: [TCP]
    conf = 2008, 'conf', TransportProtocol.get('tcp')

    #: [TCP]
    news = 2009, 'news', TransportProtocol.get('tcp')

    #: [TCP]
    search = 2010, 'search', TransportProtocol.get('tcp')

    #: - [UDP] IANA assigned this well-formed service name as a replacement for
    #: "pipe_server".
    #: - [UDP]
    pipe_server = 2010, 'pipe-server', TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [TCP] raid
    raid_cc = 2011, 'raid-cc', TransportProtocol.get('tcp')

    #: [UDP]
    servserv = 2011, 'servserv', TransportProtocol.get('udp')

    #: [TCP]
    ttyinfo = 2012, 'ttyinfo', TransportProtocol.get('tcp')

    #: [UDP]
    raid_ac = 2012, 'raid-ac', TransportProtocol.get('udp')

    #: [UDP]
    raid_am_2007 = 2007, 'raid-am', TransportProtocol.get('udp')

    #: [TCP]
    raid_am_2013 = 2013, 'raid-am', TransportProtocol.get('tcp')

    #: [UDP] raid
    raid_cd_2006 = 2006, 'raid-cd', TransportProtocol.get('udp')

    #: [UDP]
    raid_cd_2013 = 2013, 'raid-cd', TransportProtocol.get('udp')

    #: [TCP]
    troff = 2014, 'troff', TransportProtocol.get('tcp')

    #: [UDP]
    raid_sf = 2014, 'raid-sf', TransportProtocol.get('udp')

    #: [TCP]
    cypress = 2015, 'cypress', TransportProtocol.get('tcp')

    #: [UDP]
    raid_cs = 2015, 'raid-cs', TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    bootserver = 2016, 'bootserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    cypress_stat = 2017, 'cypress-stat', TransportProtocol.get('tcp')

    #: [UDP]
    bootclient = 2017, 'bootclient', TransportProtocol.get('udp')

    #: [UDP]
    terminaldb_2008 = 2008, 'terminaldb', TransportProtocol.get('udp')

    #: [TCP]
    terminaldb_2018 = 2018, 'terminaldb', TransportProtocol.get('tcp')

    #: [UDP]
    rellpack = 2018, 'rellpack', TransportProtocol.get('udp')

    #: [UDP]
    whosockami_2009 = 2009, 'whosockami', TransportProtocol.get('udp')

    #: [TCP]
    whosockami_2019 = 2019, 'whosockami', TransportProtocol.get('tcp')

    #: [UDP]
    about = 2019, 'about', TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    xinupageserver = 2020, 'xinupageserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    servexec = 2021, 'servexec', TransportProtocol.get('tcp')

    #: [UDP]
    xinuexpansion1 = 2021, 'xinuexpansion1', TransportProtocol.get('udp')

    #: [TCP]
    down = 2022, 'down', TransportProtocol.get('tcp')

    #: [UDP]
    xinuexpansion2 = 2022, 'xinuexpansion2', TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    xinuexpansion3 = 2023, 'xinuexpansion3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    xinuexpansion4 = 2024, 'xinuexpansion4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP]
    ellpack = 2025, 'ellpack', TransportProtocol.get('tcp')

    #: [UDP]
    xribs = 2025, 'xribs', TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    scrabble = 2026, 'scrabble', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    shadowserver = 2027, 'shadowserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    submitserver = 2028, 'submitserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hot Standby Router Protocol IPv6
    #: - [UDP] Hot Standby Router Protocol IPv6
    hsrpv6 = 2029, 'hsrpv6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    device2 = 2030, 'device2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mobrien-chat
    #: - [UDP] mobrien-chat
    mobrien_chat = 2031, 'mobrien-chat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    blackboard = 2032, 'blackboard', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    glogger = 2033, 'glogger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    scoremgr = 2034, 'scoremgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    imsldoc = 2035, 'imsldoc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ethernet WS DP network
    #: - [UDP] Ethernet WS DP network
    e_dpnet = 2036, 'e-dpnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APplus Application Server
    #: - [UDP] APplus Application Server
    applus = 2037, 'applus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    objectmanager = 2038, 'objectmanager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prizma Monitoring Service
    #: - [UDP] Prizma Monitoring Service
    prizma = 2039, 'prizma', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    lam = 2040, 'lam', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    interbase = 2041, 'interbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isis
    #: - [UDP] isis
    isis = 2042, 'isis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isis-bcast
    #: - [UDP] isis-bcast
    isis_bcast = 2043, 'isis-bcast', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    rimsl = 2044, 'rimsl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    cdfunc = 2045, 'cdfunc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    sdfunc = 2046, 'sdfunc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Directory Location Service
    #: - [UDP] Directory Location Service
    dls_197 = 197, 'dls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    dls_2047 = 2047, 'dls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    dls_monitor = 2048, 'dls-monitor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    shilp = 2049, 'shilp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network File System - Sun Microsystems
    #: - [UDP] Network File System - Sun Microsystems
    #: - [SCTP] Network File System [:rfc:`5665`]
    nfs = 2049, 'nfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Avaya EMB Config Port
    #: - [UDP] Avaya EMB Config Port
    av_emb_config = 2050, 'av-emb-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPNSDP
    #: - [UDP] EPNSDP
    epnsdp = 2051, 'epnsdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] clearVisn Services Port
    #: - [UDP] clearVisn Services Port
    clearvisn = 2052, 'clearvisn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lot105 DSuper Updates
    #: - [UDP] Lot105 DSuper Updates
    lot105_ds_upd = 2053, 'lot105-ds-upd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Weblogin Port
    #: - [UDP] Weblogin Port
    weblogin = 2054, 'weblogin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Iliad-Odyssey Protocol
    #: - [UDP] Iliad-Odyssey Protocol
    iop = 2055, 'iop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OmniSky Port
    #: - [UDP] OmniSky Port
    omnisky = 2056, 'omnisky', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rich Content Protocol
    #: - [UDP] Rich Content Protocol
    rich_cp = 2057, 'rich-cp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NewWaveSearchables RMI
    #: - [UDP] NewWaveSearchables RMI
    newwavesearch = 2058, 'newwavesearch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC Messaging Service
    #: - [UDP] BMC Messaging Service
    bmc_messaging = 2059, 'bmc-messaging', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Telenium Daemon IF
    #: - [UDP] Telenium Daemon IF
    teleniumdaemon = 2060, 'teleniumdaemon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetMount
    #: - [UDP] NetMount
    netmount = 2061, 'netmount', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICG SWP Port
    #: - [UDP] ICG SWP Port
    icg_swp = 2062, 'icg-swp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICG Bridge Port
    #: - [UDP] ICG Bridge Port
    icg_bridge = 2063, 'icg-bridge', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICG IP Relay Port
    #: - [UDP] ICG IP Relay Port
    icg_iprelay = 2064, 'icg-iprelay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Link Switch Read Port Number
    #: - [UDP] Data Link Switch Read Port Number
    dlsrpn = 2065, 'dlsrpn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AVM USB Remote Architecture
    #: - [UDP] AVM USB Remote Architecture
    aura = 2066, 'aura', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Link Switch Write Port Number
    #: - [UDP] Data Link Switch Write Port Number
    dlswpn = 2067, 'dlswpn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avocent AuthSrv Protocol
    #: - [UDP] Avocent AuthSrv Protocol
    avauthsrvprtcl = 2068, 'avauthsrvprtcl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HTTP Event Port
    #: - [UDP] HTTP Event Port
    event_port = 2069, 'event-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AH and ESP Encapsulated in UDP packet
    #: - [UDP] AH and ESP Encapsulated in UDP packet
    ah_esp_encap = 2070, 'ah-esp-encap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Axon Control Protocol
    #: - [UDP] Axon Control Protocol
    acp_port = 2071, 'acp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GlobeCast mSync
    #: - [UDP] GlobeCast mSync
    msync = 2072, 'msync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DataReel Database Socket
    #: - [UDP] DataReel Database Socket
    gxs_data_port = 2073, 'gxs-data-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vertel VMF SA
    #: - [UDP] Vertel VMF SA
    vrtl_vmf_sa = 2074, 'vrtl-vmf-sa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Newlix ServerWare Engine
    #: - [UDP] Newlix ServerWare Engine
    newlixengine = 2075, 'newlixengine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Newlix JSPConfig
    #: - [UDP] Newlix JSPConfig
    newlixconfig = 2076, 'newlixconfig', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Old Tivoli Storage Manager
    #: - [UDP] Old Tivoli Storage Manager
    tsrmagt = 2077, 'tsrmagt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Total Productivity Center Server
    #: - [UDP] IBM Total Productivity Center Server
    tpcsrvr = 2078, 'tpcsrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IDWARE Router Port
    #: - [UDP] IDWARE Router Port
    idware_router = 2079, 'idware-router', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Autodesk NLM (FLEXlm)
    #: - [UDP] Autodesk NLM (FLEXlm)
    autodesk_nlm = 2080, 'autodesk-nlm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KME PRINTER TRAP PORT
    #: - [UDP] KME PRINTER TRAP PORT
    kme_trap_port = 2081, 'kme-trap-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Infowave Mobility Server
    #: - [UDP] Infowave Mobility Server
    infowave = 2082, 'infowave', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Radius Service [:rfc:`6614`]
    #: - [UDP] Secure Radius Service [:rfc:`7360`]
    radsec = 2083, 'radsec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SunCluster Geographic
    #: - [UDP] SunCluster Geographic
    sunclustergeo = 2084, 'sunclustergeo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ADA Control
    #: - [UDP] ADA Control
    ada_cip = 2085, 'ada-cip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GNUnet
    #: - [UDP] GNUnet
    gnunet = 2086, 'gnunet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ELI - Event Logging Integration
    #: - [UDP] ELI - Event Logging Integration
    eli = 2087, 'eli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IP Busy Lamp Field
    #: - [UDP] IP Busy Lamp Field
    ip_blf = 2088, 'ip-blf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Security Encapsulation Protocol - SEP
    #: - [UDP] Security Encapsulation Protocol - SEP
    sep = 2089, 'sep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Load Report Protocol
    #: - [UDP] Load Report Protocol
    lrp = 2090, 'lrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PRP
    #: - [UDP] PRP
    prp = 2091, 'prp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Descent 3
    #: - [UDP] Descent 3
    descent3 = 2092, 'descent3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NBX CC
    #: - [UDP] NBX CC
    nbx_cc = 2093, 'nbx-cc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NBX AU
    #: - [UDP] NBX AU
    nbx_au = 2094, 'nbx-au', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NBX SER
    #: - [UDP] NBX SER
    nbx_ser = 2095, 'nbx-ser', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NBX DIR
    #: - [UDP] NBX DIR
    nbx_dir = 2096, 'nbx-dir', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jet Form Preview
    #: - [UDP] Jet Form Preview
    jetformpreview = 2097, 'jetformpreview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dialog Port
    #: - [UDP] Dialog Port
    dialog_port = 2098, 'dialog-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] H.225.0 Annex G Signalling
    #: - [UDP] H.225.0 Annex G Signalling
    h2250_annex_g = 2099, 'h2250-annex-g', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Amiga Network Filesystem
    #: - [UDP] Amiga Network Filesystem
    amiganetfs = 2100, 'amiganetfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rtcm-sc104
    #: - [UDP] rtcm-sc104
    rtcm_sc104 = 2101, 'rtcm-sc104', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zephyr server
    #: - [UDP] Zephyr server
    zephyr_srv = 2102, 'zephyr-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zephyr serv-hm connection
    #: - [UDP] Zephyr serv-hm connection
    zephyr_clt = 2103, 'zephyr-clt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zephyr hostmanager
    #: - [UDP] Zephyr hostmanager
    zephyr_hm = 2104, 'zephyr-hm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MiniPay
    #: - [UDP] MiniPay
    minipay = 2105, 'minipay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MZAP
    #: - [UDP] MZAP
    mzap = 2106, 'mzap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BinTec Admin
    #: - [UDP] BinTec Admin
    bintec_admin = 2107, 'bintec-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Comcam
    #: - [UDP] Comcam
    comcam = 2108, 'comcam', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ergolight
    #: - [UDP] Ergolight
    ergolight = 2109, 'ergolight', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UMSP
    #: - [UDP] UMSP
    umsp = 2110, 'umsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPNET Dynamic Sampling Agent Transaction Protocol
    #: - [UDP] OPNET Dynamic Sampling Agent Transaction Protocol
    dsatp = 2111, 'dsatp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Idonix MetaNet
    #: - [UDP] Idonix MetaNet
    idonix_metanet = 2112, 'idonix-metanet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HSL StoRM
    #: - [UDP] HSL StoRM
    hsl_storm = 2113, 'hsl-storm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Classical Music Meta-Data Access and Enhancement
    #: - [UDP] Classical Music Meta-Data Access and Enhancement
    ariascribe = 2114, 'ariascribe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Key Distribution Manager
    #: - [UDP] Key Distribution Manager
    kdm = 2115, 'kdm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CCOWCMR
    #: - [UDP] CCOWCMR
    ccowcmr = 2116, 'ccowcmr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MENTACLIENT
    #: - [UDP] MENTACLIENT
    mentaclient = 2117, 'mentaclient', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MENTASERVER
    #: - [UDP] MENTASERVER
    mentaserver = 2118, 'mentaserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GSIGATEKEEPER
    #: - [UDP] GSIGATEKEEPER
    gsigatekeeper = 2119, 'gsigatekeeper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quick Eagle Networks CP
    #: - [UDP] Quick Eagle Networks CP
    qencp = 2120, 'qencp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCIENTIA-SSDB
    #: - [UDP] SCIENTIA-SSDB
    scientia_ssdb = 2121, 'scientia-ssdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CauPC Remote Control
    #: - [UDP] CauPC Remote Control
    caupc_remote = 2122, 'caupc-remote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GTP-Control Plane (3GPP)
    #: - [UDP] GTP-Control Plane (3GPP)
    gtp_control = 2123, 'gtp-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ELATELINK
    #: - [UDP] ELATELINK
    elatelink = 2124, 'elatelink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LOCKSTEP
    #: - [UDP] LOCKSTEP
    lockstep = 2125, 'lockstep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PktCable-COPS
    #: - [UDP] PktCable-COPS
    pktcable_cops = 2126, 'pktcable-cops', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INDEX-PC-WB
    #: - [UDP] INDEX-PC-WB
    index_pc_wb = 2127, 'index-pc-wb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net Steward Control
    #: - [UDP] Net Steward Control
    net_steward = 2128, 'net-steward', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cs-live.com
    #: - [UDP] cs-live.com
    cs_live = 2129, 'cs-live', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XDS
    #: - [UDP] XDS
    xds = 2130, 'xds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avantageb2b
    #: - [UDP] Avantageb2b
    avantageb2b = 2131, 'avantageb2b', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SoleraTec End Point Map
    #: - [UDP] SoleraTec End Point Map
    solera_epmap = 2132, 'solera-epmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ZYMED-ZPP
    #: - [UDP] ZYMED-ZPP
    zymed_zpp = 2133, 'zymed-zpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AVENUE
    #: - [UDP] AVENUE
    avenue = 2134, 'avenue', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Grid Resource Information Server
    #: - [UDP] Grid Resource Information Server
    gris = 2135, 'gris', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APPWORXSRV
    #: - [UDP] APPWORXSRV
    appworxsrv = 2136, 'appworxsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CONNECT
    #: - [UDP] CONNECT
    connect = 2137, 'connect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UNBIND-CLUSTER
    #: - [UDP] UNBIND-CLUSTER
    unbind_cluster = 2138, 'unbind-cluster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IAS-AUTH
    #: - [UDP] IAS-AUTH
    ias_auth = 2139, 'ias-auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IAS-REG
    #: - [UDP] IAS-REG
    ias_reg = 2140, 'ias-reg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IAS-ADMIND
    #: - [UDP] IAS-ADMIND
    ias_admind = 2141, 'ias-admind', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TDM OVER IP [:rfc:`5087`]
    #: - [UDP] TDM OVER IP [:rfc:`5087`]
    tdmoip = 2142, 'tdmoip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Live Vault Job Control
    #: - [UDP] Live Vault Job Control
    lv_jc = 2143, 'lv-jc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Live Vault Fast Object Transfer
    #: - [UDP] Live Vault Fast Object Transfer
    lv_ffx = 2144, 'lv-ffx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Live Vault Remote Diagnostic Console Support
    #: - [UDP] Live Vault Remote Diagnostic Console Support
    lv_pici = 2145, 'lv-pici', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Live Vault Admin Event Notification
    #: - [UDP] Live Vault Admin Event Notification
    lv_not = 2146, 'lv-not', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Live Vault Authentication
    #: - [UDP] Live Vault Authentication
    lv_auth = 2147, 'lv-auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS UNIVERSAL COMMUNICATION LAYER
    #: - [UDP] VERITAS UNIVERSAL COMMUNICATION LAYER
    veritas_ucl = 2148, 'veritas-ucl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACPTSYS
    #: - [UDP] ACPTSYS
    acptsys = 2149, 'acptsys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DYNAMIC3D
    #: - [UDP] DYNAMIC3D
    dynamic3d = 2150, 'dynamic3d', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DOCENT
    #: - [UDP] DOCENT
    docent = 2151, 'docent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GTP-User Plane (3GPP)
    #: - [UDP] GTP-User Plane (3GPP)
    gtp_user = 2152, 'gtp-user', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Control Protocol
    #: - [UDP] Control Protocol
    ctlptc = 2153, 'ctlptc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Standard Protocol
    #: - [UDP] Standard Protocol
    stdptc = 2154, 'stdptc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bridge Protocol
    #: - [UDP] Bridge Protocol
    brdptc = 2155, 'brdptc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talari Reliable Protocol
    #: - [UDP] Talari Reliable Protocol
    trp = 2156, 'trp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xerox Network Document Scan Protocol
    #: - [UDP] Xerox Network Document Scan Protocol
    xnds = 2157, 'xnds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TouchNetPlus Service
    #: - [UDP] TouchNetPlus Service
    touchnetplus = 2158, 'touchnetplus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GDB Remote Debug Port
    #: - [UDP] GDB Remote Debug Port
    gdbremote = 2159, 'gdbremote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 2160
    #: - [UDP] APC 2160
    apc_2160 = 2160, 'apc-2160', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 2161
    #: - [UDP] APC 2161
    apc_2161 = 2161, 'apc-2161', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Navisphere
    #: - [UDP] Navisphere
    navisphere = 2162, 'navisphere', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Navisphere Secure
    #: - [UDP] Navisphere Secure
    navisphere_sec = 2163, 'navisphere-sec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dynamic DNS Version 3
    #: - [UDP] Dynamic DNS Version 3
    ddns_v3 = 2164, 'ddns-v3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] X-Bone API
    #: - [UDP] X-Bone API
    x_bone_api = 2165, 'x-bone-api', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iwserver
    #: - [UDP] iwserver
    iwserver = 2166, 'iwserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Raw Async Serial Link
    #: - [UDP] Raw Async Serial Link
    raw_serial = 2167, 'raw-serial', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] easy-soft Multiplexer
    #: - [UDP] easy-soft Multiplexer
    easy_soft_mux = 2168, 'easy-soft-mux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Backbone for Academic Information Notification (BRAIN)
    #: - [UDP] Backbone for Academic Information Notification (BRAIN)
    brain = 2169, 'brain', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EyeTV Server Port
    #: - [UDP] EyeTV Server Port
    eyetv = 2170, 'eyetv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS Firewall Storage
    #: - [UDP] MS Firewall Storage
    msfw_storage = 2171, 'msfw-storage', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS Firewall SecureStorage
    #: - [UDP] MS Firewall SecureStorage
    msfw_s_storage = 2172, 'msfw-s-storage', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS Firewall Replication
    #: - [UDP] MS Firewall Replication
    msfw_replica = 2173, 'msfw-replica', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS Firewall Intra Array
    #: - [UDP] MS Firewall Intra Array
    msfw_array = 2174, 'msfw-array', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Desktop AirSync Protocol
    #: - [UDP] Microsoft Desktop AirSync Protocol
    airsync = 2175, 'airsync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft ActiveSync Remote API
    #: - [UDP] Microsoft ActiveSync Remote API
    rapi = 2176, 'rapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qWAVE Bandwidth Estimate
    #: - [UDP] qWAVE Bandwidth Estimate
    qwave = 2177, 'qwave', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Peer Services for BITS
    #: - [UDP] Peer Services for BITS
    bitspeer = 2178, 'bitspeer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft RDP for virtual machines
    #: - [UDP] Microsoft RDP for virtual machines
    vmrdp = 2179, 'vmrdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Millicent Vendor Gateway Server
    #: - [UDP] Millicent Vendor Gateway Server
    mc_gt_srv = 2180, 'mc-gt-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eforward
    #: - [UDP] eforward
    eforward = 2181, 'eforward', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CGN status
    #: - [UDP] CGN status
    cgn_stat = 2182, 'cgn-stat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Code Green configuration
    #: - [UDP] Code Green configuration
    cgn_config = 2183, 'cgn-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OnBase Distributed Disk Services
    #: - [UDP] OnBase Distributed Disk Services
    onbase_dds = 2185, 'onbase-dds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Guy-Tek Automated Update Applications
    #: - [UDP] Guy-Tek Automated Update Applications
    gtaua = 2186, 'gtaua', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Sepehr System Management Control
    ssmc = 2187, 'ssmc', TransportProtocol.get('tcp')

    #: [UDP] Sepehr System Management Data
    ssmd = 2187, 'ssmd', TransportProtocol.get('udp')

    #: [TCP] Radware Resource Pool Manager
    radware_rpm = 2188, 'radware-rpm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2188 = 2188, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Secure Radware Resource Pool Manager
    radware_rpm_s = 2189, 'radware-rpm-s', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2189 = 2189, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] TiVoConnect Beacon
    #: - [UDP] TiVoConnect Beacon
    tivoconnect = 2190, 'tivoconnect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TvBus Messaging
    #: - [UDP] TvBus Messaging
    tvbus = 2191, 'tvbus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASDIS software management
    #: - [UDP] ASDIS software management
    asdis = 2192, 'asdis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dr.Web Enterprise Management Service
    #: - [UDP] Dr.Web Enterprise Management Service
    drwcs = 2193, 'drwcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MNP data exchange
    #: - [UDP] MNP data exchange
    mnp_exchange = 2197, 'mnp-exchange', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OneHome Remote Access
    #: - [UDP] OneHome Remote Access
    onehome_remote = 2198, 'onehome-remote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OneHome Service Port
    #: - [UDP] OneHome Service Port
    onehome_help = 2199, 'onehome-help', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_2200 = 2200, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Advanced Training System Program
    #: - [UDP] Advanced Training System Program
    ats = 2201, 'ats', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Int. Multimedia Teleconferencing Cosortium
    #: - [UDP] Int. Multimedia Teleconferencing Cosortium
    imtc_map = 2202, 'imtc-map', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] b2 Runtime Protocol
    #: - [UDP] b2 Runtime Protocol
    b2_runtime = 2203, 'b2-runtime', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] b2 License Server
    #: - [UDP] b2 License Server
    b2_license = 2204, 'b2-license', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Java Presentation Server
    #: - [UDP] Java Presentation Server
    jps = 2205, 'jps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP OpenCall bus
    #: - [UDP] HP OpenCall bus
    hpocbus = 2206, 'hpocbus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP Status and Services
    #: - [UDP] HP Status and Services
    hpssd = 2207, 'hpssd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP I/O Backend
    #: - [UDP] HP I/O Backend
    hpiod = 2208, 'hpiod', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP RIM for Files Portal Service
    #: - [UDP] HP RIM for Files Portal Service
    rimf_ps = 2209, 'rimf-ps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NOAAPORT Broadcast Network
    #: - [UDP] NOAAPORT Broadcast Network
    noaaport = 2210, 'noaaport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMWIN
    #: - [UDP] EMWIN
    emwin = 2211, 'emwin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LeeCO POS Server Service
    #: - [UDP] LeeCO POS Server Service
    leecoposserver = 2212, 'leecoposserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kali
    #: - [UDP] Kali
    kali = 2213, 'kali', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RDQ Protocol Interface
    #: - [UDP] RDQ Protocol Interface
    rpi = 2214, 'rpi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPCore.co.za GPRS
    #: - [UDP] IPCore.co.za GPRS
    ipcore = 2215, 'ipcore', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VTU data service
    #: - [UDP] VTU data service
    vtu_comms = 2216, 'vtu-comms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GoToDevice Device Management
    #: - [UDP] GoToDevice Device Management
    gotodevice = 2217, 'gotodevice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bounzza IRC Proxy
    #: - [UDP] Bounzza IRC Proxy
    bounzza = 2218, 'bounzza', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetIQ NCAP Protocol
    #: - [UDP] NetIQ NCAP Protocol
    netiq_ncap = 2219, 'netiq-ncap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetIQ End2End
    #: - [UDP] NetIQ End2End
    netiq = 2220, 'netiq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EtherNet/IP over TLS
    #: - [UDP] EtherNet/IP over DTLS
    ethernet_ip_s = 2221, 'ethernet-ip-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EtherNet/IP I/O IANA assigned this well-formed service name as a
    #: replacement for "EtherNet/IP-1".
    #: - [TCP] EtherNet/IP I/O
    #: - [UDP] EtherNet/IP I/O IANA assigned this well-formed service name as a
    #: replacement for "EtherNet/IP-1".
    #: - [UDP] EtherNet/IP I/O
    ethernet_ip_1 = 2222, 'ethernet-ip-1', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Rockwell CSP2
    #: - [UDP] Rockwell CSP2
    rockwell_csp2 = 2223, 'rockwell-csp2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Easy Flexible Internet/Multiplayer Games
    #: - [UDP] Easy Flexible Internet/Multiplayer Games
    efi_mg = 2224, 'efi-mg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Resource Connection Initiation Protocol
    #: - [SCTP] Resource Connection Initiation Protocol
    rcip_itu = 2225, 'rcip-itu', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_2225 = 2225, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Digital Instinct DRM
    #: - [UDP] Digital Instinct DRM
    di_drm = 2226, 'di-drm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DI Messaging Service
    #: - [UDP] DI Messaging Service
    di_msg = 2227, 'di-msg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eHome Message Server
    #: - [UDP] eHome Message Server
    ehome_ms = 2228, 'ehome-ms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DataLens Service
    #: - [UDP] DataLens Service
    datalens = 2229, 'datalens', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MetaSoft Job Queue Administration Service
    #: - [UDP] MetaSoft Job Queue Administration Service
    queueadm = 2230, 'queueadm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WiMAX ASN Control Plane Protocol
    #: - [UDP] WiMAX ASN Control Plane Protocol
    wimaxasncp = 2231, 'wimaxasncp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IVS Video default
    #: - [UDP] IVS Video default
    ivs_video = 2232, 'ivs-video', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INFOCRYPT
    #: - [UDP] INFOCRYPT
    infocrypt = 2233, 'infocrypt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DirectPlay
    #: - [UDP] DirectPlay
    directplay = 2234, 'directplay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sercomm-WLink
    #: - [UDP] Sercomm-WLink
    sercomm_wlink = 2235, 'sercomm-wlink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nani
    #: - [UDP] Nani
    nani = 2236, 'nani', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Optech Port1 License Manager
    #: - [UDP] Optech Port1 License Manager
    optech_port1_lm = 2237, 'optech-port1-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AVIVA SNA SERVER
    #: - [UDP] AVIVA SNA SERVER
    aviva_sna = 2238, 'aviva-sna', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Image Query
    #: - [UDP] Image Query
    imagequery = 2239, 'imagequery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RECIPe
    #: - [UDP] RECIPe
    recipe = 2240, 'recipe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IVS Daemon
    #: - [UDP] IVS Daemon
    ivsd = 2241, 'ivsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Folio Remote Server
    #: - [UDP] Folio Remote Server
    foliocorp = 2242, 'foliocorp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Magicom Protocol
    #: - [UDP] Magicom Protocol
    magicom = 2243, 'magicom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NMS Server
    #: - [UDP] NMS Server
    nmsserver = 2244, 'nmsserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HaO
    #: - [UDP] HaO
    hao = 2245, 'hao', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PacketCable MTA Addr Map
    #: - [UDP] PacketCable MTA Addr Map
    pc_mta_addrmap = 2246, 'pc-mta-addrmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Antidote Deployment Manager Service
    #: - [UDP] Antidote Deployment Manager Service
    antidotemgrsvr = 2247, 'antidotemgrsvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] User Management Service
    #: - [UDP] User Management Service
    ums = 2248, 'ums', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RISO File Manager Protocol
    #: - [UDP] RISO File Manager Protocol
    rfmp = 2249, 'rfmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] remote-collab
    #: - [UDP] remote-collab
    remote_collab = 2250, 'remote-collab', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Distributed Framework Port
    #: - [UDP] Distributed Framework Port
    dif_port = 2251, 'dif-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NJENET using SSL
    #: - [UDP] NJENET using SSL
    njenet_ssl = 2252, 'njenet-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DTV Channel Request
    #: - [UDP] DTV Channel Request
    dtv_chan_req = 2253, 'dtv-chan-req', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Seismic P.O.C. Port
    #: - [UDP] Seismic P.O.C. Port
    seispoc = 2254, 'seispoc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VRTP - ViRtue Transfer Protocol
    #: - [UDP] VRTP - ViRtue Transfer Protocol
    vrtp = 2255, 'vrtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCC MFP
    #: - [UDP] PCC MFP
    pcc_mfp = 2256, 'pcc-mfp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] simple text/file transfer
    #: - [UDP] simple text/file transfer
    simple_tx_rx = 2257, 'simple-tx-rx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rotorcraft Communications Test System
    #: - [UDP] Rotorcraft Communications Test System
    rcts = 2258, 'rcts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BIF identifiers resolution service
    #: - [UDP] BIF identifiers resolution service
    bid_serv = 2259, 'bid-serv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 2260
    #: - [UDP] APC 2260
    apc_2260 = 2260, 'apc-2260', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CoMotion Master Server
    #: - [UDP] CoMotion Master Server
    comotionmaster = 2261, 'comotionmaster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CoMotion Backup Server
    #: - [UDP] CoMotion Backup Server
    comotionback = 2262, 'comotionback', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ECweb Configuration Service
    #: - [UDP] ECweb Configuration Service
    ecwcfg = 2263, 'ecwcfg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Audio Precision Apx500 API Port 1
    #: - [UDP] Audio Precision Apx500 API Port 1
    apx500api_1 = 2264, 'apx500api-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Audio Precision Apx500 API Port 2
    #: - [UDP] Audio Precision Apx500 API Port 2
    apx500api_2 = 2265, 'apx500api-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] M-Files Server
    #: - [UDP] M-files Server
    mfserver = 2266, 'mfserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OntoBroker
    #: - [UDP] OntoBroker
    ontobroker = 2267, 'ontobroker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMT [:rfc:`7450`]
    #: - [UDP] AMT [:rfc:`7450`]
    amt = 2268, 'amt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MIKEY
    #: - [UDP] MIKEY
    mikey = 2269, 'mikey', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] starSchool
    #: - [UDP] starSchool
    starschool = 2270, 'starschool', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Meeting Maker Scheduling
    #: - [UDP] Secure Meeting Maker Scheduling
    mmcals = 2271, 'mmcals', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Meeting Maker Scheduling
    #: - [UDP] Meeting Maker Scheduling
    mmcal = 2272, 'mmcal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MySQL Instance Manager
    #: - [UDP] MySQL Instance Manager
    mysql_im = 2273, 'mysql-im', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCTTunneller
    #: - [UDP] PCTTunneller
    pcttunnell = 2274, 'pcttunnell', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iBridge Conferencing
    #: - [UDP] iBridge Conferencing
    ibridge_data = 2275, 'ibridge-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iBridge Management
    #: - [UDP] iBridge Management
    ibridge_mgmt = 2276, 'ibridge-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bt device control proxy
    #: - [UDP] Bt device control proxy
    bluectrlproxy = 2277, 'bluectrlproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Stacked Sequences Database
    #: - [UDP] Simple Stacked Sequences Database
    s3db = 2278, 's3db', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xmquery
    #: - [UDP] xmquery
    xmquery = 2279, 'xmquery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LNVPOLLER
    #: - [UDP] LNVPOLLER
    lnvpoller = 2280, 'lnvpoller', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LNVCONSOLE
    #: - [UDP] LNVCONSOLE
    lnvconsole = 2281, 'lnvconsole', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LNVALARM
    #: - [UDP] LNVALARM
    lnvalarm = 2282, 'lnvalarm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LNVSTATUS
    #: - [UDP] LNVSTATUS
    lnvstatus = 2283, 'lnvstatus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LNVMAPS
    #: - [UDP] LNVMAPS
    lnvmaps = 2284, 'lnvmaps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LNVMAILMON
    #: - [UDP] LNVMAILMON
    lnvmailmon = 2285, 'lnvmailmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NAS-Metering
    #: - [UDP] NAS-Metering
    nas_metering = 2286, 'nas-metering', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNA
    #: - [UDP] DNA
    dna = 2287, 'dna', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETML
    #: - [UDP] NETML
    netml = 2288, 'netml', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lookup dict server
    #: - [UDP] Lookup dict server
    dict_lookup = 2289, 'dict-lookup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sonus Logging Services
    #: - [UDP] Sonus Logging Services
    sonus_logging = 2290, 'sonus-logging', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPSON Advanced Printer Share Protocol
    #: - [UDP] EPSON Advanced Printer Share Protocol
    eapsp = 2291, 'eapsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sonus Element Management Services
    #: - [UDP] Sonus Element Management Services
    mib_streaming = 2292, 'mib-streaming', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Platform Debug Manager
    #: - [UDP] Network Platform Debug Manager
    npdbgmngr = 2293, 'npdbgmngr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Konshus License Manager (FLEX)
    #: - [UDP] Konshus License Manager (FLEX)
    konshus_lm = 2294, 'konshus-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Advant License Manager
    #: - [UDP] Advant License Manager
    advant_lm = 2295, 'advant-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Theta License Manager (Rainbow)
    #: - [UDP] Theta License Manager (Rainbow)
    theta_lm = 2296, 'theta-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D2K DataMover 1
    #: - [UDP] D2K DataMover 1
    d2k_datamover1 = 2297, 'd2k-datamover1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D2K DataMover 2
    #: - [UDP] D2K DataMover 2
    d2k_datamover2 = 2298, 'd2k-datamover2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PC Telecommute
    #: - [UDP] PC Telecommute
    pc_telecommute = 2299, 'pc-telecommute', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CVMMON
    #: - [UDP] CVMMON
    cvmmon = 2300, 'cvmmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compaq HTTP
    #: - [UDP] Compaq HTTP
    cpq_wbem = 2301, 'cpq-wbem', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bindery Support
    #: - [UDP] Bindery Support
    binderysupport = 2302, 'binderysupport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Proxy Gateway
    #: - [UDP] Proxy Gateway
    proxy_gateway = 2303, 'proxy-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Attachmate UTS
    #: - [UDP] Attachmate UTS
    attachmate_uts = 2304, 'attachmate-uts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MT ScaleServer
    #: - [UDP] MT ScaleServer
    mt_scaleserver = 2305, 'mt-scaleserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TAPPI BoxNet
    #: - [UDP] TAPPI BoxNet
    tappi_boxnet = 2306, 'tappi-boxnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pehelp
    #: - [UDP] pehelp
    pehelp = 2307, 'pehelp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sdhelp
    #: - [UDP] sdhelp
    sdhelp = 2308, 'sdhelp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SD Server
    #: - [UDP] SD Server
    sdserver = 2309, 'sdserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SD Client
    #: - [UDP] SD Client
    sdclient = 2310, 'sdclient', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Message Service
    #: - [UDP] Message Service
    messageservice = 2311, 'messageservice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WANScaler Communication Service
    #: - [UDP] WANScaler Communication Service
    wanscaler = 2312, 'wanscaler', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IAPP (Inter Access Point Protocol)
    #: - [UDP] IAPP (Inter Access Point Protocol)
    iapp = 2313, 'iapp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CR WebSystems
    #: - [UDP] CR WebSystems
    cr_websystems = 2314, 'cr-websystems', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Precise Sft.
    #: - [UDP] Precise Sft.
    precise_sft = 2315, 'precise-sft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SENT License Manager
    #: - [UDP] SENT License Manager
    sent_lm = 2316, 'sent-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Attachmate G32
    #: - [UDP] Attachmate G32
    attachmate_g32 = 2317, 'attachmate-g32', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cadence Control
    #: - [UDP] Cadence Control
    cadencecontrol = 2318, 'cadencecontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] InfoLibria
    #: - [UDP] InfoLibria
    infolibria = 2319, 'infolibria', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Siebel NS
    #: - [UDP] Siebel NS
    siebel_ns = 2320, 'siebel-ns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RDLAP
    #: - [UDP] RDLAP
    rdlap = 2321, 'rdlap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ofsd
    #: - [UDP] ofsd
    ofsd = 2322, 'ofsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3d-nfsd
    #: - [UDP] 3d-nfsd
    AppType_3d_nfsd = 2323, '3d-nfsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cosmocall
    #: - [UDP] Cosmocall
    cosmocall = 2324, 'cosmocall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ANSYS Licensing Interconnect
    #: - [UDP] ANSYS Licensing Interconnect
    ansysli = 2325, 'ansysli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IDCP
    #: - [UDP] IDCP
    idcp = 2326, 'idcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xingcsm
    #: - [UDP] xingcsm
    xingcsm = 2327, 'xingcsm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netrix SFTM
    #: - [UDP] Netrix SFTM
    netrix_sftm = 2328, 'netrix-sftm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NVD User
    #: - [UDP] NVD User
    nvd_2184 = 2184, 'nvd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NVD
    #: - [UDP] NVD
    nvd_2329 = 2329, 'nvd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TSCCHAT
    #: - [UDP] TSCCHAT
    tscchat = 2330, 'tscchat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AGENTVIEW
    #: - [UDP] AGENTVIEW
    agentview = 2331, 'agentview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RCC Host
    #: - [UDP] RCC Host
    rcc_host = 2332, 'rcc-host', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNAPP
    #: - [UDP] SNAPP
    snapp = 2333, 'snapp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACE Client Auth
    #: - [UDP] ACE Client Auth
    ace_client = 2334, 'ace-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACE Proxy
    #: - [UDP] ACE Proxy
    ace_proxy = 2335, 'ace-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apple UG Control
    #: - [UDP] Apple UG Control
    appleugcontrol = 2336, 'appleugcontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ideesrv
    #: - [UDP] ideesrv
    ideesrv = 2337, 'ideesrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Norton Lambert
    #: - [UDP] Norton Lambert
    norton_lambert = 2338, 'norton-lambert', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3Com WebView
    #: - [UDP] 3Com WebView
    AppType_3com_webview = 2339, '3com-webview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WRS Registry IANA assigned this well-formed service name as a
    #: replacement for "wrs_registry".
    #: - [TCP] WRS Registry
    #: - [UDP] WRS Registry IANA assigned this well-formed service name as a
    #: replacement for "wrs_registry".
    #: - [UDP] WRS Registry
    wrs_registry = 2340, 'wrs-registry', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] XIO Status
    #: - [UDP] XIO Status
    xiostatus = 2341, 'xiostatus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Seagate Manage Exec
    #: - [UDP] Seagate Manage Exec
    manage_exec = 2342, 'manage-exec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nati logos
    #: - [UDP] nati logos
    nati_logos = 2343, 'nati-logos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fcmsys
    #: - [UDP] fcmsys
    fcmsys = 2344, 'fcmsys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dbm
    #: - [UDP] dbm
    dbm = 2345, 'dbm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Game Connection Port IANA assigned this well-formed service name as a
    #: replacement for "redstorm_join".
    #: - [TCP] Game Connection Port
    #: - [UDP] Game Connection Port IANA assigned this well-formed service name as a
    #: replacement for "redstorm_join".
    #: - [UDP] Game Connection Port
    redstorm_join = 2346, 'redstorm-join', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Game Announcement and Location IANA assigned this well-formed service
    #: name as a replacement for "redstorm_find".
    #: - [TCP] Game Announcement and Location
    #: - [UDP] Game Announcement and Location IANA assigned this well-formed service
    #: name as a replacement for "redstorm_find".
    #: - [UDP] Game Announcement and Location
    redstorm_find = 2347, 'redstorm-find', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Information to query for game status IANA assigned this well-formed
    #: service name as a replacement for "redstorm_info".
    #: - [TCP] Information to query for game status
    #: - [UDP] Information to query for game status IANA assigned this well-formed
    #: service name as a replacement for "redstorm_info".
    #: - [UDP] Information to query for game status
    redstorm_info = 2348, 'redstorm-info', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Diagnostics Port IANA assigned this well-formed service name as a
    #: replacement for "redstorm_diag".
    #: - [TCP] Diagnostics Port
    #: - [UDP] Diagnostics Port IANA assigned this well-formed service name as a
    #: replacement for "redstorm_diag".
    #: - [UDP] Diagnostics Port
    redstorm_diag = 2349, 'redstorm-diag', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Pharos Booking Server
    #: - [UDP] Pharos Booking Server
    psbserver = 2350, 'psbserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] psrserver
    #: - [UDP] psrserver
    psrserver = 2351, 'psrserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pslserver
    #: - [UDP] pslserver
    pslserver = 2352, 'pslserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pspserver
    #: - [UDP] pspserver
    pspserver = 2353, 'pspserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] psprserver
    #: - [UDP] psprserver
    psprserver = 2354, 'psprserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] psdbserver
    #: - [UDP] psdbserver
    psdbserver = 2355, 'psdbserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GXT License Managemant
    #: - [UDP] GXT License Managemant
    gxtelmd = 2356, 'gxtelmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UniHub Server
    #: - [UDP] UniHub Server
    unihub_server = 2357, 'unihub-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Futrix
    #: - [UDP] Futrix
    futrix = 2358, 'futrix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FlukeServer
    #: - [UDP] FlukeServer
    flukeserver = 2359, 'flukeserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NexstorIndLtd
    #: - [UDP] NexstorIndLtd
    nexstorindltd = 2360, 'nexstorindltd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TL1
    #: - [UDP] TL1
    tl1 = 2361, 'tl1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] digiman
    #: - [UDP] digiman
    digiman = 2362, 'digiman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Media Central NFSD
    #: - [UDP] Media Central NFSD
    mediacntrlnfsd = 2363, 'mediacntrlnfsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OI-2000
    #: - [UDP] OI-2000
    oi_2000 = 2364, 'oi-2000', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dbref
    #: - [UDP] dbref
    dbref = 2365, 'dbref', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qip-login
    #: - [UDP] qip-login
    qip_login = 2366, 'qip-login', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Control
    #: - [UDP] Service Control
    service_ctrl = 2367, 'service-ctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenTable
    #: - [UDP] OpenTable
    opentable = 2368, 'opentable', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Blockchain Identifier InFrastructure P2P
    #: - [UDP] Blockchain Identifier InFrastructure P2P
    bif_p2p = 2369, 'bif-p2p', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] L3-HBMon
    #: - [UDP] L3-HBMon
    l3_hbmon = 2370, 'l3-hbmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RDA
    #: - [UDP] RDA
    rda_630 = 630, 'rda', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Remote Device Access
    rda_2371 = 2371, 'rda', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2371 = 2371, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] LanMessenger
    #: - [UDP] LanMessenger
    lanmessenger = 2372, 'lanmessenger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Remograph License Manager
    remographlm = 2373, 'remographlm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2373 = 2373, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Hydra RPC
    hydra = 2374, 'hydra', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2374 = 2374, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Docker REST API (plain text)
    docker = 2375, 'docker', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2375 = 2375, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Docker REST API (ssl)
    docker_s = 2376, 'docker-s', TransportProtocol.get('tcp')

    #: [TCP] RPC interface for Docker Swarm
    swarm = 2377, 'swarm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2377 = 2377, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_2378 = 2378, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] DALI lighting control
    dali = 2378, 'dali', TransportProtocol.get('udp')

    #: [TCP] etcd client communication
    etcd_client = 2379, 'etcd-client', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2379 = 2379, 'reserved', TransportProtocol.get('udp')

    #: [TCP] etcd server to server communication
    etcd_server = 2380, 'etcd-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2380 = 2380, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Compaq HTTPS
    #: - [UDP] Compaq HTTPS
    compaq_https = 2381, 'compaq-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft OLAP
    #: - [UDP] Microsoft OLAP
    ms_olap3 = 2382, 'ms-olap3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft OLAP
    #: - [UDP] Microsoft OLAP
    ms_olap4 = 2383, 'ms-olap4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SD-REQUEST
    sd_request = 2384, 'sd-request', TransportProtocol.get('tcp')

    #: [UDP] SD-CAPACITY
    sd_capacity = 2384, 'sd-capacity', TransportProtocol.get('udp')

    #: - [TCP] SD-DATA
    #: - [UDP] SD-DATA
    sd_data = 2385, 'sd-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Tape
    #: - [UDP] Virtual Tape
    virtualtape = 2386, 'virtualtape', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSAM Redirector
    #: - [UDP] VSAM Redirector
    vsamredirector = 2387, 'vsamredirector', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MYNAH AutoStart
    #: - [UDP] MYNAH AutoStart
    mynahautostart = 2388, 'mynahautostart', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView Session Mgr
    #: - [UDP] OpenView Session Mgr
    ovsessionmgr = 2389, 'ovsessionmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSMTP
    #: - [UDP] RSMTP
    rsmtp = 2390, 'rsmtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3COM Net Management
    #: - [UDP] 3COM Net Management
    AppType_3com_net_mgmt = 2391, '3com-net-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tactical Auth
    #: - [UDP] Tactical Auth
    tacticalauth = 2392, 'tacticalauth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS OLAP 1
    #: - [UDP] MS OLAP 1
    ms_olap1 = 2393, 'ms-olap1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS OLAP 2
    #: - [UDP] MS OLAP 2
    ms_olap2 = 2394, 'ms-olap2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LAN900 Remote IANA assigned this well-formed service name as a
    #: replacement for "lan900_remote".
    #: - [TCP] LAN900 Remote
    #: - [UDP] LAN900 Remote IANA assigned this well-formed service name as a
    #: replacement for "lan900_remote".
    #: - [UDP] LAN900 Remote
    lan900_remote = 2395, 'lan900-remote', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Wusage
    #: - [UDP] Wusage
    wusage = 2396, 'wusage', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NCL
    #: - [UDP] NCL
    ncl = 2397, 'ncl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Orbiter
    #: - [UDP] Orbiter
    orbiter = 2398, 'orbiter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileMaker, Inc. - Data Access Layer
    #: - [UDP] FileMaker, Inc. - Data Access Layer
    fmpro_fdal = 2399, 'fmpro-fdal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpEquus Server
    #: - [UDP] OpEquus Server
    opequus_server = 2400, 'opequus-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cvspserver
    #: - [UDP] cvspserver
    cvspserver = 2401, 'cvspserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TaskMaster 2000 Server
    #: - [UDP] TaskMaster 2000 Server
    taskmaster2000_2402 = 2402, 'taskmaster2000', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TaskMaster 2000 Web
    #: - [UDP] TaskMaster 2000 Web
    taskmaster2000_2403 = 2403, 'taskmaster2000', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IEC 60870-5-104 process control over IP
    #: - [UDP] IEC 60870-5-104 process control over IP
    iec_104 = 2404, 'iec-104', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRC Netpoll
    #: - [UDP] TRC Netpoll
    trc_netpoll = 2405, 'trc-netpoll', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JediServer
    #: - [UDP] JediServer
    jediserver = 2406, 'jediserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Orion
    #: - [UDP] Orion
    orion = 2407, 'orion', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CloudFlare Railgun Web Acceleration Protocol
    railgun_webaccl = 2408, 'railgun-webaccl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2408 = 2408, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] SNS Protocol
    #: - [UDP] SNS Protocol
    sns_protocol = 2409, 'sns-protocol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VRTS Registry
    #: - [UDP] VRTS Registry
    vrts_registry = 2410, 'vrts-registry', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netwave AP Management
    #: - [UDP] Netwave AP Management
    netwave_ap_mgmt = 2411, 'netwave-ap-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CDN
    #: - [UDP] CDN
    cdn = 2412, 'cdn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] orion-rmi-reg
    #: - [UDP] orion-rmi-reg
    orion_rmi_reg = 2413, 'orion-rmi-reg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Beeyond
    #: - [UDP] Beeyond
    beeyond = 2414, 'beeyond', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Codima Remote Transaction Protocol
    #: - [UDP] Codima Remote Transaction Protocol
    codima_rtp = 2415, 'codima-rtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RMT Server
    #: - [UDP] RMT Server
    rmtserver = 2416, 'rmtserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Composit Server
    #: - [UDP] Composit Server
    composit_server = 2417, 'composit-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cas
    #: - [UDP] cas
    cas = 2418, 'cas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Attachmate S2S
    #: - [UDP] Attachmate S2S
    attachmate_s2s = 2419, 'attachmate-s2s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DSL Remote Management
    #: - [UDP] DSL Remote Management
    dslremote_mgmt = 2420, 'dslremote-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] G-Talk
    #: - [UDP] G-Talk
    g_talk = 2421, 'g-talk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CRMSBITS
    #: - [UDP] CRMSBITS
    crmsbits = 2422, 'crmsbits', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RNRP
    #: - [UDP] RNRP
    rnrp = 2423, 'rnrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KOFAX-SVR
    #: - [UDP] KOFAX-SVR
    kofax_svr = 2424, 'kofax-svr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu App Manager
    #: - [UDP] Fujitsu App Manager
    fjitsuappmgr = 2425, 'fjitsuappmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VeloCloud MultiPath Protocol
    #: - [UDP] VeloCloud MultiPath Protocol
    vcmp = 2426, 'vcmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Media Gateway Control Protocol Gateway
    #: - [UDP] Media Gateway Control Protocol Gateway
    mgcp_gateway = 2427, 'mgcp-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] One Way Trip Time
    #: - [UDP] One Way Trip Time
    ott = 2428, 'ott', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FT-ROLE
    #: - [UDP] FT-ROLE
    ft_role = 2429, 'ft-role', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] venus
    #: - [UDP] venus
    venus = 2430, 'venus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] venus-se
    #: - [UDP] venus-se
    venus_se = 2431, 'venus-se', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] codasrv
    #: - [UDP] codasrv
    codasrv = 2432, 'codasrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] codasrv-se
    #: - [UDP] codasrv-se
    codasrv_se = 2433, 'codasrv-se', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pxc-epmap
    #: - [UDP] pxc-epmap
    pxc_epmap = 2434, 'pxc-epmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OptiLogic
    #: - [UDP] OptiLogic
    optilogic = 2435, 'optilogic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TOP/X
    #: - [UDP] TOP/X
    topx = 2436, 'topx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Message Send Protocol (historic)
    #: - [UDP] Message Send Protocol (historic)
    msp_18 = 18, 'msp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSP
    #: - [UDP] MSP
    msp_2438 = 2438, 'msp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SybaseDBSynch
    #: - [UDP] SybaseDBSynch
    sybasedbsynch = 2439, 'sybasedbsynch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Spearway Lockers
    #: - [UDP] Spearway Lockers
    spearway = 2440, 'spearway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pervasive I*net Data Server
    #: - [UDP] Pervasive I*net Data Server
    pvsw_inet = 2441, 'pvsw-inet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netangel
    #: - [UDP] Netangel
    netangel = 2442, 'netangel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PowerClient Central Storage Facility
    #: - [UDP] PowerClient Central Storage Facility
    powerclientcsf = 2443, 'powerclientcsf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BT PP2 Sectrans
    #: - [UDP] BT PP2 Sectrans
    btpp2sectrans = 2444, 'btpp2sectrans', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DTN1
    #: - [UDP] DTN1
    dtn1 = 2445, 'dtn1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bues_service IANA assigned this well-formed service name as a
    #: replacement for "bues_service".
    #: - [TCP] bues_service
    #: - [UDP] bues_service IANA assigned this well-formed service name as a
    #: replacement for "bues_service".
    #: - [UDP] bues_service
    bues_service = 2446, 'bues-service', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView NNM daemon
    #: - [UDP] OpenView NNM daemon
    ovwdb = 2447, 'ovwdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hpppsvr
    #: - [UDP] hpppsvr
    hpppssvr = 2448, 'hpppssvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RATL
    #: - [UDP] RATL
    ratl = 2449, 'ratl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netadmin
    #: - [UDP] netadmin
    netadmin = 2450, 'netadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netchat
    #: - [UDP] netchat
    netchat = 2451, 'netchat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SnifferClient
    #: - [UDP] SnifferClient
    snifferclient = 2452, 'snifferclient', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] madge ltd
    #: - [UDP] madge ltd
    madge_ltd = 2453, 'madge-ltd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IndX-DDS
    #: - [UDP] IndX-DDS
    indx_dds = 2454, 'indx-dds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAGO-IO-SYSTEM
    #: - [UDP] WAGO-IO-SYSTEM
    wago_io_system = 2455, 'wago-io-system', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] altav-remmgt
    #: - [UDP] altav-remmgt
    altav_remmgt = 2456, 'altav-remmgt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rapido_IP
    #: - [UDP] Rapido_IP
    rapido_ip = 2457, 'rapido-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] griffin
    #: - [UDP] griffin
    griffin = 2458, 'griffin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Community
    #: - [UDP] Community
    xrpl = 2459, 'xrpl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ms-theater
    #: - [UDP] ms-theater
    ms_theater = 2460, 'ms-theater', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qadmifoper
    #: - [UDP] qadmifoper
    qadmifoper = 2461, 'qadmifoper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qadmifevent
    #: - [UDP] qadmifevent
    qadmifevent = 2462, 'qadmifevent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LSI RAID Management
    #: - [UDP] LSI RAID Management
    lsi_raid_mgmt = 2463, 'lsi-raid-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DirecPC SI
    #: - [UDP] DirecPC SI
    direcpc_si = 2464, 'direcpc-si', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Load Balance Management
    #: - [UDP] Load Balance Management
    lbm = 2465, 'lbm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Load Balance Forwarding
    #: - [UDP] Load Balance Forwarding
    lbf = 2466, 'lbf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] High Criteria
    #: - [UDP] High Criteria
    high_criteria = 2467, 'high-criteria', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qip_msgd
    #: - [UDP] qip_msgd
    qip_msgd = 2468, 'qip-msgd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MTI-TCS-COMM
    #: - [UDP] MTI-TCS-COMM
    mti_tcs_comm = 2469, 'mti-tcs-comm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] taskman port
    #: - [UDP] taskman port
    taskman_port = 2470, 'taskman-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SeaODBC
    #: - [UDP] SeaODBC
    seaodbc = 2471, 'seaodbc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] C3
    #: - [UDP] C3
    c3 = 2472, 'c3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Aker-cdp
    #: - [UDP] Aker-cdp
    aker_cdp = 2473, 'aker-cdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vital Analysis
    #: - [UDP] Vital Analysis
    vitalanalysis = 2474, 'vitalanalysis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACE Server
    #: - [UDP] ACE Server
    ace_server = 2475, 'ace-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACE Server Propagation
    #: - [UDP] ACE Server Propagation
    ace_svr_prop = 2476, 'ace-svr-prop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SecurSight Certificate Valifation Service
    #: - [UDP] SecurSight Certificate Valifation Service
    ssm_cvs = 2477, 'ssm-cvs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SecurSight Authentication Server (SSL)
    #: - [UDP] SecurSight Authentication Server (SSL)
    ssm_cssps = 2478, 'ssm-cssps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SecurSight Event Logging Server (SSL)
    #: - [UDP] SecurSight Event Logging Server (SSL)
    ssm_els = 2479, 'ssm-els', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Informatica PowerExchange Listener
    #: - [UDP] Informatica PowerExchange Listener
    powerexchange = 2480, 'powerexchange', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle GIOP
    #: - [UDP] Oracle GIOP
    giop = 2481, 'giop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle GIOP SSL
    #: - [UDP] Oracle GIOP SSL
    giop_ssl = 2482, 'giop-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle TTC
    #: - [UDP] Oracle TTC
    ttc = 2483, 'ttc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle TTC SSL
    #: - [UDP] Oracle TTC SSL
    ttc_ssl = 2484, 'ttc-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net Objects1
    #: - [UDP] Net Objects1
    netobjects1 = 2485, 'netobjects1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net Objects2
    #: - [UDP] Net Objects2
    netobjects2 = 2486, 'netobjects2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Policy Notice Service
    #: - [UDP] Policy Notice Service
    pns = 2487, 'pns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Moy Corporation
    #: - [UDP] Moy Corporation
    moy_corp = 2488, 'moy-corp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TSILB
    #: - [UDP] TSILB
    tsilb = 2489, 'tsilb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qip_qdhcp
    #: - [UDP] qip_qdhcp
    qip_qdhcp = 2490, 'qip-qdhcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Conclave CPP
    #: - [UDP] Conclave CPP
    conclave_cpp = 2491, 'conclave-cpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GROOVE
    #: - [UDP] GROOVE
    groove = 2492, 'groove', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talarian MQS
    #: - [UDP] Talarian MQS
    talarian_mqs = 2493, 'talarian-mqs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC AR
    #: - [UDP] BMC AR
    bmc_ar = 2494, 'bmc-ar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fast Remote Services
    #: - [UDP] Fast Remote Services
    fast_rem_serv = 2495, 'fast-rem-serv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DIRGIS
    #: - [UDP] DIRGIS
    dirgis = 2496, 'dirgis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quad DB
    #: - [UDP] Quad DB
    quaddb = 2497, 'quaddb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ODN-CasTraq
    #: - [UDP] ODN-CasTraq
    odn_castraq = 2498, 'odn-castraq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UniControl
    #: - [UDP] UniControl
    unicontrol_2437 = 2437, 'unicontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UniControl
    #: - [UDP] UniControl
    unicontrol_2499 = 2499, 'unicontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Resource Tracking system server
    #: - [UDP] Resource Tracking system server
    rtsserv = 2500, 'rtsserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Resource Tracking system client
    #: - [UDP] Resource Tracking system client
    rtsclient = 2501, 'rtsclient', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kentrox Protocol
    #: - [UDP] Kentrox Protocol
    kentrox_prot = 2502, 'kentrox-prot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NMS-DPNSS
    #: - [UDP] NMS-DPNSS
    nms_dpnss = 2503, 'nms-dpnss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WLBS
    #: - [UDP] WLBS
    wlbs = 2504, 'wlbs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PowerPlay Control
    #: - [UDP] PowerPlay Control
    ppcontrol = 2505, 'ppcontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] jbroker
    #: - [UDP] jbroker
    jbroker = 2506, 'jbroker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] spock
    #: - [UDP] spock
    spock = 2507, 'spock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JDataStore
    #: - [UDP] JDataStore
    jdatastore = 2508, 'jdatastore', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fjmpss
    #: - [UDP] fjmpss
    fjmpss = 2509, 'fjmpss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fjappmgrbulk
    #: - [UDP] fjappmgrbulk
    fjappmgrbulk = 2510, 'fjappmgrbulk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Metastorm
    #: - [UDP] Metastorm
    metastorm = 2511, 'metastorm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Citrix IMA
    #: - [UDP] Citrix IMA
    citrixima = 2512, 'citrixima', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Citrix ADMIN
    #: - [UDP] Citrix ADMIN
    citrixadmin = 2513, 'citrixadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Facsys NTP
    #: - [UDP] Facsys NTP
    facsys_ntp = 2514, 'facsys-ntp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Facsys Router
    #: - [UDP] Facsys Router
    facsys_router = 2515, 'facsys-router', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Main Control
    #: - [UDP] Main Control
    maincontrol = 2516, 'maincontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] H.323 Annex E Call Control Signalling Transport
    #: - [UDP] H.323 Annex E Call Control Signalling Transport
    call_sig_trans = 2517, 'call-sig-trans', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Willy
    #: - [UDP] Willy
    willy = 2518, 'willy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] globmsgsvc
    #: - [UDP] globmsgsvc
    globmsgsvc = 2519, 'globmsgsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pervasive Listener
    #: - [UDP] Pervasive Listener
    pvsw = 2520, 'pvsw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Adaptec Manager
    #: - [UDP] Adaptec Manager
    adaptecmgr = 2521, 'adaptecmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WinDb
    #: - [UDP] WinDb
    windb = 2522, 'windb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Qke LLC V.3
    #: - [UDP] Qke LLC V.3
    qke_llc_v3 = 2523, 'qke-llc-v3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Optiwave License Management
    #: - [UDP] Optiwave License Management
    optiwave_lm = 2524, 'optiwave-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS V-Worlds
    #: - [UDP] MS V-Worlds
    ms_v_worlds = 2525, 'ms-v-worlds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMA License Manager
    #: - [UDP] EMA License Manager
    ema_sent_lm = 2526, 'ema-sent-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IQ Server
    #: - [UDP] IQ Server
    iqserver = 2527, 'iqserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NCR CCL IANA assigned this well-formed service name as a replacement
    #: for "ncr_ccl".
    #: - [TCP] NCR CCL
    #: - [UDP] NCR CCL IANA assigned this well-formed service name as a replacement
    #: for "ncr_ccl".
    #: - [UDP] NCR CCL
    ncr_ccl = 2528, 'ncr-ccl', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] UTS FTP
    #: - [UDP] UTS FTP
    utsftp = 2529, 'utsftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VR Commerce
    #: - [UDP] VR Commerce
    vrcommerce = 2530, 'vrcommerce', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITO-E GUI
    #: - [UDP] ITO-E GUI
    ito_e_gui = 2531, 'ito-e-gui', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OVTOPMD
    #: - [UDP] OVTOPMD
    ovtopmd = 2532, 'ovtopmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SnifferServer
    #: - [UDP] SnifferServer
    snifferserver = 2533, 'snifferserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Combox Web Access
    #: - [UDP] Combox Web Access
    combox_web_acc = 2534, 'combox-web-acc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MADCAP
    #: - [UDP] MADCAP
    madcap = 2535, 'madcap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] btpp2audctr1
    #: - [UDP] btpp2audctr1
    btpp2audctr1 = 2536, 'btpp2audctr1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Upgrade Protocol
    #: - [UDP] Upgrade Protocol
    upgrade = 2537, 'upgrade', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vnwk-prapi
    #: - [UDP] vnwk-prapi
    vnwk_prapi = 2538, 'vnwk-prapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSI Admin
    #: - [UDP] VSI Admin
    vsiadmin = 2539, 'vsiadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LonWorks2
    #: - [UDP] LonWorks2
    lonworks2 = 2541, 'lonworks2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] uDraw(Graph)
    #: - [UDP] uDraw(Graph)
    udrawgraph = 2542, 'udrawgraph', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] REFTEK
    #: - [UDP] REFTEK
    reftek = 2543, 'reftek', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Management Daemon Refresh
    #: - [UDP] Management Daemon Refresh
    novell_zen = 2544, 'novell-zen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sis-emt
    #: - [UDP] sis-emt
    sis_emt = 2545, 'sis-emt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vytalvaultbrtp
    #: - [UDP] vytalvaultbrtp
    vytalvaultbrtp = 2546, 'vytalvaultbrtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vytalvaultvsmp
    #: - [UDP] vytalvaultvsmp
    vytalvaultvsmp = 2547, 'vytalvaultvsmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vytalvaultpipe
    #: - [UDP] vytalvaultpipe
    vytalvaultpipe = 2548, 'vytalvaultpipe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPASS
    #: - [UDP] IPASS
    ipass = 2549, 'ipass', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ADS
    #: - [UDP] ADS
    ads = 2550, 'ads', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISG UDA Server
    #: - [UDP] ISG UDA Server
    isg_uda_server = 2551, 'isg-uda-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Call Logging
    #: - [UDP] Call Logging
    call_logging = 2552, 'call-logging', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] efidiningport
    #: - [UDP] efidiningport
    efidiningport = 2553, 'efidiningport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VCnet-Link v10
    #: - [UDP] VCnet-Link v10
    vcnet_link_v10 = 2554, 'vcnet-link-v10', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compaq WCP
    #: - [UDP] Compaq WCP
    compaq_wcp = 2555, 'compaq-wcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nicetec-nmsvc
    #: - [UDP] nicetec-nmsvc
    nicetec_nmsvc = 2556, 'nicetec-nmsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nicetec-mgmt
    #: - [UDP] nicetec-mgmt
    nicetec_mgmt = 2557, 'nicetec-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCLE Multi Media
    #: - [UDP] PCLE Multi Media
    pclemultimedia = 2558, 'pclemultimedia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LSTP
    #: - [UDP] LSTP
    lstp = 2559, 'lstp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] labrat
    #: - [UDP] labrat
    labrat = 2560, 'labrat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MosaixCC
    #: - [UDP] MosaixCC
    mosaixcc = 2561, 'mosaixcc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Delibo
    #: - [UDP] Delibo
    delibo = 2562, 'delibo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CTI Redwood
    #: - [UDP] CTI Redwood
    cti_redwood = 2563, 'cti-redwood', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP 3000 NS/VT block mode telnet
    #: - [UDP] HP 3000 NS/VT block mode telnet
    hp_3000_telnet = 2564, 'hp-3000-telnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Coordinator Server
    #: - [UDP] Coordinator Server
    coord_svr = 2565, 'coord-svr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pcs-pcw
    #: - [UDP] pcs-pcw
    pcs_pcw = 2566, 'pcs-pcw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cisco Line Protocol
    #: - [UDP] Cisco Line Protocol
    clp = 2567, 'clp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPAM TRAP
    #: - [UDP] SPAM TRAP
    spamtrap = 2568, 'spamtrap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sonus Call Signal
    #: - [UDP] Sonus Call Signal
    sonuscallsig = 2569, 'sonuscallsig', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HS Port
    #: - [UDP] HS Port
    hs_port = 2570, 'hs-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CECSVC
    #: - [UDP] CECSVC
    cecsvc = 2571, 'cecsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBP
    #: - [UDP] IBP
    ibp = 2572, 'ibp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trust Establish
    #: - [UDP] Trust Establish
    trustestablish = 2573, 'trustestablish', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Blockade BPSP
    #: - [UDP] Blockade BPSP
    blockade_bpsp = 2574, 'blockade-bpsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HL7
    #: - [UDP] HL7
    hl7 = 2575, 'hl7', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TCL Pro Debugger
    #: - [UDP] TCL Pro Debugger
    tclprodebugger = 2576, 'tclprodebugger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Scriptics Lsrvr
    #: - [UDP] Scriptics Lsrvr
    scipticslsrvr = 2577, 'scipticslsrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RVS ISDN DCP
    #: - [UDP] RVS ISDN DCP
    rvs_isdn_dcp = 2578, 'rvs-isdn-dcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mpfoncl
    #: - [UDP] mpfoncl
    mpfoncl = 2579, 'mpfoncl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tributary
    #: - [UDP] Tributary
    tributary = 2580, 'tributary', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARGIS TE
    #: - [UDP] ARGIS TE
    argis_te = 2581, 'argis-te', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARGIS DS
    #: - [UDP] ARGIS DS
    argis_ds = 2582, 'argis-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cyaserv
    #: - [UDP] cyaserv
    cyaserv = 2584, 'cyaserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETX Server
    #: - [UDP] NETX Server
    netx_server = 2585, 'netx-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETX Agent
    #: - [UDP] NETX Agent
    netx_agent = 2586, 'netx-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MASC
    #: - [UDP] MASC
    masc = 2587, 'masc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Privilege
    #: - [UDP] Privilege
    privilege = 2588, 'privilege', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] quartus tcl
    #: - [UDP] quartus tcl
    quartus_tcl = 2589, 'quartus-tcl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] idotdist
    #: - [UDP] idotdist
    idotdist = 2590, 'idotdist', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Maytag Shuffle
    #: - [UDP] Maytag Shuffle
    maytagshuffle = 2591, 'maytagshuffle', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netrek
    #: - [UDP] netrek
    netrek = 2592, 'netrek', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MNS Mail Notice Service
    #: - [UDP] MNS Mail Notice Service
    mns_mail = 2593, 'mns-mail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Base Server
    #: - [UDP] Data Base Server
    dts = 2594, 'dts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] World Fusion 1
    #: - [UDP] World Fusion 1
    worldfusion1 = 2595, 'worldfusion1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] World Fusion 2
    #: - [UDP] World Fusion 2
    worldfusion2 = 2596, 'worldfusion2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Homestead Glory
    #: - [UDP] Homestead Glory
    homesteadglory = 2597, 'homesteadglory', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Citrix MA Client
    #: - [UDP] Citrix MA Client
    citriximaclient = 2598, 'citriximaclient', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Snap Discovery
    #: - [UDP] Snap Discovery
    snapd = 2599, 'snapd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HPSTGMGR
    #: - [UDP] HPSTGMGR
    hpstgmgr = 2600, 'hpstgmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] discp client
    #: - [UDP] discp client
    discp_client = 2601, 'discp-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] discp server
    #: - [UDP] discp server
    discp_server = 2602, 'discp-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Meter
    #: - [UDP] Service Meter
    servicemeter = 2603, 'servicemeter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NSC CCS
    #: - [UDP] NSC CCS
    nsc_ccs = 2604, 'nsc-ccs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NSC POSA
    #: - [UDP] NSC POSA
    nsc_posa = 2605, 'nsc-posa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dell Netmon
    #: - [UDP] Dell Netmon
    netmon = 2606, 'netmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dell Connection
    #: - [UDP] Dell Connection
    connection = 2607, 'connection', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Wag Service
    #: - [UDP] Wag Service
    wag_service = 2608, 'wag-service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] System Monitor
    #: - [UDP] System Monitor
    system_monitor = 2609, 'system-monitor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VersaTek
    #: - [UDP] VersaTek
    versa_tek = 2610, 'versa-tek', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LIONHEAD
    #: - [UDP] LIONHEAD
    lionhead = 2611, 'lionhead', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Qpasa Agent
    #: - [UDP] Qpasa Agent
    qpasa_agent = 2612, 'qpasa-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMNTUBootstrap
    #: - [UDP] SMNTUBootstrap
    smntubootstrap = 2613, 'smntubootstrap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Never Offline
    #: - [UDP] Never Offline
    neveroffline = 2614, 'neveroffline', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] firepower
    #: - [UDP] firepower
    firepower = 2615, 'firepower', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] appswitch-emp
    #: - [UDP] appswitch-emp
    appswitch_emp = 2616, 'appswitch-emp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Clinical Context Managers
    #: - [UDP] Clinical Context Managers
    cmadmin = 2617, 'cmadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Priority E-Com
    #: - [UDP] Priority E-Com
    priority_e_com = 2618, 'priority-e-com', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bruce
    #: - [UDP] bruce
    bruce = 2619, 'bruce', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LPSRecommender
    #: - [UDP] LPSRecommender
    lpsrecommender = 2620, 'lpsrecommender', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Miles Apart Jukebox Server
    #: - [UDP] Miles Apart Jukebox Server
    miles_apart = 2621, 'miles-apart', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MetricaDBC
    #: - [UDP] MetricaDBC
    metricadbc = 2622, 'metricadbc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LMDP
    #: - [UDP] LMDP
    lmdp = 2623, 'lmdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Aria
    #: - [UDP] Aria
    aria = 2624, 'aria', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Blwnkl Port
    #: - [UDP] Blwnkl Port
    blwnkl_port = 2625, 'blwnkl-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gbjd816
    #: - [UDP] gbjd816
    gbjd816 = 2626, 'gbjd816', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Moshe Beeri
    #: - [UDP] Moshe Beeri
    moshebeeri = 2627, 'moshebeeri', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DICT
    #: - [UDP] DICT
    dict = 2628, 'dict', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sitara Server
    #: - [UDP] Sitara Server
    sitaraserver = 2629, 'sitaraserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sitara Management
    #: - [UDP] Sitara Management
    sitaramgmt = 2630, 'sitaramgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sitara Dir
    #: - [UDP] Sitara Dir
    sitaradir = 2631, 'sitaradir', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IRdg Post
    #: - [UDP] IRdg Post
    irdg_post = 2632, 'irdg-post', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] InterIntelli
    #: - [UDP] InterIntelli
    interintelli = 2633, 'interintelli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PK Electronics
    #: - [UDP] PK Electronics
    pk_electronics = 2634, 'pk-electronics', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Back Burner
    #: - [UDP] Back Burner
    backburner = 2635, 'backburner', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Solve
    #: - [UDP] Solve
    solve = 2636, 'solve', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Import Document Service
    #: - [UDP] Import Document Service
    imdocsvc = 2637, 'imdocsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sybase Anywhere
    #: - [UDP] Sybase Anywhere
    sybaseanywhere = 2638, 'sybaseanywhere', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMInet
    #: - [UDP] AMInet
    aminet = 2639, 'aminet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alcorn McBride Inc protocol used for device control
    #: - [UDP] Alcorn McBride Inc protocol used for device control
    ami_control = 2640, 'ami-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HDL Server
    #: - [UDP] HDL Server
    hdl_srv = 2641, 'hdl-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tragic
    #: - [UDP] Tragic
    tragic = 2642, 'tragic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GTE-SAMP
    #: - [UDP] GTE-SAMP
    gte_samp = 2643, 'gte-samp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Travsoft IPX Tunnel
    #: - [UDP] Travsoft IPX Tunnel
    travsoft_ipx_t = 2644, 'travsoft-ipx-t', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novell IPX CMD
    #: - [UDP] Novell IPX CMD
    novell_ipx_cmd = 2645, 'novell-ipx-cmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AND License Manager
    #: - [UDP] AND License Manager
    and_lm = 2646, 'and-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SyncServer
    #: - [UDP] SyncServer
    syncserver = 2647, 'syncserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Upsnotifyprot
    #: - [UDP] Upsnotifyprot
    upsnotifyprot = 2648, 'upsnotifyprot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VPSIPPORT
    #: - [UDP] VPSIPPORT
    vpsipport = 2649, 'vpsipport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eristwoguns
    #: - [UDP] eristwoguns
    eristwoguns = 2650, 'eristwoguns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EBInSite
    #: - [UDP] EBInSite
    ebinsite = 2651, 'ebinsite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] InterPathPanel
    #: - [UDP] InterPathPanel
    interpathpanel = 2652, 'interpathpanel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sonus
    #: - [UDP] Sonus
    sonus = 2653, 'sonus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Corel VNC Admin IANA assigned this well-formed service name as a
    #: replacement for "corel_vncadmin".
    #: - [TCP] Corel VNC Admin
    #: - [UDP] Corel VNC Admin IANA assigned this well-formed service name as a
    #: replacement for "corel_vncadmin".
    #: - [UDP] Corel VNC Admin
    corel_vncadmin = 2654, 'corel-vncadmin', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] UNIX Nt Glue
    #: - [UDP] UNIX Nt Glue
    unglue = 2655, 'unglue', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kana
    #: - [UDP] Kana
    kana = 2656, 'kana', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNS Dispatcher
    #: - [UDP] SNS Dispatcher
    sns_dispatcher = 2657, 'sns-dispatcher', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNS Admin
    #: - [UDP] SNS Admin
    sns_admin = 2658, 'sns-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNS Query
    #: - [UDP] SNS Query
    sns_query = 2659, 'sns-query', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GC Monitor
    #: - [UDP] GC Monitor
    gcmonitor = 2660, 'gcmonitor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OLHOST
    #: - [UDP] OLHOST
    olhost = 2661, 'olhost', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BinTec-CAPI
    #: - [UDP] BinTec-CAPI
    bintec_capi = 2662, 'bintec-capi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BinTec-TAPI
    #: - [UDP] BinTec-TAPI
    bintec_tapi = 2663, 'bintec-tapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Patrol for MQ GM
    #: - [UDP] Patrol for MQ GM
    patrol_mq_gm = 2664, 'patrol-mq-gm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Patrol for MQ NM
    #: - [UDP] Patrol for MQ NM
    patrol_mq_nm = 2665, 'patrol-mq-nm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] extensis
    #: - [UDP] extensis
    extensis = 2666, 'extensis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alarm Clock Server
    #: - [UDP] Alarm Clock Server
    alarm_clock_s = 2667, 'alarm-clock-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alarm Clock Client
    #: - [UDP] Alarm Clock Client
    alarm_clock_c = 2668, 'alarm-clock-c', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TOAD
    #: - [UDP] TOAD
    toad = 2669, 'toad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TVE Announce
    #: - [UDP] TVE Announce
    tve_announce = 2670, 'tve-announce', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] newlixreg
    #: - [UDP] newlixreg
    newlixreg = 2671, 'newlixreg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nhserver
    #: - [UDP] nhserver
    nhserver = 2672, 'nhserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] First Call 42
    #: - [UDP] First Call 42
    firstcall42 = 2673, 'firstcall42', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ewnn
    #: - [UDP] ewnn
    ewnn = 2674, 'ewnn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TTC ETAP
    #: - [UDP] TTC ETAP
    ttc_etap = 2675, 'ttc-etap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SIMSLink
    #: - [UDP] SIMSLink
    simslink = 2676, 'simslink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gadget Gate 1 Way
    #: - [UDP] Gadget Gate 1 Way
    gadgetgate1way = 2677, 'gadgetgate1way', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gadget Gate 2 Way
    #: - [UDP] Gadget Gate 2 Way
    gadgetgate2way = 2678, 'gadgetgate2way', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sync Server SSL
    #: - [UDP] Sync Server SSL
    syncserverssl = 2679, 'syncserverssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pxc-sapxom
    #: - [UDP] pxc-sapxom
    pxc_sapxom = 2680, 'pxc-sapxom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mpnjsomb
    #: - [UDP] mpnjsomb
    mpnjsomb = 2681, 'mpnjsomb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Removed
    removed_659 = 659, 'removed', TransportProtocol.get('undefined')

    #: [N/A] Removed
    removed_2682 = 2682, 'removed', TransportProtocol.get('undefined')

    #: - [TCP] NCDLoadBalance
    #: - [UDP] NCDLoadBalance
    ncdloadbalance = 2683, 'ncdloadbalance', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mpnjsosv
    #: - [UDP] mpnjsosv
    mpnjsosv = 2684, 'mpnjsosv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mpnjsocl
    #: - [UDP] mpnjsocl
    mpnjsocl = 2685, 'mpnjsocl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mpnjsomg
    #: - [UDP] mpnjsomg
    mpnjsomg = 2686, 'mpnjsomg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pq-lic-mgmt
    #: - [UDP] pq-lic-mgmt
    pq_lic_mgmt = 2687, 'pq-lic-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] md-cf-http
    #: - [UDP] md-cf-http
    md_cg_http = 2688, 'md-cg-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FastLynx
    #: - [UDP] FastLynx
    fastlynx = 2689, 'fastlynx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP NNM Embedded Database
    #: - [UDP] HP NNM Embedded Database
    hp_nnm_data = 2690, 'hp-nnm-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITInternet ISM Server
    #: - [UDP] ITInternet ISM Server
    itinternet = 2691, 'itinternet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Admins LMS
    #: - [UDP] Admins LMS
    admins_lms = 2692, 'admins-lms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unassigned
    #: - [UDP] Unassigned
    unassigned_2693 = 2693, 'unassigned', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pwrsevent
    #: - [UDP] pwrsevent
    pwrsevent = 2694, 'pwrsevent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSPREAD
    #: - [UDP] VSPREAD
    vspread = 2695, 'vspread', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unify Admin
    #: - [UDP] Unify Admin
    unifyadmin = 2696, 'unifyadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oce SNMP Trap Port
    #: - [UDP] Oce SNMP Trap Port
    oce_snmp_trap = 2697, 'oce-snmp-trap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCK-IVPIP
    #: - [UDP] MCK-IVPIP
    mck_ivpip = 2698, 'mck-ivpip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Csoft Plus Client
    #: - [UDP] Csoft Plus Client
    csoft_plusclnt = 2699, 'csoft-plusclnt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tqdata
    #: - [UDP] tqdata
    tqdata = 2700, 'tqdata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMS RCINFO
    #: - [UDP] SMS RCINFO
    sms_rcinfo = 2701, 'sms-rcinfo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMS XFER
    #: - [UDP] SMS XFER
    sms_xfer = 2702, 'sms-xfer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMS CHAT
    #: - [UDP] SMS CHAT
    sms_chat = 2703, 'sms-chat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMS REMCTRL
    #: - [UDP] SMS REMCTRL
    sms_remctrl = 2704, 'sms-remctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SDS Admin
    #: - [UDP] SDS Admin
    sds_admin = 2705, 'sds-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NCD Mirroring
    #: - [UDP] NCD Mirroring
    ncdmirroring = 2706, 'ncdmirroring', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMCSYMAPIPORT
    #: - [UDP] EMCSYMAPIPORT
    emcsymapiport = 2707, 'emcsymapiport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Banyan-Net
    #: - [UDP] Banyan-Net
    banyan_net = 2708, 'banyan-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Supermon
    #: - [UDP] Supermon
    supermon = 2709, 'supermon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSO Service
    #: - [UDP] SSO Service
    sso_service = 2710, 'sso-service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSO Control
    #: - [UDP] SSO Control
    sso_control = 2711, 'sso-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Axapta Object Communication Protocol
    #: - [UDP] Axapta Object Communication Protocol
    aocp = 2712, 'aocp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Raven Trinity Broker Service
    #: - [UDP] Raven Trinity Broker Service
    raventbs = 2713, 'raventbs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Raven Trinity Data Mover
    #: - [UDP] Raven Trinity Data Mover
    raventdm = 2714, 'raventdm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HPSTGMGR2
    #: - [UDP] HPSTGMGR2
    hpstgmgr2 = 2715, 'hpstgmgr2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inova IP Disco
    #: - [UDP] Inova IP Disco
    inova_ip_disco = 2716, 'inova-ip-disco', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PN REQUESTER
    #: - [UDP] PN REQUESTER
    pn_requester = 2717, 'pn-requester', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PN REQUESTER 2
    #: - [UDP] PN REQUESTER 2
    pn_requester2 = 2718, 'pn-requester2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Scan & Change
    #: - [UDP] Scan & Change
    scan_change = 2719, 'scan-change', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] wkars
    #: - [UDP] wkars
    wkars = 2720, 'wkars', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Smart Diagnose
    #: - [UDP] Smart Diagnose
    smart_diagnose = 2721, 'smart-diagnose', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Proactive Server
    #: - [UDP] Proactive Server
    proactivesrvr = 2722, 'proactivesrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WatchDog NT Protocol
    #: - [UDP] WatchDog NT Protocol
    watchdog_nt = 2723, 'watchdog-nt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qotps
    #: - [UDP] qotps
    qotps = 2724, 'qotps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSOLAP PTP2
    #: - [UDP] MSOLAP PTP2
    msolap_ptp2 = 2725, 'msolap-ptp2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TAMS
    #: - [UDP] TAMS
    tams = 2726, 'tams', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Media Gateway Control Protocol Call Agent
    #: - [UDP] Media Gateway Control Protocol Call Agent
    mgcp_callagent = 2727, 'mgcp-callagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SQDR
    #: - [UDP] SQDR
    sqdr = 2728, 'sqdr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TCIM Control
    #: - [UDP] TCIM Control
    tcim_control = 2729, 'tcim-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NEC RaidPlus
    #: - [UDP] NEC RaidPlus
    nec_raidplus = 2730, 'nec-raidplus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fyre Messanger
    #: - [UDP] Fyre Messanger
    fyre_messanger = 2731, 'fyre-messanger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] G5M
    #: - [UDP] G5M
    g5m = 2732, 'g5m', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Signet CTF
    #: - [UDP] Signet CTF
    signet_ctf = 2733, 'signet-ctf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CCS Software
    #: - [UDP] CCS Software
    ccs_software = 2734, 'ccs-software', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetIQ Monitor Console
    #: - [UDP] NetIQ Monitor Console
    netiq_mc = 2735, 'netiq-mc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RADWIZ NMS SRV
    #: - [UDP] RADWIZ NMS SRV
    radwiz_nms_srv = 2736, 'radwiz-nms-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SRP Feedback
    #: - [UDP] SRP Feedback
    srp_feedback = 2737, 'srp-feedback', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NDL TCP-OSI Gateway
    #: - [UDP] NDL TCP-OSI Gateway
    ndl_tcp_ois_gw = 2738, 'ndl-tcp-ois-gw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TN Timing
    #: - [UDP] TN Timing
    tn_timing = 2739, 'tn-timing', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alarm
    #: - [UDP] Alarm
    alarm = 2740, 'alarm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TSB
    #: - [UDP] TSB
    tsb = 2741, 'tsb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TSB2
    #: - [UDP] TSB2
    tsb2 = 2742, 'tsb2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] murx
    #: - [UDP] murx
    murx = 2743, 'murx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] honyaku
    #: - [UDP] honyaku
    honyaku = 2744, 'honyaku', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] URBISNET
    #: - [UDP] URBISNET
    urbisnet = 2745, 'urbisnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CPUDPENCAP
    #: - [UDP] CPUDPENCAP
    cpudpencap = 2746, 'cpudpencap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    fjippol_swrly = 2747, 'fjippol-swrly', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    fjippol_polsvr = 2748, 'fjippol-polsvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    fjippol_cnsl = 2749, 'fjippol-cnsl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    fjippol_port1 = 2750, 'fjippol-port1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    fjippol_port2 = 2751, 'fjippol-port2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSISYS ACCESS
    #: - [UDP] RSISYS ACCESS
    rsisysaccess = 2752, 'rsisysaccess', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] de-spot
    #: - [UDP] de-spot
    de_spot = 2753, 'de-spot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APOLLO CC
    #: - [UDP] APOLLO CC
    apollo_cc = 2754, 'apollo-cc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Express Pay
    #: - [UDP] Express Pay
    expresspay = 2755, 'expresspay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] simplement-tie
    #: - [UDP] simplement-tie
    simplement_tie = 2756, 'simplement-tie', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CNRP
    #: - [UDP] CNRP
    cnrp = 2757, 'cnrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APOLLO Status
    #: - [UDP] APOLLO Status
    apollo_status = 2758, 'apollo-status', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APOLLO GMS
    #: - [UDP] APOLLO GMS
    apollo_gms = 2759, 'apollo-gms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Saba MS
    #: - [UDP] Saba MS
    sabams = 2760, 'sabams', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DICOM ISCL
    #: - [UDP] DICOM ISCL
    dicom_iscl = 2761, 'dicom-iscl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DICOM TLS
    #: - [UDP] DICOM TLS
    dicom_tls = 2762, 'dicom-tls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Desktop DNA
    #: - [UDP] Desktop DNA
    desktop_dna = 2763, 'desktop-dna', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Insurance
    #: - [UDP] Data Insurance
    data_insurance = 2764, 'data-insurance', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qip-audup
    #: - [UDP] qip-audup
    qip_audup = 2765, 'qip-audup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compaq SCP
    #: - [UDP] Compaq SCP
    compaq_scp = 2766, 'compaq-scp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UADTC
    #: - [UDP] UADTC
    uadtc = 2767, 'uadtc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UACS
    #: - [UDP] UACS
    uacs = 2768, 'uacs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eXcE
    #: - [UDP] eXcE
    exce = 2769, 'exce', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Veronica
    #: - [UDP] Veronica
    veronica = 2770, 'veronica', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vergence CM
    #: - [UDP] Vergence CM
    vergencecm = 2771, 'vergencecm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] auris
    #: - [UDP] auris
    auris = 2772, 'auris', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RBackup Remote Backup
    #: - [UDP] RBackup Remote Backup
    rbakcup1 = 2773, 'rbakcup1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RBackup Remote Backup
    #: - [UDP] RBackup Remote Backup
    rbakcup2 = 2774, 'rbakcup2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMPP
    #: - [UDP] SMPP
    smpp = 2775, 'smpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ridgeway Systems & Software
    #: - [UDP] Ridgeway Systems & Software
    ridgeway1 = 2776, 'ridgeway1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ridgeway Systems & Software
    #: - [UDP] Ridgeway Systems & Software
    ridgeway2 = 2777, 'ridgeway2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gwen-Sonya
    #: - [UDP] Gwen-Sonya
    gwen_sonya = 2778, 'gwen-sonya', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LBC Sync
    #: - [UDP] LBC Sync
    lbc_sync = 2779, 'lbc-sync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LBC Control
    #: - [UDP] LBC Control
    lbc_control = 2780, 'lbc-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] whosells
    #: - [UDP] whosells
    whosells = 2781, 'whosells', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] everydayrc
    #: - [UDP] everydayrc
    everydayrc = 2782, 'everydayrc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AISES
    #: - [UDP] AISES
    aises = 2783, 'aises', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] world wide web - development
    #: - [UDP] world wide web - development
    www_dev = 2784, 'www-dev', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] aic-np
    #: - [UDP] aic-np
    aic_np = 2785, 'aic-np', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] aic-oncrpc - Destiny MCD database
    #: - [UDP] aic-oncrpc - Destiny MCD database
    aic_oncrpc = 2786, 'aic-oncrpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] piccolo - Cornerstone Software
    #: - [UDP] piccolo - Cornerstone Software
    piccolo = 2787, 'piccolo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetWare Loadable Module - Seagate Software
    #: - [UDP] NetWare Loadable Module - Seagate Software
    fryeserv = 2788, 'fryeserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Media Agent
    #: - [UDP] Media Agent
    media_agent = 2789, 'media-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PLG Proxy
    #: - [UDP] PLG Proxy
    plgproxy = 2790, 'plgproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MT Port Registrator
    #: - [UDP] MT Port Registrator
    mtport_regist = 2791, 'mtport-regist', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] f5-globalsite
    #: - [UDP] f5-globalsite
    f5_globalsite = 2792, 'f5-globalsite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] initlsmsad
    #: - [UDP] initlsmsad
    initlsmsad = 2793, 'initlsmsad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_2794 = 2794, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] LiveStats
    #: - [UDP] LiveStats
    livestats = 2795, 'livestats', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ac-tech
    #: - [UDP] ac-tech
    ac_tech = 2796, 'ac-tech', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] esp-encap
    #: - [UDP] esp-encap
    esp_encap = 2797, 'esp-encap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TMESIS-UPShot
    #: - [UDP] TMESIS-UPShot
    tmesis_upshot = 2798, 'tmesis-upshot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICON Discover
    #: - [UDP] ICON Discover
    icon_discover = 2799, 'icon-discover', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACC RAID
    #: - [UDP] ACC RAID
    acc_raid = 2800, 'acc-raid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IGCP
    #: - [UDP] IGCP
    igcp = 2801, 'igcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Veritas TCP1
    veritas_tcp1 = 2802, 'veritas-tcp1', TransportProtocol.get('tcp')

    #: [UDP] Veritas UDP1
    veritas_udp1 = 2802, 'veritas-udp1', TransportProtocol.get('udp')

    #: - [TCP] btprjctrl
    #: - [UDP] btprjctrl
    btprjctrl = 2803, 'btprjctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] March Networks Digital Video Recorders and Enterprise Service Manager
    #: products
    #: - [UDP] March Networks Digital Video Recorders and Enterprise Service Manager
    #: products
    dvr_esm = 2804, 'dvr-esm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WTA WSP-S
    #: - [UDP] WTA WSP-S
    wta_wsp_s = 2805, 'wta-wsp-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cspuni
    #: - [UDP] cspuni
    cspuni = 2806, 'cspuni', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cspmulti
    #: - [UDP] cspmulti
    cspmulti = 2807, 'cspmulti', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] J-LAN-P
    #: - [UDP] J-LAN-P
    j_lan_p = 2808, 'j-lan-p', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CORBA LOC
    #: - [UDP] CORBA LOC
    corbaloc = 2809, 'corbaloc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Active Net Steward
    #: - [UDP] Active Net Steward
    netsteward = 2810, 'netsteward', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GSI FTP
    #: - [UDP] GSI FTP
    gsiftp = 2811, 'gsiftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] atmtcp
    #: - [UDP] atmtcp
    atmtcp = 2812, 'atmtcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] llm-pass
    #: - [UDP] llm-pass
    llm_pass = 2813, 'llm-pass', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] llm-csv
    #: - [UDP] llm-csv
    llm_csv = 2814, 'llm-csv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LBC Measurement
    #: - [UDP] LBC Measurement
    lbc_measure = 2815, 'lbc-measure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LBC Watchdog
    #: - [UDP] LBC Watchdog
    lbc_watchdog = 2816, 'lbc-watchdog', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rmlnk
    #: - [UDP] rmlnk
    rmlnk = 2818, 'rmlnk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FC Fault Notification
    #: - [UDP] FC Fault Notification
    fc_faultnotify = 2819, 'fc-faultnotify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UniVision
    #: - [UDP] UniVision
    univision = 2820, 'univision', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS Authentication Service
    #: - [UDP] VERITAS Authentication Service
    vrts_at_port = 2821, 'vrts-at-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ka0wuc
    #: - [UDP] ka0wuc
    ka0wuc = 2822, 'ka0wuc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CQG Net/LAN
    #: - [UDP] CQG Net/LAN
    cqg_netlan = 2823, 'cqg-netlan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CQG Net/LAN 1
    #: - [UDP] CQG Net/Lan 1
    cqg_netlan_1 = 2824, 'cqg-netlan-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] (unassigned) Possibly assigned
    possibly_assigned = 2825, 'possibly_assigned', TransportProtocol.get('undefined')

    #: - [TCP] slc systemlog
    #: - [UDP] slc systemlog
    slc_systemlog = 2826, 'slc-systemlog', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] slc ctrlrloops
    #: - [UDP] slc ctrlrloops
    slc_ctrlrloops = 2827, 'slc-ctrlrloops', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITM License Manager
    #: - [UDP] ITM License Manager
    itm_lm = 2828, 'itm-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] silkp1
    #: - [UDP] silkp1
    silkp1 = 2829, 'silkp1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] silkp2
    #: - [UDP] silkp2
    silkp2 = 2830, 'silkp2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] silkp3
    #: - [UDP] silkp3
    silkp3 = 2831, 'silkp3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] silkp4
    #: - [UDP] silkp4
    silkp4 = 2832, 'silkp4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] glishd
    #: - [UDP] glishd
    glishd = 2833, 'glishd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EVTP
    #: - [UDP] EVTP
    evtp = 2834, 'evtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EVTP-DATA
    #: - [UDP] EVTP-DATA
    evtp_data = 2835, 'evtp-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] catalyst
    #: - [UDP] catalyst
    catalyst = 2836, 'catalyst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Repliweb
    #: - [UDP] Repliweb
    repliweb = 2837, 'repliweb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Starbot
    #: - [UDP] Starbot
    starbot = 2838, 'starbot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NMSig Port
    #: - [UDP] NMSig Port
    nmsigport_2817 = 2817, 'nmsigport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NMSigPort
    #: - [UDP] NMSigPort
    nmsigport_2839 = 2839, 'nmsigport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] l3-exprt
    #: - [UDP] l3-exprt
    l3_exprt = 2840, 'l3-exprt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] l3-ranger
    #: - [UDP] l3-ranger
    l3_ranger = 2841, 'l3-ranger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] l3-hawk
    #: - [UDP] l3-hawk
    l3_hawk = 2842, 'l3-hawk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PDnet
    #: - [UDP] PDnet
    pdnet = 2843, 'pdnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BPCP POLL
    #: - [UDP] BPCP POLL
    bpcp_poll = 2844, 'bpcp-poll', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BPCP TRAP
    #: - [UDP] BPCP TRAP
    bpcp_trap = 2845, 'bpcp-trap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AIMPP Hello
    #: - [UDP] AIMPP Hello
    aimpp_hello = 2846, 'aimpp-hello', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AIMPP Port Req
    #: - [UDP] AIMPP Port Req
    aimpp_port_req = 2847, 'aimpp-port-req', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMT-BLC-PORT
    #: - [UDP] AMT-BLC-PORT
    amt_blc_port = 2848, 'amt-blc-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FXP Communication
    #: - [UDP] FXP Communication
    fxp_286 = 286, 'fxp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FXP
    #: - [UDP] FXP
    fxp_2849 = 2849, 'fxp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MetaConsole
    #: - [UDP] MetaConsole
    metaconsole = 2850, 'metaconsole', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] webemshttp
    #: - [UDP] webemshttp
    webemshttp = 2851, 'webemshttp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bears-01
    #: - [UDP] bears-01
    bears_01 = 2852, 'bears-01', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISPipes
    #: - [UDP] ISPipes
    ispipes = 2853, 'ispipes', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] InfoMover
    #: - [UDP] InfoMover
    infomover = 2854, 'infomover', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] MSRP over TCP [:rfc:`4976`]
    msrp = 2855, 'msrp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_2855 = 2855, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] cesdinv
    #: - [UDP] cesdinv
    cesdinv = 2856, 'cesdinv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SimCtIP
    #: - [UDP] SimCtIP
    simctlp = 2857, 'simctlp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ECNP
    #: - [UDP] ECNP
    ecnp = 2858, 'ecnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Active Memory
    #: - [UDP] Active Memory
    activememory = 2859, 'activememory', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dialpad Voice 1
    #: - [UDP] Dialpad Voice 1
    dialpad_voice1 = 2860, 'dialpad-voice1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dialpad Voice 2
    #: - [UDP] Dialpad Voice 2
    dialpad_voice2 = 2861, 'dialpad-voice2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TTG Protocol
    #: - [UDP] TTG Protocol
    ttg_protocol = 2862, 'ttg-protocol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sonar Data
    #: - [UDP] Sonar Data
    sonardata = 2863, 'sonardata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] main 5001 cmd
    #: - [UDP] main 5001 cmd
    astronova_main = 2864, 'astronova-main', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pit-vpn
    #: - [UDP] pit-vpn
    pit_vpn = 2865, 'pit-vpn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iwlistener
    #: - [UDP] iwlistener
    iwlistener = 2866, 'iwlistener', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] esps-portal
    #: - [UDP] esps-portal
    esps_portal = 2867, 'esps-portal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Norman Proprietaqry Events Protocol
    #: - [UDP] Norman Proprietaqry Events Protocol
    npep_messaging = 2868, 'npep-messaging', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICSLAP
    #: - [UDP] ICSLAP
    icslap = 2869, 'icslap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] daishi
    #: - [UDP] daishi
    daishi = 2870, 'daishi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSI Select Play
    #: - [UDP] MSI Select Play
    msi_selectplay = 2871, 'msi-selectplay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RADIX
    #: - [UDP] RADIX
    radix = 2872, 'radix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PubSub Realtime Telemetry Protocol
    #: - [UDP] PubSub Realtime Telemetry Protocol
    psrt = 2873, 'psrt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DX Message Base Transport Protocol
    #: - [UDP] DX Message Base Transport Protocol
    dxmessagebase1 = 2874, 'dxmessagebase1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DX Message Base Transport Protocol
    #: - [UDP] DX Message Base Transport Protocol
    dxmessagebase2 = 2875, 'dxmessagebase2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPS Tunnel
    #: - [UDP] SPS Tunnel
    sps_tunnel = 2876, 'sps-tunnel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BLUELANCE
    #: - [UDP] BLUELANCE
    bluelance = 2877, 'bluelance', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AAP
    #: - [UDP] AAP
    aap = 2878, 'aap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ucentric-ds
    #: - [UDP] ucentric-ds
    ucentric_ds = 2879, 'ucentric-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Synapse Transport
    #: - [UDP] Synapse Transport
    synapse = 2880, 'synapse', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NDSP
    #: - [UDP] NDSP
    ndsp = 2881, 'ndsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NDTP
    #: - [UDP] NDTP
    ndtp = 2882, 'ndtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NDNP
    #: - [UDP] NDNP
    ndnp = 2883, 'ndnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Flash Msg
    #: - [UDP] Flash Msg
    flashmsg = 2884, 'flashmsg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TopFlow
    #: - [UDP] TopFlow
    topflow = 2885, 'topflow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RESPONSELOGIC
    #: - [UDP] RESPONSELOGIC
    responselogic = 2886, 'responselogic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] aironet
    #: - [UDP] aironet
    aironetddp = 2887, 'aironetddp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPCSDLOBBY
    #: - [UDP] SPCSDLOBBY
    spcsdlobby = 2888, 'spcsdlobby', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSOM
    #: - [UDP] RSOM
    rsom = 2889, 'rsom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSPCLMULTI
    #: - [UDP] CSPCLMULTI
    cspclmulti = 2890, 'cspclmulti', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CINEGRFX-ELMD License Manager
    #: - [UDP] CINEGRFX-ELMD License Manager
    cinegrfx_elmd = 2891, 'cinegrfx-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNIFFERDATA
    #: - [UDP] SNIFFERDATA
    snifferdata = 2892, 'snifferdata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSECONNECTOR
    #: - [UDP] VSECONNECTOR
    vseconnector = 2893, 'vseconnector', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ABACUS-REMOTE
    #: - [UDP] ABACUS-REMOTE
    abacus_remote = 2894, 'abacus-remote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NATUS LINK
    #: - [UDP] NATUS LINK
    natuslink = 2895, 'natuslink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ECOVISIONG6-1
    #: - [UDP] ECOVISIONG6-1
    ecovisiong6_1 = 2896, 'ecovisiong6-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Citrix RTMP
    #: - [UDP] Citrix RTMP
    citrix_rtmp = 2897, 'citrix-rtmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APPLIANCE-CFG
    #: - [UDP] APPLIANCE-CFG
    appliance_cfg = 2898, 'appliance-cfg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] POWERGEMPLUS
    #: - [UDP] POWERGEMPLUS
    powergemplus = 2899, 'powergemplus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QUICKSUITE
    #: - [UDP] QUICKSUITE
    quicksuite = 2900, 'quicksuite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ALLSTORCNS
    #: - [UDP] ALLSTORCNS
    allstorcns = 2901, 'allstorcns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NET ASPI
    #: - [UDP] NET ASPI
    netaspi = 2902, 'netaspi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SUITCASE
    #: - [UDP] SUITCASE
    suitcase = 2903, 'suitcase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] M2UA
    #: - [UDP] M2UA
    #: - [SCTP] M2UA
    m2ua = 2904, 'm2ua', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] M3UA [:rfc:`4666`]
    #: - [SCTP] M3UA [:rfc:`4666`]
    m3ua = 2905, 'm3ua', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [N/A] De-registered
    de_registered_585 = 585, 'de_registered', TransportProtocol.get('undefined')

    #: [UDP] De-registered
    de_registered_2905 = 2905, 'de_registered', TransportProtocol.get('udp')

    #: - [TCP] CALLER9
    #: - [UDP] CALLER9
    caller9 = 2906, 'caller9', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WEBMETHODS B2B
    #: - [UDP] WEBMETHODS B2B
    webmethods_b2b = 2907, 'webmethods-b2b', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mao
    #: - [UDP] mao
    mao = 2908, 'mao', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Funk Dialout
    #: - [UDP] Funk Dialout
    funk_dialout = 2909, 'funk-dialout', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TDAccess
    #: - [UDP] TDAccess
    tdaccess = 2910, 'tdaccess', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Blockade
    #: - [UDP] Blockade
    blockade = 2911, 'blockade', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Epicon
    #: - [UDP] Epicon
    epicon = 2912, 'epicon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Booster Ware
    #: - [UDP] Booster Ware
    boosterware = 2913, 'boosterware', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Game Lobby
    #: - [UDP] Game Lobby
    gamelobby = 2914, 'gamelobby', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TK Socket
    #: - [UDP] TK Socket
    tksocket = 2915, 'tksocket', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Elvin Server IANA assigned this well-formed service name as a
    #: replacement for "elvin_server".
    #: - [TCP] Elvin Server
    #: - [UDP] Elvin Server IANA assigned this well-formed service name as a
    #: replacement for "elvin_server".
    #: - [UDP] Elvin Server
    elvin_server = 2916, 'elvin-server', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Elvin Client IANA assigned this well-formed service name as a
    #: replacement for "elvin_client".
    #: - [TCP] Elvin Client
    #: - [UDP] Elvin Client IANA assigned this well-formed service name as a
    #: replacement for "elvin_client".
    #: - [UDP] Elvin Client
    elvin_client = 2917, 'elvin-client', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Kasten Chase Pad
    #: - [UDP] Kasten Chase Pad
    kastenchasepad = 2918, 'kastenchasepad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] roboER
    #: - [UDP] roboER
    roboer = 2919, 'roboer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] roboEDA
    #: - [UDP] roboEDA
    roboeda = 2920, 'roboeda', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CESD Contents Delivery Management
    #: - [UDP] CESD Contents Delivery Management
    cesdcdman = 2921, 'cesdcdman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CESD Contents Delivery Data Transfer
    #: - [UDP] CESD Contents Delivery Data Transfer
    cesdcdtrn = 2922, 'cesdcdtrn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WTA-WSP-WTP-S
    #: - [UDP] WTA-WSP-WTP-S
    wta_wsp_wtp_s = 2923, 'wta-wsp-wtp-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PRECISE-VIP
    #: - [UDP] PRECISE-VIP
    precise_vip = 2924, 'precise-vip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned (FRP-Released 12/7/00)
    unassigned_2925 = 2925, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] MOBILE-FILE-DL
    #: - [UDP] MOBILE-FILE-DL
    mobile_file_dl = 2926, 'mobile-file-dl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UNIMOBILECTRL
    #: - [UDP] UNIMOBILECTRL
    unimobilectrl = 2927, 'unimobilectrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] REDSTONE-CPSS
    #: - [UDP] REDSTONE-CPSS
    redstone_cpss = 2928, 'redstone-cpss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMX-WEBADMIN
    #: - [UDP] AMX-WEBADMIN
    amx_webadmin = 2929, 'amx-webadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMX-WEBLINX
    #: - [UDP] AMX-WEBLINX
    amx_weblinx = 2930, 'amx-weblinx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Circle-X
    #: - [UDP] Circle-X
    circle_x = 2931, 'circle-x', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INCP
    #: - [UDP] INCP
    incp = 2932, 'incp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 4-TIER OPM GW
    #: - [UDP] 4-TIER OPM GW
    AppType_4_tieropmgw = 2933, '4-tieropmgw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 4-TIER OPM CLI
    #: - [UDP] 4-TIER OPM CLI
    AppType_4_tieropmcli = 2934, '4-tieropmcli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QTP
    #: - [UDP] QTP
    qtp = 2935, 'qtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OTPatch
    #: - [UDP] OTPatch
    otpatch = 2936, 'otpatch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PNACONSULT-LM
    #: - [UDP] PNACONSULT-LM
    pnaconsult_lm = 2937, 'pnaconsult-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SM-PAS-1
    #: - [UDP] SM-PAS-1
    sm_pas_1 = 2938, 'sm-pas-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SM-PAS-2
    #: - [UDP] SM-PAS-2
    sm_pas_2 = 2939, 'sm-pas-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SM-PAS-3
    #: - [UDP] SM-PAS-3
    sm_pas_3 = 2940, 'sm-pas-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SM-PAS-4
    #: - [UDP] SM-PAS-4
    sm_pas_4 = 2941, 'sm-pas-4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SM-PAS-5
    #: - [UDP] SM-PAS-5
    sm_pas_5 = 2942, 'sm-pas-5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TTNRepository
    #: - [UDP] TTNRepository
    ttnrepository = 2943, 'ttnrepository', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Megaco H-248
    #: - [UDP] Megaco H-248
    #: - [SCTP] Megaco-H.248 text
    megaco_h248 = 2944, 'megaco-h248', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] H248 Binary
    #: - [UDP] H248 Binary
    #: - [SCTP] Megaco/H.248 binary
    h248_binary = 2945, 'h248-binary', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] FJSVmpor
    #: - [UDP] FJSVmpor
    fjsvmpor = 2946, 'fjsvmpor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GPS Daemon request/response protocol
    #: - [UDP] GPS Daemon request/response protocol
    gpsd = 2947, 'gpsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP PUSH
    #: - [UDP] WAP PUSH
    wap_push = 2948, 'wap-push', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP PUSH SECURE
    #: - [UDP] WAP PUSH SECURE
    wap_pushsecure = 2949, 'wap-pushsecure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESIP
    #: - [UDP] ESIP
    esip = 2950, 'esip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OTTP
    #: - [UDP] OTTP
    ottp = 2951, 'ottp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MPFWSAS
    #: - [UDP] MPFWSAS
    mpfwsas = 2952, 'mpfwsas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OVALARMSRV
    #: - [UDP] OVALARMSRV
    ovalarmsrv = 2953, 'ovalarmsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OVALARMSRV-CMD
    #: - [UDP] OVALARMSRV-CMD
    ovalarmsrv_cmd = 2954, 'ovalarmsrv-cmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSNOTIFY
    #: - [UDP] CSNOTIFY
    csnotify = 2955, 'csnotify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OVRIMOSDBMAN
    #: - [UDP] OVRIMOSDBMAN
    ovrimosdbman = 2956, 'ovrimosdbman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JAMCT5
    #: - [UDP] JAMCT5
    jmact5 = 2957, 'jmact5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JAMCT6
    #: - [UDP] JAMCT6
    jmact6 = 2958, 'jmact6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RMOPAGT
    #: - [UDP] RMOPAGT
    rmopagt = 2959, 'rmopagt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DFOXSERVER
    #: - [UDP] DFOXSERVER
    dfoxserver = 2960, 'dfoxserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BOLDSOFT-LM
    #: - [UDP] BOLDSOFT-LM
    boldsoft_lm = 2961, 'boldsoft-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPH-POLICY-CLI
    #: - [UDP] IPH-POLICY-CLI
    iph_policy_cli = 2962, 'iph-policy-cli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPH-POLICY-ADM
    #: - [UDP] IPH-POLICY-ADM
    iph_policy_adm = 2963, 'iph-policy-adm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BULLANT SRAP
    #: - [UDP] BULLANT SRAP
    bullant_srap = 2964, 'bullant-srap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BULLANT RAP
    #: - [UDP] BULLANT RAP
    bullant_rap = 2965, 'bullant-rap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IDP-INFOTRIEVE
    #: - [UDP] IDP-INFOTRIEVE
    idp_infotrieve = 2966, 'idp-infotrieve', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSC-AGENT
    #: - [UDP] SSC-AGENT
    ssc_agent = 2967, 'ssc-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ENPP
    #: - [UDP] ENPP
    enpp = 2968, 'enpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESSP
    #: - [UDP] ESSP
    essp = 2969, 'essp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INDEX-NET
    #: - [UDP] INDEX-NET
    index_net = 2970, 'index-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetClip clipboard daemon
    #: - [UDP] NetClip clipboard daemon
    netclip = 2971, 'netclip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PMSM Webrctl
    #: - [UDP] PMSM Webrctl
    pmsm_webrctl = 2972, 'pmsm-webrctl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SV Networks
    #: - [UDP] SV Networks
    svnetworks = 2973, 'svnetworks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Signal
    #: - [UDP] Signal
    signal = 2974, 'signal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Configuration Management Service
    #: - [UDP] Fujitsu Configuration Management Service
    fjmpcm = 2975, 'fjmpcm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CNS Server Port
    #: - [UDP] CNS Server Port
    cns_srv_port = 2976, 'cns-srv-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TTCs Enterprise Test Access Protocol - NS
    #: - [UDP] TTCs Enterprise Test Access Protocol - NS
    ttc_etap_ns = 2977, 'ttc-etap-ns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TTCs Enterprise Test Access Protocol - DS
    #: - [UDP] TTCs Enterprise Test Access Protocol - DS
    ttc_etap_ds = 2978, 'ttc-etap-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] H.263 Video Streaming
    #: - [UDP] H.263 Video Streaming
    h263_video = 2979, 'h263-video', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Instant Messaging Service
    #: - [UDP] Instant Messaging Service
    wimd = 2980, 'wimd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MYLXAMPORT
    #: - [UDP] MYLXAMPORT
    mylxamport = 2981, 'mylxamport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IWB-WHITEBOARD
    #: - [UDP] IWB-WHITEBOARD
    iwb_whiteboard = 2982, 'iwb-whiteboard', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETPLAN
    #: - [UDP] NETPLAN
    netplan = 2983, 'netplan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HPIDSADMIN
    #: - [UDP] HPIDSADMIN
    hpidsadmin = 2984, 'hpidsadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HPIDSAGENT
    #: - [UDP] HPIDSAGENT
    hpidsagent = 2985, 'hpidsagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STONEFALLS
    #: - [UDP] STONEFALLS
    stonefalls = 2986, 'stonefalls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] identify
    #: - [UDP] identify
    identify = 2987, 'identify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HIPPA Reporting Protocol
    #: - [UDP] HIPPA Reporting Protocol
    hippad = 2988, 'hippad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ZARKOV Intelligent Agent Communication
    #: - [UDP] ZARKOV Intelligent Agent Communication
    zarkov = 2989, 'zarkov', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BOSCAP
    #: - [UDP] BOSCAP
    boscap = 2990, 'boscap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WKSTN-MON
    #: - [UDP] WKSTN-MON
    wkstn_mon = 2991, 'wkstn-mon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avenyo Server
    #: - [UDP] Avenyo Server
    avenyo = 2992, 'avenyo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS VIS1
    #: - [UDP] VERITAS VIS1
    veritas_vis1 = 2993, 'veritas-vis1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS VIS2
    #: - [UDP] VERITAS VIS2
    veritas_vis2 = 2994, 'veritas-vis2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IDRS
    #: - [UDP] IDRS
    idrs = 2995, 'idrs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vsixml
    #: - [UDP] vsixml
    vsixml = 2996, 'vsixml', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] REBOL
    #: - [UDP] REBOL
    rebol = 2997, 'rebol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Real Secure
    #: - [UDP] Real Secure
    realsecure = 2998, 'realsecure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RemoteWare Unassigned
    #: - [UDP] RemoteWare Unassigned
    remoteware_un = 2999, 'remoteware-un', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HBCI
    #: - [UDP] HBCI
    hbci = 3000, 'hbci', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RemoteWare Client
    #: - [UDP] RemoteWare Client
    remoteware_cl = 3000, 'remoteware-cl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] OrigoDB Server Native Interface
    origo_native = 3001, 'origo-native', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3001 = 3001, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] EXLM Agent
    #: - [UDP] EXLM Agent
    exlm_agent = 3002, 'exlm-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RemoteWare Server
    #: - [UDP] RemoteWare Server
    remoteware_srv = 3002, 'remoteware-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CGMS
    #: - [UDP] CGMS
    cgms = 3003, 'cgms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Csoft Agent
    #: - [UDP] Csoft Agent
    csoftragent = 3004, 'csoftragent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Genius License Manager
    #: - [UDP] Genius License Manager
    geniuslm = 3005, 'geniuslm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Instant Internet Admin
    #: - [UDP] Instant Internet Admin
    ii_admin = 3006, 'ii-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lotus Mail Tracking Agent Protocol
    #: - [UDP] Lotus Mail Tracking Agent Protocol
    lotusmtap = 3007, 'lotusmtap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Midnight Technologies
    #: - [UDP] Midnight Technologies
    midnight_tech = 3008, 'midnight-tech', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PXC-NTFY
    #: - [UDP] PXC-NTFY
    pxc_ntfy = 3009, 'pxc-ntfy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Telerate Workstation
    gw = 3010, 'gw', TransportProtocol.get('tcp')

    #: [UDP] Telerate Workstation
    ping_pong = 3010, 'ping-pong', TransportProtocol.get('udp')

    #: - [TCP] Trusted Web
    #: - [UDP] Trusted Web
    trusted_web = 3011, 'trusted-web', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trusted Web Client
    #: - [UDP] Trusted Web Client
    twsdss = 3012, 'twsdss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gilat Sky Surfer
    #: - [UDP] Gilat Sky Surfer
    gilatskysurfer = 3013, 'gilatskysurfer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Broker Service IANA assigned this well-formed service name as a
    #: replacement for "broker_service".
    #: - [TCP] Broker Service
    #: - [UDP] Broker Service IANA assigned this well-formed service name as a
    #: replacement for "broker_service".
    #: - [UDP] Broker Service
    broker_service = 3014, 'broker-service', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] NATI DSTP
    #: - [UDP] NATI DSTP
    nati_dstp = 3015, 'nati-dstp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Notify Server IANA assigned this well-formed service name as a
    #: replacement for "notify_srvr".
    #: - [TCP] Notify Server
    #: - [UDP] Notify Server IANA assigned this well-formed service name as a
    #: replacement for "notify_srvr".
    #: - [UDP] Notify Server
    notify_srvr = 3016, 'notify-srvr', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Event Listener IANA assigned this well-formed service name as a
    #: replacement for "event_listener".
    #: - [TCP] Event Listener
    #: - [UDP] Event Listener IANA assigned this well-formed service name as a
    #: replacement for "event_listener".
    #: - [UDP] Event Listener
    event_listener = 3017, 'event-listener', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Service Registry IANA assigned this well-formed service name as a
    #: replacement for "srvc_registry".
    #: - [TCP] Service Registry
    #: - [UDP] Service Registry IANA assigned this well-formed service name as a
    #: replacement for "srvc_registry".
    #: - [UDP] Service Registry
    srvc_registry = 3018, 'srvc-registry', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Resource Manager IANA assigned this well-formed service name as a
    #: replacement for "resource_mgr".
    #: - [TCP] Resource Manager
    #: - [UDP] Resource Manager IANA assigned this well-formed service name as a
    #: replacement for "resource_mgr".
    #: - [UDP] Resource Manager
    resource_mgr = 3019, 'resource-mgr', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] CIFS
    #: - [UDP] CIFS
    cifs = 3020, 'cifs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AGRI Server
    #: - [UDP] AGRI Server
    agriserver = 3021, 'agriserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSREGAGENT
    #: - [UDP] CSREGAGENT
    csregagent = 3022, 'csregagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] magicnotes
    #: - [UDP] magicnotes
    magicnotes = 3023, 'magicnotes', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NDS_SSO IANA assigned this well-formed service name as a replacement
    #: for "nds_sso".
    #: - [TCP] NDS_SSO
    #: - [UDP] NDS_SSO IANA assigned this well-formed service name as a replacement
    #: for "nds_sso".
    #: - [UDP] NDS_SSO
    nds_sso = 3024, 'nds-sso', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Arepa Raft
    #: - [UDP] Arepa Raft
    arepa_raft = 3025, 'arepa-raft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AGRI Gateway
    #: - [UDP] AGRI Gateway
    agri_gateway = 3026, 'agri-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LiebDevMgmt_C IANA assigned this well-formed service name as a
    #: replacement for "LiebDevMgmt_C".
    #: - [TCP] LiebDevMgmt_C
    #: - [UDP] LiebDevMgmt_C IANA assigned this well-formed service name as a
    #: replacement for "LiebDevMgmt_C".
    #: - [UDP] LiebDevMgmt_C
    liebdevmgmt_c = 3027, 'liebdevmgmt-c', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] LiebDevMgmt_DM IANA assigned this well-formed service name as a
    #: replacement for "LiebDevMgmt_DM".
    #: - [TCP] LiebDevMgmt_DM
    #: - [UDP] LiebDevMgmt_DM IANA assigned this well-formed service name as a
    #: replacement for "LiebDevMgmt_DM".
    #: - [UDP] LiebDevMgmt_DM
    liebdevmgmt_dm = 3028, 'liebdevmgmt-dm', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] LiebDevMgmt_A IANA assigned this well-formed service name as a
    #: replacement for "LiebDevMgmt_A".
    #: - [TCP] LiebDevMgmt_A
    #: - [UDP] LiebDevMgmt_A IANA assigned this well-formed service name as a
    #: replacement for "LiebDevMgmt_A".
    #: - [UDP] LiebDevMgmt_A
    liebdevmgmt_a = 3029, 'liebdevmgmt-a', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Arepa Cas
    #: - [UDP] Arepa Cas
    arepa_cas = 3030, 'arepa-cas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote AppleEvents/PPC Toolbox
    #: - [UDP] Remote AppleEvents/PPC Toolbox
    eppc = 3031, 'eppc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Redwood Chat
    #: - [UDP] Redwood Chat
    redwood_chat = 3032, 'redwood-chat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PDB
    #: - [UDP] PDB
    pdb = 3033, 'pdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Osmosis / Helix (R) AEEA Port
    #: - [UDP] Osmosis / Helix (R) AEEA Port
    osmosis_aeea = 3034, 'osmosis-aeea', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FJSV gssagt
    #: - [UDP] FJSV gssagt
    fjsv_gssagt = 3035, 'fjsv-gssagt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hagel DUMP
    #: - [UDP] Hagel DUMP
    hagel_dump = 3036, 'hagel-dump', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP SAN Mgmt
    #: - [UDP] HP SAN Mgmt
    hp_san_mgmt = 3037, 'hp-san-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Santak UPS
    #: - [UDP] Santak UPS
    santak_ups = 3038, 'santak-ups', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cogitate, Inc.
    #: - [UDP] Cogitate, Inc.
    cogitate = 3039, 'cogitate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tomato Springs
    #: - [UDP] Tomato Springs
    tomato_springs = 3040, 'tomato-springs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] di-traceware
    #: - [UDP] di-traceware
    di_traceware = 3041, 'di-traceware', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] journee
    #: - [UDP] journee
    journee = 3042, 'journee', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Broadcast Routing Protocol
    #: - [UDP] Broadcast Routing Protocol
    brp = 3043, 'brp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Extensible Provisioning Protocol [:rfc:`5734`]
    #: - [UDP] Extensible Provisioning Protocol [:rfc:`5734`]
    epp_700 = 700, 'epp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EndPoint Protocol
    #: - [UDP] EndPoint Protocol
    epp_3044 = 3044, 'epp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ResponseNet
    #: - [UDP] ResponseNet
    responsenet = 3045, 'responsenet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] di-ase
    #: - [UDP] di-ase
    di_ase = 3046, 'di-ase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fast Security HL Server
    #: - [UDP] Fast Security HL Server
    hlserver = 3047, 'hlserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sierra Net PC Trader
    #: - [UDP] Sierra Net PC Trader
    pctrader = 3048, 'pctrader', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NSWS
    #: - [UDP] NSWS
    nsws = 3049, 'nsws', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gds_db IANA assigned this well-formed service name as a replacement
    #: for "gds_db".
    #: - [TCP] gds_db
    #: - [UDP] gds_db IANA assigned this well-formed service name as a replacement
    #: for "gds_db".
    #: - [UDP] gds_db
    gds_db = 3050, 'gds-db', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Galaxy Server
    #: - [UDP] Galaxy Server
    galaxy_server = 3051, 'galaxy-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 3052
    #: - [UDP] APC 3052
    apc_3052 = 3052, 'apc-3052', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dsom-server
    #: - [UDP] dsom-server
    dsom_server = 3053, 'dsom-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMT CNF PROT
    #: - [UDP] AMT CNF PROT
    amt_cnf_prot = 3054, 'amt-cnf-prot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Policy Server
    #: - [UDP] Policy Server
    policyserver = 3055, 'policyserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CDL Server
    #: - [UDP] CDL Server
    cdl_server = 3056, 'cdl-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GoAhead FldUp
    #: - [UDP] GoAhead FldUp
    goahead_fldup = 3057, 'goahead-fldup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] videobeans
    #: - [UDP] videobeans
    videobeans = 3058, 'videobeans', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qsoft
    #: - [UDP] qsoft
    qsoft = 3059, 'qsoft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] interserver
    #: - [UDP] interserver
    interserver = 3060, 'interserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cautcpd
    #: - [UDP] cautcpd
    cautcpd = 3061, 'cautcpd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ncacn-ip-tcp
    #: - [UDP] ncacn-ip-tcp
    ncacn_ip_tcp = 3062, 'ncacn-ip-tcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ncadg-ip-udp
    #: - [UDP] ncadg-ip-udp
    ncadg_ip_udp = 3063, 'ncadg-ip-udp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Port Redirector
    #: - [UDP] Remote Port Redirector
    rprt = 3064, 'rprt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] slinterbase
    #: - [UDP] slinterbase
    slinterbase = 3065, 'slinterbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETATTACHSDMP
    #: - [UDP] NETATTACHSDMP
    netattachsdmp = 3066, 'netattachsdmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FJHPJP
    #: - [UDP] FJHPJP
    fjhpjp = 3067, 'fjhpjp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ls3 Broadcast
    #: - [UDP] ls3 Broadcast
    ls3bcast = 3068, 'ls3bcast', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ls3
    #: - [UDP] ls3
    ls3 = 3069, 'ls3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MGXSWITCH
    #: - [UDP] MGXSWITCH
    mgxswitch = 3070, 'mgxswitch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Crossplatform replication protocol
    xplat_replicate = 3071, 'xplat-replicate', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3071 = 3071, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ContinuStor Monitor Port
    #: - [UDP] ContinuStor Monitor Port
    csd_monitor = 3072, 'csd-monitor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Very simple chatroom prot
    #: - [UDP] Very simple chatroom prot
    vcrp = 3073, 'vcrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xbox game port
    #: - [UDP] Xbox game port
    xbox = 3074, 'xbox', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Orbix 2000 Locator
    #: - [UDP] Orbix 2000 Locator
    orbix_locator = 3075, 'orbix-locator', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Orbix 2000 Config
    #: - [UDP] Orbix 2000 Config
    orbix_config = 3076, 'orbix-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Orbix 2000 Locator SSL
    #: - [UDP] Orbix 2000 Locator SSL
    orbix_loc_ssl = 3077, 'orbix-loc-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Orbix 2000 Locator SSL
    #: - [UDP] Orbix 2000 Locator SSL
    orbix_cfg_ssl = 3078, 'orbix-cfg-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LV Front Panel
    #: - [UDP] LV Front Panel
    lv_frontpanel = 3079, 'lv-frontpanel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] stm_pproc IANA assigned this well-formed service name as a replacement
    #: for "stm_pproc".
    #: - [TCP] stm_pproc
    #: - [UDP] stm_pproc IANA assigned this well-formed service name as a replacement
    #: for "stm_pproc".
    #: - [UDP] stm_pproc
    stm_pproc = 3080, 'stm-pproc', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] TL1-LV
    #: - [UDP] TL1-LV
    tl1_lv = 3081, 'tl1-lv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TL1-RAW
    #: - [UDP] TL1-RAW
    tl1_raw = 3082, 'tl1-raw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TL1-TELNET
    #: - [UDP] TL1-TELNET
    tl1_telnet = 3083, 'tl1-telnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITM-MCCS
    #: - [UDP] ITM-MCCS
    itm_mccs = 3084, 'itm-mccs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCIHReq
    #: - [UDP] PCIHReq
    pcihreq = 3085, 'pcihreq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JDL-DBKitchen
    #: - [UDP] JDL-DBKitchen
    jdl_dbkitchen = 3086, 'jdl-dbkitchen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Asoki SMA
    #: - [UDP] Asoki SMA
    asoki_sma = 3087, 'asoki-sma', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eXtensible Data Transfer Protocol
    #: - [UDP] eXtensible Data Transfer Protocol
    xdtp = 3088, 'xdtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ParaTek Agent Linking
    #: - [UDP] ParaTek Agent Linking
    ptk_alink = 3089, 'ptk-alink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Senforce Session Services
    #: - [UDP] Senforce Session Services
    stss = 3090, 'stss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 1Ci Server Management
    #: - [UDP] 1Ci Server Management
    AppType_1ci_smcs = 3091, '1ci-smcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_3092 = 3092, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Jiiva RapidMQ Center
    #: - [UDP] Jiiva RapidMQ Center
    rapidmq_center = 3093, 'rapidmq-center', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jiiva RapidMQ Registry
    #: - [UDP] Jiiva RapidMQ Registry
    rapidmq_reg = 3094, 'rapidmq-reg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Panasas rendezvous port
    #: - [UDP] Panasas rendezvous port
    panasas = 3095, 'panasas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Active Print Server Port
    #: - [UDP] Active Print Server Port
    ndl_aps = 3096, 'ndl-aps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_3097 = 3097, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] ITU-T Q.1902.1/Q.2150.3
    itu_bicc_stc = 3097, 'itu-bicc-stc', TransportProtocol.get('sctp')

    #: - [TCP] Universal Message Manager
    #: - [UDP] Universal Message Manager
    umm_port = 3098, 'umm-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CHIPSY Machine Daemon
    #: - [UDP] CHIPSY Machine Daemon
    chmd = 3099, 'chmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpCon/xps
    #: - [UDP] OpCon/xps
    opcon_xps = 3100, 'opcon-xps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP PolicyXpert PIB Server
    #: - [UDP] HP PolicyXpert PIB Server
    hp_pxpib = 3101, 'hp-pxpib', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SoftlinK Slave Mon Port
    #: - [UDP] SoftlinK Slave Mon Port
    slslavemon = 3102, 'slslavemon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Autocue SMI Protocol
    #: - [UDP] Autocue SMI Protocol
    autocuesmi = 3103, 'autocuesmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Autocue Logger Protocol
    autocuelog = 3104, 'autocuelog', TransportProtocol.get('tcp')

    #: [UDP] Autocue Time Service
    autocuetime = 3104, 'autocuetime', TransportProtocol.get('udp')

    #: - [TCP] Cardbox
    #: - [UDP] Cardbox
    cardbox = 3105, 'cardbox', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cardbox HTTP
    #: - [UDP] Cardbox HTTP
    cardbox_http = 3106, 'cardbox-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Business protocol
    #: - [UDP] Business protocol
    business = 3107, 'business', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Geolocate protocol
    #: - [UDP] Geolocate protocol
    geolocate = 3108, 'geolocate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Personnel protocol
    #: - [UDP] Personnel protocol
    personnel = 3109, 'personnel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] simulator control port
    #: - [UDP] simulator control port
    sim_control = 3110, 'sim-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Web Synchronous Services
    #: - [UDP] Web Synchronous Services
    wsynch = 3111, 'wsynch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KDE System Guard
    #: - [UDP] KDE System Guard
    ksysguard = 3112, 'ksysguard', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CS-Authenticate Svr Port
    #: - [UDP] CS-Authenticate Svr Port
    cs_auth_svr = 3113, 'cs-auth-svr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CCM AutoDiscover
    #: - [UDP] CCM AutoDiscover
    ccmad = 3114, 'ccmad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCTET Master
    #: - [UDP] MCTET Master
    mctet_master = 3115, 'mctet-master', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCTET Gateway
    #: - [UDP] MCTET Gateway
    mctet_gateway = 3116, 'mctet-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCTET Jserv
    #: - [UDP] MCTET Jserv
    mctet_jserv = 3117, 'mctet-jserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PKAgent
    #: - [UDP] PKAgent
    pkagent = 3118, 'pkagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D2000 Kernel Port
    #: - [UDP] D2000 Kernel Port
    d2000kernel = 3119, 'd2000kernel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D2000 Webserver Port
    #: - [UDP] D2000 Webserver Port
    d2000webserver = 3120, 'd2000webserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] The pacemaker remote (pcmk-remote) service extends high availability
    #: functionality outside of the Linux cluster into remote nodes.
    pcmk_remote = 3121, 'pcmk-remote', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3121 = 3121, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] MTI VTR Emulator port
    #: - [UDP] MTI VTR Emulator port
    vtr_emulator = 3122, 'vtr-emulator', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EDI Translation Protocol
    #: - [UDP] EDI Translation Protocol
    edix = 3123, 'edix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Beacon Port
    #: - [UDP] Beacon Port
    beacon_port = 3124, 'beacon-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A13-AN Interface
    #: - [UDP] A13-AN Interface
    a13_an = 3125, 'a13-an', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_3126 = 3126, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] CTX Bridge Port
    #: - [UDP] CTX Bridge Port
    ctx_bridge = 3127, 'ctx-bridge', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Active API Server Port
    #: - [UDP] Active API Server Port
    ndl_aas = 3128, 'ndl-aas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetPort Discovery Port
    #: - [UDP] NetPort Discovery Port
    netport_id = 3129, 'netport-id', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICPv2
    #: - [UDP] ICPv2
    icpv2 = 3130, 'icpv2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net Book Mark
    #: - [UDP] Net Book Mark
    netbookmark = 3131, 'netbookmark', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Business Rule Engine Update Service
    #: - [UDP] Microsoft Business Rule Engine Update Service
    ms_rule_engine = 3132, 'ms-rule-engine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prism Deploy User Port
    #: - [UDP] Prism Deploy User Port
    prism_deploy = 3133, 'prism-deploy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Extensible Code Protocol
    #: - [UDP] Extensible Code Protocol
    ecp = 3134, 'ecp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PeerBook Port
    #: - [UDP] PeerBook Port
    peerbook_port = 3135, 'peerbook-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Grub Server Port
    #: - [UDP] Grub Server Port
    grubd = 3136, 'grubd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rtnt-1 data packets
    #: - [UDP] rtnt-1 data packets
    rtnt_1 = 3137, 'rtnt-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rtnt-2 data packets
    #: - [UDP] rtnt-2 data packets
    rtnt_2 = 3138, 'rtnt-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Incognito Rendez-Vous
    #: - [UDP] Incognito Rendez-Vous
    incognitorv = 3139, 'incognitorv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Arilia Multiplexor
    #: - [UDP] Arilia Multiplexor
    ariliamulti = 3140, 'ariliamulti', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VMODEM
    #: - [UDP] VMODEM
    vmodem = 3141, 'vmodem', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RDC WH EOS
    #: - [UDP] RDC WH EOS
    rdc_wh_eos = 3142, 'rdc-wh-eos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sea View
    #: - [UDP] Sea View
    seaview = 3143, 'seaview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tarantella
    #: - [UDP] Tarantella
    tarantella = 3144, 'tarantella', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSI-LFAP
    #: - [UDP] CSI-LFAP
    csi_lfap = 3145, 'csi-lfap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bears-02
    #: - [UDP] bears-02
    bears_02 = 3146, 'bears-02', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RFIO
    #: - [UDP] RFIO
    rfio = 3147, 'rfio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetMike Game Administrator
    #: - [UDP] NetMike Game Administrator
    nm_game_admin = 3148, 'nm-game-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetMike Game Server
    #: - [UDP] NetMike Game Server
    nm_game_server = 3149, 'nm-game-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetMike Assessor Administrator
    #: - [UDP] NetMike Assessor Administrator
    nm_asses_admin = 3150, 'nm-asses-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetMike Assessor
    #: - [UDP] NetMike Assessor
    nm_assessor = 3151, 'nm-assessor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FeiTian Port
    #: - [UDP] FeiTian Port
    feitianrockey = 3152, 'feitianrockey', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] S8Cargo Client Port
    #: - [UDP] S8Cargo Client Port
    s8_client_port = 3153, 's8-client-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ON RMI Registry
    #: - [UDP] ON RMI Registry
    ccmrmi = 3154, 'ccmrmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JpegMpeg Port
    #: - [UDP] JpegMpeg Port
    jpegmpeg = 3155, 'jpegmpeg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Indura Collector
    #: - [UDP] Indura Collector
    indura = 3156, 'indura', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CCC Listener Port
    #: - [UDP] CCC Listener Port
    e3consultants = 3157, 'e3consultants', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SmashTV Protocol
    #: - [UDP] SmashTV Protocol
    stvp = 3158, 'stvp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NavegaWeb Tarification
    #: - [UDP] NavegaWeb Tarification
    navegaweb_port = 3159, 'navegaweb-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TIP Application Server
    #: - [UDP] TIP Application Server
    tip_app_server = 3160, 'tip-app-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DOC1 License Manager
    #: - [UDP] DOC1 License Manager
    doc1lm = 3161, 'doc1lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SFLM
    #: - [UDP] SFLM
    sflm = 3162, 'sflm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RES-SAP
    #: - [UDP] RES-SAP
    res_sap = 3163, 'res-sap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IMPRS
    #: - [UDP] IMPRS
    imprs = 3164, 'imprs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Newgenpay Engine Service
    #: - [UDP] Newgenpay Engine Service
    newgenpay = 3165, 'newgenpay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quest Spotlight Out-Of-Process Collector
    #: - [UDP] Quest Spotlight Out-Of-Process Collector
    sossecollector = 3166, 'sossecollector', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Now Contact Public Server
    #: - [UDP] Now Contact Public Server
    nowcontact = 3167, 'nowcontact', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Now Up-to-Date Public Server
    #: - [UDP] Now Up-to-Date Public Server
    poweronnud = 3168, 'poweronnud', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SERVERVIEW-AS
    #: - [UDP] SERVERVIEW-AS
    serverview_as = 3169, 'serverview-as', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SERVERVIEW-ASN
    #: - [UDP] SERVERVIEW-ASN
    serverview_asn = 3170, 'serverview-asn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SERVERVIEW-GF
    #: - [UDP] SERVERVIEW-GF
    serverview_gf = 3171, 'serverview-gf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SERVERVIEW-RM
    #: - [UDP] SERVERVIEW-RM
    serverview_rm = 3172, 'serverview-rm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SERVERVIEW-ICC
    #: - [UDP] SERVERVIEW-ICC
    serverview_icc = 3173, 'serverview-icc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARMI Server
    #: - [UDP] ARMI Server
    armi_server = 3174, 'armi-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] T1_E1_Over_IP
    #: - [UDP] T1_E1_Over_IP
    t1_e1_over_ip = 3175, 't1-e1-over-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARS Master
    #: - [UDP] ARS Master
    ars_master = 3176, 'ars-master', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Phonex Protocol
    #: - [UDP] Phonex Protocol
    phonex_port = 3177, 'phonex-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Radiance UltraEdge Port
    #: - [UDP] Radiance UltraEdge Port
    radclientport = 3178, 'radclientport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] H2GF W.2m Handover prot.
    #: - [UDP] H2GF W.2m Handover prot.
    h2gf_w_2m = 3179, 'h2gf-w-2m', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Millicent Broker Server
    #: - [UDP] Millicent Broker Server
    mc_brk_srv = 3180, 'mc-brk-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC Patrol Agent
    #: - [UDP] BMC Patrol Agent
    bmcpatrolagent = 3181, 'bmcpatrolagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC Patrol Rendezvous
    #: - [UDP] BMC Patrol Rendezvous
    bmcpatrolrnvu = 3182, 'bmcpatrolrnvu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] COPS/TLS
    #: - [UDP] COPS/TLS
    cops_tls = 3183, 'cops-tls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ApogeeX Port
    #: - [UDP] ApogeeX Port
    apogeex_port = 3184, 'apogeex-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SuSE Meta PPPD
    #: - [UDP] SuSE Meta PPPD
    smpppd = 3185, 'smpppd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IIW Monitor User Port
    #: - [UDP] IIW Monitor User Port
    iiw_port = 3186, 'iiw-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Open Design Listen Port
    #: - [UDP] Open Design Listen Port
    odi_port = 3187, 'odi-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Broadcom Port
    #: - [UDP] Broadcom Port
    brcm_comm_port = 3188, 'brcm-comm-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pinnacle Sys InfEx Port
    #: - [UDP] Pinnacle Sys InfEx Port
    pcle_infex = 3189, 'pcle-infex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ConServR Proxy
    #: - [UDP] ConServR Proxy
    csvr_proxy = 3190, 'csvr-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ConServR SSL Proxy
    #: - [UDP] ConServR SSL Proxy
    csvr_sslproxy = 3191, 'csvr-sslproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FireMon Revision Control
    #: - [UDP] FireMon Revision Control
    firemonrcc = 3192, 'firemonrcc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SpanDataPort
    #: - [UDP] SpanDataPort
    spandataport = 3193, 'spandataport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rockstorm MAG protocol
    #: - [UDP] Rockstorm MAG protocol
    magbind = 3194, 'magbind', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Control Unit
    #: - [UDP] Network Control Unit
    ncu_1 = 3195, 'ncu-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Control Unit
    #: - [UDP] Network Control Unit
    ncu_2 = 3196, 'ncu-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Embrace Device Protocol Server
    #: - [UDP] Embrace Device Protocol Server
    embrace_dp_s = 3197, 'embrace-dp-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Embrace Device Protocol Client
    #: - [UDP] Embrace Device Protocol Client
    embrace_dp_c = 3198, 'embrace-dp-c', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DMOD WorkSpace
    #: - [UDP] DMOD WorkSpace
    dmod_workspace = 3199, 'dmod-workspace', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Press-sense Tick Port
    #: - [UDP] Press-sense Tick Port
    tick_port = 3200, 'tick-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CPQ-TaskSmart
    #: - [UDP] CPQ-TaskSmart
    cpq_tasksmart = 3201, 'cpq-tasksmart', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IntraIntra
    #: - [UDP] IntraIntra
    intraintra = 3202, 'intraintra', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Watcher Monitor
    #: - [UDP] Network Watcher Monitor
    netwatcher_mon = 3203, 'netwatcher-mon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Watcher DB Access
    #: - [UDP] Network Watcher DB Access
    netwatcher_db = 3204, 'netwatcher-db', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iSNS Server Port [:rfc:`4171`]
    #: - [UDP] iSNS Server Port [:rfc:`4171`]
    isns = 3205, 'isns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IronMail POP Proxy
    #: - [UDP] IronMail POP Proxy
    ironmail = 3206, 'ironmail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Veritas Authentication Port
    #: - [UDP] Veritas Authentication Port
    vx_auth_port = 3207, 'vx-auth-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PFU PR Callback
    #: - [UDP] PFU PR Callback
    pfu_prcallback = 3208, 'pfu-prcallback', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP OpenView Network Path Engine Server
    #: - [UDP] HP OpenView Network Path Engine Server
    netwkpathengine = 3209, 'netwkpathengine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Flamenco Networks Proxy
    #: - [UDP] Flamenco Networks Proxy
    flamenco_proxy = 3210, 'flamenco-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avocent Secure Management
    #: - [UDP] Avocent Secure Management
    avsecuremgmt = 3211, 'avsecuremgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Survey Instrument
    #: - [UDP] Survey Instrument
    surveyinst = 3212, 'surveyinst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NEON 24X7 Mission Control
    #: - [UDP] NEON 24X7 Mission Control
    neon24x7 = 3213, 'neon24x7', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JMQ Daemon Port 1
    #: - [UDP] JMQ Daemon Port 1
    jmq_daemon_1 = 3214, 'jmq-daemon-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JMQ Daemon Port 2
    #: - [UDP] JMQ Daemon Port 2
    jmq_daemon_2 = 3215, 'jmq-daemon-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ferrari electronic FOAM
    #: - [UDP] Ferrari electronic FOAM
    ferrari_foam = 3216, 'ferrari-foam', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unified IP & Telecom Environment
    #: - [UDP] Unified IP & Telecom Environment
    unite = 3217, 'unite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMC SmartPackets
    #: - [UDP] EMC SmartPackets
    smartpackets = 3218, 'smartpackets', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WMS Messenger
    #: - [UDP] WMS Messenger
    wms_messenger = 3219, 'wms-messenger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XML NM over SSL
    #: - [UDP] XML NM over SSL
    xnm_ssl = 3220, 'xnm-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XML NM over TCP
    #: - [UDP] XML NM over TCP
    xnm_clear_text = 3221, 'xnm-clear-text', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Gateway Load Balancing Pr
    #: - [UDP] Gateway Load Balancing Pr
    glbp = 3222, 'glbp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DIGIVOTE (R) Vote-Server
    #: - [UDP] DIGIVOTE (R) Vote-Server
    digivote = 3223, 'digivote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AES Discovery Port
    #: - [UDP] AES Discovery Port
    aes_discovery = 3224, 'aes-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FCIP [:rfc:`3821`]
    #: - [UDP] FCIP [:rfc:`3821`]
    fcip_port = 3225, 'fcip-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISI Industry Software IRP
    #: - [UDP] ISI Industry Software IRP
    isi_irp = 3226, 'isi-irp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DiamondWave NMS Server
    #: - [UDP] DiamondWave NMS Server
    dwnmshttp = 3227, 'dwnmshttp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DiamondWave MSG Server
    #: - [UDP] DiamondWave MSG Server
    dwmsgserver = 3228, 'dwmsgserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Global CD Port
    #: - [UDP] Global CD Port
    global_cd_port = 3229, 'global-cd-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Software Distributor Port
    #: - [UDP] Software Distributor Port
    sftdst_port = 3230, 'sftdst-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VidiGo communication (previous was: Delta Solutions Direct)
    #: - [UDP] VidiGo communication (previous was: Delta Solutions Direct)
    vidigo = 3231, 'vidigo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MDT port [:rfc:`6513`]
    #: - [UDP] MDT port [:rfc:`6513`]
    mdtp = 3232, 'mdtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WhiskerControl main port
    #: - [UDP] WhiskerControl main port
    whisker = 3233, 'whisker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alchemy Server
    #: - [UDP] Alchemy Server
    alchemy = 3234, 'alchemy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MDAP port
    #: - [UDP] MDAP Port
    mdap_port = 3235, 'mdap-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] appareNet Test Server
    #: - [UDP] appareNet Test Server
    apparenet_ts = 3236, 'apparenet-ts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] appareNet Test Packet Sequencer
    #: - [UDP] appareNet Test Packet Sequencer
    apparenet_tps = 3237, 'apparenet-tps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] appareNet Analysis Server
    #: - [UDP] appareNet Analysis Server
    apparenet_as = 3238, 'apparenet-as', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] appareNet User Interface
    #: - [UDP] appareNet User Interface
    apparenet_ui = 3239, 'apparenet-ui', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trio Motion Control Port
    #: - [UDP] Trio Motion Control Port
    triomotion = 3240, 'triomotion', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SysOrb Monitoring Server
    #: - [UDP] SysOrb Monitoring Server
    sysorb = 3241, 'sysorb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Session Description ID
    #: - [UDP] Session Description ID
    sdp_id_port = 3242, 'sdp-id-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Timelot Port
    #: - [UDP] Timelot Port
    timelot = 3243, 'timelot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OneSAF
    #: - [UDP] OneSAF
    onesaf = 3244, 'onesaf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VIEO Fabric Executive
    #: - [UDP] VIEO Fabric Executive
    vieo_fe = 3245, 'vieo-fe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DVT SYSTEM PORT
    #: - [UDP] DVT SYSTEM PORT
    dvt_system = 3246, 'dvt-system', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DVT DATA LINK
    #: - [UDP] DVT DATA LINK
    dvt_data = 3247, 'dvt-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PROCOS LM
    #: - [UDP] PROCOS LM
    procos_lm = 3248, 'procos-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] State Sync Protocol
    #: - [UDP] State Sync Protocol
    ssp = 3249, 'ssp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HMS hicp port
    #: - [UDP] HMS hicp port
    hicp = 3250, 'hicp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sys Scanner
    #: - [UDP] Sys Scanner
    sysscanner = 3251, 'sysscanner', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DHE port
    #: - [UDP] DHE port
    dhe = 3252, 'dhe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PDA Data
    #: - [UDP] PDA Data
    pda_data = 3253, 'pda-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PDA System
    #: - [UDP] PDA System
    pda_sys = 3254, 'pda-sys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Semaphore Connection Port
    #: - [UDP] Semaphore Connection Port
    semaphore = 3255, 'semaphore', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compaq RPM Agent Port
    #: - [UDP] Compaq RPM Agent Port
    cpqrpm_agent = 3256, 'cpqrpm-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compaq RPM Server Port
    #: - [UDP] Compaq RPM Server Port
    cpqrpm_server = 3257, 'cpqrpm-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ivecon Server Port
    #: - [UDP] Ivecon Server Port
    ivecon_port = 3258, 'ivecon-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Epson Network Common Devi
    #: - [UDP] Epson Network Common Devi
    epncdp2 = 3259, 'epncdp2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iSCSI port [:rfc:`7143`]
    #: - [UDP] iSCSI port [:rfc:`7143`]
    iscsi_target = 3260, 'iscsi-target', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] winShadow
    #: - [UDP] winShadow
    winshadow = 3261, 'winshadow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NECP
    #: - [UDP] NECP
    necp = 3262, 'necp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] E-Color Enterprise Imager
    #: - [UDP] E-Color Enterprise Imager
    ecolor_imager = 3263, 'ecolor-imager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cc:mail/lotus
    #: - [UDP] cc:mail/lotus
    ccmail = 3264, 'ccmail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Altav Tunnel
    #: - [UDP] Altav Tunnel
    altav_tunnel = 3265, 'altav-tunnel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NS CFG Server
    #: - [UDP] NS CFG Server
    ns_cfg_server = 3266, 'ns-cfg-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Dial Out
    #: - [UDP] IBM Dial Out
    ibm_dial_out = 3267, 'ibm-dial-out', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Global Catalog
    #: - [UDP] Microsoft Global Catalog
    msft_gc = 3268, 'msft-gc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Global Catalog with LDAP/SSL
    #: - [UDP] Microsoft Global Catalog with LDAP/SSL
    msft_gc_ssl = 3269, 'msft-gc-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Verismart
    #: - [UDP] Verismart
    verismart = 3270, 'verismart', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSoft Prev Port
    #: - [UDP] CSoft Prev Port
    csoft_prev = 3271, 'csoft-prev', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu User Manager
    #: - [UDP] Fujitsu User Manager
    user_manager = 3272, 'user-manager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Extensible Multiplexed Protocol
    #: - [UDP] Simple Extensible Multiplexed Protocol
    sxmp = 3273, 'sxmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ordinox Server
    #: - [UDP] Ordinox Server
    ordinox_server = 3274, 'ordinox-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SAMD
    #: - [UDP] SAMD
    samd = 3275, 'samd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Maxim ASICs
    #: - [UDP] Maxim ASICs
    maxim_asics = 3276, 'maxim-asics', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AWG Proxy
    #: - [UDP] AWG Proxy
    awg_proxy = 3277, 'awg-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LKCM Server
    #: - [UDP] LKCM Server
    lkcmserver = 3278, 'lkcmserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VS Server
    #: - [UDP] VS Server
    vs_server = 3280, 'vs-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SYSOPT
    #: - [UDP] SYSOPT
    sysopt = 3281, 'sysopt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Datusorb
    #: - [UDP] Datusorb
    datusorb = 3282, 'datusorb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net Assistant
    #: - [UDP] Net Assistant
    apple_remote_desktop = 3283, 'apple remote desktop (net assistant)', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 4Talk
    #: - [UDP] 4Talk
    AppType_4talk = 3284, '4talk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Plato
    #: - [UDP] Plato
    plato = 3285, 'plato', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] E-Net
    #: - [UDP] E-Net
    e_net = 3286, 'e-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DIRECTVDATA
    #: - [UDP] DIRECTVDATA
    directvdata = 3287, 'directvdata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] COPS
    #: - [UDP] COPS
    cops = 3288, 'cops', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ENPC
    #: - [UDP] ENPC
    enpc = 3289, 'enpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CAPS LOGISTICS TOOLKIT - LM
    #: - [UDP] CAPS LOGISTICS TOOLKIT - LM
    caps_lm = 3290, 'caps-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] S A Holditch & Associates - LM
    #: - [UDP] S A Holditch & Associates - LM
    sah_lm = 3291, 'sah-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cart O Rama
    #: - [UDP] Cart O Rama
    cart_o_rama = 3292, 'cart-o-rama', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fg-fps
    #: - [UDP] fg-fps
    fg_fps = 3293, 'fg-fps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fg-gip
    #: - [UDP] fg-gip
    fg_gip = 3294, 'fg-gip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dynamic IP Lookup
    #: - [UDP] Dynamic IP Lookup
    dyniplookup = 3295, 'dyniplookup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rib License Manager
    #: - [UDP] Rib License Manager
    rib_slm = 3296, 'rib-slm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cytel License Manager
    #: - [UDP] Cytel License Manager
    cytel_lm = 3297, 'cytel-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DeskView
    #: - [UDP] DeskView
    deskview = 3298, 'deskview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pdrncs
    #: - [UDP] pdrncs
    pdrncs = 3299, 'pdrncs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Ceph monitor
    ceph = 3300, 'ceph', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3300 = 3300, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Tarantool in-memory computing platform
    #: - [UDP] Tarantool in-memory computing platform
    tarantool = 3301, 'tarantool', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCS Fastmail
    #: - [UDP] MCS Fastmail
    mcs_fastmail = 3302, 'mcs-fastmail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OP Session Client
    #: - [UDP] OP Session Client
    opsession_clnt = 3303, 'opsession-clnt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OP Session Server
    #: - [UDP] OP Session Server
    opsession_srvr = 3304, 'opsession-srvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ODETTE-FTP [:rfc:`5024`]
    #: - [UDP] ODETTE-FTP [:rfc:`5024`]
    odette_ftp = 3305, 'odette-ftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MySQL
    #: - [UDP] MySQL
    mysql = 3306, 'mysql', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OP Session Proxy
    #: - [UDP] OP Session Proxy
    opsession_prxy = 3307, 'opsession-prxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TNS Server
    #: - [UDP] TNS Server
    tns_server = 3308, 'tns-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TNS ADV
    #: - [UDP] TNS ADV
    tns_adv = 3309, 'tns-adv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dyna Access
    #: - [UDP] Dyna Access
    dyna_access = 3310, 'dyna-access', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCNS Tel Ret
    #: - [UDP] MCNS Tel Ret
    mcns_tel_ret = 3311, 'mcns-tel-ret', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Application Management Server
    #: - [UDP] Application Management Server
    appman_server = 3312, 'appman-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unify Object Broker
    #: - [UDP] Unify Object Broker
    uorb = 3313, 'uorb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unify Object Host
    #: - [UDP] Unify Object Host
    uohost = 3314, 'uohost', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CDID
    #: - [UDP] CDID
    cdid = 3315, 'cdid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AICC/CMI
    #: - [UDP] AICC/CMI
    aicc_cmi = 3316, 'aicc-cmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSAI PORT
    #: - [UDP] VSAI PORT
    vsaiport = 3317, 'vsaiport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Swith to Swith Routing Information Protocol
    #: - [UDP] Swith to Swith Routing Information Protocol
    ssrip = 3318, 'ssrip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SDT License Manager
    #: - [UDP] SDT License Manager
    sdt_lmd = 3319, 'sdt-lmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Office Link 2000
    #: - [UDP] Office Link 2000
    officelink2000 = 3320, 'officelink2000', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VNSSTR
    #: - [UDP] VNSSTR
    vnsstr = 3321, 'vnsstr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SFTU
    #: - [UDP] SFTU
    sftu = 3326, 'sftu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BBARS
    #: - [UDP] BBARS
    bbars = 3327, 'bbars', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Eaglepoint License Manager
    #: - [UDP] Eaglepoint License Manager
    egptlm = 3328, 'egptlm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP Device Disc
    #: - [UDP] HP Device Disc
    hp_device_disc = 3329, 'hp-device-disc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCS Calypso ICF
    #: - [UDP] MCS Calypso ICF
    mcs_calypsoicf = 3330, 'mcs-calypsoicf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCS Messaging
    #: - [UDP] MCS Messaging
    mcs_messaging = 3331, 'mcs-messaging', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCS Mail Server
    #: - [UDP] MCS Mail Server
    mcs_mailsvr = 3332, 'mcs-mailsvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DEC Notes
    #: - [UDP] DEC Notes
    dec_notes = 3333, 'dec-notes', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Direct TV Webcasting
    #: - [UDP] Direct TV Webcasting
    directv_web = 3334, 'directv-web', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Direct TV Software Updates
    #: - [UDP] Direct TV Software Updates
    directv_soft = 3335, 'directv-soft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Direct TV Tickers
    #: - [UDP] Direct TV Tickers
    directv_tick = 3336, 'directv-tick', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Direct TV Data Catalog
    #: - [UDP] Direct TV Data Catalog
    directv_catlg = 3337, 'directv-catlg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMF data b
    #: - [UDP] OMF data b
    anet_b = 3338, 'anet-b', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMF data l
    #: - [UDP] OMF data l
    anet_l = 3339, 'anet-l', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMF data m
    #: - [UDP] OMF data m
    anet_m = 3340, 'anet-m', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMF data h
    #: - [UDP] OMF data h
    anet_h = 3341, 'anet-h', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WebTIE
    #: - [UDP] WebTIE
    webtie = 3342, 'webtie', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS Cluster Net
    #: - [UDP] MS Cluster Net
    ms_cluster_net = 3343, 'ms-cluster-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BNT Manager
    #: - [UDP] BNT Manager
    bnt_manager = 3344, 'bnt-manager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Influence
    #: - [UDP] Influence
    influence = 3345, 'influence', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trnsprnt Proxy
    #: - [UDP] Trnsprnt Proxy
    trnsprntproxy = 3346, 'trnsprntproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Phoenix RPC
    #: - [UDP] Phoenix RPC
    phoenix_rpc = 3347, 'phoenix-rpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pangolin Laser
    #: - [UDP] Pangolin Laser
    pangolin_laser = 3348, 'pangolin-laser', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Chevin Services
    #: - [UDP] Chevin Services
    chevinservices = 3349, 'chevinservices', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FINDVIATV
    #: - [UDP] FINDVIATV
    findviatv = 3350, 'findviatv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Btrieve port
    #: - [UDP] Btrieve port
    btrieve = 3351, 'btrieve', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Scalable SQL
    #: - [UDP] Scalable SQL
    ssql = 3352, 'ssql', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FATPIPE
    #: - [UDP] FATPIPE
    fatpipe = 3353, 'fatpipe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SUITJD
    #: - [UDP] SUITJD
    suitjd = 3354, 'suitjd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ordinox Dbase
    #: - [UDP] Ordinox Dbase
    ordinox_dbase = 3355, 'ordinox-dbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UPNOTIFYPS
    #: - [UDP] UPNOTIFYPS
    upnotifyps = 3356, 'upnotifyps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Adtech Test IP
    #: - [UDP] Adtech Test IP
    adtech_test = 3357, 'adtech-test', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mp Sys Rmsvr
    #: - [UDP] Mp Sys Rmsvr
    mpsysrmsvr = 3358, 'mpsysrmsvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WG NetForce
    #: - [UDP] WG NetForce
    wg_netforce = 3359, 'wg-netforce', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KV Server
    #: - [UDP] KV Server
    kv_server = 3360, 'kv-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KV Agent
    #: - [UDP] KV Agent
    kv_agent = 3361, 'kv-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DJ ILM
    #: - [UDP] DJ ILM
    dj_ilm = 3362, 'dj-ilm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NATI Vi Server
    #: - [UDP] NATI Vi Server
    nati_vi_server = 3363, 'nati-vi-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CreativeServer
    #: - [UDP] CreativeServer
    creativeserver_453 = 453, 'creativeserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Creative Server
    #: - [UDP] Creative Server
    creativeserver_3364 = 3364, 'creativeserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ContentServer
    #: - [UDP] ContentServer
    contentserver_454 = 454, 'contentserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Content Server
    #: - [UDP] Content Server
    contentserver_3365 = 3365, 'contentserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CreativePartnr
    #: - [UDP] CreativePartnr
    creativepartnr_455 = 455, 'creativepartnr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Creative Partner
    #: - [UDP] Creative Partner
    creativepartnr_3366 = 3366, 'creativepartnr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TIP 2
    #: - [UDP] TIP 2
    tip2 = 3372, 'tip2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lavenir License Manager
    #: - [UDP] Lavenir License Manager
    lavenir_lm = 3373, 'lavenir-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cluster Disc
    #: - [UDP] Cluster Disc
    cluster_disc = 3374, 'cluster-disc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSNM Agent
    #: - [UDP] VSNM Agent
    vsnm_agent = 3375, 'vsnm-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CD Broker
    #: - [UDP] CD Broker
    cdbroker = 3376, 'cdbroker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cogsys Network License Manager
    #: - [UDP] Cogsys Network License Manager
    cogsys_lm = 3377, 'cogsys-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WSICOPY
    #: - [UDP] WSICOPY
    wsicopy = 3378, 'wsicopy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SOCORFS
    #: - [UDP] SOCORFS
    socorfs = 3379, 'socorfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNS Channels
    #: - [UDP] SNS Channels
    sns_channels = 3380, 'sns-channels', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Geneous
    #: - [UDP] Geneous
    geneous = 3381, 'geneous', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Network Enhanced Antitheft function
    #: - [UDP] Fujitsu Network Enhanced Antitheft function
    fujitsu_neat = 3382, 'fujitsu-neat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Enterprise Software Products License Manager
    #: - [UDP] Enterprise Software Products License Manager
    esp_lm = 3383, 'esp-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cluster Management Services
    #: - [UDP] Hardware Management
    hp_clic = 3384, 'hp-clic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qnxnetman
    #: - [UDP] qnxnetman
    qnxnetman = 3385, 'qnxnetman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] GPRS Data
    gprs_data = 3386, 'gprs-data', TransportProtocol.get('tcp')

    #: [UDP] GPRS SIG
    gprs_sig = 3386, 'gprs-sig', TransportProtocol.get('udp')

    #: - [TCP] Back Room Net
    #: - [UDP] Back Room Net
    backroomnet = 3387, 'backroomnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CB Server
    #: - [UDP] CB Server
    cbserver = 3388, 'cbserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS WBT Server
    #: - [UDP] MS WBT Server
    ms_wbt_server = 3389, 'ms-wbt-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Distributed Service Coordinator
    #: - [UDP] Distributed Service Coordinator
    dsc = 3390, 'dsc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SAVANT
    #: - [UDP] SAVANT
    savant = 3391, 'savant', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EFI License Management
    #: - [UDP] EFI License Management
    efi_lm = 3392, 'efi-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D2K Tapestry Client to Server
    #: - [UDP] D2K Tapestry Client to Server
    d2k_tapestry1 = 3393, 'd2k-tapestry1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D2K Tapestry Server to Server
    #: - [UDP] D2K Tapestry Server to Server
    d2k_tapestry2 = 3394, 'd2k-tapestry2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dyna License Manager (Elam)
    #: - [UDP] Dyna License Manager (Elam)
    dyna_lm = 3395, 'dyna-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Printer Agent IANA assigned this well-formed service name as a
    #: replacement for "printer_agent".
    #: - [TCP] Printer Agent
    #: - [UDP] Printer Agent IANA assigned this well-formed service name as a
    #: replacement for "printer_agent".
    #: - [UDP] Printer Agent
    printer_agent = 3396, 'printer-agent', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Cloanto License Manager
    #: - [UDP] Cloanto License Manager
    cloanto_lm = 3397, 'cloanto-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mercantile
    #: - [UDP] Mercantile
    mercantile = 3398, 'mercantile', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSMS
    #: - [UDP] CSMS
    csms = 3399, 'csms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSMS2
    #: - [UDP] CSMS2
    csms2 = 3400, 'csms2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] filecast
    #: - [UDP] filecast
    filecast = 3401, 'filecast', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FXa Engine Network Port
    #: - [UDP] FXa Engine Network Port
    fxaengine_net = 3402, 'fxaengine-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Removed
    removed = 3404, 'removed', TransportProtocol.get('undefined')

    #: - [TCP] Nokia Announcement ch 1
    #: - [UDP] Nokia Announcement ch 1
    nokia_ann_ch1 = 3405, 'nokia-ann-ch1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nokia Announcement ch 2
    #: - [UDP] Nokia Announcement ch 2
    nokia_ann_ch2 = 3406, 'nokia-ann-ch2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LDAP admin server port
    #: - [UDP] LDAP admin server port
    ldap_admin = 3407, 'ldap-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BES Api Port
    #: - [UDP] BES Api Port
    besapi = 3408, 'besapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetworkLens Event Port
    #: - [UDP] NetworkLens Event Port
    networklens = 3409, 'networklens', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetworkLens SSL Event
    #: - [UDP] NetworkLens SSL Event
    networklenss = 3410, 'networklenss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BioLink Authenteon server
    #: - [UDP] BioLink Authenteon server
    biolink_auth = 3411, 'biolink-auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xmlBlaster
    #: - [UDP] xmlBlaster
    xmlblaster = 3412, 'xmlblaster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SpecView Networking
    #: - [UDP] SpecView Networking
    svnet = 3413, 'svnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BroadCloud WIP Port
    #: - [UDP] BroadCloud WIP Port
    wip_port = 3414, 'wip-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BCI Name Service
    #: - [UDP] BCI Name Service
    bcinameservice = 3415, 'bcinameservice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AirMobile IS Command Port
    #: - [UDP] AirMobile IS Command Port
    commandport = 3416, 'commandport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ConServR file translation
    #: - [UDP] ConServR file translation
    csvr = 3417, 'csvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote nmap
    #: - [UDP] Remote nmap
    rnmap = 3418, 'rnmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Isogon SoftAudit
    #: - [UDP] ISogon SoftAudit
    softaudit = 3419, 'softaudit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iFCP User Port [:rfc:`4172`]
    #: - [UDP] iFCP User Port [:rfc:`4172`]
    ifcp_port = 3420, 'ifcp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bull Apprise portmapper
    #: - [UDP] Bull Apprise portmapper
    bmap = 3421, 'bmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote USB System Port
    #: - [UDP] Remote USB System Port
    rusb_sys_port = 3422, 'rusb-sys-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xTrade Reliable Messaging
    #: - [UDP] xTrade Reliable Messaging
    xtrm = 3423, 'xtrm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xTrade over TLS/SSL
    #: - [UDP] xTrade over TLS/SSL
    xtrms = 3424, 'xtrms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AGPS Access Port
    #: - [UDP] AGPS Access Port
    agps_port = 3425, 'agps-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Arkivio Storage Protocol
    #: - [UDP] Arkivio Storage Protocol
    arkivio = 3426, 'arkivio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WebSphere SNMP
    #: - [UDP] WebSphere SNMP
    websphere_snmp = 3427, 'websphere-snmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 2Wire CSS
    #: - [UDP] 2Wire CSS
    twcss = 3428, 'twcss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GCSP user port
    #: - [UDP] GCSP user port
    gcsp = 3429, 'gcsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Scott Studios Dispatch
    #: - [UDP] Scott Studios Dispatch
    ssdispatch = 3430, 'ssdispatch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Active License Server Port
    #: - [UDP] Active License Server Port
    ndl_als = 3431, 'ndl-als', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Device Protocol
    #: - [UDP] Secure Device Protocol
    osdcp = 3432, 'osdcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPNET Service Management Platform
    #: - [UDP] OPNET Service Management Platform
    opnet_smp = 3433, 'opnet-smp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenCM Server
    #: - [UDP] OpenCM Server
    opencm = 3434, 'opencm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pacom Security User Port
    #: - [UDP] Pacom Security User Port
    pacom = 3435, 'pacom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GuardControl Exchange Protocol
    #: - [UDP] GuardControl Exchange Protocol
    gc_config = 3436, 'gc-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Autocue Directory Service
    #: - [UDP] Autocue Directory Service
    autocueds = 3437, 'autocueds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Spiralcraft Admin
    #: - [UDP] Spiralcraft Admin
    spiral_admin = 3438, 'spiral-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HRI Interface Port
    #: - [UDP] HRI Interface Port
    hri_port = 3439, 'hri-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net Steward Mgmt Console
    #: - [UDP] Net Steward Mgmt Console
    ans_console = 3440, 'ans-console', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OC Connect Client
    #: - [UDP] OC Connect Client
    connect_client = 3441, 'connect-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OC Connect Server
    #: - [UDP] OC Connect Server
    connect_server = 3442, 'connect-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView Network Node Manager WEB Server
    #: - [UDP] OpenView Network Node Manager WEB Server
    ov_nnm_websrv = 3443, 'ov-nnm-websrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Denali Server
    #: - [UDP] Denali Server
    denali_server = 3444, 'denali-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Media Object Network Protocol
    #: - [UDP] Media Object Network Protocol
    monp = 3445, 'monp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3Com FAX RPC port
    #: - [UDP] 3Com FAX RPC port
    AppType_3comfaxrpc = 3446, '3comfaxrpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DirectNet IM System
    #: - [UDP] DirectNet IM System
    directnet = 3447, 'directnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Discovery and Net Config
    #: - [UDP] Discovery and Net Config
    dnc_port = 3448, 'dnc-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HotU Chat
    #: - [UDP] HotU Chat
    hotu_chat = 3449, 'hotu-chat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CAStorProxy
    #: - [UDP] CAStorProxy
    castorproxy = 3450, 'castorproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASAM Services
    #: - [UDP] ASAM Services
    asam = 3451, 'asam', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SABP-Signalling Protocol
    #: - [UDP] SABP-Signalling Protocol
    sabp_signal = 3452, 'sabp-signal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PSC Update
    #: - [UDP] PSC Update
    pscupd = 3453, 'pscupd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apple Remote Access Protocol
    #: - [UDP] Apple Remote Access Protocol
    mira = 3454, 'mira', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSVP Port
    #: - [UDP] RSVP Port
    prsvp = 3455, 'prsvp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VAT default data
    #: - [UDP] VAT default data
    vat = 3456, 'vat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VAT default control
    #: - [UDP] VAT default control
    vat_control = 3457, 'vat-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D3WinOSFI
    #: - [UDP] D3WinOSFI
    d3winosfi = 3458, 'd3winosfi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TIP Integral
    #: - [UDP] TIP Integral
    integral = 3459, 'integral', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EDM Manger
    #: - [UDP] EDM Manger
    edm_manager = 3460, 'edm-manager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EDM Stager
    #: - [UDP] EDM Stager
    edm_stager = 3461, 'edm-stager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EDM STD Notify
    #: - [UDP] EDM STD Notify
    edm_std_notify = 3462, 'edm-std-notify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EDM ADM Notify
    #: - [UDP] EDM ADM Notify
    edm_adm_notify = 3463, 'edm-adm-notify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EDM MGR Sync
    #: - [UDP] EDM MGR Sync
    edm_mgr_sync = 3464, 'edm-mgr-sync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EDM MGR Cntrl
    #: - [UDP] EDM MGR Cntrl
    edm_mgr_cntrl = 3465, 'edm-mgr-cntrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WORKFLOW
    #: - [UDP] WORKFLOW
    workflow = 3466, 'workflow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RCST
    #: - [UDP] RCST
    rcst = 3467, 'rcst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TTCM Remote Controll
    #: - [UDP] TTCM Remote Controll
    ttcmremotectrl = 3468, 'ttcmremotectrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pluribus
    #: - [UDP] Pluribus
    pluribus = 3469, 'pluribus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] jt400
    #: - [UDP] jt400
    jt400 = 3470, 'jt400', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] jt400-ssl
    #: - [UDP] jt400-ssl
    jt400_ssl = 3471, 'jt400-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JAUGS N-G Remotec 1
    #: - [UDP] JAUGS N-G Remotec 1
    jaugsremotec_1 = 3472, 'jaugsremotec-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JAUGS N-G Remotec 2
    #: - [UDP] JAUGS N-G Remotec 2
    jaugsremotec_2 = 3473, 'jaugsremotec-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TSP Automation
    #: - [UDP] TSP Automation
    ttntspauto = 3474, 'ttntspauto', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Genisar Comm Port
    #: - [UDP] Genisar Comm Port
    genisar_port = 3475, 'genisar-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NVIDIA Mgmt Protocol
    #: - [UDP] NVIDIA Mgmt Protocol
    nppmp = 3476, 'nppmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eComm link port
    #: - [UDP] eComm link port
    ecomm = 3477, 'ecomm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Session Traversal Utilities for NAT (STUN) port [:rfc:`8489`]
    #: - [UDP] Session Traversal Utilities for NAT (STUN) port [:rfc:`8489`]
    stun = 3478, 'stun', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TURN over TCP [:rfc:`8656`]
    #: - [UDP] TURN over UDP [:rfc:`8656`]
    turn = 3478, 'turn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STUN Behavior Discovery over TCP [:rfc:`5780`]
    #: - [UDP] STUN Behavior Discovery over UDP [:rfc:`5780`]
    stun_behavior = 3478, 'stun-behavior', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 2Wire RPC
    #: - [UDP] 2Wire RPC
    twrpc = 3479, 'twrpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Virtual Workspace
    #: - [UDP] Secure Virtual Workspace
    plethora = 3480, 'plethora', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CleanerLive remote ctrl
    #: - [UDP] CleanerLive remote ctrl
    cleanerliverc = 3481, 'cleanerliverc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vulture Monitoring System
    #: - [UDP] Vulture Monitoring System
    vulture = 3482, 'vulture', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Slim Devices Protocol
    #: - [UDP] Slim Devices Protocol
    slim_devices = 3483, 'slim-devices', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GBS SnapTalk Protocol
    #: - [UDP] GBS SnapTalk Protocol
    gbs_stp = 3484, 'gbs-stp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CelaTalk
    #: - [UDP] CelaTalk
    celatalk = 3485, 'celatalk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IFSF Heartbeat Port
    #: - [UDP] IFSF Heartbeat Port
    ifsf_hb_port = 3486, 'ifsf-hb-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] LISA TCP Transfer Channel
    ltctcp = 3487, 'ltctcp', TransportProtocol.get('tcp')

    #: [UDP] LISA UDP Transfer Channel
    ltcudp = 3487, 'ltcudp', TransportProtocol.get('udp')

    #: - [TCP] FS Remote Host Server
    #: - [UDP] FS Remote Host Server
    fs_rh_srv = 3488, 'fs-rh-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DTP/DIA
    #: - [UDP] DTP/DIA
    dtp_dia = 3489, 'dtp-dia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Colubris Management Port
    #: - [UDP] Colubris Management Port
    colubris = 3490, 'colubris', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SWR Port
    #: - [UDP] SWR Port
    swr_port = 3491, 'swr-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TVDUM Tray Port
    #: - [UDP] TVDUM Tray Port
    tvdumtray_port = 3492, 'tvdumtray-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network UPS Tools [:rfc:`9271`]
    #: - [UDP] Network UPS Tools
    nut = 3493, 'nut', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM 3494
    #: - [UDP] IBM 3494
    ibm3494 = 3494, 'ibm3494', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] securitylayer over tcp
    #: - [UDP] securitylayer over tcp
    seclayer_tcp = 3495, 'seclayer-tcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] securitylayer over tls
    #: - [UDP] securitylayer over tls
    seclayer_tls = 3496, 'seclayer-tls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ipEther232Port
    #: - [UDP] ipEther232Port
    ipether232port = 3497, 'ipether232port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DASHPAS user port
    #: - [UDP] DASHPAS user port
    dashpas_port = 3498, 'dashpas-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SccIP Media
    #: - [UDP] SccIP Media
    sccip_media = 3499, 'sccip-media', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RTMP Port
    #: - [UDP] RTMP Port
    rtmp_port = 3500, 'rtmp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iSoft-P2P
    #: - [UDP] iSoft-P2P
    isoft_p2p = 3501, 'isoft-p2p', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avocent Install Discovery
    #: - [UDP] Avocent Install Discovery
    avinstalldisc = 3502, 'avinstalldisc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MPLS LSP-echo Port [:rfc:`8029`]
    #: - [UDP] MPLS LSP-echo Port [:rfc:`8029`]
    lsp_ping = 3503, 'lsp-ping', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IronStorm game server
    #: - [UDP] IronStorm game server
    ironstorm = 3504, 'ironstorm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CCM communications port
    #: - [UDP] CCM communications port
    ccmcomm = 3505, 'ccmcomm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 3506
    #: - [UDP] APC 3506
    apc_3506 = 3506, 'apc-3506', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nesh Broker Port
    #: - [UDP] Nesh Broker Port
    nesh_broker = 3507, 'nesh-broker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Interaction Web
    #: - [UDP] Interaction Web
    interactionweb = 3508, 'interactionweb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Token SSL Port
    #: - [UDP] Virtual Token SSL Port
    vt_ssl = 3509, 'vt-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XSS Port
    #: - [UDP] XSS Port
    xss_port = 3510, 'xss-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WebMail/2
    #: - [UDP] WebMail/2
    webmail_2 = 3511, 'webmail-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Aztec Distribution Port
    #: - [UDP] Aztec Distribution Port
    aztec = 3512, 'aztec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Adaptec Remote Protocol
    #: - [UDP] Adaptec Remote Protocol
    arcpd = 3513, 'arcpd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MUST Peer to Peer
    #: - [UDP] MUST Peer to Peer
    must_p2p = 3514, 'must-p2p', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MUST Backplane
    #: - [UDP] MUST Backplane
    must_backplane = 3515, 'must-backplane', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Smartcard Port
    #: - [UDP] Smartcard Port
    smartcard_port = 3516, 'smartcard-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IEEE 802.11 WLANs WG IAPP
    #: - [UDP] IEEE 802.11 WLANs WG IAPP
    AppType_802_11_iapp = 3517, '802-11-iapp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Artifact Message Server
    #: - [UDP] Artifact Message Server
    artifact_msg = 3518, 'artifact-msg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Netvion Messenger Port
    nvmsgd = 3519, 'nvmsgd', TransportProtocol.get('tcp')

    #: [UDP] Netvion Galileo Port
    galileo = 3519, 'galileo', TransportProtocol.get('udp')

    #: - [TCP] Netvion Galileo Log Port
    #: - [UDP] Netvion Galileo Log Port
    galileolog = 3520, 'galileolog', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Telequip Labs MC3SS
    #: - [UDP] Telequip Labs MC3SS
    mc3ss = 3521, 'mc3ss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DO over NSSocketPort
    #: - [UDP] DO over NSSocketPort
    nssocketport = 3522, 'nssocketport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Odeum Serverlink
    #: - [UDP] Odeum Serverlink
    odeumservlink = 3523, 'odeumservlink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ECM Server port
    #: - [UDP] ECM Server port
    ecmport = 3524, 'ecmport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EIS Server port
    #: - [UDP] EIS Server port
    eisport = 3525, 'eisport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] starQuiz Port
    #: - [UDP] starQuiz Port
    starquiz_port = 3526, 'starquiz-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS Backup Exec Server
    #: - [UDP] VERITAS Backup Exec Server
    beserver_msg_q = 3527, 'beserver-msg-q', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JBoss IIOP
    #: - [UDP] JBoss IIOP
    jboss_iiop = 3528, 'jboss-iiop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JBoss IIOP/SSL
    #: - [UDP] JBoss IIOP/SSL
    jboss_iiop_ssl = 3529, 'jboss-iiop-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Grid Friendly
    #: - [UDP] Grid Friendly
    gf = 3530, 'gf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Joltid
    #: - [UDP] Joltid
    joltid = 3531, 'joltid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Raven Remote Management Control
    #: - [UDP] Raven Remote Management Control
    raven_rmp = 3532, 'raven-rmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Raven Remote Management Data
    #: - [UDP] Raven Remote Management Data
    raven_rdp = 3533, 'raven-rdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] URL Daemon Port
    #: - [UDP] URL Daemon Port
    urld_port = 3534, 'urld-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS-LA
    #: - [UDP] MS-LA
    ms_la = 3535, 'ms-la', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNAC
    #: - [UDP] SNAC
    snac = 3536, 'snac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote NI-VISA port
    #: - [UDP] Remote NI-VISA port
    ni_visa_remote = 3537, 'ni-visa-remote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Directory Server
    #: - [UDP] IBM Directory Server
    ibm_diradm = 3538, 'ibm-diradm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Directory Server SSL
    #: - [UDP] IBM Directory Server SSL
    ibm_diradm_ssl = 3539, 'ibm-diradm-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PNRP User Port
    #: - [UDP] PNRP User Port
    pnrp_port = 3540, 'pnrp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VoiSpeed Port
    #: - [UDP] VoiSpeed Port
    voispeed_port = 3541, 'voispeed-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HA cluster monitor
    #: - [UDP] HA cluster monitor
    hacl_monitor = 3542, 'hacl-monitor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] qftest Lookup Port
    #: - [UDP] qftest Lookup Port
    qftest_lookup = 3543, 'qftest-lookup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Teredo Port [:rfc:`4380`]
    #: - [UDP] Teredo Port [:rfc:`4380`]
    teredo = 3544, 'teredo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CAMAC equipment
    #: - [UDP] CAMAC equipment
    camac = 3545, 'camac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_3546 = 3546, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Symantec SIM
    #: - [UDP] Symantec SIM
    symantec_sim = 3547, 'symantec-sim', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Interworld
    #: - [UDP] Interworld
    interworld = 3548, 'interworld', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tellumat MDR NMS
    #: - [UDP] Tellumat MDR NMS
    tellumat_nms = 3549, 'tellumat-nms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure SMPP
    #: - [UDP] Secure SMPP
    ssmpp = 3550, 'ssmpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apcupsd Information Port
    #: - [UDP] Apcupsd Information Port
    apcupsd = 3551, 'apcupsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TeamAgenda Server Port
    #: - [UDP] TeamAgenda Server Port
    taserver = 3552, 'taserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Red Box Recorder ADP
    #: - [UDP] Red Box Recorder ADP
    rbr_discovery = 3553, 'rbr-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quest Notification Server
    #: - [UDP] Quest Notification Server
    questnotify = 3554, 'questnotify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vipul's Razor
    #: - [UDP] Vipul's Razor
    razor = 3555, 'razor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sky Transport Protocol
    #: - [UDP] Sky Transport Protocol
    sky_transport = 3556, 'sky-transport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PersonalOS Comm Port
    #: - [UDP] PersonalOS Comm Port
    personalos_001 = 3557, 'personalos-001', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCP user port
    #: - [UDP] MCP user port
    mcp_port = 3558, 'mcp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CCTV control port
    #: - [UDP] CCTV control port
    cctv_port = 3559, 'cctv-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INIServe port
    #: - [UDP] INIServe port
    iniserve_port = 3560, 'iniserve-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC-OneKey
    #: - [UDP] BMC-OneKey
    bmc_onekey = 3561, 'bmc-onekey', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SDBProxy
    #: - [UDP] SDBProxy
    sdbproxy = 3562, 'sdbproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Watcom Debug
    #: - [UDP] Watcom Debug
    watcomdebug = 3563, 'watcomdebug', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Electromed SIM port
    #: - [UDP] Electromed SIM port
    esimport = 3564, 'esimport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] M2PA [:rfc:`4165`]
    #: - [SCTP] M2PA [:rfc:`4165`]
    m2pa = 3565, 'm2pa', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_3565 = 3565, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Quest Data Hub
    quest_data_hub = 3566, 'quest-data-hub', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3566 = 3566, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] DOF Protocol Stack
    #: - [UDP] DOF Protocol Stack
    dof_eps = 3567, 'dof-eps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DOF Secure Tunnel
    #: - [UDP] DOF Secure Tunnel
    dof_tunnel_sec = 3568, 'dof-tunnel-sec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Meinberg Control Service
    #: - [UDP] Meinberg Control Service
    mbg_ctrl = 3569, 'mbg-ctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCC Web Server Port
    #: - [UDP] MCC Web Server Port
    mccwebsvr_port = 3570, 'mccwebsvr-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MegaRAID Server Port
    #: - [UDP] MegaRAID Server Port
    megardsvr_port = 3571, 'megardsvr-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Registration Server Port
    #: - [UDP] Registration Server Port
    megaregsvrport = 3572, 'megaregsvrport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Advantage Group UPS Suite
    #: - [UDP] Advantage Group UPS Suite
    tag_ups_1 = 3573, 'tag-ups-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DMAF Server
    dmaf_server = 3574, 'dmaf-server', TransportProtocol.get('tcp')

    #: [UDP] DMAF Caster
    dmaf_caster = 3574, 'dmaf-caster', TransportProtocol.get('udp')

    #: - [TCP] Coalsere CCM Port
    #: - [UDP] Coalsere CCM Port
    ccm_port = 3575, 'ccm-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Coalsere CMC Port
    #: - [UDP] Coalsere CMC Port
    cmc_port = 3576, 'cmc-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Configuration Port
    #: - [UDP] Configuration Port
    config_port = 3577, 'config-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Port
    #: - [UDP] Data Port
    data_port = 3578, 'data-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tarantella Load Balancing
    #: - [UDP] Tarantella Load Balancing
    ttat3lb = 3579, 'ttat3lb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NATI-ServiceLocator
    #: - [UDP] NATI-ServiceLocator
    nati_svrloc = 3580, 'nati-svrloc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ascent Capture Licensing
    #: - [UDP] Ascent Capture Licensing
    kfxaclicensing = 3581, 'kfxaclicensing', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PEG PRESS Server
    #: - [UDP] PEG PRESS Server
    press = 3582, 'press', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CANEX Watch System
    #: - [UDP] CANEX Watch System
    canex_watch = 3583, 'canex-watch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] U-DBase Access Protocol
    #: - [UDP] U-DBase Access Protocol
    u_dbap = 3584, 'u-dbap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Emprise License Server
    #: - [UDP] Emprise License Server
    emprise_lls = 3585, 'emprise-lls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] License Server Console
    #: - [UDP] License Server Console
    emprise_lsc = 3586, 'emprise-lsc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Peer to Peer Grouping
    #: - [UDP] Peer to Peer Grouping
    p2pgroup = 3587, 'p2pgroup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sentinel Server
    #: - [UDP] Sentinel Server
    sentinel = 3588, 'sentinel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] isomair
    #: - [UDP] isomair
    isomair = 3589, 'isomair', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WV CSP SMS Binding
    #: - [UDP] WV CSP SMS Binding
    wv_csp_sms = 3590, 'wv-csp-sms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LOCANIS G-TRACK Server
    #: - [UDP] LOCANIS G-TRACK Server
    gtrack_server = 3591, 'gtrack-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LOCANIS G-TRACK NE Port
    #: - [UDP] LOCANIS G-TRACK NE Port
    gtrack_ne = 3592, 'gtrack-ne', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BP Model Debugger
    #: - [UDP] BP Model Debugger
    bpmd = 3593, 'bpmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MediaSpace
    #: - [UDP] MediaSpace
    mediaspace = 3594, 'mediaspace', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ShareApp
    #: - [UDP] ShareApp
    shareapp = 3595, 'shareapp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Illusion Wireless MMOG
    #: - [UDP] Illusion Wireless MMOG
    iw_mmogame = 3596, 'iw-mmogame', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A14 (AN-to-SC/MM)
    #: - [UDP] A14 (AN-to-SC/MM)
    a14 = 3597, 'a14', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A15 (AN-to-AN)
    #: - [UDP] A15 (AN-to-AN)
    a15 = 3598, 'a15', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quasar Accounting Server
    #: - [UDP] Quasar Accounting Server
    quasar_server = 3599, 'quasar-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] text relay-answer
    #: - [UDP] text relay-answer
    trap_daemon = 3600, 'trap-daemon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Visinet Gui
    #: - [UDP] Visinet Gui
    visinet_gui = 3601, 'visinet-gui', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] InfiniSwitch Mgr Client
    #: - [UDP] InfiniSwitch Mgr Client
    infiniswitchcl = 3602, 'infiniswitchcl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Integrated Rcvr Control
    #: - [UDP] Integrated Rcvr Control
    int_rcv_cntrl = 3603, 'int-rcv-cntrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC JMX Port
    #: - [UDP] BMC JMX Port
    bmc_jmx_port = 3604, 'bmc-jmx-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ComCam IO Port
    #: - [UDP] ComCam IO Port
    comcam_io = 3605, 'comcam-io', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Splitlock Server
    #: - [UDP] Splitlock Server
    splitlock = 3606, 'splitlock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Precise I3
    #: - [UDP] Precise I3
    precise_i3 = 3607, 'precise-i3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trendchip control protocol
    #: - [UDP] Trendchip control protocol
    trendchip_dcp = 3608, 'trendchip-dcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CPDI PIDAS Connection Mon
    #: - [UDP] CPDI PIDAS Connection Mon
    cpdi_pidas_cm = 3609, 'cpdi-pidas-cm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ECHONET
    #: - [UDP] ECHONET
    echonet = 3610, 'echonet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Six Degrees Port
    #: - [UDP] Six Degrees Port
    six_degrees = 3611, 'six-degrees', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Micro Focus Data Protector
    #: - [UDP] Micro Focus Data Protector
    dataprotector = 3612, 'dataprotector', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alaris Device Discovery
    #: - [UDP] Alaris Device Discovery
    alaris_disc = 3613, 'alaris-disc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Satchwell Sigma
    #: - [UDP] Satchwell Sigma
    sigma_port = 3614, 'sigma-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Start Messaging Network
    #: - [UDP] Start Messaging Network
    start_network = 3615, 'start-network', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] cd3o Control Protocol
    #: - [UDP] cd3o Control Protocol
    cd3o_protocol = 3616, 'cd3o-protocol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ATI SHARP Logic Engine
    #: - [UDP] ATI SHARP Logic Engine
    sharp_server = 3617, 'sharp-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AAIR-Network 1
    #: - [UDP] AAIR-Network 1
    aairnet_1 = 3618, 'aairnet-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AAIR-Network 2
    #: - [UDP] AAIR-Network 2
    aairnet_2 = 3619, 'aairnet-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPSON Projector Control Port
    #: - [UDP] EPSON Projector Control Port
    ep_pcp = 3620, 'ep-pcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPSON Network Screen Port
    #: - [UDP] EPSON Network Screen Port
    ep_nsp = 3621, 'ep-nsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FF LAN Redundancy Port
    #: - [UDP] FF LAN Redundancy Port
    ff_lr_port = 3622, 'ff-lr-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HAIPIS Dynamic Discovery
    #: - [UDP] HAIPIS Dynamic Discovery
    haipe_discover = 3623, 'haipe-discover', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Distributed Upgrade Port
    #: - [UDP] Distributed Upgrade Port
    dist_upgrade = 3624, 'dist-upgrade', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Volley
    #: - [UDP] Volley
    volley = 3625, 'volley', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bvControl Daemon
    #: - [UDP] bvControl Daemon
    bvcdaemon_port = 3626, 'bvcdaemon-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jam Server Port
    #: - [UDP] Jam Server Port
    jamserverport = 3627, 'jamserverport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPT Machine Interface
    #: - [UDP] EPT Machine Interface
    ept_machine = 3628, 'ept-machine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESC/VP.net
    #: - [UDP] ESC/VP.net
    escvpnet = 3629, 'escvpnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] C&S Remote Database Port
    #: - [UDP] C&S Remote Database Port
    cs_remote_db = 3630, 'cs-remote-db', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] C&S Web Services Port
    #: - [UDP] C&S Web Services Port
    cs_services = 3631, 'cs-services', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] distributed compiler
    #: - [UDP] distributed compiler
    distcc = 3632, 'distcc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Wyrnix AIS port
    #: - [UDP] Wyrnix AIS port
    wacp = 3633, 'wacp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hNTSP Library Manager
    #: - [UDP] hNTSP Library Manager
    hlibmgr = 3634, 'hlibmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Distributed Objects
    #: - [UDP] Simple Distributed Objects
    sdo = 3635, 'sdo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SerVistaITSM
    #: - [UDP] SerVistaITSM
    servistaitsm = 3636, 'servistaitsm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Customer Service Port
    #: - [UDP] Customer Service Port
    scservp = 3637, 'scservp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EHP Backup Protocol
    #: - [UDP] EHP Backup Protocol
    ehp_backup = 3638, 'ehp-backup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Extensible Automation
    #: - [UDP] Extensible Automation
    xap_ha = 3639, 'xap-ha', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netplay Port 1
    #: - [UDP] Netplay Port 1
    netplay_port1 = 3640, 'netplay-port1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netplay Port 2
    #: - [UDP] Netplay Port 2
    netplay_port2 = 3641, 'netplay-port2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Juxml Replication port
    #: - [UDP] Juxml Replication port
    juxml_port = 3642, 'juxml-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AudioJuggler
    #: - [UDP] AudioJuggler
    audiojuggler = 3643, 'audiojuggler', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ssowatch
    #: - [UDP] ssowatch
    ssowatch = 3644, 'ssowatch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cyc
    #: - [UDP] Cyc
    cyc = 3645, 'cyc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XSS Server Port
    #: - [UDP] XSS Server Port
    xss_srv_port = 3646, 'xss-srv-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Splitlock Gateway
    #: - [UDP] Splitlock Gateway
    splitlock_gw = 3647, 'splitlock-gw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fujitsu Cooperation Port
    #: - [UDP] Fujitsu Cooperation Port
    fjcp = 3648, 'fjcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nishioka Miyuki Msg Protocol
    #: - [UDP] Nishioka Miyuki Msg Protocol
    nmmp = 3649, 'nmmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PRISMIQ VOD plug-in
    #: - [UDP] PRISMIQ VOD plug-in
    prismiq_plugin = 3650, 'prismiq-plugin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XRPC Registry
    #: - [UDP] XRPC Registry
    xrpc_registry = 3651, 'xrpc-registry', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VxCR NBU Default Port
    #: - [UDP] VxCR NBU Default Port
    vxcrnbuport = 3652, 'vxcrnbuport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tunnel Setup Protocol [:rfc:`5572`]
    #: - [UDP] Tunnel Setup Protocol [:rfc:`5572`]
    tsp = 3653, 'tsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VAP RealTime Messenger
    #: - [UDP] VAP RealTime Messenger
    vaprtm = 3654, 'vaprtm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ActiveBatch Exec Agent
    #: - [UDP] ActiveBatch Exec Agent
    abatemgr = 3655, 'abatemgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ActiveBatch Job Scheduler
    #: - [UDP] ActiveBatch Job Scheduler
    abatjss = 3656, 'abatjss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ImmediaNet Beacon
    #: - [UDP] ImmediaNet Beacon
    immedianet_bcn = 3657, 'immedianet-bcn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PlayStation AMS (Secure)
    #: - [UDP] PlayStation AMS (Secure)
    ps_ams = 3658, 'ps-ams', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apple SASL
    #: - [UDP] Apple SASL
    apple_sasl = 3659, 'apple-sasl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Tivoli Directory Service using SSL
    #: - [UDP] IBM Tivoli Directory Service using SSL
    can_nds_ssl = 3660, 'can-nds-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Tivoli Directory Service using SSL
    #: - [UDP] IBM Tivoli Directory Service using SSL
    can_ferret_ssl = 3661, 'can-ferret-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pserver
    #: - [UDP] pserver
    pserver = 3662, 'pserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DIRECWAY Tunnel Protocol
    #: - [UDP] DIRECWAY Tunnel Protocol
    dtp = 3663, 'dtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UPS Engine Port
    #: - [UDP] UPS Engine Port
    ups_engine = 3664, 'ups-engine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Enterprise Engine Port
    #: - [UDP] Enterprise Engine Port
    ent_engine = 3665, 'ent-engine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM eServer PAP
    #: - [UDP] IBM EServer PAP
    eserver_pap = 3666, 'eserver-pap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Information Exchange
    #: - [UDP] IBM Information Exchange
    infoexch = 3667, 'infoexch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dell Remote Management
    #: - [UDP] Dell Remote Management
    dell_rm_port = 3668, 'dell-rm-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA SAN Switch Management
    #: - [UDP] CA SAN Switch Management
    casanswmgmt = 3669, 'casanswmgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMILE TCP/UDP Interface
    #: - [UDP] SMILE TCP/UDP Interface
    smile = 3670, 'smile', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] e Field Control (EIBnet)
    #: - [UDP] e Field Control (EIBnet)
    efcp = 3671, 'efcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LispWorks ORB
    #: - [UDP] LispWorks ORB
    lispworks_orb = 3672, 'lispworks-orb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Openview Media Vault GUI
    #: - [UDP] Openview Media Vault GUI
    mediavault_gui = 3673, 'mediavault-gui', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WinINSTALL IPC Port
    #: - [UDP] WinINSTALL IPC Port
    wininstall_ipc = 3674, 'wininstall-ipc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CallTrax Data Port
    #: - [UDP] CallTrax Data Port
    calltrax = 3675, 'calltrax', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VisualAge Pacbase server
    #: - [UDP] VisualAge Pacbase server
    va_pacbase = 3676, 'va-pacbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RoverLog IPC
    #: - [UDP] RoverLog IPC
    roverlog = 3677, 'roverlog', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DataGuardianLT
    #: - [UDP] DataGuardianLT
    ipr_dglt = 3678, 'ipr-dglt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Newton Dock
    #: - [UDP] Newton Dock
    escale = 3679, 'escale (newton dock)', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NPDS Tracker
    #: - [UDP] NPDS Tracker
    npds_tracker = 3680, 'npds-tracker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BTS X73 Port
    #: - [UDP] BTS X73 Port
    bts_x73 = 3681, 'bts-x73', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMC SmartPackets-MAPI
    #: - [UDP] EMC SmartPackets-MAPI
    cas_mapi = 3682, 'cas-mapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC EDV/EA
    #: - [UDP] BMC EDV/EA
    bmc_ea = 3683, 'bmc-ea', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FAXstfX
    #: - [UDP] FAXstfX
    faxstfx_port = 3684, 'faxstfx-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DS Expert Agent
    #: - [UDP] DS Expert Agent
    dsx_agent = 3685, 'dsx-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trivial Network Management
    #: - [UDP] Trivial Network Management
    tnmpv2 = 3686, 'tnmpv2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] simple-push
    #: - [UDP] simple-push
    simple_push = 3687, 'simple-push', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] simple-push Secure
    #: - [UDP] simple-push Secure
    simple_push_s = 3688, 'simple-push-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Digital Audio Access Protocol (iTunes)
    #: - [UDP] Digital Audio Access Protocol (iTunes)
    daap = 3689, 'daap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Magaya Network Port
    #: - [UDP] Magaya Network Port
    magaya_network = 3691, 'magaya-network', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Brimstone IntelSync
    #: - [UDP] Brimstone IntelSync
    intelsync = 3692, 'intelsync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Emergency Automatic Structure Lockdown System
    easl = 3693, 'easl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3693 = 3693, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_3694 = 3694, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] BMC Data Collection
    #: - [UDP] BMC Data Collection
    bmc_data_coll = 3695, 'bmc-data-coll', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Telnet Com Port Control
    #: - [UDP] Telnet Com Port Control
    telnetcpcd = 3696, 'telnetcpcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NavisWorks License System
    #: - [UDP] NavisWorks License System
    nw_license = 3697, 'nw-license', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SAGECTLPANEL
    #: - [UDP] SAGECTLPANEL
    sagectlpanel = 3698, 'sagectlpanel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Internet Call Waiting
    #: - [UDP] Internet Call Waiting
    kpn_icw = 3699, 'kpn-icw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LRS NetPage
    #: - [UDP] LRS NetPage
    lrs_paging = 3700, 'lrs-paging', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetCelera
    #: - [UDP] NetCelera
    netcelera = 3701, 'netcelera', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Web Service Discovery
    #: - [UDP] Web Service Discovery
    ws_discovery = 3702, 'ws-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Adobe Server 3
    #: - [UDP] Adobe Server 3
    adobeserver_3 = 3703, 'adobeserver-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Adobe Server 4
    #: - [UDP] Adobe Server 4
    adobeserver_4 = 3704, 'adobeserver-4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Adobe Server 5
    #: - [UDP] Adobe Server 5
    adobeserver_5 = 3705, 'adobeserver-5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Real-Time Event Port
    #: - [UDP] Real-Time Event Port
    rt_event = 3706, 'rt-event', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Real-Time Event Secure Port
    #: - [UDP] Real-Time Event Secure Port
    rt_event_s = 3707, 'rt-event-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun App Svr - Naming
    #: - [UDP] Sun App Svr - Naming
    sun_as_iiops = 3708, 'sun-as-iiops', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA-IDMS Server
    #: - [UDP] CA-IDMS Server
    ca_idms = 3709, 'ca-idms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PortGate Authentication
    #: - [UDP] PortGate Authentication
    portgate_auth = 3710, 'portgate-auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EBD Server 2
    #: - [UDP] EBD Server 2
    edb_server2 = 3711, 'edb-server2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sentinel Enterprise
    #: - [UDP] Sentinel Enterprise
    sentinel_ent = 3712, 'sentinel-ent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TFTP over TLS
    #: - [UDP] TFTP over TLS
    tftps = 3713, 'tftps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DELOS Direct Messaging
    #: - [UDP] DELOS Direct Messaging
    delos_dms = 3714, 'delos-dms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Anoto Rendezvous Port
    #: - [UDP] Anoto Rendezvous Port
    anoto_rendezv = 3715, 'anoto-rendezv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WV CSP SMS CIR Channel
    #: - [UDP] WV CSP SMS CIR Channel
    wv_csp_sms_cir = 3716, 'wv-csp-sms-cir', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WV CSP UDP/IP CIR Channel
    #: - [UDP] WV CSP UDP/IP CIR Channel
    wv_csp_udp_cir = 3717, 'wv-csp-udp-cir', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPUS Server Port
    #: - [UDP] OPUS Server Port
    opus_services = 3718, 'opus-services', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iTel Server Port
    #: - [UDP] iTel Server Port
    itelserverport = 3719, 'itelserverport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UF Astro. Instr. Services
    #: - [UDP] UF Astro. Instr. Services
    ufastro_instr = 3720, 'ufastro-instr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xsync
    #: - [UDP] Xsync
    xsync = 3721, 'xsync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xserve RAID
    #: - [UDP] Xserve RAID
    xserveraid = 3722, 'xserveraid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sychron Service Daemon
    #: - [UDP] Sychron Service Daemon
    sychrond = 3723, 'sychrond', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] World of Warcraft
    #: - [UDP] World of Warcraft
    blizwow = 3724, 'blizwow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netia NA-ER Port
    #: - [UDP] Netia NA-ER Port
    na_er_tip = 3725, 'na-er-tip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xyratex Array Manager
    #: - [UDP] Xyratex Array Manager
    array_manager = 3726, 'array-manager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ericsson Mobile Data Unit
    #: - [UDP] Ericsson Mobile Data Unit
    e_mdu = 3727, 'e-mdu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ericsson Web on Air
    #: - [UDP] Ericsson Web on Air
    e_woa = 3728, 'e-woa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fireking Audit Port
    #: - [UDP] Fireking Audit Port
    fksp_audit = 3729, 'fksp-audit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Client Control
    #: - [UDP] Client Control
    client_ctrl = 3730, 'client-ctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Manager
    #: - [UDP] Service Manager
    smap = 3731, 'smap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mobile Wnn
    #: - [UDP] Mobile Wnn
    m_wnn = 3732, 'm-wnn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multipuesto Msg Port
    #: - [UDP] Multipuesto Msg Port
    multip_msg = 3733, 'multip-msg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Synel Data Collection Port
    #: - [UDP] Synel Data Collection Port
    synel_data = 3734, 'synel-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Password Distribution
    #: - [UDP] Password Distribution
    pwdis = 3735, 'pwdis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RealSpace RMI
    #: - [UDP] RealSpace RMI
    rs_rmi = 3736, 'rs-rmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] XPanel Daemon
    xpanel = 3737, 'xpanel', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3737 = 3737, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] versaTalk Server Port
    #: - [UDP] versaTalk Server Port
    versatalk = 3738, 'versatalk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Launchbird LicenseManager
    #: - [UDP] Launchbird LicenseManager
    launchbird_lm = 3739, 'launchbird-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Heartbeat Protocol
    #: - [UDP] Heartbeat Protocol
    heartbeat = 3740, 'heartbeat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WysDM Agent
    #: - [UDP] WysDM Agent
    wysdma = 3741, 'wysdma', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CST - Configuration & Service Tracker
    #: - [UDP] CST - Configuration & Service Tracker
    cst_port = 3742, 'cst-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IP Control Systems Ltd.
    #: - [UDP] IP Control Systems Ltd.
    ipcs_command = 3743, 'ipcs-command', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SASG
    #: - [UDP] SASG
    sasg = 3744, 'sasg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GWRTC Call Port
    #: - [UDP] GWRTC Call Port
    gw_call_port = 3745, 'gw-call-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LXPRO.COM LinkTest
    #: - [UDP] LXPRO.COM LinkTest
    linktest = 3746, 'linktest', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LXPRO.COM LinkTest SSL
    #: - [UDP] LXPRO.COM LinkTest SSL
    linktest_s = 3747, 'linktest-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] webData
    #: - [UDP] webData
    webdata = 3748, 'webdata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CimTrak
    #: - [UDP] CimTrak
    cimtrak = 3749, 'cimtrak', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CBOS/IP ncapsalation port
    #: - [UDP] CBOS/IP ncapsalation port
    cbos_ip_port = 3750, 'cbos-ip-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CommLinx GPRS Cube
    #: - [UDP] CommLinx GPRS Cube
    gprs_cube = 3751, 'gprs-cube', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vigil-IP RemoteAgent
    #: - [UDP] Vigil-IP RemoteAgent
    vipremoteagent = 3752, 'vipremoteagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NattyServer Port
    #: - [UDP] NattyServer Port
    nattyserver = 3753, 'nattyserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TimesTen Broker Port
    #: - [UDP] TimesTen Broker Port
    timestenbroker = 3754, 'timestenbroker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SAS Remote Help Server
    #: - [UDP] SAS Remote Help Server
    sas_remote_hlp = 3755, 'sas-remote-hlp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Canon CAPT Port
    #: - [UDP] Canon CAPT Port
    canon_capt = 3756, 'canon-capt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GRF Server Port
    #: - [UDP] GRF Server Port
    grf_port = 3757, 'grf-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] apw RMI registry
    #: - [UDP] apw RMI registry
    apw_registry = 3758, 'apw-registry', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Exapt License Manager
    #: - [UDP] Exapt License Manager
    exapt_lmgr = 3759, 'exapt-lmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] adTempus Client
    #: - [UDP] adTEmpus Client
    adtempusclient = 3760, 'adtempusclient', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gsakmp port [:rfc:`4535`]
    #: - [UDP] gsakmp port [:rfc:`4535`]
    gsakmp = 3761, 'gsakmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GBS SnapMail Protocol
    #: - [UDP] GBS SnapMail Protocol
    gbs_smp = 3762, 'gbs-smp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XO Wave Control Port
    #: - [UDP] XO Wave Control Port
    xo_wave = 3763, 'xo-wave', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MNI Protected Routing
    #: - [UDP] MNI Protected Routing
    mni_prot_rout = 3764, 'mni-prot-rout', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Traceroute
    #: - [UDP] Remote Traceroute
    rtraceroute = 3765, 'rtraceroute', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SSL e-watch sitewatch server
    sitewatch_s = 3766, 'sitewatch-s', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3766 = 3766, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ListMGR Port
    #: - [UDP] ListMGR Port
    listmgr_port = 3767, 'listmgr-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rblcheckd server daemon
    #: - [UDP] rblcheckd server daemon
    rblcheckd = 3768, 'rblcheckd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HAIPE Network Keying
    #: - [UDP] HAIPE Network Keying
    haipe_otnk = 3769, 'haipe-otnk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cinderella Collaboration
    #: - [UDP] Cinderella Collaboration
    cindycollab = 3770, 'cindycollab', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RTP Paging Port
    #: - [UDP] RTP Paging Port
    paging_port = 3771, 'paging-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Chantry Tunnel Protocol
    #: - [UDP] Chantry Tunnel Protocol
    ctp = 3772, 'ctp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ctdhercules
    #: - [UDP] ctdhercules
    ctdhercules = 3773, 'ctdhercules', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ZICOM
    #: - [UDP] ZICOM
    zicom = 3774, 'zicom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISPM Manager Port
    #: - [UDP] ISPM Manager Port
    ispmmgr = 3775, 'ispmmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Device Provisioning Port
    #: - [UDP] Device Provisioning Port
    dvcprov_port = 3776, 'dvcprov-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jibe EdgeBurst
    #: - [UDP] Jibe EdgeBurst
    jibe_eb = 3777, 'jibe-eb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cutler-Hammer IT Port
    #: - [UDP] Cutler-Hammer IT Port
    c_h_it_port = 3778, 'c-h-it-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cognima Replication
    #: - [UDP] Cognima Replication
    cognima = 3779, 'cognima', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nuzzler Network Protocol
    #: - [UDP] Nuzzler Network Protocol
    nnp = 3780, 'nnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ABCvoice server port
    #: - [UDP] ABCvoice server port
    abcvoice_port = 3781, 'abcvoice-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure ISO TP0 port
    #: - [UDP] Secure ISO TP0 port
    iso_tp0s = 3782, 'iso-tp0s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Impact Mgr./PEM Gateway
    #: - [UDP] Impact Mgr./PEM Gateway
    bim_pem = 3783, 'bim-pem', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BFD Control Protocol [:rfc:`5881`]
    #: - [UDP] BFD Control Protocol [:rfc:`5881`]
    bfd_control = 3784, 'bfd-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BFD Echo Protocol [:rfc:`5881`]
    #: - [UDP] BFD Echo Protocol [:rfc:`5881`]
    bfd_echo = 3785, 'bfd-echo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSW Upstrigger port
    #: - [UDP] VSW Upstrigger port
    upstriggervsw = 3786, 'upstriggervsw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fintrx
    #: - [UDP] Fintrx
    fintrx = 3787, 'fintrx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPACEWAY Routing port
    #: - [UDP] SPACEWAY Routing port
    isrp_port = 3788, 'isrp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RemoteDeploy Administration Port [July 2003]
    #: - [UDP] RemoteDeploy Administration Port [July 2003]
    remotedeploy = 3789, 'remotedeploy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QuickBooks RDS
    #: - [UDP] QuickBooks RDS
    quickbooksrds = 3790, 'quickbooksrds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TV NetworkVideo Data port
    #: - [UDP] TV NetworkVideo Data port
    tvnetworkvideo = 3791, 'tvnetworkvideo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] e-Watch Corporation SiteWatch
    #: - [UDP] e-Watch Corporation SiteWatch
    sitewatch = 3792, 'sitewatch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DataCore Software
    #: - [UDP] DataCore Software
    dcsoftware = 3793, 'dcsoftware', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JAUS Robots
    #: - [UDP] JAUS Robots
    jaus = 3794, 'jaus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] myBLAST Mekentosj port
    #: - [UDP] myBLAST Mekentosj port
    myblast = 3795, 'myblast', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Spaceway Dialer
    #: - [UDP] Spaceway Dialer
    spw_dialer = 3796, 'spw-dialer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] idps
    #: - [UDP] idps
    idps = 3797, 'idps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Minilock
    #: - [UDP] Minilock
    minilock = 3798, 'minilock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RADIUS Dynamic Authorization [:rfc:`3576`]
    #: - [UDP] RADIUS Dynamic Authorization [:rfc:`3576`]
    radius_dynauth = 3799, 'radius-dynauth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Print Services Interface
    #: - [UDP] Print Services Interface
    pwgpsi = 3800, 'pwgpsi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ibm manager service
    #: - [UDP] ibm manager service
    ibm_mgr = 3801, 'ibm-mgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VHD
    #: - [UDP] VHD
    vhd = 3802, 'vhd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SoniqSync
    #: - [UDP] SoniqSync
    soniqsync = 3803, 'soniqsync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Harman IQNet Port
    #: - [UDP] Harman IQNet Port
    iqnet_port = 3804, 'iqnet-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ThorGuard Server Port
    #: - [UDP] ThorGuard Server Port
    tcpdataserver = 3805, 'tcpdataserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote System Manager
    #: - [UDP] Remote System Manager
    wsmlb = 3806, 'wsmlb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SpuGNA Communication Port
    #: - [UDP] SpuGNA Communication Port
    spugna = 3807, 'spugna', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun App Svr-IIOPClntAuth
    #: - [UDP] Sun App Svr-IIOPClntAuth
    sun_as_iiops_ca = 3808, 'sun-as-iiops-ca', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Java Desktop System Configuration Agent
    #: - [UDP] Java Desktop System Configuration Agent
    apocd = 3809, 'apocd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WLAN AS server
    #: - [UDP] WLAN AS server
    wlanauth = 3810, 'wlanauth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMP
    #: - [UDP] AMP
    amp = 3811, 'amp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netO WOL Server
    #: - [UDP] netO WOL Server
    neto_wol_server = 3812, 'neto-wol-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rhapsody Interface Protocol
    #: - [UDP] Rhapsody Interface Protocol
    rap_ip = 3813, 'rap-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] netO DCS
    #: - [UDP] netO DCS
    neto_dcs = 3814, 'neto-dcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LANsurveyor XML
    #: - [UDP] LANsurveyor XML
    lansurveyorxml = 3815, 'lansurveyorxml', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun Local Patch Server
    #: - [UDP] Sun Local Patch Server
    sunlps_http = 3816, 'sunlps-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Yosemite Tech Tapeware
    #: - [UDP] Yosemite Tech Tapeware
    tapeware = 3817, 'tapeware', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Crinis Heartbeat
    #: - [UDP] Crinis Heartbeat
    crinis_hb = 3818, 'crinis-hb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPL Sequ Layer Protocol
    #: - [UDP] EPL Sequ Layer Protocol
    epl_slp = 3819, 'epl-slp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Siemens AuD SCP
    #: - [UDP] Siemens AuD SCP
    scp = 3820, 'scp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ATSC PMCP Standard
    #: - [UDP] ATSC PMCP Standard
    pmcp = 3821, 'pmcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compute Pool Discovery
    #: - [UDP] Compute Pool Discovery
    acp_discovery = 3822, 'acp-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compute Pool Conduit
    #: - [UDP] Compute Pool Conduit
    acp_conduit = 3823, 'acp-conduit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Compute Pool Policy
    #: - [UDP] Compute Pool Policy
    acp_policy = 3824, 'acp-policy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Antera FlowFusion Process Simulation
    #: - [UDP] Antera FlowFusion Process Simulation
    ffserver = 3825, 'ffserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WarMUX game server
    #: - [UDP] WarMUX game server
    warmux = 3826, 'warmux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netadmin Systems MPI service
    #: - [UDP] Netadmin Systems MPI service
    netmpi = 3827, 'netmpi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netadmin Systems Event Handler
    #: - [UDP] Netadmin Systems Event Handler
    neteh = 3828, 'neteh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netadmin Systems Event Handler External
    #: - [UDP] Netadmin Systems Event Handler External
    neteh_ext = 3829, 'neteh-ext', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cerner System Management Agent
    #: - [UDP] Cerner System Management Agent
    cernsysmgmtagt = 3830, 'cernsysmgmtagt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Docsvault Application Service
    #: - [UDP] Docsvault Application Service
    dvapps = 3831, 'dvapps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xxNETserver
    #: - [UDP] xxNETserver
    xxnetserver = 3832, 'xxnetserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AIPN LS Authentication
    #: - [UDP] AIPN LS Authentication
    aipn_auth = 3833, 'aipn-auth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Spectar Data Stream Service
    #: - [UDP] Spectar Data Stream Service
    spectardata = 3834, 'spectardata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Spectar Database Rights Service
    #: - [UDP] Spectar Database Rights Service
    spectardb = 3835, 'spectardb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MARKEM NEXTGEN DCP
    #: - [UDP] MARKEM NEXTGEN DCP
    markem_dcp = 3836, 'markem-dcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MARKEM Auto-Discovery
    #: - [UDP] MARKEM Auto-Discovery
    mkm_discovery = 3837, 'mkm-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Scito Object Server
    #: - [UDP] Scito Object Server
    sos = 3838, 'sos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMX Resource Management Suite
    #: - [UDP] AMX Resource Management Suite
    amx_rms = 3839, 'amx-rms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] www.FlirtMitMir.de
    #: - [UDP] www.FlirtMitMir.de
    flirtmitmir = 3840, 'flirtmitmir', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ShipRush Database Server
    shiprush_db_svr = 3841, 'shiprush-db-svr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3841 = 3841, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] NHCI status port
    #: - [UDP] NHCI status port
    nhci = 3842, 'nhci', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quest Common Agent
    #: - [UDP] Quest Common Agent
    quest_agent = 3843, 'quest-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RNM
    #: - [UDP] RNM
    rnm = 3844, 'rnm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] V-ONE Single Port Proxy
    #: - [UDP] V-ONE Single Port Proxy
    v_one_spp = 3845, 'v-one-spp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Astare Network PCP
    #: - [UDP] Astare Network PCP
    an_pcp = 3846, 'an-pcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS Firewall Control
    #: - [UDP] MS Firewall Control
    msfw_control = 3847, 'msfw-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IT Environmental Monitor
    #: - [UDP] IT Environmental Monitor
    item = 3848, 'item', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPACEWAY DNS Preload
    #: - [UDP] SPACEWAY DNS Preload
    spw_dnspreload = 3849, 'spw-dnspreload', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QTMS Bootstrap Protocol
    #: - [UDP] QTMS Bootstrap Protocol
    qtms_bootstrap = 3850, 'qtms-bootstrap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SpectraTalk Port
    #: - [UDP] SpectraTalk Port
    spectraport = 3851, 'spectraport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSE App Configuration
    #: - [UDP] SSE App Configuration
    sse_app_config = 3852, 'sse-app-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SONY scanning protocol
    #: - [UDP] SONY scanning protocol
    sscan = 3853, 'sscan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Stryker Comm Port
    #: - [UDP] Stryker Comm Port
    stryker_com = 3854, 'stryker-com', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenTRAC
    #: - [UDP] OpenTRAC
    opentrac = 3855, 'opentrac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INFORMER
    #: - [UDP] INFORMER
    informer = 3856, 'informer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trap Port
    #: - [UDP] Trap Port
    trap_port = 3857, 'trap-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trap Port MOM
    #: - [UDP] Trap Port MOM
    trap_port_mom = 3858, 'trap-port-mom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Navini Port
    #: - [UDP] Navini Port
    nav_port = 3859, 'nav-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Server/Application State Protocol (SASP)
    #: - [UDP] Server/Application State Protocol (SASP)
    sasp = 3860, 'sasp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] winShadow Host Discovery
    #: - [UDP] winShadow Host Discovery
    winshadow_hd = 3861, 'winshadow-hd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GIGA-POCKET
    #: - [UDP] GIGA-POCKET
    giga_pocket = 3862, 'giga-pocket', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] asap tcp port
    asap_tcp = 3863, 'asap-tcp', TransportProtocol.get('tcp')

    #: [UDP] asap udp port [:rfc:`5352`]
    asap_udp = 3863, 'asap-udp', TransportProtocol.get('udp')

    #: [SCTP] asap sctp [:rfc:`5352`]
    asap_sctp = 3863, 'asap-sctp', TransportProtocol.get('sctp')

    #: [TCP] asap/tls tcp port [:rfc:`5352`]
    asap_tcp_tls = 3864, 'asap-tcp-tls', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_3864 = 3864, 'reserved', TransportProtocol.get('udp')

    #: [SCTP] asap-sctp/tls [:rfc:`5352`]
    asap_sctp_tls = 3864, 'asap-sctp-tls', TransportProtocol.get('sctp')

    #: - [TCP] xpl automation protocol
    #: - [UDP] xpl automation protocol
    xpl = 3865, 'xpl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun SDViz DZDAEMON Port
    #: - [UDP] Sun SDViz DZDAEMON Port
    dzdaemon = 3866, 'dzdaemon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun SDViz DZOGLSERVER Port
    #: - [UDP] Sun SDViz DZOGLSERVER Port
    dzoglserver = 3867, 'dzoglserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DIAMETER
    #: - [SCTP] DIAMETER [:rfc:`3588`]
    diameter = 3868, 'diameter', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_3868 = 3868, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] hp OVSAM MgmtServer Disco
    #: - [UDP] hp OVSAM MgmtServer Disco
    ovsam_mgmt = 3869, 'ovsam-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp OVSAM HostAgent Disco
    #: - [UDP] hp OVSAM HostAgent Disco
    ovsam_d_agent = 3870, 'ovsam-d-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avocent DS Authorization
    #: - [UDP] Avocent DS Authorization
    avocent_adsap = 3871, 'avocent-adsap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OEM Agent
    #: - [UDP] OEM Agent
    oem_agent = 3872, 'oem-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fagordnc
    #: - [UDP] fagordnc
    fagordnc = 3873, 'fagordnc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SixXS Configuration
    #: - [UDP] SixXS Configuration
    sixxsconfig = 3874, 'sixxsconfig', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PNBSCADA
    #: - [UDP] PNBSCADA
    pnbscada = 3875, 'pnbscada', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DirectoryLockdown Agent IANA assigned this well-formed service name as
    #: a replacement for "dl_agent".
    #: - [TCP] DirectoryLockdown Agent
    #: - [UDP] DirectoryLockdown Agent IANA assigned this well-formed service name as
    #: a replacement for "dl_agent".
    #: - [UDP] DirectoryLockdown Agent
    dl_agent = 3876, 'dl-agent', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] XMPCR Interface Port
    #: - [UDP] XMPCR Interface Port
    xmpcr_interface = 3877, 'xmpcr-interface', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FotoG CAD interface
    #: - [UDP] FotoG CAD interface
    fotogcad = 3878, 'fotogcad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] appss license manager
    #: - [UDP] appss license manager
    appss_lm = 3879, 'appss-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IGRS
    #: - [UDP] IGRS
    igrs = 3880, 'igrs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Acquisition and Control
    #: - [UDP] Data Acquisition and Control
    idac = 3881, 'idac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DTS Service Port
    #: - [UDP] DTS Service Port
    msdts1 = 3882, 'msdts1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VR Peripheral Network
    #: - [UDP] VR Peripheral Network
    vrpn = 3883, 'vrpn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SofTrack Metering
    #: - [UDP] SofTrack Metering
    softrack_meter = 3884, 'softrack-meter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TopFlow SSL
    #: - [UDP] TopFlow SSL
    topflow_ssl = 3885, 'topflow-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NEI management port
    #: - [UDP] NEI management port
    nei_management = 3886, 'nei-management', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ciphire Data Transport
    #: - [UDP] Ciphire Data Transport
    ciphire_data = 3887, 'ciphire-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ciphire Services
    #: - [UDP] Ciphire Services
    ciphire_serv = 3888, 'ciphire-serv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D and V Tester Control Port
    #: - [UDP] D and V Tester Control Port
    dandv_tester = 3889, 'dandv-tester', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Niche Data Server Connect
    #: - [UDP] Niche Data Server Connect
    ndsconnect = 3890, 'ndsconnect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle RTC-PM port
    #: - [UDP] Oracle RTC-PM port
    rtc_pm_port = 3891, 'rtc-pm-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCC-image-port
    #: - [UDP] PCC-image-port
    pcc_image_port = 3892, 'pcc-image-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CGI StarAPI Server
    #: - [UDP] CGI StarAPI Server
    cgi_starapi = 3893, 'cgi-starapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SyAM Agent Port
    #: - [UDP] SyAM Agent Port
    syam_agent = 3894, 'syam-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SyAm SMC Service Port
    #: - [UDP] SyAm SMC Service Port
    syam_smc = 3895, 'syam-smc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Distributed Objects over TLS
    #: - [UDP] Simple Distributed Objects over TLS
    sdo_tls = 3896, 'sdo-tls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Distributed Objects over SSH
    #: - [UDP] Simple Distributed Objects over SSH
    sdo_ssh = 3897, 'sdo-ssh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IAS, Inc. SmartEye NET Internet Protocol
    #: - [UDP] IAS, Inc. SmartEye NET Internet Protocol
    senip = 3898, 'senip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITV Port
    #: - [UDP] ITV Port
    itv_control = 3899, 'itv-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] udt_os IANA assigned this well-formed service name as a replacement
    #: for "udt_os".
    #: - [TCP] udt_os
    #: - [UDP] udt_os IANA assigned this well-formed service name as a replacement
    #: for "udt_os".
    #: - [UDP] udt_os
    udt_os_1382 = 1382, 'udt-os', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Unidata UDT OS IANA assigned this well-formed service name as a
    #: replacement for "udt_os".
    #: - [TCP] Unidata UDT OS
    #: - [UDP] Unidata UDT OS IANA assigned this well-formed service name as a
    #: replacement for "udt_os".
    #: - [UDP] Unidata UDT OS
    udt_os_3900 = 3900, 'udt-os', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] NIM Service Handler
    #: - [UDP] NIM Service Handler
    nimsh = 3901, 'nimsh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NIMsh Auxiliary Port
    #: - [UDP] NIMsh Auxiliary Port
    nimaux = 3902, 'nimaux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CharsetMGR
    #: - [UDP] CharsetMGR
    charsetmgr = 3903, 'charsetmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Arnet Omnilink Port
    #: - [UDP] Arnet Omnilink Port
    omnilink_port = 3904, 'omnilink-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mailbox Update (MUPDATE) protocol [:rfc:`3656`]
    #: - [UDP] Mailbox Update (MUPDATE) protocol [:rfc:`3656`]
    mupdate = 3905, 'mupdate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TopoVista elevation data
    #: - [UDP] TopoVista elevation data
    topovista_data = 3906, 'topovista-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Imoguia Port
    #: - [UDP] Imoguia Port
    imoguia_port = 3907, 'imoguia-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP Procurve NetManagement
    #: - [UDP] HP Procurve NetManagement
    hppronetman = 3908, 'hppronetman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SurfControl CPA
    #: - [UDP] SurfControl CPA
    surfcontrolcpa = 3909, 'surfcontrolcpa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Printer Request Port
    #: - [UDP] Printer Request Port
    prnrequest = 3910, 'prnrequest', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Printer Status Port
    #: - [UDP] Printer Status Port
    prnstatus = 3911, 'prnstatus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Global Maintech Stars
    #: - [UDP] Global Maintech Stars
    gbmt_stars = 3912, 'gbmt-stars', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ListCREATOR Port
    #: - [UDP] ListCREATOR Port
    listcrt_port = 3913, 'listcrt-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ListCREATOR Port 2
    #: - [UDP] ListCREATOR Port 2
    listcrt_port_2 = 3914, 'listcrt-port-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Auto-Graphics Cataloging
    #: - [UDP] Auto-Graphics Cataloging
    agcat = 3915, 'agcat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WysDM Controller
    #: - [UDP] WysDM Controller
    wysdmc = 3916, 'wysdmc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AFT multiplex port
    #: - [UDP] AFT multiplex port
    aftmux = 3917, 'aftmux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PacketCableMultimediaCOPS
    #: - [UDP] PacketCableMultimediaCOPS
    pktcablemmcops = 3918, 'pktcablemmcops', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HyperIP
    #: - [UDP] HyperIP
    hyperip = 3919, 'hyperip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Exasoft IP Port
    #: - [UDP] Exasoft IP Port
    exasoftport1 = 3920, 'exasoftport1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Herodotus Net
    #: - [UDP] Herodotus Net
    herodotus_net = 3921, 'herodotus-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Soronti Update Port
    #: - [UDP] Soronti Update Port
    sor_update = 3922, 'sor-update', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Symbian Service Broker
    #: - [UDP] Symbian Service Broker
    symb_sb_port = 3923, 'symb-sb-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MPL_GPRS_PORT
    #: - [UDP] MPL_GPRS_Port
    mpl_gprs_port = 3924, 'mpl-gprs-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zoran Media Port
    #: - [UDP] Zoran Media Port
    zmp = 3925, 'zmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WINPort
    #: - [UDP] WINPort
    winport = 3926, 'winport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ScsTsr
    #: - [UDP] ScsTsr
    natdataservice = 3927, 'natdataservice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PXE NetBoot Manager
    #: - [UDP] PXE NetBoot Manager
    netboot_pxe = 3928, 'netboot-pxe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMS Port
    #: - [UDP] AMS Port
    smauth_port = 3929, 'smauth-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Syam Web Server Port
    #: - [UDP] Syam Web Server Port
    syam_webserver = 3930, 'syam-webserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MSR Plugin Port
    #: - [UDP] MSR Plugin Port
    msr_plugin_port = 3931, 'msr-plugin-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dynamic Site System
    #: - [UDP] Dynamic Site System
    dyn_site = 3932, 'dyn-site', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PL/B App Server User Port
    #: - [UDP] PL/B App Server User Port
    plbserve_port = 3933, 'plbserve-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PL/B File Manager Port
    #: - [UDP] PL/B File Manager Port
    sunfm_port = 3934, 'sunfm-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SDP Port Mapper Protocol
    #: - [UDP] SDP Port Mapper Protocol
    sdp_portmapper = 3935, 'sdp-portmapper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mailprox
    #: - [UDP] Mailprox
    mailprox = 3936, 'mailprox', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DVB Service Discovery
    #: - [UDP] DVB Service Discovery
    dvbservdsc = 3937, 'dvbservdsc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle dbControl Agent po IANA assigned this well-formed service name
    #: as a replacement for "dbcontrol_agent".
    #: - [TCP] Oracle dbControl Agent po
    #: - [UDP] Oracle dbControl Agent po IANA assigned this well-formed service name
    #: as a replacement for "dbcontrol_agent".
    #: - [UDP] Oracle dbControl Agent po
    dbcontrol_agent = 3938, 'dbcontrol-agent', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Anti-virus Application Management Port
    #: - [UDP] Anti-virus Application Management Port
    aamp = 3939, 'aamp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XeCP Node Service
    #: - [UDP] XeCP Node Service
    xecp_node = 3940, 'xecp-node', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Home Portal Web Server
    #: - [UDP] Home Portal Web Server
    homeportal_web = 3941, 'homeportal-web', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] satellite distribution
    #: - [UDP] satellite distribution
    srdp = 3942, 'srdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TetraNode Ip Gateway
    #: - [UDP] TetraNode Ip Gateway
    tig = 3943, 'tig', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] S-Ops Management
    #: - [UDP] S-Ops Management
    sops = 3944, 'sops', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMCADS Server Port
    #: - [UDP] EMCADS Server Port
    emcads = 3945, 'emcads', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BackupEDGE Server
    #: - [UDP] BackupEDGE Server
    backupedge = 3946, 'backupedge', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Connect and Control Protocol for Consumer, Commercial, and Industrial
    #: Electronic Devices
    #: - [UDP] Connect and Control Protocol for Consumer, Commercial, and Industrial
    #: Electronic Devices
    ccp = 3947, 'ccp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Anton Paar Device Administration Protocol
    #: - [UDP] Anton Paar Device Administration Protocol
    apdap = 3948, 'apdap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dynamic Routing Information Protocol
    #: - [UDP] Dynamic Routing Information Protocol
    drip = 3949, 'drip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Name Munging
    #: - [UDP] Name Munging
    namemunge = 3950, 'namemunge', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PWG IPP Facsimile
    #: - [UDP] PWG IPP Facsimile
    pwgippfax = 3951, 'pwgippfax', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] I3 Session Manager
    #: - [UDP] I3 Session Manager
    i3_sessionmgr = 3952, 'i3-sessionmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Eydeas XMLink Connect
    #: - [UDP] Eydeas XMLink Connect
    xmlink_connect = 3953, 'xmlink-connect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AD Replication RPC
    #: - [UDP] AD Replication RPC
    adrep = 3954, 'adrep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] p2pCommunity
    #: - [UDP] p2pCommunity
    p2pcommunity = 3955, 'p2pcommunity', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GigE Vision Control
    #: - [UDP] GigE Vision Control
    gvcp = 3956, 'gvcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MQEnterprise Broker
    #: - [UDP] MQEnterprise Broker
    mqe_broker = 3957, 'mqe-broker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MQEnterprise Agent
    #: - [UDP] MQEnterprise Agent
    mqe_agent = 3958, 'mqe-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tree Hopper Networking
    #: - [UDP] Tree Hopper Networking
    treehopper = 3959, 'treehopper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bess Peer Assessment
    #: - [UDP] Bess Peer Assessment
    bess = 3960, 'bess', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ProAxess Server
    #: - [UDP] ProAxess Server
    proaxess = 3961, 'proaxess', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SBI Agent Protocol
    #: - [UDP] SBI Agent Protocol
    sbi_agent = 3962, 'sbi-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Teran Hybrid Routing Protocol
    #: - [UDP] Teran Hybrid Routing Protocol
    thrp = 3963, 'thrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SASG GPRS
    #: - [UDP] SASG GPRS
    sasggprs = 3964, 'sasggprs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avanti IP to NCPE API
    #: - [UDP] Avanti IP to NCPE API
    ati_ip_to_ncpe = 3965, 'ati-ip-to-ncpe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BuildForge Lock Manager
    #: - [UDP] BuildForge Lock Manager
    bflckmgr = 3966, 'bflckmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PPS Message Service
    #: - [UDP] PPS Message Service
    ppsms = 3967, 'ppsms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iAnywhere DBNS
    #: - [UDP] iAnywhere DBNS
    ianywhere_dbns = 3968, 'ianywhere-dbns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Landmark Messages
    #: - [UDP] Landmark Messages
    landmarks = 3969, 'landmarks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LANrev Agent
    #: - [UDP] LANrev Agent
    lanrevagent = 3970, 'lanrevagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LANrev Server
    #: - [UDP] LANrev Server
    lanrevserver = 3971, 'lanrevserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ict-control Protocol
    #: - [UDP] ict-control Protocol
    iconp = 3972, 'iconp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ConnectShip Progistics
    #: - [UDP] ConnectShip Progistics
    progistics = 3973, 'progistics', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Applicant Tracking Service
    #: - [UDP] Remote Applicant Tracking Service
    xk22 = 3974, 'xk22', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Air Shot
    #: - [UDP] Air Shot
    airshot = 3975, 'airshot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Server Automation Agent
    #: - [UDP] Server Automation Agent
    opswagent = 3976, 'opswagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Opsware Manager
    #: - [UDP] Opsware Manager
    opswmanager = 3977, 'opswmanager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secured Configuration Server
    #: - [UDP] Secured Configuration Server
    secure_cfg_svr = 3978, 'secure-cfg-svr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Smith Micro Wide Area Network Service
    #: - [UDP] Smith Micro Wide Area Network Service
    smwan = 3979, 'smwan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_3980 = 3980, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Starfish System Admin
    #: - [UDP] Starfish System Admin
    starfish = 3981, 'starfish', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESRI Image Server
    #: - [UDP] ESRI Image Server
    eis = 3982, 'eis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESRI Image Service
    #: - [UDP] ESRI Image Service
    eisp = 3983, 'eisp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MAPPER network node manager
    #: - [UDP] MAPPER network node manager
    mapper_nodemgr = 3984, 'mapper-nodemgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MAPPER TCP/IP server
    #: - [UDP] MAPPER TCP/IP server
    mapper_mapethd = 3985, 'mapper-mapethd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MAPPER workstation server IANA assigned this well-formed service name
    #: as a replacement for "mapper-ws_ethd".
    #: - [TCP] MAPPER workstation server
    #: - [UDP] MAPPER workstation server IANA assigned this well-formed service name
    #: as a replacement for "mapper-ws_ethd".
    #: - [UDP] MAPPER workstation server
    mapper_ws_ethd = 3986, 'mapper-ws-ethd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Centerline
    #: - [UDP] Centerline
    centerline = 3987, 'centerline', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DCS Configuration Port
    #: - [UDP] DCS Configuration Port
    dcs_config = 3988, 'dcs-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BindView-Query Engine
    #: - [UDP] BindView-Query Engine
    bv_queryengine = 3989, 'bv-queryengine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BindView-IS
    #: - [UDP] BindView-IS
    bv_is = 3990, 'bv-is', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BindView-SMCServer
    #: - [UDP] BindView-SMCServer
    bv_smcsrv = 3991, 'bv-smcsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BindView-DirectoryServer
    #: - [UDP] BindView-DirectoryServer
    bv_ds = 3992, 'bv-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BindView-Agent
    #: - [UDP] BindView-Agent
    bv_agent = 3993, 'bv-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_3994 = 3994, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] ISS Management Svcs SSL
    #: - [UDP] ISS Management Svcs SSL
    iss_mgmt_ssl = 3995, 'iss-mgmt-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] abcsoftware-01
    #: - [UDP] abcsoftware-01
    abcsoftware = 3996, 'abcsoftware', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] aes_db
    #: - [UDP] aes_db
    agentsease_db = 3997, 'agentsease-db', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Distributed Nagios Executor Service
    #: - [UDP] Distributed Nagios Executor Service
    dnx = 3998, 'dnx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Norman distributes scanning service
    #: - [UDP] Norman distributes scanning service
    nvcnet = 3999, 'nvcnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Terabase
    #: - [UDP] Terabase
    terabase = 4000, 'terabase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NewOak
    #: - [UDP] NewOak
    newoak = 4001, 'newoak', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pxc-spvr-ft
    #: - [UDP] pxc-spvr-ft
    pxc_spvr_ft = 4002, 'pxc-spvr-ft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pxc-splr-ft
    #: - [UDP] pxc-splr-ft
    pxc_splr_ft = 4003, 'pxc-splr-ft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pxc-roid
    #: - [UDP] pxc-roid
    pxc_roid = 4004, 'pxc-roid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pxc-pin
    #: - [UDP] pxc-pin
    pxc_pin = 4005, 'pxc-pin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pxc-spvr
    #: - [UDP] pxc-spvr
    pxc_spvr = 4006, 'pxc-spvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pxc-splr
    #: - [UDP] pxc-splr
    pxc_splr = 4007, 'pxc-splr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetCheque accounting
    #: - [UDP] NetCheque accounting
    netcheque = 4008, 'netcheque', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Chimera HWM
    #: - [UDP] Chimera HWM
    chimera_hwm = 4009, 'chimera-hwm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Samsung Unidex
    #: - [UDP] Samsung Unidex
    samsung_unidex = 4010, 'samsung-unidex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alternate Service Boot
    #: - [UDP] Alternate Service Boot
    altserviceboot = 4011, 'altserviceboot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PDA Gate
    #: - [UDP] PDA Gate
    pda_gate = 4012, 'pda-gate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACL Manager
    #: - [UDP] ACL Manager
    acl_manager = 4013, 'acl-manager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TAICLOCK
    #: - [UDP] TAICLOCK
    taiclock = 4014, 'taiclock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talarian Mcast
    #: - [UDP] Talarian Mcast
    talarian_mcast1 = 4015, 'talarian-mcast1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talarian Mcast
    #: - [UDP] Talarian Mcast
    talarian_mcast2 = 4016, 'talarian-mcast2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talarian Mcast
    #: - [UDP] Talarian Mcast
    talarian_mcast3 = 4017, 'talarian-mcast3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talarian Mcast
    #: - [UDP] Talarian Mcast
    talarian_mcast4 = 4018, 'talarian-mcast4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talarian Mcast
    #: - [UDP] Talarian Mcast
    talarian_mcast5 = 4019, 'talarian-mcast5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRAP Port
    #: - [UDP] TRAP Port
    trap = 4020, 'trap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nexus Portal
    #: - [UDP] Nexus Portal
    nexus_portal = 4021, 'nexus-portal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DNOX
    #: - [UDP] DNOX
    dnox = 4022, 'dnox', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESNM Zoning Port
    #: - [UDP] ESNM Zoning Port
    esnm_zoning = 4023, 'esnm-zoning', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TNP1 User Port
    #: - [UDP] TNP1 User Port
    tnp1_port = 4024, 'tnp1-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Partition Image Port
    #: - [UDP] Partition Image Port
    partimage = 4025, 'partimage', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Graphical Debug Server
    #: - [UDP] Graphical Debug Server
    as_debug = 4026, 'as-debug', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bitxpress
    #: - [UDP] bitxpress
    bxp = 4027, 'bxp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DTServer Port
    #: - [UDP] DTServer Port
    dtserver_port = 4028, 'dtserver-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IP Q signaling protocol
    #: - [UDP] IP Q signaling protocol
    ip_qsig = 4029, 'ip-qsig', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Accell/JSP Daemon Port
    #: - [UDP] Accell/JSP Daemon Port
    jdmn_port = 4030, 'jdmn-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UUCP over SSL
    #: - [UDP] UUCP over SSL
    suucp = 4031, 'suucp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS Authorization Service
    #: - [UDP] VERITAS Authorization Service
    vrts_auth_port = 4032, 'vrts-auth-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SANavigator Peer Port
    #: - [UDP] SANavigator Peer Port
    sanavigator = 4033, 'sanavigator', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ubiquinox Daemon
    #: - [UDP] Ubiquinox Daemon
    ubxd = 4034, 'ubxd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP Push OTA-HTTP port
    #: - [UDP] WAP Push OTA-HTTP port
    wap_push_http = 4035, 'wap-push-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP Push OTA-HTTP secure
    #: - [UDP] WAP Push OTA-HTTP secure
    wap_push_https = 4036, 'wap-push-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RaveHD network control
    #: - [UDP] RaveHD network control
    ravehd = 4037, 'ravehd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fazzt Point-To-Point
    #: - [UDP] Fazzt Point-To-Point
    fazzt_ptp = 4038, 'fazzt-ptp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fazzt Administration
    #: - [UDP] Fazzt Administration
    fazzt_admin = 4039, 'fazzt-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Yo.net main service
    #: - [UDP] Yo.net main service
    yo_main = 4040, 'yo-main', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rocketeer-Houston
    #: - [UDP] Rocketeer-Houston
    houston = 4041, 'houston', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LDXP
    #: - [UDP] LDXP
    ldxp = 4042, 'ldxp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Neighbour Identity Resolution
    #: - [UDP] Neighbour Identity Resolution
    nirp = 4043, 'nirp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Location Tracking Protocol
    #: - [UDP] Location Tracking Protocol
    ltp = 4044, 'ltp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Printing Protocol
    #: - [UDP] Network Printing Protocol
    npp_92 = 92, 'npp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Paging Protocol
    #: - [UDP] Network Paging Protocol
    npp_4045 = 4045, 'npp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Accounting Protocol
    #: - [UDP] Accounting Protocol
    acp_proto = 4046, 'acp-proto', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Context Transfer Protocol
    #: - [UDP] Context Transfer Protocol
    ctp_state = 4047, 'ctp-state', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_4048 = 4048, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Wide Area File Services
    #: - [UDP] Wide Area File Services
    wafs = 4049, 'wafs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Wide Area File Services
    #: - [UDP] Wide Area File Services
    cisco_wafs = 4050, 'cisco-wafs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cisco Peer to Peer Distribution Protocol
    #: - [UDP] Cisco Peer to Peer Distribution Protocol
    cppdp = 4051, 'cppdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VoiceConnect Interact
    #: - [UDP] VoiceConnect Interact
    interact = 4052, 'interact', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CosmoCall Universe Communications Port 1
    #: - [UDP] CosmoCall Universe Communications Port 1
    ccu_comm_1 = 4053, 'ccu-comm-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CosmoCall Universe Communications Port 2
    #: - [UDP] CosmoCall Universe Communications Port 2
    ccu_comm_2 = 4054, 'ccu-comm-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CosmoCall Universe Communications Port 3
    #: - [UDP] CosmoCall Universe Communications Port 3
    ccu_comm_3 = 4055, 'ccu-comm-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Location Message Service
    #: - [UDP] Location Message Service
    lms = 4056, 'lms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Servigistics WFM server
    #: - [UDP] Servigistics WFM server
    wfm = 4057, 'wfm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kingfisher protocol
    #: - [UDP] Kingfisher protocol
    kingfisher = 4058, 'kingfisher', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DLMS/COSEM
    #: - [UDP] DLMS/COSEM
    dlms_cosem = 4059, 'dlms-cosem', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DSMETER Inter-Agent Transfer Channel IANA assigned this well-formed
    #: service name as a replacement for "dsmeter_iatc".
    #: - [TCP] DSMETER Inter-Agent Transfer Channel
    #: - [UDP] DSMETER Inter-Agent Transfer Channel IANA assigned this well-formed
    #: service name as a replacement for "dsmeter_iatc".
    #: - [UDP] DSMETER Inter-Agent Transfer Channel
    dsmeter_iatc = 4060, 'dsmeter-iatc', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Ice Location Service (TCP)
    #: - [UDP] Ice Location Service (TCP)
    ice_location = 4061, 'ice-location', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ice Location Service (SSL)
    #: - [UDP] Ice Location Service (SSL)
    ice_slocation = 4062, 'ice-slocation', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ice Firewall Traversal Service (TCP)
    #: - [UDP] Ice Firewall Traversal Service (TCP)
    ice_router = 4063, 'ice-router', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ice Firewall Traversal Service (SSL)
    #: - [UDP] Ice Firewall Traversal Service (SSL)
    ice_srouter = 4064, 'ice-srouter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Avanti Common Data IANA assigned this well-formed service name as a
    #: replacement for "avanti_cdp".
    #: - [TCP] Avanti Common Data
    #: - [UDP] Avanti Common Data IANA assigned this well-formed service name as a
    #: replacement for "avanti_cdp".
    #: - [UDP] Avanti Common Data
    avanti_cdp = 4065, 'avanti-cdp', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Performance Measurement and Analysis
    #: - [UDP] Performance Measurement and Analysis
    pmas = 4066, 'pmas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Information Distribution Protocol
    #: - [UDP] Information Distribution Protocol
    idp = 4067, 'idp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IP Fleet Broadcast
    #: - [UDP] IP Fleet Broadcast
    ipfltbcst = 4068, 'ipfltbcst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Minger Email Address Validation Service
    #: - [UDP] Minger Email Address Validation Service
    minger = 4069, 'minger', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trivial IP Encryption (TrIPE)
    #: - [UDP] Trivial IP Encryption (TrIPE)
    tripe = 4070, 'tripe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Automatically Incremental Backup
    #: - [UDP] Automatically Incremental Backup
    aibkup = 4071, 'aibkup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zieto Socket Communications
    #: - [UDP] Zieto Socket Communications
    zieto_sock = 4072, 'zieto-sock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Interactive Remote Application Pairing Protocol
    #: - [UDP] Interactive Remote Application Pairing Protocol
    irapp = 4073, 'irapp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cequint City ID UI trigger
    #: - [UDP] Cequint City ID UI trigger
    cequint_cityid = 4074, 'cequint-cityid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ISC Alarm Message Service
    #: - [UDP] ISC Alarm Message Service
    perimlan = 4075, 'perimlan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Seraph DCS
    #: - [UDP] Seraph DCS
    seraph = 4076, 'seraph', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4077 = 4077, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Ascom IP Alarming
    ascomalarm = 4077, 'ascomalarm', TransportProtocol.get('udp')

    #: [TCP] Coordinated Security Service Protocol
    cssp = 4078, 'cssp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4078 = 4078, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] SANtools Diagnostic Server
    #: - [UDP] SANtools Diagnostic Server
    santools = 4079, 'santools', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lorica inside facing
    #: - [UDP] Lorica inside facing
    lorica_in = 4080, 'lorica-in', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lorica inside facing (SSL)
    #: - [UDP] Lorica inside facing (SSL)
    lorica_in_sec = 4081, 'lorica-in-sec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lorica outside facing
    #: - [UDP] Lorica outside facing
    lorica_out = 4082, 'lorica-out', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lorica outside facing (SSL)
    #: - [UDP] Lorica outside facing (SSL)
    lorica_out_sec = 4083, 'lorica-out-sec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4084 = 4084, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Fortisphere VM Service
    fortisphere_vm = 4084, 'fortisphere-vm', TransportProtocol.get('udp')

    #: [TCP] EZNews Newsroom Message Service
    ezmessagesrv = 4085, 'ezmessagesrv', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4085 = 4085, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4086 = 4086, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Firewall/NAT state table synchronization
    ftsync = 4086, 'ftsync', TransportProtocol.get('udp')

    #: [TCP] APplus Service
    applusservice = 4087, 'applusservice', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4087 = 4087, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Noah Printing Service Protocol
    npsp = 4088, 'npsp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4088 = 4088, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] OpenCORE Remote Control Service
    #: - [UDP] OpenCORE Remote Control Service
    opencore = 4089, 'opencore', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA BCAST Service Guide
    #: - [UDP] OMA BCAST Service Guide
    omasgport = 4090, 'omasgport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EminentWare Installer
    #: - [UDP] EminentWare Installer
    ewinstaller = 4091, 'ewinstaller', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EminentWare DGS
    #: - [UDP] EminentWare DGS
    ewdgs = 4092, 'ewdgs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pvx Plus CS Host
    #: - [UDP] Pvx Plus CS Host
    pvxpluscs = 4093, 'pvxpluscs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sysrq daemon
    #: - [UDP] sysrq daemon
    sysrqd = 4094, 'sysrqd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xtgui information service
    #: - [UDP] xtgui information service
    xtgui = 4095, 'xtgui', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BRE (Bridge Relay Element)
    #: - [UDP] BRE (Bridge Relay Element)
    bre = 4096, 'bre', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Patrol View
    #: - [UDP] Patrol View
    patrolview = 4097, 'patrolview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] drmsfsd
    #: - [UDP] drmsfsd
    drmsfsd = 4098, 'drmsfsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DPCP
    #: - [UDP] DPCP
    dpcp = 4099, 'dpcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IGo Incognito Data Port
    #: - [UDP] IGo Incognito Data Port
    igo_incognito = 4100, 'igo-incognito', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Braille protocol
    #: - [UDP] Braille protocol
    brlp_0 = 4101, 'brlp-0', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Braille protocol
    #: - [UDP] Braille protocol
    brlp_1 = 4102, 'brlp-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Braille protocol
    #: - [UDP] Braille protocol
    brlp_2 = 4103, 'brlp-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Braille protocol
    #: - [UDP] Braille protocol
    brlp_3 = 4104, 'brlp-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Shofar
    #: - [UDP] Shofar
    shofar = 4105, 'shofar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Synchronite
    #: - [UDP] Synchronite
    synchronite = 4106, 'synchronite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JDL Accounting LAN Service
    #: - [UDP] JDL Accounting LAN Service
    j_ac = 4107, 'j-ac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACCEL
    #: - [UDP] ACCEL
    accel = 4108, 'accel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Instantiated Zero-control Messaging
    #: - [UDP] Instantiated Zero-control Messaging
    izm = 4109, 'izm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] G2 RFID Tag Telemetry Data
    #: - [UDP] G2 RFID Tag Telemetry Data
    g2tag = 4110, 'g2tag', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xgrid
    #: - [UDP] Xgrid
    xgrid = 4111, 'xgrid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apple VPN Server Reporting Protocol
    #: - [UDP] Apple VPN Server Reporting Protocol
    apple_vpns_rp = 4112, 'apple-vpns-rp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AIPN LS Registration
    #: - [UDP] AIPN LS Registration
    aipn_reg = 4113, 'aipn-reg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JomaMQMonitor
    #: - [UDP] JomaMQMonitor
    jomamqmonitor = 4114, 'jomamqmonitor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CDS Transfer Agent
    #: - [UDP] CDS Transfer Agent
    cds = 4115, 'cds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] smartcard-TLS
    #: - [UDP] smartcard-TLS
    smartcard_tls = 4116, 'smartcard-tls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hillr Connection Manager
    #: - [UDP] Hillr Connection Manager
    hillrserv = 4117, 'hillrserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netadmin Systems NETscript service
    #: - [UDP] Netadmin Systems NETscript service
    netscript = 4118, 'netscript', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Assuria Log Manager
    #: - [UDP] Assuria Log Manager
    assuria_slm = 4119, 'assuria-slm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] MiniRem Remote Telemetry and Control
    minirem = 4120, 'minirem', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4120 = 4120, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] e-Builder Application Communication
    #: - [UDP] e-Builder Application Communication
    e_builder = 4121, 'e-builder', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fiber Patrol Alarm Service
    #: - [UDP] Fiber Patrol Alarm Service
    fprams = 4122, 'fprams', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rohill TetraNode Ip Gateway v2
    #: - [UDP] Rohill TetraNode Ip Gateway v2
    tigv2 = 4124, 'tigv2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Opsview Envoy
    #: - [UDP] Opsview Envoy
    opsview_envoy = 4125, 'opsview-envoy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Data Domain Replication Service
    #: - [UDP] Data Domain Replication Service
    ddrepl = 4126, 'ddrepl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetUniKeyServer
    #: - [UDP] NetUniKeyServer
    unikeypro = 4127, 'unikeypro', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NuFW decision delegation protocol
    #: - [UDP] NuFW decision delegation protocol
    nufw = 4128, 'nufw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NuFW authentication protocol
    #: - [UDP] NuFW authentication protocol
    nuauth = 4129, 'nuauth', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FRONET message protocol
    #: - [UDP] FRONET message protocol
    fronet = 4130, 'fronet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Global Maintech Stars
    #: - [UDP] Global Maintech Stars
    stars = 4131, 'stars', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NUTS Daemon IANA assigned this well-formed service name as a
    #: replacement for "nuts_dem".
    #: - [TCP] NUTS Daemon
    #: - [UDP] NUTS Daemon IANA assigned this well-formed service name as a
    #: replacement for "nuts_dem".
    #: - [UDP] NUTS Daemon
    nuts_dem = 4132, 'nuts-dem', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] NUTS Bootp Server IANA assigned this well-formed service name as a
    #: replacement for "nuts_bootp".
    #: - [TCP] NUTS Bootp Server
    #: - [UDP] NUTS Bootp Server IANA assigned this well-formed service name as a
    #: replacement for "nuts_bootp".
    #: - [UDP] NUTS Bootp Server
    nuts_bootp = 4133, 'nuts-bootp', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] NIFTY-Serve HMI protocol
    #: - [UDP] NIFTY-Serve HMI protocol
    nifty_hmi = 4134, 'nifty-hmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Classic Line Database Server Attach
    #: - [UDP] Classic Line Database Server Attach
    cl_db_attach = 4135, 'cl-db-attach', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Classic Line Database Server Request
    #: - [UDP] Classic Line Database Server Request
    cl_db_request = 4136, 'cl-db-request', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Classic Line Database Server Remote
    #: - [UDP] Classic Line Database Server Remote
    cl_db_remote = 4137, 'cl-db-remote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nettest
    #: - [UDP] nettest
    nettest = 4138, 'nettest', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Imperfect Networks Server
    #: - [UDP] Imperfect Networks Server
    thrtx = 4139, 'thrtx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cedros Fraud Detection System IANA assigned this well-formed service
    #: name as a replacement for "cedros_fds".
    #: - [TCP] Cedros Fraud Detection System
    #: - [UDP] Cedros Fraud Detection System IANA assigned this well-formed service
    #: name as a replacement for "cedros_fds".
    #: - [UDP] Cedros Fraud Detection System
    cedros_fds = 4140, 'cedros-fds', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Workflow Server
    #: - [UDP] Workflow Server
    oirtgsvc = 4141, 'oirtgsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Document Server
    #: - [UDP] Document Server
    oidocsvc = 4142, 'oidocsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Document Replication
    #: - [UDP] Document Replication
    oidsr = 4143, 'oidsr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_4144 = 4144, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] VVR Control
    #: - [UDP] VVR Control
    vvr_control = 4145, 'vvr-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TGCConnect Beacon
    #: - [UDP] TGCConnect Beacon
    tgcconnect = 4146, 'tgcconnect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multum Service Manager
    #: - [UDP] Multum Service Manager
    vrxpservman = 4147, 'vrxpservman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HHB Handheld Client
    #: - [UDP] HHB Handheld Client
    hhb_handheld = 4148, 'hhb-handheld', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A10 GSLB Service
    #: - [UDP] A10 GSLB Service
    agslb = 4149, 'agslb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PowerAlert Network Shutdown Agent
    #: - [UDP] PowerAlert Network Shutdown Agent
    poweralert_nsa = 4150, 'poweralert-nsa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Men & Mice Remote Control IANA assigned this well-formed service name
    #: as a replacement for "menandmice_noh".
    #: - [TCP] Men & Mice Remote Control
    #: - [UDP] Men & Mice Remote Control IANA assigned this well-formed service name
    #: as a replacement for "menandmice_noh".
    #: - [UDP] Men & Mice Remote Control
    menandmice_noh = 4151, 'menandmice-noh', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] iDigTech Multiplex IANA assigned this well-formed service name as a
    #: replacement for "idig_mux".
    #: - [TCP] iDigTech Multiplex
    #: - [UDP] iDigTech Multiplex IANA assigned this well-formed service name as a
    #: replacement for "idig_mux".
    #: - [UDP] iDigTech Multiplex
    idig_mux = 4152, 'idig-mux', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] MBL Remote Battery Monitoring
    #: - [UDP] MBL Remote Battery Monitoring
    mbl_battd = 4153, 'mbl-battd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] atlinks device discovery
    #: - [UDP] atlinks device discovery
    atlinks = 4154, 'atlinks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bazaar version control system
    #: - [UDP] Bazaar version control system
    bzr = 4155, 'bzr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STAT Results
    #: - [UDP] STAT Results
    stat_results = 4156, 'stat-results', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STAT Scanner Control
    #: - [UDP] STAT Scanner Control
    stat_scanner = 4157, 'stat-scanner', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STAT Command Center
    #: - [UDP] STAT Command Center
    stat_cc = 4158, 'stat-cc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Security Service
    #: - [UDP] Network Security Service
    nss = 4159, 'nss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jini Discovery
    #: - [UDP] Jini Discovery
    jini_discovery = 4160, 'jini-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMS Contact
    #: - [UDP] OMS Contact
    omscontact = 4161, 'omscontact', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMS Topology
    #: - [UDP] OMS Topology
    omstopology = 4162, 'omstopology', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Silver Peak Peer Protocol
    #: - [UDP] Silver Peak Peer Protocol
    silverpeakpeer = 4163, 'silverpeakpeer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Silver Peak Communication Protocol
    #: - [UDP] Silver Peak Communication Protocol
    silverpeakcomm = 4164, 'silverpeakcomm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ArcLink over Ethernet
    #: - [UDP] ArcLink over Ethernet
    altcp = 4165, 'altcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Joost Peer to Peer Protocol
    #: - [UDP] Joost Peer to Peer Protocol
    joost = 4166, 'joost', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DeskDirect Global Network
    #: - [UDP] DeskDirect Global Network
    ddgn = 4167, 'ddgn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PrintSoft License Server
    #: - [UDP] PrintSoft License Server
    pslicser = 4168, 'pslicser', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Automation Drive Interface Transport
    iadt = 4169, 'iadt', TransportProtocol.get('tcp')

    #: [UDP] Internet ADT Discovery Protocol
    iadt_disc = 4169, 'iadt-disc', TransportProtocol.get('udp')

    #: [TCP] SMPTE Content Synchonization Protocol
    d_cinema_csp = 4170, 'd-cinema-csp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4170 = 4170, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Maxlogic Supervisor Communication
    ml_svnet = 4171, 'ml-svnet', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4171 = 4171, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] PC over IP
    #: - [UDP] PC over IP
    pcoip = 4172, 'pcoip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4173 = 4173, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] MMA Device Discovery
    mma_discovery = 4173, 'mma-discovery', TransportProtocol.get('udp')

    #: [TCP] StorMagic Cluster Services
    smcluster = 4174, 'smcluster', TransportProtocol.get('tcp')

    #: [UDP] StorMagic Discovery
    sm_disc = 4174, 'sm-disc', TransportProtocol.get('udp')

    #: [TCP] Brocade Cluster Communication Protocol
    bccp = 4175, 'bccp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4175 = 4175, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Translattice Cluster IPC Proxy
    tl_ipcproxy = 4176, 'tl-ipcproxy', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4176 = 4176, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Wello P2P pubsub service
    #: - [UDP] Wello P2P pubsub service
    wello = 4177, 'wello', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] StorMan
    #: - [UDP] StorMan
    storman = 4178, 'storman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Maxum Services
    #: - [UDP] Maxum Services
    maxumsp = 4179, 'maxumsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HTTPX
    #: - [UDP] HTTPX
    httpx = 4180, 'httpx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MacBak
    #: - [UDP] MacBak
    macbak = 4181, 'macbak', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Production Company Pro TCP Service
    #: - [UDP] Production Company Pro TCP Service
    pcptcpservice = 4182, 'pcptcpservice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CyborgNet communications protocol
    #: - [UDP] CyborgNet communications protocol
    cyborgnet = 4183, 'cyborgnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UNIVERSE SUITE MESSAGE SERVICE IANA assigned this well-formed service
    #: name as a replacement for "universe_suite".
    #: - [TCP] UNIVERSE SUITE MESSAGE SERVICE
    #: - [UDP] UNIVERSE SUITE MESSAGE SERVICE IANA assigned this well-formed service
    #: name as a replacement for "universe_suite".
    #: - [UDP] UNIVERSE SUITE MESSAGE SERVICE
    universe_suite = 4184, 'universe-suite', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Woven Control Plane Protocol
    #: - [UDP] Woven Control Plane Protocol
    wcpp = 4185, 'wcpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Box Backup Store Service
    boxbackupstore = 4186, 'boxbackupstore', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4186 = 4186, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Cascade Proxy IANA assigned this well-formed service name as a
    #: replacement for "csc_proxy".
    #: - [TCP] Cascade Proxy
    csc_proxy = 4187, 'csc-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4187 = 4187, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Vatata Peer to Peer Protocol
    #: - [UDP] Vatata Peer to Peer Protocol
    vatata = 4188, 'vatata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Path Computation Element Communication Protocol
    pcep = 4189, 'pcep', TransportProtocol.get('tcp')

    #: [UDP] Reserved [:rfc:`5440`]
    reserved_4189 = 4189, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ManageSieve Protocol
    sieve = 4190, 'sieve', TransportProtocol.get('tcp')

    #: [UDP] Reserved [:rfc:`5804`]
    reserved_4190 = 4190, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4191 = 4191, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Dual Stack MIPv6 NAT Traversal [:rfc:`5555`]
    dsmipv6 = 4191, 'dsmipv6', TransportProtocol.get('udp')

    #: [TCP] Azeti Agent Service
    azeti = 4192, 'azeti', TransportProtocol.get('tcp')

    #: [UDP] azeti blinddate
    azeti_bd = 4192, 'azeti-bd', TransportProtocol.get('udp')

    #: [TCP] PxPlus remote file srvr
    pvxplusio = 4193, 'pvxplusio', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4193 = 4193, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Security Protocol and Data Model
    spdm = 4194, 'spdm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4194 = 4194, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] AWS protocol for cloud remoting solution
    #: - [UDP] AWS protocol for cloud remoting solution
    #: - [SCTP] AWS protocol for cloud remoting solution
    #: - [DCCP] AWS protocol for cloud remoting solution
    aws_wsp = 4195, 'aws-wsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp') | TransportProtocol.get('dccp')

    #: [N/A] Unassigned
    unassigned_4196 = 4196, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Harman HControl Protocol
    #: - [UDP] Harman HControl Protocol
    hctl = 4197, 'hctl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_4198 = 4198, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] EIMS ADMIN
    #: - [UDP] EIMS ADMIN
    eims_admin = 4199, 'eims-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Corel CCam
    #: - [UDP] Corel CCam
    corelccam = 4300, 'corelccam', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Diagnostic Data
    #: - [UDP] Diagnostic Data
    d_data = 4301, 'd-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Diagnostic Data Control
    #: - [UDP] Diagnostic Data Control
    d_data_control = 4302, 'd-data-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Railroad Command Protocol
    #: - [UDP] Simple Railroad Command Protocol
    srcp = 4303, 'srcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] One-Wire Filesystem Server
    #: - [UDP] One-Wire Filesystem Server
    owserver = 4304, 'owserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] better approach to mobile ad-hoc networking
    #: - [UDP] better approach to mobile ad-hoc networking
    batman = 4305, 'batman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hellgate London
    #: - [UDP] Hellgate London
    pinghgl = 4306, 'pinghgl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TrueConf Videoconference Service
    #: - [UDP] TrueConf Videoconference Service
    trueconf = 4307, 'trueconf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CompX-LockView
    #: - [UDP] CompX-LockView
    compx_lockview = 4308, 'compx-lockview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Exsequi Appliance Discovery
    #: - [UDP] Exsequi Appliance Discovery
    dserver = 4309, 'dserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mir-RT exchange service
    #: - [UDP] Mir-RT exchange service
    mirrtex = 4310, 'mirrtex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] P6R Secure Server Management Console
    p6ssmc = 4311, 'p6ssmc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4311 = 4311, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Parascale Membership Manager
    pscl_mgt = 4312, 'pscl-mgt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4312 = 4312, 'reserved', TransportProtocol.get('udp')

    #: [TCP] PERRLA User Services
    perrla = 4313, 'perrla', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4313 = 4313, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ChoiceView Agent
    choiceview_agt = 4314, 'choiceview-agt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4314 = 4314, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_4315 = 4315, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] ChoiceView Client
    choiceview_clt = 4316, 'choiceview-clt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4316 = 4316, 'reserved', TransportProtocol.get('udp')

    #: [TCP] OpenTelemetry Protocol
    opentelemetry = 4317, 'opentelemetry', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4317 = 4317, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_4318 = 4318, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Fox SkyTale encrypted communication
    #: - [UDP] Fox SkyTale encrypted communication
    fox_skytale = 4319, 'fox-skytale', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FDT Remote Categorization Protocol
    #: - [UDP] FDT Remote Categorization Protocol
    fdt_rcatp = 4320, 'fdt-rcatp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Who Is [:rfc:`2167`]
    #: - [UDP] Remote Who Is [:rfc:`2167`]
    rwhois = 4321, 'rwhois', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRIM Event Service
    #: - [UDP] TRIM Event Service
    trim_event = 4322, 'trim-event', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRIM ICE Service
    #: - [UDP] TRIM ICE Service
    trim_ice = 4323, 'trim-ice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Reserved
    reserved_4324 = 4324, 'reserved', TransportProtocol.get('undefined')

    #: [TCP] Cadcorp GeognoSIS Administrator
    geognosisadmin = 4325, 'geognosisadmin', TransportProtocol.get('tcp')

    #: [UDP] Cadcorp GeognoSIS Administrator
    geognosisman = 4325, 'geognosisman', TransportProtocol.get('udp')

    #: - [TCP] Cadcorp GeognoSIS
    #: - [UDP] Cadcorp GeognoSIS
    geognosis = 4326, 'geognosis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jaxer Web Protocol
    #: - [UDP] Jaxer Web Protocol
    jaxer_web = 4327, 'jaxer-web', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Jaxer Manager Command Protocol
    #: - [UDP] Jaxer Manager Command Protocol
    jaxer_manager = 4328, 'jaxer-manager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PubliQare Distributed Environment Synchronisation Engine
    publiqare_sync = 4329, 'publiqare-sync', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4329 = 4329, 'reserved', TransportProtocol.get('udp')

    #: [TCP] DEY Storage Administration REST API
    dey_sapi = 4330, 'dey-sapi', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4330 = 4330, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ktickets REST API for event management and ticketing systems (embedded
    #: POS devices)
    ktickets_rest = 4331, 'ktickets-rest', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4331 = 4331, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Getty Images FOCUS service
    getty_focus = 4332, 'getty-focus', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4332 = 4332, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ArrowHead Service Protocol (AHSP)
    #: - [UDP] ArrowHead Service Protocol (AHSP)
    #: - [SCTP] ArrowHead Service Protocol (AHSP)
    ahsp = 4333, 'ahsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: [TCP] NETCONF Call Home (SSH) [:rfc:`8071`]
    netconf_ch_ssh = 4334, 'netconf-ch-ssh', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4334 = 4334, 'reserved', TransportProtocol.get('udp')

    #: [TCP] NETCONF Call Home (TLS) [:rfc:`8071`]
    netconf_ch_tls = 4335, 'netconf-ch-tls', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4335 = 4335, 'reserved', TransportProtocol.get('udp')

    #: [TCP] RESTCONF Call Home (TLS) [:rfc:`8071`]
    restconf_ch_tls = 4336, 'restconf-ch-tls', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4336 = 4336, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Gaia Connector Protocol
    #: - [UDP] Gaia Connector Protocol
    gaia = 4340, 'gaia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4341 = 4341, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] LISP Data Packets [:rfc:`9300`]
    lisp_data = 4341, 'lisp-data', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4342 = 4342, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] LISP Control Packets [:rfc:`9301`]
    lisp_control = 4342, 'lisp-control', TransportProtocol.get('udp')

    #: - [TCP] UNICALL
    #: - [UDP] UNICALL
    unicall = 4343, 'unicall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VinaInstall
    #: - [UDP] VinaInstall
    vinainstall = 4344, 'vinainstall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Macro 4 Network AS
    #: - [UDP] Macro 4 Network AS
    m4_network_as = 4345, 'm4-network-as', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ELAN LM
    #: - [UDP] ELAN LM
    elanlm = 4346, 'elanlm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LAN Surveyor
    #: - [UDP] LAN Surveyor
    lansurveyor = 4347, 'lansurveyor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITOSE
    #: - [UDP] ITOSE
    itose = 4348, 'itose', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] File System Port Map
    #: - [UDP] File System Port Map
    fsportmap = 4349, 'fsportmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net Device
    #: - [UDP] Net Device
    net_device = 4350, 'net-device', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PLCY Net Services
    #: - [UDP] PLCY Net Services
    plcy_net_svcs = 4351, 'plcy-net-svcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Projector Link
    #: - [UDP] Projector Link
    pjlink = 4352, 'pjlink', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] F5 iQuery
    #: - [UDP] F5 iQuery
    f5_iquery = 4353, 'f5-iquery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSNet Transmitter
    #: - [UDP] QSNet Transmitter
    qsnet_trans = 4354, 'qsnet-trans', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSNet Workstation
    #: - [UDP] QSNet Workstation
    qsnet_workst = 4355, 'qsnet-workst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSNet Assistant
    #: - [UDP] QSNet Assistant
    qsnet_assist = 4356, 'qsnet-assist', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSNet Conductor
    #: - [UDP] QSNet Conductor
    qsnet_cond = 4357, 'qsnet-cond', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSNet Nucleus
    #: - [UDP] QSNet Nucleus
    qsnet_nucl = 4358, 'qsnet-nucl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA BCAST Long-Term Key Messages
    #: - [UDP] OMA BCAST Long-Term Key Messages
    omabcastltkm = 4359, 'omabcastltkm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Matrix VNet Communication Protocol IANA assigned this well-formed
    #: service name as a replacement for "matrix_vnet".
    #: - [TCP] Matrix VNet Communication Protocol
    matrix_vnet = 4360, 'matrix-vnet', TransportProtocol.get('tcp') | TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4360 = 4360, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4361 = 4361, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] NavCom Discovery and Control Port
    nacnl = 4361, 'nacnl', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4362 = 4362, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] AFORE vNode Discovery protocol
    afore_vdp_disc = 4362, 'afore-vdp-disc', TransportProtocol.get('udp')

    #: [UDP] ShadowStream System
    shadowstream = 4366, 'shadowstream', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4366 = 4366, 'reserved', TransportProtocol.get('tcp')

    #: [N/A] Unassigned
    unassigned_4367 = 4367, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] WeatherBrief Direct
    #: - [UDP] WeatherBrief Direct
    wxbrief = 4368, 'wxbrief', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Erlang Port Mapper Daemon
    #: - [UDP] Erlang Port Mapper Daemon
    epmd = 4369, 'epmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ELPRO V2 Protocol Tunnel IANA assigned this well-formed service name
    #: as a replacement for "elpro_tunnel".
    #: - [TCP] ELPRO V2 Protocol Tunnel
    #: - [UDP] ELPRO V2 Protocol Tunnel IANA assigned this well-formed service name
    #: as a replacement for "elpro_tunnel".
    #: - [UDP] ELPRO V2 Protocol Tunnel
    elpro_tunnel = 4370, 'elpro-tunnel', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [TCP] LAN2CAN Control
    l2c_control = 4371, 'l2c-control', TransportProtocol.get('tcp')

    #: [UDP] LAN2CAN Discovery
    l2c_disc = 4371, 'l2c-disc', TransportProtocol.get('udp')

    #: - [TCP] LAN2CAN Data
    #: - [UDP] LAN2CAN Data
    l2c_data = 4372, 'l2c-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Authenticated Command Service
    #: - [UDP] Remote Authenticated Command Service
    remctl = 4373, 'remctl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PSI Push-to-Talk Protocol
    psi_ptt = 4374, 'psi-ptt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4374 = 4374, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Toltec EasyShare
    #: - [UDP] Toltec EasyShare
    tolteces = 4375, 'tolteces', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BioAPI Interworking
    #: - [UDP] BioAPI Interworking
    bip = 4376, 'bip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cambridge Pixel SPx Server
    #: - [UDP] Cambridge Pixel SPx Server
    cp_spxsvr = 4377, 'cp-spxsvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cambridge Pixel SPx Display
    #: - [UDP] Cambridge Pixel SPx Display
    cp_spxdpy = 4378, 'cp-spxdpy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CTDB
    #: - [UDP] CTDB
    ctdb = 4379, 'ctdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xandros Community Management Service
    #: - [UDP] Xandros Community Management Service
    xandros_cms = 4389, 'xandros-cms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Physical Access Control
    #: - [UDP] Physical Access Control
    wiegand = 4390, 'wiegand', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] American Printware IMServer Protocol
    apwi_imserver = 4391, 'apwi-imserver', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4391 = 4391, 'reserved', TransportProtocol.get('udp')

    #: [TCP] American Printware RXServer Protocol
    apwi_rxserver = 4392, 'apwi-rxserver', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4392 = 4392, 'reserved', TransportProtocol.get('udp')

    #: [TCP] American Printware RXSpooler Protocol
    apwi_rxspooler = 4393, 'apwi-rxspooler', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4393 = 4393, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4394 = 4394, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] American Printware Discovery
    apwi_disc = 4394, 'apwi-disc', TransportProtocol.get('udp')

    #: - [TCP] OmniVision communication for Virtual environments
    #: - [UDP] OmniVision communication for Virtual environments
    omnivisionesx = 4395, 'omnivisionesx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Fly Object Space
    fly = 4396, 'fly', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4396 = 4396, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ASIGRA Services
    #: - [UDP] ASIGRA Services
    ds_srv = 4400, 'ds-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASIGRA Televaulting DS-System Service
    #: - [UDP] ASIGRA Televaulting DS-System Service
    ds_srvr = 4401, 'ds-srvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASIGRA Televaulting DS-Client Service
    #: - [UDP] ASIGRA Televaulting DS-Client Service
    ds_clnt = 4402, 'ds-clnt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASIGRA Televaulting DS-Client Monitoring/Management
    #: - [UDP] ASIGRA Televaulting DS-Client Monitoring/Management
    ds_user = 4403, 'ds-user', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASIGRA Televaulting DS-System Monitoring/Management
    #: - [UDP] ASIGRA Televaulting DS-System Monitoring/Management
    ds_admin = 4404, 'ds-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASIGRA Televaulting Message Level Restore service
    #: - [UDP] ASIGRA Televaulting Message Level Restore service
    ds_mail = 4405, 'ds-mail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASIGRA Televaulting DS-Sleeper Service
    #: - [UDP] ASIGRA Televaulting DS-Sleeper Service
    ds_slp = 4406, 'ds-slp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Network Access Control Agent
    nacagent = 4407, 'nacagent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4407 = 4407, 'reserved', TransportProtocol.get('udp')

    #: [TCP] SLS Technology Control Centre
    slscc = 4408, 'slscc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4408 = 4408, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Net-Cabinet comunication
    netcabinet_com = 4409, 'netcabinet-com', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4409 = 4409, 'reserved', TransportProtocol.get('udp')

    #: [TCP] RIB iTWO Application Server
    itwo_server = 4410, 'itwo-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4410 = 4410, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Found Messaging Protocol
    found = 4411, 'found', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4411 = 4411, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4412 = 4412, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] SmallChat
    smallchat = 4412, 'smallchat', TransportProtocol.get('udp')

    #: [TCP] AVI Systems NMS
    avi_nms = 4413, 'avi-nms', TransportProtocol.get('tcp')

    #: [UDP] AVI Systems NMS
    avi_nms_disc = 4413, 'avi-nms-disc', TransportProtocol.get('udp')

    #: [TCP] Updog Monitoring and Status Framework
    updog = 4414, 'updog', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4414 = 4414, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Brocade Virtual Router Request
    brcd_vr_req = 4415, 'brcd-vr-req', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4415 = 4415, 'reserved', TransportProtocol.get('udp')

    #: [TCP] PJJ Media Player
    pjj_player = 4416, 'pjj-player', TransportProtocol.get('tcp')

    #: [UDP] PJJ Media Player discovery
    pjj_player_disc = 4416, 'pjj-player-disc', TransportProtocol.get('udp')

    #: [TCP] Workflow Director Communication
    workflowdir = 4417, 'workflowdir', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4417 = 4417, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4418 = 4418, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] AXYS communication protocol
    axysbridge = 4418, 'axysbridge', TransportProtocol.get('udp')

    #: [TCP] Colnod Binary Protocol
    cbp = 4419, 'cbp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4419 = 4419, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] NVM Express over Fabrics storage access
    #: - [UDP] NVM Express over Fabrics storage access
    nvme = 4420, 'nvme', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Multi-Platform Remote Management for Cloud Infrastructure
    scaleft = 4421, 'scaleft', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4421 = 4421, 'reserved', TransportProtocol.get('udp')

    #: [TCP] TSEP Installation Service Protocol
    tsepisp = 4422, 'tsepisp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4422 = 4422, 'reserved', TransportProtocol.get('udp')

    #: [TCP] thingkit secure mesh
    thingkit = 4423, 'thingkit', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4423 = 4423, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_4424 = 4424, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] NetROCKEY6 SMART Plus Service
    #: - [UDP] NetROCKEY6 SMART Plus Service
    netrockey6 = 4425, 'netrockey6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMARTS Beacon Port
    #: - [UDP] SMARTS Beacon Port
    beacon_port_2 = 4426, 'beacon-port-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Drizzle database server
    drizzle = 4427, 'drizzle', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4427 = 4427, 'reserved', TransportProtocol.get('udp')

    #: [TCP] OMV-Investigation Server-Client
    omviserver = 4428, 'omviserver', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4428 = 4428, 'reserved', TransportProtocol.get('udp')

    #: [TCP] OMV Investigation Agent-Server
    omviagent = 4429, 'omviagent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4429 = 4429, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] REAL SQL Server
    #: - [UDP] REAL SQL Server
    rsqlserver = 4430, 'rsqlserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] adWISE Pipe
    wspipe = 4431, 'wspipe', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4431 = 4431, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] L-ACOUSTICS management
    #: - [UDP] L-ACOUSTICS management
    l_acoustics = 4432, 'l-acoustics', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Versile Object Protocol
    vop = 4433, 'vop', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4433 = 4433, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4441 = 4441, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Netblox Protocol
    netblox = 4441, 'netblox', TransportProtocol.get('udp')

    #: - [TCP] Saris
    #: - [UDP] Saris
    saris = 4442, 'saris', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pharos
    #: - [UDP] Pharos
    pharos = 4443, 'pharos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] KRB524
    #: - [UDP] KRB524
    krb524 = 4444, 'krb524', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NV Video default
    #: - [UDP] NV Video default
    nv_video = 4444, 'nv-video', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UPNOTIFYP
    #: - [UDP] UPNOTIFYP
    upnotifyp = 4445, 'upnotifyp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] N1-FWP
    #: - [UDP] N1-FWP
    n1_fwp = 4446, 'n1-fwp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] N1-RMGMT
    #: - [UDP] N1-RMGMT
    n1_rmgmt = 4447, 'n1-rmgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ASC Licence Manager
    #: - [UDP] ASC Licence Manager
    asc_slmd = 4448, 'asc-slmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PrivateWire
    #: - [UDP] PrivateWire
    privatewire = 4449, 'privatewire', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Common ASCII Messaging Protocol
    #: - [UDP] Common ASCII Messaging Protocol
    camp = 4450, 'camp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CTI System Msg
    #: - [UDP] CTI System Msg
    ctisystemmsg = 4451, 'ctisystemmsg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CTI Program Load
    #: - [UDP] CTI Program Load
    ctiprogramload = 4452, 'ctiprogramload', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NSS Alert Manager
    #: - [UDP] NSS Alert Manager
    nssalertmgr = 4453, 'nssalertmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NSS Agent Manager
    #: - [UDP] NSS Agent Manager
    nssagentmgr = 4454, 'nssagentmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PR Chat User
    #: - [UDP] PR Chat User
    prchat_user = 4455, 'prchat-user', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PR Chat Server
    #: - [UDP] PR Chat Server
    prchat_server = 4456, 'prchat-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PR Register
    #: - [UDP] PR Register
    prregister = 4457, 'prregister', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Matrix Configuration Protocol
    #: - [UDP] Matrix Configuration Protocol
    mcp = 4458, 'mcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_4459 = 4459, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Network Time Security Key Establishment [:rfc:`8915`]
    ntske = 4460, 'ntske', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4460 = 4460, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] hpssmgmt service
    #: - [UDP] hpssmgmt service
    hpssmgmt = 4484, 'hpssmgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Assyst Data Repository Service
    assyst_dr = 4485, 'assyst-dr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4485 = 4485, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Integrated Client Message Service
    #: - [UDP] Integrated Client Message Service
    icms = 4486, 'icms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Protocol for Remote Execution over TCP
    prex_tcp = 4487, 'prex-tcp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4487 = 4487, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Apple Wide Area Connectivity Service ICE Bootstrap
    #: - [UDP] Apple Wide Area Connectivity Service ICE Bootstrap
    awacs_ice = 4488, 'awacs-ice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPsec NAT-Traversal [:rfc:`9329`]
    #: - [UDP] IPsec NAT-Traversal [:rfc:`3948`][:rfc:`7296`]
    ipsec_nat_t = 4500, 'ipsec-nat-t', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_4501 = 4501, 'unassigned', TransportProtocol.get('undefined')

    #: [SCTP] A25 (FAP-FGW)
    a25_fap_fgw = 4502, 'a25-fap-fgw', TransportProtocol.get('sctp')

    #: [TCP] Reserved
    reserved_4534 = 4534, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Armagetron Advanced Game Server
    armagetronad = 4534, 'armagetronad', TransportProtocol.get('udp')

    #: - [TCP] Event Heap Server
    #: - [UDP] Event Heap Server
    ehs = 4535, 'ehs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Event Heap Server SSL
    #: - [UDP] Event Heap Server SSL
    ehs_ssl = 4536, 'ehs-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WSS Security Service
    #: - [UDP] WSS Security Service
    wssauthsvc = 4537, 'wssauthsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Software Data Exchange Gateway
    #: - [UDP] Software Data Exchange Gateway
    swx_gate = 4538, 'swx-gate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WorldScores
    #: - [UDP] WorldScores
    worldscores = 4545, 'worldscores', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SF License Manager (Sentinel)
    #: - [UDP] SF License Manager (Sentinel)
    sf_lm = 4546, 'sf-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lanner License Manager
    #: - [UDP] Lanner License Manager
    lanner_lm = 4547, 'lanner-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Synchromesh
    #: - [UDP] Synchromesh
    synchromesh = 4548, 'synchromesh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Aegate PMR Service
    #: - [UDP] Aegate PMR Service
    aegate = 4549, 'aegate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Perman I Interbase Server
    #: - [UDP] Perman I Interbase Server
    gds_adppiw_db = 4550, 'gds-adppiw-db', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MIH Services [:rfc:`5677`]
    #: - [UDP] MIH Services [:rfc:`5677`]
    ieee_mih = 4551, 'ieee-mih', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Men and Mice Monitoring
    #: - [UDP] Men and Mice Monitoring
    menandmice_mon = 4552, 'menandmice-mon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ICS host services
    icshostsvc = 4553, 'icshostsvc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4553 = 4553, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] MS FRS Replication
    #: - [UDP] MS FRS Replication
    msfrs = 4554, 'msfrs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RSIP Port [:rfc:`3103`]
    #: - [UDP] RSIP Port [:rfc:`3103`]
    rsip = 4555, 'rsip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DTN Bundle TCP CL Protocol [:rfc:`9174`]
    #: - [UDP] DTN Bundle UDP CL Protocol [:rfc:`7122`]
    #: - [DCCP] DTN Bundle DCCP CL Protocol [:rfc:`7122`]
    dtn_bundle = 4556, 'dtn-bundle', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('dccp')

    #: [TCP] Reserved
    reserved_4557 = 4557, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Marathon everRun Quorum Service Server
    mtcevrunqss = 4557, 'mtcevrunqss', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4558 = 4558, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Marathon everRun Quorum Service Manager
    mtcevrunqman = 4558, 'mtcevrunqman', TransportProtocol.get('udp')

    #: - [TCP] HylaFAX
    #: - [UDP] HylaFAX
    hylafax = 4559, 'hylafax', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Amahi Anywhere
    amahi_anywhere = 4563, 'amahi-anywhere', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4563 = 4563, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Kids Watch Time Control Service
    #: - [UDP] Kids Watch Time Control Service
    kwtc = 4566, 'kwtc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRAM
    #: - [UDP] TRAM
    tram = 4567, 'tram', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC Reporting
    #: - [UDP] BMC Reporting
    bmc_reporting = 4568, 'bmc-reporting', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inter-Asterisk eXchange [:rfc:`5456`]
    #: - [UDP] Inter-Asterisk eXchange [:rfc:`5456`]
    iax = 4569, 'iax', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Service to distribute and update within a site deployment information
    #: for Oracle Communications Suite
    deploymentmap = 4570, 'deploymentmap', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4570 = 4570, 'reserved', TransportProtocol.get('udp')

    #: [TCP] A port for communication between a server and client for a custom
    #: backup system
    cardifftec_back = 4573, 'cardifftec-back', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4573 = 4573, 'reserved', TransportProtocol.get('udp')

    #: [TCP] RID over HTTP/TLS [:rfc:`6546`]
    rid = 4590, 'rid', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4590 = 4590, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] HRPD L3T (AT-AN)
    #: - [UDP] HRPD L3T (AT-AN)
    l3t_at_an = 4591, 'l3t-at-an', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4592 = 4592, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] HRPD-ITH (AT-AN)
    hrpd_ith_at_an = 4592, 'hrpd-ith-at-an', TransportProtocol.get('udp')

    #: - [TCP] IPT (ANRI-ANRI)
    #: - [UDP] IPT (ANRI-ANRI)
    ipt_anri_anri = 4593, 'ipt-anri-anri', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IAS-Session (ANRI-ANRI)
    #: - [UDP] IAS-Session (ANRI-ANRI)
    ias_session = 4594, 'ias-session', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IAS-Paging (ANRI-ANRI)
    #: - [UDP] IAS-Paging (ANRI-ANRI)
    ias_paging = 4595, 'ias-paging', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IAS-Neighbor (ANRI-ANRI)
    #: - [UDP] IAS-Neighbor (ANRI-ANRI)
    ias_neighbor = 4596, 'ias-neighbor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A21 (AN-1xBS)
    #: - [UDP] A21 (AN-1xBS)
    a21_an_1xbs = 4597, 'a21-an-1xbs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A16 (AN-AN)
    #: - [UDP] A16 (AN-AN)
    a16_an_an = 4598, 'a16-an-an', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A17 (AN-AN)
    #: - [UDP] A17 (AN-AN)
    a17_an_an = 4599, 'a17-an-an', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Piranha1
    #: - [UDP] Piranha1
    piranha1 = 4600, 'piranha1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Piranha2
    #: - [UDP] Piranha2
    piranha2 = 4601, 'piranha2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] EAX MTS Server
    mtsserver = 4602, 'mtsserver', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4602 = 4602, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Men & Mice Upgrade Agent
    menandmice_upg = 4603, 'menandmice-upg', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4603 = 4603, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Identity Registration Protocol
    irp = 4604, 'irp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4604 = 4604, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Direct End to End Secure Chat Protocol
    sixchat = 4605, 'sixchat', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4605 = 4605, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Secure ID to IP registration and lookup
    sixid = 4606, 'sixid', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4606 = 4606, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4621 = 4621, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Bidirectional single port remote radio VOIP and Control stream
    ventoso = 4621, 'ventoso', TransportProtocol.get('udp')

    #: - [TCP] Distributed Denial-of-Service Open Threat Signaling (DOTS) Signal
    #: Channel Protocol. The service name is used to construct the SRV service
    #: names "_dots-signal._udp" and "_dots-signal._tcp" for discovering DOTS
    #: servers used to establish DOTS signal channel. [:rfc:`8973`][:rfc:`9132`]
    #: - [UDP] Distributed Denial-of-Service Open Threat Signaling (DOTS) Signal
    #: Channel Protocol. The service name is used to construct the SRV service
    #: names "_dots-signal._udp" and "_dots-signal._tcp" for discovering DOTS
    #: servers used to establish DOTS signal channel. [:rfc:`8973`][:rfc:`9132`]
    dots_signal = 4646, 'dots-signal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PlayStation2 App Port
    #: - [UDP] PlayStation2 App Port
    playsta2_app = 4658, 'playsta2-app', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PlayStation2 Lobby Port
    #: - [UDP] PlayStation2 Lobby Port
    playsta2_lob = 4659, 'playsta2-lob', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] smaclmgr
    #: - [UDP] smaclmgr
    smaclmgr = 4660, 'smaclmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kar2ouche Peer location service
    #: - [UDP] Kar2ouche Peer location service
    kar2ouche = 4661, 'kar2ouche', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OrbitNet Message Service
    #: - [UDP] OrbitNet Message Service
    oms = 4662, 'oms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Note It! Message Service
    #: - [UDP] Note It! Message Service
    noteit = 4663, 'noteit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Rimage Messaging Server
    #: - [UDP] Rimage Messaging Server
    ems = 4664, 'ems', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Container Client Message Service
    #: - [UDP] Container Client Message Service
    contclientms = 4665, 'contclientms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] E-Port Message Service
    #: - [UDP] E-Port Message Service
    eportcomm = 4666, 'eportcomm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MMA Comm Services
    #: - [UDP] MMA Comm Services
    mmacomm = 4667, 'mmacomm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MMA EDS Service
    #: - [UDP] MMA EDS Service
    mmaeds = 4668, 'mmaeds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] E-Port Data Service
    #: - [UDP] E-Port Data Service
    eportcommdata = 4669, 'eportcommdata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Light packets transfer protocol
    #: - [UDP] Light packets transfer protocol
    light = 4670, 'light', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bull RSF action server
    #: - [UDP] Bull RSF action server
    acter = 4671, 'acter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] remote file access server
    #: - [UDP] remote file access server
    rfa = 4672, 'rfa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CXWS Operations
    #: - [UDP] CXWS Operations
    cxws = 4673, 'cxws', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppIQ Agent Management
    #: - [UDP] AppIQ Agent Management
    appiq_mgmt = 4674, 'appiq-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BIAP Device Status
    #: - [UDP] BIAP Device Status
    dhct_status = 4675, 'dhct-status', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BIAP Generic Alert
    #: - [UDP] BIAP Generic Alert
    dhct_alerts = 4676, 'dhct-alerts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Business Continuity Servi
    #: - [UDP] Business Continuity Servi
    bcs = 4677, 'bcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] boundary traversal
    #: - [UDP] boundary traversal
    traversal = 4678, 'traversal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MGE UPS Supervision
    #: - [UDP] MGE UPS Supervision
    mgesupervision = 4679, 'mgesupervision', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MGE UPS Management
    #: - [UDP] MGE UPS Management
    mgemanagement = 4680, 'mgemanagement', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Parliant Telephony System
    #: - [UDP] Parliant Telephony System
    parliant = 4681, 'parliant', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] finisar
    #: - [UDP] finisar
    finisar = 4682, 'finisar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Spike Clipboard Service
    #: - [UDP] Spike Clipboard Service
    spike = 4683, 'spike', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RFID Reader Protocol 1.0
    #: - [UDP] RFID Reader Protocol 1.0
    rfid_rp1 = 4684, 'rfid-rp1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Autopac Protocol
    #: - [UDP] Autopac Protocol
    autopac = 4685, 'autopac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Manina Service Protocol
    #: - [UDP] Manina Service Protocol
    msp_os = 4686, 'msp-os', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Scanner Tool FTP
    #: - [UDP] Network Scanner Tool FTP
    nst = 4687, 'nst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mobile P2P Service
    #: - [UDP] Mobile P2P Service
    mobile_p2p = 4688, 'mobile-p2p', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Altova DatabaseCentral
    #: - [UDP] Altova DatabaseCentral
    altovacentral = 4689, 'altovacentral', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prelude IDS message proto
    #: - [UDP] Prelude IDS message proto
    prelude = 4690, 'prelude', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] monotone Netsync Protocol
    #: - [UDP] monotone Netsync Protocol
    mtn = 4691, 'mtn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Conspiracy messaging
    #: - [UDP] Conspiracy messaging
    conspiracy = 4692, 'conspiracy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetXMS Agent
    #: - [UDP] NetXMS Agent
    netxms_agent = 4700, 'netxms-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetXMS Management
    #: - [UDP] NetXMS Management
    netxms_mgmt = 4701, 'netxms-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetXMS Server Synchronization
    #: - [UDP] NetXMS Server Synchronization
    netxms_sync = 4702, 'netxms-sync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Network Performance Quality Evaluation System Test Service
    npqes_test = 4703, 'npqes-test', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4703 = 4703, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Assuria Insider
    assuria_ins = 4704, 'assuria-ins', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4704 = 4704, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Trinity Trust Network Node Communication
    #: - [UDP] Trinity Trust Network Node Communication
    #: - [SCTP] Trinity Trust Network Node Communication
    trinity_dist = 4711, 'trinity-dist', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] TruckStar Service
    #: - [UDP] TruckStar Service
    truckstar = 4725, 'truckstar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4726 = 4726, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] A26 (FAP-FGW)
    a26_fap_fgw = 4726, 'a26-fap-fgw', TransportProtocol.get('udp')

    #: [TCP] F-Link Client Information Service
    fcis = 4727, 'fcis', TransportProtocol.get('tcp')

    #: [UDP] F-Link Client Information Service Discovery
    fcis_disc = 4727, 'fcis-disc', TransportProtocol.get('udp')

    #: - [TCP] CA Port Multiplexer
    #: - [UDP] CA Port Multiplexer
    capmux = 4728, 'capmux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4729 = 4729, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] GSM Interface Tap
    gsmtap = 4729, 'gsmtap', TransportProtocol.get('udp')

    #: - [TCP] Gearman Job Queue System
    #: - [UDP] Gearman Job Queue System
    gearman = 4730, 'gearman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Remote Capture Protocol
    remcap = 4731, 'remcap', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4731 = 4731, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4732 = 4732, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] OHM server trigger
    ohmtrigger = 4732, 'ohmtrigger', TransportProtocol.get('udp')

    #: [TCP] RES Orchestration Catalog Services
    resorcs = 4733, 'resorcs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4733 = 4733, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] IPDR/SP
    #: - [UDP] IPDR/SP
    ipdr_sp = 4737, 'ipdr-sp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SoleraTec Locator
    #: - [UDP] SoleraTec Locator
    solera_lpn = 4738, 'solera-lpn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IP Flow Info Export
    #: - [UDP] IP Flow Info Export
    #: - [SCTP] IP Flow Info Export
    ipfix = 4739, 'ipfix', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] ipfix protocol over TLS
    #: - [SCTP] ipfix protocol over DTLS
    #: - [UDP] ipfix protocol over DTLS
    ipfixs = 4740, 'ipfixs', TransportProtocol.get('tcp') | TransportProtocol.get('sctp') | TransportProtocol.get('udp')

    #: - [TCP] Luminizer Manager
    #: - [UDP] Luminizer Manager
    lumimgrd = 4741, 'lumimgrd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SICCT
    sicct = 4742, 'sicct', TransportProtocol.get('tcp')

    #: [UDP] SICCT Service Discovery Protocol
    sicct_sdp = 4742, 'sicct-sdp', TransportProtocol.get('udp')

    #: - [TCP] openhpi HPI service
    #: - [UDP] openhpi HPI service
    openhpid = 4743, 'openhpid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Internet File Synchronization Protocol
    #: - [UDP] Internet File Synchronization Protocol
    ifsp = 4744, 'ifsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Funambol Mobile Push
    #: - [UDP] Funambol Mobile Push
    fmp = 4745, 'fmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4746 = 4746, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] IntelliAdmin Discovery
    intelliadm_disc = 4746, 'intelliadm-disc', TransportProtocol.get('udp')

    #: [UDP] peer-to-peer file exchange protocol
    buschtrommel = 4747, 'buschtrommel', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4747 = 4747, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Simple Service Auto Discovery
    #: - [UDP] Simple Service Auto Discovery
    ssad = 4750, 'ssad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Policy Control Protocol
    #: - [UDP] Simple Policy Control Protocol
    spocp = 4751, 'spocp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Network Audio Protocol
    #: - [UDP] Simple Network Audio Protocol
    snap = 4752, 'snap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Simple Invocation of Methods Over Network (SIMON)
    simon = 4753, 'simon', TransportProtocol.get('tcp')

    #: [UDP] Simple Invocation of Methods Over Network (SIMON) Discovery
    simon_disc = 4753, 'simon-disc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4754 = 4754, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] GRE-in-UDP Encapsulation [:rfc:`8086`]
    gre_in_udp = 4754, 'gre-in-udp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4755 = 4755, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] GRE-in-UDP Encapsulation with DTLS [:rfc:`8086`]
    gre_udp_dtls = 4755, 'gre-udp-dtls', TransportProtocol.get('udp')

    #: [TCP] Reticle Decision Center
    rdcenter = 4756, 'rdcenter', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4756 = 4756, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Converge RPC
    converge = 4774, 'converge', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4774 = 4774, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] BFD Multihop Control
    #: - [UDP] BFD Multihop Control
    bfd_multi_ctl = 4784, 'bfd-multi-ctl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4785 = 4785, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Cisco Nexus Control Protocol
    cncp = 4785, 'cncp', TransportProtocol.get('udp')

    #: [TCP] Smart Install Service
    smart_install = 4786, 'smart-install', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4786 = 4786, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Service Insertion Architecture (SIA) Control-Plane
    sia_ctrl_plane = 4787, 'sia-ctrl-plane', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4787 = 4787, 'reserved', TransportProtocol.get('udp')

    #: [TCP] eXtensible Messaging Client Protocol
    xmcp = 4788, 'xmcp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4788 = 4788, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Virtual eXtensible Local Area Network (VXLAN) [:rfc:`7348`]
    vxlan = 4789, 'vxlan', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4789 = 4789, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Generic Protocol Extension for Virtual eXtensible Local Area Network
    #: (VXLAN)
    vxlan_gpe = 4790, 'vxlan-gpe', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4790 = 4790, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] IP Routable RocE
    roce = 4791, 'roce', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4791 = 4791, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] IP Routable Unified Bus
    #: - [UDP] IP Routable Unified Bus
    unified_bus = 4792, 'unified-bus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Icona Instant Messenging System
    #: - [UDP] Icona Instant Messenging System
    iims = 4800, 'iims', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Icona Web Embedded Chat
    #: - [UDP] Icona Web Embedded Chat
    iwec = 4801, 'iwec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Icona License System Server
    #: - [UDP] Icona License System Server
    ilss = 4802, 'ilss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Notateit Messaging
    notateit = 4803, 'notateit', TransportProtocol.get('tcp')

    #: [UDP] Notateit Messaging Discovery
    notateit_disc = 4803, 'notateit-disc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4804 = 4804, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] AJA ntv4 Video System Discovery
    aja_ntv4_disc = 4804, 'aja-ntv4-disc', TransportProtocol.get('udp')

    #: - [TCP] HTCP
    #: - [UDP] HTCP
    htcp = 4827, 'htcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Varadero-0
    #: - [UDP] Varadero-0
    varadero_0 = 4837, 'varadero-0', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Varadero-1
    #: - [UDP] Varadero-1
    varadero_1 = 4838, 'varadero-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Varadero-2
    #: - [UDP] Varadero-2
    varadero_2 = 4839, 'varadero-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] OPC UA Connection Protocol
    opcua_tcp = 4840, 'opcua-tcp', TransportProtocol.get('tcp')

    #: [UDP] OPC UA Multicast Datagram Protocol
    opcua_udp = 4840, 'opcua-udp', TransportProtocol.get('udp')

    #: - [TCP] QUOSA Virtual Library Service
    #: - [UDP] QUOSA Virtual Library Service
    quosa = 4841, 'quosa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nCode ICE-flow Library AppServer
    #: - [UDP] nCode ICE-flow Library AppServer
    gw_asv = 4842, 'gw-asv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPC UA TCP Protocol over TLS/SSL
    #: - [UDP] OPC UA TCP Protocol over TLS/SSL
    opcua_tls = 4843, 'opcua-tls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nCode ICE-flow Library LogServer
    #: - [UDP] nCode ICE-flow Library LogServer
    gw_log = 4844, 'gw-log', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WordCruncher Remote Library Service
    #: - [UDP] WordCruncher Remote Library Service
    wcr_remlib = 4845, 'wcr-remlib', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Contamac ICM Service IANA assigned this well-formed service name as a
    #: replacement for "contamac_icm".
    #: - [TCP] Contamac ICM Service
    #: - [UDP] Contamac ICM Service IANA assigned this well-formed service name as a
    #: replacement for "contamac_icm".
    #: - [UDP] Contamac ICM Service
    contamac_icm = 4846, 'contamac-icm', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Web Fresh Communication
    #: - [UDP] Web Fresh Communication
    wfc = 4847, 'wfc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] App Server - Admin HTTP
    #: - [UDP] App Server - Admin HTTP
    appserv_http = 4848, 'appserv-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] App Server - Admin HTTPS
    #: - [UDP] App Server - Admin HTTPS
    appserv_https = 4849, 'appserv-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun App Server - NA
    #: - [UDP] Sun App Server - NA
    sun_as_nodeagt = 4850, 'sun-as-nodeagt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apache Derby Replication
    #: - [UDP] Apache Derby Replication
    derby_repli = 4851, 'derby-repli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Unify Debugger
    #: - [UDP] Unify Debugger
    unify_debug = 4867, 'unify-debug', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Photon Relay
    #: - [UDP] Photon Relay
    phrelay = 4868, 'phrelay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Photon Relay Debug
    #: - [UDP] Photon Relay Debug
    phrelaydbg = 4869, 'phrelaydbg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Citcom Tracking Service
    #: - [UDP] Citcom Tracking Service
    cc_tracking = 4870, 'cc-tracking', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Wired
    #: - [UDP] Wired
    wired = 4871, 'wired', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tritium CAN Bus Bridge Service
    #: - [UDP] Tritium CAN Bus Bridge Service
    tritium_can = 4876, 'tritium-can', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lighting Management Control System
    #: - [UDP] Lighting Management Control System
    lmcs = 4877, 'lmcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4878 = 4878, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Agilent Instrument Discovery
    inst_discovery = 4878, 'inst-discovery', TransportProtocol.get('udp')

    #: [TCP] WSDL Event Receiver
    wsdl_event = 4879, 'wsdl-event', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4879 = 4879, 'reserved', TransportProtocol.get('udp')

    #: [TCP] IVI High-Speed LAN Instrument Protocol
    hislip = 4880, 'hislip', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4880 = 4880, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4881 = 4881, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] SOCP Time Synchronization Protocol
    socp_t = 4881, 'socp-t', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4882 = 4882, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] SOCP Control Protocol
    socp_c = 4882, 'socp-c', TransportProtocol.get('udp')

    #: [TCP] Meier-Phelps License Server
    wmlserver = 4883, 'wmlserver', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4883 = 4883, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] HiveStor Distributed File System
    #: - [UDP] HiveStor Distributed File System
    hivestor = 4884, 'hivestor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ABBS
    #: - [UDP] ABBS
    abbs = 4885, 'abbs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] xcap code analysis portal public user access
    xcap_portal = 4888, 'xcap-portal', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4888 = 4888, 'reserved', TransportProtocol.get('udp')

    #: [TCP] xcap code analysis portal cluster control and administration
    xcap_control = 4889, 'xcap-control', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4889 = 4889, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] LysKOM Protocol A
    #: - [UDP] LysKOM Protocol A
    lyskom = 4894, 'lyskom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RAdmin Port
    #: - [UDP] RAdmin Port
    radmin_port = 4899, 'radmin-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HFSQL Client/Server Database Engine
    #: - [UDP] HFSQL Client/Server Database Engine
    hfcs = 4900, 'hfcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileLocator Remote Search Agent IANA assigned this well-formed service
    #: name as a replacement for "flr_agent".
    #: - [TCP] FileLocator Remote Search Agent
    flr_agent = 4901, 'flr-agent', TransportProtocol.get('tcp') | TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4901 = 4901, 'reserved', TransportProtocol.get('udp')

    #: [TCP] magicCONROL RF and Data Interface
    magiccontrol = 4902, 'magiccontrol', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4902 = 4902, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Technicolor LUT Access Protocol
    lutap = 4912, 'lutap', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4912 = 4912, 'reserved', TransportProtocol.get('udp')

    #: [TCP] LUTher Control Protocol
    lutcp = 4913, 'lutcp', TransportProtocol.get('tcp')

    #: - [TCP] Bones Remote Control
    #: - [UDP] Bones Remote Control
    bones = 4914, 'bones', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Fibics Remote Control Service
    frcs = 4915, 'frcs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4915 = 4915, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Signal protocol port for autonomic networking
    an_signaling = 4936, 'an-signaling', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4936 = 4936, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] Reserved
    reserved_4937 = 4937, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] ATSC-M/H Service Signaling Channel
    atsc_mh_ssc = 4937, 'atsc-mh-ssc', TransportProtocol.get('udp')

    #: - [TCP] Equitrac Office
    #: - [UDP] Equitrac Office
    eq_office_4940 = 4940, 'eq-office-4940', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Equitrac Office
    #: - [UDP] Equitrac Office
    eq_office_4941 = 4941, 'eq-office-4941', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Equitrac Office
    #: - [UDP] Equitrac Office
    eq_office_4942 = 4942, 'eq-office-4942', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Munin Graphing Framework
    #: - [UDP] Munin Graphing Framework
    munin = 4949, 'munin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sybase Server Monitor
    #: - [UDP] Sybase Server Monitor
    sybasesrvmon = 4950, 'sybasesrvmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PWG WIMS
    #: - [UDP] PWG WIMS
    pwgwims = 4951, 'pwgwims', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SAG Directory Server
    #: - [UDP] SAG Directory Server
    sagxtsds = 4952, 'sagxtsds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Synchronization Arbiter
    dbsyncarbiter = 4953, 'dbsyncarbiter', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4953 = 4953, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] CCSS QMessageMonitor
    #: - [UDP] CCSS QMessageMonitor
    ccss_qmm = 4969, 'ccss-qmm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CCSS QSystemMonitor
    #: - [UDP] CCSS QSystemMonitor
    ccss_qsm = 4970, 'ccss-qsm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] BackUp and Restore Program
    burp = 4971, 'burp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4971 = 4971, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Citrix Virtual Path
    ctxs_vpp = 4980, 'ctxs-vpp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_4980 = 4980, 'reserved', TransportProtocol.get('tcp')

    #: [N/A] Unassigned
    unassigned_4983 = 4983, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] WebYast
    webyast = 4984, 'webyast', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4984 = 4984, 'reserved', TransportProtocol.get('udp')

    #: [TCP] GER HC Standard
    gerhcs = 4985, 'gerhcs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_4985 = 4985, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Model Railway Interface Program
    #: - [UDP] Model Railway Interface Program
    mrip = 4986, 'mrip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMAR Ethernet Port 1
    #: - [UDP] SMAR Ethernet Port 1
    smar_se_port1 = 4987, 'smar-se-port1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMAR Ethernet Port 2
    #: - [UDP] SMAR Ethernet Port 2
    smar_se_port2 = 4988, 'smar-se-port2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Parallel for GAUSS (tm)
    #: - [UDP] Parallel for GAUSS (tm)
    parallel = 4989, 'parallel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BusySync Calendar Synch. Protocol
    #: - [UDP] BusySync Calendar Synch. Protocol
    busycal = 4990, 'busycal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VITA Radio Transport
    #: - [UDP] VITA Radio Transport
    vrt = 4991, 'vrt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HFSQL Client/Server Database Engine Manager
    #: - [UDP] HFSQL Client/Server Database Engine Manager
    hfcs_manager = 4999, 'hfcs-manager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    commplex_main = 5000, 'commplex-main', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    commplex_link = 5001, 'commplex-link', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] radio free ethernet
    #: - [UDP] radio free ethernet
    rfe = 5002, 'rfe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileMaker, Inc. - Proprietary transport
    #: - [UDP] FileMaker, Inc. - Proprietary name binding
    fmpro_internal = 5003, 'fmpro-internal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RTP media data [:rfc:`3551`][:rfc:`4571`]
    #: - [UDP] RTP media data [:rfc:`3551`]
    #: - [DCCP] RTP media data [:rfc:`3551`][:rfc:`5762`]
    avt_profile_1 = 5004, 'avt-profile-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('dccp')

    #: - [TCP] RTP control protocol [:rfc:`3551`][:rfc:`4571`]
    #: - [UDP] RTP control protocol [:rfc:`3551`]
    #: - [DCCP] RTP control protocol [:rfc:`3551`][:rfc:`5762`]
    avt_profile_2 = 5005, 'avt-profile-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('dccp')

    #: - [TCP] wsm server
    #: - [UDP] wsm server
    wsm_server = 5006, 'wsm-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] wsm server ssl
    #: - [UDP] wsm server ssl
    wsm_server_ssl = 5007, 'wsm-server-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Synapsis EDGE
    #: - [UDP] Synapsis EDGE
    synapsis_edge = 5008, 'synapsis-edge', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Windows Filesystem
    #: - [UDP] Microsoft Windows Filesystem
    winfs = 5009, 'winfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TelepathStart
    #: - [UDP] TelepathStart
    telelpathstart = 5010, 'telelpathstart', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TelepathAttack
    #: - [UDP] TelepathAttack
    telelpathattack = 5011, 'telelpathattack', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetOnTap Service
    #: - [UDP] NetOnTap Service
    nsp = 5012, 'nsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileMaker, Inc. - Proprietary transport
    #: - [UDP] FileMaker, Inc. - Proprietary transport
    fmpro_v6 = 5013, 'fmpro-v6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5014 = 5014, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Overlay Network Protocol
    onpsocket = 5014, 'onpsocket', TransportProtocol.get('udp')

    #: [TCP] FileMaker, Inc. - Web publishing
    fmwp = 5015, 'fmwp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5015 = 5015, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] zenginkyo-1
    #: - [UDP] zenginkyo-1
    zenginkyo_1 = 5020, 'zenginkyo-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] zenginkyo-2
    #: - [UDP] zenginkyo-2
    zenginkyo_2 = 5021, 'zenginkyo-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mice server
    #: - [UDP] mice server
    mice = 5022, 'mice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Htuil Server for PLD2
    #: - [UDP] Htuil Server for PLD2
    htuilsrv = 5023, 'htuilsrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Storix I/O daemon (data)
    #: - [UDP] Storix I/O daemon (data)
    strexec_d = 5026, 'strexec-d', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Storix I/O daemon (stat)
    #: - [UDP] Storix I/O daemon (stat)
    strexec_s = 5027, 'strexec-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Quiqum Virtual Relais
    qvr = 5028, 'qvr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5028 = 5028, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Infobright Database Server
    #: - [UDP] Infobright Database Server
    infobright = 5029, 'infobright', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_5030 = 5030, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5031 = 5031, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Direct Message Protocol
    dmp = 5031, 'dmp', TransportProtocol.get('udp')

    #: [TCP] SignaCert Enterprise Trust Server Agent
    signacert_agent = 5032, 'signacert-agent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5032 = 5032, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Janstor Secure Data
    jtnetd_server = 5033, 'jtnetd-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5033 = 5033, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Janstor Status
    jtnetd_status = 5034, 'jtnetd-status', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5034 = 5034, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] asnaacceler8db
    #: - [UDP] asnaacceler8db
    asnaacceler8db = 5042, 'asnaacceler8db', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ShopWorX Administration
    #: - [UDP] ShopWorX Administration
    swxadmin = 5043, 'swxadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LXI Event Service
    #: - [UDP] LXI Event Service
    lxi_evntsvc = 5044, 'lxi-evntsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Open Settlement Protocol
    osp = 5045, 'osp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5045 = 5045, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5046 = 5046, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Vishay PM UDP Service
    vpm_udp = 5046, 'vpm-udp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5047 = 5047, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] iSCAPE Data Broadcasting
    iscape = 5047, 'iscape', TransportProtocol.get('udp')

    #: [TCP] Texai Message Service
    texai = 5048, 'texai', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5048 = 5048, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] iVocalize Web Conference
    #: - [UDP] iVocalize Web Conference
    ivocalize = 5049, 'ivocalize', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] multimedia conference control tool
    #: - [UDP] multimedia conference control tool
    mmcc = 5050, 'mmcc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITA Agent
    #: - [UDP] ITA Agent
    ita_agent = 5051, 'ita-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITA Manager
    #: - [UDP] ITA Manager
    ita_manager = 5052, 'ita-manager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] RLM License Server
    rlm = 5053, 'rlm', TransportProtocol.get('tcp')

    #: [UDP] RLM Discovery Server
    rlm_disc = 5053, 'rlm-disc', TransportProtocol.get('udp')

    #: [TCP] RLM administrative interface
    rlm_admin = 5054, 'rlm-admin', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5054 = 5054, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] UNOT
    #: - [UDP] UNOT
    unot = 5055, 'unot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intecom Pointspan 1
    #: - [UDP] Intecom Pointspan 1
    intecom_ps1 = 5056, 'intecom-ps1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intecom Pointspan 2
    #: - [UDP] Intecom Pointspan 2
    intecom_ps2 = 5057, 'intecom-ps2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5058 = 5058, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Locus Discovery
    locus_disc = 5058, 'locus-disc', TransportProtocol.get('udp')

    #: - [TCP] SIP Directory Services
    #: - [UDP] SIP Directory Services
    sds = 5059, 'sds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SIP [:rfc:`3263`]
    #: - [UDP] SIP [:rfc:`3263`]
    #: - [SCTP] SIP [:rfc:`4168`]
    sip = 5060, 'sip', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] SIP-TLS [:rfc:`3263`]
    #: - [UDP] SIP-TLS [:rfc:`3263`]
    #: - [SCTP] SIP-TLS [:rfc:`4168`]
    sips = 5061, 'sips', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Localisation access
    #: - [UDP] Localisation access
    na_localise = 5062, 'na-localise', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] centrify secure RPC
    csrpc = 5063, 'csrpc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5063 = 5063, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Channel Access 1
    #: - [UDP] Channel Access 1
    ca_1 = 5064, 'ca-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Channel Access 2
    #: - [UDP] Channel Access 2
    ca_2 = 5065, 'ca-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STANAG-5066-SUBNET-INTF
    #: - [UDP] STANAG-5066-SUBNET-INTF
    stanag_5066 = 5066, 'stanag-5066', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Authentx Service
    #: - [UDP] Authentx Service
    authentx = 5067, 'authentx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Bitforest Data Service
    bitforestsrv = 5068, 'bitforestsrv', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5068 = 5068, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] I/Net 2000-NPR
    #: - [UDP] I/Net 2000-NPR
    i_net_2000_npr = 5069, 'i-net-2000-npr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VersaTrans Server Agent Service
    #: - [UDP] VersaTrans Server Agent Service
    vtsas = 5070, 'vtsas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PowerSchool
    #: - [UDP] PowerSchool
    powerschool = 5071, 'powerschool', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Anything In Anything
    #: - [UDP] Anything In Anything
    ayiya = 5072, 'ayiya', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Advantage Group Port Mgr
    #: - [UDP] Advantage Group Port Mgr
    tag_pm = 5073, 'tag-pm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ALES Query
    #: - [UDP] ALES Query
    alesquery = 5074, 'alesquery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Reserved
    reserved_5075 = 5075, 'reserved', TransportProtocol.get('udp')

    #: [UDP] PixelPusher pixel data
    pixelpusher = 5078, 'pixelpusher', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5078 = 5078, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] Reserved
    reserved_5079 = 5079, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Cambridge Pixel SPx Reports
    cp_spxrpts = 5079, 'cp-spxrpts', TransportProtocol.get('udp')

    #: - [TCP] OnScreen Data Collection Service
    #: - [UDP] OnScreen Data Collection Service
    onscreen = 5080, 'onscreen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SDL - Ent Trans Server
    #: - [UDP] SDL - Ent Trans Server
    sdl_ets = 5081, 'sdl-ets', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Qpur Communication Protocol
    #: - [UDP] Qpur Communication Protocol
    qcp = 5082, 'qcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Qpur File Protocol
    #: - [UDP] Qpur File Protocol
    qfp = 5083, 'qfp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPCglobal Low-Level Reader Protocol
    #: - [UDP] EPCglobal Low-Level Reader Protocol
    llrp = 5084, 'llrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EPCglobal Encrypted LLRP
    #: - [UDP] EPCglobal Encrypted LLRP
    encrypted_llrp = 5085, 'encrypted-llrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Aprigo Collection Service
    aprigo_cs = 5086, 'aprigo-cs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5086 = 5086, 'reserved', TransportProtocol.get('udp')

    #: [TCP] BIOTIC - Binary Internet of Things Interoperable Communication
    biotic = 5087, 'biotic', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5087 = 5087, 'reserved', TransportProtocol.get('udp')

    #: [SCTP] Candidate AR
    car = 5090, 'car', TransportProtocol.get('sctp')

    #: [SCTP] Context Transfer Protocol [:rfc:`4065`]
    cxtp = 5091, 'cxtp', TransportProtocol.get('sctp')

    #: [TCP] Reserved
    reserved_5092 = 5092, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Magpie Binary
    magpie = 5092, 'magpie', TransportProtocol.get('udp')

    #: - [TCP] Sentinel LM
    #: - [UDP] Sentinel LM
    sentinel_lm = 5093, 'sentinel-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HART-IP
    #: - [UDP] HART-IP
    hart_ip = 5094, 'hart-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SentLM Srv2Srv
    #: - [UDP] SentLM Srv2Srv
    sentlm_srv2srv = 5099, 'sentlm-srv2srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Socalia service mux
    #: - [UDP] Socalia service mux
    socalia = 5100, 'socalia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Talarian_TCP
    talarian_tcp = 5101, 'talarian-tcp', TransportProtocol.get('tcp')

    #: [UDP] Talarian_UDP
    talarian_udp = 5101, 'talarian-udp', TransportProtocol.get('udp')

    #: - [TCP] Oracle OMS non-secure
    #: - [UDP] Oracle OMS non-secure
    oms_nonsecure = 5102, 'oms-nonsecure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Actifio C2C
    actifio_c2c = 5103, 'actifio-c2c', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5103 = 5103, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5104 = 5104, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] TinyMessage
    tinymessage = 5104, 'tinymessage', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5105 = 5105, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Hughes Association Protocol
    hughes_ap = 5105, 'hughes-ap', TransportProtocol.get('udp')

    #: [TCP] Actifio UDS Agent
    actifioudsagent = 5106, 'actifioudsagent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5106 = 5106, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Disk to Disk replication between Actifio Clusters
    actifioreplic = 5107, 'actifioreplic', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5107 = 5107, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] TAEP AS service
    #: - [UDP] TAEP AS service
    taep_as_svc = 5111, 'taep-as-svc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PeerMe Msg Cmd Service
    #: - [UDP] PeerMe Msg Cmd Service
    pm_cmdsvr = 5112, 'pm-cmdsvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_5113 = 5113, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Enterprise Vault Services
    ev_services = 5114, 'ev-services', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5114 = 5114, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Symantec Autobuild Service
    autobuild = 5115, 'autobuild', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5115 = 5115, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5116 = 5116, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] EPSON Projecter Image Transfer
    emb_proj_cmd = 5116, 'emb-proj-cmd', TransportProtocol.get('udp')

    #: [TCP] GradeCam Image Processing
    gradecam = 5117, 'gradecam', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5117 = 5117, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Barracuda Backup Protocol
    #: - [UDP] Barracuda Backup Protocol
    barracuda_bbs = 5120, 'barracuda-bbs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Policy Commander
    #: - [UDP] Policy Commander
    nbt_pc = 5133, 'nbt-pc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PP ActivationServer
    ppactivation = 5134, 'ppactivation', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5134 = 5134, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ERP-Scale
    erp_scale = 5135, 'erp-scale', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5135 = 5135, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5136 = 5136, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Minotaur SA
    minotaur_sa = 5136, 'minotaur-sa', TransportProtocol.get('udp')

    #: - [TCP] MyCTS server port
    #: - [UDP] MyCTS server port
    ctsd = 5137, 'ctsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RMONITOR SECURE IANA assigned this well-formed service name as a
    #: replacement for "rmonitor_secure".
    #: - [TCP] RMONITOR SECURE
    #: - [UDP] RMONITOR SECURE IANA assigned this well-formed service name as a
    #: replacement for "rmonitor_secure".
    #: - [UDP] RMONITOR SECURE
    rmonitor_secure = 5145, 'rmonitor-secure', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [TCP] Social Alarm Service
    social_alarm = 5146, 'social-alarm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5146 = 5146, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Ascend Tunnel Management Protocol
    #: - [UDP] Ascend Tunnel Management Protocol
    atmp = 5150, 'atmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ESRI SDE Instance IANA assigned this well-formed service name as a
    #: replacement for "esri_sde".
    #: - [TCP] ESRI SDE Instance
    #: - [UDP] ESRI SDE Remote Start IANA assigned this well-formed service name as a
    #: replacement for "esri_sde".
    #: - [UDP] ESRI SDE Remote Start
    esri_sde = 5151, 'esri-sde', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] ESRI SDE Instance Discovery
    #: - [UDP] ESRI SDE Instance Discovery
    sde_discovery = 5152, 'sde-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_5153 = 5153, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BZFlag game server
    #: - [UDP] BZFlag game server
    bzflag = 5154, 'bzflag', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle asControl Agent
    #: - [UDP] Oracle asControl Agent
    asctrl_agent = 5155, 'asctrl-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Russian Online Game
    rugameonline = 5156, 'rugameonline', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5156 = 5156, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Mediat Remote Object Exchange
    mediat = 5157, 'mediat', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5157 = 5157, 'reserved', TransportProtocol.get('udp')

    #: [TCP] SNMP over SSH Transport Model [:rfc:`5592`]
    snmpssh = 5161, 'snmpssh', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5161 = 5161, 'reserved', TransportProtocol.get('udp')

    #: [TCP] SNMP Notification over SSH Transport Model [:rfc:`5592`]
    snmpssh_trap = 5162, 'snmpssh-trap', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5162 = 5162, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Shadow Backup
    sbackup = 5163, 'sbackup', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5163 = 5163, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Virtual Protocol Adapter
    vpa = 5164, 'vpa', TransportProtocol.get('tcp')

    #: [UDP] Virtual Protocol Adapter Discovery
    vpa_disc = 5164, 'vpa-disc', TransportProtocol.get('udp')

    #: - [TCP] ife_1corp IANA assigned this well-formed service name as a replacement
    #: for "ife_icorp".
    #: - [TCP] ife_1corp
    #: - [UDP] ife_1corp IANA assigned this well-formed service name as a replacement
    #: for "ife_icorp".
    #: - [UDP] ife_1corp
    ife_icorp = 5165, 'ife-icorp', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] WinPCS Service Connection
    #: - [UDP] WinPCS Service Connection
    winpcs = 5166, 'winpcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCTE104 Connection
    #: - [UDP] SCTE104 Connection
    scte104 = 5167, 'scte104', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCTE30 Connection
    #: - [UDP] SCTE30 Connection
    scte30 = 5168, 'scte30', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PC over IP Endpoint Management
    pcoip_mgmt = 5172, 'pcoip-mgmt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5172 = 5172, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] America-Online
    #: - [UDP] America-Online
    aol = 5190, 'aol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AmericaOnline1
    #: - [UDP] AmericaOnline1
    aol_1 = 5191, 'aol-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AmericaOnline2
    #: - [UDP] AmericaOnline2
    aol_2 = 5192, 'aol-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AmericaOnline3
    #: - [UDP] AmericaOnline3
    aol_3 = 5193, 'aol-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CipherPoint Config Service
    cpscomm = 5194, 'cpscomm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5194 = 5194, 'reserved', TransportProtocol.get('udp')

    #: [TCP] The protocol is used by a license server and client programs to
    #: control use of program licenses that float to networked machines
    ampl_lic = 5195, 'ampl-lic', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5195 = 5195, 'reserved', TransportProtocol.get('udp')

    #: [TCP] The protocol is used by two programs that exchange "table" data used
    #: in the AMPL modeling language
    ampl_tableproxy = 5196, 'ampl-tableproxy', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5196 = 5196, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Tunstall Lone worker device interface
    tunstall_lwp = 5197, 'tunstall-lwp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5197 = 5197, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] TARGUS GetData
    #: - [UDP] TARGUS GetData
    targus_getdata = 5200, 'targus-getdata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TARGUS GetData 1
    #: - [UDP] TARGUS GetData 1
    targus_getdata1 = 5201, 'targus-getdata1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TARGUS GetData 2
    #: - [UDP] TARGUS GetData 2
    targus_getdata2 = 5202, 'targus-getdata2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TARGUS GetData 3
    #: - [UDP] TARGUS GetData 3
    targus_getdata3 = 5203, 'targus-getdata3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Nomad Device Video Transfer
    nomad = 5209, 'nomad', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5209 = 5209, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] NOTEZA Data Safety Service
    #: - [SCTP] NOTEZA Data Safety Service
    noteza = 5215, 'noteza', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_5215 = 5215, 'reserved', TransportProtocol.get('udp')

    #: [TCP] 3eTI Extensible Management Protocol for OAMP
    AppType_3exmp = 5221, '3exmp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5221 = 5221, 'reserved', TransportProtocol.get('udp')

    #: [TCP] XMPP Client Connection [:rfc:`6120`]
    xmpp_client = 5222, 'xmpp-client', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5222 = 5222, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] HP Virtual Machine Group Management
    #: - [UDP] HP Virtual Machine Group Management
    hpvirtgrp = 5223, 'hpvirtgrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP Virtual Machine Console Operations
    #: - [UDP] HP Virtual Machine Console Operations
    hpvirtctrl = 5224, 'hpvirtctrl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP Server
    #: - [UDP] HP Server
    hp_server = 5225, 'hp-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP Status
    #: - [UDP] HP Status
    hp_status = 5226, 'hp-status', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP System Performance Metric Service
    #: - [UDP] HP System Performance Metric Service
    perfd = 5227, 'perfd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] HP Virtual Room Service
    hpvroom = 5228, 'hpvroom', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5228 = 5228, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Netflow/IPFIX/sFlow Collector and Forwarder Management
    jaxflow = 5229, 'jaxflow', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5229 = 5229, 'reserved', TransportProtocol.get('udp')

    #: [TCP] JaxMP RealFlow application and protocol data
    jaxflow_data = 5230, 'jaxflow-data', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5230 = 5230, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Remote Control of Scan Software for Cruse Scanners
    crusecontrol = 5231, 'crusecontrol', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5231 = 5231, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Cruse Scanning System Service
    csedaemon = 5232, 'csedaemon', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5232 = 5232, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Etinnae Network File Service
    enfs = 5233, 'enfs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5233 = 5233, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] EEnet communications
    #: - [UDP] EEnet communications
    eenet = 5234, 'eenet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Galaxy Network Service
    #: - [UDP] Galaxy Network Service
    galaxy_network = 5235, 'galaxy-network', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    padl2sim = 5236, 'padl2sim', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] m-net discovery
    #: - [UDP] m-net discovery
    mnet_discovery = 5237, 'mnet-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ATTUne API
    attune = 5242, 'attune', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5242 = 5242, 'reserved', TransportProtocol.get('udp')

    #: [TCP] xyClient Status API and rendevous point
    xycstatus = 5243, 'xycstatus', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5243 = 5243, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_5244 = 5244, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] DownTools Control Protocol
    downtools = 5245, 'downtools', TransportProtocol.get('tcp')

    #: [UDP] DownTools Discovery Protocol
    downtools_disc = 5245, 'downtools-disc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5246 = 5246, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] CAPWAP Control Protocol [:rfc:`5415`]
    capwap_control = 5246, 'capwap-control', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5247 = 5247, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] CAPWAP Data Protocol [:rfc:`5415`]
    capwap_data = 5247, 'capwap-data', TransportProtocol.get('udp')

    #: - [TCP] CA Access Control Web Service
    #: - [UDP] CA Access Control Web Service
    caacws = 5248, 'caacws', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA AC Lang Service
    #: - [UDP] CA AC Lang Service
    caaclang2 = 5249, 'caaclang2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] soaGateway
    #: - [UDP] soaGateway
    soagateway = 5250, 'soagateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA eTrust VM Service
    #: - [UDP] CA eTrust VM Service
    caevms = 5251, 'caevms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Movaz SSC
    #: - [UDP] Movaz SSC
    movaz_ssc = 5252, 'movaz-ssc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Kohler Power Device Protocol
    kpdp = 5253, 'kpdp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5253 = 5253, 'reserved', TransportProtocol.get('udp')

    #: [TCP] LogCabin storage service
    logcabin = 5254, 'logcabin', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5254 = 5254, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] 3Com Network Jack Port 1
    #: - [UDP] 3Com Network Jack Port 1
    AppType_3com_njack_1 = 5264, '3com-njack-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3Com Network Jack Port 2
    #: - [UDP] 3Com Network Jack Port 2
    AppType_3com_njack_2 = 5265, '3com-njack-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] XMPP Server Connection [:rfc:`6120`]
    xmpp_server = 5269, 'xmpp-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5269 = 5269, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Cartographer XMP
    #: - [UDP] Cartographer XMP
    cartographerxmp = 5270, 'cartographerxmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] StageSoft CueLink messaging
    cuelink = 5271, 'cuelink', TransportProtocol.get('tcp')

    #: [UDP] StageSoft CueLink discovery
    cuelink_disc = 5271, 'cuelink-disc', TransportProtocol.get('udp')

    #: - [TCP] PK
    #: - [UDP] PK
    pk = 5272, 'pk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Bidirectional-streams Over Synchronous HTTP (BOSH)
    xmpp_bosh = 5280, 'xmpp-bosh', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5280 = 5280, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Undo License Manager
    undo_lm = 5281, 'undo-lm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5281 = 5281, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Marimba Transmitter Port
    #: - [UDP] Marimba Transmitter Port
    transmit_port = 5282, 'transmit-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NLG Data Service
    #: - [UDP] NLG Data Service
    nlg_data = 5299, 'nlg-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HA cluster heartbeat
    #: - [UDP] HA cluster heartbeat
    hacl_hb = 5300, 'hacl-hb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HA cluster general services
    #: - [UDP] HA cluster general services
    hacl_gs = 5301, 'hacl-gs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HA cluster configuration
    #: - [UDP] HA cluster configuration
    hacl_cfg = 5302, 'hacl-cfg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HA cluster probing
    #: - [UDP] HA cluster probing
    hacl_probe = 5303, 'hacl-probe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HA Cluster Commands
    #: - [UDP] HA Cluster Commands
    hacl_local = 5304, 'hacl-local', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HA Cluster Test
    #: - [UDP] HA Cluster Test
    hacl_test = 5305, 'hacl-test', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun MC Group
    #: - [UDP] Sun MC Group
    sun_mc_grp = 5306, 'sun-mc-grp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCO AIP
    #: - [UDP] SCO AIP
    sco_aip = 5307, 'sco-aip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CFengine
    #: - [UDP] CFengine
    cfengine = 5308, 'cfengine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] J Printer
    #: - [UDP] J Printer
    jprinter = 5309, 'jprinter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Outlaws
    #: - [UDP] Outlaws
    outlaws = 5310, 'outlaws', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_5311 = 5311, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Permabit Client-Server
    #: - [UDP] Permabit Client-Server
    permabit_cs = 5312, 'permabit-cs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Real-time & Reliable Data
    #: - [UDP] Real-time & Reliable Data
    rrdp = 5313, 'rrdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] opalis-rbt-ipc
    #: - [UDP] opalis-rbt-ipc
    opalis_rbt_ipc = 5314, 'opalis-rbt-ipc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HA Cluster UDP Polling
    #: - [UDP] HA Cluster UDP Polling
    hacl_poll = 5315, 'hacl-poll', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] HPBladeSystem Monitor Service
    hpbladems = 5316, 'hpbladems', TransportProtocol.get('tcp')

    #: [UDP] Unassigned
    unassigned_5316 = 5316, 'unassigned', TransportProtocol.get('udp')

    #: [TCP] HP Device Monitor Service
    hpdevms = 5317, 'hpdevms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5317 = 5317, 'reserved', TransportProtocol.get('udp')

    #: [TCP] PKIX Certificate Management using CMS (CMC) [:rfc:`6402`]
    pkix_cmc = 5318, 'pkix-cmc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5318 = 5318, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_5319 = 5319, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Webservices-based Zn interface of BSF
    bsfserver_zn = 5320, 'bsfserver-zn', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5320 = 5320, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Webservices-based Zn interface of BSF over SSL
    bsfsvr_zn_ssl = 5321, 'bsfsvr-zn-ssl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5321 = 5321, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Sculptor Database Server
    #: - [UDP] Sculptor Database Server
    kfserver = 5343, 'kfserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xkoto DRCP
    #: - [UDP] xkoto DRCP
    xkotodrcp = 5344, 'xkotodrcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Session Traversal Utilities for NAT (STUN) port [:rfc:`8489`]
    #: - [UDP] STUN over DTLS [:rfc:`7350`]
    stuns = 5349, 'stuns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TURN over TLS [:rfc:`8656`]
    #: - [UDP] TURN over DTLS [:rfc:`7350`]
    turns = 5349, 'turns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] STUN Behavior Discovery over TLS [:rfc:`5780`]
    #: - [UDP] Reserved for a future enhancement of STUN-BEHAVIOR [:rfc:`5780`]
    stun_behaviors = 5349, 'stun-behaviors', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5350 = 5350, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Port Control Protocol Multicast [:rfc:`6887`]
    pcp_multicast = 5350, 'pcp-multicast', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5351 = 5351, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Port Control Protocol [:rfc:`6887`]
    pcp = 5351, 'pcp', TransportProtocol.get('udp')

    #: - [TCP] DNS Long-Lived Queries [:rfc:`8764`]
    #: - [UDP] DNS Long-Lived Queries [:rfc:`8764`]
    dns_llq = 5352, 'dns-llq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multicast DNS [:rfc:`6762`]
    #: - [UDP] Multicast DNS [:rfc:`6762`]
    mdns = 5353, 'mdns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multicast DNS Responder IPC
    #: - [UDP] Multicast DNS Responder IPC
    mdnsresponder = 5354, 'mdnsresponder', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LLMNR
    #: - [UDP] LLMNR
    llmnr = 5355, 'llmnr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Small Business
    #: - [UDP] Microsoft Small Business
    ms_smlbiz = 5356, 'ms-smlbiz', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Web Services for Devices
    #: - [UDP] Web Services for Devices
    wsdapi = 5357, 'wsdapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WS for Devices Secured
    #: - [UDP] WS for Devices Secured
    wsdapi_s = 5358, 'wsdapi-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Alerter
    #: - [UDP] Microsoft Alerter
    ms_alerter = 5359, 'ms-alerter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Protocol for Windows SideShow
    #: - [UDP] Protocol for Windows SideShow
    ms_sideshow = 5360, 'ms-sideshow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Protocol for Windows SideShow
    #: - [UDP] Secure Protocol for Windows SideShow
    ms_s_sideshow = 5361, 'ms-s-sideshow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Windows Server WSD2 Service
    #: - [UDP] Microsoft Windows Server WSD2 Service
    serverwsd2 = 5362, 'serverwsd2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Windows Network Projection
    #: - [UDP] Windows Network Projection
    net_projection = 5363, 'net-projection', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Microsoft Kernel Debugger
    kdnet = 5364, 'kdnet', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5364 = 5364, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] StressTester(tm) Injector
    #: - [UDP] StressTester(tm) Injector
    stresstester = 5397, 'stresstester', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Elektron Administration
    #: - [UDP] Elektron Administration
    elektron_admin = 5398, 'elektron-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SecurityChase
    #: - [UDP] SecurityChase
    securitychase = 5399, 'securitychase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Excerpt Search
    #: - [UDP] Excerpt Search
    excerpt = 5400, 'excerpt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Excerpt Search Secure
    #: - [UDP] Excerpt Search Secure
    excerpts = 5401, 'excerpts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mftp
    #: - [UDP] mftp
    mftp_349 = 349, 'mftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OmniCast MFTP
    #: - [UDP] OmniCast MFTP
    mftp_5402 = 5402, 'mftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HPOMS-CI-LSTN
    #: - [UDP] HPOMS-CI-LSTN
    hpoms_ci_lstn = 5403, 'hpoms-ci-lstn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HPOMS-DPS-LSTN
    #: - [UDP] HPOMS-DPS-LSTN
    hpoms_dps_lstn = 5404, 'hpoms-dps-lstn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetSupport
    #: - [UDP] NetSupport
    netsupport = 5405, 'netsupport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Systemics Sox
    #: - [UDP] Systemics Sox
    systemics_sox = 5406, 'systemics-sox', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Foresyte-Clear
    #: - [UDP] Foresyte-Clear
    foresyte_clear = 5407, 'foresyte-clear', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Foresyte-Sec
    #: - [UDP] Foresyte-Sec
    foresyte_sec = 5408, 'foresyte-sec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Salient Data Server
    #: - [UDP] Salient Data Server
    salient_dtasrv = 5409, 'salient-dtasrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Salient User Manager
    #: - [UDP] Salient User Manager
    salient_usrmgr = 5410, 'salient-usrmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ActNet
    #: - [UDP] ActNet
    actnet = 5411, 'actnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Continuus
    #: - [UDP] Continuus
    continuus = 5412, 'continuus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WWIOTALK
    #: - [UDP] WWIOTALK
    wwiotalk = 5413, 'wwiotalk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] StatusD
    #: - [UDP] StatusD
    statusd = 5414, 'statusd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NS Server
    #: - [UDP] NS Server
    ns_server = 5415, 'ns-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNS Gateway
    #: - [UDP] SNS Gateway
    sns_gateway = 5416, 'sns-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNS Agent
    #: - [UDP] SNS Agent
    sns_agent = 5417, 'sns-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCNTP
    #: - [UDP] MCNTP
    mcntp = 5418, 'mcntp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DJ-ICE
    #: - [UDP] DJ-ICE
    dj_ice = 5419, 'dj-ice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cylink-C
    #: - [UDP] Cylink-C
    cylink_c = 5420, 'cylink-c', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net Support 2
    #: - [UDP] Net Support 2
    netsupport2 = 5421, 'netsupport2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Salient MUX
    #: - [UDP] Salient MUX
    salient_mux = 5422, 'salient-mux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VIRTUALUSER
    #: - [UDP] VIRTUALUSER
    virtualuser = 5423, 'virtualuser', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Beyond Remote
    #: - [UDP] Beyond Remote
    beyond_remote = 5424, 'beyond-remote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Beyond Remote Command Channel
    #: - [UDP] Beyond Remote Command Channel
    br_channel = 5425, 'br-channel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DEVBASIC
    #: - [UDP] DEVBASIC
    devbasic = 5426, 'devbasic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCO-PEER-TTA
    #: - [UDP] SCO-PEER-TTA
    sco_peer_tta = 5427, 'sco-peer-tta', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TELACONSOLE
    #: - [UDP] TELACONSOLE
    telaconsole = 5428, 'telaconsole', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Billing and Accounting System Exchange
    #: - [UDP] Billing and Accounting System Exchange
    base = 5429, 'base', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RADEC CORP
    #: - [UDP] RADEC CORP
    radec_corp = 5430, 'radec-corp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PARK AGENT
    #: - [UDP] PARK AGENT
    park_agent = 5431, 'park-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PostgreSQL Database
    #: - [UDP] PostgreSQL Database
    postgresql = 5432, 'postgresql', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pyrrho DBMS
    #: - [UDP] Pyrrho DBMS
    pyrrho = 5433, 'pyrrho', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SGI Array Services Daemon
    #: - [UDP] SGI Array Services Daemon
    sgi_arrayd = 5434, 'sgi-arrayd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCEANICS situation and action notification
    #: - [UDP] SCEANICS situation and action notification
    sceanics = 5435, 'sceanics', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5436 = 5436, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] pmip6-cntl [:rfc:`5844`]
    pmip6_cntl = 5436, 'pmip6-cntl', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5437 = 5437, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] pmip6-data [:rfc:`5844`]
    pmip6_data = 5437, 'pmip6-data', TransportProtocol.get('udp')

    #: - [TCP] Pearson HTTPS
    #: - [UDP] Pearson HTTPS
    spss = 5443, 'spss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_5444 = 5444, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Server Message Block over Remote Direct Memory Access
    #: - [SCTP] Server Message Block over Remote Direct Memory Access
    smbdirect = 5445, 'smbdirect', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_5445 = 5445, 'reserved', TransportProtocol.get('udp')

    #: [TCP] TiePie engineering data acquisition
    tiepie = 5450, 'tiepie', TransportProtocol.get('tcp')

    #: [UDP] TiePie engineering data acquisition (discovery)
    tiepie_disc = 5450, 'tiepie-disc', TransportProtocol.get('udp')

    #: - [TCP] SureBox
    #: - [UDP] SureBox
    surebox = 5453, 'surebox', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 5454
    #: - [UDP] APC 5454
    apc_5454 = 5454, 'apc-5454', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 5455
    #: - [UDP] APC 5455
    apc_5455 = 5455, 'apc-5455', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 5456
    #: - [UDP] APC 5456
    apc_5456 = 5456, 'apc-5456', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SILKMETER
    #: - [UDP] SILKMETER
    silkmeter = 5461, 'silkmeter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TTL Publisher
    #: - [UDP] TTL Publisher
    ttl_publisher = 5462, 'ttl-publisher', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TTL Price Proxy
    #: - [UDP] TTL Price Proxy
    ttlpriceproxy = 5463, 'ttlpriceproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quail Networks Object Broker
    #: - [UDP] Quail Networks Object Broker
    quailnet = 5464, 'quailnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NETOPS-BROKER
    #: - [UDP] NETOPS-BROKER
    netops_broker = 5465, 'netops-broker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] The Apsolab company's data collection protocol (native api)
    apsolab_col = 5470, 'apsolab-col', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5470 = 5470, 'reserved', TransportProtocol.get('udp')

    #: [TCP] The Apsolab company's secure data collection protocol (native api)
    apsolab_cols = 5471, 'apsolab-cols', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5471 = 5471, 'reserved', TransportProtocol.get('udp')

    #: [TCP] The Apsolab company's dynamic tag protocol
    apsolab_tag = 5472, 'apsolab-tag', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5472 = 5472, 'reserved', TransportProtocol.get('udp')

    #: [TCP] The Apsolab company's secure dynamic tag protocol
    apsolab_tags = 5473, 'apsolab-tags', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5473 = 5473, 'reserved', TransportProtocol.get('udp')

    #: [UDP] The Apsolab company's status query protocol
    apsolab_rpc = 5474, 'apsolab-rpc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5474 = 5474, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] The Apsolab company's data retrieval protocol
    apsolab_data = 5475, 'apsolab-data', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5475 = 5475, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] fcp-addr-srvr1
    #: - [UDP] fcp-addr-srvr1
    fcp_addr_srvr1 = 5500, 'fcp-addr-srvr1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fcp-addr-srvr2
    #: - [UDP] fcp-addr-srvr2
    fcp_addr_srvr2 = 5501, 'fcp-addr-srvr2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fcp-srvr-inst1
    #: - [UDP] fcp-srvr-inst1
    fcp_srvr_inst1 = 5502, 'fcp-srvr-inst1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fcp-srvr-inst2
    #: - [UDP] fcp-srvr-inst2
    fcp_srvr_inst2 = 5503, 'fcp-srvr-inst2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fcp-cics-gw1
    #: - [UDP] fcp-cics-gw1
    fcp_cics_gw1 = 5504, 'fcp-cics-gw1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Checkout Database
    #: - [UDP] Checkout Database
    checkoutdb = 5505, 'checkoutdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Amcom Mobile Connect
    #: - [UDP] Amcom Mobile Connect
    amc = 5506, 'amc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PowerSysLab Electrical Management
    psl_management = 5507, 'psl-management', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5507 = 5507, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Matter Operational Discovery and Communi
    #: - [UDP] Matter Operational Discovery and Communi
    matter = 5540, 'matter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Model Railway control using the CBUS message protocol
    cbus = 5550, 'cbus', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5550 = 5550, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] SGI Eventmond Port
    #: - [UDP] SGI Eventmond Port
    sgi_eventmond = 5553, 'sgi-eventmond', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SGI ESP HTTP
    #: - [UDP] SGI ESP HTTP
    sgi_esphttp = 5554, 'sgi-esphttp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Personal Agent
    #: - [UDP] Personal Agent
    personal_agent = 5555, 'personal-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Freeciv gameplay
    #: - [UDP] Freeciv gameplay
    freeciv = 5556, 'freeciv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Sandlab FARENET
    farenet = 5557, 'farenet', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5557 = 5557, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Data Protector BURA
    dp_bura = 5565, 'dp-bura', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5565 = 5565, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Westec Connect
    westec_connect = 5566, 'westec-connect', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5566 = 5566, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] DOF Protocol Stack Multicast/Secure Transport
    #: - [UDP] DOF Protocol Stack Multicast/Secure Transport
    dof_dps_mc_sec = 5567, 'dof-dps-mc-sec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Session Data Transport Multicast
    #: - [UDP] Session Data Transport Multicast
    sdt = 5568, 'sdt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PLASA E1.33, Remote Device Management (RDM) controller status
    #: notifications
    rdmnet_ctrl = 5569, 'rdmnet-ctrl', TransportProtocol.get('tcp')

    #: [UDP] PLASA E1.33, Remote Device Management (RDM) messages
    rdmnet_device = 5569, 'rdmnet-device', TransportProtocol.get('udp')

    #: - [TCP] SAS Domain Management Messaging Protocol
    #: - [UDP] SAS Domain Management Messaging Protocol
    sdmmp = 5573, 'sdmmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SAS IO Forwarding
    lsi_bobcat = 5574, 'lsi-bobcat', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5574 = 5574, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Oracle Access Protocol
    ora_oap = 5575, 'ora-oap', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5575 = 5575, 'reserved', TransportProtocol.get('udp')

    #: [TCP] FleetDisplay Tracking Service
    fdtracks = 5579, 'fdtracks', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5579 = 5579, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] T-Mobile SMS Protocol Message 0
    #: - [UDP] T-Mobile SMS Protocol Message 0
    tmosms0 = 5580, 'tmosms0', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] T-Mobile SMS Protocol Message 1
    #: - [UDP] T-Mobile SMS Protocol Message 1
    tmosms1 = 5581, 'tmosms1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] T-Mobile SMS Protocol Message 3
    #: - [UDP] T-Mobile SMS Protocol Message 3
    fac_restore = 5582, 'fac-restore', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] T-Mobile SMS Protocol Message 2
    #: - [UDP] T-Mobile SMS Protocol Message 2
    tmo_icon_sync = 5583, 'tmo-icon-sync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BeInSync-Web
    #: - [UDP] BeInSync-Web
    bis_web = 5584, 'bis-web', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BeInSync-sync
    #: - [UDP] BeInSync-sync
    bis_sync = 5585, 'bis-sync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Planning to send mobile terminated SMS to the specific port so that
    #: the SMS is not visible to the client
    att_mt_sms = 5586, 'att-mt-sms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5586 = 5586, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] inin secure messaging
    #: - [UDP] inin secure messaging
    ininmessaging = 5597, 'ininmessaging', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCT Market Data Feed
    #: - [UDP] MCT Market Data Feed
    mctfeed = 5598, 'mctfeed', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Enterprise Security Remote Install
    #: - [UDP] Enterprise Security Remote Install
    esinstall = 5599, 'esinstall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Enterprise Security Manager
    #: - [UDP] Enterprise Security Manager
    esmmanager = 5600, 'esmmanager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Enterprise Security Agent
    #: - [UDP] Enterprise Security Agent
    esmagent = 5601, 'esmagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A1-MSC
    #: - [UDP] A1-MSC
    a1_msc = 5602, 'a1-msc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A1-BS
    #: - [UDP] A1-BS
    a1_bs = 5603, 'a1-bs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A3-SDUNode
    #: - [UDP] A3-SDUNode
    a3_sdunode = 5604, 'a3-sdunode', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] A4-SDUNode
    #: - [UDP] A4-SDUNode
    a4_sdunode = 5605, 'a4-sdunode', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Fiscal Registering Protocol
    efr = 5618, 'efr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5618 = 5618, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Node Initiated Network Association Forma
    #: - [UDP] Node Initiated Network Association Forma
    ninaf = 5627, 'ninaf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HTrust API
    #: - [UDP] HTrust API
    htrust = 5628, 'htrust', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Symantec Storage Foundation for Database
    #: - [UDP] Symantec Storage Foundation for Database
    symantec_sfdb = 5629, 'symantec-sfdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PreciseCommunication
    #: - [UDP] PreciseCommunication
    precise_comm = 5630, 'precise-comm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pcANYWHEREdata
    #: - [UDP] pcANYWHEREdata
    pcanywheredata = 5631, 'pcanywheredata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pcANYWHEREstat
    #: - [UDP] pcANYWHEREstat
    pcanywherestat = 5632, 'pcanywherestat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BE Operations Request Listener
    #: - [UDP] BE Operations Request Listener
    beorl = 5633, 'beorl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SF Message Service
    #: - [UDP] SF Message Service
    xprtld = 5634, 'xprtld', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SFM Authentication Subsystem
    sfmsso = 5635, 'sfmsso', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5635 = 5635, 'reserved', TransportProtocol.get('udp')

    #: [TCP] SFMdb - SFM DB server
    sfm_db_server = 5636, 'sfm-db-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5636 = 5636, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Symantec CSSC
    cssc = 5637, 'cssc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5637 = 5637, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Symantec Fingerprint Lookup and Container Reference Service
    flcrs = 5638, 'flcrs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5638 = 5638, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Symantec Integrity Checking Service
    ics = 5639, 'ics', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5639 = 5639, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Ventureforth Mobile
    vfmobile = 5646, 'vfmobile', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5646 = 5646, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Nagios Remote Plugin Executor
    nrpe = 5666, 'nrpe', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5666 = 5666, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ZeroMQ file publish-subscribe protocol
    filemq = 5670, 'filemq', TransportProtocol.get('tcp')

    #: [UDP] Local area discovery and messaging over ZeroMQ
    zre_disc = 5670, 'zre-disc', TransportProtocol.get('udp')

    #: - [TCP] amqp protocol over TLS/SSL
    #: - [UDP] amqp protocol over TLS/SSL
    amqps = 5671, 'amqps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AMQP
    #: - [UDP] AMQP
    #: - [SCTP] AMQP
    amqp = 5672, 'amqp', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] JACL Message Server
    #: - [UDP] JACL Message Server
    jms = 5673, 'jms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HyperSCSI Port
    #: - [UDP] HyperSCSI Port
    hyperscsi_port = 5674, 'hyperscsi-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] V5UA application port [:rfc:`3807`]
    #: - [UDP] V5UA application port [:rfc:`3807`]
    #: - [SCTP] V5UA application port [:rfc:`3807`]
    v5ua = 5675, 'v5ua', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] RA Administration
    #: - [UDP] RA Administration
    raadmin = 5676, 'raadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quest Central DB2 Launchr
    #: - [UDP] Quest Central DB2 Launchr
    questdb2_lnchr = 5677, 'questdb2-lnchr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Remote Replication Agent Connection
    #: - [UDP] Remote Replication Agent Connection
    rrac = 5678, 'rrac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Direct Cable Connect Manager
    #: - [UDP] Direct Cable Connect Manager
    dccm = 5679, 'dccm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Auriga Router Service
    #: - [UDP] Auriga Router Service
    auriga_router = 5680, 'auriga-router', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Net-coneX Control Protocol
    #: - [UDP] Net-coneX Control Protocol
    ncxcp = 5681, 'ncxcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5682 = 5682, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] BrightCore control & data transfer exchange
    brightcore = 5682, 'brightcore', TransportProtocol.get('udp')

    #: - [TCP] Constrained Application Protocol (CoAP) [:rfc:`8323`]
    #: - [UDP] Constrained Application Protocol [:rfc:`7252`]
    coap = 5683, 'coap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Constrained Application Protocol (CoAP) [:rfc:`7301`][:rfc:`8323`]
    #: - [UDP] DTLS-secured CoAP [:rfc:`7252`]
    coaps = 5684, 'coaps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] GOG multiplayer game protocol
    gog_multiplayer = 5687, 'gog-multiplayer', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5687 = 5687, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] GGZ Gaming Zone
    #: - [UDP] GGZ Gaming Zone
    ggz = 5688, 'ggz', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QM video network management protocol
    #: - [UDP] QM video network management protocol
    qmvideo = 5689, 'qmvideo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Robert Bosch Data Transfer
    rbsystem = 5693, 'rbsystem', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5693 = 5693, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Key Management Interoperability Protocol
    kmip = 5696, 'kmip', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5696 = 5696, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Dell SupportAssist data center management
    supportassist = 5700, 'supportassist', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5700 = 5700, 'reserved', TransportProtocol.get('udp')

    #: [TCP] StorageOS REST API
    storageos = 5705, 'storageos', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5705 = 5705, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] proshare conf audio
    #: - [UDP] proshare conf audio
    proshareaudio = 5713, 'proshareaudio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] proshare conf video
    #: - [UDP] proshare conf video
    prosharevideo = 5714, 'prosharevideo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] proshare conf data
    #: - [UDP] proshare conf data
    prosharedata = 5715, 'prosharedata', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] proshare conf request
    #: - [UDP] proshare conf request
    prosharerequest = 5716, 'prosharerequest', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] proshare conf notify
    #: - [UDP] proshare conf notify
    prosharenotify = 5717, 'prosharenotify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DPM Communication Server
    #: - [UDP] DPM Communication Server
    dpm = 5718, 'dpm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DPM Agent Coordinator
    #: - [UDP] DPM Agent Coordinator
    dpm_agent = 5719, 'dpm-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MS-Licensing
    #: - [UDP] MS-Licensing
    ms_licensing = 5720, 'ms-licensing', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Desktop Passthru Service
    #: - [UDP] Desktop Passthru Service
    dtpt = 5721, 'dtpt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft DFS Replication Service
    #: - [UDP] Microsoft DFS Replication Service
    msdfsr = 5722, 'msdfsr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Operations Manager - Health Service
    #: - [UDP] Operations Manager - Health Service
    omhs = 5723, 'omhs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Operations Manager - SDK Service
    #: - [UDP] Operations Manager - SDK Service
    omsdk = 5724, 'omsdk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Microsoft Identity Lifecycle Manager
    ms_ilm = 5725, 'ms-ilm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5725 = 5725, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Microsoft Lifecycle Manager Secure Token Service
    ms_ilm_sts = 5726, 'ms-ilm-sts', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5726 = 5726, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ASG Event Notification Framework
    asgenf = 5727, 'asgenf', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5727 = 5727, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Dist. I/O Comm. Service Data and Control
    io_dist_data = 5728, 'io-dist-data', TransportProtocol.get('tcp')

    #: [UDP] Dist. I/O Comm. Service Group Membership
    io_dist_group = 5728, 'io-dist-group', TransportProtocol.get('udp')

    #: - [TCP] Openmail User Agent Layer
    #: - [UDP] Openmail User Agent Layer
    openmail = 5729, 'openmail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Steltor's calendar access
    #: - [UDP] Steltor's calendar access
    unieng = 5730, 'unieng', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IDA Discover Port 1
    #: - [UDP] IDA Discover Port 1
    ida_discover1 = 5741, 'ida-discover1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IDA Discover Port 2
    #: - [UDP] IDA Discover Port 2
    ida_discover2 = 5742, 'ida-discover2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Watchdoc NetPOD Protocol
    #: - [UDP] Watchdoc NetPOD Protocol
    watchdoc_pod = 5743, 'watchdoc-pod', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Watchdoc Server
    #: - [UDP] Watchdoc Server
    watchdoc = 5744, 'watchdoc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fcopy-server
    #: - [UDP] fcopy-server
    fcopy_server = 5745, 'fcopy-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fcopys-server
    #: - [UDP] fcopys-server
    fcopys_server = 5746, 'fcopys-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Wildbits Tunatic
    #: - [UDP] Wildbits Tunatic
    tunatic = 5747, 'tunatic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Wildbits Tunalyzer
    #: - [UDP] Wildbits Tunalyzer
    tunalyzer = 5748, 'tunalyzer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_5749 = 5749, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Bladelogic Agent Service
    #: - [UDP] Bladelogic Agent Service
    rscd = 5750, 'rscd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenMail Desk Gateway server
    #: - [UDP] OpenMail Desk Gateway server
    openmailg = 5755, 'openmailg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_5756 = 5756, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] OpenMail X.500 Directory Server
    #: - [UDP] OpenMail X.500 Directory Server
    x500ms = 5757, 'x500ms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenMail NewMail Server
    #: - [UDP] OpenMail NewMail Server
    openmailns = 5766, 'openmailns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenMail Suer Agent Layer (Secure)
    #: - [UDP] OpenMail Suer Agent Layer (Secure)
    s_openmail = 5767, 's-openmail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenMail CMTS Server
    #: - [UDP] OpenMail CMTS Server
    openmailpxy = 5768, 'openmailpxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] x509solutions Internal CA
    #: - [UDP] x509solutions Internal CA
    spramsca = 5769, 'spramsca', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] x509solutions Secure Data
    #: - [UDP] x509solutions Secure Data
    spramsd = 5770, 'spramsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetAgent
    #: - [UDP] NetAgent
    netagent = 5771, 'netagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Control commands and responses
    #: - [UDP] Control commands and responses
    starfield_io = 5777, 'starfield-io', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Visual Tag System RPC
    vts_rpc = 5780, 'vts-rpc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5780 = 5780, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] 3PAR Event Reporting Service
    #: - [UDP] 3PAR Event Reporting Service
    AppType_3par_evts = 5781, '3par-evts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3PAR Management Service
    #: - [UDP] 3PAR Management Service
    AppType_3par_mgmt = 5782, '3par-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3PAR Management Service with SSL
    #: - [UDP] 3PAR Management Service with SSL
    AppType_3par_mgmt_ssl = 5783, '3par-mgmt-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5784 = 5784, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Cisco Interbox Application Redundancy
    ibar = 5784, 'ibar', TransportProtocol.get('udp')

    #: - [TCP] 3PAR Inform Remote Copy
    #: - [UDP] 3PAR Inform Remote Copy
    AppType_3par_rcopy = 5785, '3par-rcopy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5786 = 5786, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] redundancy notification
    cisco_redu = 5786, 'cisco-redu', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5787 = 5787, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Cisco WAAS Cluster Protocol
    waascluster = 5787, 'waascluster', TransportProtocol.get('udp')

    #: - [TCP] XtreamX Supervised Peer message
    #: - [UDP] XtreamX Supervised Peer message
    xtreamx = 5793, 'xtreamx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_5794 = 5794, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Simple Peered Discovery Protocol
    spdp = 5794, 'spdp', TransportProtocol.get('udp')

    #: [TCP] Proprietary Website deployment service
    enlabel_dpl = 5798, 'enlabel-dpl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5798 = 5798, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ICMPD
    #: - [UDP] ICMPD
    icmpd = 5813, 'icmpd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Support Automation
    #: - [UDP] Support Automation
    spt_automation = 5814, 'spt-automation', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] AutoPass licensing
    autopassdaemon = 5820, 'autopassdaemon', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5820 = 5820, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Z-firm ShipRush interface for web access and bidirectional data
    shiprush_d_ch = 5841, 'shiprush-d-ch', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5841 = 5841, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reversion Backup/Restore
    reversion = 5842, 'reversion', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5842 = 5842, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] WHEREHOO
    #: - [UDP] WHEREHOO
    wherehoo = 5859, 'wherehoo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PlanetPress Suite Messeng
    #: - [UDP] PlanetPress Suite Messeng
    ppsuitemsg = 5863, 'ppsuitemsg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Diameter over TLS/TCP [:rfc:`6733`]
    #: - [SCTP] Diameter over DTLS/SCTP [:rfc:`6733`]
    diameters = 5868, 'diameters', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_5868 = 5868, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Javascript Unit Test Environment
    jute = 5883, 'jute', TransportProtocol.get('tcp')

    #: - [TCP] Remote Framebuffer [:rfc:`6143`]
    #: - [UDP] Remote Framebuffer [:rfc:`6143`]
    rfb = 5900, 'rfb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Flight & Flow Info for Collaborative Env
    #: - [UDP] Flight & Flow Info for Collaborative Env
    #: - [SCTP] Flight & Flow Info for Collaborative Env
    ff_ice = 5903, 'ff-ice', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Air-Ground SWIM
    #: - [UDP] Air-Ground SWIM
    #: - [SCTP] Air-Ground SWIM
    ag_swim = 5904, 'ag-swim', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Adv Surface Mvmnt and Guidance Cont Sys
    #: - [UDP] Adv Surface Mvmnt and Guidance Cont Sys
    #: - [SCTP] Adv Surface Mvmnt and Guidance Cont Sys
    asmgcs = 5905, 'asmgcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Remotely Piloted Vehicle C&C
    #: - [UDP] Remotely Piloted Vehicle C&C
    #: - [SCTP] Remotely Piloted Vehicle C&C
    rpas_c2 = 5906, 'rpas-c2', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Distress and Safety Data App
    #: - [UDP] Distress and Safety Data App
    #: - [SCTP] Distress and Safety Data App
    dsd = 5907, 'dsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] IPS Management Application
    #: - [UDP] IPS Management Application
    #: - [SCTP] IPS Management Application
    ipsma = 5908, 'ipsma', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Air-ground media advisory
    #: - [UDP] Air-ground media advisory
    #: - [SCTP] Air-ground media advisory
    agma = 5909, 'agma', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Air Traffic Services applications using ATN
    #: - [UDP] Air Traffic Services applications using ATN
    ats_atn = 5910, 'ats-atn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] Context Management
    cm = 5910, 'cm', TransportProtocol.get('sctp')

    #: - [TCP] Air Traffic Services applications using ACARS
    #: - [UDP] Air Traffic Services applications using ACARS
    ats_acars = 5911, 'ats-acars', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] Controller Pilot Data Link Communication
    cpdlc = 5911, 'cpdlc', TransportProtocol.get('sctp')

    #: - [TCP] Aeronautical Information Service/Meteorological applications using
    #: ACARS
    #: - [UDP] Aeronautical Information Service/Meteorological applications using
    #: ACARS
    ais_met = 5912, 'ais-met', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] Flight Information Services
    fis = 5912, 'fis', TransportProtocol.get('sctp')

    #: - [TCP] Airline operational communications applications using ACARS
    #: - [UDP] Airline operational communications applications using ACARS
    aoc_acars = 5913, 'aoc-acars', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] Automatic Dependent Surveillance
    ads_c = 5913, 'ads-c', TransportProtocol.get('sctp')

    #: - [TCP] Indy Application Server
    #: - [UDP] Indy Application Server
    indy = 5963, 'indy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mppolicy-v5
    #: - [UDP] mppolicy-v5
    mppolicy_v5 = 5968, 'mppolicy-v5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mppolicy-mgr
    #: - [UDP] mppolicy-mgr
    mppolicy_mgr = 5969, 'mppolicy-mgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CouchDB
    #: - [UDP] CouchDB
    couchdb = 5984, 'couchdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WBEM WS-Management HTTP
    #: - [UDP] WBEM WS-Management HTTP
    wsman = 5985, 'wsman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WBEM WS-Management HTTP over TLS/SSL
    #: - [UDP] WBEM WS-Management HTTP over TLS/SSL
    wsmans = 5986, 'wsmans', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WBEM RMI
    #: - [UDP] WBEM RMI
    wbem_rmi = 5987, 'wbem-rmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WBEM CIM-XML (HTTP)
    #: - [UDP] WBEM CIM-XML (HTTP)
    wbem_http = 5988, 'wbem-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WBEM CIM-XML (HTTPS)
    #: - [UDP] WBEM CIM-XML (HTTPS)
    wbem_https = 5989, 'wbem-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WBEM Export HTTPS
    #: - [UDP] WBEM Export HTTPS
    wbem_exp_https = 5990, 'wbem-exp-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NUXSL
    #: - [UDP] NUXSL
    nuxsl = 5991, 'nuxsl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Consul InSight Security
    #: - [UDP] Consul InSight Security
    consul_insight = 5992, 'consul-insight', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DMTF WBEM CIM REST
    cim_rs = 5993, 'cim-rs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5993 = 5993, 'reserved', TransportProtocol.get('udp')

    #: [TCP] RMS Agent Listening Service
    rms_agent = 5994, 'rms-agent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_5994 = 5994, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] CVSup
    #: - [UDP] CVSup
    cvsup = 5999, 'cvsup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NDL-AHP-SVC
    #: - [UDP] NDL-AHP-SVC
    ndl_ahp_svc = 6064, 'ndl-ahp-svc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WinPharaoh
    #: - [UDP] WinPharaoh
    winpharaoh = 6065, 'winpharaoh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EWCTSP
    #: - [UDP] EWCTSP
    ewctsp = 6066, 'ewctsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6067 = 6067, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] GSMP/ANCP [:rfc:`6320`]
    gsmp_ancp = 6068, 'gsmp-ancp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6068 = 6068, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] TRIP
    #: - [UDP] TRIP
    trip = 6069, 'trip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Messageasap
    #: - [UDP] Messageasap
    messageasap = 6070, 'messageasap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSDTP
    #: - [UDP] SSDTP
    ssdtp = 6071, 'ssdtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DIAGNOSE-PROC
    #: - [UDP] DIAGNOSE-PROC
    diagnose_proc = 6072, 'diagnose-proc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DirectPlay8
    #: - [UDP] DirectPlay8
    directplay8 = 6073, 'directplay8', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Max
    #: - [UDP] Microsoft Max
    max = 6074, 'max', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Microsoft DPM Access Control Manager
    dpm_acm = 6075, 'dpm-acm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6075 = 6075, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Microsoft DPM WCF Certificates
    msft_dpm_cert = 6076, 'msft-dpm-cert', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6076 = 6076, 'reserved', TransportProtocol.get('udp')

    #: [TCP] iConstruct Server
    iconstructsrv = 6077, 'iconstructsrv', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6077 = 6077, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Generic UDP Encapsulation [draft-herbert-gue]
    gue = 6080, 'gue', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6080 = 6080, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Generic Network Virtualization Encapsulation (Geneve) [:rfc:`8926`]
    geneve = 6081, 'geneve', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6081 = 6081, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] Reserved
    reserved_6082 = 6082, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] APCO Project 25 Common Air Interface - UDP encapsulation
    p25cai = 6082, 'p25cai', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6083 = 6083, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] telecomsoftware miami broadcast
    miami_bcast = 6083, 'miami-bcast', TransportProtocol.get('udp')

    #: [TCP] Peer to Peer Infrastructure Configuration [:rfc:`6940`]
    reload_config = 6084, 'reload-config', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6084 = 6084, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] konspire2b p2p network
    #: - [UDP] konspire2b p2p network
    konspire2b = 6085, 'konspire2b', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PDTP P2P
    #: - [UDP] PDTP P2P
    pdtp = 6086, 'pdtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Local Download Sharing Service
    #: - [UDP] Local Download Sharing Service
    ldss = 6087, 'ldss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SuperDog License Manager
    doglms = 6088, 'doglms', TransportProtocol.get('tcp')

    #: [UDP] SuperDog License Manager Notifier
    doglms_notify = 6088, 'doglms-notify', TransportProtocol.get('udp')

    #: [TCP] RAXA Management
    raxa_mgmt = 6099, 'raxa-mgmt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6099 = 6099, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] SynchroNet-db
    #: - [UDP] SynchroNet-db
    synchronet_db = 6100, 'synchronet-db', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SynchroNet-rtc
    #: - [UDP] SynchroNet-rtc
    synchronet_rtc = 6101, 'synchronet-rtc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SynchroNet-upd
    #: - [UDP] SynchroNet-upd
    synchronet_upd = 6102, 'synchronet-upd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RETS
    #: - [UDP] RETS
    rets = 6103, 'rets', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DBDB
    #: - [UDP] DBDB
    dbdb = 6104, 'dbdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Prima Server
    #: - [UDP] Prima Server
    primaserver = 6105, 'primaserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MPS Server
    #: - [UDP] MPS Server
    mpsserver = 6106, 'mpsserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ETC Control
    #: - [UDP] ETC Control
    etc_control = 6107, 'etc-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sercomm-SCAdmin
    #: - [UDP] Sercomm-SCAdmin
    sercomm_scadmin = 6108, 'sercomm-scadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GLOBECAST-ID
    #: - [UDP] GLOBECAST-ID
    globecast_id = 6109, 'globecast-id', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP SoftBench CM
    #: - [UDP] HP SoftBench CM
    softcm = 6110, 'softcm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP SoftBench Sub-Process Control
    #: - [UDP] HP SoftBench Sub-Process Control
    spc = 6111, 'spc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Desk-Top Sub-Process Control Daemon
    #: - [UDP] Desk-Top Sub-Process Control Daemon
    dtspcd = 6112, 'dtspcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Daylite Server
    dayliteserver = 6113, 'dayliteserver', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6113 = 6113, 'reserved', TransportProtocol.get('udp')

    #: [TCP] WRspice IPC Service
    wrspice = 6114, 'wrspice', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6114 = 6114, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Xic IPC Service
    xic = 6115, 'xic', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6115 = 6115, 'reserved', TransportProtocol.get('udp')

    #: [TCP] XicTools License Manager Service
    xtlserv = 6116, 'xtlserv', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6116 = 6116, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Daylite Touch Sync
    daylitetouch = 6117, 'daylitetouch', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6117 = 6117, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Transparent Inter Process Communication
    tipc = 6118, 'tipc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6118 = 6118, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] SPDY for a faster web
    spdy = 6121, 'spdy', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6121 = 6121, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Backup Express Web Server
    #: - [UDP] Backup Express Web Server
    bex_webadmin = 6122, 'bex-webadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Backup Express
    #: - [UDP] Backup Express
    backup_express = 6123, 'backup-express', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Phlexible Network Backup Service
    #: - [UDP] Phlexible Network Backup Service
    pnbs = 6124, 'pnbs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] The DameWare Mobile Gateway Service
    damewaremobgtwy = 6130, 'damewaremobgtwy', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6130 = 6130, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] New Boundary Tech WOL
    #: - [UDP] New Boundary Tech WOL
    nbt_wol = 6133, 'nbt-wol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pulsonix Network License Service
    #: - [UDP] Pulsonix Network License Service
    pulsonixnls = 6140, 'pulsonixnls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Meta Corporation License Manager
    #: - [UDP] Meta Corporation License Manager
    meta_corp = 6141, 'meta-corp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Aspen Technology License Manager
    #: - [UDP] Aspen Technology License Manager
    aspentec_lm = 6142, 'aspentec-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Watershed License Manager
    #: - [UDP] Watershed License Manager
    watershed_lm = 6143, 'watershed-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] StatSci License Manager - 1
    #: - [UDP] StatSci License Manager - 1
    statsci1_lm = 6144, 'statsci1-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] StatSci License Manager - 2
    #: - [UDP] StatSci License Manager - 2
    statsci2_lm = 6145, 'statsci2-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Lone Wolf Systems License Manager
    #: - [UDP] Lone Wolf Systems License Manager
    lonewolf_lm = 6146, 'lonewolf-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Montage License Manager
    #: - [UDP] Montage License Manager
    montage_lm = 6147, 'montage-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ricardo North America License Manager
    #: - [UDP] Ricardo North America License Manager
    ricardo_lm_1522 = 1522, 'ricardo-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ricardo North America License Manager
    #: - [UDP] Ricardo North America License Manager
    ricardo_lm_6148 = 6148, 'ricardo-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tal-pod
    #: - [UDP] tal-pod
    tal_pod = 6149, 'tal-pod', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] EFB Application Control Interface
    efb_aci = 6159, 'efb-aci', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6159 = 6159, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Emerson Extensible Control and Management Protocol
    ecmp = 6160, 'ecmp', TransportProtocol.get('tcp')

    #: [UDP] Emerson Extensible Control and Management Protocol Data
    ecmp_data = 6160, 'ecmp-data', TransportProtocol.get('udp')

    #: - [TCP] PATROL Internet Srv Mgr
    #: - [UDP] PATROL Internet Srv Mgr
    patrol_ism = 6161, 'patrol-ism', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PATROL Collector
    #: - [UDP] PATROL Collector
    patrol_coll = 6162, 'patrol-coll', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Precision Scribe Cnx Port
    #: - [UDP] Precision Scribe Cnx Port
    pscribe = 6163, 'pscribe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LM-X License Manager by X-Formation
    #: - [UDP] LM-X License Manager by X-Formation
    lm_x = 6200, 'lm-x', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6201 = 6201, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Management of service nodes in a processing grid for thermodynamic
    #: calculations
    thermo_calc = 6201, 'thermo-calc', TransportProtocol.get('udp')

    #: - [TCP] QMTP over TLS
    #: - [UDP] QMTP over TLS
    qmtps = 6209, 'qmtps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Radmind Access Protocol
    #: - [UDP] Radmind Access Protocol
    radmind = 6222, 'radmind', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] JEOL Network Services Data Transport Protocol 1
    jeol_nsdtp_1 = 6241, 'jeol-nsdtp-1', TransportProtocol.get('tcp')

    #: [UDP] JEOL Network Services Dynamic Discovery Protocol 1
    jeol_nsddp_1 = 6241, 'jeol-nsddp-1', TransportProtocol.get('udp')

    #: [TCP] JEOL Network Services Data Transport Protocol 2
    jeol_nsdtp_2 = 6242, 'jeol-nsdtp-2', TransportProtocol.get('tcp')

    #: [UDP] JEOL Network Services Dynamic Discovery Protocol 2
    jeol_nsddp_2 = 6242, 'jeol-nsddp-2', TransportProtocol.get('udp')

    #: [TCP] JEOL Network Services Data Transport Protocol 3
    jeol_nsdtp_3 = 6243, 'jeol-nsdtp-3', TransportProtocol.get('tcp')

    #: [UDP] JEOL Network Services Dynamic Discovery Protocol 3
    jeol_nsddp_3 = 6243, 'jeol-nsddp-3', TransportProtocol.get('udp')

    #: [TCP] JEOL Network Services Data Transport Protocol 4
    jeol_nsdtp_4 = 6244, 'jeol-nsdtp-4', TransportProtocol.get('tcp')

    #: [UDP] JEOL Network Services Dynamic Discovery Protocol 4
    jeol_nsddp_4 = 6244, 'jeol-nsddp-4', TransportProtocol.get('udp')

    #: - [TCP] TL1 Raw Over SSL/TLS
    #: - [UDP] TL1 Raw Over SSL/TLS
    tl1_raw_ssl = 6251, 'tl1-raw-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TL1 over SSH
    #: - [UDP] TL1 over SSH
    tl1_ssh = 6252, 'tl1-ssh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CRIP
    #: - [UDP] CRIP
    crip = 6253, 'crip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] GridLAB-D User Interface
    gld = 6267, 'gld', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6267 = 6267, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Grid Authentication
    #: - [UDP] Grid Authentication
    grid = 6268, 'grid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Grid Authentication Alt
    #: - [UDP] Grid Authentication Alt
    grid_alt = 6269, 'grid-alt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC GRX
    #: - [UDP] BMC GRX
    bmc_grx = 6300, 'bmc-grx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC CONTROL-D LDAP SERVER IANA assigned this well-formed service name
    #: as a replacement for "bmc_ctd_ldap".
    #: - [TCP] BMC CONTROL-D LDAP SERVER
    #: - [UDP] BMC CONTROL-D LDAP SERVER IANA assigned this well-formed service name
    #: as a replacement for "bmc_ctd_ldap".
    #: - [UDP] BMC CONTROL-D LDAP SERVER
    bmc_ctd_ldap = 6301, 'bmc-ctd-ldap', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Unified Fabric Management Protocol
    #: - [UDP] Unified Fabric Management Protocol
    ufmp = 6306, 'ufmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Sensor Control Unit Protocol
    scup = 6315, 'scup', TransportProtocol.get('tcp')

    #: [UDP] Sensor Control Unit Protocol Discovery Protocol
    scup_disc = 6315, 'scup-disc', TransportProtocol.get('udp')

    #: - [TCP] Ethernet Sensor Communications Protocol
    #: - [UDP] Ethernet Sensor Communications Protocol
    abb_escp = 6316, 'abb-escp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Navtech Radar Sensor Data Command
    nav_data_cmd = 6317, 'nav-data-cmd', TransportProtocol.get('tcp')

    #: [UDP] Navtech Radar Sensor Data
    nav_data = 6317, 'nav-data', TransportProtocol.get('udp')

    #: - [TCP] Double-Take Replication Service
    #: - [UDP] Double-Take Replication Service
    repsvc = 6320, 'repsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Empress Software Connectivity Server 1
    #: - [UDP] Empress Software Connectivity Server 1
    emp_server1 = 6321, 'emp-server1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Empress Software Connectivity Server 2
    #: - [UDP] Empress Software Connectivity Server 2
    emp_server2 = 6322, 'emp-server2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6323 = 6323, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] HR Device Network Configuration Service
    hrd_ncs = 6324, 'hrd-ncs', TransportProtocol.get('tcp')

    #: [UDP] HR Device Network service
    hrd_ns_disc = 6324, 'hrd-ns-disc', TransportProtocol.get('udp')

    #: [TCP] Double-Take Management Service
    dt_mgmtsvc = 6325, 'dt-mgmtsvc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6325 = 6325, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Double-Take Virtual Recovery Assistant
    dt_vra = 6326, 'dt-vra', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6326 = 6326, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] sFlow traffic monitoring
    #: - [UDP] sFlow traffic monitoring
    sflow = 6343, 'sflow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Argus-Spectr security and fire-prevention systems service
    streletz = 6344, 'streletz', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6344 = 6344, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] gnutella-svc
    #: - [UDP] gnutella-svc
    gnutella_svc = 6346, 'gnutella-svc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] gnutella-rtr
    #: - [UDP] gnutella-rtr
    gnutella_rtr = 6347, 'gnutella-rtr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] App Discovery and Access Protocol
    #: - [UDP] App Discovery and Access Protocol
    adap = 6350, 'adap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PMCS applications
    #: - [UDP] PMCS applications
    pmcs = 6355, 'pmcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MetaEdit+ Multi-User
    #: - [UDP] MetaEdit+ Multi-User
    metaedit_mu = 6360, 'metaedit-mu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Named Data Networking
    ndn = 6363, 'ndn', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6363 = 6363, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] MetaEdit+ Server Administration
    #: - [UDP] MetaEdit+ Server Administration
    metaedit_se = 6370, 'metaedit-se', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] An advanced key-value cache and store
    redis = 6379, 'redis', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6379 = 6379, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Metatude Dialogue Server
    #: - [UDP] Metatude Dialogue Server
    metatude_mds = 6382, 'metatude-mds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] clariion-evr01
    #: - [UDP] clariion-evr01
    clariion_evr01 = 6389, 'clariion-evr01', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MetaEdit+ WebService API
    #: - [UDP] MetaEdit+ WebService API
    metaedit_ws = 6390, 'metaedit-ws', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Business Objects CMS contact port
    boe_cms = 6400, 'boe-cms', TransportProtocol.get('undefined')

    #: [N/A] boe-was
    boe_was = 6401, 'boe-was', TransportProtocol.get('undefined')

    #: [N/A] boe-eventsrv
    boe_eventsrv = 6402, 'boe-eventsrv', TransportProtocol.get('undefined')

    #: [N/A] boe-cachesvr
    boe_cachesvr = 6403, 'boe-cachesvr', TransportProtocol.get('undefined')

    #: [N/A] Business Objects Enterprise internal server
    boe_filesvr = 6404, 'boe-filesvr', TransportProtocol.get('undefined')

    #: [N/A] Business Objects Enterprise internal server
    boe_pagesvr = 6405, 'boe-pagesvr', TransportProtocol.get('undefined')

    #: [N/A] Business Objects Enterprise internal server
    boe_processsvr = 6406, 'boe-processsvr', TransportProtocol.get('undefined')

    #: [N/A] Business Objects Enterprise internal server
    boe_resssvr1 = 6407, 'boe-resssvr1', TransportProtocol.get('undefined')

    #: [N/A] Business Objects Enterprise internal server
    boe_resssvr2 = 6408, 'boe-resssvr2', TransportProtocol.get('undefined')

    #: [N/A] Business Objects Enterprise internal server
    boe_resssvr3 = 6409, 'boe-resssvr3', TransportProtocol.get('undefined')

    #: [N/A] Business Objects Enterprise internal server
    boe_resssvr4 = 6410, 'boe-resssvr4', TransportProtocol.get('undefined')

    #: - [TCP] Faxcom Message Service
    #: - [UDP] Faxcom Message Service
    faxcomservice = 6417, 'faxcomservice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SYserver remote commands
    syserverremote = 6418, 'syserverremote', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6418 = 6418, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Simple VDR Protocol
    svdrp = 6419, 'svdrp', TransportProtocol.get('tcp')

    #: [UDP] Simple VDR Protocol Discovery
    svdrp_disc = 6419, 'svdrp-disc', TransportProtocol.get('udp')

    #: - [TCP] NIM_VDRShell
    #: - [UDP] NIM_VDRShell
    nim_vdrshell = 6420, 'nim-vdrshell', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NIM_WAN
    #: - [UDP] NIM_WAN
    nim_wan = 6421, 'nim-wan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PgBouncer
    pgbouncer = 6432, 'pgbouncer', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6432 = 6432, 'reserved', TransportProtocol.get('udp')

    #: [TCP] heliosd daemon
    heliosd = 6440, 'heliosd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6440 = 6440, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6441 = 6441, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Transitory Application Request Protocol
    tarp = 6442, 'tarp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6442 = 6442, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Service Registry Default HTTPS Domain
    #: - [UDP] Service Registry Default HTTPS Domain
    sun_sr_https = 6443, 'sun-sr-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Grid Engine Execution Service IANA assigned this well-formed service
    #: name as a replacement for "sge_execd".
    #: - [TCP] Grid Engine Execution Service
    #: - [UDP] Grid Engine Execution Service IANA assigned this well-formed service
    #: name as a replacement for "sge_execd".
    #: - [UDP] Grid Engine Execution Service
    sge_execd = 6445, 'sge-execd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] MySQL Proxy
    #: - [UDP] MySQL Proxy
    mysql_proxy = 6446, 'mysql-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SKIP Certificate Receive
    #: - [UDP] SKIP Certificate Receive
    skip_cert_recv = 6455, 'skip-cert-recv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SKIP Certificate Send
    #: - [UDP] SKIP Certificate Send
    skip_cert_send = 6456, 'skip-cert-send', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Port assignment for medical device communication in accordance to IEEE
    #: 11073-20701
    #: - [UDP] Port assignment for medical device communication in accordance to IEEE
    #: 11073-20701
    ieee11073_20701 = 6464, 'ieee11073-20701', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LVision License Manager
    #: - [UDP] LVision License Manager
    lvision_lm = 6471, 'lvision-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Registry Default HTTP Domain
    #: - [UDP] Service Registry Default HTTP Domain
    sun_sr_http = 6480, 'sun-sr-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Tags
    #: - [UDP] Service Tags
    servicetags = 6481, 'servicetags', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Logical Domains Management Interface
    #: - [UDP] Logical Domains Management Interface
    ldoms_mgmt = 6482, 'ldoms-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SunVTS RMI
    #: - [UDP] SunVTS RMI
    sunvts_rmi = 6483, 'sunvts-rmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Registry Default JMS Domain
    #: - [UDP] Service Registry Default JMS Domain
    sun_sr_jms = 6484, 'sun-sr-jms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Registry Default IIOP Domain
    #: - [UDP] Service Registry Default IIOP Domain
    sun_sr_iiop = 6485, 'sun-sr-iiop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Registry Default IIOPS Domain
    #: - [UDP] Service Registry Default IIOPS Domain
    sun_sr_iiops = 6486, 'sun-sr-iiops', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Registry Default IIOPAuth Domain
    #: - [UDP] Service Registry Default IIOPAuth Domain
    sun_sr_iiop_aut = 6487, 'sun-sr-iiop-aut', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Registry Default JMX Domain
    #: - [UDP] Service Registry Default JMX Domain
    sun_sr_jmx = 6488, 'sun-sr-jmx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Service Registry Default Admin Domain
    #: - [UDP] Service Registry Default Admin Domain
    sun_sr_admin = 6489, 'sun-sr-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BoKS Master
    #: - [UDP] BoKS Master
    boks = 6500, 'boks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BoKS Servc IANA assigned this well-formed service name as a
    #: replacement for "boks_servc".
    #: - [TCP] BoKS Servc
    #: - [UDP] BoKS Servc IANA assigned this well-formed service name as a
    #: replacement for "boks_servc".
    #: - [UDP] BoKS Servc
    boks_servc = 6501, 'boks-servc', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] BoKS Servm IANA assigned this well-formed service name as a
    #: replacement for "boks_servm".
    #: - [TCP] BoKS Servm
    #: - [UDP] BoKS Servm IANA assigned this well-formed service name as a
    #: replacement for "boks_servm".
    #: - [UDP] BoKS Servm
    boks_servm = 6502, 'boks-servm', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] BoKS Clntd IANA assigned this well-formed service name as a
    #: replacement for "boks_clntd".
    #: - [TCP] BoKS Clntd
    #: - [UDP] BoKS Clntd IANA assigned this well-formed service name as a
    #: replacement for "boks_clntd".
    #: - [UDP] BoKS Clntd
    boks_clntd = 6503, 'boks-clntd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6504 = 6504, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] BoKS Admin Private Port IANA assigned this well-formed service name as
    #: a replacement for "badm_priv".
    #: - [TCP] BoKS Admin Private Port
    #: - [UDP] BoKS Admin Private Port IANA assigned this well-formed service name as
    #: a replacement for "badm_priv".
    #: - [UDP] BoKS Admin Private Port
    badm_priv = 6505, 'badm-priv', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] BoKS Admin Public Port IANA assigned this well-formed service name as
    #: a replacement for "badm_pub".
    #: - [TCP] BoKS Admin Public Port
    #: - [UDP] BoKS Admin Public Port IANA assigned this well-formed service name as
    #: a replacement for "badm_pub".
    #: - [UDP] BoKS Admin Public Port
    badm_pub = 6506, 'badm-pub', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] BoKS Dir Server, Private Port IANA assigned this well-formed service
    #: name as a replacement for "bdir_priv".
    #: - [TCP] BoKS Dir Server, Private Port
    #: - [UDP] BoKS Dir Server, Private Port IANA assigned this well-formed service
    #: name as a replacement for "bdir_priv".
    #: - [UDP] BoKS Dir Server, Private Port
    bdir_priv = 6507, 'bdir-priv', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] BoKS Dir Server, Public Port IANA assigned this well-formed service
    #: name as a replacement for "bdir_pub".
    #: - [TCP] BoKS Dir Server, Public Port
    #: - [UDP] BoKS Dir Server, Public Port IANA assigned this well-formed service
    #: name as a replacement for "bdir_pub".
    #: - [UDP] BoKS Dir Server, Public Port
    bdir_pub = 6508, 'bdir-pub', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] MGCS-MFP Port
    #: - [UDP] MGCS-MFP Port
    mgcs_mfp_port = 6509, 'mgcs-mfp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MCER Port
    #: - [UDP] MCER Port
    mcer_port = 6510, 'mcer-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6511 = 6511, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Datagram Congestion Control Protocol Encapsulation for NAT Traversal
    #: [:rfc:`6773`]
    dccp_udp = 6511, 'dccp-udp', TransportProtocol.get('udp')

    #: [TCP] NETCONF over TLS [:rfc:`7589`]
    netconf_tls = 6513, 'netconf-tls', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6513 = 6513, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Syslog over TLS [:rfc:`5425`]
    #: - [UDP] syslog over DTLS [:rfc:`6012`]
    #: - [DCCP] syslog over DTLS [:rfc:`6012`]
    syslog_tls = 6514, 'syslog-tls', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('dccp')

    #: - [TCP] Elipse RPC Protocol
    #: - [UDP] Elipse RPC Protocol
    elipse_rec = 6515, 'elipse-rec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] lds_distrib
    #: - [UDP] lds_distrib
    lds_distrib = 6543, 'lds-distrib', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LDS Dump Service
    #: - [UDP] LDS Dump Service
    lds_dump = 6544, 'lds-dump', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 6547
    #: - [UDP] APC 6547
    apc_6547 = 6547, 'apc-6547', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 6548
    #: - [UDP] APC 6548
    apc_6548 = 6548, 'apc-6548', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 6549
    #: - [UDP] APC 6549
    apc_6549 = 6549, 'apc-6549', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fg-sysupdate
    #: - [UDP] fg-sysupdate
    fg_sysupdate = 6550, 'fg-sysupdate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Software Update Manager
    #: - [UDP] Software Update Manager
    sum = 6551, 'sum', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Checkmk Monitoring Agent
    checkmk_agent = 6556, 'checkmk-agent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6556 = 6556, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6557 = 6557, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP]
    #: - [UDP]
    xdsxdm = 6558, 'xdsxdm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SANE Control Port
    #: - [UDP] SANE Control Port
    sane_port = 6566, 'sane-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Reserved
    reserved_6567 = 6567, 'reserved', TransportProtocol.get('undefined')

    #: - [TCP] CanIt Storage Manager IANA assigned this well-formed service name as a
    #: replacement for "canit_store".
    #: - [TCP] CanIt Storage Manager
    canit_store = 6568, 'canit-store', TransportProtocol.get('tcp') | TransportProtocol.get('tcp')

    #: [UDP] Roaring Penguin IP Address Reputation Collection
    rp_reputation = 6568, 'rp-reputation', TransportProtocol.get('udp')

    #: - [TCP] Affiliate
    #: - [UDP] Affiliate
    affiliate = 6579, 'affiliate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Parsec Masterserver
    #: - [UDP] Parsec Masterserver
    parsec_master = 6580, 'parsec-master', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Parsec Peer-to-Peer
    #: - [UDP] Parsec Peer-to-Peer
    parsec_peer = 6581, 'parsec-peer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Parsec Gameserver
    #: - [UDP] Parsec Gameserver
    parsec_game = 6582, 'parsec-game', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JOA Jewel Suite
    #: - [UDP] JOA Jewel Suite
    joajewelsuite = 6583, 'joajewelsuite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6588 = 6588, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Microsoft Hyper-V Live Migration
    mshvlm = 6600, 'mshvlm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6600 = 6600, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Microsoft Threat Management Gateway SSTP
    mstmg_sstp = 6601, 'mstmg-sstp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6601 = 6601, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Windows WSS Communication Framework
    wsscomfrmwk = 6602, 'wsscomfrmwk', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6602 = 6602, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ODETTE-FTP over TLS/SSL [:rfc:`5024`]
    #: - [UDP] ODETTE-FTP over TLS/SSL [:rfc:`5024`]
    odette_ftps = 6619, 'odette-ftps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kerberos V5 FTP Data
    #: - [UDP] Kerberos V5 FTP Data
    kftp_data = 6620, 'kftp-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kerberos V5 FTP Control
    #: - [UDP] Kerberos V5 FTP Control
    kftp = 6621, 'kftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multicast FTP
    #: - [UDP] Multicast FTP
    mcftp = 6622, 'mcftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kerberos V5 Telnet
    #: - [UDP] Kerberos V5 Telnet
    ktelnet = 6623, 'ktelnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DataScaler database
    datascaler_db = 6624, 'datascaler-db', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6624 = 6624, 'reserved', TransportProtocol.get('udp')

    #: [TCP] DataScaler control
    datascaler_ctl = 6625, 'datascaler-ctl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6625 = 6625, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] WAGO Service and Update
    #: - [UDP] WAGO Service and Update
    wago_service = 6626, 'wago-service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Allied Electronics NeXGen
    #: - [UDP] Allied Electronics NeXGen
    nexgen = 6627, 'nexgen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AFE Stock Channel M/C
    #: - [UDP] AFE Stock Channel M/C
    afesc_mc = 6628, 'afesc-mc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secondary, (non ANDI) multi-protocol multi-function interface to
    #: the Allied ANDI-based family of forecourt controllers
    #: - [UDP] Secondary, (non ANDI) multi-protocol multi-function interface to
    #: the Allied ANDI-based family of forecourt controllers
    nexgen_aux = 6629, 'nexgen-aux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6630 = 6630, 'unassigned', TransportProtocol.get('undefined')

    #: [N/A] Unassigned
    unassigned_6631 = 6631, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] eGenix mxODBC Connect
    mxodbc_connect = 6632, 'mxodbc-connect', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6632 = 6632, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6633 = 6633, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Cisco vPath Services Overlay
    cisco_vpath_tun = 6633, 'cisco-vpath-tun', TransportProtocol.get('udp')

    #: [UDP] MPLS Performance Measurement out-of-band response
    mpls_pm = 6634, 'mpls-pm', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6634 = 6634, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] Reserved
    reserved_6635 = 6635, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Encapsulate MPLS packets in UDP tunnels. [:rfc:`7510`]
    mpls_udp = 6635, 'mpls-udp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6636 = 6636, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Encapsulate MPLS packets in UDP tunnels with DTLS. [:rfc:`7510`]
    mpls_udp_dtls = 6636, 'mpls-udp-dtls', TransportProtocol.get('udp')

    #: [TCP] Open vSwitch Database protocol [:rfc:`7047`]
    ovsdb = 6640, 'ovsdb', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6640 = 6640, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] OpenFlow
    #: - [UDP] OpenFlow
    openflow = 6653, 'openflow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6654 = 6654, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] PC SOFT - Software factory UI/manager
    pcs_sf_ui_man = 6655, 'pcs-sf-ui-man', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6655 = 6655, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Emergency Message Control Service
    emgmsg = 6656, 'emgmsg', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6656 = 6656, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6657 = 6657, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] PalCom Discovery
    palcom_disc = 6657, 'palcom-disc', TransportProtocol.get('udp')

    #: - [TCP] Vocaltec Global Online Directory
    #: - [UDP] Vocaltec Global Online Directory
    vocaltec_gold = 6670, 'vocaltec-gold', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] P4P Portal Service
    #: - [UDP] P4P Portal Service
    p4p_portal = 6671, 'p4p-portal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] vision_server IANA assigned this well-formed service name as a
    #: replacement for "vision_server".
    #: - [TCP] vision_server
    #: - [UDP] vision_server IANA assigned this well-formed service name as a
    #: replacement for "vision_server".
    #: - [UDP] vision_server
    vision_server = 6672, 'vision-server', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] vision_elmd IANA assigned this well-formed service name as a
    #: replacement for "vision_elmd".
    #: - [TCP] vision_elmd
    #: - [UDP] vision_elmd IANA assigned this well-formed service name as a
    #: replacement for "vision_elmd".
    #: - [UDP] vision_elmd
    vision_elmd = 6673, 'vision-elmd', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [TCP] Viscount Freedom Bridge Protocol
    vfbp = 6678, 'vfbp', TransportProtocol.get('tcp')

    #: [UDP] Viscount Freedom Bridge Discovery
    vfbp_disc = 6678, 'vfbp-disc', TransportProtocol.get('udp')

    #: - [TCP] Osorno Automation
    #: - [UDP] Osorno Automation
    osaut = 6679, 'osaut', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CleverView for cTrace Message Service
    clever_ctrace = 6687, 'clever-ctrace', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6687 = 6687, 'reserved', TransportProtocol.get('udp')

    #: [TCP] CleverView for TCP/IP Message Service
    clever_tcpip = 6688, 'clever-tcpip', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6688 = 6688, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Tofino Security Appliance
    #: - [UDP] Tofino Security Appliance
    tsa = 6689, 'tsa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CLEVERDetect Message Service
    cleverdetect = 6690, 'cleverdetect', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6690 = 6690, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6696 = 6696, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Babel Routing Protocol [:rfc:`8966`]
    babel = 6696, 'babel', TransportProtocol.get('udp')

    #: [TCP] Internet Relay Chat via TLS/SSL [:rfc:`7194`]
    ircs_u = 6697, 'ircs-u', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6697 = 6697, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6698 = 6698, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Reserved
    reserved_6699 = 6699, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Babel Routing Protocol over DTLS [:rfc:`8968`]
    babel_dtls = 6699, 'babel-dtls', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_6700 = 6700, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] KTI/ICAD Nameserver
    #: - [UDP] KTI/ICAD Nameserver
    kti_icad_srvr = 6701, 'kti-icad-srvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] Unassigned
    unassigned_6701 = 6701, 'unassigned', TransportProtocol.get('sctp')

    #: - [TCP] e-Design network
    #: - [UDP] e-Design network
    e_design_net = 6702, 'e-design-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] Unassigned
    unassigned_6702 = 6702, 'unassigned', TransportProtocol.get('sctp')

    #: - [TCP] e-Design web
    #: - [UDP] e-Design web
    e_design_web = 6703, 'e-design-web', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [UDP] Reserved
    #: - [TCP] Reserved
    reserved_6704 = 6704, 'reserved', TransportProtocol.get('udp') | TransportProtocol.get('tcp')

    #: [SCTP] ForCES HP (High Priority) channel [:rfc:`5811`]
    frc_hp = 6704, 'frc-hp', TransportProtocol.get('sctp')

    #: - [UDP] Reserved
    #: - [TCP] Reserved
    reserved_6705 = 6705, 'reserved', TransportProtocol.get('udp') | TransportProtocol.get('tcp')

    #: [SCTP] ForCES MP (Medium Priority) channel [:rfc:`5811`]
    frc_mp = 6705, 'frc-mp', TransportProtocol.get('sctp')

    #: - [UDP] Reserved
    #: - [TCP] Reserved
    reserved_6706 = 6706, 'reserved', TransportProtocol.get('udp') | TransportProtocol.get('tcp')

    #: [SCTP] ForCES LP (Low priority) channel [:rfc:`5811`]
    frc_lp = 6706, 'frc-lp', TransportProtocol.get('sctp')

    #: - [TCP] Internet Backplane Protocol
    #: - [UDP] Internet Backplane Protocol
    ibprotocol = 6714, 'ibprotocol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fibotrader Communications
    #: - [UDP] Fibotrader Communications
    fibotrader_com = 6715, 'fibotrader-com', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Princity Agent
    princity_agent = 6716, 'princity-agent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6716 = 6716, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] BMC PERFORM AGENT
    #: - [UDP] BMC PERFORM AGENT
    bmc_perf_agent = 6767, 'bmc-perf-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BMC PERFORM MGRD
    #: - [UDP] BMC PERFORM MGRD
    bmc_perf_mgrd = 6768, 'bmc-perf-mgrd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ADInstruments GxP Server
    #: - [UDP] ADInstruments GxP Server
    adi_gxp_srvprt = 6769, 'adi-gxp-srvprt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PolyServe http
    #: - [UDP] PolyServe http
    plysrv_http = 6770, 'plysrv-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PolyServe https
    #: - [UDP] PolyServe https
    plysrv_https = 6771, 'plysrv-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] netTsunami Tracker
    ntz_tracker = 6777, 'ntz-tracker', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6777 = 6777, 'reserved', TransportProtocol.get('udp')

    #: [TCP] netTsunami p2p storage system
    ntz_p2p_storage = 6778, 'ntz-p2p-storage', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6778 = 6778, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_6784 = 6784, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Bidirectional Forwarding Detection (BFD) on Link Aggregation Group
    #: (LAG) Interfaces [:rfc:`7130`]
    bfd_lag = 6784, 'bfd-lag', TransportProtocol.get('udp')

    #: - [TCP] DGPF Individual Exchange
    #: - [UDP] DGPF Individual Exchange
    dgpf_exchg = 6785, 'dgpf-exchg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun Java Web Console JMX
    #: - [UDP] Sun Java Web Console JMX
    smc_jmx = 6786, 'smc-jmx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun Web Console Admin
    #: - [UDP] Sun Web Console Admin
    smc_admin = 6787, 'smc-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SMC-HTTP
    #: - [UDP] SMC-HTTP
    smc_http = 6788, 'smc-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] GSS-API for the Oracle Remote Administration Daemon
    radg = 6789, 'radg', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6789 = 6789, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] HNMP
    #: - [UDP] HNMP
    hnmp = 6790, 'hnmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Halcyon Network Manager
    #: - [UDP] Halcyon Network Manager
    hnm = 6791, 'hnm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ACNET Control System Protocol
    #: - [UDP] ACNET Control System Protocol
    acnet = 6801, 'acnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PenTBox Secure IM Protocol
    pentbox_sim = 6817, 'pentbox-sim', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6817 = 6817, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ambit-lm
    #: - [UDP] ambit-lm
    ambit_lm = 6831, 'ambit-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netmo Default
    #: - [UDP] Netmo Default
    netmo_default = 6841, 'netmo-default', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Netmo HTTP
    #: - [UDP] Netmo HTTP
    netmo_http = 6842, 'netmo-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ICCRUSHMORE
    #: - [UDP] ICCRUSHMORE
    iccrushmore = 6850, 'iccrushmore', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Acctopus Command Channel
    acctopus_cc = 6868, 'acctopus-cc', TransportProtocol.get('tcp')

    #: [UDP] Acctopus Status
    acctopus_st = 6868, 'acctopus-st', TransportProtocol.get('udp')

    #: - [TCP] MUSE
    #: - [UDP] MUSE
    muse = 6888, 'muse', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] R*TIME Viewer Data Interface
    rtimeviewer = 6900, 'rtimeviewer', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6900 = 6900, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Novell Jetstream messaging protocol
    jetstream = 6901, 'jetstream', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6901 = 6901, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Ping with RX/TX latency/loss split
    #: - [UDP] Ping with RX/TX latency/loss split
    split_ping = 6924, 'split-ping', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EthoScan Service
    #: - [UDP] EthoScan Service
    ethoscan = 6935, 'ethoscan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XenSource Management Service
    #: - [UDP] XenSource Management Service
    xsmsvc = 6936, 'xsmsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Biometrics Server
    #: - [UDP] Biometrics Server
    bioserver = 6946, 'bioserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OTLP
    #: - [UDP] OTLP
    otlp = 6951, 'otlp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JMACT3
    #: - [UDP] JMACT3
    jmact3 = 6961, 'jmact3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] jmevt2
    #: - [UDP] jmevt2
    jmevt2 = 6962, 'jmevt2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] swismgr1
    #: - [UDP] swismgr1
    swismgr1 = 6963, 'swismgr1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] swismgr2
    #: - [UDP] swismgr2
    swismgr2 = 6964, 'swismgr2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] swistrap
    #: - [UDP] swistrap
    swistrap = 6965, 'swistrap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] swispol
    #: - [UDP] swispol
    swispol = 6966, 'swispol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] acmsoda
    #: - [UDP] acmsoda
    acmsoda = 6969, 'acmsoda', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Conductor test coordination protocol
    conductor = 6970, 'conductor', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_6970 = 6970, 'reserved', TransportProtocol.get('udp')

    #: [SCTP] conductor for multiplex
    conductor_mpx = 6970, 'conductor-mpx', TransportProtocol.get('sctp')

    #: [TCP] Reserved
    reserved_6980 = 6980, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] QoS-extended OLSR protocol
    qolyester = 6980, 'qolyester', TransportProtocol.get('udp')

    #: - [TCP] Mobility XE Protocol
    #: - [UDP] Mobility XE Protocol
    mobilitysrv = 6997, 'mobilitysrv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IATP-highPri
    #: - [UDP] IATP-highPri
    iatp_highpri = 6998, 'iatp-highpri', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IATP-normalPri
    #: - [UDP] IATP-normalPri
    iatp_normalpri = 6999, 'iatp-normalpri', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] file server itself
    #: - [UDP] file server itself
    afs3_fileserver = 7000, 'afs3-fileserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] callbacks to cache managers
    #: - [UDP] callbacks to cache managers
    afs3_callback = 7001, 'afs3-callback', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] users & groups database
    #: - [UDP] users & groups database
    afs3_prserver = 7002, 'afs3-prserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] volume location database
    #: - [UDP] volume location database
    afs3_vlserver = 7003, 'afs3-vlserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AFS/Kerberos authentication service
    #: - [UDP] AFS/Kerberos authentication service
    afs3_kaserver = 7004, 'afs3-kaserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] volume managment server
    #: - [UDP] volume managment server
    afs3_volser = 7005, 'afs3-volser', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] error interpretation service
    #: - [UDP] error interpretation service
    afs3_errors = 7006, 'afs3-errors', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] basic overseer process
    #: - [UDP] basic overseer process
    afs3_bos = 7007, 'afs3-bos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] server-to-server updater
    #: - [UDP] server-to-server updater
    afs3_update = 7008, 'afs3-update', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] remote cache manager service
    #: - [UDP] remote cache manager service
    afs3_rmtsys = 7009, 'afs3-rmtsys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] onlinet uninterruptable power supplies
    #: - [UDP] onlinet uninterruptable power supplies
    ups_onlinet = 7010, 'ups-onlinet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talon Discovery Port
    #: - [UDP] Talon Discovery Port
    talon_disc = 7011, 'talon-disc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talon Engine
    #: - [UDP] Talon Engine
    talon_engine = 7012, 'talon-engine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microtalon Discovery
    #: - [UDP] Microtalon Discovery
    microtalon_dis = 7013, 'microtalon-dis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microtalon Communications
    #: - [UDP] Microtalon Communications
    microtalon_com = 7014, 'microtalon-com', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Talon Webserver
    #: - [UDP] Talon Webserver
    talon_webserver = 7015, 'talon-webserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SPG Controls Carrier
    #: - [UDP] SPG Controls Carrier
    spg = 7016, 'spg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] GeneRic Autonomic Signaling Protocol [:rfc:`8990`]
    #: - [UDP] GeneRic Autonomic Signaling Protocol [:rfc:`8990`]
    grasp = 7017, 'grasp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] FISA Service
    fisa_svc = 7018, 'fisa-svc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7018 = 7018, 'reserved', TransportProtocol.get('udp')

    #: [TCP] doceri drawing service control
    doceri_ctl = 7019, 'doceri-ctl', TransportProtocol.get('tcp')

    #: [UDP] doceri drawing service screen view
    doceri_view = 7019, 'doceri-view', TransportProtocol.get('udp')

    #: - [TCP] DP Serve
    #: - [UDP] DP Serve
    dpserve = 7020, 'dpserve', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DP Serve Admin
    #: - [UDP] DP Serve Admin
    dpserveadmin = 7021, 'dpserveadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CT Discovery Protocol
    #: - [UDP] CT Discovery Protocol
    ctdp = 7022, 'ctdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Comtech T2 NMCS
    #: - [UDP] Comtech T2 NMCS
    ct2nmcs = 7023, 'ct2nmcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vormetric service
    #: - [UDP] Vormetric service
    vmsvc = 7024, 'vmsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vormetric Service II
    #: - [UDP] Vormetric Service II
    vmsvc_2 = 7025, 'vmsvc-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Loreji Webhosting Panel
    loreji_panel = 7026, 'loreji-panel', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7026 = 7026, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ObjectPlanet probe
    #: - [UDP] ObjectPlanet probe
    op_probe = 7030, 'op-probe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] IPOSPLANET retailing multi devices protocol
    iposplanet = 7031, 'iposplanet', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7031 = 7031, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7040 = 7040, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Quest application level network service discovery
    quest_disc = 7040, 'quest-disc', TransportProtocol.get('udp')

    #: - [TCP] ARCP
    #: - [UDP] ARCP
    arcp = 7070, 'arcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IWGADTS Aircraft Housekeeping Message
    #: - [UDP] IWGADTS Aircraft Housekeeping Message
    iwg1 = 7071, 'iwg1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] iba Device Configuration Protocol
    iba_cfg = 7072, 'iba-cfg', TransportProtocol.get('tcp')

    #: [UDP] iba Device Configuration Protocol
    iba_cfg_disc = 7072, 'iba-cfg-disc', TransportProtocol.get('udp')

    #: [TCP] MarTalk protocol
    martalk = 7073, 'martalk', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7073 = 7073, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] EmpowerID Communication
    #: - [UDP] EmpowerID Communication
    empowerid = 7080, 'empowerid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7088 = 7088, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Zixi live video transport protocol
    zixi_transport = 7088, 'zixi-transport', TransportProtocol.get('udp')

    #: [UDP] Java Discovery Protocol
    jdp_disc = 7095, 'jdp-disc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7095 = 7095, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] lazy-ptop
    #: - [UDP] lazy-ptop
    lazy_ptop = 7099, 'lazy-ptop', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] X Font Service
    #: - [UDP] X Font Service
    font_service = 7100, 'font-service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Embedded Light Control Network
    #: - [UDP] Embedded Light Control Network
    elcn = 7101, 'elcn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7107 = 7107, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] AES-X170
    aes_x170 = 7107, 'aes-x170', TransportProtocol.get('udp')

    #: [TCP] Encrypted chat and file transfer service
    rothaga = 7117, 'rothaga', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7117 = 7117, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Virtual Prototypes License Manager
    #: - [UDP] Virtual Prototypes License Manager
    virprot_lm = 7121, 'virprot-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7122 = 7122, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] End-to-end TLS Relay Control Connection
    snif = 7123, 'snif', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7123 = 7123, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] intelligent data manager
    #: - [UDP] intelligent data manager
    scenidm = 7128, 'scenidm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Catalog Content Search
    #: - [UDP] Catalog Content Search
    scenccs = 7129, 'scenccs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA BSM Comm
    #: - [UDP] CA BSM Comm
    cabsm_comm = 7161, 'cabsm-comm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA Storage Manager
    #: - [UDP] CA Storage Manager
    caistoragemgr = 7162, 'caistoragemgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA Connection Broker
    #: - [UDP] CA Connection Broker
    cacsambroker = 7163, 'cacsambroker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] File System Repository Agent
    #: - [UDP] File System Repository Agent
    fsr = 7164, 'fsr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Document WCF Server
    #: - [UDP] Document WCF Server
    doc_server = 7165, 'doc-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Aruba eDiscovery Server
    #: - [UDP] Aruba eDiscovery Server
    aruba_server = 7166, 'aruba-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CA SRM Agent
    casrmagent = 7167, 'casrmagent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7167 = 7167, 'reserved', TransportProtocol.get('udp')

    #: [TCP] cncKadServer DB & Inventory Services
    cnckadserver = 7168, 'cnckadserver', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7168 = 7168, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Consequor Consulting Process Integration Bridge
    #: - [UDP] Consequor Consulting Process Integration Bridge
    ccag_pib = 7169, 'ccag-pib', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Adaptive Name/Service Resolution
    #: - [UDP] Adaptive Name/Service Resolution
    nsrp = 7170, 'nsrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Discovery and Retention Mgt Production
    #: - [UDP] Discovery and Retention Mgt Production
    drm_production = 7171, 'drm-production', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Port used for MetalBend programmable interface
    metalbend = 7172, 'metalbend', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7172 = 7172, 'reserved', TransportProtocol.get('udp')

    #: [TCP] zSecure Server
    zsecure = 7173, 'zsecure', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7173 = 7173, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Clutild
    #: - [UDP] Clutild
    clutild = 7174, 'clutild', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Janus Guidewire Enterprise Discovery Service Bus
    janus_disc = 7181, 'janus-disc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7181 = 7181, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] FODMS FLIP
    #: - [UDP] FODMS FLIP
    fodms = 7200, 'fodms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DLIP
    #: - [UDP] DLIP
    dlip = 7201, 'dlip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Inter-Channel Termination Protocol (ICTP) for multi-wavelength PON
    #: (Passive Optical Network) systems [TR-352][Inter-Channel-Termination
    #: protocol (ICTP) is a peer-to-peer protocol with       a binary message
    #: format that is used by a set of optical line terminal (OLT) channel
    #: terminations (CT) comprising a 40-Gigabit-capable next-generation passive
    #: optical network (NG-PON2) system to       exchange channel profile
    #: configuration and status information and to coordinate the operation of
    #: tunable optical network units (ONUs) in a multi-wavelength channel
    #: environment.]
    pon_ictp = 7202, 'pon-ictp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7202 = 7202, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Communication ports for PaperStream Server services
    ps_server = 7215, 'ps-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7215 = 7215, 'reserved', TransportProtocol.get('udp')

    #: [TCP] PaperStream Capture Professional
    ps_capture_pro = 7216, 'ps-capture-pro', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7216 = 7216, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Registry A & M Protocol
    #: - [UDP] Registry A & M Protocol
    ramp = 7227, 'ramp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Citrix Universal Printing Port
    citrixupp = 7228, 'citrixupp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7228 = 7228, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Citrix UPP Gateway
    citrixuppg = 7229, 'citrixuppg', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7229 = 7229, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Traffic forwarding for Okta cloud infra
    asa_gateways = 7234, 'asa-gateways', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7234 = 7234, 'reserved', TransportProtocol.get('udp')

    #: [UDP] ASP Coordination Protocol
    aspcoordination = 7235, 'aspcoordination', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7235 = 7235, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] Wi-Fi Alliance Wi-Fi Display Protocol
    display = 7236, 'display', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7236 = 7236, 'reserved', TransportProtocol.get('udp')

    #: [TCP] PADS (Public Area Display System) Server
    pads = 7237, 'pads', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7237 = 7237, 'reserved', TransportProtocol.get('udp')

    #: [TCP] FrontRow Calypso Human Interface Control Protocol
    frc_hicp = 7244, 'frc-hicp', TransportProtocol.get('tcp')

    #: [UDP] FrontRow Calypso Human Interface Control Protocol
    frc_hicp_disc = 7244, 'frc-hicp-disc', TransportProtocol.get('udp')

    #: - [TCP] Calypso Network Access Protocol
    #: - [UDP] Calypso Network Access Protocol
    cnap = 7262, 'cnap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WatchMe Monitoring 7272
    #: - [UDP] WatchMe Monitoring 7272
    watchme_7272 = 7272, 'watchme-7272', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA Roaming Location
    #: - [UDP] OMA Roaming Location
    oma_rlp = 7273, 'oma-rlp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA Roaming Location SEC
    #: - [UDP] OMA Roaming Location SEC
    oma_rlp_s = 7274, 'oma-rlp-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA UserPlane Location
    #: - [UDP] OMA UserPlane Location
    oma_ulp = 7275, 'oma-ulp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA Internal Location Protocol
    #: - [UDP] OMA Internal Location Protocol
    oma_ilp = 7276, 'oma-ilp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA Internal Location Secure Protocol
    #: - [UDP] OMA Internal Location Secure Protocol
    oma_ilp_s = 7277, 'oma-ilp-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA Dynamic Content Delivery over CBS
    #: - [UDP] OMA Dynamic Content Delivery over CBS
    oma_dcdocbs = 7278, 'oma-dcdocbs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Citrix Licensing
    #: - [UDP] Citrix Licensing
    ctxlic = 7279, 'ctxlic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITACTIONSERVER 1
    #: - [UDP] ITACTIONSERVER 1
    itactionserver1 = 7280, 'itactionserver1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ITACTIONSERVER 2
    #: - [UDP] ITACTIONSERVER 2
    itactionserver2 = 7281, 'itactionserver2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] eventACTION/ussACTION (MZCA) server
    mzca_action = 7282, 'mzca-action', TransportProtocol.get('tcp')

    #: [UDP] eventACTION/ussACTION (MZCA) alert
    mzca_alert = 7282, 'mzca-alert', TransportProtocol.get('udp')

    #: [TCP] General Statistics Rendezvous Protocol
    genstat = 7283, 'genstat', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7283 = 7283, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] LifeKeeper Communications
    #: - [UDP] LifeKeeper Communications
    lcm_server = 7365, 'lcm-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mind-file system server
    #: - [UDP] mind-file system server
    mindfilesys = 7391, 'mindfilesys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mrss-rendezvous server
    #: - [UDP] mrss-rendezvous server
    mrssrendezvous = 7392, 'mrssrendezvous', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] nFoldMan Remote Publish
    #: - [UDP] nFoldMan Remote Publish
    nfoldman = 7393, 'nfoldman', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] File system export of backup images
    #: - [UDP] File system export of backup images
    fse = 7394, 'fse', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] winqedit
    #: - [UDP] winqedit
    winqedit = 7395, 'winqedit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7396 = 7396, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Hexarc Command Language
    #: - [UDP] Hexarc Command Language
    hexarc = 7397, 'hexarc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RTPS Discovery
    #: - [UDP] RTPS Discovery
    rtps_discovery = 7400, 'rtps-discovery', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RTPS Data-Distribution User-Traffic
    #: - [UDP] RTPS Data-Distribution User-Traffic
    rtps_dd_ut = 7401, 'rtps-dd-ut', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RTPS Data-Distribution Meta-Traffic
    #: - [UDP] RTPS Data-Distribution Meta-Traffic
    rtps_dd_mt = 7402, 'rtps-dd-mt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ionix Network Monitor
    #: - [UDP] Ionix Network Monitor
    ionixnetmon = 7410, 'ionixnetmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Streaming of measurement data
    #: - [UDP] Streaming of measurement data
    daqstream = 7411, 'daqstream', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7420 = 7420, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Multichannel real-time lighting control
    ipluminary = 7420, 'ipluminary', TransportProtocol.get('udp')

    #: - [TCP] Matisse Port Monitor
    #: - [UDP] Matisse Port Monitor
    mtportmon = 7421, 'mtportmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView DM Postmaster Manager
    #: - [UDP] OpenView DM Postmaster Manager
    pmdmgr = 7426, 'pmdmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView DM Event Agent Manager
    #: - [UDP] OpenView DM Event Agent Manager
    oveadmgr = 7427, 'oveadmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView DM Log Agent Manager
    #: - [UDP] OpenView DM Log Agent Manager
    ovladmgr = 7428, 'ovladmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView DM rqt communication
    #: - [UDP] OpenView DM rqt communication
    opi_sock = 7429, 'opi-sock', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView DM xmpv7 api pipe
    #: - [UDP] OpenView DM xmpv7 api pipe
    xmpv7 = 7430, 'xmpv7', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenView DM ovc/xmpv3 api pipe
    #: - [UDP] OpenView DM ovc/xmpv3 api pipe
    pmd = 7431, 'pmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Faximum
    #: - [UDP] Faximum
    faximum = 7437, 'faximum', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Oracle Application Server HTTPS
    #: - [UDP] Oracle Application Server HTTPS
    oracleas_https = 7443, 'oracleas-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Stateless Transport Tunneling Protocol
    sttunnel = 7471, 'sttunnel', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7471 = 7471, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7472 = 7472, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Rise: The Vieneo Province
    #: - [UDP] Rise: The Vieneo Province
    rise = 7473, 'rise', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Neo4j Graph Database
    neo4j = 7474, 'neo4j', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7474 = 7474, 'reserved', TransportProtocol.get('udp')

    #: [TCP] IT Asset Management
    openit = 7478, 'openit', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7478 = 7478, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] telops-lmd
    #: - [UDP] telops-lmd
    telops_lmd = 7491, 'telops-lmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Silhouette User
    #: - [UDP] Silhouette User
    silhouette = 7500, 'silhouette', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP OpenView Bus Daemon
    #: - [UDP] HP OpenView Bus Daemon
    ovbus = 7501, 'ovbus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Automation Device Configuration Protocol
    adcp = 7508, 'adcp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7508 = 7508, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ACPLT - process automation service
    acplt = 7509, 'acplt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7509 = 7509, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] HP OpenView Application Server
    #: - [UDP] HP OpenView Application Server
    ovhpas = 7510, 'ovhpas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] pafec-lm
    #: - [UDP] pafec-lm
    pafec_lm = 7511, 'pafec-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Saratoga Transfer Protocol
    #: - [UDP] Saratoga Transfer Protocol
    saratoga = 7542, 'saratoga', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] atul server
    #: - [UDP] atul server
    atul = 7543, 'atul', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FlowAnalyzer DisplayServer
    #: - [UDP] FlowAnalyzer DisplayServer
    nta_ds = 7544, 'nta-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FlowAnalyzer UtilityServer
    #: - [UDP] FlowAnalyzer UtilityServer
    nta_us = 7545, 'nta-us', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cisco Fabric service
    #: - [UDP] Cisco Fabric service
    cfs = 7546, 'cfs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Broadband Forum CWMP
    #: - [UDP] Broadband Forum CWMP
    cwmp = 7547, 'cwmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Threat Information Distribution Protocol
    #: - [UDP] Threat Information Distribution Protocol
    tidp = 7548, 'tidp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Layer Signaling Transport Layer
    #: - [UDP] Network Layer Signaling Transport Layer
    nls_tl = 7549, 'nls-tl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7550 = 7550, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Cloud Signaling Service
    cloudsignaling = 7550, 'cloudsignaling', TransportProtocol.get('udp')

    #: [TCP] ControlONE Console signaling
    controlone_con = 7551, 'controlone-con', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7551 = 7551, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Sniffer Command Protocol
    #: - [UDP] Sniffer Command Protocol
    sncp = 7560, 'sncp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Control Framework [:rfc:`6230`]
    cfw = 7563, 'cfw', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7563 = 7563, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] VSI Omega
    #: - [UDP] VSI Omega
    vsi_omega = 7566, 'vsi-omega', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Dell EqualLogic Host Group Management
    dell_eql_asm = 7569, 'dell-eql-asm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7569 = 7569, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Aries Kfinder
    #: - [UDP] Aries Kfinder
    aries_kfinder = 7570, 'aries-kfinder', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Oracle Coherence Cluster Service
    coherence = 7574, 'coherence', TransportProtocol.get('tcp')

    #: [UDP] Oracle Coherence Cluster discovery service
    coherence_disc = 7574, 'coherence-disc', TransportProtocol.get('udp')

    #: - [TCP] Sun License Manager
    #: - [UDP] Sun License Manager
    sun_lm = 7588, 'sun-lm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MIPI Alliance Debug
    #: - [UDP] MIPI Alliance Debug
    mipi_debug = 7606, 'mipi-debug', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Instrument Neutral Distributed Interface
    #: - [UDP] Instrument Neutral Distributed Interface
    indi = 7624, 'indi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7625 = 7625, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] SImple Middlebox COnfiguration (SIMCO) Server [:rfc:`4540`]
    #: - [SCTP] SImple Middlebox COnfiguration (SIMCO)
    simco = 7626, 'simco', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [N/A] De-registered
    de_registered_3403 = 3403, 'de_registered', TransportProtocol.get('undefined')

    #: [UDP] De-registered
    de_registered_7626 = 7626, 'de_registered', TransportProtocol.get('udp')

    #: - [TCP] SOAP Service Port
    #: - [UDP] SOAP Service Port
    soap_http = 7627, 'soap-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Primary Agent Work Notification
    #: - [UDP] Primary Agent Work Notification
    zen_pawn = 7628, 'zen-pawn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenXDAS Wire Protocol
    #: - [UDP] OpenXDAS Wire Protocol
    xdas = 7629, 'xdas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] HA Web Konsole
    hawk = 7630, 'hawk', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7630 = 7630, 'reserved', TransportProtocol.get('udp')

    #: [TCP] TESLA System Messaging
    tesla_sys_msg = 7631, 'tesla-sys-msg', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7631 = 7631, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7632 = 7632, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] PMDF Management
    #: - [UDP] PMDF Management
    pmdfmgt = 7633, 'pmdfmgt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] bonjour-cuseeme
    #: - [UDP] bonjour-cuseeme
    cuseeme = 7648, 'cuseeme', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Proprietary immutable distributed data storage
    #: - [UDP] Proprietary immutable distributed data storage
    rome = 7663, 'rome', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] iMQ STOMP Server
    imqstomp = 7672, 'imqstomp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7672 = 7672, 'reserved', TransportProtocol.get('udp')

    #: [TCP] iMQ STOMP Server over SSL
    imqstomps = 7673, 'imqstomps', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7673 = 7673, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] iMQ SSL tunnel
    #: - [UDP] iMQ SSL tunnel
    imqtunnels = 7674, 'imqtunnels', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iMQ Tunnel
    #: - [UDP] iMQ Tunnel
    imqtunnel = 7675, 'imqtunnel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iMQ Broker Rendezvous
    #: - [UDP] iMQ Broker Rendezvous
    imqbrokerd = 7676, 'imqbrokerd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun App Server - HTTPS
    #: - [UDP] Sun App Server - HTTPS
    sun_user_https = 7677, 'sun-user-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Microsoft Delivery Optimization Peer-to-Peer
    #: - [UDP] Microsoft Delivery Optimization Peer-to-Peer
    ms_do = 7680, 'ms-do', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Cleondris DMT
    dmt = 7683, 'dmt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7683 = 7683, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Bolt database connection
    bolt = 7687, 'bolt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7687 = 7687, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7688 = 7688, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Collaber Network Service
    #: - [UDP] Collaber Network Service
    collaber = 7689, 'collaber', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Service-Oriented Vehicle Diagnostics
    sovd = 7690, 'sovd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7690 = 7690, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] KLIO communications
    #: - [UDP] KLIO communications
    klio = 7697, 'klio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] EM7 Secure Communications
    em7_secom = 7700, 'em7-secom', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7700 = 7700, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_7701 = 7701, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] SCF nFAPI defining MAC/PHY split
    nfapi = 7701, 'nfapi', TransportProtocol.get('sctp')

    #: - [TCP] EM7 Dynamic Updates
    #: - [UDP] EM7 Dynamic Updates
    sync_em7 = 7707, 'sync-em7', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] scientia.net
    #: - [UDP] scientia.net
    scinet = 7708, 'scinet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MedImage Portal
    #: - [UDP] MedImage Portal
    medimageportal = 7720, 'medimageportal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novell Snap-in Deep Freeze Control
    #: - [UDP] Novell Snap-in Deep Freeze Control
    nsdeepfreezectl = 7724, 'nsdeepfreezectl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nitrogen Service
    #: - [UDP] Nitrogen Service
    nitrogen = 7725, 'nitrogen', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FreezeX Console Service
    #: - [UDP] FreezeX Console Service
    freezexservice = 7726, 'freezexservice', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trident Systems Data
    #: - [UDP] Trident Systems Data
    trident_data = 7727, 'trident-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Open-Source Virtual Reality
    #: - [UDP] Open-Source Virtual Reality
    #: - [SCTP] Open-Source Virtual Reality
    osvr = 7728, 'osvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] Smith Protocol over IP
    #: - [UDP] Smith Protocol over IP
    smip = 7734, 'smip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP Enterprise Discovery Agent
    #: - [UDP] HP Enterprise Discovery Agent
    aiagent = 7738, 'aiagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ScriptView Network
    #: - [UDP] ScriptView Network
    scriptview = 7741, 'scriptview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Mugginsoft Script Server Service
    msss = 7742, 'msss', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7742 = 7742, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Sakura Script Transfer Protocol
    #: - [UDP] Sakura Script Transfer Protocol
    sstp_1 = 7743, 'sstp-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RAQMON PDU [:rfc:`4712`]
    #: - [UDP] RAQMON PDU [:rfc:`4712`]
    raqmon_pdu = 7744, 'raqmon-pdu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Put/Run/Get Protocol
    #: - [UDP] Put/Run/Get Protocol
    prgp = 7747, 'prgp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] A File System using TLS over a wide area network
    inetfs = 7775, 'inetfs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7775 = 7775, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7776 = 7776, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] cbt
    #: - [UDP] cbt
    cbt = 7777, 'cbt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Interwise
    #: - [UDP] Interwise
    interwise = 7778, 'interwise', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VSTAT
    #: - [UDP] VSTAT
    vstat = 7779, 'vstat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7780 = 7780, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] accu-lmgr
    #: - [UDP] accu-lmgr
    accu_lmgr = 7781, 'accu-lmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Seamless Bidirectional Forwarding Detection (S-BFD) [:rfc:`7881`]
    s_bfd = 7784, 's-bfd', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7784 = 7784, 'reserved', TransportProtocol.get('tcp')

    #: [N/A] Unassigned
    unassigned_7785 = 7785, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] MINIVEND
    #: - [UDP] MINIVEND
    minivend = 7786, 'minivend', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Popup Reminders Receive
    #: - [UDP] Popup Reminders Receive
    popup_reminders = 7787, 'popup-reminders', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7788 = 7788, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Office Tools Pro Receive
    #: - [UDP] Office Tools Pro Receive
    office_tools = 7789, 'office-tools', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Q3ADE Cluster Service
    #: - [UDP] Q3ADE Cluster Service
    q3ade = 7794, 'q3ade', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Propel Connector port
    #: - [UDP] Propel Connector port
    pnet_conn = 7797, 'pnet-conn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Propel Encoder port
    #: - [UDP] Propel Encoder port
    pnet_enc = 7798, 'pnet-enc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Alternate BSDP Service
    #: - [UDP] Alternate BSDP Service
    altbsdp = 7799, 'altbsdp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apple Software Restore
    #: - [UDP] Apple Software Restore
    asr = 7800, 'asr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Server Protocol - client
    #: - [UDP] Secure Server Protocol - client
    ssp_client = 7801, 'ssp-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Virtualized Network Services Tunnel Protocol
    vns_tp = 7802, 'vns-tp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7802 = 7802, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Riverbed WAN Optimization Protocol
    #: - [UDP] Riverbed WAN Optimization Protocol
    rbt_wanopt = 7810, 'rbt-wanopt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 7845
    #: - [UDP] APC 7845
    apc_7845 = 7845, 'apc-7845', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 7846
    #: - [UDP] APC 7846
    apc_7846 = 7846, 'apc-7846', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] A product key authentication protocol made by CSO
    csoauth = 7847, 'csoauth', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7847 = 7847, 'reserved', TransportProtocol.get('udp')

    #: [TCP] MobileAnalyzer& MobileMonitor
    mobileanalyzer = 7869, 'mobileanalyzer', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7869 = 7869, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Riverbed Steelhead Mobile Service
    rbt_smc = 7870, 'rbt-smc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7870 = 7870, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Mobile Device Management
    mdm = 7871, 'mdm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7871 = 7871, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7872 = 7872, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] TLS-based Mobile IPv6 Security [:rfc:`6618`]
    mipv6tls = 7872, 'mipv6tls', TransportProtocol.get('udp')

    #: [TCP] Opswise Message Service
    owms = 7878, 'owms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7878 = 7878, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_7879 = 7879, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Pearson
    #: - [UDP] Pearson
    pss = 7880, 'pss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Universal Broker
    #: - [UDP] Universal Broker
    ubroker = 7887, 'ubroker', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multicast Event
    #: - [UDP] Multicast Event
    mevent = 7900, 'mevent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TNOS Service Protocol
    #: - [UDP] TNOS Service Protocol
    tnos_sp = 7901, 'tnos-sp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TNOS shell Protocol
    #: - [UDP] TNOS shell Protocol
    tnos_dp = 7902, 'tnos-dp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TNOS Secure DiaguardProtocol
    #: - [UDP] TNOS Secure DiaguardProtocol
    tnos_dps = 7903, 'tnos-dps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QuickObjects secure port
    #: - [UDP] QuickObjects secure port
    qo_secure = 7913, 'qo-secure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tier 2 Data Resource Manager
    #: - [UDP] Tier 2 Data Resource Manager
    t2_drm = 7932, 't2-drm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Tier 2 Business Rules Manager
    #: - [UDP] Tier 2 Business Rules Manager
    t2_brm = 7933, 't2-brm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Encrypted, extendable, general-purpose synchronization protocol
    #: - [UDP] Encrypted, extendable, general-purpose synchronization protocol
    generalsync = 7962, 'generalsync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Supercell
    #: - [UDP] Supercell
    supercell = 7967, 'supercell', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Micromuse-ncps
    #: - [UDP] Micromuse-ncps
    micromuse_ncps = 7979, 'micromuse-ncps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Quest Vista
    #: - [UDP] Quest Vista
    quest_vista = 7980, 'quest-vista', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Spotlight on SQL Server Desktop Collect
    sossd_collect = 7981, 'sossd-collect', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7981 = 7981, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Spotlight on SQL Server Desktop Agent
    sossd_agent = 7982, 'sossd-agent', TransportProtocol.get('tcp')

    #: [UDP] Spotlight on SQL Server Desktop Agent Discovery
    sossd_disc = 7982, 'sossd-disc', TransportProtocol.get('udp')

    #: [TCP] PUSH Notification Service
    pushns = 7997, 'pushns', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_7997 = 7997, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_7998 = 7998, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] USI Content Push Service
    usicontentpush = 7998, 'usicontentpush', TransportProtocol.get('udp')

    #: - [TCP] iRDMI2
    #: - [UDP] iRDMI2
    irdmi2 = 7999, 'irdmi2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iRDMI
    #: - [UDP] iRDMI
    irdmi = 8000, 'irdmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VCOM Tunnel
    #: - [UDP] VCOM Tunnel
    vcom_tunnel = 8001, 'vcom-tunnel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Teradata ORDBMS
    #: - [UDP] Teradata ORDBMS
    teradataordbms = 8002, 'teradataordbms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mulberry Connect Reporting Service
    #: - [UDP] Mulberry Connect Reporting Service
    mcreport = 8003, 'mcreport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Opensource Evolv Enterprise Platform P2P Network Node Connection
    #: Protocol
    p2pevolvenet = 8004, 'p2pevolvenet', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8004 = 8004, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] MXI Generation II for z/OS
    #: - [UDP] MXI Generation II for z/OS
    mxi = 8005, 'mxi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] World Programming analytics
    wpl_analytics = 8006, 'wpl-analytics', TransportProtocol.get('tcp')

    #: [UDP] World Programming analytics discovery
    wpl_disc = 8006, 'wpl-disc', TransportProtocol.get('udp')

    #: - [TCP] I/O oriented cluster computing software
    #: - [UDP] I/O oriented cluster computing software
    warppipe = 8007, 'warppipe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileMaker, Inc. - HTTP Alternate (see Port 80)
    #: - [UDP] FileMaker, Inc. - HTTP Alternate (see Port 80)
    http_alt_591 = 591, 'http-alt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HTTP Alternate
    #: - [UDP] HTTP Alternate
    http_alt_8008 = 8008, 'http-alt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] NVMe over Fabrics Discovery Service
    nvme_disc = 8009, 'nvme-disc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8009 = 8009, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Reserved
    reserved_8010 = 8010, 'reserved', TransportProtocol.get('undefined')

    #: [TCP] Configuration Cloud Service
    cfg_cloud = 8015, 'cfg-cloud', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8015 = 8015, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Beckhoff Automation Device Specification
    ads_s = 8016, 'ads-s', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8016 = 8016, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8017 = 8017, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Cisco Cloudsec Dataplane Port Number
    cisco_cloudsec = 8017, 'cisco-cloudsec', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8018 = 8018, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] QB DB Dynamic Port
    #: - [UDP] QB DB Dynamic Port
    qbdb = 8019, 'qbdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intuit Entitlement Service and Discovery
    #: - [UDP] Intuit Entitlement Service and Discovery
    intu_ec_svcdisc = 8020, 'intu-ec-svcdisc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intuit Entitlement Client
    #: - [UDP] Intuit Entitlement Client
    intu_ec_client = 8021, 'intu-ec-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] oa-system
    #: - [UDP] oa-system
    oa_system = 8022, 'oa-system', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARCATrust vault API
    #: - [UDP] ARCATrust vault API
    arca_api = 8023, 'arca-api', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8024 = 8024, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] CA Audit Distribution Agent
    #: - [UDP] CA Audit Distribution Agent
    ca_audit_da = 8025, 'ca-audit-da', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA Audit Distribution Server
    #: - [UDP] CA Audit Distribution Server
    ca_audit_ds = 8026, 'ca-audit-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] peer tracker and data relay service
    #: - [UDP] peer tracker and data relay service
    papachi_p2p_srv = 8027, 'papachi-p2p-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ProEd
    #: - [UDP] ProEd
    pro_ed = 8032, 'pro-ed', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MindPrint
    #: - [UDP] MindPrint
    mindprint = 8033, 'mindprint', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] .vantronix Management
    #: - [UDP] .vantronix Management
    vantronix_mgmt = 8034, 'vantronix-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ampify Messaging Protocol
    #: - [UDP] Ampify Messaging Protocol
    ampify = 8040, 'ampify', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xcorpeon ASIC Carrier Ethernet Transport
    #: - [UDP] Xcorpeon ASIC Carrier Ethernet Transport
    enguity_xccetp = 8041, 'enguity-xccetp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] FireScope Agent
    fs_agent = 8042, 'fs-agent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8042 = 8042, 'reserved', TransportProtocol.get('udp')

    #: [TCP] FireScope Server
    fs_server = 8043, 'fs-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8043 = 8043, 'reserved', TransportProtocol.get('udp')

    #: [TCP] FireScope Management Interface
    fs_mgmt = 8044, 'fs-mgmt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8044 = 8044, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Rocrail Client Service
    rocrail = 8051, 'rocrail', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8051 = 8051, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Senomix Timesheets Server
    #: - [UDP] Senomix Timesheets Server
    senomix01 = 8052, 'senomix01', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Senomix Timesheets Client [1 year assignment]
    #: - [UDP] Senomix Timesheets Client [1 year assignment]
    senomix02 = 8053, 'senomix02', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Senomix Timesheets Server [1 year assignment]
    #: - [UDP] Senomix Timesheets Server [1 year assignment]
    senomix03 = 8054, 'senomix03', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Senomix Timesheets Server [1 year assignment]
    #: - [UDP] Senomix Timesheets Server [1 year assignment]
    senomix04 = 8055, 'senomix04', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Senomix Timesheets Server [1 year assignment]
    #: - [UDP] Senomix Timesheets Server [1 year assignment]
    senomix05 = 8056, 'senomix05', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Senomix Timesheets Client [1 year assignment]
    #: - [UDP] Senomix Timesheets Client [1 year assignment]
    senomix06 = 8057, 'senomix06', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Senomix Timesheets Client [1 year assignment]
    #: - [UDP] Senomix Timesheets Client [1 year assignment]
    senomix07 = 8058, 'senomix07', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Senomix Timesheets Client [1 year assignment]
    #: - [UDP] Senomix Timesheets Client [1 year assignment]
    senomix08 = 8059, 'senomix08', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8060 = 8060, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Asymmetric Extended Route Optimization (AERO) [:rfc:`6706`]
    aero = 8060, 'aero', TransportProtocol.get('udp')

    #: [TCP] Toad BI Application Server
    toad_bi_appsrvr = 8066, 'toad-bi-appsrvr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8066 = 8066, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Infinidat async replication
    infi_async = 8067, 'infi-async', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8067 = 8067, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Oracle Unified Communication Suite's Indexed Search Converter
    ucs_isc = 8070, 'ucs-isc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8070 = 8070, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Gadu-Gadu
    #: - [UDP] Gadu-Gadu
    gadugadu = 8074, 'gadugadu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Mles is a client-server data distribution protocol targeted to serve
    #: as a lightweight and reliable distributed publish/subscribe database
    #: service.
    mles = 8077, 'mles', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8077 = 8077, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] HTTP Alternate (see port 80)
    #: - [UDP] HTTP Alternate (see port 80)
    http_alt = 8080, 'http-alt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun Proxy Admin Service
    #: - [UDP] Sun Proxy Admin Service
    sunproxyadmin = 8081, 'sunproxyadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Utilistor (Client)
    #: - [UDP] Utilistor (Client)
    us_cli = 8082, 'us-cli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Utilistor (Server)
    #: - [UDP] Utilistor (Server)
    us_srv = 8083, 'us-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Snarl Network Protocol over HTTP
    websnp = 8084, 'websnp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8084 = 8084, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8085 = 8085, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Distributed SCADA Networking Rendezvous Port
    #: - [UDP] Distributed SCADA Networking Rendezvous Port
    d_s_n = 8086, 'd-s-n', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simplify Media SPP Protocol
    #: - [UDP] Simplify Media SPP Protocol
    simplifymedia = 8087, 'simplifymedia', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Radan HTTP
    #: - [UDP] Radan HTTP
    radan_http = 8088, 'radan-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8089 = 8089, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Vehicle to station messaging
    opsmessaging = 8090, 'opsmessaging', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8090 = 8090, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Jam Link Framework
    jamlink = 8091, 'jamlink', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8091 = 8091, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] SAC Port Id
    #: - [UDP] SAC Port Id
    sac = 8097, 'sac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xprint Server
    #: - [UDP] Xprint Server
    xprint_server = 8100, 'xprint-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Logical Domains Migration
    ldoms_migr = 8101, 'ldoms-migr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8101 = 8101, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Oracle Kernel zones migration server
    kz_migr = 8102, 'kz-migr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8102 = 8102, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8111 = 8111, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Skynetflow network services
    skynetflow = 8111, 'skynetflow', TransportProtocol.get('udp')

    #: - [TCP] MTL8000 Matrix
    #: - [UDP] MTL8000 Matrix
    mtl8000_matrix = 8115, 'mtl8000-matrix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Check Point Clustering
    #: - [UDP] Check Point Clustering
    cp_cluster = 8116, 'cp-cluster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Purity replication clustering and remote management
    purityrpc = 8117, 'purityrpc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8117 = 8117, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Privoxy HTTP proxy
    #: - [UDP] Privoxy HTTP proxy
    privoxy = 8118, 'privoxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apollo Data Port
    #: - [UDP] Apollo Data Port
    apollo_data = 8121, 'apollo-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Apollo Admin Port
    #: - [UDP] Apollo Admin Port
    apollo_admin = 8122, 'apollo-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PayCash Online Protocol
    #: - [UDP] PayCash Online Protocol
    paycash_online = 8128, 'paycash-online', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PayCash Wallet-Browser
    #: - [UDP] PayCash Wallet-Browser
    paycash_wbp = 8129, 'paycash-wbp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INDIGO-VRMI
    #: - [UDP] INDIGO-VRMI
    indigo_vrmi = 8130, 'indigo-vrmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INDIGO-VBCP
    #: - [UDP] INDIGO-VBCP
    indigo_vbcp = 8131, 'indigo-vbcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] dbabble
    #: - [UDP] dbabble
    dbabble = 8132, 'dbabble', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] The Puppet master service
    puppet = 8140, 'puppet', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8140 = 8140, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] i-SDD file transfer
    #: - [UDP] i-SDD file transfer
    isdd = 8148, 'isdd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8149 = 8149, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Edge of Reality game data
    eor_game = 8149, 'eor-game', TransportProtocol.get('udp')

    #: [TCP] QuantaStor Management Interface
    quantastor = 8153, 'quantastor', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8153 = 8153, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Patrol
    #: - [UDP] Patrol
    patrol = 8160, 'patrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Patrol SNMP
    #: - [UDP] Patrol SNMP
    patrol_snmp = 8161, 'patrol-snmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] LPAR2RRD client server communication
    lpar2rrd = 8162, 'lpar2rrd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8162 = 8162, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Intermapper network management system
    intermapper = 8181, 'intermapper', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8181 = 8181, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] VMware Fault Domain Manager
    #: - [UDP] VMware Fault Domain Manager
    vmware_fdm = 8182, 'vmware-fdm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ProRemote
    proremote = 8183, 'proremote', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8183 = 8183, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Remote iTach Connection
    #: - [UDP] Remote iTach Connection
    itach = 8184, 'itach', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Generic control plane for RPHY
    gcp_rphy = 8190, 'gcp-rphy', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8190 = 8190, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Limner Pressure
    limnerpressure = 8191, 'limnerpressure', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8191 = 8191, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] SpyTech Phone Service
    #: - [UDP] SpyTech Phone Service
    spytechphone = 8192, 'spytechphone', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8193 = 8193, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Bloomberg data API
    #: - [UDP] Bloomberg data API
    blp1 = 8194, 'blp1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bloomberg feed
    #: - [UDP] Bloomberg feed
    blp2 = 8195, 'blp2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VVR DATA
    #: - [UDP] VVR DATA
    vvr_data = 8199, 'vvr-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRIVNET
    #: - [UDP] TRIVNET
    trivnet1 = 8200, 'trivnet1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TRIVNET
    #: - [UDP] TRIVNET
    trivnet2 = 8201, 'trivnet2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8202 = 8202, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Audio+Ethernet Standard Open Protocol
    aesop = 8202, 'aesop', TransportProtocol.get('udp')

    #: - [TCP] LM Perfworks
    #: - [UDP] LM Perfworks
    lm_perfworks = 8204, 'lm-perfworks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LM Instmgr
    #: - [UDP] LM Instmgr
    lm_instmgr = 8205, 'lm-instmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LM Dta
    #: - [UDP] LM Dta
    lm_dta = 8206, 'lm-dta', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LM SServer
    #: - [UDP] LM SServer
    lm_sserver = 8207, 'lm-sserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LM Webwatcher
    #: - [UDP] LM Webwatcher
    lm_webwatcher = 8208, 'lm-webwatcher', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8211 = 8211, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Aruba Networks AP management
    aruba_papi = 8211, 'aruba-papi', TransportProtocol.get('udp')

    #: - [TCP] RexecJ Server
    #: - [UDP] RexecJ Server
    rexecj = 8230, 'rexecj', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8231 = 8231, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] HNCP [:rfc:`7788`]
    hncp_udp_port = 8231, 'hncp-udp-port', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8232 = 8232, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] HNCP over DTLS [:rfc:`7788`]
    hncp_dtls_port = 8232, 'hncp-dtls-port', TransportProtocol.get('udp')

    #: - [TCP] Synapse Non Blocking HTTPS
    #: - [UDP] Synapse Non Blocking HTTPS
    synapse_nhttps = 8243, 'synapse-nhttps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8266 = 8266, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] ESPeasy peer-2-peer communication
    espeasy_p2p = 8266, 'espeasy-p2p', TransportProtocol.get('udp')

    #: [TCP] Robot Framework Remote Library Interface
    robot_remote = 8270, 'robot-remote', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8270 = 8270, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Microsoft Connected Cache
    #: - [UDP] Microsoft Connected Cache
    ms_mcc = 8276, 'ms-mcc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Synapse Non Blocking HTTP
    #: - [UDP] Synapse Non Blocking HTTP
    synapse_nhttp = 8280, 'synapse-nhttp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8281 = 8281, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Libelle EnterpriseBus
    libelle = 8282, 'libelle', TransportProtocol.get('tcp')

    #: [UDP] Libelle EnterpriseBus discovery
    libelle_disc = 8282, 'libelle-disc', TransportProtocol.get('udp')

    #: - [SCTP] Reserved
    #: - [DCCP] Reserved
    reserved_8282 = 8282, 'reserved', TransportProtocol.get('sctp') | TransportProtocol.get('dccp')

    #: - [TCP] Bloomberg professional
    #: - [UDP] Bloomberg professional
    blp3 = 8292, 'blp3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Hiperscan Identification Service
    hiperscan_id = 8293, 'hiperscan-id', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8293 = 8293, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Bloomberg intelligent client
    #: - [UDP] Bloomberg intelligent client
    blp4 = 8294, 'blp4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Transport Management Interface
    #: - [UDP] Transport Management Interface
    tmi = 8300, 'tmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Amberon PPC/PPS
    #: - [UDP] Amberon PPC/PPS
    amberon = 8301, 'amberon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Hub Open Network
    hub_open_net = 8313, 'hub-open-net', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8313 = 8313, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Thin(ium) Network Protocol
    #: - [UDP] Thin(ium) Network Protocol
    tnp_discover = 8320, 'tnp-discover', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Thin(ium) Network Protocol
    #: - [UDP] Thin(ium) Network Protocol
    tnp = 8321, 'tnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Garmin Marine
    #: - [UDP] Garmin Marine
    garmin_marine = 8322, 'garmin-marine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Server Find
    #: - [UDP] Server Find
    server_find = 8351, 'server-find', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cruise ENUM
    #: - [UDP] Cruise ENUM
    cruise_enum = 8376, 'cruise-enum', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cruise SWROUTE
    #: - [UDP] Cruise SWROUTE
    cruise_swroute = 8377, 'cruise-swroute', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cruise CONFIG
    #: - [UDP] Cruise CONFIG
    cruise_config = 8378, 'cruise-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cruise DIAGS
    #: - [UDP] Cruise DIAGS
    cruise_diags = 8379, 'cruise-diags', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cruise UPDATE
    #: - [UDP] Cruise UPDATE
    cruise_update = 8380, 'cruise-update', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] M2m Services
    #: - [UDP] M2m Services
    m2mservices = 8383, 'm2mservices', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Marathon Transport Protocol
    marathontp = 8384, 'marathontp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8384 = 8384, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] cvd
    #: - [UDP] cvd
    cvd = 8400, 'cvd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sabarsd
    #: - [UDP] sabarsd
    sabarsd = 8401, 'sabarsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] abarsd
    #: - [UDP] abarsd
    abarsd = 8402, 'abarsd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] admind
    #: - [UDP] admind
    admind_3279 = 3279, 'admind', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] admind
    #: - [UDP] admind
    admind_8403 = 8403, 'admind', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SuperVault Cloud
    svcloud = 8404, 'svcloud', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8404 = 8404, 'reserved', TransportProtocol.get('udp')

    #: [TCP] SuperVault Backup
    svbackup = 8405, 'svbackup', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8405 = 8405, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Reserved
    reserved_8415 = 8415, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] eSpeech Session Protocol
    #: - [UDP] eSpeech Session Protocol
    espeech = 8416, 'espeech', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eSpeech RTP Protocol
    #: - [UDP] eSpeech RTP Protocol
    espeech_rtp = 8417, 'espeech-rtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Aristech text-to-speech server
    aritts = 8423, 'aritts', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8423 = 8423, 'reserved', TransportProtocol.get('udp')

    #: [TCP] PostgreSQL Backup
    pgbackrest = 8432, 'pgbackrest', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8432 = 8432, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8433 = 8433, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Non Persistent Desktop and Application Streaming
    aws_as2 = 8433, 'aws-as2', TransportProtocol.get('udp')

    #: - [TCP] CyBro A-bus Protocol
    #: - [UDP] CyBro A-bus Protocol
    cybro_a_bus = 8442, 'cybro-a-bus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCsync HTTPS
    #: - [UDP] PCsync HTTPS
    pcsync_https = 8443, 'pcsync-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCsync HTTP
    #: - [UDP] PCsync HTTP
    pcsync_http = 8444, 'pcsync-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Port for copy peer sync feature
    copy = 8445, 'copy', TransportProtocol.get('tcp')

    #: [UDP] Port for copy discovery
    copy_disc = 8445, 'copy-disc', TransportProtocol.get('udp')

    #: - [TCP] npmp
    #: - [UDP] npmp
    npmp = 8450, 'npmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Nexenta Management GUI
    nexentamv = 8457, 'nexentamv', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8457 = 8457, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Cisco Address Validation Protocol
    cisco_avp = 8470, 'cisco-avp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8470 = 8470, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] PIM over Reliable Transport [:rfc:`6559`]
    #: - [SCTP] PIM over Reliable Transport [:rfc:`6559`]
    pim_port = 8471, 'pim-port', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_8471 = 8471, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Overlay Transport Virtualization (OTV)
    #: - [UDP] Overlay Transport Virtualization (OTV)
    otv = 8472, 'otv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Point to Point
    #: - [UDP] Virtual Point to Point
    vp2p = 8473, 'vp2p', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AquaMinds NoteShare
    #: - [UDP] AquaMinds NoteShare
    noteshare = 8474, 'noteshare', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Flight Message Transfer Protocol
    #: - [UDP] Flight Message Transfer Protocol
    fmtp = 8500, 'fmtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CYTEL Message Transfer Management
    cmtp_mgt = 8501, 'cmtp-mgt', TransportProtocol.get('tcp')

    #: [UDP] CYTEL Message Transfer Audio and Video
    cmtp_av = 8501, 'cmtp-av', TransportProtocol.get('udp')

    #: [TCP] FTN Message Transfer Protocol
    ftnmtp = 8502, 'ftnmtp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8502 = 8502, 'reserved', TransportProtocol.get('udp')

    #: [UDP] MPLS LSP Self-Ping [:rfc:`7746`]
    lsp_self_ping = 8503, 'lsp-self-ping', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8503 = 8503, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] RTSP Alternate (see port 554)
    #: - [UDP] RTSP Alternate (see port 554)
    rtsp_alt = 8554, 'rtsp-alt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SYMAX D-FENCE
    #: - [UDP] SYMAX D-FENCE
    d_fence = 8555, 'd-fence', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DOF Tunneling Protocol
    #: - [UDP] DOF Tunneling Protocol
    dof_tunnel = 8567, 'dof-tunnel', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Surveillance Data
    #: - [UDP] Surveillance Data
    asterix = 8600, 'asterix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8609 = 8609, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Canon Compact Printer Protocol Discovery
    canon_cpp_disc = 8609, 'canon-cpp-disc', TransportProtocol.get('udp')

    #: - [TCP] Canon MFNP Service
    #: - [UDP] Canon MFNP Service
    canon_mfnp = 8610, 'canon-mfnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Canon BJNP Port 1
    #: - [UDP] Canon BJNP Port 1
    canon_bjnp1 = 8611, 'canon-bjnp1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Canon BJNP Port 2
    #: - [UDP] Canon BJNP Port 2
    canon_bjnp2 = 8612, 'canon-bjnp2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Canon BJNP Port 3
    #: - [UDP] Canon BJNP Port 3
    canon_bjnp3 = 8613, 'canon-bjnp3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Canon BJNP Port 4
    #: - [UDP] Canon BJNP Port 4
    canon_bjnp4 = 8614, 'canon-bjnp4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Imink Service Control
    imink = 8615, 'imink', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8615 = 8615, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Monetra
    monetra = 8665, 'monetra', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8665 = 8665, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Monetra Administrative Access
    monetra_admin = 8666, 'monetra-admin', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8666 = 8666, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Motorola Solutions Customer Programming Software for Radio Management
    msi_cps_rm = 8675, 'msi-cps-rm', TransportProtocol.get('tcp')

    #: [UDP] Motorola Solutions Customer Programming Software for Radio Management
    #: Discovery
    msi_cps_rm_disc = 8675, 'msi-cps-rm-disc', TransportProtocol.get('udp')

    #: - [TCP] Sun App Server - JMX/RMI
    #: - [UDP] Sun App Server - JMX/RMI
    sun_as_jmxrmi = 8686, 'sun-as-jmxrmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8687 = 8687, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] OpenRemote Controller HTTP/REST
    openremote_ctrl = 8688, 'openremote-ctrl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8688 = 8688, 'reserved', TransportProtocol.get('udp')

    #: [TCP] VNYX Primary Port
    vnyx = 8699, 'vnyx', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8699 = 8699, 'reserved', TransportProtocol.get('udp')

    #: [TCP] gRPC for SEMI Standards implementations
    semi_grpc = 8710, 'semi-grpc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8710 = 8710, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Nuance Voice Control
    nvc = 8711, 'nvc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8711 = 8711, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8732 = 8732, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] DASGIP Net Services
    dtp_net = 8732, 'dtp-net', TransportProtocol.get('udp')

    #: - [TCP] iBus
    #: - [UDP] iBus
    ibus = 8733, 'ibus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DEY Storage Key Negotiation
    dey_keyneg = 8750, 'dey-keyneg', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8750 = 8750, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] MC-APPSERVER
    #: - [UDP] MC-APPSERVER
    mc_appserver = 8763, 'mc-appserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPENQUEUE
    #: - [UDP] OPENQUEUE
    openqueue = 8764, 'openqueue', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ultraseek HTTP
    #: - [UDP] Ultraseek HTTP
    ultraseek_http = 8765, 'ultraseek-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Agilent Connectivity Service
    #: - [UDP] Agilent Connectivity Service
    amcs = 8766, 'amcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Online mobile multiplayer game
    core_of_source = 8767, 'core-of-source', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8767 = 8767, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Sandpolis Server
    sandpolis = 8768, 'sandpolis', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8768 = 8768, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Okta MultiPlatform Access Mgmt for Cloud Svcs
    oktaauthenticat = 8769, 'oktaauthenticat', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8769 = 8769, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Digital Photo Access Protocol (iPhoto)
    #: - [UDP] Digital Photo Access Protocol (iPhoto)
    dpap = 8770, 'dpap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Stonebranch Universal Enterprise Controller
    uec = 8778, 'uec', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8778 = 8778, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Message Client
    #: - [UDP] Message Client
    msgclnt = 8786, 'msgclnt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Message Server
    #: - [UDP] Message Server
    msgsrvr = 8787, 'msgsrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Accedian Performance Measurement
    #: - [UDP] Accedian Performance Measurement
    acd_pm = 8793, 'acd-pm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun Web Server Admin Service
    #: - [UDP] Sun Web Server Admin Service
    sunwebadmin = 8800, 'sunwebadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] truecm
    #: - [UDP] truecm
    truecm = 8804, 'truecm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8805 = 8805, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Destination Port number for PFCP
    pfcp = 8805, 'pfcp', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8806 = 8806, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Reserved
    reserved_8807 = 8807, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] HES-CLIP Interoperability protocol
    hes_clip = 8807, 'hes-clip', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8808 = 8808, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] STATSports Broadcast Service
    ssports_bcast = 8808, 'ssports-bcast', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8809 = 8809, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] MCPTT Off-Network Protocol (MONP)
    AppType_3gpp_monp = 8809, '3gpp-monp', TransportProtocol.get('udp')

    #: - [TCP] dxspider linking protocol
    #: - [UDP] dxspider linking protocol
    dxspider = 8873, 'dxspider', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CDDBP
    #: - [UDP] CDDBP
    cddbp_alt = 8880, 'cddbp-alt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Galaxy4D Online Game Engine
    galaxy4d = 8881, 'galaxy4d', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8881 = 8881, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8882 = 8882, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Secure MQTT
    #: - [UDP] Secure MQTT
    secure_mqtt = 8883, 'secure-mqtt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] NewsEDGE server TCP (TCP 1)
    ddi_tcp_1 = 8888, 'ddi-tcp-1', TransportProtocol.get('tcp')

    #: [UDP] NewsEDGE server UDP (UDP 1)
    ddi_udp_1 = 8888, 'ddi-udp-1', TransportProtocol.get('udp')

    #: [TCP] Desktop Data TCP 1
    ddi_tcp_2 = 8889, 'ddi-tcp-2', TransportProtocol.get('tcp')

    #: [UDP] NewsEDGE server broadcast
    ddi_udp_2 = 8889, 'ddi-udp-2', TransportProtocol.get('udp')

    #: [TCP] Desktop Data TCP 2
    ddi_tcp_3 = 8890, 'ddi-tcp-3', TransportProtocol.get('tcp')

    #: [UDP] NewsEDGE client broadcast
    ddi_udp_3 = 8890, 'ddi-udp-3', TransportProtocol.get('udp')

    #: [TCP] Desktop Data TCP 3: NESS application
    ddi_tcp_4 = 8891, 'ddi-tcp-4', TransportProtocol.get('tcp')

    #: [UDP] Desktop Data UDP 3: NESS application
    ddi_udp_4 = 8891, 'ddi-udp-4', TransportProtocol.get('udp')

    #: [TCP] Desktop Data TCP 4: FARM product
    ddi_tcp_5 = 8892, 'ddi-tcp-5', TransportProtocol.get('tcp')

    #: [UDP] Desktop Data UDP 4: FARM product
    ddi_udp_5 = 8892, 'ddi-udp-5', TransportProtocol.get('udp')

    #: [TCP] Desktop Data TCP 5: NewsEDGE/Web application
    ddi_tcp_6 = 8893, 'ddi-tcp-6', TransportProtocol.get('tcp')

    #: [UDP] Desktop Data UDP 5: NewsEDGE/Web application
    ddi_udp_6 = 8893, 'ddi-udp-6', TransportProtocol.get('udp')

    #: [TCP] Desktop Data TCP 6: COAL application
    ddi_tcp_7 = 8894, 'ddi-tcp-7', TransportProtocol.get('tcp')

    #: [UDP] Desktop Data UDP 6: COAL application
    ddi_udp_7 = 8894, 'ddi-udp-7', TransportProtocol.get('udp')

    #: - [TCP] ospf-lite
    #: - [UDP] ospf-lite
    ospf_lite = 8899, 'ospf-lite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JMB-CDS 1
    #: - [UDP] JMB-CDS 1
    jmb_cds1 = 8900, 'jmb-cds1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JMB-CDS 2
    #: - [UDP] JMB-CDS 2
    jmb_cds2 = 8901, 'jmb-cds2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] WFA Device Provisioning Protocol
    dpp = 8908, 'dpp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8908 = 8908, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_8909 = 8909, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] manyone-http
    #: - [UDP] manyone-http
    manyone_http = 8910, 'manyone-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] manyone-xml
    #: - [UDP] manyone-xml
    manyone_xml = 8911, 'manyone-xml', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Windows Client Backup
    #: - [UDP] Windows Client Backup
    wcbackup = 8912, 'wcbackup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Dragonfly System Service
    #: - [UDP] Dragonfly System Service
    dragonfly = 8913, 'dragonfly', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Transaction Warehouse Data Service
    twds = 8937, 'twds', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8937 = 8937, 'reserved', TransportProtocol.get('udp')

    #: [TCP] unbound dns nameserver control
    ub_dns_control = 8953, 'ub-dns-control', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8953 = 8953, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Cumulus Admin Port
    #: - [UDP] Cumulus Admin Port
    cumulus_admin = 8954, 'cumulus-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network of Devices Provider
    #: - [UDP] Network of Devices Provider
    nod_provider = 8980, 'nod-provider', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Network of Devices Client
    nod_client = 8981, 'nod-client', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_8981 = 8981, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Sun Web Server SSL Admin Service
    #: - [UDP] Sun Web Server SSL Admin Service
    sunwebadmins = 8989, 'sunwebadmins', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] webmail HTTP service
    #: - [UDP] webmail HTTP service
    http_wmap = 8990, 'http-wmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] webmail HTTPS service
    #: - [UDP] webmail HTTPS service
    https_wmap = 8991, 'https-wmap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Oracle Messaging Server Event Notification Service
    oracle_ms_ens = 8997, 'oracle-ms-ens', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8997 = 8997, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Canto RoboFlow Control
    canto_roboflow = 8998, 'canto-roboflow', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_8998 = 8998, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Brodos Crypto Trade Protocol
    #: - [UDP] Brodos Crypto Trade Protocol
    bctp = 8999, 'bctp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSlistener
    #: - [UDP] CSlistener
    cslistener = 9000, 'cslistener', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ETL Service Manager
    #: - [UDP] ETL Service Manager
    etlservicemgr = 9001, 'etlservicemgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DynamID authentication
    #: - [UDP] DynamID authentication
    dynamid = 9002, 'dynamid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Golem Inter-System RPC
    golem = 9005, 'golem', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9005 = 9005, 'reserved', TransportProtocol.get('udp')

    #: [N/A] De-Commissioned Port
    de_commissioned_port = 9006, 'de_commissioned_port', TransportProtocol.get('undefined')

    #: [TCP] Reserved
    reserved_9007 = 9007, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Open Grid Services Client
    ogs_client = 9007, 'ogs-client', TransportProtocol.get('udp')

    #: [TCP] Open Grid Services Server
    ogs_server = 9008, 'ogs-server', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9008 = 9008, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Pichat Server
    #: - [UDP] Pichat Server
    pichat = 9009, 'pichat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Secure Data Replicator Protocol
    sdr = 9010, 'sdr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9010 = 9010, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9011 = 9011, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] D-Star Routing digital voice+data for amateur radio
    d_star = 9011, 'd-star', TransportProtocol.get('udp')

    #: - [TCP] TAMBORA
    #: - [UDP] TAMBORA
    tambora = 9020, 'tambora', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pangolin Identification
    #: - [UDP] Pangolin Identification
    panagolin_ident = 9021, 'panagolin-ident', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PrivateArk Remote Agent
    #: - [UDP] PrivateArk Remote Agent
    paragent = 9022, 'paragent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Web Access - 1
    #: - [UDP] Secure Web Access - 1
    swa_1 = 9023, 'swa-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Web Access - 2
    #: - [UDP] Secure Web Access - 2
    swa_2 = 9024, 'swa-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Web Access - 3
    #: - [UDP] Secure Web Access - 3
    swa_3 = 9025, 'swa-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure Web Access - 4
    #: - [UDP] Secure Web Access - 4
    swa_4 = 9026, 'swa-4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Versiera Agent Listener
    versiera = 9050, 'versiera', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9050 = 9050, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Fusion-io Central Manager Service
    fio_cmgmt = 9051, 'fio-cmgmt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9051 = 9051, 'reserved', TransportProtocol.get('udp')

    #: [TCP] CardWeb request-response I/O exchange
    cardweb_io = 9060, 'cardweb-io', TransportProtocol.get('tcp')

    #: [UDP] CardWeb realtime device data
    cardweb_rt = 9060, 'cardweb-rt', TransportProtocol.get('udp')

    #: - [TCP] Groove GLRPC
    #: - [UDP] Groove GLRPC
    glrpc = 9080, 'glrpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9081 = 9081, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Required for Adaptive Quality of Service
    cisco_aqos = 9081, 'cisco-aqos', TransportProtocol.get('udp')

    #: [SCTP] LCS Application Protocol
    lcs_ap = 9082, 'lcs-ap', TransportProtocol.get('sctp')

    #: [TCP] EMC PowerPath Mgmt Service
    emc_pp_mgmtsvc = 9083, 'emc-pp-mgmtsvc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9083 = 9083, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] IBM AURORA Performance Visualizer
    #: - [UDP] IBM AURORA Performance Visualizer
    #: - [SCTP] IBM AURORA Performance Visualizer
    aurora = 9084, 'aurora', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] IBM Remote System Console
    #: - [UDP] IBM Remote System Console
    ibm_rsyscon = 9085, 'ibm-rsyscon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vesa Net2Display
    #: - [UDP] Vesa Net2Display
    net2display = 9086, 'net2display', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Classic Data Server
    #: - [UDP] Classic Data Server
    classic = 9087, 'classic', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Informix SQL Interface
    #: - [UDP] IBM Informix SQL Interface
    sqlexec = 9088, 'sqlexec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Informix SQL Interface - Encrypted
    #: - [UDP] IBM Informix SQL Interface - Encrypted
    sqlexec_ssl = 9089, 'sqlexec-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WebSM
    #: - [UDP] WebSM
    websm = 9090, 'websm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] xmltec-xmlmail
    #: - [UDP] xmltec-xmlmail
    xmltec_xmlmail = 9091, 'xmltec-xmlmail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xml-Ipc Server Reg
    #: - [UDP] Xml-Ipc Server Reg
    xmlipcregsvc = 9092, 'xmlipcregsvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Copycat database replication service
    copycat = 9093, 'copycat', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9093 = 9093, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] PDL Data Streaming Port
    #: - [UDP] PDL Data Streaming Port
    hp_pdl_datastr = 9100, 'hp-pdl-datastr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Printer PDL Data Stream
    #: - [UDP] Printer PDL Data Stream
    pdl_datastream = 9100, 'pdl-datastream', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bacula Director
    #: - [UDP] Bacula Director
    bacula_dir = 9101, 'bacula-dir', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bacula File Daemon
    #: - [UDP] Bacula File Daemon
    bacula_fd = 9102, 'bacula-fd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bacula Storage Daemon
    #: - [UDP] Bacula Storage Daemon
    bacula_sd = 9103, 'bacula-sd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PeerWire
    #: - [UDP] PeerWire
    peerwire = 9104, 'peerwire', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Xadmin Control Service [Ari√´n Huisken <xadmin&huisken-systems.nl> 15
    #: June 2009]
    #: - [UDP] Xadmin Control Service [Ari√´n Huisken <xadmin&huisken-systems.nl> 15
    #: June 2009]
    xadmin = 9105, 'xadmin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Astergate Control Service
    astergate = 9106, 'astergate', TransportProtocol.get('tcp')

    #: [UDP] Astergate Discovery Service
    astergate_disc = 9106, 'astergate-disc', TransportProtocol.get('udp')

    #: [TCP] AstergateFax Control Service
    astergatefax = 9107, 'astergatefax', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9107 = 9107, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Multiple Purpose, Distributed Message Bus
    #: - [UDP] Multiple Purpose, Distributed Message Bus
    hexxorecore = 9111, 'hexxorecore', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MXit Instant Messaging
    #: - [UDP] MXit Instant Messaging
    mxit = 9119, 'mxit', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Global Relay compliant mobile instant messaging protocol
    grcmp = 9122, 'grcmp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9122 = 9122, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Global Relay compliant instant messaging protocol
    grcp = 9123, 'grcp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9123 = 9123, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Dynamic Device Discovery
    #: - [UDP] Dynamic Device Discovery
    dddp = 9131, 'dddp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] apani1
    #: - [UDP] apani1
    apani1 = 9160, 'apani1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] apani2
    #: - [UDP] apani2
    apani2 = 9161, 'apani2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] apani3
    #: - [UDP] apani3
    apani3 = 9162, 'apani3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] apani4
    #: - [UDP] apani4
    apani4 = 9163, 'apani4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] apani5
    #: - [UDP] apani5
    apani5 = 9164, 'apani5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sun AppSvr JPDA
    #: - [UDP] Sun AppSvr JPDA
    sun_as_jpda = 9191, 'sun-as-jpda', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP connectionless session service
    #: - [UDP] WAP connectionless session service
    wap_wsp = 9200, 'wap-wsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP session service
    #: - [UDP] WAP session service
    wap_wsp_wtp = 9201, 'wap-wsp-wtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP secure connectionless session service
    #: - [UDP] WAP secure connectionless session service
    wap_wsp_s = 9202, 'wap-wsp-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP secure session service
    #: - [UDP] WAP secure session service
    wap_wsp_wtp_s = 9203, 'wap-wsp-wtp-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP vCard
    #: - [UDP] WAP vCard
    wap_vcard = 9204, 'wap-vcard', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP vCal
    #: - [UDP] WAP vCal
    wap_vcal = 9205, 'wap-vcal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP vCard Secure
    #: - [UDP] WAP vCard Secure
    wap_vcard_s = 9206, 'wap-vcard-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WAP vCal Secure
    #: - [UDP] WAP vCal Secure
    wap_vcal_s = 9207, 'wap-vcal-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rjcdb vCard
    #: - [UDP] rjcdb vCard
    rjcdb_vcards = 9208, 'rjcdb-vcards', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ALMobile System Service
    #: - [UDP] ALMobile System Service
    almobile_system = 9209, 'almobile-system', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA Mobile Location Protocol
    #: - [UDP] OMA Mobile Location Protocol
    oma_mlp = 9210, 'oma-mlp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OMA Mobile Location Protocol Secure
    #: - [UDP] OMA Mobile Location Protocol Secure
    oma_mlp_s = 9211, 'oma-mlp-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Server View dbms access
    #: - [UDP] Server View dbms access
    serverviewdbms = 9212, 'serverviewdbms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ServerStart RemoteControl
    #: - [UDP] ServerStart RemoteControl
    serverstart = 9213, 'serverstart', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPDC ESG BootstrapService
    #: - [UDP] IPDC ESG BootstrapService
    ipdcesgbs = 9214, 'ipdcesgbs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Integrated Setup and Install Service
    #: - [UDP] Integrated Setup and Install Service
    insis = 9215, 'insis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Aionex Communication Management Engine
    #: - [UDP] Aionex Communication Management Engine
    acme = 9216, 'acme', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FSC Communication Port
    #: - [UDP] FSC Communication Port
    fsc_port = 9217, 'fsc-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] QSC Team Coherence
    #: - [UDP] QSC Team Coherence
    teamcoherence = 9222, 'teamcoherence', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MON
    #: - [UDP] MON
    mon_2583 = 2583, 'mon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Manager On Network
    #: - [UDP] Manager On Network
    mon_9255 = 9255, 'mon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] GPS Data transmitted from train to ground network
    traingpsdata = 9277, 'traingpsdata', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9277 = 9277, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Pegasus GPS Platform
    #: - [UDP] Pegasus GPS Platform
    pegasus = 9278, 'pegasus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pegaus GPS System Control Interface
    #: - [UDP] Pegaus GPS System Control Interface
    pegasus_ctl = 9279, 'pegasus-ctl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Predicted GPS
    #: - [UDP] Predicted GPS
    pgps = 9280, 'pgps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SofaWare transport port 1
    #: - [UDP] SofaWare transport port 1
    swtp_port1 = 9281, 'swtp-port1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SofaWare transport port 2
    #: - [UDP] SofaWare transport port 2
    swtp_port2 = 9282, 'swtp-port2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CallWaveIAM
    #: - [UDP] CallWaveIAM
    callwaveiam = 9283, 'callwaveiam', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS Information Serve
    #: - [UDP] VERITAS Information Serve
    visd = 9284, 'visd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] N2H2 Filter Service Port
    #: - [UDP] N2H2 Filter Service Port
    n2h2server = 9285, 'n2h2server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9286 = 9286, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] n2 monitoring receiver
    n2receive = 9286, 'n2receive', TransportProtocol.get('udp')

    #: - [TCP] Cumulus
    #: - [UDP] Cumulus
    cumulus = 9287, 'cumulus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ArmTech Daemon
    #: - [UDP] ArmTech Daemon
    armtechdaemon = 9292, 'armtechdaemon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] StorView Client
    #: - [UDP] StorView Client
    storview = 9293, 'storview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARMCenter http Service
    #: - [UDP] ARMCenter http Service
    armcenterhttp = 9294, 'armcenterhttp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARMCenter https Service
    #: - [UDP] ARMCenter https Service
    armcenterhttps = 9295, 'armcenterhttps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Virtual Racing Service
    #: - [UDP] Virtual Racing Service
    vrace = 9300, 'vrace', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Sphinx search server (MySQL listener)
    sphinxql = 9306, 'sphinxql', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9306 = 9306, 'reserved', TransportProtocol.get('udp')

    #: [TCP] SAP Message Server
    sapms = 9310, 'sapms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9310 = 9310, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_9311 = 9311, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Sphinx search server
    sphinxapi = 9312, 'sphinxapi', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9312 = 9312, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] PKIX TimeStamp over TLS
    #: - [UDP] PKIX TimeStamp over TLS
    secure_ts = 9318, 'secure-ts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] guibase
    #: - [UDP] guibase
    guibase = 9321, 'guibase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] gRPC Network Mgmt/Operations Interface
    gnmi_gnoi = 9339, 'gnmi-gnoi', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9339 = 9339, 'reserved', TransportProtocol.get('udp')

    #: [TCP] gRPC Routing Information Base Interface
    gribi = 9340, 'gribi', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9340 = 9340, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] MpIdcMgr
    #: - [UDP] MpIdcMgr
    mpidcmgr = 9343, 'mpidcmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mphlpdmc
    #: - [UDP] Mphlpdmc
    mphlpdmc = 9344, 'mphlpdmc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Rancher Agent
    rancher = 9345, 'rancher', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9345 = 9345, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] C Tech Licensing
    #: - [UDP] C Tech Licensing
    ctechlicensing = 9346, 'ctechlicensing', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fjdmimgr
    #: - [UDP] fjdmimgr
    fjdmimgr = 9374, 'fjdmimgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Brivs! Open Extensible Protocol
    #: - [UDP] Brivs! Open Extensible Protocol
    boxp = 9380, 'boxp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] D2D Configuration Service
    d2dconfig = 9387, 'd2dconfig', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9387 = 9387, 'reserved', TransportProtocol.get('udp')

    #: [TCP] D2D Data Transfer Service
    d2ddatatrans = 9388, 'd2ddatatrans', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9388 = 9388, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Active Directory Web Services
    adws = 9389, 'adws', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9389 = 9389, 'reserved', TransportProtocol.get('udp')

    #: [TCP] OpenVAS Transfer Protocol
    otp = 9390, 'otp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9390 = 9390, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] fjinvmgr
    #: - [UDP] fjinvmgr
    fjinvmgr = 9396, 'fjinvmgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MpIdcAgt
    #: - [UDP] MpIdcAgt
    mpidcagt = 9397, 'mpidcagt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Samsung Twain for Network Server
    #: - [UDP] Samsung Twain for Network Server
    sec_t4net_srv = 9400, 'sec-t4net-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Samsung Twain for Network Client
    #: - [UDP] Samsung Twain for Network Client
    sec_t4net_clt = 9401, 'sec-t4net-clt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Samsung PC2FAX for Network Server
    #: - [UDP] Samsung PC2FAX for Network Server
    sec_pc2fax_srv = 9402, 'sec-pc2fax-srv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] git pack transfer service
    #: - [UDP] git pack transfer service
    git = 9418, 'git', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WSO2 Tungsten HTTPS
    #: - [UDP] WSO2 Tungsten HTTPS
    tungsten_https = 9443, 'tungsten-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WSO2 ESB Administration Console HTTPS
    #: - [UDP] WSO2 ESB Administration Console HTTPS
    wso2esb_console = 9444, 'wso2esb-console', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] MindArray Systems Console Agent
    mindarray_ca = 9445, 'mindarray-ca', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9445 = 9445, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Sentinel Keys Server
    #: - [UDP] Sentinel Keys Server
    sntlkeyssrvr = 9450, 'sntlkeyssrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ismserver
    #: - [UDP] ismserver
    ismserver = 9500, 'ismserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9522 = 9522, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] SMA Speedwire
    sma_spw = 9522, 'sma-spw', TransportProtocol.get('udp')

    #: - [TCP] Management Suite Remote Control
    #: - [UDP] Management Suite Remote Control
    mngsuite = 9535, 'mngsuite', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Surveillance buffering function
    #: - [UDP] Surveillance buffering function
    laes_bf = 9536, 'laes-bf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Trispen Secure Remote Access
    #: - [UDP] Trispen Secure Remote Access
    trispen_sra = 9555, 'trispen-sra', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] P4Runtime gRPC Service
    p4runtime = 9559, 'p4runtime', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9559 = 9559, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] LANDesk Gateway
    #: - [UDP] LANDesk Gateway
    ldgateway = 9592, 'ldgateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LANDesk Management Agent (cba8)
    #: - [UDP] LANDesk Management Agent (cba8)
    cba8 = 9593, 'cba8', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Message System
    #: - [UDP] Message System
    msgsys = 9594, 'msgsys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Ping Discovery Service
    #: - [UDP] Ping Discovery Service
    pds = 9595, 'pds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Mercury Discovery
    #: - [UDP] Mercury Discovery
    mercury_disc = 9596, 'mercury-disc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PD Administration
    #: - [UDP] PD Administration
    pd_admin = 9597, 'pd-admin', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Very Simple Ctrl Protocol
    #: - [UDP] Very Simple Ctrl Protocol
    vscp = 9598, 'vscp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Robix
    #: - [UDP] Robix
    robix = 9599, 'robix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MICROMUSE-NCPW
    #: - [UDP] MICROMUSE-NCPW
    micromuse_ncpw = 9600, 'micromuse-ncpw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] StreamComm User Directory
    #: - [UDP] StreamComm User Directory
    streamcomm_ds = 9612, 'streamcomm-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_9613 = 9613, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] iADT Protocol over TLS
    iadt_tls = 9614, 'iadt-tls', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9614 = 9614, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_9615 = 9615, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] eRunbook Agent IANA assigned this well-formed service name as a
    #: replacement for "erunbook_agent".
    #: - [TCP] eRunbook Agent
    erunbook_agent = 9616, 'erunbook-agent', TransportProtocol.get('tcp') | TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9616 = 9616, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] eRunbook Server IANA assigned this well-formed service name as a
    #: replacement for "erunbook_server".
    #: - [TCP] eRunbook Server
    erunbook_server = 9617, 'erunbook-server', TransportProtocol.get('tcp') | TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9617 = 9617, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Condor Collector Service
    #: - [UDP] Condor Collector Service
    condor = 9618, 'condor', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ODBC Pathway Service
    #: - [UDP] ODBC Pathway Service
    odbcpathway = 9628, 'odbcpathway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UniPort SSO Controller
    #: - [UDP] UniPort SSO Controller
    uniport = 9629, 'uniport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Peovica Controller
    peoctlr = 9630, 'peoctlr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9630 = 9630, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Peovica Collector
    peocoll = 9631, 'peocoll', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9631 = 9631, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9632 = 9632, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Mobile-C Communications
    mc_comm = 9632, 'mc-comm', TransportProtocol.get('udp')

    #: [TCP] ProQueSys Flows Service
    pqsflows = 9640, 'pqsflows', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9640 = 9640, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Zoom Control Panel Game Server Management
    zoomcp = 9666, 'zoomcp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9666 = 9666, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Cross-platform Music Multiplexing System
    #: - [UDP] Cross-platform Music Multiplexing System
    xmms2 = 9667, 'xmms2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] tec5 Spectral Device Control Protocol
    #: - [UDP] tec5 Spectral Device Control Protocol
    tec5_sdctp = 9668, 'tec5-sdctp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] T-Mobile Client Wakeup Message
    #: - [UDP] T-Mobile Client Wakeup Message
    client_wakeup = 9694, 'client-wakeup', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Content Centric Networking
    #: - [UDP] Content Centric Networking
    ccnx = 9695, 'ccnx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Board M.I.T. Service
    #: - [UDP] Board M.I.T. Service
    board_roar = 9700, 'board-roar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] L5NAS Parallel Channel
    #: - [UDP] L5NAS Parallel Channel
    l5nas_parchan = 9747, 'l5nas-parchan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Board M.I.T. Synchronous Collaboration
    #: - [UDP] Board M.I.T. Synchronous Collaboration
    board_voip = 9750, 'board-voip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] rasadv
    #: - [UDP] rasadv
    rasadv = 9753, 'rasadv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WSO2 Tungsten HTTP
    #: - [UDP] WSO2 Tungsten HTTP
    tungsten_http = 9762, 'tungsten-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WebDav Source Port
    #: - [UDP] WebDav Source Port
    davsrc = 9800, 'davsrc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Sakura Script Transfer Protocol-2
    #: - [UDP] Sakura Script Transfer Protocol-2
    sstp_2 = 9801, 'sstp-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WebDAV Source TLS/SSL
    #: - [UDP] WebDAV Source TLS/SSL
    davsrcs = 9802, 'davsrcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Session Announcement v1 [:rfc:`2974`]
    #: - [UDP] Session Announcement v1 [:rfc:`2974`]
    sapv1 = 9875, 'sapv1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Session Director
    sd = 9876, 'sd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9877 = 9877, 'reserved', TransportProtocol.get('udp')

    #: [UDP] The KX509 Kerberized Certificate Issuance Protocol in Use in 2012
    #: [:rfc:`6717`]
    kca_service = 9878, 'kca-service', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9878 = 9878, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] CYBORG Systems
    #: - [UDP] CYBORG Systems
    cyborg_systems = 9888, 'cyborg-systems', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Port for Cable network related data proxy or repeater
    #: - [UDP] Port for Cable network related data proxy or repeater
    gt_proxy = 9889, 'gt-proxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MonkeyCom
    #: - [UDP] MonkeyCom
    monkeycom = 9898, 'monkeycom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9899 = 9899, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] SCTP TUNNELING [:rfc:`6951`]
    sctp_tunneling = 9899, 'sctp-tunneling', TransportProtocol.get('udp')

    #: - [TCP] IUA
    #: - [UDP] IUA
    #: - [SCTP] IUA
    iua = 9900, 'iua', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: [UDP] enrp server channel
    enrp = 9901, 'enrp', TransportProtocol.get('udp')

    #: [SCTP] enrp server channel [:rfc:`5353`]
    enrp_sctp = 9901, 'enrp-sctp', TransportProtocol.get('sctp')

    #: [SCTP] enrp/tls server channel [:rfc:`5353`]
    enrp_sctp_tls = 9902, 'enrp-sctp-tls', TransportProtocol.get('sctp')

    #: [TCP] Reserved
    reserved_9903 = 9903, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Multicast Ping Protocol [:rfc:`6450`]
    multicast_ping = 9903, 'multicast-ping', TransportProtocol.get('udp')

    #: - [TCP] domaintime
    #: - [UDP] domaintime
    domaintime = 9909, 'domaintime', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_9910 = 9910, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] SYPECom Transport Protocol
    #: - [UDP] SYPECom Transport Protocol
    sype_transport = 9911, 'sype-transport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] XYBRID Cloud
    xybrid_cloud = 9925, 'xybrid-cloud', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9925 = 9925, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] APC 9950
    #: - [UDP] APC 9950
    apc_9950 = 9950, 'apc-9950', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 9951
    #: - [UDP] APC 9951
    apc_9951 = 9951, 'apc-9951', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APC 9952
    #: - [UDP] APC 9952
    apc_9952 = 9952, 'apc-9952', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 9953
    #: - [UDP] 9953
    acis = 9953, 'acis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] HaloteC Instrument Network Protocol
    hinp = 9954, 'hinp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9954 = 9954, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Contact Port for AllJoyn standard messaging
    alljoyn_stm = 9955, 'alljoyn-stm', TransportProtocol.get('tcp')

    #: [UDP] Contact Port for AllJoyn multiplexed constrained messaging
    alljoyn_mcm = 9955, 'alljoyn-mcm', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_9956 = 9956, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Alljoyn Name Service
    alljoyn = 9956, 'alljoyn', TransportProtocol.get('udp')

    #: - [TCP] OKI Data Network Setting Protocol
    #: - [UDP] OKI Data Network Setting Protocol
    odnsp = 9966, 'odnsp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] XYBRID RT Server
    xybrid_rt = 9978, 'xybrid-rt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9978 = 9978, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Valley Information Systems Weather station data
    visweather = 9979, 'visweather', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9979 = 9979, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_9980 = 9980, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Event sourcing database engine with a built-in programming language
    pumpkindb = 9981, 'pumpkindb', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9981 = 9981, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] DSM/SCM Target Interface
    #: - [UDP] DSM/SCM Target Interface
    dsm_scm_target = 9987, 'dsm-scm-target', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Software Essentials Secure HTTP server
    nsesrvr = 9988, 'nsesrvr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_9988 = 9988, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] OSM Applet Server
    #: - [UDP] OSM Applet Server
    osm_appsrvr = 9990, 'osm-appsrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OSM Event Server
    #: - [UDP] OSM Event Server
    osm_oev = 9991, 'osm-oev', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OnLive-1
    #: - [UDP] OnLive-1
    palace_1 = 9992, 'palace-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OnLive-2
    #: - [UDP] OnLive-2
    palace_2 = 9993, 'palace-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OnLive-3
    #: - [UDP] OnLive-3
    palace_3 = 9994, 'palace-3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Palace-4
    #: - [UDP] Palace-4
    palace_4 = 9995, 'palace-4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Palace-5
    #: - [UDP] Palace-5
    palace_5 = 9996, 'palace-5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Palace-6
    #: - [UDP] Palace-6
    palace_6 = 9997, 'palace-6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Distinct32
    #: - [UDP] Distinct32
    distinct32 = 9998, 'distinct32', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] distinct
    #: - [UDP] distinct
    distinct = 9999, 'distinct', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Data Management Protocol
    #: - [UDP] Network Data Management Protocol
    ndmp = 10000, 'ndmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCP Configuration
    #: - [UDP] SCP Configuration
    scp_config = 10001, 'scp-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMC-Documentum Content Server Product
    #: - [UDP] EMC-Documentum Content Server Product
    documentum = 10002, 'documentum', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMC-Documentum Content Server Product IANA assigned this well-formed
    #: service name as a replacement for "documentum_s".
    #: - [TCP] EMC-Documentum Content Server Product
    #: - [UDP] EMC-Documentum Content Server Product IANA assigned this well-formed
    #: service name as a replacement for "documentum_s".
    #: - [UDP] EMC-Documentum Content Server Product
    documentum_s = 10003, 'documentum-s', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [TCP] EMC Replication Manager Client
    emcrmirccd = 10004, 'emcrmirccd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10004 = 10004, 'reserved', TransportProtocol.get('udp')

    #: [TCP] EMC Replication Manager Server
    emcrmird = 10005, 'emcrmird', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10005 = 10005, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Sync replication protocol among different NetApp platforms
    netapp_sync = 10006, 'netapp-sync', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10006 = 10006, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] MVS Capacity
    #: - [UDP] MVS Capacity
    mvs_capacity = 10007, 'mvs-capacity', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Octopus Multiplexer
    #: - [UDP] Octopus Multiplexer
    octopus = 10008, 'octopus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Systemwalker Desktop Patrol
    #: - [UDP] Systemwalker Desktop Patrol
    swdtp_sv = 10009, 'swdtp-sv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ooRexx rxapi services
    rxapi = 10010, 'rxapi', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10010 = 10010, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Hardware configuration and maintenance
    abb_hw = 10020, 'abb-hw', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10020 = 10020, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Comtech EF-Data's Vipersat Management Protocol
    cefd_vmp = 10023, 'cefd-vmp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_10023 = 10023, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Zabbix Agent
    #: - [UDP] Zabbix Agent
    zabbix_agent = 10050, 'zabbix-agent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Zabbix Trapper
    #: - [UDP] Zabbix Trapper
    zabbix_trapper = 10051, 'zabbix-trapper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Quantapoint FLEXlm Licensing Service
    qptlmd = 10055, 'qptlmd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10055 = 10055, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Amanda
    #: - [UDP] Amanda
    amanda = 10080, 'amanda', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FAM Archive Server
    #: - [UDP] FAM Archive Server
    famdc = 10081, 'famdc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS ITAP DDTP
    #: - [UDP] VERITAS ITAP DDTP
    itap_ddtp = 10100, 'itap-ddtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eZmeeting
    #: - [UDP] eZmeeting
    ezmeeting_2 = 10101, 'ezmeeting-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eZproxy
    #: - [UDP] eZproxy
    ezproxy_2 = 10102, 'ezproxy-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eZrelay
    #: - [UDP] eZrelay
    ezrelay = 10103, 'ezrelay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Systemwalker Desktop Patrol
    #: - [UDP] Systemwalker Desktop Patrol
    swdtp = 10104, 'swdtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS BCTP, server
    #: - [UDP] VERITAS BCTP, server
    bctp_server = 10107, 'bctp-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NMEA-0183 Navigational Data
    #: - [UDP] NMEA-0183 Navigational Data
    nmea_0183 = 10110, 'nmea-0183', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_10111 = 10111, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] NMEA OneNet multicast messaging
    nmea_onenet = 10111, 'nmea-onenet', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_10112 = 10112, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] NetIQ Endpoint
    #: - [UDP] NetIQ Endpoint
    netiq_endpoint = 10113, 'netiq-endpoint', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetIQ Qcheck
    #: - [UDP] NetIQ Qcheck
    netiq_qcheck = 10114, 'netiq-qcheck', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetIQ Endpoint
    #: - [UDP] NetIQ Endpoint
    netiq_endpt = 10115, 'netiq-endpt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetIQ VoIP Assessor
    #: - [UDP] NetIQ VoIP Assessor
    netiq_voipa = 10116, 'netiq-voipa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetIQ IQCResource Managament Svc
    #: - [UDP] NetIQ IQCResource Managament Svc
    iqrm = 10117, 'iqrm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] HotLink CIMple REST API
    cimple = 10125, 'cimple', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10125 = 10125, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] BMC-PERFORM-SERVICE DAEMON
    #: - [UDP] BMC-PERFORM-SERVICE DAEMON
    bmc_perf_sd = 10128, 'bmc-perf-sd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] BMC General Manager Server
    bmc_gms = 10129, 'bmc-gms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10129 = 10129, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] QB Database Server
    #: - [UDP] QB Database Server
    qb_db_server = 10160, 'qb-db-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SNMP-TLS [:rfc:`6353`]
    snmptls = 10161, 'snmptls', TransportProtocol.get('tcp')

    #: [UDP] SNMP-DTLS [:rfc:`6353`]
    snmpdtls = 10161, 'snmpdtls', TransportProtocol.get('udp')

    #: [TCP] SNMP-Trap-TLS [:rfc:`6353`]
    snmptls_trap = 10162, 'snmptls-trap', TransportProtocol.get('tcp')

    #: [UDP] SNMP-Trap-DTLS [:rfc:`6353`]
    snmpdtls_trap = 10162, 'snmpdtls-trap', TransportProtocol.get('udp')

    #: - [TCP] Trigence AE Soap Service
    #: - [UDP] Trigence AE Soap Service
    trisoap = 10200, 'trisoap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Remote Server Management Service
    rsms = 10201, 'rsms', TransportProtocol.get('tcp')

    #: [UDP] Remote Server Control and Test Service
    rscs = 10201, 'rscs', TransportProtocol.get('udp')

    #: - [TCP] Apollo Relay Port
    #: - [UDP] Apollo Relay Port
    apollo_relay = 10252, 'apollo-relay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Relay of EAPOL frames
    eapol_relay = 10253, 'eapol-relay', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_10253 = 10253, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Axis WIMP Port
    #: - [UDP] Axis WIMP Port
    axis_wimp_port = 10260, 'axis-wimp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Tile remote machine learning
    tile_ml = 10261, 'tile-ml', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10261 = 10261, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Blocks
    #: - [UDP] Blocks
    blocks = 10288, 'blocks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Computer Op System Information Report
    cosir = 10321, 'cosir', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10321 = 10321, 'reserved', TransportProtocol.get('udp')

    #: [UDP] BalanceNG session table synchronization protocol
    bngsync = 10439, 'bngsync', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_10439 = 10439, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] CirrosSP Workstation Communication
    cirrossp = 10443, 'cirrossp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10443 = 10443, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_10500 = 10500, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] HIP NAT-Traversal [:rfc:`5770`][:rfc:`9028`]
    hip_nat_t = 10500, 'hip-nat-t', TransportProtocol.get('udp')

    #: - [TCP] MOS Media Object Metadata Port
    #: - [UDP] MOS Media Object Metadata Port
    mos_lower = 10540, 'mos-lower', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MOS Running Order Port
    #: - [UDP] MOS Running Order Port
    mos_upper = 10541, 'mos-upper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MOS Low Priority Port
    #: - [UDP] MOS Low Priority Port
    mos_aux = 10542, 'mos-aux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MOS SOAP Default Port
    #: - [UDP] MOS SOAP Default Port
    mos_soap = 10543, 'mos-soap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] MOS SOAP Optional Port
    #: - [UDP] MOS SOAP Optional Port
    mos_soap_opt = 10544, 'mos-soap-opt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Apple Document Sharing Service
    serverdocs = 10548, 'serverdocs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10548 = 10548, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Reserved
    reserved_10631 = 10631, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Gestor de Acaparamiento para Pocket PCs
    #: - [UDP] Gestor de Acaparamiento para Pocket PCs
    gap = 10800, 'gap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LUCIA Pareja Data Group
    #: - [UDP] LUCIA Pareja Data Group
    lpdg = 10805, 'lpdg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Linux Network Block Device
    nbd = 10809, 'nbd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10809 = 10809, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_10810 = 10810, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Nuance Mobile Care Discovery
    nmc_disc = 10810, 'nmc-disc', TransportProtocol.get('udp')

    #: - [TCP] Helix Client/Server
    #: - [UDP] Helix Client/Server
    helix = 10860, 'helix', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BVEssentials HTTP API
    #: - [UDP] BVEssentials HTTP API
    bveapi = 10880, 'bveapi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Listen port used by the Octopus Deploy Tentacle deployment agent
    octopustentacle = 10933, 'octopustentacle', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_10933 = 10933, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Auxiliary RMI Port
    #: - [UDP] Auxiliary RMI Port
    rmiaux = 10990, 'rmiaux', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IRISA
    #: - [UDP] IRISA
    irisa = 11000, 'irisa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Metasys
    #: - [UDP] Metasys
    metasys = 11001, 'metasys', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nest device-to-device and device-to-service application protocol
    #: - [UDP] Nest device-to-device and device-to-service application protocol
    weave = 11095, 'weave', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] OrigoDB Server Sync Interface
    origo_sync = 11103, 'origo-sync', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11103 = 11103, 'reserved', TransportProtocol.get('udp')

    #: [TCP] NetApp Intercluster Management
    netapp_icmgmt = 11104, 'netapp-icmgmt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11104 = 11104, 'reserved', TransportProtocol.get('udp')

    #: [TCP] NetApp Intercluster Data
    netapp_icdata = 11105, 'netapp-icdata', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11105 = 11105, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] SGI LK Licensing service
    #: - [UDP] SGI LK Licensing service
    sgi_lk = 11106, 'sgi-lk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_11107 = 11107, 'unassigned', TransportProtocol.get('undefined')

    #: [UDP] Hardware Terminals Discovery and Low-Level Communication Protocol
    myq_termlink = 11108, 'myq-termlink', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_11108 = 11108, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] Data migration facility Manager (DMF) is a browser based interface to
    #: DMF
    sgi_dmfmgr = 11109, 'sgi-dmfmgr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11109 = 11109, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Data migration facility (DMF) SOAP is a web server protocol to support
    #: remote access to DMF
    sgi_soap = 11110, 'sgi-soap', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11110 = 11110, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Viral Computing Environment (VCE)
    #: - [UDP] Viral Computing Environment (VCE)
    vce = 11111, 'vce', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DICOM
    #: - [UDP] DICOM
    dicom = 11112, 'dicom', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sun cacao snmp access point
    #: - [UDP] sun cacao snmp access point
    suncacao_snmp = 11161, 'suncacao-snmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sun cacao JMX-remoting access point
    #: - [UDP] sun cacao JMX-remoting access point
    suncacao_jmxmp = 11162, 'suncacao-jmxmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sun cacao rmi registry access point
    #: - [UDP] sun cacao rmi registry access point
    suncacao_rmi = 11163, 'suncacao-rmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sun cacao command-streaming access point
    #: - [UDP] sun cacao command-streaming access point
    suncacao_csa = 11164, 'suncacao-csa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sun cacao web service access point
    #: - [UDP] sun cacao web service access point
    suncacao_websvc = 11165, 'suncacao-websvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_11171 = 11171, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Surgical Notes Security Service Discovery (SNSS)
    snss = 11171, 'snss', TransportProtocol.get('udp')

    #: [TCP] OEM cacao JMX-remoting access point
    oemcacao_jmxmp = 11172, 'oemcacao-jmxmp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11172 = 11172, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Straton Runtime Programing
    t5_straton = 11173, 't5-straton', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11173 = 11173, 'reserved', TransportProtocol.get('udp')

    #: [TCP] OEM cacao rmi registry access point
    oemcacao_rmi = 11174, 'oemcacao-rmi', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11174 = 11174, 'reserved', TransportProtocol.get('udp')

    #: [TCP] OEM cacao web service access point
    oemcacao_websvc = 11175, 'oemcacao-websvc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11175 = 11175, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] smsqp
    #: - [UDP] smsqp
    smsqp = 11201, 'smsqp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DCSL Network Backup Services
    dcsl_backup = 11202, 'dcsl-backup', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11202 = 11202, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] WiFree Service
    #: - [UDP] WiFree Service
    wifree = 11208, 'wifree', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Memory cache service
    #: - [UDP] Memory cache service
    memcache = 11211, 'memcache', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] numerical systems messaging
    #: - [SCTP] numerical systems messaging
    xcompute = 11235, 'xcompute', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_11235 = 11235, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] IMIP
    #: - [UDP] IMIP
    imip = 11319, 'imip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IMIP Channels Port
    #: - [UDP] IMIP Channels Port
    imip_channels = 11320, 'imip-channels', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Arena Server Listen
    #: - [UDP] Arena Server Listen
    arena_server = 11321, 'arena-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ATM UHAS
    #: - [UDP] ATM UHAS
    atm_uhas = 11367, 'atm-uhas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OpenPGP HTTP Keyserver
    #: - [UDP] OpenPGP HTTP Keyserver
    hkp = 11371, 'hkp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Lenbrook Service Discovery Protocol
    lsdp = 11430, 'lsdp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_11430 = 11430, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] ASG Cypress Secure Only
    asgcypresstcps = 11489, 'asgcypresstcps', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11489 = 11489, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Tempest Protocol Port
    #: - [UDP] Tempest Protocol Port
    tempest_port = 11600, 'tempest-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] EMC XtremSW distributed config
    emc_xsw_dconfig = 11623, 'emc-xsw-dconfig', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11623 = 11623, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] H.323 Call Control Signalling Alternate
    #: - [UDP] H.323 Call Control Signalling Alternate
    h323callsigalt = 11720, 'h323callsigalt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EMC XtremSW distributed cache
    #: - [UDP] EMC XtremSW distributed cache
    emc_xsw_dcache = 11723, 'emc-xsw-dcache', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intrepid SSL
    #: - [UDP] Intrepid SSL
    intrepid_ssl = 11751, 'intrepid-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] LanSchool
    lanschool = 11796, 'lanschool', TransportProtocol.get('tcp')

    #: [UDP] Lanschool Multipoint
    lanschool_mpt = 11796, 'lanschool-mpt', TransportProtocol.get('udp')

    #: - [TCP] X2E Xoraya Multichannel protocol
    #: - [UDP] X2E Xoraya Multichannel protocol
    xoraya = 11876, 'xoraya', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_11877 = 11877, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] X2E service discovery protocol
    x2e_disc = 11877, 'x2e-disc', TransportProtocol.get('udp')

    #: - [TCP] SysInfo Service Protocol
    #: - [UDP] SysInfo Service Protocol
    sysinfo_sp = 11967, 'sysinfo-sp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] TiBS Service
    tibsd = 11971, 'tibsd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_11971 = 11971, 'reserved', TransportProtocol.get('udp')

    #: [SCTP] WorldMailExpress
    wmereceiving = 11997, 'wmereceiving', TransportProtocol.get('sctp')

    #: [SCTP] WorldMailExpress
    wmedistribution = 11998, 'wmedistribution', TransportProtocol.get('sctp')

    #: [SCTP] WorldMailExpress
    wmereporting = 11999, 'wmereporting', TransportProtocol.get('sctp')

    #: - [TCP] IBM Enterprise Extender SNA XID Exchange
    #: - [UDP] IBM Enterprise Extender SNA XID Exchange
    entextxid = 12000, 'entextxid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Enterprise Extender SNA COS Network Priority
    #: - [UDP] IBM Enterprise Extender SNA COS Network Priority
    entextnetwk = 12001, 'entextnetwk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Enterprise Extender SNA COS High Priority
    #: - [UDP] IBM Enterprise Extender SNA COS High Priority
    entexthigh = 12002, 'entexthigh', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Enterprise Extender SNA COS Medium Priority
    #: - [UDP] IBM Enterprise Extender SNA COS Medium Priority
    entextmed = 12003, 'entextmed', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IBM Enterprise Extender SNA COS Low Priority
    #: - [UDP] IBM Enterprise Extender SNA COS Low Priority
    entextlow = 12004, 'entextlow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DBISAM Database Server - Regular
    #: - [UDP] DBISAM Database Server - Regular
    dbisamserver1 = 12005, 'dbisamserver1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DBISAM Database Server - Admin
    #: - [UDP] DBISAM Database Server - Admin
    dbisamserver2 = 12006, 'dbisamserver2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Accuracer Database System Server
    #: - [UDP] Accuracer Database System Server
    accuracer = 12007, 'accuracer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Accuracer Database System Admin
    #: - [UDP] Accuracer Database System Admin
    accuracer_dbms = 12008, 'accuracer-dbms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_12009 = 12009, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Green Hills VPN
    ghvpn = 12009, 'ghvpn', TransportProtocol.get('udp')

    #: [TCP] ElevateDB Server
    edbsrvr = 12010, 'edbsrvr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_12010 = 12010, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_12011 = 12011, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Vipera Messaging Service
    #: - [UDP] Vipera Messaging Service
    vipera = 12012, 'vipera', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vipera Messaging Service over SSL Communication
    #: - [UDP] Vipera Messaging Service over SSL Communication
    vipera_ssl = 12013, 'vipera-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] RETS over SSL
    #: - [UDP] RETS over SSL
    rets_ssl = 12109, 'rets-ssl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NuPaper Session Service
    #: - [UDP] NuPaper Session Service
    nupaper_ss = 12121, 'nupaper-ss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA Web Access Service
    #: - [UDP] CA Web Access Service
    cawas = 12168, 'cawas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HiveP
    #: - [UDP] HiveP
    hivep = 12172, 'hivep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] LinoGrid Engine
    #: - [UDP] LinoGrid Engine
    linogridengine = 12300, 'linogridengine', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_12301 = 12301, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Remote Administration Daemon (RAD) is a system service that offers
    #: secure, remote, programmatic access to Solaris system configuration and run-
    #: time state
    rads = 12302, 'rads', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_12302 = 12302, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Warehouse Monitoring Syst SSS
    #: - [UDP] Warehouse Monitoring Syst SSS
    warehouse_sss = 12321, 'warehouse-sss', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Warehouse Monitoring Syst
    #: - [UDP] Warehouse Monitoring Syst
    warehouse = 12322, 'warehouse', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Italk Chat System
    #: - [UDP] Italk Chat System
    italk = 12345, 'italk', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Carbonite Server Replication Control
    carb_repl_ctrl = 12546, 'carb-repl-ctrl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_12546 = 12546, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] tsaf port
    #: - [UDP] tsaf port
    tsaf = 12753, 'tsaf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] control port for the netperf benchmark
    netperf = 12865, 'netperf', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_12865 = 12865, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] I-ZIPQD
    #: - [UDP] I-ZIPQD
    i_zipqd = 13160, 'i-zipqd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Black Crow Software application logging
    #: - [UDP] Black Crow Software application logging
    bcslogc = 13216, 'bcslogc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] R&S Proxy Installation Assistant Service
    #: - [UDP] R&S Proxy Installation Assistant Service
    rs_pias = 13217, 'rs-pias', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] EMC Virtual CAS Service
    emc_vcas_tcp = 13218, 'emc-vcas-tcp', TransportProtocol.get('tcp')

    #: [UDP] EMV Virtual CAS Service Discovery
    emc_vcas_udp = 13218, 'emc-vcas-udp', TransportProtocol.get('udp')

    #: - [TCP] PowWow Client
    #: - [UDP] PowWow Client
    powwow_client = 13223, 'powwow-client', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PowWow Server
    #: - [UDP] PowWow Server
    powwow_server = 13224, 'powwow-server', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DoIP Data
    doip_data = 13400, 'doip-data', TransportProtocol.get('tcp')

    #: [UDP] DoIP Discovery
    doip_disc = 13400, 'doip-disc', TransportProtocol.get('udp')

    #: - [TCP] BPRD Protocol (VERITAS NetBackup)
    #: - [UDP] BPRD Protocol (VERITAS NetBackup)
    bprd = 13720, 'bprd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BPDBM Protocol (VERITAS NetBackup)
    #: - [UDP] BPDBM Protocol (VERITAS NetBackup)
    bpdbm = 13721, 'bpdbm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BP Java MSVC Protocol
    #: - [UDP] BP Java MSVC Protocol
    bpjava_msvc = 13722, 'bpjava-msvc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_13723 = 13723, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Veritas Network Utility
    #: - [UDP] Veritas Network Utility
    vnetd = 13724, 'vnetd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VERITAS NetBackup
    #: - [UDP] VERITAS NetBackup
    bpcd = 13782, 'bpcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VOPIED Protocol
    #: - [UDP] VOPIED Protocol
    vopied = 13783, 'vopied', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_13784 = 13784, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] NetBackup Database
    #: - [UDP] NetBackup Database
    nbdb = 13785, 'nbdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Veritas-nomdb
    #: - [UDP] Veritas-nomdb
    nomdb = 13786, 'nomdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DSMCC Config [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    #: - [UDP] DSMCC Config [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    dsmcc_config = 13818, 'dsmcc-config', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DSMCC Session Messages [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    #: - [UDP] DSMCC Session Messages [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    dsmcc_session = 13819, 'dsmcc-session', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DSMCC Pass-Thru Messages [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    #: - [UDP] DSMCC Pass-Thru Messages [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    dsmcc_passthru = 13820, 'dsmcc-passthru', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DSMCC Download Protocol [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    #: - [UDP] DSMCC Download Protocol [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    dsmcc_download = 13821, 'dsmcc-download', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DSMCC Channel Change Protocol [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    #: - [UDP] DSMCC Channel Change Protocol [ISO/IEC 13818-6 MPEG-2 DSM-CC]
    dsmcc_ccp = 13822, 'dsmcc-ccp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Blackmagic Design Streaming Server
    bmdss = 13823, 'bmdss', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_13823 = 13823, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Certificate Management and Issuing
    a_trust_rpc = 13832, 'a-trust-rpc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_13832 = 13832, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Ultimate Control communication protocol
    #: - [UDP] Ultimate Control communication protocol
    ucontrol = 13894, 'ucontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] D-TA SYSTEMS
    #: - [UDP] D-TA SYSTEMS
    dta_systems = 13929, 'dta-systems', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] MedEvolve Port Requester
    medevolve = 13930, 'medevolve', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_13930 = 13930, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] SCOTTY High-Speed Filetransfer
    #: - [UDP] SCOTTY High-Speed Filetransfer
    scotty_ft = 14000, 'scotty-ft', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SUA
    #: - [UDP] De-Registered
    #: - [SCTP] SUA
    sua = 14001, 'sua', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: [UDP] Discovery of a SCOTTY hardware codec board
    scotty_disc = 14002, 'scotty-disc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_14002 = 14002, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] sage Best! Config Server 1
    #: - [UDP] sage Best! Config Server 1
    sage_best_com1 = 14033, 'sage-best-com1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] sage Best! Config Server 2
    #: - [UDP] sage Best! Config Server 2
    sage_best_com2 = 14034, 'sage-best-com2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VCS Application
    #: - [UDP] VCS Application
    vcs_app = 14141, 'vcs-app', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IceWall Cert Protocol
    #: - [UDP] IceWall Cert Protocol
    icpp = 14142, 'icpp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] IceWall Cert Protocol over TLS
    icpps = 14143, 'icpps', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_14143 = 14143, 'reserved', TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_14144 = 14144, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] GCM Application
    #: - [UDP] GCM Application
    gcm_app = 14145, 'gcm-app', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Veritas Traffic Director
    #: - [UDP] Veritas Traffic Director
    vrts_tdd = 14149, 'vrts-tdd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Veritas Cluster Server Command Server
    vcscmd = 14150, 'vcscmd', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_14150 = 14150, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Veritas Application Director
    #: - [UDP] Veritas Application Director
    vad = 14154, 'vad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fencing Server
    #: - [UDP] Fencing Server
    cps = 14250, 'cps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA eTrust Web Update Service
    #: - [UDP] CA eTrust Web Update Service
    ca_web_update = 14414, 'ca-web-update', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] xpra network protocol
    xpra = 14500, 'xpra', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_14500 = 14500, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] hde-lcesrvr-1
    #: - [UDP] hde-lcesrvr-1
    hde_lcesrvr_1 = 14936, 'hde-lcesrvr-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hde-lcesrvr-2
    #: - [UDP] hde-lcesrvr-2
    hde_lcesrvr_2 = 14937, 'hde-lcesrvr-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Hypack Data Aquisition
    #: - [UDP] Hypack Data Aquisition
    hydap = 15000, 'hydap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_15001 = 15001, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Open Network Environment TLS
    onep_tls = 15002, 'onep-tls', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_15002 = 15002, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_15118 = 15118, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] v2g Supply Equipment Communication Controller Discovery Protocol
    v2g_secc = 15118, 'v2g-secc', TransportProtocol.get('udp')

    #: - [TCP] XPilot Contact Port
    #: - [UDP] XPilot Contact Port
    xpilot = 15345, 'xpilot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] 3Link Negotiation
    #: - [UDP] 3Link Negotiation
    AppType_3link = 15363, '3link', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Cisco Stateful NAT
    #: - [UDP] Cisco Stateful NAT
    cisco_snat = 15555, 'cisco-snat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Backup Express Restore Server
    #: - [UDP] Backup Express Restore Server
    bex_xr = 15660, 'bex-xr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Picture Transfer Protocol
    #: - [UDP] Picture Transfer Protocol
    ptp = 15740, 'ptp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_15998 = 15998, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] 2ping Bi-Directional Ping Service
    AppType_2ping = 15998, '2ping', TransportProtocol.get('udp')

    #: [TCP] ProGrammar Enterprise
    programmar = 15999, 'programmar', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_15999 = 15999, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Administration Server Access
    fmsas = 16000, 'fmsas', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16000 = 16000, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Administration Server Connector
    fmsascon = 16001, 'fmsascon', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16001 = 16001, 'reserved', TransportProtocol.get('udp')

    #: [TCP] GoodSync Mediation Service
    gsms = 16002, 'gsms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16002 = 16002, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_16003 = 16003, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Automation and Control by REGULACE.ORG
    alfin = 16003, 'alfin', TransportProtocol.get('udp')

    #: [TCP] Filemaker Java Web Publishing Core
    jwpc = 16020, 'jwpc', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16020 = 16020, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Filemaker Java Web Publishing Core Binary
    jwpc_bin = 16021, 'jwpc-bin', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16021 = 16021, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Solaris SEA Port
    #: - [UDP] Solaris SEA Port
    sun_sea_port = 16161, 'sun-sea-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Solaris Audit - secure remote audit log
    solaris_audit = 16162, 'solaris-audit', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16162 = 16162, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] etb4j
    #: - [UDP] etb4j
    etb4j = 16309, 'etb4j', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Policy Distribute, Update Notification
    #: - [UDP] Policy Distribute, Update Notification
    pduncs = 16310, 'pduncs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Policy definition and update management
    #: - [UDP] Policy definition and update management
    pdefmns = 16311, 'pdefmns', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Serial Extension Ports One
    #: - [UDP] Network Serial Extension Ports One
    netserialext1 = 16360, 'netserialext1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Serial Extension Ports Two
    #: - [UDP] Network Serial Extension Ports Two
    netserialext2 = 16361, 'netserialext2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Serial Extension Ports Three
    #: - [UDP] Network Serial Extension Ports Three
    netserialext3 = 16367, 'netserialext3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network Serial Extension Ports Four
    #: - [UDP] Network Serial Extension Ports Four
    netserialext4 = 16368, 'netserialext4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Connected Corp
    #: - [UDP] Connected Corp
    connected = 16384, 'connected', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reliable Datagram Sockets
    rdgs = 16385, 'rdgs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16385 = 16385, 'reserved', TransportProtocol.get('udp')

    #: [TCP] X509 Objects Management Service
    xoms = 16619, 'xoms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16619 = 16619, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reliable multipath data transport for high latencies
    axon_tunnel = 16665, 'axon-tunnel', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16665 = 16665, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_16666 = 16666, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Vidder Tunnel Protocol
    vtp = 16666, 'vtp', TransportProtocol.get('udp')

    #: [TCP] This server provides callable services to mainframe External Security
    #: Managers from any TCP/IP platform
    cadsisvr = 16789, 'cadsisvr', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_16789 = 16789, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Newbay Mobile Client Update Service
    #: - [UDP] Newbay Mobile Client Update Service
    newbay_snc_mc = 16900, 'newbay-snc-mc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Simple Generic Client Interface Protocol
    #: - [UDP] Simple Generic Client Interface Protocol
    sgcip = 16950, 'sgcip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] INTEL-RCI-MP
    #: - [UDP] INTEL-RCI-MP
    intel_rci_mp = 16991, 'intel-rci-mp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intel(R) AMT SOAP/HTTP
    #: - [UDP] Intel(R) AMT SOAP/HTTP
    amt_soap_http = 16992, 'amt-soap-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intel(R) AMT SOAP/HTTPS
    #: - [UDP] Intel(R) AMT SOAP/HTTPS
    amt_soap_https = 16993, 'amt-soap-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intel(R) AMT Redirection/TCP
    #: - [UDP] Intel(R) AMT Redirection/TCP
    amt_redir_tcp = 16994, 'amt-redir-tcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Intel(R) AMT Redirection/TLS
    #: - [UDP] Intel(R) AMT Redirection/TLS
    amt_redir_tls = 16995, 'amt-redir-tls', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP]
    #: - [UDP]
    isode_dua = 17007, 'isode-dua', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Plan 9 cpu port
    ncpu = 17010, 'ncpu', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_17010 = 17010, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Vestas Data Layer Protocol
    vestasdlp = 17184, 'vestasdlp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_17184 = 17184, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Sounds Virtual
    #: - [UDP] Sounds Virtual
    soundsvirtual = 17185, 'soundsvirtual', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Chipper
    #: - [UDP] Chipper
    chipper = 17219, 'chipper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IEEE 1722 Transport Protocol for Time Sensitive Applications
    #: - [UDP] IEEE 1722 Transport Protocol for Time Sensitive Applications
    avtp = 17220, 'avtp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IEEE 1722.1 AVB Discovery, Enumeration, Connection management, and
    #: Control
    #: - [UDP] IEEE 1722.1 AVB Discovery, Enumeration, Connection management, and
    #: Control
    avdecc = 17221, 'avdecc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_17222 = 17222, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Control Plane Synchronization Protocol (SPSP)
    cpsp = 17222, 'cpsp', TransportProtocol.get('udp')

    #: [TCP] ISA100 GCI is a service utilizing a common interface between an ISA100
    #: Wireless gateway and a client application
    isa100_gci = 17223, 'isa100-gci', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_17223 = 17223, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Train Realtime Data Protocol (TRDP) Process Data
    trdp_pd = 17224, 'trdp-pd', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_17224 = 17224, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Train Realtime Data Protocol (TRDP) Message Data
    #: - [UDP] Train Realtime Data Protocol (TRDP) Message Data
    trdp_md = 17225, 'trdp-md', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Integrius Secure Tunnel Protocol
    #: - [UDP] Integrius Secure Tunnel Protocol
    integrius_stp = 17234, 'integrius-stp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSH Tectia Manager
    #: - [UDP] SSH Tectia Manager
    ssh_mgmt = 17235, 'ssh-mgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Dropbox LanSync Protocol
    db_lsp = 17500, 'db-lsp', TransportProtocol.get('tcp')

    #: [UDP] Dropbox LanSync Discovery
    db_lsp_disc = 17500, 'db-lsp-disc', TransportProtocol.get('udp')

    #: [TCP] Ailith management of routers
    ailith = 17555, 'ailith', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_17555 = 17555, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Eclipse Aviation
    #: - [UDP] Eclipse Aviation
    ea = 17729, 'ea', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Encap. ZigBee Packets
    #: - [UDP] Encap. ZigBee Packets
    zep = 17754, 'zep', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ZigBee IP Transport Service
    #: - [UDP] ZigBee IP Transport Service
    zigbee_ip = 17755, 'zigbee-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ZigBee IP Transport Secure Service
    #: - [UDP] ZigBee IP Transport Secure Service
    zigbee_ips = 17756, 'zigbee-ips', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] SolarWinds Orion
    sw_orion = 17777, 'sw-orion', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_17777 = 17777, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Beckman Instruments, Inc.
    #: - [UDP] Beckman Instruments, Inc.
    biimenu = 18000, 'biimenu', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] RAD PDF Service
    radpdf = 18104, 'radpdf', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_18104 = 18104, 'reserved', TransportProtocol.get('udp')

    #: [TCP] z/OS Resource Access Control Facility
    racf = 18136, 'racf', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_18136 = 18136, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] OPSEC CVP
    #: - [UDP] OPSEC CVP
    opsec_cvp = 18181, 'opsec-cvp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPSEC UFP
    #: - [UDP] OPSEC UFP
    opsec_ufp = 18182, 'opsec-ufp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPSEC SAM
    #: - [UDP] OPSEC SAM
    opsec_sam = 18183, 'opsec-sam', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPSEC LEA
    #: - [UDP] OPSEC LEA
    opsec_lea = 18184, 'opsec-lea', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPSEC OMI
    #: - [UDP] OPSEC OMI
    opsec_omi = 18185, 'opsec-omi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Occupational Health SC
    #: - [UDP] Occupational Health Sc
    ohsc = 18186, 'ohsc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OPSEC ELA
    #: - [UDP] OPSEC ELA
    opsec_ela = 18187, 'opsec-ela', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Check Point RTM
    #: - [UDP] Check Point RTM
    checkpoint_rtm = 18241, 'checkpoint-rtm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Checkpoint router monitoring
    iclid = 18242, 'iclid', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_18242 = 18242, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Checkpoint router state backup
    clusterxl = 18243, 'clusterxl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_18243 = 18243, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] GV NetConfig Service
    #: - [UDP] GV NetConfig Service
    gv_pf = 18262, 'gv-pf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AC Cluster
    #: - [UDP] AC Cluster
    ac_cluster = 18463, 'ac-cluster', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_18516 = 18516, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] HeyThings Device communicate service
    heythings = 18516, 'heythings', TransportProtocol.get('udp')

    #: - [TCP] Reliable Datagram Service
    #: - [UDP] Reliable Datagram Service
    rds_ib = 18634, 'rds-ib', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reliable Datagram Service over IP
    #: - [UDP] Reliable Datagram Service over IP
    rds_ip = 18635, 'rds-ip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Manufacturing Execution Systems Mesh Communication
    vdmmesh = 18668, 'vdmmesh', TransportProtocol.get('tcp')

    #: [UDP] Manufacturing Execution Systems Mesh Communication
    vdmmesh_disc = 18668, 'vdmmesh-disc', TransportProtocol.get('udp')

    #: - [TCP] IQue Protocol
    #: - [UDP] IQue Protocol
    ique = 18769, 'ique', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Infotos
    #: - [UDP] Infotos
    infotos = 18881, 'infotos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] APCNECMP
    #: - [UDP] APCNECMP
    apc_necmp = 18888, 'apc-necmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iGrid Server
    #: - [UDP] iGrid Server
    igrid = 19000, 'igrid', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Scintilla protocol for device services
    #: - [UDP] Scintilla protocol for device services
    scintilla = 19007, 'scintilla', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] J-Link TCP/IP Protocol
    j_link = 19020, 'j-link', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_19020 = 19020, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] OPSEC UAA
    #: - [UDP] OPSEC UAA
    opsec_uaa = 19191, 'opsec-uaa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] UserAuthority SecureAgent
    #: - [UDP] UserAuthority SecureAgent
    ua_secureagent = 19194, 'ua-secureagent', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Client Connection Management and Data Exchange Service
    cora = 19220, 'cora', TransportProtocol.get('tcp')

    #: [UDP] Discovery for Client Connection Management and Data Exchange Service
    cora_disc = 19220, 'cora-disc', TransportProtocol.get('udp')

    #: - [TCP] Key Server for SASSAFRAS
    #: - [UDP] Key Server for SASSAFRAS
    keysrvr = 19283, 'keysrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Key Shadow for SASSAFRAS
    #: - [UDP] Key Shadow for SASSAFRAS
    keyshadow = 19315, 'keyshadow', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] mtrgtrans
    #: - [UDP] mtrgtrans
    mtrgtrans = 19398, 'mtrgtrans', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp-sco
    #: - [UDP] hp-sco
    hp_sco = 19410, 'hp-sco', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] hp-sca
    #: - [UDP] hp-sca
    hp_sca = 19411, 'hp-sca', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] HP-SESSMON
    #: - [UDP] HP-SESSMON
    hp_sessmon = 19412, 'hp-sessmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FXUPTP
    #: - [UDP] FXUPTP
    fxuptp = 19539, 'fxuptp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SXUPTP
    #: - [UDP] SXUPTP
    sxuptp = 19540, 'sxuptp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] JCP Client
    #: - [UDP] JCP Client
    jcp = 19541, 'jcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Mesh Link Establishment
    mle = 19788, 'mle', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_19788 = 19788, 'reserved', TransportProtocol.get('tcp')

    #: [N/A] Unassigned
    unassigned_19789 = 19789, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] FairCom Database
    faircom_db = 19790, 'faircom-db', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_19790 = 19790, 'reserved', TransportProtocol.get('udp')

    #: [TCP] IEC 60870-5-104 process control - secure
    iec_104_sec = 19998, 'iec-104-sec', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_19998 = 19998, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Distributed Network Protocol - Secure
    #: - [UDP] Distributed Network Protocol - Secure
    #: - [SCTP] Distributed Network Protocol - secured
    dnp_sec = 19999, 'dnp-sec', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] DNP
    #: - [UDP] DNP
    #: - [SCTP] Distributed Network Protocol
    dnp = 20000, 'dnp', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: - [TCP] MicroSAN
    #: - [UDP] MicroSAN
    microsan = 20001, 'microsan', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Commtact HTTP
    #: - [UDP] Commtact HTTP
    commtact_http = 20002, 'commtact-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Commtact HTTPS
    #: - [UDP] Commtact HTTPS
    commtact_https = 20003, 'commtact-https', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_20004 = 20004, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] OpenWebNet protocol for electric network
    #: - [UDP] OpenWebNet protocol for electric network
    openwebnet = 20005, 'openwebnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_20012 = 20012, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Samsung Interdevice Interaction discovery
    ss_idi_disc = 20012, 'ss-idi-disc', TransportProtocol.get('udp')

    #: [TCP] Samsung Interdevice Interaction
    ss_idi = 20013, 'ss-idi', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_20013 = 20013, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] OpenDeploy Listener
    #: - [UDP] OpenDeploy Listener
    opendeploy = 20014, 'opendeploy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetBurner ID Port IANA assigned this well-formed service name as a
    #: replacement for "nburn_id".
    #: - [TCP] NetBurner ID Port
    #: - [UDP] NetBurner ID Port IANA assigned this well-formed service name as a
    #: replacement for "nburn_id".
    #: - [UDP] NetBurner ID Port
    nburn_id = 20034, 'nburn-id', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] TMOP HL7 Message Transfer Service
    #: - [UDP] TMOP HL7 Message Transfer Service
    tmophl7mts = 20046, 'tmophl7mts', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NFS mount protocol
    #: - [UDP] NFS mount protocol
    mountd = 20048, 'mountd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Network File System (NFS) over RDMA [:rfc:`8267`]
    #: - [UDP] Network File System (NFS) over RDMA [:rfc:`8267`]
    #: - [SCTP] Network File System (NFS) over RDMA [:rfc:`8267`]
    nfsrdma = 20049, 'nfsrdma', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: [TCP] AvesTerra Hypergraph Transfer Protocol (HGTP)
    avesterra = 20057, 'avesterra', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_20057 = 20057, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] TOLfab Data Change
    #: - [UDP] TOLfab Data Change
    tolfab = 20167, 'tolfab', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IPD Tunneling Port
    #: - [UDP] IPD Tunneling Port
    ipdtp_port = 20202, 'ipdtp-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iPulse-ICS
    #: - [UDP] iPulse-ICS
    ipulse_ics = 20222, 'ipulse-ics', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] emWave Message Service
    #: - [UDP] emWave Message Service
    emwavemsg = 20480, 'emwavemsg', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Track
    #: - [UDP] Track
    track = 20670, 'track', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CRTech NLM
    crtech_nlm = 20810, 'crtech-nlm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_20810 = 20810, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] At Hand MMP
    #: - [UDP] AT Hand MMP
    athand_mmp = 20999, 'athand-mmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IRTrans Control
    #: - [UDP] IRTrans Control
    irtrans = 21000, 'irtrans', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Notezilla.Lan Server
    notezilla_lan = 21010, 'notezilla-lan', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_21010 = 21010, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Distributed artificial intelligence
    trinket_agent = 21212, 'trinket-agent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_21212 = 21212, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Cohesity backup agents
    cohesity_agent = 21213, 'cohesity-agent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_21213 = 21213, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Services for Air Server
    aigairserver = 21221, 'aigairserver', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_21221 = 21221, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Raima RDM TFS
    rdm_tfs = 21553, 'rdm-tfs', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_21553 = 21553, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] MineScape Design File Server
    #: - [UDP] MineScape Design File Server
    dfserver = 21554, 'dfserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] VoFR Gateway
    #: - [UDP] VoFR Gateway
    vofr_gateway = 21590, 'vofr-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TVNC Pro Multiplexing
    #: - [UDP] TVNC Pro Multiplexing
    tvpm = 21800, 'tvpm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Safe AutoLogon
    sal = 21801, 'sal', TransportProtocol.get('tcp')

    #: - [TCP] webphone
    #: - [UDP] webphone
    webphone = 21845, 'webphone', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetSpeak Corp. Directory Services
    #: - [UDP] NetSpeak Corp. Directory Services
    netspeak_is = 21846, 'netspeak-is', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetSpeak Corp. Connection Services
    #: - [UDP] NetSpeak Corp. Connection Services
    netspeak_cs = 21847, 'netspeak-cs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetSpeak Corp. Automatic Call Distribution
    #: - [UDP] NetSpeak Corp. Automatic Call Distribution
    netspeak_acd = 21848, 'netspeak-acd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NetSpeak Corp. Credit Processing System
    #: - [UDP] NetSpeak Corp. Credit Processing System
    netspeak_cps = 21849, 'netspeak-cps', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SNAPenetIO
    #: - [UDP] SNAPenetIO
    snapenetio = 22000, 'snapenetio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] OptoControl
    #: - [UDP] OptoControl
    optocontrol = 22001, 'optocontrol', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Opto Host Port 2
    #: - [UDP] Opto Host Port 2
    optohost002 = 22002, 'optohost002', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Opto Host Port 3
    #: - [UDP] Opto Host Port 3
    optohost003 = 22003, 'optohost003', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Opto Host Port 4
    #: - [UDP] Opto Host Port 4
    optohost004_22004 = 22004, 'optohost004', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Opto Host Port 5
    #: - [UDP] Opto Host Port 5
    optohost004_22005 = 22005, 'optohost004', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] dCache Access Protocol
    dcap = 22125, 'dcap', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_22125 = 22125, 'reserved', TransportProtocol.get('udp')

    #: [TCP] GSI dCache Access Protocol
    gsidcap = 22128, 'gsidcap', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_22128 = 22128, 'reserved', TransportProtocol.get('udp')

    #: [TCP] EasyEngine is CLI tool to manage WordPress Sites on Nginx server
    easyengine = 22222, 'easyengine', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_22222 = 22222, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] wnn6
    #: - [UDP] wnn6
    wnn6 = 22273, 'wnn6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CompactIS Tunnel
    #: - [UDP] CompactIS Tunnel
    cis = 22305, 'cis', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ShowCockpit Networking
    #: - [UDP] ShowCockpit Networking
    showcockpit_net = 22333, 'showcockpit-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_22334 = 22334, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Initium Labs Security and Automation Control
    shrewd_control = 22335, 'shrewd-control', TransportProtocol.get('tcp')

    #: [UDP] Initium Labs Security and Automation Streaming
    shrewd_stream = 22335, 'shrewd-stream', TransportProtocol.get('udp')

    #: - [TCP] CompactIS Secure Tunnel
    #: - [UDP] CompactIS Secure Tunnel
    cis_secure = 22343, 'cis-secure', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] WibuKey Standard WkLan
    #: - [UDP] WibuKey Standard WkLan
    wibukey = 22347, 'wibukey', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CodeMeter Standard
    #: - [UDP] CodeMeter Standard
    codemeter = 22350, 'codemeter', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] TPC/IP requests of copy protection software to a server
    codemeter_cmwan = 22351, 'codemeter-cmwan', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_22351 = 22351, 'reserved', TransportProtocol.get('udp')

    #: [TCP] CaldSoft Backup server file transfer
    caldsoft_backup = 22537, 'caldsoft-backup', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_22537 = 22537, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Vocaltec Web Conference
    vocaltec_wconf = 22555, 'vocaltec-wconf', TransportProtocol.get('tcp')

    #: [UDP] Vocaltec Internet Phone
    vocaltec_phone = 22555, 'vocaltec-phone', TransportProtocol.get('udp')

    #: - [TCP] Talika Main Server
    #: - [UDP] Talika Main Server
    talikaserver = 22763, 'talikaserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Telerate Information Platform LAN
    #: - [UDP] Telerate Information Platform LAN
    aws_brf = 22800, 'aws-brf', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Telerate Information Platform WAN
    #: - [UDP] Telerate Information Platform WAN
    brf_gw = 22951, 'brf-gw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inova LightLink Server Type 1
    #: - [UDP] Inova LightLink Server Type 1
    inovaport1 = 23000, 'inovaport1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inova LightLink Server Type 2
    #: - [UDP] Inova LightLink Server Type 2
    inovaport2 = 23001, 'inovaport2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inova LightLink Server Type 3
    #: - [UDP] Inova LightLink Server Type 3
    inovaport3 = 23002, 'inovaport3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inova LightLink Server Type 4
    #: - [UDP] Inova LightLink Server Type 4
    inovaport4 = 23003, 'inovaport4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inova LightLink Server Type 5
    #: - [UDP] Inova LightLink Server Type 5
    inovaport5 = 23004, 'inovaport5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Inova LightLink Server Type 6
    #: - [UDP] Inova LightLink Server Type 6
    inovaport6 = 23005, 'inovaport6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Generic Notification Transport Protocol
    gntp = 23053, 'gntp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_23053 = 23053, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_23272 = 23272, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] S102 application
    s102 = 23272, 's102', TransportProtocol.get('udp')

    #: [TCP] 5AFE SDN Directory
    AppType_5afe_dir = 23294, '5afe-dir', TransportProtocol.get('tcp')

    #: [UDP] 5AFE SDN Directory discovery
    AppType_5afe_disc = 23294, '5afe-disc', TransportProtocol.get('udp')

    #: - [TCP] Emulex HBAnyware Remote Management
    #: - [UDP] Emulex HBAnyware Remote Management
    elxmgmt = 23333, 'elxmgmt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novar Data
    #: - [UDP] Novar Data
    novar_dbase = 23400, 'novar-dbase', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novar Alarm
    #: - [UDP] Novar Alarm
    novar_alarm = 23401, 'novar-alarm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Novar Global
    #: - [UDP] Novar Global
    novar_global = 23402, 'novar-global', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Aequus Service
    aequus = 23456, 'aequus', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_23456 = 23456, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Aequus Service Mgmt
    aequus_alt = 23457, 'aequus-alt', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_23457 = 23457, 'reserved', TransportProtocol.get('udp')

    #: [TCP] AreaGuard Neo - WebServer
    areaguard_neo = 23546, 'areaguard-neo', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_23546 = 23546, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] med-ltp
    #: - [UDP] med-ltp
    med_ltp = 24000, 'med-ltp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] med-fsp-rx
    #: - [UDP] med-fsp-rx
    med_fsp_rx = 24001, 'med-fsp-rx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] med-fsp-tx
    #: - [UDP] med-fsp-tx
    med_fsp_tx = 24002, 'med-fsp-tx', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] med-supp
    #: - [UDP] med-supp
    med_supp = 24003, 'med-supp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] med-ovw
    #: - [UDP] med-ovw
    med_ovw = 24004, 'med-ovw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] med-ci
    #: - [UDP] med-ci
    med_ci = 24005, 'med-ci', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] med-net-svc
    #: - [UDP] med-net-svc
    med_net_svc = 24006, 'med-net-svc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] fileSphere
    #: - [UDP] fileSphere
    filesphere = 24242, 'filesphere', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Vista 4GL
    #: - [UDP] Vista 4GL
    vista_4gl = 24249, 'vista-4gl', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Isolv Local Directory
    #: - [UDP] Isolv Local Directory
    ild = 24321, 'ild', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Transport of Human Interface Device data streams
    hid = 24322, 'hid', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_24322 = 24322, 'reserved', TransportProtocol.get('tcp')

    #: [TCP] Verimag mobile class protocol over TCP
    vrmg_ip = 24323, 'vrmg-ip', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_24323 = 24323, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Intel RCI IANA assigned this well-formed service name as a replacement
    #: for "intel_rci".
    #: - [TCP] Intel RCI
    #: - [UDP] Intel RCI IANA assigned this well-formed service name as a replacement
    #: for "intel_rci".
    #: - [UDP] Intel RCI
    intel_rci = 24386, 'intel-rci', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Tonido Domain Server
    #: - [UDP] Tonido Domain Server
    tonidods = 24465, 'tonidods', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] BINKP
    #: - [UDP] BINKP
    binkp = 24554, 'binkp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] bilobit Service
    bilobit = 24577, 'bilobit', TransportProtocol.get('tcp')

    #: [UDP] bilobit Service Update
    bilobit_update = 24577, 'bilobit-update', TransportProtocol.get('udp')

    #: [TCP] Service used by SmarDTV to communicate between a CAM and a second
    #: screen application
    sdtvwcam = 24666, 'sdtvwcam', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_24666 = 24666, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Canditv Message Service
    #: - [UDP] Canditv Message Service
    canditv = 24676, 'canditv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FlashFiler
    #: - [UDP] FlashFiler
    flashfiler = 24677, 'flashfiler', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Turbopower Proactivate
    #: - [UDP] Turbopower Proactivate
    proactivate = 24678, 'proactivate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_24679 = 24679, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] TCC User HTTP Service
    #: - [UDP] TCC User HTTP Service
    tcc_http = 24680, 'tcc-http', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_24727 = 24727, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Citrix StorageLink Gateway
    cslg = 24754, 'cslg', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_24754 = 24754, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_24850 = 24850, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Device Association Discovery
    assoc_disc = 24850, 'assoc-disc', TransportProtocol.get('udp')

    #: - [TCP] Find Identification of Network Devices
    #: - [UDP] Find Identification of Network Devices
    find = 24922, 'find', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase1
    #: - [UDP] icl-twobase1
    icl_twobase1 = 25000, 'icl-twobase1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase2
    #: - [UDP] icl-twobase2
    icl_twobase2 = 25001, 'icl-twobase2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase3
    #: - [UDP] icl-twobase3
    icl_twobase3 = 25002, 'icl-twobase3', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase4
    #: - [UDP] icl-twobase4
    icl_twobase4 = 25003, 'icl-twobase4', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase5
    #: - [UDP] icl-twobase5
    icl_twobase5 = 25004, 'icl-twobase5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase6
    #: - [UDP] icl-twobase6
    icl_twobase6 = 25005, 'icl-twobase6', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase7
    #: - [UDP] icl-twobase7
    icl_twobase7 = 25006, 'icl-twobase7', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase8
    #: - [UDP] icl-twobase8
    icl_twobase8 = 25007, 'icl-twobase8', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase9
    #: - [UDP] icl-twobase9
    icl_twobase9 = 25008, 'icl-twobase9', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] icl-twobase10
    #: - [UDP] icl-twobase10
    icl_twobase10 = 25009, 'icl-twobase10', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] RNSAP User Adaptation for Iurh
    rna = 25471, 'rna', TransportProtocol.get('sctp')

    #: [TCP] Sauter Dongle
    sauterdongle = 25576, 'sauterdongle', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_25576 = 25576, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Identifier Tracing Protocol
    idtp = 25604, 'idtp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_25604 = 25604, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Vocaltec Address Server
    #: - [UDP] Vocaltec Address Server
    vocaltec_hos = 25793, 'vocaltec-hos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TASP Network Comm
    #: - [UDP] TASP Network Comm
    tasp_net = 25900, 'tasp-net', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NIObserver
    #: - [UDP] NIObserver
    niobserver = 25901, 'niobserver', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NILinkAnalyst
    #: - [UDP] NILinkAnalyst
    nilinkanalyst = 25902, 'nilinkanalyst', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NIProbe
    #: - [UDP] NIProbe
    niprobe = 25903, 'niprobe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Bitfighter game server
    bf_game = 25954, 'bf-game', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_25954 = 25954, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] Bitfighter master server
    bf_master = 25955, 'bf-master', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_25955 = 25955, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] quake
    #: - [UDP] quake
    quake = 26000, 'quake', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Symbolic Computation Software Composability Protocol
    #: - [UDP] Symbolic Computation Software Composability Protocol
    scscp = 26133, 'scscp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] wnn6-ds
    #: - [UDP] wnn6-ds
    wnn6_ds = 26208, 'wnn6-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] CockroachDB
    cockroach = 26257, 'cockroach', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_26257 = 26257, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] eZproxy
    #: - [UDP] eZproxy
    ezproxy = 26260, 'ezproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] eZmeeting
    #: - [UDP] eZmeeting
    ezmeeting = 26261, 'ezmeeting', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] K3 Software-Server
    #: - [UDP] K3 Software-Server
    k3software_svr = 26262, 'k3software-svr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] K3 Software-Client
    #: - [UDP] K3 Software-Client
    k3software_cli = 26263, 'k3software-cli', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] De-registered
    de_registered = 26264, 'de_registered', TransportProtocol.get('undefined')

    #: [TCP] EXOline-TCP
    exoline_tcp = 26486, 'exoline-tcp', TransportProtocol.get('tcp')

    #: [UDP] EXOline-UDP
    exoline_udp = 26486, 'exoline-udp', TransportProtocol.get('udp')

    #: - [TCP] EXOconfig
    #: - [UDP] EXOconfig
    exoconfig = 26487, 'exoconfig', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_26488 = 26488, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] EXOnet
    #: - [UDP] EXOnet
    exonet = 26489, 'exonet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] A protocol for managing license services
    flex_lmadmin = 27010, 'flex-lmadmin', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_27010 = 27010, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Mongo database system
    mongodb = 27017, 'mongodb', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_27017 = 27017, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ImagePump
    #: - [UDP] ImagePump
    imagepump = 27345, 'imagepump', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Job controller service
    #: - [UDP] Job controller service
    jesmsjc = 27442, 'jesmsjc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Kopek HTTP Head Port
    #: - [UDP] Kopek HTTP Head Port
    kopek_httphead = 27504, 'kopek-httphead', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ARS VISTA Application
    #: - [UDP] ARS VISTA Application
    ars_vista = 27782, 'ars-vista', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Astrolink Protocol
    astrolink = 27876, 'astrolink', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_27876 = 27876, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] TW Authentication/Key Distribution and
    #: - [UDP] Attribute Certificate Services
    tw_auth_key = 27999, 'tw-auth-key', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NX License Manager
    #: - [UDP] NX License Manager
    nxlmd = 28000, 'nxlmd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PQ Service
    pqsp = 28001, 'pqsp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_28001 = 28001, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Gruber cash registry protocol
    gruber_cashreg = 28010, 'gruber-cashreg', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_28010 = 28010, 'reserved', TransportProtocol.get('udp')

    #: [TCP] thor/server - ML engine
    thor_engine = 28080, 'thor-engine', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_28080 = 28080, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_28119 = 28119, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] A27 cdma2000 RAN Management
    a27_ran_ran = 28119, 'a27-ran-ran', TransportProtocol.get('udp')

    #: - [TCP] VoxelStorm game server
    #: - [UDP] VoxelStorm game server
    voxelstorm = 28200, 'voxelstorm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Siemens GSM
    #: - [UDP] Siemens GSM
    siemensgsm = 28240, 'siemensgsm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Building operating system services wide area verified exchange
    bosswave = 28589, 'bosswave', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_28589 = 28589, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Siemens Licensing Server
    saltd_licensing = 29000, 'saltd-licensing', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_29000 = 29000, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_29118 = 29118, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] SGsAP in 3GPP
    sgsap = 29118, 'sgsap', TransportProtocol.get('sctp')

    #: - [TCP] ObTools Message Protocol
    #: - [UDP] ObTools Message Protocol
    otmp = 29167, 'otmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_29168 = 29168, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] SBcAP in 3GPP
    sbcap = 29168, 'sbcap', TransportProtocol.get('sctp')

    #: [SCTP] HNBAP and RUA Common Association
    iuhsctpassoc = 29169, 'iuhsctpassoc', TransportProtocol.get('sctp')

    #: [TCP] data exchange protocol for IEC61850 in wind power plants
    bingbang = 29999, 'bingbang', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_29999 = 29999, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Secure Network Data Management Protocol
    ndmps = 30000, 'ndmps', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_30000 = 30000, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Pago Services 1
    #: - [UDP] Pago Services 1
    pago_services1 = 30001, 'pago-services1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Pago Services 2
    #: - [UDP] Pago Services 2
    pago_services2 = 30002, 'pago-services2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Amicon FPSU-IP Remote Administration
    #: - [UDP] Amicon FPSU-IP Remote Administration
    amicon_fpsu_ra = 30003, 'amicon-fpsu-ra', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Amicon FPSU-IP VPN
    amicon_fpsu_s = 30004, 'amicon-fpsu-s', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_30004 = 30004, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Remote Window Protocol
    #: - [SCTP] Remote Window Protocol
    rwp = 30100, 'rwp', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [UDP] Reserved
    reserved_30100 = 30100, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Kingdoms Online (CraigAvenue)
    #: - [UDP] Kingdoms Online (CraigAvenue)
    kingdomsonline = 30260, 'kingdomsonline', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] GroundStar RealTime System
    gs_realtime = 30400, 'gs-realtime', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_30400 = 30400, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Samsung Convergence Discovery Protocol
    samsung_disc = 30832, 'samsung-disc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_30832 = 30832, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] OpenView Service Desk Client
    #: - [UDP] OpenView Service Desk Client
    ovobs = 30999, 'ovobs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Kollective Agent Secure Distributed Delivery Protocol
    ka_sddp = 31016, 'ka-sddp', TransportProtocol.get('tcp')

    #: [UDP] Kollective Agent Kollective Delivery Protocol
    ka_kdp = 31016, 'ka-kdp', TransportProtocol.get('udp')

    #: [TCP] Autotrac ACP 245
    autotrac_acp = 31020, 'autotrac-acp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_31020 = 31020, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_31029 = 31029, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] YaWN - Yet Another Windows Notifier
    yawn = 31029, 'yawn', TransportProtocol.get('udp')

    #: - [TCP] eldim is a secure file upload proxy
    #: - [UDP] eldim is a secure file upload proxy
    eldim = 31337, 'eldim', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] PACE license server
    pace_licensed = 31400, 'pace-licensed', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_31400 = 31400, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] XQoS network monitor
    #: - [UDP] XQoS network monitor
    xqosd = 31416, 'xqosd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] TetriNET Protocol
    #: - [UDP] TetriNET Protocol
    tetrinet = 31457, 'tetrinet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] lm mon
    #: - [UDP] lm mon
    lm_mon = 31620, 'lm-mon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DS Expert Monitor IANA assigned this well-formed service name as a
    #: replacement for "dsx_monitor".
    #: - [TCP] DS Expert Monitor
    dsx_monitor = 31685, 'dsx-monitor', TransportProtocol.get('tcp') | TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_31685 = 31685, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] GameSmith Port
    #: - [UDP] GameSmith Port
    gamesmith_port = 31765, 'gamesmith-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Embedded Device Configuration Protocol TX IANA assigned this well-
    #: formed service name as a replacement for "iceedcp_tx".
    #: - [TCP] Embedded Device Configuration Protocol TX
    #: - [UDP] Embedded Device Configuration Protocol TX IANA assigned this well-
    #: formed service name as a replacement for "iceedcp_tx".
    #: - [UDP] Embedded Device Configuration Protocol TX
    iceedcp_tx = 31948, 'iceedcp-tx', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] Embedded Device Configuration Protocol RX IANA assigned this well-
    #: formed service name as a replacement for "iceedcp_rx".
    #: - [TCP] Embedded Device Configuration Protocol RX
    #: - [UDP] Embedded Device Configuration Protocol RX IANA assigned this well-
    #: formed service name as a replacement for "iceedcp_rx".
    #: - [UDP] Embedded Device Configuration Protocol RX
    iceedcp_rx = 31949, 'iceedcp-rx', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: - [TCP] iRacing helper service
    #: - [UDP] iRacing helper service
    iracinghelper = 32034, 'iracinghelper', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] T1 Distributed Processor
    #: - [UDP] T1 Distributed Processor
    t1distproc60 = 32249, 't1distproc60', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Plex multimedia
    plex = 32400, 'plex', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_32400 = 32400, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Access Point Manager Link
    #: - [UDP] Access Point Manager Link
    apm_link = 32483, 'apm-link', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SecureNotebook-CLNT
    #: - [UDP] SecureNotebook-CLNT
    sec_ntb_clnt = 32635, 'sec-ntb-clnt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DMExpress
    #: - [UDP] DMExpress
    dmexpress = 32636, 'dmexpress', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileNet BPM WS-ReliableMessaging Client
    #: - [UDP] FileNet BPM WS-ReliableMessaging Client
    filenet_powsrm = 32767, 'filenet-powsrm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Filenet TMS
    #: - [UDP] Filenet TMS
    filenet_tms = 32768, 'filenet-tms', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Filenet RPC
    #: - [UDP] Filenet RPC
    filenet_rpc = 32769, 'filenet-rpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Filenet NCH
    #: - [UDP] Filenet NCH
    filenet_nch = 32770, 'filenet-nch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileNET RMI
    #: - [UDP] FileNet RMI
    filenet_rmi = 32771, 'filenet-rmi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileNET Process Analyzer
    #: - [UDP] FileNET Process Analyzer
    filenet_pa = 32772, 'filenet-pa', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileNET Component Manager
    #: - [UDP] FileNET Component Manager
    filenet_cm = 32773, 'filenet-cm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileNET Rules Engine
    #: - [UDP] FileNET Rules Engine
    filenet_re = 32774, 'filenet-re', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Performance Clearinghouse
    #: - [UDP] Performance Clearinghouse
    filenet_pch = 32775, 'filenet-pch', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileNET BPM IOR
    #: - [UDP] FileNET BPM IOR
    filenet_peior = 32776, 'filenet-peior', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] FileNet BPM CORBA
    #: - [UDP] FileNet BPM CORBA
    filenet_obrok = 32777, 'filenet-obrok', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Multiple Listing Service Network
    #: - [UDP] Multiple Listing Service Network
    mlsn = 32801, 'mlsn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Real Estate Transport Protocol
    retp = 32811, 'retp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_32811 = 32811, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Attachmate ID Manager
    #: - [UDP] Attachmate ID Manager
    idmgratm = 32896, 'idmgratm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] WatchGuard Endpoint Communications
    wg_endpt_comms = 33000, 'wg-endpt-comms', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_33000 = 33000, 'reserved', TransportProtocol.get('udp')

    #: [TCP] MySQL Database Extended Interface
    mysqlx = 33060, 'mysqlx', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_33060 = 33060, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Aurora (Balaena Ltd)
    #: - [UDP] Aurora (Balaena Ltd)
    aurora_balaena = 33123, 'aurora-balaena', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] DiamondCentral Interface
    #: - [UDP] DiamondCentral Interface
    diamondport = 33331, 'diamondport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_33332 = 33332, 'unassigned', TransportProtocol.get('undefined')

    #: [TCP] Digital Gaslight Service
    dgi_serv = 33333, 'dgi-serv', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_33333 = 33333, 'reserved', TransportProtocol.get('udp')

    #: [TCP] SpeedTrace TraceAgent
    speedtrace = 33334, 'speedtrace', TransportProtocol.get('tcp')

    #: [UDP] SpeedTrace TraceAgent Discovery
    speedtrace_disc = 33334, 'speedtrace-disc', TransportProtocol.get('udp')

    #: - [TCP] traceroute use
    #: - [UDP] traceroute use
    traceroute = 33434, 'traceroute', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_33435 = 33435, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] IP Multicast Traceroute [:rfc:`8487`]
    mtrace = 33435, 'mtrace', TransportProtocol.get('udp')

    #: - [TCP] SNIP Slave
    #: - [UDP] SNIP Slave
    snip_slave = 33656, 'snip-slave', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Adept IP protocol
    digilent_adept = 33890, 'digilent-adept', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_33890 = 33890, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] TurboNote Relay Server Default Port
    #: - [UDP] TurboNote Relay Server Default Port
    turbonote_2 = 34249, 'turbonote-2', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] P-Net on IP local
    #: - [UDP] P-Net on IP local
    p_net_local = 34378, 'p-net-local', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] P-Net on IP remote
    #: - [UDP] P-Net on IP remote
    p_net_remote = 34379, 'p-net-remote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] dhanalakshmi.org EDI Service
    dhanalakshmi = 34567, 'dhanalakshmi', TransportProtocol.get('tcp')

    #: [UDP] dhanalakshmi.org EDI Service
    edi_service = 34567, 'edi_service', TransportProtocol.get('udp')

    #: - [TCP] PROFInet RT Unicast
    #: - [UDP] PROFInet RT Unicast
    profinet_rt = 34962, 'profinet-rt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PROFInet RT Multicast
    #: - [UDP] PROFInet RT Multicast
    profinet_rtm = 34963, 'profinet-rtm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PROFInet Context Manager
    #: - [UDP] PROFInet Context Manager
    profinet_cm = 34964, 'profinet-cm', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] EtherCAT Port
    #: - [UDP] EtherCAT Port
    ethercat = 34980, 'ethercat', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] HeathView
    heathview = 35000, 'heathview', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_35000 = 35000, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ReadyTech Viewer
    #: - [UDP] ReadyTech Viewer
    rt_viewer = 35001, 'rt-viewer', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ReadyTech Sound Server
    rt_sound = 35002, 'rt-sound', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_35002 = 35002, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ReadyTech DeviceMapper Server
    rt_devicemapper = 35003, 'rt-devicemapper', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_35003 = 35003, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] ReadyTech ClassManager
    #: - [UDP] ReadyTech ClassManager
    rt_classmanager = 35004, 'rt-classmanager', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ReadyTech LabTracker
    rt_labtracker = 35005, 'rt-labtracker', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_35005 = 35005, 'reserved', TransportProtocol.get('udp')

    #: [TCP] ReadyTech Helper Service
    rt_helper = 35006, 'rt-helper', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_35006 = 35006, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Axiomatic discovery protocol
    #: - [UDP] Axiomatic discovery protocol
    axio_disc = 35100, 'axio-disc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] KIT Messenger
    kitim = 35354, 'kitim', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_35354 = 35354, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Altova License Management
    altova_lm = 35355, 'altova-lm', TransportProtocol.get('tcp')

    #: [UDP] Altova License Management Discovery
    altova_lm_disc = 35355, 'altova-lm-disc', TransportProtocol.get('udp')

    #: [TCP] Gutters Note Exchange
    guttersnex = 35356, 'guttersnex', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_35356 = 35356, 'reserved', TransportProtocol.get('udp')

    #: [TCP] OpenStack ID Service
    openstack_id = 35357, 'openstack-id', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_35357 = 35357, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] AllPeers Network
    #: - [UDP] AllPeers Network
    allpeers = 36001, 'allpeers', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Wireless LAN Control plane Protocol (WLCP)
    wlcp = 36411, 'wlcp', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_36411 = 36411, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_36412 = 36412, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] S1-Control Plane (3GPP)
    s1_control = 36412, 's1-control', TransportProtocol.get('sctp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_36422 = 36422, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] X2-Control Plane (3GPP)
    x2_control = 36422, 'x2-control', TransportProtocol.get('sctp')

    #: [SCTP] SLm Interface Application Protocol
    slmap = 36423, 'slmap', TransportProtocol.get('sctp')

    #: [SCTP] Nq and Nq' Application Protocol
    nq_ap = 36424, 'nq-ap', TransportProtocol.get('sctp')

    #: [SCTP] M2 Application Part
    m2ap = 36443, 'm2ap', TransportProtocol.get('sctp')

    #: [SCTP] M3 Application Part
    m3ap = 36444, 'm3ap', TransportProtocol.get('sctp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_36462 = 36462, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] Xw-Control Plane (3GPP)
    xw_control = 36462, 'xw-control', TransportProtocol.get('sctp')

    #: [TCP] Febooti Automation Workshop
    febooti_aw = 36524, 'febooti-aw', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_36524 = 36524, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Observium statistics collection agent
    observium_agent = 36602, 'observium-agent', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_36602 = 36602, 'reserved', TransportProtocol.get('udp')

    #: [TCP] MapX communication
    mapx = 36700, 'mapx', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_36700 = 36700, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] KastenX Pipe
    #: - [UDP] KastenX Pipe
    kastenxpipe = 36865, 'kastenxpipe', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_37472 = 37472, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] W1 signalling transport
    AppType_3gpp_w1ap = 37472, '3gpp-w1ap', TransportProtocol.get('sctp')

    #: - [TCP] science + computing's Venus Administration Port
    #: - [UDP] science + computing's Venus Administration Port
    neckar = 37475, 'neckar', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Google Drive Sync
    gdrive_sync = 37483, 'gdrive-sync', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_37483 = 37483, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Epipole File Transfer Protocol
    eftp = 37601, 'eftp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_37601 = 37601, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Unisys ClearPath ePortal
    #: - [UDP] Unisys ClearPath ePortal
    unisys_eportal = 37654, 'unisys-eportal', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] InfoVista Server Database
    ivs_database = 38000, 'ivs-database', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_38000 = 38000, 'reserved', TransportProtocol.get('udp')

    #: [TCP] InfoVista Server Insertion
    ivs_insertion = 38001, 'ivs-insertion', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_38001 = 38001, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Cresco Controller
    cresco_control = 38002, 'cresco-control', TransportProtocol.get('tcp')

    #: [UDP] Cresco Controller Discovery
    crescoctrl_disc = 38002, 'crescoctrl-disc', TransportProtocol.get('udp')

    #: - [TCP] Galaxy7 Data Tunnel
    #: - [UDP] Galaxy7 Data Tunnel
    galaxy7_data = 38201, 'galaxy7-data', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Fairview Message Service
    #: - [UDP] Fairview Message Service
    fairview = 38202, 'fairview', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] AppGate Policy Server
    #: - [UDP] AppGate Policy Server
    agpolicy = 38203, 'agpolicy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_38412 = 38412, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] NG Control Plane (3GPP)
    ng_control = 38412, 'ng-control', TransportProtocol.get('sctp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_38422 = 38422, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] Xn Control Plane (3GPP)
    xn_control = 38422, 'xn-control', TransportProtocol.get('sctp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_38462 = 38462, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] E1 signalling transport (3GPP)
    e1_interface = 38462, 'e1-interface', TransportProtocol.get('sctp')

    #: - [TCP] Reserved
    #: - [UDP] Reserved
    reserved_38472 = 38472, 'reserved', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [SCTP] F1 Control Plane (3GPP)
    f1_control = 38472, 'f1-control', TransportProtocol.get('sctp')

    #: [TCP] Premier SQL Middleware Server
    psqlmws = 38638, 'psqlmws', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_38638 = 38638, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Sruth is a service for the distribution of routinely-       generated
    #: but arbitrary files based on a publish/subscribe       distribution model
    #: and implemented using a peer-to-peer transport       mechanism
    sruth = 38800, 'sruth', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_38800 = 38800, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Security approval process for use of the secRMM SafeCopy program
    secrmmsafecopya = 38865, 'secrmmsafecopya', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_38865 = 38865, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Children's hearing test/Telemedicine
    vroa = 39063, 'vroa', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_39063 = 39063, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] TurboNote Default Port
    #: - [UDP] TurboNote Default Port
    turbonote_1 = 39681, 'turbonote-1', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SafetyNET p
    #: - [UDP] SafetyNET p
    safetynetp = 40000, 'safetynetp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_40023 = 40023, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] K-PatentsSensorInformation
    k_patentssensor = 40023, 'k-patentssensor', TransportProtocol.get('udp')

    #: [TCP] Simplify Printing TX
    sptx = 40404, 'sptx', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_40404 = 40404, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] CSCP
    #: - [UDP] CSCP
    cscp = 40841, 'cscp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSCCREDIR
    #: - [UDP] CSCCREDIR
    csccredir = 40842, 'csccredir', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CSCCFIREWALL
    #: - [UDP] CSCCFIREWALL
    csccfirewall = 40843, 'csccfirewall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_40853 = 40853, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] ORTEC Service Discovery
    ortec_disc = 40853, 'ortec-disc', TransportProtocol.get('udp')

    #: - [TCP] Foursticks QoS Protocol
    #: - [UDP] Foursticks QoS Protocol
    fs_qos = 41111, 'fs-qos', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Tentacle Server
    tentacle = 41121, 'tentacle', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_41121 = 41121, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Z-Wave Protocol over SSL/TLS
    #: - [UDP] Z-Wave Protocol over DTLS
    z_wave_s = 41230, 'z-wave-s', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Crestron Control Port
    #: - [UDP] Crestron Control Port
    crestron_cip = 41794, 'crestron-cip', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Crestron Terminal Port
    #: - [UDP] Crestron Terminal Port
    crestron_ctp = 41795, 'crestron-ctp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Crestron Secure Control Port
    crestron_cips = 41796, 'crestron-cips', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_41796 = 41796, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Crestron Secure Terminal Port
    crestron_ctps = 41797, 'crestron-ctps', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_41797 = 41797, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Computer Associates network discovery protocol
    #: - [UDP] Computer Associates network discovery protocol
    candp = 42508, 'candp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA discovery response
    #: - [UDP] CA discovery response
    candrp = 42509, 'candrp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] CA eTrust RPC
    #: - [UDP] CA eTrust RPC
    caerpc = 42510, 'caerpc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] API endpoint for search application
    curiosity = 42999, 'curiosity', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_42999 = 42999, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Receiver Remote Control
    recvr_rc = 43000, 'recvr-rc', TransportProtocol.get('tcp')

    #: [UDP] Receiver Remote Control Discovery
    recvr_rc_disc = 43000, 'recvr-rc-disc', TransportProtocol.get('udp')

    #: - [TCP] REACHOUT
    #: - [UDP] REACHOUT
    reachout = 43188, 'reachout', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NDM-AGENT-PORT
    #: - [UDP] NDM-AGENT-PORT
    ndm_agent_port = 43189, 'ndm-agent-port', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] IP-PROVISION
    #: - [UDP] IP-PROVISION
    ip_provision = 43190, 'ip-provision', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reconnoiter Agent Data Transport
    noit_transport = 43191, 'noit-transport', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_43191 = 43191, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Shaper Automation Server Management
    shaperai = 43210, 'shaperai', TransportProtocol.get('tcp')

    #: [UDP] Shaper Automation Server Management Discovery
    shaperai_disc = 43210, 'shaperai-disc', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_43438 = 43438, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] HmIP LAN Routing
    hmip_routing = 43438, 'hmip-routing', TransportProtocol.get('udp')

    #: [TCP] EQ3 firmware update
    eq3_update = 43439, 'eq3-update', TransportProtocol.get('tcp')

    #: [UDP] EQ3 discovery and configuration
    eq3_config = 43439, 'eq3-config', TransportProtocol.get('udp')

    #: [TCP] Cisco EnergyWise Management
    ew_mgmt = 43440, 'ew-mgmt', TransportProtocol.get('tcp')

    #: [UDP] Cisco EnergyWise Discovery and Command Flooding
    ew_disc_cmd = 43440, 'ew-disc-cmd', TransportProtocol.get('udp')

    #: - [TCP] Cisco NetMgmt DB Ports
    #: - [UDP] Cisco NetMgmt DB Ports
    ciscocsdb = 43441, 'ciscocsdb', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Z-Wave Secure Tunnel
    z_wave_tunnel = 44123, 'z-wave-tunnel', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_44123 = 44123, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] PCP server (pmcd)
    #: - [UDP] PCP server (pmcd)
    pmcd = 44321, 'pmcd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] PCP server (pmcd) proxy
    #: - [UDP] PCP server (pmcd) proxy
    pmcdproxy = 44322, 'pmcdproxy', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] HTTP binding for Performance Co-Pilot client API
    pmwebapi = 44323, 'pmwebapi', TransportProtocol.get('tcp')

    #: [UDP] Unassigned
    unassigned_44323 = 44323, 'unassigned', TransportProtocol.get('udp')

    #: [TCP] Cognex DataMan Management Protocol
    cognex_dataman = 44444, 'cognex-dataman', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_44444 = 44444, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Acronis Backup Gateway service port
    acronis_backup = 44445, 'acronis-backup', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_44445 = 44445, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_44544 = 44544, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] DOMIQ Building Automation
    domiq = 44544, 'domiq', TransportProtocol.get('udp')

    #: - [TCP] REALbasic Remote Debug
    #: - [UDP] REALbasic Remote Debug
    rbr_debug = 44553, 'rbr-debug', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_44600 = 44600, 'reserved', TransportProtocol.get('tcp')

    #: [UDP] AudioScience HPI
    asihpi = 44600, 'asihpi', TransportProtocol.get('udp')

    #: - [TCP] EtherNet/IP messaging IANA assigned this well-formed service name as a
    #: replacement for "EtherNet/IP-2".
    #: - [TCP] EtherNet/IP messaging
    #: - [UDP] EtherNet/IP messaging IANA assigned this well-formed service name as a
    #: replacement for "EtherNet/IP-2".
    #: - [UDP] EtherNet/IP messaging
    ethernet_ip_2 = 44818, 'ethernet-ip-2', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [TCP] M3DA is used for efficient machine-to-machine communications
    m3da = 44900, 'm3da', TransportProtocol.get('tcp')

    #: [UDP] M3DA Discovery is used for efficient machine-to-machine communications
    m3da_disc = 44900, 'm3da-disc', TransportProtocol.get('udp')

    #: [TCP] Nuance AutoStore Status Monitoring Protocol (data transfer)
    asmp = 45000, 'asmp', TransportProtocol.get('tcp')

    #: [UDP] Nuance AutoStore Status Monitoring Protocol (device monitoring)
    asmp_mon = 45000, 'asmp-mon', TransportProtocol.get('udp')

    #: [TCP] Nuance AutoStore Status Monitoring Protocol (secure data transfer)
    asmps = 45001, 'asmps', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_45001 = 45001, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Redspeed Status Monitor
    rs_status = 45002, 'rs-status', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_45002 = 45002, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Remote application control protocol
    synctest = 45045, 'synctest', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_45045 = 45045, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] InVision AG
    #: - [UDP] InVision AG
    invision_ag = 45054, 'invision-ag', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ASSIA CloudCheck WiFi Management System
    cloudcheck = 45514, 'cloudcheck', TransportProtocol.get('tcp')

    #: [UDP] ASSIA CloudCheck WiFi Management keepalive
    cloudcheck_ping = 45514, 'cloudcheck-ping', TransportProtocol.get('udp')

    #: - [TCP] EBA PRISE
    #: - [UDP] EBA PRISE
    eba = 45678, 'eba', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Server for the DAI family of client-server products
    dai_shell = 45824, 'dai-shell', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_45824 = 45824, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Qpuncture Data Access Service
    #: - [UDP] Qpuncture Data Access Service
    qdb2service = 45825, 'qdb2service', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SSRServerMgr
    #: - [UDP] SSRServerMgr
    ssr_servermgr = 45966, 'ssr-servermgr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Listen port used for Inedo agent communication
    inedo = 46336, 'inedo', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_46336 = 46336, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Connection between a desktop computer or server and a signature tablet
    #: to capture handwritten signatures
    spremotetablet = 46998, 'spremotetablet', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_46998 = 46998, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] MediaBox Server
    #: - [UDP] MediaBox Server
    mediabox = 46999, 'mediabox', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Message Bus
    #: - [UDP] Message Bus
    mbus = 47000, 'mbus', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Windows Remote Management Service
    winrm = 47001, 'winrm', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_47001 = 47001, 'reserved', TransportProtocol.get('udp')

    #: [UDP] Configuration of motors connected to Industrial Ethernet
    jvl_mactalk = 47100, 'jvl-mactalk', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_47100 = 47100, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Databeam Corporation
    #: - [UDP] Databeam Corporation
    dbbrowse = 47557, 'dbbrowse', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Direct Play Server
    #: - [UDP] Direct Play Server
    directplaysrvr = 47624, 'directplaysrvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ALC Protocol
    #: - [UDP] ALC Protocol
    ap = 47806, 'ap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Unassigned
    unassigned_47807 = 47807, 'unassigned', TransportProtocol.get('undefined')

    #: - [TCP] Building Automation and Control Networks
    #: - [UDP] Building Automation and Control Networks
    bacnet = 47808, 'bacnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] PreSonus Universal Control Network Protocol
    presonus_ucnet = 47809, 'presonus-ucnet', TransportProtocol.get('udp')

    #: [TCP] Reserved
    reserved_47809 = 47809, 'reserved', TransportProtocol.get('tcp')

    #: - [TCP] Nimbus Controller
    #: - [UDP] Nimbus Controller
    nimcontroller = 48000, 'nimcontroller', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nimbus Spooler
    #: - [UDP] Nimbus Spooler
    nimspooler = 48001, 'nimspooler', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nimbus Hub
    #: - [UDP] Nimbus Hub
    nimhub = 48002, 'nimhub', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Nimbus Gateway
    #: - [UDP] Nimbus Gateway
    nimgtw = 48003, 'nimgtw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] NimbusDB Connector
    nimbusdb = 48004, 'nimbusdb', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_48004 = 48004, 'reserved', TransportProtocol.get('udp')

    #: [TCP] NimbusDB Control
    nimbusdbctrl = 48005, 'nimbusdbctrl', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_48005 = 48005, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Juliar Programming Language Protocol
    juka = 48048, 'juka', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_48048 = 48048, 'reserved', TransportProtocol.get('udp')

    #: [TCP] 3GPP Cell Broadcast Service Protocol
    AppType_3gpp_cbsp = 48049, '3gpp-cbsp', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_48049 = 48049, 'reserved', TransportProtocol.get('udp')

    #: [TCP] WeFi Access Network Discovery and Selection Function
    weandsf = 48050, 'weandsf', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_48050 = 48050, 'reserved', TransportProtocol.get('udp')

    #: - [TCP] Image Systems Network Services
    #: - [UDP] Image Systems Network Services
    isnetserv = 48128, 'isnetserv', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Bloomberg locator
    #: - [UDP] Bloomberg locator
    blp5 = 48129, 'blp5', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] com-bardac-dw
    #: - [UDP] com-bardac-dw
    com_bardac_dw = 48556, 'com-bardac-dw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] iqobject
    #: - [UDP] iqobject
    iqobject = 48619, 'iqobject', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Robot Raconteur transport
    #: - [UDP] Robot Raconteur transport
    robotraconteur = 48653, 'robotraconteur', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Matahari Broker
    matahari = 49000, 'matahari', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_49000 = 49000, 'reserved', TransportProtocol.get('udp')

    #: [TCP] Nuance Unity Service Request Protocol
    nusrp = 49001, 'nusrp', TransportProtocol.get('tcp')

    #: [UDP] Nuance Unity Service Discovery Protocol
    nusdp_disc = 49001, 'nusdp-disc', TransportProtocol.get('udp')

    #: [TCP] InSpider System
    inspider = 49150, 'inspider', TransportProtocol.get('tcp')

    #: [UDP] Reserved
    reserved_49150 = 49150, 'reserved', TransportProtocol.get('udp')

    #: [N/A] IANA Reserved
    iana_reserved = 49151, 'iana_reserved', TransportProtocol.get('undefined')

    #: [N/A] ARGUS Protocol
    argus = -1, 'argus', TransportProtocol.get('undefined')

    #: [N/A] Address Resolution Protocol
    arp = -1, 'arp', TransportProtocol.get('undefined')

    #: [N/A] BBN RCC Monitoring
    bbn_rcc_mon = -1, 'bbn-rcc-mon', TransportProtocol.get('undefined')

    #: [N/A] Bootstrap Protocol
    bootp = -1, 'bootp', TransportProtocol.get('undefined')

    #: [N/A] Backroom SATNET Monitoring
    br_sat_mon = -1, 'br-sat-mon', TransportProtocol.get('undefined')

    #: [N/A] CFTP
    cftp = -1, 'cftp', TransportProtocol.get('undefined')

    #: [N/A] CHAOS Protocol
    chaos = -1, 'chaos', TransportProtocol.get('undefined')

    #: [N/A] DCNET Time Server Protocol
    clock = -1, 'clock', TransportProtocol.get('undefined')

    #: [N/A] Common Mgmnt Info Ser and Prot over TCP/IP
    cmot = -1, 'cmot', TransportProtocol.get('undefined')

    #: [N/A] Authentication Scheme
    cookie_jar = -1, 'cookie-jar', TransportProtocol.get('undefined')

    #: [N/A] DCN Measurement Subsystems Protocol
    dcn_meas = -1, 'dcn-meas', TransportProtocol.get('undefined')

    #: [N/A] Dissimilar Gateway Protocol
    dgp = -1, 'dgp', TransportProtocol.get('undefined')

    #: [N/A] Digest Message Format for Mail
    dmf_mail = -1, 'dmf-mail', TransportProtocol.get('undefined')

    #: [N/A] Exterior Gateway Protocol
    egp = -1, 'egp', TransportProtocol.get('undefined')

    #: [N/A] Encoding Header Field for Mail
    ehf_mail = -1, 'ehf-mail', TransportProtocol.get('undefined')

    #: [N/A] Emission Control Protocol
    emcon = -1, 'emcon', TransportProtocol.get('undefined')

    #: [N/A] Fujitsu Config Protocol
    fconfig = -1, 'fconfig', TransportProtocol.get('undefined')

    #: [N/A] Gateway Gateway Protocol
    ggp = -1, 'ggp', TransportProtocol.get('undefined')

    #: [N/A] Host Monitoring Protocol
    hmp = -1, 'hmp', TransportProtocol.get('undefined')

    #: [N/A] Host2 Name Server
    host2_ns = -1, 'host2-ns', TransportProtocol.get('undefined')

    #: [N/A] Internet Control Message Protocol
    icmp = -1, 'icmp', TransportProtocol.get('undefined')

    #: [N/A] Internet Group Management Protocol
    igmp = -1, 'igmp', TransportProtocol.get('undefined')

    #: [N/A] Interior Gateway Protocol
    igp = -1, 'igp', TransportProtocol.get('undefined')

    #: [N/A] Interim Mail Access Protocol version 2
    imap2 = -1, 'imap2', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol
    ip = -1, 'ip', TransportProtocol.get('undefined')

    #: [N/A] Internet Packet Core Utility
    ipcu = -1, 'ipcu', TransportProtocol.get('undefined')

    #: [N/A] Internet Pluribus Packet Core
    ippc = -1, 'ippc', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on ARCNET
    ip_arc = -1, 'ip-arc', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on ARPANET
    ip_arpa = -1, 'ip-arpa', TransportProtocol.get('undefined')

    #: [N/A] Compressing TCP/IP Headers
    ip_cmprs = -1, 'ip-cmprs', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on DC Networks
    ip_dc = -1, 'ip-dc', TransportProtocol.get('undefined')

    #: [N/A] Distance Vector Multicast Routing Protocol
    ip_dvmrp = -1, 'ip-dvmrp', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on Ethernet Networks
    ip_e = -1, 'ip-e', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on Exp. Ethernet Nets
    ip_ee = -1, 'ip-ee', TransportProtocol.get('undefined')

    #: [N/A] Transmission of IP over FDDI
    ip_fddi = -1, 'ip-fddi', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on Hyperchannnel
    ip_hc = -1, 'ip-hc', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on IEEE 802
    ip_ieee = -1, 'ip-ieee', TransportProtocol.get('undefined')

    #: [N/A] Transmission of 802.2 over IPX Networks
    ip_ipx = -1, 'ip-ipx', TransportProtocol.get('undefined')

    #: [N/A] IP MTU Discovery Options
    ip_mtu = -1, 'ip-mtu', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol over NetBIOS Networks
    ip_netbios = -1, 'ip-netbios', TransportProtocol.get('undefined')

    #: [N/A] Transmission of IP over Serial Lines
    ip_slip = -1, 'ip-slip', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on Wideband Network
    ip_wb = -1, 'ip-wb', TransportProtocol.get('undefined')

    #: [N/A] Internet Protocol on X.25 Networks
    ip_x25 = -1, 'ip-x25', TransportProtocol.get('undefined')

    #: [N/A] Internet Reliable Transaction Protocol
    irtp = -1, 'irtp', TransportProtocol.get('undefined')

    #: [N/A] ISO Transport Protocol Class 4
    iso_tp4 = -1, 'iso-tp4', TransportProtocol.get('undefined')

    #: [N/A] Locus Address Resoultion Protocol
    larp = -1, 'larp', TransportProtocol.get('undefined')

    #: [N/A] Leaf-1 Protocol
    leaf_1 = -1, 'leaf-1', TransportProtocol.get('undefined')

    #: [N/A] Leaf-2 Protocol
    leaf_2 = -1, 'leaf-2', TransportProtocol.get('undefined')

    #: [N/A] Location Service
    loc_srv = -1, 'loc-srv', TransportProtocol.get('undefined')

    #: [N/A] Format of Electronic Mail Messages
    mail = -1, 'mail', TransportProtocol.get('undefined')

    #: [N/A] MERIT Internodal Protocol
    merit_inp = -1, 'merit-inp', TransportProtocol.get('undefined')

    #: [N/A] Management Information Base
    mib = -1, 'mib', TransportProtocol.get('undefined')

    #: [N/A] MIH Command Services [:rfc:`5679`]
    mihcs = -1, 'mihcs', TransportProtocol.get('undefined')

    #: [N/A] MIH Event Services [:rfc:`5679`]
    mihes = -1, 'mihes', TransportProtocol.get('undefined')

    #: [N/A] MIH Information Services [:rfc:`5679`]
    mihis = -1, 'mihis', TransportProtocol.get('undefined')

    #: [N/A] MFE Network Services Protocol
    mfe_nsp = -1, 'mfe-nsp', TransportProtocol.get('undefined')

    #: [N/A] MIT Subnet Support
    mit_subnet = -1, 'mit-subnet', TransportProtocol.get('undefined')

    #: [N/A] Multiplexing Protocol
    mux = -1, 'mux', TransportProtocol.get('undefined')

    #: [N/A] Bulk Data Transfer Protocol
    netblt = -1, 'netblt', TransportProtocol.get('undefined')

    #: [N/A] Network Standard Text Editor
    neted = -1, 'neted', TransportProtocol.get('undefined')

    #: [N/A] Remote Job Service
    netrjs = -1, 'netrjs', TransportProtocol.get('undefined')

    #: [N/A] A File Access Protocol
    nfile = -1, 'nfile', TransportProtocol.get('undefined')

    #: [N/A] Network Voice Protocol
    nvp_ii = -1, 'nvp-ii', TransportProtocol.get('undefined')

    #: [N/A] Open Shortest Path First Interior GW Protocol
    ospf = -1, 'ospf', TransportProtocol.get('undefined')

    #: [N/A] Pcmail Transport Protocol
    pcmail = -1, 'pcmail', TransportProtocol.get('undefined')

    #: [N/A] Point-to-Point Protocol
    ppp = -1, 'ppp', TransportProtocol.get('undefined')

    #: [N/A] Packet Radio Measurement
    prm = -1, 'prm', TransportProtocol.get('undefined')

    #: [N/A] PUP Protocol
    pup = -1, 'pup', TransportProtocol.get('undefined')

    #: [N/A] Quote of the Day Protocol
    quote = -1, 'quote', TransportProtocol.get('undefined')

    #: [N/A] A Reverse Address Resolution Protocol
    rarp = -1, 'rarp', TransportProtocol.get('undefined')

    #: [N/A] Reliable Asynchronous Transfer Protocol
    ratp = -1, 'ratp', TransportProtocol.get('undefined')

    #: - [N/A] Reliable Data Protocol
    #: - [N/A] Windows Remote Desktop Protocol
    rdp = -1, 'rdp', TransportProtocol.get('undefined') | TransportProtocol.get('undefined')

    #: [N/A] Routing Information Protocol
    rip = -1, 'rip', TransportProtocol.get('undefined')

    #: [N/A] Remote Virtual Disk Protocol
    rvd = -1, 'rvd', TransportProtocol.get('undefined')

    #: [N/A] Satnet and Backroom EXPAK
    sat_expak = -1, 'sat-expak', TransportProtocol.get('undefined')

    #: [N/A] SATNET Monitoring
    sat_mon = -1, 'sat-mon', TransportProtocol.get('undefined')

    #: [N/A] Structure of Management Information
    smi = -1, 'smi', TransportProtocol.get('undefined')

    #: [N/A] Stream Protocol
    stp = -1, 'stp', TransportProtocol.get('undefined')

    #: [N/A] SUN Remote Procedure Call
    sun_rpc = -1, 'sun-rpc', TransportProtocol.get('undefined')

    #: [N/A] Transmission Control Protocol
    tcp = -1, 'tcp', TransportProtocol.get('undefined')

    #: [N/A] TCP Alternate Checksum Option
    tcp_aco = -1, 'tcp-aco', TransportProtocol.get('undefined')

    #: [N/A] Thinwire Protocol
    thinwire = -1, 'thinwire', TransportProtocol.get('undefined')

    #: [N/A] ISO Transport Service on top of the TCP
    tp_tcp = -1, 'tp-tcp', TransportProtocol.get('undefined')

    #: [N/A] Trunk-1 Protocol
    trunk_1 = -1, 'trunk-1', TransportProtocol.get('undefined')

    #: [N/A] Trunk-2 Protocol
    trunk_2 = -1, 'trunk-2', TransportProtocol.get('undefined')

    #: [N/A] University College London Protocol
    ucl = -1, 'ucl', TransportProtocol.get('undefined')

    #: [N/A] User Datagram Protocol
    udp = -1, 'udp', TransportProtocol.get('undefined')

    #: [N/A] Active Users Protocol
    users = -1, 'users', TransportProtocol.get('undefined')

    #: [N/A] VIA Systems-File Transfer Protocol
    via_ftp = -1, 'via-ftp', TransportProtocol.get('undefined')

    #: [N/A] VISA Protocol
    visa = -1, 'visa', TransportProtocol.get('undefined')

    #: [N/A] Versatile Message Transaction Protocol
    vmtp = -1, 'vmtp', TransportProtocol.get('undefined')

    #: [N/A] Wideband EXPAK
    wb_expak = -1, 'wb-expak', TransportProtocol.get('undefined')

    #: [N/A] Wideband Monitoring
    wb_mon = -1, 'wb-mon', TransportProtocol.get('undefined')

    #: [N/A] Cross Net Debugger
    xnet = -1, 'xnet', TransportProtocol.get('undefined')

    #: [N/A] Xerox NS IDP
    xns_idp = -1, 'xns-idp', TransportProtocol.get('undefined')

    #: [N/A] 1Password Password Manager data sharing and synchronization protocol
    AppType_1password = -1, '1password', TransportProtocol.get('undefined')

    #: [N/A] Altos Design Synchronization protocol
    a_d_sync = -1, 'a-d-sync', TransportProtocol.get('undefined')

    #: [N/A] Applied Biosystems Universal Instrument Framework
    abi_instrument = -1, 'abi-instrument', TransportProtocol.get('undefined')

    #: [N/A] FTK2 Database Discovery Service
    accessdata_f2d = -1, 'accessdata-f2d', TransportProtocol.get('undefined')

    #: [N/A] FTK2 Backend Processing Agent Service
    accessdata_f2w = -1, 'accessdata-f2w', TransportProtocol.get('undefined')

    #: [N/A] Strix Systems 5S/AccessOne protocol
    accessone = -1, 'accessone', TransportProtocol.get('undefined')

    #: [N/A] MYOB AccountEdge
    accountedge = -1, 'accountedge', TransportProtocol.get('undefined')

    #: [N/A] Adobe Acrobat
    acrobatsrv = -1, 'acrobatsrv', TransportProtocol.get('undefined')

    #: - [TCP] Access Control Device
    #: - [UDP] Access Control Device
    acs_ctl_ds = -1, 'acs-ctl-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Access Control Gateway
    #: - [UDP] Access Control Gateway
    acs_ctl_gw = -1, 'acs-ctl-gw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] ActionItems
    actionitems = -1, 'actionitems', TransportProtocol.get('undefined')

    #: [N/A] Active Storage Proprietary Device Management Protocol
    activeraid = -1, 'activeraid', TransportProtocol.get('undefined')

    #: [N/A] Encrypted transport of Active Storage Proprietary Device Management
    #: Protocol
    activeraid_ssl = -1, 'activeraid-ssl', TransportProtocol.get('undefined')

    #: - [TCP] Adam Hall network control and monitoring
    #: - [UDP] Adam Hall network control and monitoring
    adamhall = -1, 'adamhall', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Address-O-Matic
    addressbook = -1, 'addressbook', TransportProtocol.get('undefined')

    #: [TCP] Address Book Server used for contacts and calendar synchronisation
    addressbooksrv = -1, 'addressbooksrv', TransportProtocol.get('tcp')

    #: [N/A] difusi Cloud based plug & play network       synchronization protocol,
    #: content pool database discovery, and       cloudOS SAaS discovery protocol.
    adnodes = -1, 'adnodes', TransportProtocol.get('undefined')

    #: [TCP] Adobe Shadow Server
    adobe_shadow = -1, 'adobe-shadow', TransportProtocol.get('tcp')

    #: [N/A] Adobe Version Cue
    adobe_vc = -1, 'adobe-vc', TransportProtocol.get('undefined')

    #: [N/A] Automatic Disk Discovery
    adisk = -1, 'adisk', TransportProtocol.get('undefined')

    #: [N/A] ADPRO Security Device Setup
    adpro_setup = -1, 'adpro-setup', TransportProtocol.get('undefined')

    #: [N/A] Apple Application Engineering Services
    aecoretech = -1, 'aecoretech', TransportProtocol.get('undefined')

    #: [N/A] Aeroflex instrumentation and software
    aeroflex = -1, 'aeroflex', TransportProtocol.get('undefined')

    #: [TCP] Aerohive Proxy Configuration Service
    aerohive_proxy = -1, 'aerohive-proxy', TransportProtocol.get('tcp')

    #: [TCP] Affinity Designer document sharing
    affinity_docs0 = -1, 'affinity-docs0', TransportProtocol.get('tcp')

    #: [TCP] Affinity Photo document sharing
    affinity_docs1 = -1, 'affinity-docs1', TransportProtocol.get('tcp')

    #: [TCP] Affinity Publisher document sharing
    affinity_docs2 = -1, 'affinity-docs2', TransportProtocol.get('tcp')

    #: [TCP] Affinity content sharing
    affinity_cntent = -1, 'affinity-cntent', TransportProtocol.get('tcp')

    #: [TCP] Airdrop
    airdrop = -1, 'airdrop', TransportProtocol.get('tcp')

    #: - [TCP] Airmate interworking protocol
    #: - [UDP] Airmate interworking protocol
    airmate = -1, 'airmate', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Protocol for streaming of audio/video content
    #: - [UDP] Protocol for streaming of audio/video content
    airplay = -1, 'airplay', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] AirPort Base Station
    airport = -1, 'airport', TransportProtocol.get('undefined')

    #: [TCP] Coda AirPreview
    airpreview = -1, 'airpreview', TransportProtocol.get('tcp')

    #: [N/A] AirProjector
    airprojector = -1, 'airprojector', TransportProtocol.get('undefined')

    #: [N/A] Air Sharing
    airsharing = -1, 'airsharing', TransportProtocol.get('undefined')

    #: [N/A] Air Sharing Pro
    airsharingpro = -1, 'airsharingpro', TransportProtocol.get('undefined')

    #: [TCP] Alazar Technologies ATN Communication Protocol
    alazartech_atn = -1, 'alazartech-atn', TransportProtocol.get('tcp')

    #: [TCP] Aloe Gateway Protocol
    aloe_gwp = -1, 'aloe-gwp', TransportProtocol.get('tcp')

    #: [TCP] Aloe Pairing Protocol
    aloe_pp = -1, 'aloe-pp', TransportProtocol.get('tcp')

    #: [UDP] ASCOM Alpaca Device Discovery
    alpacadiscovery = -1, 'alpacadiscovery', TransportProtocol.get('udp')

    #: [UDP] Ambarella Cameras
    amba_cam = -1, 'amba-cam', TransportProtocol.get('udp')

    #: [N/A] P2PTapWar Sample Application from "iPhone SDK Development" Book
    amiphd_p2p = -1, 'amiphd-p2p', TransportProtocol.get('undefined')

    #: [N/A] Proprietary protocol for Accu-Med HTM
    ams_htm = -1, 'ams-htm', TransportProtocol.get('undefined')

    #: [N/A] Animo License Manager
    animolmd = -1, 'animolmd', TransportProtocol.get('undefined')

    #: [N/A] Animo Batch Server
    animobserver = -1, 'animobserver', TransportProtocol.get('undefined')

    #: [N/A] Anquet map synchronization between desktop and handheld devices
    anquetsync = -1, 'anquetsync', TransportProtocol.get('undefined')

    #: [TCP] ANT Galio web services
    antrmai = -1, 'antrmai', TransportProtocol.get('tcp')

    #: [TCP] remote control of Linux PCs from Android and J2ME
    anyremote = -1, 'anyremote', TransportProtocol.get('tcp')

    #: [N/A] Appelezvous
    appelezvous = -1, 'appelezvous', TransportProtocol.get('undefined')

    #: [N/A] Apple Audio Units
    apple_ausend = -1, 'apple-ausend', TransportProtocol.get('undefined')

    #: [TCP] Apple Mobile Device Protocol
    apple_mobdev = -1, 'apple-mobdev', TransportProtocol.get('tcp')

    #: [UDP] Apple MIDI
    apple_midi = -1, 'apple-midi', TransportProtocol.get('udp')

    #: [N/A] Apple Remote Debug Services (OpenGL Profiler)
    applerdbg = -1, 'applerdbg', TransportProtocol.get('undefined')

    #: [N/A] Apple TV
    appletv = -1, 'appletv', TransportProtocol.get('undefined')

    #: [N/A] Apple TV discovery of iTunes
    appletv_itunes = -1, 'appletv-itunes', TransportProtocol.get('undefined')

    #: [N/A] Apple TV Pairing
    appletv_pair = -1, 'appletv-pair', TransportProtocol.get('undefined')

    #: [N/A] AquaMon
    aquamon = -1, 'aquamon', TransportProtocol.get('undefined')

    #: [UDP] Arcturus Networks Inc. Hardware Services
    arcnet = -1, 'arcnet', TransportProtocol.get('udp')

    #: [TCP] Active Registry Network for distribution of values and streams
    arn = -1, 'arn', TransportProtocol.get('tcp')

    #: [N/A] AroundSound's information sharing protocol
    aroundsound = -1, 'aroundsound', TransportProtocol.get('undefined')

    #: [UDP] ASAM Capture Module Protocol
    asam_cmp = -1, 'asam-cmp', TransportProtocol.get('udp')

    #: [TCP] protocol for an event-based architecture for distributed control of
    #: mobile robots
    aseba = -1, 'aseba', TransportProtocol.get('tcp')

    #: [UDP] Asterisk Caller-ID Notification Service
    astnotify = -1, 'astnotify', TransportProtocol.get('udp')

    #: [N/A] Astralite
    astralite = -1, 'astralite', TransportProtocol.get('undefined')

    #: [N/A] address-o-sync
    async_ = -1, 'async', TransportProtocol.get('undefined')

    #: [UDP] AT protocol over IP
    atnet = -1, 'atnet', TransportProtocol.get('udp')

    #: [N/A] Atlassian Application (JIRA, Confluence, Fisheye, Crucible, Crowd,
    #: Bamboo) discovery service
    atlassianapp = -1, 'atlassianapp', TransportProtocol.get('undefined')

    #: - [TCP] Attero Tech Audio Device
    #: - [UDP] Attero Tech Audio Device
    attero_ad = -1, 'attero-ad', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Audirvana Remote Access Protocol
    audirvana_ap = -1, 'audirvana-ap', TransportProtocol.get('tcp')

    #: [TCP] automatic sending Image data protocol
    autosendimages = -1, 'autosendimages', TransportProtocol.get('tcp')

    #: [TCP] Access Point for AutoTargets system
    autotargets_ap = -1, 'autotargets-ap', TransportProtocol.get('tcp')

    #: [N/A] Allen Vanguard Hardware Service
    av = -1, 'av', TransportProtocol.get('undefined')

    #: [TCP] TCP SpyChat Stream Message Exchange
    av_chat_ring_01 = -1, 'av-chat-ring-01', TransportProtocol.get('tcp')

    #: [TCP] Libravatar federated avatar hosting service.
    avatars = -1, 'avatars', TransportProtocol.get('tcp')

    #: [TCP] Libravatar federated avatar hosting service.
    avatars_sec = -1, 'avatars-sec', TransportProtocol.get('tcp')

    #: - [TCP] Axis Network Video Recorders
    #: - [UDP] Axis Network Video Recorders
    axis_nvr = -1, 'axis-nvr', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Axis Video Cameras
    axis_video = -1, 'axis-video', TransportProtocol.get('undefined')

    #: [UDP] IPSEC VPN tunnel over UDP [:rfc:`6281`]
    autotunnel = -1, 'autotunnel', TransportProtocol.get('udp')

    #: [N/A] 3M Unitek Digital Orthodontic System
    b3d_convince = -1, 'b3d-convince', TransportProtocol.get('undefined')

    #: [N/A] BabyPhone
    babyphone = -1, 'babyphone', TransportProtocol.get('undefined')

    #: [TCP] Remote Control for BandHelper app
    bandhelper_rc = -1, 'bandhelper-rc', TransportProtocol.get('tcp')

    #: [TCP] Screen sharing for BandHelper app
    bandhelper_ss = -1, 'bandhelper-ss', TransportProtocol.get('tcp')

    #: [TCP] Peer to peer file sharing for a media player application
    barroomcomedy = -1, 'barroomcomedy', TransportProtocol.get('tcp')

    #: [TCP] Buddycloud Server Delegation
    bcloud_server = -1, 'bcloud-server', TransportProtocol.get('tcp')

    #: [N/A] BibDesk Sharing
    bdsk = -1, 'bdsk', TransportProtocol.get('undefined')

    #: [N/A] Beacon Remote Service
    beacon = -1, 'beacon', TransportProtocol.get('undefined')

    #: [N/A] Beamer Data Sharing Protocol
    beamer = -1, 'beamer', TransportProtocol.get('undefined')

    #: [N/A] BeatPack Synchronization Server for BeatMaker
    beatpack = -1, 'beatpack', TransportProtocol.get('undefined')

    #: [TCP] Beats Direct allows for the discovery and control of devices
    beatsdirect = -1, 'beatsdirect', TransportProtocol.get('tcp')

    #: [N/A] Xgrid Technology Preview
    beep = -1, 'beep', TransportProtocol.get('undefined')

    #: [TCP] Behnke doorphones / emergency phones
    behnke_cc = -1, 'behnke-cc', TransportProtocol.get('tcp')

    #: [TCP] Behnke audio / video intercom systems
    behnke_station = -1, 'behnke-station', TransportProtocol.get('tcp')

    #: [TCP] Behnke doorphones / video intercoms
    behnke_video = -1, 'behnke-video', TransportProtocol.get('tcp')

    #: [N/A] Bender Communication Protocol
    bender = -1, 'bender', TransportProtocol.get('undefined')

    #: [N/A] BuildForge Agent
    bfagent = -1, 'bfagent', TransportProtocol.get('undefined')

    #: - [TCP] Becker & Hickl Inter Process Communication
    #: - [UDP] Becker & Hickl Inter Process Communication
    bhipc = -1, 'bhipc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Model Railway Control via netBiDiB
    bidib = -1, 'bidib', TransportProtocol.get('tcp')

    #: [N/A] Big Bang Chess
    bigbangchess = -1, 'bigbangchess', TransportProtocol.get('undefined')

    #: [N/A] Big Bang Mancala
    bigbangmancala = -1, 'bigbangmancala', TransportProtocol.get('undefined')

    #: [TCP] Data transfer service
    bitflit = -1, 'bitflit', TransportProtocol.get('tcp')

    #: [N/A] BitTorrent Zeroconf Peer Discovery Protocol
    bittorrent = -1, 'bittorrent', TransportProtocol.get('undefined')

    #: [N/A] Little Black Book Information Exchange Protocol
    blackbook = -1, 'blackbook', TransportProtocol.get('undefined')

    #: [UDP] BlueVertise Network Protocol (BNP)
    bluevertise = -1, 'bluevertise', TransportProtocol.get('udp')

    #: [TCP] board plus application transfer protocol
    boardplus = -1, 'boardplus', TransportProtocol.get('tcp')

    #: [TCP] Booked communication protocol - Sharing And Sync Service
    booked_sync = -1, 'booked-sync', TransportProtocol.get('tcp')

    #: [N/A] Bookworm Client Discovery
    bookworm = -1, 'bookworm', TransportProtocol.get('undefined')

    #: - [TCP] Bootstrap service discovery
    #: - [UDP] Bootstrap service discovery
    bootstrap = -1, 'bootstrap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Proprietary
    boundaryscan = -1, 'boundaryscan', TransportProtocol.get('udp')

    #: [N/A] Bag Of Unusual Strategy Games
    bousg = -1, 'bousg', TransportProtocol.get('undefined')

    #: - [TCP] Bout Fitness Synchronization Service
    #: - [UDP] Bout Fitness Synchronization Service
    boutfitness = -1, 'boutfitness', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Boxray Devices Host Server
    boxraysrvr = -1, 'boxraysrvr', TransportProtocol.get('tcp')

    #: - [TCP] bq Cromo protocol
    #: - [UDP] bq Cromo protocol
    bq_cromo = -1, 'bq-cromo', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Breas
    #: - [UDP] Breas
    breas = -1, 'breas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] RFID Reader Basic Reader Interface
    bri = -1, 'bri', TransportProtocol.get('undefined')

    #: [TCP] JSON RPC Bridge Protocol
    bridgeprotocol = -1, 'bridgeprotocol', TransportProtocol.get('tcp')

    #: [TCP] The Bootstrapping Remote Secure Key Infrastructure Proxy [:rfc:`8995`]
    brski_proxy = -1, 'brski-proxy', TransportProtocol.get('tcp')

    #: [TCP] The Bootstrapping Remote Secure Key Infrastructure Registrar
    #: [:rfc:`8995`]
    brski_registrar = -1, 'brski-registrar', TransportProtocol.get('tcp')

    #: [N/A] Backup Simplicity
    bsqdea = -1, 'bsqdea', TransportProtocol.get('undefined')

    #: [TCP] Beats Transfer Protocol allows for the discovery and control of
    #: devices
    btp = -1, 'btp', TransportProtocol.get('tcp')

    #: [TCP] buddycloud API
    buddycloud_api = -1, 'buddycloud-api', TransportProtocol.get('tcp')

    #: [TCP] Service for opening electric doors
    buzzer = -1, 'buzzer', TransportProtocol.get('tcp')

    #: [TCP] Calendaring Extensions to WebDAV (CalDAV) - non-TLS [:rfc:`6764`]
    caldav = -1, 'caldav', TransportProtocol.get('tcp')

    #: [TCP] Calendaring Extensions to WebDAV (CalDAV) - over TLS [:rfc:`6764`]
    caldavs = -1, 'caldavs', TransportProtocol.get('tcp')

    #: [N/A] CalTalk
    caltalk = -1, 'caltalk', TransportProtocol.get('undefined')

    #: [TCP] Canon HTTP Management Protocol
    canon_chmp = -1, 'canon-chmp', TransportProtocol.get('tcp')

    #: [TCP] vCard Extensions to WebDAV (CardDAV) - non-TLS [:rfc:`6352`]
    carddav = -1, 'carddav', TransportProtocol.get('tcp')

    #: [TCP] vCard Extensions to WebDAV (CardDAV) - over TLS [:rfc:`6352`]
    carddavs = -1, 'carddavs', TransportProtocol.get('tcp')

    #: [N/A] Card Send Protocol
    cardsend = -1, 'cardsend', TransportProtocol.get('undefined')

    #: [TCP] Carousel Player Protocol
    carousel = -1, 'carousel', TransportProtocol.get('tcp')

    #: [N/A] IP and Closed-Circuit Television for Securitiy applications
    cctv = -1, 'cctv', TransportProtocol.get('undefined')

    #: [TCP] Control the Cerebra Aquarium Controller
    cerebra = -1, 'cerebra', TransportProtocol.get('tcp')

    #: [TCP] Server for S/MIME and PGP certificates [draft-bhjl-x509-srv-01]
    certificates = -1, 'certificates', TransportProtocol.get('tcp')

    #: [N/A] The Cheat
    cheat = -1, 'cheat', TransportProtocol.get('undefined')

    #: [N/A] Project Gridlock
    chess = -1, 'chess', TransportProtocol.get('undefined')

    #: [N/A] Fluid Theme Server
    chfts = -1, 'chfts', TransportProtocol.get('undefined')

    #: [N/A] The CHILI Radiology System
    chili = -1, 'chili', TransportProtocol.get('undefined')

    #: [TCP] Ciao Arduino Protocol
    ciao = -1, 'ciao', TransportProtocol.get('tcp')

    #: [N/A] Discovery of JDF (CIP4 Job Definition Format) enabled devices
    cip4discovery = -1, 'cip4discovery', TransportProtocol.get('undefined')

    #: [N/A] Clipboard Sharing
    clipboard = -1, 'clipboard', TransportProtocol.get('undefined')

    #: [UDP] Clique Link-Local Multicast Chat Room
    clique = -1, 'clique', TransportProtocol.get('udp')

    #: [N/A] Oracle CLS Cluster Topology Service
    clscts = -1, 'clscts', TransportProtocol.get('undefined')

    #: - [TCP] Collabio
    #: - [UDP] Collabio
    collabio = -1, 'collabio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Published Collection Object
    collection = -1, 'collection', TransportProtocol.get('undefined')

    #: [TCP] A client-server chat for LAN or Internet with video chat support
    commfort = -1, 'commfort', TransportProtocol.get('tcp')

    #: [N/A] ElectraStar media centre control protocol
    com_ocs_es_mcc = -1, 'com-ocs-es-mcc', TransportProtocol.get('undefined')

    #: [UDP] DNS SRV service for smarthome server
    conecube = -1, 'conecube', TransportProtocol.get('udp')

    #: [N/A] Now Contact
    contactserver = -1, 'contactserver', TransportProtocol.get('undefined')

    #: [TCP] game controller forwarding protocol
    controllerplus = -1, 'controllerplus', TransportProtocol.get('tcp')

    #: [TCP] 0-RTT TCP Convert Protocol [:rfc:`8803`]
    convert = -1, 'convert', TransportProtocol.get('tcp')

    #: [TCP] Coolatoola Network Audio
    coolanetaudio = -1, 'coolanetaudio', TransportProtocol.get('tcp')

    #: - [TCP] Resource Directory accessed using CoAP over TCP [:rfc:`9176`]
    #: - [UDP] Resource Directory accessed using CoAP [:rfc:`9176`]
    core_rd = -1, 'core-rd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Resource Directory accessed using CoAP over TLS [:rfc:`9176`]
    core_rd_tls = -1, 'core-rd-tls', TransportProtocol.get('tcp')

    #: [UDP] Resource Directory accessed using CoAP over DTLS [:rfc:`9176`]
    core_rd_dtls = -1, 'core-rd-dtls', TransportProtocol.get('udp')

    #: [N/A] Corroboree Server
    corroboree = -1, 'corroboree', TransportProtocol.get('undefined')

    #: - [TCP] Service for coviot branded devices
    #: - [UDP] Service for coviot branded devices
    coviot = -1, 'coviot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] NoteBook 2
    cpnotebook2 = -1, 'cpnotebook2', TransportProtocol.get('undefined')

    #: [TCP] Canfield Scientific Inc - Mirror Mobile Web Services
    csi_mmws = -1, 'csi-mmws', TransportProtocol.get('tcp')

    #: [N/A] CodeWarrior HTI Xscale PowerTAP
    cw_codetap = -1, 'cw-codetap', TransportProtocol.get('undefined')

    #: [N/A] CodeWarrior HTI DPI PowerTAP
    cw_dpitap = -1, 'cw-dpitap', TransportProtocol.get('undefined')

    #: [N/A] CodeWarrior HTI OnCE PowerTAP
    cw_oncetap = -1, 'cw-oncetap', TransportProtocol.get('undefined')

    #: [N/A] CodeWarrior HTI COP PowerTAP
    cw_powertap = -1, 'cw-powertap', TransportProtocol.get('undefined')

    #: [N/A] CyTV - Network streaming for Elgato EyeTV
    cytv = -1, 'cytv', TransportProtocol.get('undefined')

    #: [N/A] Digital Audio Control Protocol (iTunes)
    dacp = -1, 'dacp', TransportProtocol.get('undefined')

    #: [N/A] Dance partner application for iPhone
    dancepartner = -1, 'dancepartner', TransportProtocol.get('undefined')

    #: [TCP] Timer LAN service
    darkhorsetimer = -1, 'darkhorsetimer', TransportProtocol.get('tcp')

    #: [N/A] Open Source DataTurbine Streaming Data Middleware
    dataturbine = -1, 'dataturbine', TransportProtocol.get('undefined')

    #: - [TCP] d&b audiotechnik remote network
    #: - [UDP] d&b audiotechnik remote network
    dbaudio = -1, 'dbaudio', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [DCCP] ping/traceroute using DCCP
    dccp_ping = -1, 'dccp-ping', TransportProtocol.get('dccp')

    #: - [TCP] Spotlight on Oracle Diagnostic Server
    #: - [UDP] Spotlight on Oracle Diagnostic Server
    dell_soo_ds = -1, 'dell-soo-ds', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Spotlight on SQL Server Diagnostic Server HTTP
    dell_soss_ds_w = -1, 'dell-soss-ds-w', TransportProtocol.get('tcp')

    #: [TCP] Spotlight SSMS Plugin Diagnostic Server
    dell_ssms_ds = -1, 'dell-ssms-ds', TransportProtocol.get('tcp')

    #: - [TCP] Peer-to-peer real-time video streaming
    #: - [UDP] Peer-to-peer real-time video streaming
    demoncamremote = -1, 'demoncamremote', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Device Info
    device_info = -1, 'device-info', TransportProtocol.get('undefined')

    #: [TCP] DEVONthink synchronization protocol
    devonsync = -1, 'devonsync', TransportProtocol.get('tcp')

    #: - [TCP] DHANDAg is going for a site
    #: - [UDP] DHANDAg is going for a site
    #: - [SCTP] DHANDAg is going for a site
    dhanda_g = -1, 'dhanda-g', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: [N/A] Use of a dictation service by a hand-held device
    dictation = -1, 'dictation', TransportProtocol.get('undefined')

    #: [N/A] EyeHome
    difi = -1, 'difi', TransportProtocol.get('undefined')

    #: [N/A] DisConnect Peer to Peer Game Protocol
    disconnect = -1, 'disconnect', TransportProtocol.get('undefined')

    #: [TCP] Wi-Fi Alliance Display Source service
    displaysrc = -1, 'displaysrc', TransportProtocol.get('tcp')

    #: [N/A] Distributed OpenCL discovery protocol
    dist_opencl = -1, 'dist-opencl', TransportProtocol.get('undefined')

    #: [N/A] Ditrios SOA Framework Protocol
    ditrios = -1, 'ditrios', TransportProtocol.get('undefined')

    #: [N/A] Dive Log Data Sharing and Synchronization Protocol
    divelogsync = -1, 'divelogsync', TransportProtocol.get('undefined')

    #: [TCP] Delphix Session Protocol
    dlpx_sp_8415 = 8415, 'dlpx-sp', TransportProtocol.get('tcp')

    #: [TCP] Delphix Session Protocol
    dlpx_sp_null = -1, 'dlpx-sp', TransportProtocol.get('tcp')

    #: [UDP] Local Area Dynamic Time Synchronisation Protocol
    dltimesync = -1, 'dltimesync', TransportProtocol.get('udp')

    #: [TCP] DNS Long-Lived Queries over TLS [:rfc:`6281`]
    dns_llq_tls = -1, 'dns-llq-tls', TransportProtocol.get('tcp')

    #: [TCP] DNS Push Notification Service Type [:rfc:`8765, Section 6.1`]
    dns_push_tls = -1, 'dns-push-tls', TransportProtocol.get('tcp')

    #: [TCP] DNS queries to the authoritative server over TLS
    dns_query_tls = -1, 'dns-query-tls', TransportProtocol.get('tcp')

    #: [N/A] DNS Service Discovery
    dns_sd = -1, 'dns-sd', TransportProtocol.get('undefined')

    #: - [TCP] DNS Dynamic Update Service
    #: - [UDP] DNS Dynamic Update Service
    dns_update = -1, 'dns-update', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DNS Dynamic Update Service over TLS [:rfc:`6281`]
    dns_update_tls = -1, 'dns-update-tls', TransportProtocol.get('tcp')

    #: [N/A] Roar (Death of Productivity)
    dop = -1, 'dop', TransportProtocol.get('undefined')

    #: - [TCP] DOTS Signal Channel Call Home Protocol.       The service name is used
    #: to construct the SRV service names "_dots-call-home._udp"       and "_dots-
    #: call-home._tcp" for discovering Call Home DOTS clients used to establish
    #: DOTS signal channel call home. [:rfc:`8973`]
    #: - [UDP] DOTS Signal Channel Call Home Protocol.       The service name is used
    #: to construct the SRV service names "_dots-call-home._udp"       and "_dots-
    #: call-home._tcp" for discovering Call Home DOTS clients used to establish
    #: DOTS signal channel call home. [:rfc:`8973`]
    dots_call_home = -1, 'dots-call-home', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] DOTS Data Channel Protocol.       The service name is used to
    #: construct the SRV service name "_dots-data._tcp" for       discovering DOTS
    #: servers used to establish DOTS data channel. [:rfc:`8973`]
    dots_data = -1, 'dots-data', TransportProtocol.get('tcp')

    #: [N/A] DropCopy
    dropcopy = -1, 'dropcopy', TransportProtocol.get('undefined')

    #: [N/A] Datacolor SpyderGallery Desktop Sync Protocol
    dsgsync = -1, 'dsgsync', TransportProtocol.get('undefined')

    #: [N/A] Data Synchronization Protocol for Discovery Software products
    dsl_sync = -1, 'dsl-sync', TransportProtocol.get('undefined')

    #: [N/A] Desktop Transporter Remote Desktop Protocol
    dtrmtdesktop = -1, 'dtrmtdesktop', TransportProtocol.get('undefined')

    #: [TCP] A communication protocol that allows a school teacher to set work
    #: activities to students over a LAN.
    duckrace = -1, 'duckrace', TransportProtocol.get('tcp')

    #: [N/A] Documents To Go Desktop Sync Protocol
    dxtgsync = -1, 'dxtgsync', TransportProtocol.get('undefined')

    #: [N/A] Protocol for EA Downtown Texas Hold 'em
    ea_dttx_poker = -1, 'ea-dttx-poker', TransportProtocol.get('undefined')

    #: [N/A] Earphoria
    earphoria = -1, 'earphoria', TransportProtocol.get('undefined')

    #: [TCP] Sync service for the Easy Spend Log app
    easyspndlg_sync = -1, 'easyspndlg-sync', TransportProtocol.get('tcp')

    #: [N/A] Amuzi peer-to-peer session synchronization protocol
    eb_amuzi = -1, 'eb-amuzi', TransportProtocol.get('undefined')

    #: - [TCP] Easy Books App data sync helper for Mac OS X and iOS
    #: - [UDP] Easy Books App data sync helper for Mac OS X and iOS
    eb_sync = -1, 'eb-sync', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] ebXML Messaging
    ebms = -1, 'ebms', TransportProtocol.get('undefined')

    #: [N/A] Northrup Grumman/Mission Systems/ESL Data Flow Protocol
    ecms = -1, 'ecms', TransportProtocol.get('undefined')

    #: [N/A] ebXML Registry
    ebreg = -1, 'ebreg', TransportProtocol.get('undefined')

    #: [N/A] Net Monitor Anti-Piracy Service
    ecbyesfsgksc = -1, 'ecbyesfsgksc', TransportProtocol.get('undefined')

    #: [UDP] LaCie Ethernet Disk Configuration Protocol
    edcp = -1, 'edcp', TransportProtocol.get('udp')

    #: [TCP] Edge1 Base Station
    edge1 = -1, 'edge1', TransportProtocol.get('tcp')

    #: [TCP] Provides resolution of EDS services available on a given network
    edsservice = -1, 'edsservice', TransportProtocol.get('tcp')

    #: [TCP] EEG System Discovery across local and wide area networks
    eeg = -1, 'eeg', TransportProtocol.get('tcp')

    #: [TCP] EFKON Lightweight Interface to Traffic Events
    efkon_elite = -1, 'efkon-elite', TransportProtocol.get('tcp')

    #: [N/A] Egistix Auto-Discovery
    egistix = -1, 'egistix', TransportProtocol.get('undefined')

    #: [N/A] Interactive Room Software Infrastructure (Event Sharing)
    eheap = -1, 'eheap', TransportProtocol.get('undefined')

    #: [N/A] DataEnvoy
    embrace = -1, 'embrace', TransportProtocol.get('undefined')

    #: [TCP] Enphase Energy Envoy
    enphase_envoy = -1, 'enphase-envoy', TransportProtocol.get('tcp')

    #: [N/A] Endpoint Protocol (EP) for use in Home Automation systems
    ep = -1, 'ep', TransportProtocol.get('undefined')

    #: [TCP] Extensis Server Protocol
    esp = -1, 'esp', TransportProtocol.get('tcp')

    #: [TCP] Enrollment Over Secure Transport
    est = -1, 'est', TransportProtocol.get('tcp')

    #: [UDP] EST over secure CoAP (EST-coaps)
    est_coaps = -1, 'est-coaps', TransportProtocol.get('udp')

    #: [N/A] Eucalyptus Discovery
    eucalyptus = -1, 'eucalyptus', TransportProtocol.get('undefined')

    #: [N/A] Now Up-to-Date
    eventserver = -1, 'eventserver', TransportProtocol.get('undefined')

    #: [TCP] EvP - Generic EVENT protocol
    evp = -1, 'evp', TransportProtocol.get('tcp')

    #: [N/A] EVS Notification Center Protocol
    evs_notif = -1, 'evs-notif', TransportProtocol.get('undefined')

    #: [N/A] Synchronization Protocol for Ilium Software's eWallet
    ewalletsync = -1, 'ewalletsync', TransportProtocol.get('undefined')

    #: [TCP] eWeLink devices supporting LAN Control
    ewelink = -1, 'ewelink', TransportProtocol.get('tcp')

    #: [N/A] Example Service Type
    example = -1, 'example', TransportProtocol.get('undefined')

    #: [N/A] Exbiblio Cascading Service Protocol
    exb = -1, 'exb', TransportProtocol.get('undefined')

    #: [N/A] Extensis Serial Number
    extensissn = -1, 'extensissn', TransportProtocol.get('undefined')

    #: [N/A] EyeTV Sharing
    eyetvsn = -1, 'eyetvsn', TransportProtocol.get('undefined')

    #: [N/A] FaceSpan
    facespan = -1, 'facespan', TransportProtocol.get('undefined')

    #: [N/A] FAXstf
    faxstfx = -1, 'faxstfx', TransportProtocol.get('undefined')

    #: [N/A] NetNewsWire 2.0
    feed_sharing = -1, 'feed-sharing', TransportProtocol.get('undefined')

    #: [TCP] proprietary
    feldwesen = -1, 'feldwesen', TransportProtocol.get('tcp')

    #: [N/A] Firetask task sharing and synchronization protocol
    firetask = -1, 'firetask', TransportProtocol.get('undefined')

    #: [N/A] Fish
    fish = -1, 'fish', TransportProtocol.get('undefined')

    #: [TCP] Fiscalization service of Click.al
    fisk = -1, 'fisk', TransportProtocol.get('tcp')

    #: [N/A] Financial Information Exchange (FIX) Protocol
    fix = -1, 'fix', TransportProtocol.get('undefined')

    #: [N/A] Fjork
    fjork = -1, 'fjork', TransportProtocol.get('undefined')

    #: [UDP] FilmLight Cluster Power Control Service
    fl_purr = -1, 'fl-purr', TransportProtocol.get('udp')

    #: - [TCP] Flight Data Monitoring Protocol
    #: - [UDP] Flight Data Monitoring Protocol
    flightdmp = -1, 'flightdmp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] FLIR Infrared Camera
    flir_ircam = -1, 'flir-ircam', TransportProtocol.get('tcp')

    #: [N/A] FileMaker Server Administration Communication Service
    fmserver_admin = -1, 'fmserver-admin', TransportProtocol.get('undefined')

    #: [N/A] FontAgent Pro
    fontagentnode = -1, 'fontagentnode', TransportProtocol.get('undefined')

    #: [N/A] FoxTrot Search Server Discovery Service
    foxtrot_serv = -1, 'foxtrot-serv', TransportProtocol.get('undefined')

    #: [N/A] FoxTrot Professional Search Discovery Service
    foxtrot_start = -1, 'foxtrot-start', TransportProtocol.get('undefined')

    #: [N/A] FrameForge License
    frameforge_lic = -1, 'frameforge-lic', TransportProtocol.get('undefined')

    #: [N/A] FreeHand MusicPad Pro Interface Protocol
    freehand = -1, 'freehand', TransportProtocol.get('undefined')

    #: [N/A] Frog Navigation Systems
    frog = -1, 'frog', TransportProtocol.get('undefined')

    #: [N/A] Crocodile FTP Server
    ftpcroco = -1, 'ftpcroco', TransportProtocol.get('undefined')

    #: [UDP] Fairview Certificate
    fv_cert = -1, 'fv-cert', TransportProtocol.get('udp')

    #: [UDP] Fairview Key
    fv_key = -1, 'fv-key', TransportProtocol.get('udp')

    #: [UDP] Fairview Time/Date
    fv_time = -1, 'fv-time', TransportProtocol.get('udp')

    #: [TCP] Used for the Lynx System
    g2metric_lynx = -1, 'g2metric-lynx', TransportProtocol.get('tcp')

    #: [N/A] Entrackment Client Service
    garagepad = -1, 'garagepad', TransportProtocol.get('undefined')

    #: [TCP] gif collection browsing
    giffer = -1, 'giffer', TransportProtocol.get('tcp')

    #: [N/A] G-Force Control via SoundSpectrum's SSMP TCP Protocol
    gforce_ssmp = -1, 'gforce-ssmp', TransportProtocol.get('undefined')

    #: [N/A] GlassPad Data Exchange Protocol
    glasspad = -1, 'glasspad', TransportProtocol.get('undefined')

    #: [N/A] GlassPadServer Data Exchange Protocol
    glasspadserver = -1, 'glasspadserver', TransportProtocol.get('undefined')

    #: [N/A] OpenGL Driver Monitor
    glrdrvmon = -1, 'glrdrvmon', TransportProtocol.get('undefined')

    #: - [TCP] Service related to Google Expeditions which is a technology for
    #: enabling multi-participant virtual fieldtrip experiences over a local
    #: wireless network. See http://g.co/expeditions for more details
    #: - [UDP] Service related to Google Expeditions which is a technology for
    #: enabling multi-participant virtual fieldtrip experiences over a local
    #: wireless network. See http://g.co/expeditions for more details
    googexpeditions = -1, 'googexpeditions', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Service related to Google Cast which is a technology for enabling
    #: multi-screen experiences. See developers.google.com/cast for more details
    googlecast = -1, 'googlecast', TransportProtocol.get('tcp')

    #: [TCP] For iOS Application named GoOrStop
    goorstop = -1, 'goorstop', TransportProtocol.get('tcp')

    #: [UDP] GoPro proprietary protocol to wake devices
    gopro_wake = -1, 'gopro-wake', TransportProtocol.get('udp')

    #: [UDP] GoPro proprietary protocol for devices
    gopro_web = -1, 'gopro-web', TransportProtocol.get('udp')

    #: [TCP] Network name Got It!
    gotit = -1, 'gotit', TransportProtocol.get('tcp')

    #: [N/A] Grid Plug and Play
    gpnp = -1, 'gpnp', TransportProtocol.get('undefined')

    #: [N/A] Roxio ToastAnywhere(tm) Recorder Sharing
    grillezvous = -1, 'grillezvous', TransportProtocol.get('undefined')

    #: [TCP] Groovesquid Democratic Music Control Protocol
    groovesquid = -1, 'groovesquid', TransportProtocol.get('tcp')

    #: [N/A] Growl
    growl = -1, 'growl', TransportProtocol.get('undefined')

    #: [TCP] GS Labs set-top box remote control
    gsremotecontrol = -1, 'gsremotecontrol', TransportProtocol.get('tcp')

    #: [TCP] Image Data Transfer API for Wi-Fi Camera Devices over HTTP
    gucam_http = -1, 'gucam-http', TransportProtocol.get('tcp')

    #: [N/A] Special service type for resolving by GUID (Globally Unique
    #: Identifier)
    guid = -1, 'guid', TransportProtocol.get('undefined')

    #: [N/A] H.323 Real-time audio, video and data communication call setup
    #: protocol
    h323 = -1, 'h323', TransportProtocol.get('undefined')

    #: [TCP] HbbTV Application Information Table
    hbbtv_ait = -1, 'hbbtv-ait', TransportProtocol.get('tcp')

    #: [N/A] HELP command [:rfc:`1078`]
    help = -1, 'help', TransportProtocol.get('undefined')

    #: [N/A] Mercurial web-based repository access
    hg = -1, 'hg', TransportProtocol.get('undefined')

    #: [N/A] HINZMobil Synchronization protocol
    hinz = -1, 'hinz', TransportProtocol.get('undefined')

    #: [N/A] Home Media Control Protocol
    hmcp = -1, 'hmcp', TransportProtocol.get('undefined')

    #: [TCP] An application to communicate with other
    hola = -1, 'hola', TransportProtocol.get('tcp')

    #: [N/A] iTunes Home Sharing
    home_sharing = -1, 'home-sharing', TransportProtocol.get('undefined')

    #: [N/A] iDo Technology Home Automation Protocol
    homeauto = -1, 'homeauto', TransportProtocol.get('undefined')

    #: [TCP] Home Connect Protocol
    homeconnect = -1, 'homeconnect', TransportProtocol.get('tcp')

    #: [TCP] Protocol for home hub communication
    homekit = -1, 'homekit', TransportProtocol.get('tcp')

    #: [UDP] Honeywell Video Systems
    honeywell_vid = -1, 'honeywell-vid', TransportProtocol.get('udp')

    #: [N/A] Hotwayd
    hotwayd = -1, 'hotwayd', TransportProtocol.get('undefined')

    #: [N/A] Howdy messaging and notification protocol
    howdy = -1, 'howdy', TransportProtocol.get('undefined')

    #: [TCP] The service provides information about connected HP devices
    hpdeviceinfo = -1, 'hpdeviceinfo', TransportProtocol.get('tcp')

    #: [N/A] HP Remote Build System for Linux-based Systems
    hpr_bldlnx = -1, 'hpr-bldlnx', TransportProtocol.get('undefined')

    #: [N/A] HP Remote Build System for Microsoft Windows Systems
    hpr_bldwin = -1, 'hpr-bldwin', TransportProtocol.get('undefined')

    #: [N/A] Identifies systems that house databases for the Remote Build System
    #: and Remote Test System
    hpr_db = -1, 'hpr-db', TransportProtocol.get('undefined')

    #: [N/A] HP Remote Repository for Build and Test Results
    hpr_rep = -1, 'hpr-rep', TransportProtocol.get('undefined')

    #: [N/A] HP Remote System that houses compilers and tools for Linux-based
    #: Systems
    hpr_toollnx = -1, 'hpr-toollnx', TransportProtocol.get('undefined')

    #: [N/A] HP Remote System that houses compilers and tools for Microsoft Windows
    #: Systems
    hpr_toolwin = -1, 'hpr-toolwin', TransportProtocol.get('undefined')

    #: [N/A] HP Remote Test System for Linux-based Systems
    hpr_tstlnx = -1, 'hpr-tstlnx', TransportProtocol.get('undefined')

    #: [N/A] HP Remote Test System for Microsoft Windows Systems
    hpr_tstwin = -1, 'hpr-tstwin', TransportProtocol.get('undefined')

    #: [N/A] Hobbyist Software Off Discovery
    hs_off = -1, 'hs-off', TransportProtocol.get('undefined')

    #: [N/A] Home Tv Streaming Protocol
    htsp = -1, 'htsp', TransportProtocol.get('undefined')

    #: - [TCP] http protocol over TLS/SSL [:rfc:`9110`]
    #: - [UDP] http protocol over TLS/SSL [:rfc:`9110`]
    #: - [SCTP] HTTPS [:rfc:`9260`]
    https_443 = 443, 'https', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: [TCP] HTTP over SSL/TLS
    https_null = -1, 'https', TransportProtocol.get('tcp')

    #: [UDP] HomeTouch Vnc Configuration
    htvncconf = -1, 'htvncconf', TransportProtocol.get('udp')

    #: [TCP] Philips hue protocol
    hue = -1, 'hue', TransportProtocol.get('tcp')

    #: [TCP] Philips Hue Sync control protocol
    huesync = -1, 'huesync', TransportProtocol.get('tcp')

    #: [N/A] Atempo HyperStream deduplication server
    hyperstream = -1, 'hyperstream', TransportProtocol.get('undefined')

    #: [N/A] BBN IAD
    iad1 = -1, 'iad1', TransportProtocol.get('undefined')

    #: [N/A] BBN IAD
    iad2 = -1, 'iad2', TransportProtocol.get('undefined')

    #: [N/A] BBN IAD
    iad3 = -1, 'iad3', TransportProtocol.get('undefined')

    #: [TCP] Braemar Inventory audit
    iaudit = -1, 'iaudit', TransportProtocol.get('tcp')

    #: [N/A] iBiz Server
    ibiz = -1, 'ibiz', TransportProtocol.get('undefined')

    #: [N/A] Image Capture Networking
    ica_networking = -1, 'ica-networking', TransportProtocol.get('undefined')

    #: [N/A] Northrup Grumman/TASC/ICAN Protocol
    ican = -1, 'ican', TransportProtocol.get('undefined')

    #: [N/A] iChalk
    ichalkboard = -1, 'ichalkboard', TransportProtocol.get('undefined')

    #: [N/A] iChat 1.0
    ichat = -1, 'ichat', TransportProtocol.get('undefined')

    #: [N/A] ICI
    ici = -1, 'ici', TransportProtocol.get('undefined')

    #: [N/A] iConquer
    iconquer = -1, 'iconquer', TransportProtocol.get('undefined')

    #: [TCP] A Remote Control Application service used to control Computers on a
    #: Local Area Network
    icontrolbox = -1, 'icontrolbox', TransportProtocol.get('tcp')

    #: [N/A] Generic Data Acquisition and Control Protocol
    idata = -1, 'idata', TransportProtocol.get('undefined')

    #: [N/A] Intermec Device Configuration Web Services
    idcws = -1, 'idcws', TransportProtocol.get('undefined')

    #: [TCP] IDEAQUEST Safety Monitoring System
    ideaquest = -1, 'ideaquest', TransportProtocol.get('tcp')

    #: [N/A] SplashID Synchronization Service
    idsync = -1, 'idsync', TransportProtocol.get('undefined')

    #: [N/A] iFFL Bonjour service for communication between client and server
    #: applications.
    iffl = -1, 'iffl', TransportProtocol.get('undefined')

    #: [N/A] Published iFolder
    ifolder = -1, 'ifolder', TransportProtocol.get('undefined')

    #: [N/A] Idle Hands iHouse Protocol
    ihouse = -1, 'ihouse', TransportProtocol.get('undefined')

    #: [N/A] Instant Interactive Drills
    ii_drills = -1, 'ii-drills', TransportProtocol.get('undefined')

    #: [N/A] Instant Interactive Konane
    ii_konane = -1, 'ii-konane', TransportProtocol.get('undefined')

    #: [N/A] iLynX
    ilynx = -1, 'ilynx', TransportProtocol.get('undefined')

    #: - [TCP] Instant Messaging [:rfc:`3861`]
    #: - [UDP] Instant Messaging [:rfc:`3861`]
    im = -1, 'im', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] ImagesCal App Data Sharing
    imagescal = -1, 'imagescal', TransportProtocol.get('tcp')

    #: [N/A] iMidi
    imidi = -1, 'imidi', TransportProtocol.get('undefined')

    #: [TCP] Protocol synchronizing Image data
    imgsync = -1, 'imgsync', TransportProtocol.get('tcp')

    #: [TCP] Noise Inspector
    ims_ni = -1, 'ims-ni', TransportProtocol.get('tcp')

    #: [N/A] Indigo Security Digital Video Recorders
    indigo_dvr = -1, 'indigo-dvr', TransportProtocol.get('undefined')

    #: [TCP] InfBoard interactive whiteboard protocol
    infboard = -1, 'infboard', TransportProtocol.get('tcp')

    #: [TCP] Listing Singlewire InformaCast servers
    informacast = -1, 'informacast', TransportProtocol.get('tcp')

    #: [TCP] Service for a Communications App
    innergroup = -1, 'innergroup', TransportProtocol.get('tcp')

    #: [N/A] Inova Solutions OnTrack Display Protocol
    inova_ontrack = -1, 'inova-ontrack', TransportProtocol.get('undefined')

    #: [TCP] Internet-of-things (IoT) JSON telegram service
    iot = -1, 'iot', TransportProtocol.get('tcp')

    #: [TCP] iotaMed medical records server
    iota = -1, 'iota', TransportProtocol.get('tcp')

    #: [N/A] IP Broadcaster
    ipbroadcaster = -1, 'ipbroadcaster', TransportProtocol.get('undefined')

    #: - [TCP] Network socket performance test
    #: - [UDP] Network socket performance test
    iperfd = -1, 'iperfd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] IP Speaker Control Protocol
    ipspeaker = -1, 'ipspeaker', TransportProtocol.get('undefined')

    #: [TCP] Control protocol for Phase One devices
    iqp = -1, 'iqp', TransportProtocol.get('tcp')

    #: [TCP] The universal protocol for building IoT
    irc_iot = -1, 'irc-iot', TransportProtocol.get('tcp')

    #: [TCP] HVAC SMIL Server
    ir_hvac_000 = -1, 'ir-hvac-000', TransportProtocol.get('tcp')

    #: [N/A] iRelay application discovery service
    irelay = -1, 'irelay', TransportProtocol.get('undefined')

    #: [N/A] Intego Remote Management Console
    irmc = -1, 'irmc', TransportProtocol.get('undefined')

    #: - [TCP] iRobot Monitor and Control Service
    #: - [UDP] iRobot Monitor and Control Service
    irobotmcs = -1, 'irobotmcs', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Iroid Data Service Discovery
    iroid_sd = -1, 'iroid-sd', TransportProtocol.get('tcp')

    #: [TCP] iClip clipboard transfer
    irradiatd_iclip = -1, 'irradiatd-iclip', TransportProtocol.get('tcp')

    #: [TCP] Image Resizer Toolbox preview sharing service
    irt_sharing = -1, 'irt-sharing', TransportProtocol.get('tcp')

    #: [N/A] iSparx
    isparx = -1, 'isparx', TransportProtocol.get('undefined')

    #: [N/A] iSpQ VideoChat
    ispq_vc = -1, 'ispq-vc', TransportProtocol.get('undefined')

    #: [N/A] iShare
    ishare = -1, 'ishare', TransportProtocol.get('undefined')

    #: [N/A] iSticky
    isticky = -1, 'isticky', TransportProtocol.get('undefined')

    #: [N/A] iStorm
    istorm = -1, 'istorm', TransportProtocol.get('undefined')

    #: [TCP] iSynchronize data synchronization protocol
    isynchronize = -1, 'isynchronize', TransportProtocol.get('tcp')

    #: [TCP] iTap Publishing Service
    itap_publish = -1, 'itap-publish', TransportProtocol.get('tcp')

    #: [N/A] IT-IS International Ltd. Device
    itis_device = -1, 'itis-device', TransportProtocol.get('undefined')

    #: [N/A] iTunes Socket Remote Control
    itsrc = -1, 'itsrc', TransportProtocol.get('undefined')

    #: [N/A] Inter VTS Exchange Format
    ivef = -1, 'ivef', TransportProtocol.get('undefined')

    #: [N/A] iWork Server
    iwork = -1, 'iwork', TransportProtocol.get('undefined')

    #: [TCP] Integrated Business Data Exchange
    izira = -1, 'izira', TransportProtocol.get('tcp')

    #: [N/A] Northrup Grumman/TASC/JCAN Protocol
    jcan = -1, 'jcan', TransportProtocol.get('undefined')

    #: [N/A] Jedit X
    jeditx = -1, 'jeditx', TransportProtocol.get('undefined')

    #: [N/A] Jini Service Discovery
    jini = -1, 'jini', TransportProtocol.get('undefined')

    #: [TCP] JSON Meta Application Protocol [:rfc:`8620`]
    jmap = -1, 'jmap', TransportProtocol.get('tcp')

    #: [TCP] jollys keychain cloud sync protocol
    jnx_kcsync = -1, 'jnx-kcsync', TransportProtocol.get('tcp')

    #: [N/A] Proprietary
    jtag = -1, 'jtag', TransportProtocol.get('undefined')

    #: [TCP] Jukebox Request Service
    jukebox = -1, 'jukebox', TransportProtocol.get('tcp')

    #: [TCP] Music sharing protocol
    jukejoint = -1, 'jukejoint', TransportProtocol.get('tcp')

    #: [TCP] KeynoteAccess is used for sending remote requests/responses when
    #: controlling a slideshow with Keynote Remote
    keynoteaccess = -1, 'keynoteaccess', TransportProtocol.get('tcp')

    #: [TCP] KeynotePairing is used to pair Keynote Remote with Keynote
    keynotepairing = -1, 'keynotepairing', TransportProtocol.get('tcp')

    #: [TCP] Topology Discovery
    kiwin = -1, 'kiwin', TransportProtocol.get('tcp')

    #: [UDP] Discovery in KNX IoT Point API
    knx = -1, 'knx', TransportProtocol.get('udp')

    #: [N/A] Kabira Transaction Platform
    ktp = -1, 'ktp', TransportProtocol.get('undefined')

    #: [TCP] Symbolic Sound Kyma Service Discovery
    kyma = -1, 'kyma', TransportProtocol.get('tcp')

    #: [N/A] IMP Logical Address Maintenance
    la_maint = -1, 'la-maint', TransportProtocol.get('undefined')

    #: [UDP] Labyrinth local multiplayer protocol
    labyrinth = -1, 'labyrinth', TransportProtocol.get('udp')

    #: [N/A] Lan2P Peer-to-Peer Network Protocol
    lan2p = -1, 'lan2p', TransportProtocol.get('undefined')

    #: [N/A] Gawker
    lapse = -1, 'lapse', TransportProtocol.get('undefined')

    #: - [TCP] Larva IP Controller
    #: - [UDP] Larva IP Controller
    larvaio_control = -1, 'larvaio-control', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Lua Embedded Application Framework
    leaf = -1, 'leaf', TransportProtocol.get('undefined')

    #: [N/A] Lexicon Vocabulary Sharing
    lexicon = -1, 'lexicon', TransportProtocol.get('undefined')

    #: [N/A] Liaison
    liaison = -1, 'liaison', TransportProtocol.get('undefined')

    #: [N/A] Delicious Library 2 Collection Data Sharing Protocol
    library = -1, 'library', TransportProtocol.get('undefined')

    #: [N/A] Protocol for setup and control of Libratone products
    libratone = -1, 'libratone', TransportProtocol.get('undefined')

    #: [N/A] LI-COR Biosciences instrument discovery
    licor = -1, 'licor', TransportProtocol.get('undefined')

    #: [N/A] RFID reader Low Level Reader Protocol over SSL/TLS
    llrp_secure = -1, 'llrp-secure', TransportProtocol.get('undefined')

    #: [N/A] Gobby
    lobby = -1, 'lobby', TransportProtocol.get('undefined')

    #: [UDP] Logic Pro Distributed Audio
    logicnode = -1, 'logicnode', TransportProtocol.get('udp')

    #: [TCP] logSheet Server Synchronization Server
    logsheetserver = -1, 'logsheetserver', TransportProtocol.get('tcp')

    #: [N/A] Echelon LonBridge Server
    lonbridge = -1, 'lonbridge', TransportProtocol.get('undefined')

    #: [N/A] LonTalk over IP (ANSI 852)
    lontalk = -1, 'lontalk', TransportProtocol.get('undefined')

    #: - [TCP] LonWorks
    #: - [UDP] LonWorks
    lonworks_2540 = 2540, 'lonworks', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Echelon LNS Remote Client
    lonworks_null = -1, 'lonworks', TransportProtocol.get('undefined')

    #: [N/A] Linksys One Application Server API
    lsys_appserver = -1, 'lsys-appserver', TransportProtocol.get('undefined')

    #: [N/A] Linksys One Camera API
    lsys_camera = -1, 'lsys-camera', TransportProtocol.get('undefined')

    #: [N/A] LinkSys EZ Configuration
    lsys_ezcfg = -1, 'lsys-ezcfg', TransportProtocol.get('undefined')

    #: [N/A] LinkSys Operations, Administration, Management, and Provisioning
    lsys_oamp = -1, 'lsys-oamp', TransportProtocol.get('undefined')

    #: [TCP] A protocol to remotely control DMX512 devices over the network
    lumiere = -1, 'lumiere', TransportProtocol.get('tcp')

    #: [TCP] Lumis Cache Appliance Protocol
    lumis_lca = -1, 'lumis-lca', TransportProtocol.get('tcp')

    #: [N/A] Lux Solis Data Transport Protocol
    lux_dtp = -1, 'lux-dtp', TransportProtocol.get('undefined')

    #: [N/A] LXI
    lxi = -1, 'lxi', TransportProtocol.get('undefined')

    #: [N/A] iPod Lyrics Service
    lyrics = -1, 'lyrics', TransportProtocol.get('undefined')

    #: - [TCP] Moritz30-Project Standard protocol 1 Plain Text
    #: - [UDP] Moritz30-Project Standard protocol 1 Plain Text
    m30s1pt = -1, 'm30s1pt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] MacFOH
    macfoh = -1, 'macfoh', TransportProtocol.get('undefined')

    #: [N/A] MacFOH admin services
    macfoh_admin = -1, 'macfoh-admin', TransportProtocol.get('undefined')

    #: [UDP] MacFOH audio stream
    macfoh_audio = -1, 'macfoh-audio', TransportProtocol.get('udp')

    #: [UDP] MacFOH show control events
    macfoh_events = -1, 'macfoh-events', TransportProtocol.get('udp')

    #: [UDP] MacFOH realtime data
    macfoh_data = -1, 'macfoh-data', TransportProtocol.get('udp')

    #: [N/A] MacFOH database
    macfoh_db = -1, 'macfoh-db', TransportProtocol.get('undefined')

    #: [N/A] MacFOH Remote
    macfoh_remote = -1, 'macfoh-remote', TransportProtocol.get('undefined')

    #: [N/A] Mac Minder
    macminder = -1, 'macminder', TransportProtocol.get('undefined')

    #: [N/A] Maestro Music Sharing Service
    maestro = -1, 'maestro', TransportProtocol.get('undefined')

    #: [N/A] Magic Dice Game Protocol
    magicdice = -1, 'magicdice', TransportProtocol.get('undefined')

    #: [N/A] Mandos Password Server
    mandos = -1, 'mandos', TransportProtocol.get('undefined')

    #: - [TCP] Pravala Mobility and Aggregation Service
    #: - [UDP] Pravala Mobility and Aggregation Service
    mas = -1, 'mas', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] MATRIX Remote AV Switching
    matrix = -1, 'matrix', TransportProtocol.get('undefined')

    #: - [TCP] Matter Commissionable Node Discovery
    #: - [UDP] Matter Commissionable Node Discovery
    matterc = -1, 'matterc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Matter Commissioner Discovery
    #: - [UDP] Matter Commissioner Discovery
    matterd = -1, 'matterd', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] MAVLink Micro Air Vehicle Communication Protocol
    mavlink = -1, 'mavlink', TransportProtocol.get('tcp')

    #: [UDP] Peer to peer communication between instances of the Maze Pseudo game.
    mazepseudo_game = -1, 'mazepseudo-game', TransportProtocol.get('udp')

    #: [N/A] MediaBroker++ Consumer
    mbconsumer = -1, 'mbconsumer', TransportProtocol.get('undefined')

    #: [N/A] MediaBroker++ Producer
    mbproducer = -1, 'mbproducer', TransportProtocol.get('undefined')

    #: [N/A] MediaBroker++ Server
    mbserver = -1, 'mbserver', TransportProtocol.get('undefined')

    #: [N/A] ClairMail Connect
    mconnect = -1, 'mconnect', TransportProtocol.get('undefined')

    #: [N/A] MediaCentral
    mcrcp = -1, 'mcrcp', TransportProtocol.get('undefined')

    #: [N/A] MediaBoardONE Asset and Information Manager data sharing and
    #: synchronization protocol
    mediaboard1 = -1, 'mediaboard1', TransportProtocol.get('undefined')

    #: [TCP] Mediatap streaming protocol
    mediatap = -1, 'mediatap', TransportProtocol.get('tcp')

    #: [TCP] Mercury Dock Assistant
    mercurydock = -1, 'mercurydock', TransportProtocol.get('tcp')

    #: [N/A] Mes Amis
    mesamis = -1, 'mesamis', TransportProtocol.get('undefined')

    #: [UDP] Thread Mesh Commissioning
    meshcop = -1, 'meshcop', TransportProtocol.get('udp')

    #: [N/A] Mental Ray for Maya
    mi_raysat = -1, 'mi-raysat', TransportProtocol.get('undefined')

    #: [TCP] Protocol for connected accessories
    mieleacs = -1, 'mieleacs', TransportProtocol.get('tcp')

    #: [TCP] Miele@home Protocol
    mieleathome = -1, 'mieleathome', TransportProtocol.get('tcp')

    #: [TCP] Miele protocol for prof. appliances
    mieleprof = -1, 'mieleprof', TransportProtocol.get('tcp')

    #: [TCP] Miele protocol for semi prof. appliances
    mielesemiprof = -1, 'mielesemiprof', TransportProtocol.get('tcp')

    #: [TCP] A Medical Interoperability Service Type, used to identify medical
    #: devices providing network interfaces.
    mist = -1, 'mist', TransportProtocol.get('tcp')

    #: [TCP] Provides a client with access to the Mathematical Mesh, a user-focused
    #: PKI.
    mmm = -1, 'mmm', TransportProtocol.get('tcp')

    #: - [TCP] A Remote Control Application service used to control Computers on a
    #: Local Area Network
    #: - [UDP] A Remote Control Application service used to control Computers on a
    #: Local Area Network
    mn_passage = -1, 'mn-passage', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] modo LAN Services
    modolansrv = -1, 'modolansrv', TransportProtocol.get('undefined')

    #: [TCP] Authentication service for Mogeneti Software Applications
    mogeneti_auth = -1, 'mogeneti-auth', TransportProtocol.get('tcp')

    #: [UDP] Sonnox MCON monitor controller protocol
    moncon = -1, 'moncon', TransportProtocol.get('udp')

    #: [N/A] SplashMoney Synchronization Service
    moneysync = -1, 'moneysync', TransportProtocol.get('undefined')

    #: [N/A] MoneyWorks Gold and MoneyWorks Datacentre network service
    moneyworks = -1, 'moneyworks', TransportProtocol.get('undefined')

    #: [N/A] Bonjour Mood Ring tutorial program
    moodring = -1, 'moodring', TransportProtocol.get('undefined')

    #: [N/A] Mother script server protocol
    mother = -1, 'mother', TransportProtocol.get('undefined')

    #: [N/A] MovieSlate digital clapperboard
    movieslate = -1, 'movieslate', TransportProtocol.get('undefined')

    #: [N/A] MP3 Sushi
    mp3sushi = -1, 'mp3sushi', TransportProtocol.get('undefined')

    #: - [TCP] Message Queuing Telemetry Transport Protocol
    #: - [UDP] Message Queuing Telemetry Transport Protocol
    mqtt_1883 = 1883, 'mqtt', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] IBM MQ Telemetry Transport Broker
    mqtt_null = -1, 'mqtt', TransportProtocol.get('undefined')

    #: [N/A] Martian SlingShot
    mslingshot = -1, 'mslingshot', TransportProtocol.get('undefined')

    #: [TCP] MSRP protocol over TLS [:rfc:`4976`]
    msrps = -1, 'msrps', TransportProtocol.get('tcp')

    #: [N/A] Mumble VoIP communication protocol
    mumble = -1, 'mumble', TransportProtocol.get('undefined')

    #: [N/A] Protocol for a distributed music playing service
    musicmachine = -1, 'musicmachine', TransportProtocol.get('undefined')

    #: [N/A] MySync Protocol
    mysync = -1, 'mysync', TransportProtocol.get('undefined')

    #: [N/A] MenuTunes Sharing
    mttp = -1, 'mttp', TransportProtocol.get('undefined')

    #: [N/A] Maxim Integrated Products Automated Roadtest Mk II
    mxim_art2 = -1, 'mxim-art2', TransportProtocol.get('undefined')

    #: [N/A] Maxim Integrated Products In-circuit Emulator
    mxim_ice = -1, 'mxim-ice', TransportProtocol.get('undefined')

    #: [N/A] MatrixStore
    mxs = -1, 'mxs', TransportProtocol.get('undefined')

    #: [TCP] NetAcquire server input/output protocol
    naio = -1, 'naio', TransportProtocol.get('tcp')

    #: - [TCP] Proprietary communication protocol for NAS Monitor
    #: - [UDP] Proprietary communication protocol for NAS Monitor
    nasmon = -1, 'nasmon', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] This DNS-SD service is used by mobile clients to locate the Nasuni
    #: Filer (a storage product) for a given company.
    nasunifiler = -1, 'nasunifiler', TransportProtocol.get('tcp')

    #: [N/A] Network Clipboard Broadcasts
    ncbroadcast = -1, 'ncbroadcast', TransportProtocol.get('undefined')

    #: [N/A] Network Clipboard Direct Transfers
    ncdirect = -1, 'ncdirect', TransportProtocol.get('undefined')

    #: [TCP] The issuer service in the n-Count electronic value transfer system
    ncount_issuer = -1, 'ncount-issuer', TransportProtocol.get('tcp')

    #: [N/A] Network Clipboard Sync Server
    ncsyncserver = -1, 'ncsyncserver', TransportProtocol.get('undefined')

    #: - [TCP] IP based video discovery and usage
    #: - [UDP] IP based video discovery and usage
    ndi = -1, 'ndi', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Nedap AEOS processing unit
    nedap_aepu = -1, 'nedap-aepu', TransportProtocol.get('tcp')

    #: [UDP] NeoRiders Client Discovery Protocol
    neoriders = -1, 'neoriders', TransportProtocol.get('udp')

    #: [TCP] UpdateLogic NetReady Device Identification
    netready = -1, 'netready', TransportProtocol.get('tcp')

    #: [N/A] NetRestore
    netrestore = -1, 'netrestore', TransportProtocol.get('undefined')

    #: [TCP] AD Group NetVu Connected Video
    netvu_video = -1, 'netvu-video', TransportProtocol.get('tcp')

    #: - [TCP] Proprietary communication protocol for NextCap capture solution
    #: - [UDP] Proprietary communication protocol for NextCap capture solution
    nextcap = -1, 'nextcap', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] NFS service for the domain root, the root of an organization's
    #: published file namespace. [:rfc:`6641`]
    nfs_domainroot = -1, 'nfs-domainroot', TransportProtocol.get('tcp')

    #: [TCP] NGR Key Distribution
    ngr_keydist = -1, 'ngr-keydist', TransportProtocol.get('tcp')

    #: [TCP] National Instruments Network Device
    ni = -1, 'ni', TransportProtocol.get('tcp')

    #: - [TCP] NI FTP
    #: - [UDP] NI FTP
    ni_ftp = -1, 'ni-ftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] NI MAIL
    #: - [UDP] NI MAIL
    ni_mail = -1, 'ni-mail', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] National Instruments Real-Time Target
    ni_rt = -1, 'ni-rt', TransportProtocol.get('tcp')

    #: [TCP] National Instruments System API Service
    ni_sysapi = -1, 'ni-sysapi', TransportProtocol.get('tcp')

    #: [TCP] Lightweight event based control protocol utilising JavaScript Object
    #: Notation
    nodel = -1, 'nodel', TransportProtocol.get('tcp')

    #: [TCP] Network Quality test server endpoint
    nq = -1, 'nq', TransportProtocol.get('tcp')

    #: [N/A] American Dynamics Intellex Archive Management Service
    ntlx_arch = -1, 'ntlx-arch', TransportProtocol.get('undefined')

    #: [N/A] American Dynamics Intellex Enterprise Management Service
    ntlx_ent = -1, 'ntlx-ent', TransportProtocol.get('undefined')

    #: [N/A] American Dynamics Intellex Video Service
    ntlx_video = -1, 'ntlx-video', TransportProtocol.get('undefined')

    #: [UDP] Tenasys
    ntx = -1, 'ntx', TransportProtocol.get('udp')

    #: [TCP] Nuance Unity Service Discovery Protocol
    nusdp = -1, 'nusdp', TransportProtocol.get('tcp')

    #: - [TCP] Oak Device Services
    #: - [UDP] Oak Device Services
    oak = -1, 'oak', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Observations Framework
    obf = -1, 'obf', TransportProtocol.get('undefined')

    #: [N/A] Means for clients to locate servers in an Objective
    #: (http://www.objective.com) instance.
    objective = -1, 'objective', TransportProtocol.get('undefined')

    #: - [TCP] Insecure OCP.1 protocol, which is the insecure TCP/IP implementation
    #: of the Object Control Architecture
    #: - [UDP] Insecure OCP.1 protocol, which is the insecure TCP/IP implementation
    #: of the Object Control Architecture
    oca = -1, 'oca', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Secure OCP.1 protocol, which is the secure TCP/IP implementation of
    #: the Object Control Architecture
    #: - [UDP] Secure OCP.1 protocol, which is the secure TCP/IP implementation of
    #: the Object Control Architecture
    ocasec = -1, 'ocasec', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Insecure OCP.1 hosted on a WebSocket
    ocaws = -1, 'ocaws', TransportProtocol.get('tcp')

    #: [N/A] Oce Common Exchange Protocol
    oce = -1, 'oce', TransportProtocol.get('undefined')

    #: [N/A] OpenDirectory Master
    od_master = -1, 'od-master', TransportProtocol.get('undefined')

    #: [N/A] OD4Contact
    odabsharing = -1, 'odabsharing', TransportProtocol.get('undefined')

    #: [N/A] Optical Disk Sharing
    odisk = -1, 'odisk', TransportProtocol.get('undefined')

    #: [N/A] OfficeTime Synchronization Protocol
    officetime_sync = -1, 'officetime-sync', TransportProtocol.get('undefined')

    #: [N/A] OmniFocus setting configuration
    ofocus_conf = -1, 'ofocus-conf', TransportProtocol.get('undefined')

    #: [N/A] OmniFocus document synchronization
    ofocus_sync = -1, 'ofocus-sync', TransportProtocol.get('undefined')

    #: [TCP] Web Interface for the Open Lighting Architecture Software
    ola = -1, 'ola', TransportProtocol.get('tcp')

    #: [UDP] One Laptop per Child activity
    olpc_activity1 = -1, 'olpc-activity1', TransportProtocol.get('udp')

    #: [N/A] OMA BCAST Service Guide Discovery Service
    oma_bcast_sg = -1, 'oma-bcast-sg', TransportProtocol.get('undefined')

    #: [TCP] Open Mobile Alliance (OMA) Device Management (DM) Bootstrap Server
    #: Discovery Service
    omadm_bootstrap = -1, 'omadm-bootstrap', TransportProtocol.get('tcp')

    #: [N/A] OmniWeb
    omni_bookmark = -1, 'omni-bookmark', TransportProtocol.get('undefined')

    #: [N/A] Service for remote control of Omnisphere virtual instrument
    omni_live = -1, 'omni-live', TransportProtocol.get('undefined')

    #: [TCP] Resource state update notifications
    omnistate = -1, 'omnistate', TransportProtocol.get('tcp')

    #: [TCP] OneNet Application Information Service
    onenet_info = -1, 'onenet-info', TransportProtocol.get('tcp')

    #: [TCP] OneNet Pairing
    onenet_pairing = -1, 'onenet-pairing', TransportProtocol.get('tcp')

    #: [UDP] OneNet PGN Transport Service
    onenet_pgn = -1, 'onenet-pgn', TransportProtocol.get('udp')

    #: [N/A] OpenBase SQL
    openbase = -1, 'openbase', TransportProtocol.get('undefined')

    #: [UDP] Conferencing Protocol
    opencu = -1, 'opencu', TransportProtocol.get('udp')

    #: [TCP] Discovery of OpenPath appliances
    openpath = -1, 'openpath', TransportProtocol.get('tcp')

    #: [N/A] oprofile server protocol
    oprofile = -1, 'oprofile', TransportProtocol.get('undefined')

    #: [UDP] Open Sound Control Interface Transfer
    oscit = -1, 'oscit', TransportProtocol.get('udp')

    #: [N/A] ObjectVideo OV Ready Protocol
    ovready = -1, 'ovready', TransportProtocol.get('undefined')

    #: [N/A] OWFS (1-wire file system) web server
    owhttpd = -1, 'owhttpd', TransportProtocol.get('undefined')

    #: [N/A] Remote Parental Controls
    parentcontrol = -1, 'parentcontrol', TransportProtocol.get('undefined')

    #: [N/A] PasswordWallet Data Synchronization Protocol
    passwordwallet = -1, 'passwordwallet', TransportProtocol.get('undefined')

    #: [N/A] Mac OS X Podcast Producer Server
    pcast = -1, 'pcast', TransportProtocol.get('undefined')

    #: [UDP] Peer-to-Peer Chat (Sample Java Bonjour application)
    p2pchat = -1, 'p2pchat', TransportProtocol.get('udp')

    #: [TCP] DataBOND p2p storage
    p2pstorage_sec = -1, 'p2pstorage-sec', TransportProtocol.get('tcp')

    #: [TCP] Pair & Share data protocol
    pairandshare = -1, 'pairandshare', TransportProtocol.get('tcp')

    #: [TCP] Panoply multimedia composite transfer protocol
    panoply = -1, 'panoply', TransportProtocol.get('tcp')

    #: [TCP] Parabay P2P protocol
    parabay_p2p = -1, 'parabay-p2p', TransportProtocol.get('tcp')

    #: - [TCP] PA-R-I-Ty (Public Address - Radio - Intercom - Telefony)
    #: - [UDP] PA-R-I-Ty (Public Address - Radio - Intercom - Telefony)
    parity = -1, 'parity', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Local and remote file transfers
    #: - [UDP] Local and remote file transfers
    payload_app = -1, 'payload-app', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Horowitz Key Protocol (HKP)
    pgpkey_hkp = -1, 'pgpkey-hkp', TransportProtocol.get('undefined')

    #: [N/A] PGP Keyserver using HTTP/1.1
    pgpkey_http = -1, 'pgpkey-http', TransportProtocol.get('undefined')

    #: [N/A] PGP Keyserver using HTTPS
    pgpkey_https = -1, 'pgpkey-https', TransportProtocol.get('undefined')

    #: [N/A] PGP Keyserver using LDAP
    pgpkey_ldap = -1, 'pgpkey-ldap', TransportProtocol.get('undefined')

    #: [N/A] PGP Key submission using SMTP
    pgpkey_mailto = -1, 'pgpkey-mailto', TransportProtocol.get('undefined')

    #: [N/A] Photo Parata Event Photography Software
    photoparata = -1, 'photoparata', TransportProtocol.get('undefined')

    #: [TCP] Show Photos over TCP
    photoshow = -1, 'photoshow', TransportProtocol.get('tcp')

    #: [TCP] Photosmith's iPad to Lightroom sync protocol
    photosmithsync = -1, 'photosmithsync', TransportProtocol.get('tcp')

    #: [N/A] Pictua Intercommunication Protocol
    pictua = -1, 'pictua', TransportProtocol.get('undefined')

    #: [N/A] pieSync Computer to Computer Synchronization
    piesync = -1, 'piesync', TransportProtocol.get('undefined')

    #: [TCP] Pi Pedal Web Interface
    pipedal = -1, 'pipedal', TransportProtocol.get('tcp')

    #: [N/A] Pedestal Interface Unit by RPM-PSI
    piu = -1, 'piu', TransportProtocol.get('undefined')

    #: [N/A] Public Key Infrastructure Repository Locator Service [:rfc:`4386`]
    pkixrep = -1, 'pkixrep', TransportProtocol.get('undefined')

    #: [N/A] Parallel OperatiOn and Control Heuristic (Pooch)
    poch = -1, 'poch', TransportProtocol.get('undefined')

    #: [TCP] A presenter to audience transfer service
    pochi = -1, 'pochi', TransportProtocol.get('tcp')

    #: [TCP] Protocol for communication between Podcast
    podcastproxy = -1, 'podcastproxy', TransportProtocol.get('tcp')

    #: [N/A] Communication channel for "Poke Eye" Elgato EyeTV remote controller
    pokeeye = -1, 'pokeeye', TransportProtocol.get('undefined')

    #: [N/A] PowerEasy ERP
    powereasy_erp = -1, 'powereasy-erp', TransportProtocol.get('undefined')

    #: [N/A] PowerEasy Point of Sale
    powereasy_pos = -1, 'powereasy-pos', TransportProtocol.get('undefined')

    #: [N/A] Piano Player Remote Control
    pplayer_ctrl = -1, 'pplayer-ctrl', TransportProtocol.get('undefined')

    #: - [TCP] Presence [:rfc:`3861`]
    #: - [UDP] Presence [:rfc:`3861`]
    pres = -1, 'pres', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] XMPP Link-Local Messaging
    #: - [UDP] XMPP Link-Local Messaging
    presence_5298 = 5298, 'presence', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Peer-to-peer messaging / Link-Local Messaging
    presence_null = -1, 'presence', TransportProtocol.get('undefined')

    #: [N/A] Retrieve a description of a device's print capabilities
    print_caps = -1, 'print-caps', TransportProtocol.get('undefined')

    #: [TCP] Printopia Serve
    printopia_10631 = 10631, 'printopia', TransportProtocol.get('tcp')

    #: [TCP] Printopia Server
    printopia_null = -1, 'printopia', TransportProtocol.get('tcp')

    #: - [TCP] Profile for Mac
    #: - [UDP] Profile for Mac
    profilemac_4749 = 4749, 'profilemac', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Profile for Mac medical practice management software
    profilemac_null = -1, 'profilemac', TransportProtocol.get('undefined')

    #: [N/A] Prolog
    prolog = -1, 'prolog', TransportProtocol.get('undefined')

    #: [N/A] Protonet node and service discovery protocol
    protonet = -1, 'protonet', TransportProtocol.get('undefined')

    #: [UDP] Progal Service Advertising Protocol
    psap = -1, 'psap', TransportProtocol.get('udp')

    #: [N/A] Physical Security Interoperability Alliance Protocol
    psia = -1, 'psia', TransportProtocol.get('undefined')

    #: [TCP] File synchronization protocol for Pst Mail Sync
    pstmailsync = -1, 'pstmailsync', TransportProtocol.get('tcp')

    #: [TCP] Secured file synchronization protocol for Pst Mail Sync
    pstmailsync_ssl = -1, 'pstmailsync-ssl', TransportProtocol.get('tcp')

    #: [N/A] PTNetPro Service
    ptnetprosrv2 = -1, 'ptnetprosrv2', TransportProtocol.get('undefined')

    #: [TCP] Picture Transfer Protocol(PTP) Initiator
    ptp_init = -1, 'ptp-init', TransportProtocol.get('tcp')

    #: [N/A] PTP Initiation Request Protocol
    ptp_req = -1, 'ptp-req', TransportProtocol.get('undefined')

    #: [N/A] Protocol used for puzzle games
    puzzle = -1, 'puzzle', TransportProtocol.get('undefined')

    #: [TCP] Experimental Physics and Industrial Control System
    pvaccess_5075 = 5075, 'pvaccess', TransportProtocol.get('tcp')

    #: [TCP] Experimental Physics and Industrial Control System
    pvaccess_null = -1, 'pvaccess', TransportProtocol.get('tcp')

    #: [N/A] QBox Appliance Locator
    qbox = -1, 'qbox', TransportProtocol.get('undefined')

    #: [N/A] QuickTime Transfer Protocol
    qttp = -1, 'qttp', TransportProtocol.get('undefined')

    #: [TCP] Distributed Game Data
    quad = -1, 'quad', TransportProtocol.get('tcp')

    #: [N/A] Quinn Game Server
    quinn = -1, 'quinn', TransportProtocol.get('undefined')

    #: [TCP] Broadcast of Qwizdom Presentation sessions for joining by a client,
    #: such as Qwizdom Notes+.
    qwizcollab = -1, 'qwizcollab', TransportProtocol.get('tcp')

    #: [N/A] Rakket Client Protocol
    rakket = -1, 'rakket', TransportProtocol.get('undefined')

    #: [N/A] RadioTAG: Event tagging for radio services
    radiotag = -1, 'radiotag', TransportProtocol.get('undefined')

    #: [N/A] RadioVIS: Visualisation for radio services
    radiovis = -1, 'radiovis', TransportProtocol.get('undefined')

    #: [N/A] RadioEPG: Electronic Programme Guide for radio services
    radioepg = -1, 'radioepg', TransportProtocol.get('undefined')

    #: - [TCP] RadioPort Message Service
    #: - [UDP] RadioPort Message Service
    radioport = -1, 'radioport', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [UDP] Authentication, Accounting, and Dynamic Authorization via the RADIUS
    #: protocol. This service name is used to construct the SRV service label
    #: "_radiusdtls" for discovery of RADIUS/DTLS servers. [:rfc:`7585`]
    radiusdtls = -1, 'radiusdtls', TransportProtocol.get('udp')

    #: [TCP] Authentication, Accounting, and Dynamic Authorization via the RADIUS
    #: protocol. This service name is used to construct the SRV service label
    #: "_radiustls" for discovery of RADIUS/TLS servers. [:rfc:`7585`]
    radiustls = -1, 'radiustls', TransportProtocol.get('tcp')

    #: [TCP] Model Railroad Messaging
    railduino = -1, 'railduino', TransportProtocol.get('tcp')

    #: [N/A] Remote Audio Output Protocol (AirTunes)
    raop = -1, 'raop', TransportProtocol.get('undefined')

    #: [TCP] Industrial IOT self-discovery network
    rapta = -1, 'rapta', TransportProtocol.get('tcp')

    #: [N/A] RBR Instrument Communication
    rbr = -1, 'rbr', TransportProtocol.get('undefined')

    #: [N/A] PowerCard
    rce = -1, 'rce', TransportProtocol.get('undefined')

    #: [N/A] RealPlayer Shared Favorites
    realplayfavs = -1, 'realplayfavs', TransportProtocol.get('undefined')

    #: [TCP] The Recipe Box Exchange
    recipe_box = -1, 'recipe-box', TransportProtocol.get('tcp')

    #: [TCP] Recipe Sharing Protocol
    recipe_sharing = -1, 'recipe-sharing', TransportProtocol.get('tcp')

    #: - [TCP] Remote Camera Control
    #: - [UDP] Remote Camera Control
    recolive_cc = -1, 'recolive-cc', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] Recordit Image Transport Protocol
    recordit_itp = -1, 'recordit-itp', TransportProtocol.get('tcp')

    #: [N/A] Remote Device Control Protocol
    remote = -1, 'remote', TransportProtocol.get('undefined')

    #: [TCP] Remote Buddy remote control software command and data exchange
    remotebuddy = -1, 'remotebuddy', TransportProtocol.get('tcp')

    #: [N/A] LaCie Remote Burn
    remoteburn = -1, 'remoteburn', TransportProtocol.get('undefined')

    #: [N/A] ARTvps RenderDrive/PURE Renderer Protocol
    renderpipe = -1, 'renderpipe', TransportProtocol.get('undefined')

    #: [N/A] RendezvousPong
    rendezvouspong = -1, 'rendezvouspong', TransportProtocol.get('undefined')

    #: [N/A] Renkara synchronization protocol
    renkara_sync = -1, 'renkara-sync', TransportProtocol.get('undefined')

    #: [N/A] RESOL VBus
    resol_vbus = -1, 'resol-vbus', TransportProtocol.get('undefined')

    #: - [TCP] Retrospect backup and restore service
    #: - [UDP] Retrospect backup and restore service
    retrospect_497 = 497, 'retrospect', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Retrospect backup and restore service
    retrospect_null = -1, 'retrospect', TransportProtocol.get('undefined')

    #: [N/A] Remote Frame Buffer Client (Used by VNC viewers in listen-mode)
    rfbc = -1, 'rfbc', TransportProtocol.get('undefined')

    #: [N/A] RFID Reader Mach1(tm) Protocol
    rfid = -1, 'rfid', TransportProtocol.get('undefined')

    #: [TCP] RGB Spectrum Device Discovery
    rgb = -1, 'rgb', TransportProtocol.get('tcp')

    #: [N/A] Remote I/O USB Printer Protocol
    riousbprint = -1, 'riousbprint', TransportProtocol.get('undefined')

    #: [TCP] Roambot communication
    roambot = -1, 'roambot', TransportProtocol.get('tcp')

    #: [TCP] Like ircu (RFC1459), but failure tolerant due to strong consistency
    #: among n>=3 servers
    robustirc = -1, 'robustirc', TransportProtocol.get('tcp')

    #: [N/A] Roku Control Protocol
    roku_rcp = -1, 'roku-rcp', TransportProtocol.get('undefined')

    #: [TCP] RoomCast Control Protocol
    roomcast_capi = -1, 'roomcast-capi', TransportProtocol.get('tcp')

    #: [TCP] RoomCast Management Protocol
    roomcast_mapi = -1, 'roomcast-mapi', TransportProtocol.get('tcp')

    #: [N/A] RemoteQuickLaunch
    rql = -1, 'rql', TransportProtocol.get('undefined')

    #: [N/A] Robot Raconteur discovery
    rr_disc = -1, 'rr-disc', TransportProtocol.get('undefined')

    #: [TCP] Ruckus Radio Access Device, Installation, Commissioning and
    #: Troubleshooting service.
    rradict = -1, 'rradict', TransportProtocol.get('tcp')

    #: [N/A] Remote System Management Protocol (Server Instance)
    rsmp_server = -1, 'rsmp-server', TransportProtocol.get('undefined')

    #: [N/A] RubyGems GemServer
    rubygems = -1, 'rubygems', TransportProtocol.get('undefined')

    #: [TCP] Miele protocol robot cleaners
    rxxmiele = -1, 'rxxmiele', TransportProtocol.get('tcp')

    #: [TCP] Raymarine remote control protocol
    rym_rrc = -1, 'rym-rrc', TransportProtocol.get('tcp')

    #: [N/A] Safari Menu
    safarimenu = -1, 'safarimenu', TransportProtocol.get('undefined')

    #: [N/A] Salling Clicker Sharing
    sallingbridge = -1, 'sallingbridge', TransportProtocol.get('undefined')

    #: [N/A] Salling Clicker Service
    sallingclicker = -1, 'sallingclicker', TransportProtocol.get('undefined')

    #: [N/A] Salutafugi Peer-To-Peer Java Message Service Implementation
    salutafugijms = -1, 'salutafugijms', TransportProtocol.get('undefined')

    #: [N/A] Sandvox
    sandvox = -1, 'sandvox', TransportProtocol.get('undefined')

    #: [TCP] Proprietary Client Server Protocol
    savagesoft = -1, 'savagesoft', TransportProtocol.get('tcp')

    #: [N/A] StrawberryCat Golf Protocol
    sc_golf = -1, 'sc-golf', TransportProtocol.get('undefined')

    #: [N/A] Bonjour Scanning
    scanner = -1, 'scanner', TransportProtocol.get('undefined')

    #: [TCP] schindler internal messaging service
    schdca = -1, 'schdca', TransportProtocol.get('tcp')

    #: [N/A] Schick
    schick = -1, 'schick', TransportProtocol.get('undefined')

    #: [TCP] schindler internal messaging service
    schims = -1, 'schims', TransportProtocol.get('tcp')

    #: [TCP] logging service
    schlog = -1, 'schlog', TransportProtocol.get('tcp')

    #: [TCP] Schindler maintenance portal protocol
    schmpp = -1, 'schmpp', TransportProtocol.get('tcp')

    #: [TCP] schindler object messaging system
    schoms = -1, 'schoms', TransportProtocol.get('tcp')

    #: [TCP] Schindler service authentication portal
    schsap = -1, 'schsap', TransportProtocol.get('tcp')

    #: [TCP] Schindler internal messaging service
    schsrmp = -1, 'schsrmp', TransportProtocol.get('tcp')

    #: [TCP] schindler internal messaging service
    schvpp = -1, 'schvpp', TransportProtocol.get('tcp')

    #: [N/A] Scone
    scone = -1, 'scone', TransportProtocol.get('undefined')

    #: - [TCP] The service name is used by the SFTP protocol to upload log files from
    #: vehicles to road side units in a securely way in a cooperative intelligent
    #: transportation system.
    #: - [UDP] The service name is used by the SFTP protocol to upload log files from
    #: vehicles to road side units in a securely way in a cooperative intelligent
    #: transportation system.
    scoop_sftp = -1, 'scoop-sftp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] SCPI-RAW
    #: - [UDP] SCPI-RAW
    scpi_raw_5025 = 5025, 'scpi-raw', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] IEEE 488.2 (SCPI) Socket
    scpi_raw_null = -1, 'scpi-raw', TransportProtocol.get('undefined')

    #: - [TCP] SCPI-TELNET
    #: - [UDP] SCPI-TELNET
    scpi_telnet_5024 = 5024, 'scpi-telnet', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] IEEE 488.2 (SCPI) Telnet
    scpi_telnet_null = -1, 'scpi-telnet', TransportProtocol.get('undefined')

    #: [N/A] Speed Download
    sdsharing = -1, 'sdsharing', TransportProtocol.get('undefined')

    #: [N/A] SubEthaEdit 2
    see = -1, 'see', TransportProtocol.get('undefined')

    #: [N/A] seeCard
    seecard = -1, 'seecard', TransportProtocol.get('undefined')

    #: [N/A] Senteo Assessment Software Protocol
    senteo_http = -1, 'senteo-http', TransportProtocol.get('undefined')

    #: [N/A] Sentillion Vault System
    sentillion_vlc = -1, 'sentillion-vlc', TransportProtocol.get('undefined')

    #: [N/A] Sentillion Vault Systems Cluster
    sentillion_vlt = -1, 'sentillion-vlt', TransportProtocol.get('undefined')

    #: [N/A] SEPV Application Data Synchronization Protocol
    sepvsync = -1, 'sepvsync', TransportProtocol.get('undefined')

    #: [N/A] serendiPd Shared Patches for Pure Data
    serendipd = -1, 'serendipd', TransportProtocol.get('undefined')

    #: [N/A] ServerEye AgentContainer Communication Protocol
    servereye = -1, 'servereye', TransportProtocol.get('undefined')

    #: [N/A] Mac OS X Server Admin
    servermgr = -1, 'servermgr', TransportProtocol.get('undefined')

    #: [N/A] DNS Service Discovery
    services = -1, 'services', TransportProtocol.get('undefined')

    #: [N/A] Session File Sharing
    sessionfs = -1, 'sessionfs', TransportProtocol.get('undefined')

    #: [TCP] Remote Control for Set List Maker app
    setlistmaker_rc = -1, 'setlistmaker-rc', TransportProtocol.get('tcp')

    #: [TCP] Screen sharing for Set List Maker app
    setlistmaker_ss = -1, 'setlistmaker-ss', TransportProtocol.get('tcp')

    #: [N/A] Secure File Transfer Protocol over SSH
    sftp_ssh = -1, 'sftp-ssh', TransportProtocol.get('undefined')

    #: [N/A] Sun Grid Engine (Execution Host)
    sge_exec = -1, 'sge-exec', TransportProtocol.get('undefined')

    #: - [TCP] Grid Engine Qmaster Service IANA assigned this well-formed service
    #: name as a replacement for "sge_qmaster".
    #: - [TCP] Grid Engine Qmaster Service
    #: - [UDP] Grid Engine Qmaster Service IANA assigned this well-formed service
    #: name as a replacement for "sge_qmaster".
    #: - [UDP] Grid Engine Qmaster Service
    sge_qmaster_6444 = 6444, 'sge-qmaster', TransportProtocol.get('tcp') | TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('udp')

    #: [N/A] Sun Grid Engine (Master)
    sge_qmaster_null = -1, 'sge-qmaster', TransportProtocol.get('undefined')

    #: [N/A] Window Shifter server protocol
    shifter = -1, 'shifter', TransportProtocol.get('undefined')

    #: [TCP] SHIP (Smart Home IP)
    ship = -1, 'ship', TransportProtocol.get('tcp')

    #: [N/A] Swift Office Ships
    shipsgm = -1, 'shipsgm', TransportProtocol.get('undefined')

    #: [N/A] Swift Office Ships
    shipsinvit = -1, 'shipsinvit', TransportProtocol.get('undefined')

    #: [N/A] SplashShopper Synchronization Service
    shoppersync = -1, 'shoppersync', TransportProtocol.get('undefined')

    #: [UDP] The protocol is used to sync database among iOS devices and Mac OS X
    #: computers.
    shots_sync = -1, 'shots-sync', TransportProtocol.get('udp')

    #: [N/A] Nicecast
    shoutcast = -1, 'shoutcast', TransportProtocol.get('undefined')

    #: [TCP] Simulated insufflator synchronisation protocol
    siminsufflator = -1, 'siminsufflator', TransportProtocol.get('tcp')

    #: [N/A] Medical simulation patient monitor syncronisation protocol
    simmon = -1, 'simmon', TransportProtocol.get('undefined')

    #: [N/A] simusoftpong iPhone game protocol
    simusoftpong = -1, 'simusoftpong', TransportProtocol.get('undefined')

    #: [N/A] Session Initiation Protocol Uniform Resource Identifier
    sipuri = -1, 'sipuri', TransportProtocol.get('undefined')

    #: [N/A] Sirona Xray Protocol
    sironaxray = -1, 'sironaxray', TransportProtocol.get('undefined')

    #: - [TCP] The protocol is used to transfer database records between an iOS
    #: device to a Mac OS X computer
    #: - [UDP] The protocol is used to transfer database records between an iOS
    #: device to a Mac OS X computer
    skillscapture = -1, 'skillscapture', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Skype
    skype = -1, 'skype', TransportProtocol.get('undefined')

    #: [UDP] Sleep Proxy Server
    sleep_proxy = -1, 'sleep-proxy', TransportProtocol.get('udp')

    #: [TCP] Sleeptracker(R) The loT Smartbed Platform
    sleeptracker = -1, 'sleeptracker', TransportProtocol.get('tcp')

    #: [N/A] SliMP3 Server Command-Line Interface
    slimcli = -1, 'slimcli', TransportProtocol.get('undefined')

    #: [N/A] SliMP3 Server Web Interface
    slimhttp = -1, 'slimhttp', TransportProtocol.get('undefined')

    #: - [TCP] Remote Service Discovery in the Service Location [:rfc:`3832`]
    #: - [UDP] Remote Service Discovery in the Service Location [:rfc:`3832`]
    slpda = -1, 'slpda', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] terminal access to laundry appliances
    smag = -1, 'smag', TransportProtocol.get('tcp')

    #: [TCP] Smart Home Device Setup
    smaho = -1, 'smaho', TransportProtocol.get('tcp')

    #: [N/A] Smart Energy Profile
    smartenergy = -1, 'smartenergy', TransportProtocol.get('undefined')

    #: [TCP] home control
    smartsocket = -1, 'smartsocket', TransportProtocol.get('tcp')

    #: [N/A] Server Message Block over TCP/IP
    smb = -1, 'smb', TransportProtocol.get('undefined')

    #: [TCP] Domain signing certificate for S/MIME keys [draft-bhjl-x509-srv-01]
    smimeca = -1, 'smimeca', TransportProtocol.get('tcp')

    #: [N/A] Short Text Message Sending and Delivery Status Service
    sms = -1, 'sms', TransportProtocol.get('undefined')

    #: [N/A] Syncellence file synchronization protocol
    smsync = -1, 'smsync', TransportProtocol.get('undefined')

    #: [TCP] End-to-end TLS Relay Client Connection [https://snif.host/relay-
    #: proto#snif-cln]
    snif_cln = -1, 'snif-cln', TransportProtocol.get('tcp')

    #: [TCP] End-to-end TLS Relay Cluster [https://snif.host/relay-proto#snif-fifo]
    snif_fifo = -1, 'snif-fifo', TransportProtocol.get('tcp')

    #: [TCP] End-to-end TLS Relay Service Connection [https://snif.host/relay-
    #: proto#snif-srv]
    snif_srv = -1, 'snif-srv', TransportProtocol.get('tcp')

    #: [N/A] Simple Object Access Protocol
    soap = -1, 'soap', TransportProtocol.get('undefined')

    #: [N/A] Socketcloud distributed application framework
    socketcloud = -1, 'socketcloud', TransportProtocol.get('undefined')

    #: [TCP] Secure On Device API
    soda = -1, 'soda', TransportProtocol.get('tcp')

    #: [N/A] SousChef Recipe Sharing Protocol
    souschef = -1, 'souschef', TransportProtocol.get('undefined')

    #: [N/A] Simple Object eXchange
    sox = -1, 'sox', TransportProtocol.get('undefined')

    #: [N/A] SpareChange data sharing protocol
    sparechange = -1, 'sparechange', TransportProtocol.get('undefined')

    #: [N/A] SPARQL Protocol and RDF Query Language
    sparql = -1, 'sparql', TransportProtocol.get('undefined')

    #: [N/A] sPearCat Host Discovery
    spearcat = -1, 'spearcat', TransportProtocol.get('undefined')

    #: [TCP] Sierra Photonics Inc. data protocol
    spidap = -1, 'spidap', TransportProtocol.get('tcp')

    #: [N/A] Spin Crisis
    spincrisis = -1, 'spincrisis', TransportProtocol.get('undefined')

    #: [TCP] Binary message passing protocol
    spiderelectron = -1, 'spiderelectron', TransportProtocol.get('tcp')

    #: [N/A] launchTunes
    spl_itunes = -1, 'spl-itunes', TransportProtocol.get('undefined')

    #: [N/A] netTunes
    spr_itunes = -1, 'spr-itunes', TransportProtocol.get('undefined')

    #: [N/A] SplashData Synchronization Service
    splashsync = -1, 'splashsync', TransportProtocol.get('undefined')

    #: [TCP] DNS64 in split configuration
    split_dns64 = -1, 'split-dns64', TransportProtocol.get('tcp')

    #: [TCP] SongPresenter
    spres = -1, 'spres', TransportProtocol.get('tcp')

    #: [TCP] SpinetiX HMP
    spx_hmp = -1, 'spx-hmp', TransportProtocol.get('tcp')

    #: [TCP] Square Connect Control Protocol
    sqp = -1, 'sqp', TransportProtocol.get('tcp')

    #: [TCP] Samsung Smart Interaction for Group Network
    ss_sign = -1, 'ss-sign', TransportProtocol.get('tcp')

    #: [UDP] Samsung Smart Interaction for Group Network Discovery
    ss_sign_disc = -1, 'ss-sign-disc', TransportProtocol.get('udp')

    #: [TCP] Studio Six Digital Wireless Audio
    ssd_audio = -1, 'ssd-audio', TransportProtocol.get('tcp')

    #: - [TCP] The Secure Shell (SSH) Protocol [:rfc:`4251`]
    #: - [UDP] The Secure Shell (SSH) Protocol [:rfc:`4251`]
    #: - [SCTP] SSH [:rfc:`9260`]
    ssh_22 = 22, 'ssh', TransportProtocol.get('tcp') | TransportProtocol.get('udp') | TransportProtocol.get('sctp')

    #: [TCP] SSH Remote Login Protocol
    ssh_null = -1, 'ssh', TransportProtocol.get('tcp')

    #: [N/A] Screen Sharing
    ssscreenshare = -1, 'ssscreenshare', TransportProtocol.get('undefined')

    #: [TCP] Remote Controlled Multimedia Recorder Network
    startrecapp = -1, 'startrecapp', TransportProtocol.get('tcp')

    #: [TCP] Stingray Remote Procedure Call
    stingray_rpc = -1, 'stingray-rpc', TransportProtocol.get('tcp')

    #: [TCP] Stingray remote control
    stingray_remote = -1, 'stingray-remote', TransportProtocol.get('tcp')

    #: [N/A] Strateges
    strateges = -1, 'strateges', TransportProtocol.get('undefined')

    #: [N/A] Lexcycle Stanza service for discovering shared books
    stanza = -1, 'stanza', TransportProtocol.get('undefined')

    #: [N/A] Sticky Notes
    stickynotes = -1, 'stickynotes', TransportProtocol.get('undefined')

    #: [TCP] One Time Pad Synchronisation
    stotp = -1, 'stotp', TransportProtocol.get('tcp')

    #: [TCP] Strobe Synchronization
    strobe_sync = -1, 'strobe-sync', TransportProtocol.get('tcp')

    #: - [TCP] Remote control protocol for Sugarlock consumer electronics devices
    #: - [UDP] Remote control protocol for Sugarlock consumer electronics devices
    sugarlock_rcp = -1, 'sugarlock-rcp', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Supple Service protocol
    supple = -1, 'supple', TransportProtocol.get('undefined')

    #: [N/A] Surveillus Networks Discovery Protocol
    surveillus = -1, 'surveillus', TransportProtocol.get('undefined')

    #: - [TCP] Subversion
    #: - [UDP] Subversion
    svn_3690 = 3690, 'svn', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Subversion
    svn_null = -1, 'svn', TransportProtocol.get('undefined')

    #: [N/A] Signwave Card Sharing Protocol
    swcards = -1, 'swcards', TransportProtocol.get('undefined')

    #: [N/A] Wireless home control remote control protocol
    switcher = -1, 'switcher', TransportProtocol.get('undefined')

    #: [N/A] Swordfish Protocol for Input/Output
    swordfish = -1, 'swordfish', TransportProtocol.get('undefined')

    #: [N/A] Framework for transferring any file from any app, to       any app on
    #: any device: simply with a swÿp.
    swyp = -1, 'swyp', TransportProtocol.get('undefined')

    #: [N/A] Synchronize! Pro X
    sxqdea = -1, 'sxqdea', TransportProtocol.get('undefined')

    #: [N/A] Sybase Server
    sybase_tds = -1, 'sybase-tds', TransportProtocol.get('undefined')

    #: [TCP] Remote control of Synclavier Digital Audio Workstation over local area
    #: network.
    synclavier = -1, 'synclavier', TransportProtocol.get('tcp')

    #: [N/A] Syncopation Synchronization Protocol by Sonzea
    syncopation = -1, 'syncopation', TransportProtocol.get('undefined')

    #: [N/A] Synchronize! X Plus 2.0
    syncqdea = -1, 'syncqdea', TransportProtocol.get('undefined')

    #: [N/A] Synergy Peer Discovery
    synergy = -1, 'synergy', TransportProtocol.get('undefined')

    #: [N/A] SynkSharing synchronization protocol
    synksharing = -1, 'synksharing', TransportProtocol.get('undefined')

    #: [TCP] This service name is used to construct the       SRV service label
    #: "_sztp" for discovering       SZTP bootstrap servers. [:rfc:`8572`]
    sztp = -1, 'sztp', TransportProtocol.get('tcp')

    #: [N/A] Data Transmission and Synchronization
    taccounting = -1, 'taccounting', TransportProtocol.get('undefined')

    #: [N/A] Tango Remote Control Protocol
    tango = -1, 'tango', TransportProtocol.get('undefined')

    #: [N/A] Tapinoma Easycontact receiver
    tapinoma_ecs = -1, 'tapinoma-ecs', TransportProtocol.get('undefined')

    #: [N/A] Task Coach Two-way Synchronization Protocol for iPhone
    taskcoachsync = -1, 'taskcoachsync', TransportProtocol.get('undefined')

    #: [N/A] tbricks internal protocol
    tbricks = -1, 'tbricks', TransportProtocol.get('undefined')

    #: [N/A] Time Code
    tcode = -1, 'tcode', TransportProtocol.get('undefined')

    #: [N/A] Tracking Control Unit by RPM-PSI
    tcu = -1, 'tcu', TransportProtocol.get('undefined')

    #: [N/A] TE-SYSTEMS GmbH Fax Server Daemon
    te_faxserver = -1, 'te-faxserver', TransportProtocol.get('undefined')

    #: [N/A] ARTIS Team Task
    teamlist = -1, 'teamlist', TransportProtocol.get('undefined')

    #: [UDP] Teddington Controls
    ted = -1, 'ted', TransportProtocol.get('udp')

    #: [UDP] teleport
    teleport = -1, 'teleport', TransportProtocol.get('udp')

    #: [TCP] Proprietary
    tenir_rc = -1, 'tenir-rc', TransportProtocol.get('tcp')

    #: [N/A] Terascala Filesystem Manager Protocol
    tera_fsmgr = -1, 'tera-fsmgr', TransportProtocol.get('undefined')

    #: [N/A] Terascala Maintenance Protocol
    tera_mp = -1, 'tera-mp', TransportProtocol.get('undefined')

    #: [TCP] Test Controller Card
    test_ok = -1, 'test-ok', TransportProtocol.get('tcp')

    #: [N/A] ThinkFlood RedEye IR bridge
    tf_redeye = -1, 'tf-redeye', TransportProtocol.get('undefined')

    #: - [TCP] Internet of things service discovery
    #: - [UDP] Internet of things service discovery
    thing = -1, 'thing', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] tinkerbuilt Thumb Wrestling game
    thumbwrestling = -1, 'thumbwrestling', TransportProtocol.get('undefined')

    #: [N/A] TI Connect Manager Discovery Service
    ticonnectmgr = -1, 'ticonnectmgr', TransportProtocol.get('undefined')

    #: [TCP] Tic Tac Toe game
    tic_tac_toe = -1, 'tic-tac-toe', TransportProtocol.get('tcp')

    #: [TCP] Time Zone Data Distribution Service - non-TLS [:rfc:`7808`]
    timezone = -1, 'timezone', TransportProtocol.get('tcp')

    #: [TCP] Time Zone Data Distribution Service - over TLS [:rfc:`7808`]
    timezones = -1, 'timezones', TransportProtocol.get('tcp')

    #: [N/A] TI Navigator Hub 1.0 Discovery Service
    tinavigator = -1, 'tinavigator', TransportProtocol.get('undefined')

    #: [TCP] TiVo Device Protocol
    tivo_device = -1, 'tivo-device', TransportProtocol.get('tcp')

    #: [N/A] TiVo Home Media Engine Protocol
    tivo_hme = -1, 'tivo-hme', TransportProtocol.get('undefined')

    #: [TCP] TiVo RPC Protocol
    tivo_mindrpc = -1, 'tivo-mindrpc', TransportProtocol.get('tcp')

    #: [N/A] TiVo Music Protocol
    tivo_music = -1, 'tivo-music', TransportProtocol.get('undefined')

    #: [N/A] TiVo Photos Protocol
    tivo_photos = -1, 'tivo-photos', TransportProtocol.get('undefined')

    #: [N/A] TiVo Remote Protocol
    tivo_remote = -1, 'tivo-remote', TransportProtocol.get('undefined')

    #: [N/A] TiVo Videos Protocol
    tivo_videos = -1, 'tivo-videos', TransportProtocol.get('undefined')

    #: [N/A] 2Do Sync Helper Tool for Mac OS X and PCs
    todogwa = -1, 'todogwa', TransportProtocol.get('undefined')

    #: [N/A] Tomboy
    tomboy = -1, 'tomboy', TransportProtocol.get('undefined')

    #: [N/A] ToothPics Dental Office Support Server
    toothpicserver = -1, 'toothpicserver', TransportProtocol.get('undefined')

    #: [N/A] iPhone and iPod touch Remote Controllable
    touch_able = -1, 'touch-able', TransportProtocol.get('undefined')

    #: [N/A] iPhone and iPod touch Remote Pairing
    touch_remote = -1, 'touch-remote', TransportProtocol.get('undefined')

    #: [TCP] Coordination service for client users of the TotalPraisTrax iPad
    #: application
    tptx_console = -1, 'tptx-console', TransportProtocol.get('tcp')

    #: [TCP] Service discovery and media transfer for peer to peer mobile media
    #: transfer app
    transmitr = -1, 'transmitr', TransportProtocol.get('tcp')

    #: [UDP] Thread Radio Encapsulation Link
    trel = -1, 'trel', TransportProtocol.get('udp')

    #: [N/A] triCerat Simplify Visibility Client
    tri_vis_client = -1, 'tri-vis-client', TransportProtocol.get('undefined')

    #: [N/A] triCerat Simplify Visibility Server
    tri_vis_server = -1, 'tri-vis-server', TransportProtocol.get('undefined')

    #: [N/A] Tryst
    tryst = -1, 'tryst', TransportProtocol.get('undefined')

    #: [TCP] The Social Broadband Interference Information Sharing
    tsbiis = -1, 'tsbiis', TransportProtocol.get('tcp')

    #: [N/A] Trivial Technology's 4 in a Row
    tt4inarow = -1, 'tt4inarow', TransportProtocol.get('undefined')

    #: [N/A] Trivial Technology's Checkers
    ttcheckers = -1, 'ttcheckers', TransportProtocol.get('undefined')

    #: [N/A] TechTool Pro 4 Anti-Piracy Service
    ttp4daemon = -1, 'ttp4daemon', TransportProtocol.get('undefined')

    #: [N/A] Tunage Media Control Service
    tunage = -1, 'tunage', TransportProtocol.get('undefined')

    #: [N/A] TuneRanger
    tuneranger = -1, 'tuneranger', TransportProtocol.get('undefined')

    #: [TCP] Vogel's TV mount control
    tvm = -1, 'tvm', TransportProtocol.get('tcp')

    #: [TCP] detect sanitary product
    twinlevel = -1, 'twinlevel', TransportProtocol.get('tcp')

    #: - [TCP] Service to enable multiplayer game called two snakes.
    #: - [SCTP] Service to enable multiplayer game called two snakes.
    twosnakes = -1, 'twosnakes', TransportProtocol.get('tcp') | TransportProtocol.get('sctp')

    #: [TCP] Tyba control
    tyba = -1, 'tyba', TransportProtocol.get('tcp')

    #: [TCP] TZ-Software remote procedure call based synchronization protocol
    tzrpc = -1, 'tzrpc', TransportProtocol.get('tcp')

    #: [N/A] Ubertragen
    ubertragen = -1, 'ubertragen', TransportProtocol.get('undefined')

    #: [UDP] Tactical Unified Communicator
    ucdynamics_tuc = -1, 'ucdynamics-tuc', TransportProtocol.get('udp')

    #: [N/A] Universal Description, Discovery and Integration
    uddi = -1, 'uddi', TransportProtocol.get('undefined')

    #: [N/A] Universal Description, Discovery and Integration Inquiry
    uddi_inq = -1, 'uddi-inq', TransportProtocol.get('undefined')

    #: [N/A] Universal Description, Discovery and Integration Publishing
    uddi_pub = -1, 'uddi-pub', TransportProtocol.get('undefined')

    #: [N/A] Universal Description, Discovery and Integration Subscription
    uddi_sub = -1, 'uddi-sub', TransportProtocol.get('undefined')

    #: [N/A] Universal Description, Discovery and Integration Security
    uddi_sec = -1, 'uddi-sec', TransportProtocol.get('undefined')

    #: [N/A] Universal Plug and Play
    upnp = -1, 'upnp', TransportProtocol.get('undefined')

    #: [N/A] URL Advertising
    urlbookmark = -1, 'urlbookmark', TransportProtocol.get('undefined')

    #: [UDP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_agt_coap = -1, 'usp-agt-coap', TransportProtocol.get('udp')

    #: [TCP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_agt_http = -1, 'usp-agt-http', TransportProtocol.get('tcp')

    #: [TCP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_agt_mqtt = -1, 'usp-agt-mqtt', TransportProtocol.get('tcp')

    #: [TCP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_agt_stomp = -1, 'usp-agt-stomp', TransportProtocol.get('tcp')

    #: [TCP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_agt_ws = -1, 'usp-agt-ws', TransportProtocol.get('tcp')

    #: [UDP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_ctr_coap = -1, 'usp-ctr-coap', TransportProtocol.get('udp')

    #: [TCP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_ctr_http = -1, 'usp-ctr-http', TransportProtocol.get('tcp')

    #: [TCP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_ctr_mqtt = -1, 'usp-ctr-mqtt', TransportProtocol.get('tcp')

    #: [TCP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_ctr_stomp = -1, 'usp-ctr-stomp', TransportProtocol.get('tcp')

    #: [TCP] USP discovery [http://www.broadband-forum.org/assignments]
    usp_ctr_ws = -1, 'usp-ctr-ws', TransportProtocol.get('tcp')

    #: [N/A] Universal Switching Corporation products
    uswi = -1, 'uswi', TransportProtocol.get('undefined')

    #: [N/A] uTest
    utest = -1, 'utest', TransportProtocol.get('undefined')

    #: [N/A] Unbit Web Server Gateway Interface
    uwsgi = -1, 'uwsgi', TransportProtocol.get('undefined')

    #: [TCP] Indicator for VAPIX support over HTTP
    vapix_http = -1, 'vapix-http', TransportProtocol.get('tcp')

    #: [TCP] Indicator for VAPIX support over HTTPS
    vapix_https = -1, 'vapix-https', TransportProtocol.get('tcp')

    #: [N/A] American Dynamics VideoEdge Decoder Control Service
    ve_decoder = -1, 've-decoder', TransportProtocol.get('undefined')

    #: [N/A] American Dynamics VideoEdge Encoder Control Service
    ve_encoder = -1, 've-encoder', TransportProtocol.get('undefined')

    #: [N/A] American Dynamics VideoEdge Recorder Control Service
    ve_recorder = -1, 've-recorder', TransportProtocol.get('undefined')

    #: [TCP] Application specific synchronization protocol
    vedabase = -1, 'vedabase', TransportProtocol.get('tcp')

    #: [TCP] USB over IP Sharing System
    vhusb = -1, 'vhusb', TransportProtocol.get('tcp')

    #: [N/A] VirtualDJ Remote Control protocol
    virtualdj = -1, 'virtualdj', TransportProtocol.get('undefined')

    #: [N/A] visel Q-System services
    visel = -1, 'visel', TransportProtocol.get('undefined')

    #: [TCP] Server location via DNS-SD
    voalte2 = -1, 'voalte2', TransportProtocol.get('tcp')

    #: [TCP] Server location via DNS-SD
    voalte3 = -1, 'voalte3', TransportProtocol.get('tcp')

    #: [TCP] Internal Communication Service
    volterio = -1, 'volterio', TransportProtocol.get('tcp')

    #: [N/A] Virtual Object System (using VOP/TCP)
    vos = -1, 'vos', TransportProtocol.get('undefined')

    #: [TCP] RTP Mixer/Summation Resource
    voxidahmp = -1, 'voxidahmp', TransportProtocol.get('tcp')

    #: [TCP] Verimag mobile class protocol over P2P
    vrmg_p2p = -1, 'vrmg-p2p', TransportProtocol.get('tcp')

    #: [N/A] VueProRenderCow
    vue4rendercow = -1, 'vue4rendercow', TransportProtocol.get('undefined')

    #: [N/A] VXI-11 TCP/IP Instrument Protocol
    vxi_11 = -1, 'vxi-11', TransportProtocol.get('undefined')

    #: [TCP] Proprietary
    wakeywakey = -1, 'wakeywakey', TransportProtocol.get('tcp')

    #: [N/A] Walkie Talkie
    walkietalkie = -1, 'walkietalkie', TransportProtocol.get('undefined')

    #: [TCP] NAS Service Protocol
    wd_2go = -1, 'wd-2go', TransportProtocol.get('tcp')

    #: [TCP] Windows Device Portal
    wdp = -1, 'wdp', TransportProtocol.get('tcp')

    #: [TCP] HTTP-based protocol for DAQ devices
    web_xi = -1, 'web-xi', TransportProtocol.get('tcp')

    #: - [TCP] Cisco WebEx serials products will release Bonjour based service
    #: - [UDP] Cisco WebEx serials products will release Bonjour based service
    webex = -1, 'webex', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [N/A] Proprietary collaborative messaging protocol
    we_jell = -1, 'we-jell', TransportProtocol.get('undefined')

    #: [N/A] World Wide Web Distributed Authoring and Versioning (WebDAV)
    webdav = -1, 'webdav', TransportProtocol.get('undefined')

    #: [N/A] WebDAV over SSL/TLS
    webdavs = -1, 'webdavs', TransportProtocol.get('undefined')

    #: [N/A] WebIS Sync Protocol
    webissync = -1, 'webissync', TransportProtocol.get('undefined')

    #: [N/A] weDraw document sharing protocol
    wedraw = -1, 'wedraw', TransportProtocol.get('undefined')

    #: [N/A] Whamb
    whamb = -1, 'whamb', TransportProtocol.get('undefined')

    #: [N/A] Honeywell Video Systems
    whistler = -1, 'whistler', TransportProtocol.get('undefined')

    #: [UDP] WiFi Control Platform
    wicop = -1, 'wicop', TransportProtocol.get('udp')

    #: [TCP] System for transferring files between mobile device and computer in a
    #: local network
    wifile = -1, 'wifile', TransportProtocol.get('tcp')

    #: [N/A] WiTap Sample Game Protocol
    witap = -1, 'witap', TransportProtocol.get('undefined')

    #: [N/A] witapvoice
    witapvoice = -1, 'witapvoice', TransportProtocol.get('undefined')

    #: [N/A] Workgroup Server Discovery
    wkgrpsvr = -1, 'wkgrpsvr', TransportProtocol.get('undefined')

    #: [N/A] Workgroup Manager
    workstation = -1, 'workstation', TransportProtocol.get('undefined')

    #: [N/A] Roku Cascade Wormhole Protocol
    wormhole = -1, 'wormhole', TransportProtocol.get('undefined')

    #: [N/A] Novell collaboration workgroup
    workgroup = -1, 'workgroup', TransportProtocol.get('undefined')

    #: - [TCP] W3C WoT Thing Description or Directory
    #: - [UDP] W3C WoT Thing Description or Directory
    wot = -1, 'wot', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: [TCP] World Programming repository server
    wpl_ers_http = -1, 'wpl-ers-http', TransportProtocol.get('tcp')

    #: [TCP] World Programming repository server
    wpl_ers_zmq = -1, 'wpl-ers-zmq', TransportProtocol.get('tcp')

    #: [N/A] Writie Talkie Data Sharing
    writietalkie = -1, 'writietalkie', TransportProtocol.get('undefined')

    #: [N/A] Web Services
    ws = -1, 'ws', TransportProtocol.get('undefined')

    #: [N/A] Wyatt Technology Corporation HELEOS
    wtc_heleos = -1, 'wtc-heleos', TransportProtocol.get('undefined')

    #: [N/A] Wyatt Technology Corporation QELS
    wtc_qels = -1, 'wtc-qels', TransportProtocol.get('undefined')

    #: [N/A] Wyatt Technology Corporation Optilab rEX
    wtc_rex = -1, 'wtc-rex', TransportProtocol.get('undefined')

    #: [N/A] Wyatt Technology Corporation ViscoStar
    wtc_viscostar = -1, 'wtc-viscostar', TransportProtocol.get('undefined')

    #: [N/A] Wyatt Technology Corporation DynaPro Plate Reader
    wtc_wpr = -1, 'wtc-wpr', TransportProtocol.get('undefined')

    #: [N/A] PictureSharing sample code
    wwdcpic = -1, 'wwdcpic', TransportProtocol.get('undefined')

    #: [TCP] The X.510 wrapper protocol [ITU-T X.510 / ISO/IEC 9584-11]
    x510_9877 = 9877, 'x510', TransportProtocol.get('tcp')

    #: [TCP] Security service to protocols
    x510_null = -1, 'x510', TransportProtocol.get('tcp')

    #: [N/A] x-on services synchronisation protocol
    x_on = -1, 'x-on', TransportProtocol.get('undefined')

    #: [UDP] x-plane9
    x_plane9 = -1, 'x-plane9', TransportProtocol.get('udp')

    #: [N/A] Xcode Distributed Compiler
    xcodedistcc = -1, 'xcodedistcc', TransportProtocol.get('undefined')

    #: [TCP] xential document creation services
    xential = -1, 'xential', TransportProtocol.get('tcp')

    #: [N/A] xGate Remote Management Interface
    xgate_rmi = -1, 'xgate-rmi', TransportProtocol.get('undefined')

    #: [TCP] XMI Systems home terminal local connection
    xmiserver = -1, 'xmiserver', TransportProtocol.get('tcp')

    #: [N/A] Xperientia Mobile Protocol
    xmp = -1, 'xmp', TransportProtocol.get('undefined')

    #: [N/A] Xsan Client
    xsanclient = -1, 'xsanclient', TransportProtocol.get('undefined')

    #: [N/A] Xsan Server
    xsanserver = -1, 'xsanserver', TransportProtocol.get('undefined')

    #: [N/A] Xsan System
    xsansystem = -1, 'xsansystem', TransportProtocol.get('undefined')

    #: [N/A] xTime License
    xtimelicence = -1, 'xtimelicence', TransportProtocol.get('undefined')

    #: [N/A] xTime Project
    xtshapro = -1, 'xtshapro', TransportProtocol.get('undefined')

    #: [N/A] XUL (XML User Interface Language) transported over HTTP
    xul_http = -1, 'xul-http', TransportProtocol.get('undefined')

    #: [UDP] Yakumo iPhone OS Device Control Protocol
    yakumo = -1, 'yakumo', TransportProtocol.get('udp')

    #: - [TCP] Z-Wave Protocol
    #: - [UDP] Z-Wave Protocol
    z_wave_4123 = 4123, 'z-wave', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] Z-Wave Service Discovery
    #: - [UDP] Z-Wave Service Discovery
    z_wave_null = -1, 'z-wave', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] High performance brokerless messaging
    #: - [UDP] High performance brokerless messaging
    zeromq = -1, 'zeromq', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ZigBee Bridge device
    #: - [UDP] ZigBee Bridge device
    zigbee_bridge = -1, 'zigbee-bridge', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    #: - [TCP] ZigBee IP Gateway
    #: - [UDP] ZigBee IP Gateway
    zigbee_gateway = -1, 'zigbee-gateway', TransportProtocol.get('tcp') | TransportProtocol.get('udp')

    @staticmethod
    def get(key: 'int | str', default: 'int' = -1, *,
            proto: 'TransportProtocol | str' = TransportProtocol.undefined) -> 'AppType':
        """Backport support for original codes.

        Args:
            key: Key to get enum item.
            default: Default value if not found.
            proto: Transport protocol of the enum item.

        :meta private:
        """
        if isinstance(key, int):
            if isinstance(proto, str):
                proto = TransportProtocol.get(proto.lower())
            temp_ns = AppType.__members_proto__.get(proto, {})
            if key in temp_ns:
                return temp_ns[key]
            try:
                ret = AppType._missing_(key)
                if ret is None:
                    raise ValueError
            except ValueError:

                ret = extend_enum(AppType, 'PORT_%d_%s' % (key, proto.name), key, 'unknown', proto)
            return ret
        if key in AppType.__members_proto__:
            return getattr(AppType, key)
        return extend_enum(AppType, key, default, key)

    @classmethod
    def _missing_(cls, value: 'int') -> 'AppType':
        """Lookup function used when value is not found.

        Args:
            value: Value to get enum item.

        """
        if not (isinstance(value, int) and 0 <= value <= 65535):
            raise ValueError('%r is not a valid %s' % (value, cls.__name__))
        if value in cls.__members_proto__.get(TransportProtocol.undefined, {}):  # type: ignore[call-overload]
            return cls.__members_proto__[TransportProtocol.undefined][value]  # type: ignore[index]
        if 225 <= value <= 241:
            #: [N/A] Reserved
            return extend_enum(cls, 'reserved_%d' % value, value, 'reserved', TransportProtocol.get('undefined'))
        if 249 <= value <= 255:
            #: [N/A] Reserved
            return extend_enum(cls, 'reserved_%d' % value, value, 'reserved', TransportProtocol.get('undefined'))
        if 272 <= value <= 279:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 288 <= value <= 307:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 325 <= value <= 332:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 334 <= value <= 343:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 717 <= value <= 728:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 732 <= value <= 740:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 745 <= value <= 746:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 755 <= value <= 757:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 778 <= value <= 779:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 781 <= value <= 785:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 788 <= value <= 799:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 803 <= value <= 809:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 811 <= value <= 827:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 834 <= value <= 846:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 849 <= value <= 852:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 855 <= value <= 859:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 863 <= value <= 872:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 874 <= value <= 885:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 889 <= value <= 899:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 904 <= value <= 909:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 916 <= value <= 952:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 954 <= value <= 988:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 1002 <= value <= 1007:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 1011 <= value <= 1020:
            #: [N/A] Reserved
            return extend_enum(cls, 'reserved_%d' % value, value, 'reserved', TransportProtocol.get('undefined'))
        if 2194 <= value <= 2196:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 3322 <= value <= 3325:
            #: [N/A] Active Networks
            return extend_enum(cls, 'active_net_%d' % value, value, 'active-net', TransportProtocol.get('undefined'))
        if 3367 <= value <= 3371:
            #: [N/A] Satellite Video Data Link
            return extend_enum(cls, 'satvid_datalnk_%d' % value, value, 'satvid-datalnk', TransportProtocol.get('undefined'))
        if 4200 <= value <= 4299:
            #: [N/A] VRML Multi User Systems
            return extend_enum(cls, 'vrml_multi_use_%d' % value, value, 'vrml-multi-use', TransportProtocol.get('undefined'))
        if 4337 <= value <= 4339:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4363 <= value <= 4365:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4380 <= value <= 4388:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4397 <= value <= 4399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4434 <= value <= 4440:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4461 <= value <= 4483:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4489 <= value <= 4499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4503 <= value <= 4533:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4539 <= value <= 4544:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4560 <= value <= 4562:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4564 <= value <= 4565:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4571 <= value <= 4572:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4574 <= value <= 4589:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4607 <= value <= 4620:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4622 <= value <= 4645:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4647 <= value <= 4657:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4693 <= value <= 4699:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4705 <= value <= 4710:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4712 <= value <= 4724:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4734 <= value <= 4736:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4748 <= value <= 4748:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4757 <= value <= 4773:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4775 <= value <= 4783:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4793 <= value <= 4799:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4805 <= value <= 4826:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4828 <= value <= 4836:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4852 <= value <= 4866:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4872 <= value <= 4875:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4886 <= value <= 4887:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4890 <= value <= 4893:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4895 <= value <= 4898:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4903 <= value <= 4911:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4916 <= value <= 4935:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4938 <= value <= 4939:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4943 <= value <= 4948:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4954 <= value <= 4968:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4972 <= value <= 4979:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4981 <= value <= 4982:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 4992 <= value <= 4998:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5016 <= value <= 5019:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5035 <= value <= 5041:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5076 <= value <= 5077:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5088 <= value <= 5089:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5095 <= value <= 5098:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5108 <= value <= 5110:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5118 <= value <= 5119:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5121 <= value <= 5132:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5138 <= value <= 5144:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5147 <= value <= 5149:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5158 <= value <= 5160:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5169 <= value <= 5171:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5173 <= value <= 5189:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5198 <= value <= 5199:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5204 <= value <= 5208:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5210 <= value <= 5214:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5216 <= value <= 5220:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5238 <= value <= 5241:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5255 <= value <= 5263:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5266 <= value <= 5268:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5273 <= value <= 5279:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5283 <= value <= 5297:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5322 <= value <= 5342:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5345 <= value <= 5348:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5365 <= value <= 5396:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5438 <= value <= 5442:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5446 <= value <= 5449:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5451 <= value <= 5452:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5457 <= value <= 5460:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5466 <= value <= 5469:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5476 <= value <= 5499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5508 <= value <= 5539:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5541 <= value <= 5549:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5551 <= value <= 5552:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5558 <= value <= 5564:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5570 <= value <= 5572:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5576 <= value <= 5578:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5587 <= value <= 5596:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5606 <= value <= 5617:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5619 <= value <= 5626:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5640 <= value <= 5645:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5647 <= value <= 5665:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5667 <= value <= 5669:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5685 <= value <= 5686:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5690 <= value <= 5692:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5694 <= value <= 5695:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5697 <= value <= 5699:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5701 <= value <= 5704:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5706 <= value <= 5712:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5731 <= value <= 5740:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5751 <= value <= 5754:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5758 <= value <= 5765:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5772 <= value <= 5776:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5778 <= value <= 5779:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5788 <= value <= 5792:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5795 <= value <= 5797:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5799 <= value <= 5812:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5815 <= value <= 5819:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5821 <= value <= 5840:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5843 <= value <= 5858:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5860 <= value <= 5862:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5864 <= value <= 5867:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5869 <= value <= 5882:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5884 <= value <= 5899:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5901 <= value <= 5902:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5914 <= value <= 5962:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5964 <= value <= 5967:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5970 <= value <= 5983:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 5995 <= value <= 5998:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6000 <= value <= 6063:
            #: [TCP] X Window System
            return extend_enum(cls, 'x11_%d' % value, value, 'x11', TransportProtocol.get('tcp'))
        if 6000 <= value <= 6063:
            #: [UDP] X Window System
            return extend_enum(cls, 'x11_%d' % value, value, 'x11', TransportProtocol.get('udp'))
        if 6078 <= value <= 6079:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6089 <= value <= 6098:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6119 <= value <= 6120:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6125 <= value <= 6129:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6131 <= value <= 6132:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6134 <= value <= 6139:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6150 <= value <= 6158:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6164 <= value <= 6199:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6202 <= value <= 6208:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6210 <= value <= 6221:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6223 <= value <= 6240:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6245 <= value <= 6250:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6254 <= value <= 6266:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6270 <= value <= 6299:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6302 <= value <= 6305:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6307 <= value <= 6314:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6318 <= value <= 6319:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6327 <= value <= 6342:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6345 <= value <= 6345:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6348 <= value <= 6349:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6351 <= value <= 6354:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6356 <= value <= 6359:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6361 <= value <= 6362:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6364 <= value <= 6369:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6371 <= value <= 6378:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6380 <= value <= 6381:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6383 <= value <= 6388:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6391 <= value <= 6399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6411 <= value <= 6416:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6422 <= value <= 6431:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6433 <= value <= 6439:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6447 <= value <= 6454:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6457 <= value <= 6463:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6465 <= value <= 6470:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6472 <= value <= 6479:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6490 <= value <= 6499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6512 <= value <= 6512:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6516 <= value <= 6542:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6545 <= value <= 6546:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6552 <= value <= 6555:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6559 <= value <= 6565:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6569 <= value <= 6578:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6584 <= value <= 6587:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6589 <= value <= 6599:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6603 <= value <= 6618:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6637 <= value <= 6639:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6641 <= value <= 6652:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6658 <= value <= 6664:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6665 <= value <= 6669:
            #: [TCP] IRCU
            return extend_enum(cls, 'ircu_%d' % value, value, 'ircu', TransportProtocol.get('tcp'))
        if 6665 <= value <= 6669:
            #: [UDP] Reserved
            return extend_enum(cls, 'reserved_%d' % value, value, 'reserved', TransportProtocol.get('udp'))
        if 6674 <= value <= 6677:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6680 <= value <= 6686:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6691 <= value <= 6695:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6707 <= value <= 6713:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6717 <= value <= 6766:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6772 <= value <= 6776:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6779 <= value <= 6783:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6792 <= value <= 6800:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6802 <= value <= 6816:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6818 <= value <= 6830:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6832 <= value <= 6840:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6843 <= value <= 6849:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6851 <= value <= 6867:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6869 <= value <= 6887:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6889 <= value <= 6899:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6902 <= value <= 6923:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6925 <= value <= 6934:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6937 <= value <= 6945:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6947 <= value <= 6950:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6952 <= value <= 6960:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6967 <= value <= 6968:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6971 <= value <= 6979:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 6981 <= value <= 6996:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7027 <= value <= 7029:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7032 <= value <= 7039:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7041 <= value <= 7069:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7074 <= value <= 7079:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7081 <= value <= 7087:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7089 <= value <= 7094:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7096 <= value <= 7098:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7102 <= value <= 7106:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7108 <= value <= 7116:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7118 <= value <= 7120:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7124 <= value <= 7127:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7130 <= value <= 7160:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7175 <= value <= 7180:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7182 <= value <= 7199:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7203 <= value <= 7214:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7217 <= value <= 7226:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7230 <= value <= 7233:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7238 <= value <= 7243:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7245 <= value <= 7261:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7263 <= value <= 7271:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7284 <= value <= 7299:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7300 <= value <= 7359:
            #: [N/A] The Swiss Exchange
            return extend_enum(cls, 'swx_%d' % value, value, 'swx', TransportProtocol.get('undefined'))
        if 7360 <= value <= 7364:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7366 <= value <= 7390:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7398 <= value <= 7399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7403 <= value <= 7409:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7412 <= value <= 7419:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7422 <= value <= 7425:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7432 <= value <= 7436:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7438 <= value <= 7442:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7444 <= value <= 7470:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7475 <= value <= 7477:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7479 <= value <= 7490:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7492 <= value <= 7499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7502 <= value <= 7507:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7512 <= value <= 7541:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7552 <= value <= 7559:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7561 <= value <= 7562:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7564 <= value <= 7565:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7567 <= value <= 7568:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7571 <= value <= 7573:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7575 <= value <= 7587:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7589 <= value <= 7605:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7607 <= value <= 7623:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7634 <= value <= 7647:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7649 <= value <= 7662:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7664 <= value <= 7671:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7678 <= value <= 7679:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7681 <= value <= 7682:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7684 <= value <= 7686:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7691 <= value <= 7696:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7698 <= value <= 7699:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7702 <= value <= 7706:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7709 <= value <= 7719:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7721 <= value <= 7723:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7729 <= value <= 7733:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7735 <= value <= 7737:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7739 <= value <= 7740:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7745 <= value <= 7746:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7748 <= value <= 7776:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7782 <= value <= 7783:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7790 <= value <= 7793:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7795 <= value <= 7796:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7803 <= value <= 7809:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7811 <= value <= 7844:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7848 <= value <= 7868:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7873 <= value <= 7877:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7881 <= value <= 7886:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7888 <= value <= 7899:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7904 <= value <= 7912:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7914 <= value <= 7931:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7934 <= value <= 7961:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7963 <= value <= 7966:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7968 <= value <= 7978:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 7983 <= value <= 7997:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8011 <= value <= 8014:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8028 <= value <= 8031:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8035 <= value <= 8039:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8045 <= value <= 8050:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8061 <= value <= 8065:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8068 <= value <= 8069:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8071 <= value <= 8073:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8075 <= value <= 8076:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8078 <= value <= 8079:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8092 <= value <= 8096:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8098 <= value <= 8099:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8103 <= value <= 8110:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8112 <= value <= 8114:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8119 <= value <= 8120:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8123 <= value <= 8127:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8133 <= value <= 8139:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8141 <= value <= 8147:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8150 <= value <= 8152:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8154 <= value <= 8159:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8163 <= value <= 8180:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8185 <= value <= 8189:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8196 <= value <= 8198:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8203 <= value <= 8203:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8209 <= value <= 8210:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8212 <= value <= 8229:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8233 <= value <= 8242:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8244 <= value <= 8265:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8267 <= value <= 8269:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8271 <= value <= 8275:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8277 <= value <= 8279:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8283 <= value <= 8291:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8295 <= value <= 8299:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8302 <= value <= 8312:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8314 <= value <= 8319:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8323 <= value <= 8350:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8352 <= value <= 8375:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8381 <= value <= 8382:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8385 <= value <= 8399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8406 <= value <= 8414:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8418 <= value <= 8422:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8424 <= value <= 8431:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8434 <= value <= 8441:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8446 <= value <= 8449:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8451 <= value <= 8456:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8458 <= value <= 8469:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8475 <= value <= 8499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8504 <= value <= 8553:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8556 <= value <= 8566:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8568 <= value <= 8599:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8601 <= value <= 8608:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8616 <= value <= 8664:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8667 <= value <= 8674:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8676 <= value <= 8685:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8689 <= value <= 8698:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8700 <= value <= 8709:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8712 <= value <= 8731:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8734 <= value <= 8749:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8751 <= value <= 8762:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8771 <= value <= 8777:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8779 <= value <= 8785:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8788 <= value <= 8792:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8794 <= value <= 8799:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8801 <= value <= 8803:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8810 <= value <= 8872:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8874 <= value <= 8879:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8884 <= value <= 8887:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8895 <= value <= 8898:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8902 <= value <= 8907:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8914 <= value <= 8936:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8938 <= value <= 8952:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8955 <= value <= 8979:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8982 <= value <= 8988:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 8992 <= value <= 8996:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9003 <= value <= 9004:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9012 <= value <= 9019:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9027 <= value <= 9049:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9052 <= value <= 9059:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9061 <= value <= 9079:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9094 <= value <= 9099:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9108 <= value <= 9110:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9112 <= value <= 9118:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9120 <= value <= 9121:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9124 <= value <= 9130:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9132 <= value <= 9159:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9165 <= value <= 9190:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9192 <= value <= 9199:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9218 <= value <= 9221:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9223 <= value <= 9254:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9256 <= value <= 9276:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9288 <= value <= 9291:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9296 <= value <= 9299:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9301 <= value <= 9305:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9307 <= value <= 9309:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9313 <= value <= 9317:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9319 <= value <= 9320:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9322 <= value <= 9338:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9341 <= value <= 9342:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9347 <= value <= 9373:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9375 <= value <= 9379:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9381 <= value <= 9386:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9391 <= value <= 9395:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9398 <= value <= 9399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9403 <= value <= 9417:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9419 <= value <= 9442:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9446 <= value <= 9449:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9451 <= value <= 9499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9501 <= value <= 9521:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9523 <= value <= 9534:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9537 <= value <= 9554:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9556 <= value <= 9558:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9560 <= value <= 9591:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9601 <= value <= 9611:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9619 <= value <= 9627:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9633 <= value <= 9639:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9641 <= value <= 9665:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9669 <= value <= 9693:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9696 <= value <= 9699:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9701 <= value <= 9746:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9748 <= value <= 9749:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9751 <= value <= 9752:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9754 <= value <= 9761:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9763 <= value <= 9799:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9803 <= value <= 9874:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9879 <= value <= 9887:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9890 <= value <= 9897:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9904 <= value <= 9908:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9912 <= value <= 9924:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9926 <= value <= 9949:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9957 <= value <= 9965:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9967 <= value <= 9977:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9982 <= value <= 9986:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 9989 <= value <= 9989:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10011 <= value <= 10019:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10021 <= value <= 10022:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10024 <= value <= 10049:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10052 <= value <= 10054:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10056 <= value <= 10079:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10082 <= value <= 10099:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10105 <= value <= 10106:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10108 <= value <= 10109:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10118 <= value <= 10124:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10126 <= value <= 10127:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10130 <= value <= 10159:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10163 <= value <= 10199:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10202 <= value <= 10251:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10254 <= value <= 10259:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10262 <= value <= 10287:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10289 <= value <= 10320:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10322 <= value <= 10438:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10440 <= value <= 10442:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10444 <= value <= 10499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10501 <= value <= 10539:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10545 <= value <= 10547:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10549 <= value <= 10630:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10632 <= value <= 10799:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10801 <= value <= 10804:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10806 <= value <= 10808:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10811 <= value <= 10859:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10861 <= value <= 10879:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10881 <= value <= 10932:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10934 <= value <= 10989:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 10991 <= value <= 10999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11002 <= value <= 11094:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11096 <= value <= 11102:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11113 <= value <= 11160:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11166 <= value <= 11170:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11176 <= value <= 11200:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11203 <= value <= 11207:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11209 <= value <= 11210:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11212 <= value <= 11234:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11236 <= value <= 11318:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11322 <= value <= 11366:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11368 <= value <= 11370:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11372 <= value <= 11429:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11431 <= value <= 11488:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11490 <= value <= 11599:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11601 <= value <= 11622:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11624 <= value <= 11719:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11721 <= value <= 11722:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11724 <= value <= 11750:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11752 <= value <= 11795:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11797 <= value <= 11875:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11878 <= value <= 11966:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11968 <= value <= 11970:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 11972 <= value <= 11996:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12014 <= value <= 12108:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12110 <= value <= 12120:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12122 <= value <= 12167:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12169 <= value <= 12171:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12173 <= value <= 12299:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12303 <= value <= 12320:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12323 <= value <= 12344:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12346 <= value <= 12545:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12547 <= value <= 12752:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12754 <= value <= 12864:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 12866 <= value <= 13159:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13161 <= value <= 13215:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13219 <= value <= 13222:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13225 <= value <= 13399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13401 <= value <= 13719:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13725 <= value <= 13781:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13787 <= value <= 13817:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13824 <= value <= 13831:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13833 <= value <= 13893:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13895 <= value <= 13928:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 13931 <= value <= 13999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14003 <= value <= 14032:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14035 <= value <= 14140:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14146 <= value <= 14148:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14151 <= value <= 14153:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14155 <= value <= 14249:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14251 <= value <= 14413:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14415 <= value <= 14499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14501 <= value <= 14935:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 14938 <= value <= 14999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 15003 <= value <= 15117:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 15119 <= value <= 15344:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 15346 <= value <= 15362:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 15364 <= value <= 15554:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 15556 <= value <= 15659:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 15661 <= value <= 15739:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 15741 <= value <= 15997:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16004 <= value <= 16019:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16022 <= value <= 16160:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16163 <= value <= 16308:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16312 <= value <= 16359:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16362 <= value <= 16366:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16369 <= value <= 16383:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16386 <= value <= 16618:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16620 <= value <= 16664:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16667 <= value <= 16788:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16790 <= value <= 16899:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16901 <= value <= 16949:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16951 <= value <= 16990:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 16996 <= value <= 17006:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17008 <= value <= 17009:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17011 <= value <= 17183:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17186 <= value <= 17218:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17226 <= value <= 17233:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17236 <= value <= 17499:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17501 <= value <= 17554:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17556 <= value <= 17728:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17730 <= value <= 17753:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17757 <= value <= 17776:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 17778 <= value <= 17999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18001 <= value <= 18103:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18105 <= value <= 18135:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18137 <= value <= 18180:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18188 <= value <= 18240:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18244 <= value <= 18261:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18263 <= value <= 18462:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18464 <= value <= 18515:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18517 <= value <= 18633:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18636 <= value <= 18667:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18669 <= value <= 18768:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18770 <= value <= 18880:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18882 <= value <= 18887:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 18889 <= value <= 18999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19001 <= value <= 19006:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19008 <= value <= 19019:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19021 <= value <= 19190:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19192 <= value <= 19193:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19195 <= value <= 19219:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19221 <= value <= 19282:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19284 <= value <= 19314:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19316 <= value <= 19397:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19399 <= value <= 19409:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19413 <= value <= 19538:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19542 <= value <= 19787:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 19791 <= value <= 19997:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20006 <= value <= 20011:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20015 <= value <= 20033:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20035 <= value <= 20045:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20047 <= value <= 20047:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20050 <= value <= 20056:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20058 <= value <= 20166:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20168 <= value <= 20201:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20203 <= value <= 20221:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20223 <= value <= 20479:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20481 <= value <= 20669:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20671 <= value <= 20809:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 20811 <= value <= 20998:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 21001 <= value <= 21009:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 21011 <= value <= 21211:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 21214 <= value <= 21220:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 21222 <= value <= 21552:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 21555 <= value <= 21589:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 21591 <= value <= 21799:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 21802 <= value <= 21844:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 21850 <= value <= 21999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22006 <= value <= 22124:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22126 <= value <= 22127:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22129 <= value <= 22221:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22223 <= value <= 22272:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22274 <= value <= 22304:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22306 <= value <= 22332:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22336 <= value <= 22342:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22344 <= value <= 22346:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22348 <= value <= 22349:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22352 <= value <= 22536:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22538 <= value <= 22554:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22556 <= value <= 22762:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22764 <= value <= 22799:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22801 <= value <= 22950:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 22952 <= value <= 22999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 23006 <= value <= 23052:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 23054 <= value <= 23271:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 23273 <= value <= 23293:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 23295 <= value <= 23332:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 23334 <= value <= 23399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 23403 <= value <= 23455:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 23458 <= value <= 23545:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 23547 <= value <= 23999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24007 <= value <= 24241:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24243 <= value <= 24248:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24250 <= value <= 24320:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24324 <= value <= 24385:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24387 <= value <= 24464:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24466 <= value <= 24553:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24555 <= value <= 24576:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24578 <= value <= 24665:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24667 <= value <= 24675:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24681 <= value <= 24726:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24728 <= value <= 24753:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24755 <= value <= 24849:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24851 <= value <= 24921:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 24923 <= value <= 24999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 25010 <= value <= 25470:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 25472 <= value <= 25575:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 25577 <= value <= 25603:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 25605 <= value <= 25792:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 25794 <= value <= 25899:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 25904 <= value <= 25953:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 25956 <= value <= 25999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 26001 <= value <= 26132:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 26134 <= value <= 26207:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 26209 <= value <= 26256:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 26258 <= value <= 26259:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 26265 <= value <= 26485:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 26490 <= value <= 26999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 27000 <= value <= 27009:
            #: [N/A] FLEX LM (1-10)
            return extend_enum(cls, 'flex_lm_%d' % value, value, 'flex-lm', TransportProtocol.get('undefined'))
        if 27011 <= value <= 27016:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 27018 <= value <= 27344:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 27346 <= value <= 27441:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 27443 <= value <= 27503:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 27505 <= value <= 27781:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 27783 <= value <= 27875:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 27877 <= value <= 27998:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 28002 <= value <= 28009:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 28011 <= value <= 28079:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 28081 <= value <= 28118:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 28120 <= value <= 28199:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 28201 <= value <= 28239:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 28241 <= value <= 28588:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 28590 <= value <= 28999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 29001 <= value <= 29117:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 29119 <= value <= 29166:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 29170 <= value <= 29998:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 30005 <= value <= 30099:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 30101 <= value <= 30259:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 30261 <= value <= 30399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 30401 <= value <= 30831:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 30833 <= value <= 30998:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31000 <= value <= 31015:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31017 <= value <= 31019:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31021 <= value <= 31028:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31030 <= value <= 31336:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31338 <= value <= 31399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31401 <= value <= 31415:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31417 <= value <= 31456:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31458 <= value <= 31619:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31621 <= value <= 31684:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31686 <= value <= 31764:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31766 <= value <= 31947:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 31950 <= value <= 32033:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32035 <= value <= 32248:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32250 <= value <= 32399:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32401 <= value <= 32482:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32484 <= value <= 32634:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32637 <= value <= 32766:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32778 <= value <= 32800:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32802 <= value <= 32810:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32812 <= value <= 32895:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 32897 <= value <= 32999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 33001 <= value <= 33059:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 33061 <= value <= 33122:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 33124 <= value <= 33330:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 33335 <= value <= 33433:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 33436 <= value <= 33655:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 33657 <= value <= 33889:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 33891 <= value <= 34248:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 34250 <= value <= 34377:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 34380 <= value <= 34566:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 34568 <= value <= 34961:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 34965 <= value <= 34979:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 34981 <= value <= 34999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 35007 <= value <= 35099:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 35101 <= value <= 35353:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 35358 <= value <= 36000:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36002 <= value <= 36410:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36413 <= value <= 36421:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36425 <= value <= 36442:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36445 <= value <= 36461:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36463 <= value <= 36523:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36525 <= value <= 36601:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36603 <= value <= 36699:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36701 <= value <= 36864:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 36866 <= value <= 37471:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 37473 <= value <= 37474:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 37476 <= value <= 37482:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 37484 <= value <= 37600:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 37602 <= value <= 37653:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 37655 <= value <= 37999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38003 <= value <= 38200:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38204 <= value <= 38411:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38413 <= value <= 38421:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38423 <= value <= 38461:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38463 <= value <= 38471:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38473 <= value <= 38637:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38639 <= value <= 38799:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38801 <= value <= 38864:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 38866 <= value <= 39062:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 39064 <= value <= 39680:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 39682 <= value <= 39999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 40001 <= value <= 40022:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 40024 <= value <= 40403:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 40405 <= value <= 40840:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 40844 <= value <= 40852:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 40854 <= value <= 41110:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 41112 <= value <= 41120:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 41122 <= value <= 41229:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 41231 <= value <= 41793:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 41798 <= value <= 42507:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 42511 <= value <= 42998:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 43001 <= value <= 43187:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 43192 <= value <= 43209:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 43211 <= value <= 43437:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 43442 <= value <= 44122:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 44124 <= value <= 44320:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 44324 <= value <= 44443:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 44446 <= value <= 44543:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 44545 <= value <= 44552:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 44554 <= value <= 44599:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 44601 <= value <= 44817:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 44819 <= value <= 44899:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 44901 <= value <= 44999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 45003 <= value <= 45044:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 45046 <= value <= 45053:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 45055 <= value <= 45513:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 45515 <= value <= 45677:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 45679 <= value <= 45823:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 45826 <= value <= 45965:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 45967 <= value <= 46335:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 46337 <= value <= 46997:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 47002 <= value <= 47099:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 47101 <= value <= 47556:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 47558 <= value <= 47623:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 47625 <= value <= 47805:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 47810 <= value <= 47999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 48006 <= value <= 48047:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 48051 <= value <= 48127:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 48130 <= value <= 48555:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 48557 <= value <= 48618:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 48620 <= value <= 48652:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 48654 <= value <= 48999:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        if 49002 <= value <= 49149:
            #: [N/A] Unassigned
            return extend_enum(cls, 'unassigned_%d' % value, value, 'unassigned', TransportProtocol.get('undefined'))
        return super()._missing_(value)
